﻿/*
    This file is part of Flycast.

    Flycast is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    Flycast is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Flycast.  If not, see <https://www.gnu.org/licenses/>.
 */
#ifndef LIBRETRO_CORE_OPTIONS_INTL_H__
#define LIBRETRO_CORE_OPTIONS_INTL_H__

#if defined(_MSC_VER) && (_MSC_VER >= 1500 && _MSC_VER < 1900)
/* https://support.microsoft.com/en-us/kb/980263 */
#pragma execution_character_set("utf-8")
#pragma warning(disable:4566)
#endif

#include <libretro.h>

#include "libretro_core_option_defines.h"

/*
 ********************************
 * VERSION: 2.0
 ********************************
 *
 * - 2.0: Add support for core options v2 interface
 * - 1.3: Move translations to libretro_core_options_intl.h
 *        - libretro_core_options_intl.h includes BOM and utf-8
 *          fix for MSVC 2010-2013
 *        - Added HAVE_NO_LANGEXTRA flag to disable translations
 *          on platforms/compilers without BOM support
 * - 1.2: Use core options v1 interface when
 *        RETRO_ENVIRONMENT_GET_CORE_OPTIONS_VERSION is >= 1
 *        (previously required RETRO_ENVIRONMENT_GET_CORE_OPTIONS_VERSION == 1)
 * - 1.1: Support generation of core options v0 retro_core_option_value
 *        arrays containing options with a single value
 * - 1.0: First commit
*/

#ifdef __cplusplus
extern "C" {
#endif

/*
 ********************************
 * Core Option Definitions
 ********************************
*/
/* RETRO_LANGUAGE_AR */

#define CATEGORY_SYSTEM_LABEL_AR NULL
#define CATEGORY_SYSTEM_INFO_0_AR NULL
#define CATEGORY_VIDEO_LABEL_AR "فيديو"
#define CATEGORY_VIDEO_INFO_0_AR NULL
#define CATEGORY_PERFORMANCE_LABEL_AR NULL
#define CATEGORY_PERFORMANCE_INFO_0_AR NULL
#define CATEGORY_HACKS_LABEL_AR NULL
#define CATEGORY_HACKS_INFO_0_AR NULL
#define CATEGORY_INPUT_LABEL_AR "الإدخال"
#define CATEGORY_INPUT_INFO_0_AR NULL
#define CATEGORY_VMU_LABEL_AR NULL
#define CATEGORY_VMU_INFO_0_AR NULL
#define CORE_OPTION_NAME_REGION_LABEL_AR "المنطقة"
#define OPTION_VAL_JAPAN_AR NULL
#define OPTION_VAL_USA_AR NULL
#define OPTION_VAL_EUROPE_AR NULL
#define OPTION_VAL_DEFAULT_AR NULL
#define CORE_OPTION_NAME_LANGUAGE_LABEL_AR "اللّغة"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_AR NULL
#define OPTION_VAL_JAPANESE_AR "Japanese - 日本語"
#define OPTION_VAL_ENGLISH_AR NULL
#define OPTION_VAL_GERMAN_AR "German - Deutsch"
#define OPTION_VAL_FRENCH_AR "French - Français"
#define OPTION_VAL_SPANISH_AR "Spanish - Español"
#define OPTION_VAL_ITALIAN_AR "Italian - Italiano"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_AR NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_AR NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_AR NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_AR NULL
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_AR NULL
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_AR NULL
#define CORE_OPTION_NAME_FORCE_WINCE_LABEL_AR NULL
#define CORE_OPTION_NAME_FORCE_WINCE_INFO_0_AR NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_AR NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_AR NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_AR NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_AR NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_AR NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_AR NULL
#define CORE_OPTION_NAME_UPNP_LABEL_AR NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_AR NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_AR NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_AR NULL
#define OPTION_VAL_320X240_AR NULL
#define OPTION_VAL_640X480_AR NULL
#define OPTION_VAL_800X600_AR NULL
#define OPTION_VAL_960X720_AR NULL
#define OPTION_VAL_1024X768_AR NULL
#define OPTION_VAL_1280X960_AR NULL
#define OPTION_VAL_1440X1080_AR NULL
#define OPTION_VAL_1600X1200_AR NULL
#define OPTION_VAL_1920X1440_AR NULL
#define OPTION_VAL_2560X1920_AR NULL
#define OPTION_VAL_2880X2160_AR NULL
#define OPTION_VAL_3200X2400_AR NULL
#define OPTION_VAL_3840X2880_AR NULL
#define OPTION_VAL_4480X3360_AR NULL
#define OPTION_VAL_5120X3840_AR NULL
#define OPTION_VAL_5760X4320_AR NULL
#define OPTION_VAL_6400X4800_AR NULL
#define OPTION_VAL_7040X5280_AR NULL
#define OPTION_VAL_7680X5760_AR NULL
#define OPTION_VAL_8320X6240_AR NULL
#define OPTION_VAL_8960X6720_AR NULL
#define OPTION_VAL_9600X7200_AR NULL
#define OPTION_VAL_10240X7680_AR NULL
#define OPTION_VAL_10880X8160_AR NULL
#define OPTION_VAL_11520X8640_AR NULL
#define OPTION_VAL_12160X9120_AR NULL
#define OPTION_VAL_12800X9600_AR NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_AR NULL
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_AR NULL
#define OPTION_VAL_VGA_AR NULL
#define OPTION_VAL_TV_RGB_AR NULL
#define OPTION_VAL_TV_COMPOSITE_AR NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_AR NULL
#define OPTION_VAL_NTSC_AR NULL
#define OPTION_VAL_PAL_AR NULL
#define OPTION_VAL_PAL_N_AR NULL
#define OPTION_VAL_PAL_M_AR NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_AR "اتجاه الشاشة"
#define OPTION_VAL_HORIZONTAL_AR NULL
#define OPTION_VAL_VERTICAL_AR NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_AR NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_AR NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_AR NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_AR NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_AR NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_AR NULL
#define OPTION_VAL_512MB_AR NULL
#define OPTION_VAL_1GB_AR NULL
#define OPTION_VAL_2GB_AR NULL
#define OPTION_VAL_4GB_AR NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_AR NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_AR NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_AR NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_AR NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_AR NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_AR NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_AR NULL
#define CORE_OPTION_NAME_FOG_LABEL_AR NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_AR NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_AR NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_AR NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_AR NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_AR NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_AR NULL
#define OPTION_VAL_1_AR NULL
#define OPTION_VAL_2_AR NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_AR NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_AR NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_AR NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_AR NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_AR NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_AR NULL
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_AR NULL
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_AR NULL
#define OPTION_VAL_2_O27_AR NULL
#define OPTION_VAL_4_AR NULL
#define OPTION_VAL_6_AR NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_AR NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_AR NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_AR NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_AR NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_AR NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_AR NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_AR NULL
#define OPTION_VAL_SOME_AR "عادي"
#define OPTION_VAL_MORE_AR NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_AR NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_AR NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_AR NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_AR NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_AR NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_AR NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_AR NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_AR NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_AR NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_AR NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_AR NULL
#define OPTION_VAL_0_AR NULL
#define OPTION_VAL_5_AR NULL
#define OPTION_VAL_10_AR NULL
#define OPTION_VAL_15_AR NULL
#define OPTION_VAL_20_AR NULL
#define OPTION_VAL_25_AR NULL
#define OPTION_VAL_30_AR NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_AR NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_AR NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_AR NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_AR NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_AR NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_AR NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_AR NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_AR NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_AR NULL
#define OPTION_VAL_WHITE_AR NULL
#define OPTION_VAL_RED_AR "أحمر"
#define OPTION_VAL_GREEN_AR "أخضر"
#define OPTION_VAL_BLUE_AR "أزرق"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_AR NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_AR NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_AR NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_AR NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_AR NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_AR NULL
#define OPTION_VAL_VMU_A1_AR NULL
#define OPTION_VAL_ALL_VMUS_AR NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_AR NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_AR NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_AR NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_AR NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_AR NULL
#define OPTION_VAL_UPPER_LEFT_AR NULL
#define OPTION_VAL_UPPER_RIGHT_AR NULL
#define OPTION_VAL_LOWER_LEFT_AR NULL
#define OPTION_VAL_LOWER_RIGHT_AR NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_AR NULL
#define OPTION_VAL_1X_AR NULL
#define OPTION_VAL_3X_AR NULL
#define OPTION_VAL_5X_AR NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_AR NULL
#define OPTION_VAL_DEFAULT_ON_00_AR NULL
#define OPTION_VAL_DEFAULT_OFF_01_AR NULL
#define OPTION_VAL_BLACK_02_AR NULL
#define OPTION_VAL_LIGHT_BLUE_04_AR NULL
#define OPTION_VAL_CYAN_06_AR NULL
#define OPTION_VAL_CYAN_BLUE_07_AR NULL
#define OPTION_VAL_LIGHT_GREEN_08_AR NULL
#define OPTION_VAL_CYAN_GREEN_09_AR NULL
#define OPTION_VAL_LIGHT_CYAN_10_AR NULL
#define OPTION_VAL_PURPLE_12_AR NULL
#define OPTION_VAL_LIGHT_PURPLE_13_AR NULL
#define OPTION_VAL_YELLOW_14_AR "أصفر"
#define OPTION_VAL_GRAY_15_AR NULL
#define OPTION_VAL_LIGHT_PURPLE_2_16_AR NULL
#define OPTION_VAL_LIGHT_GREEN_2_17_AR NULL
#define OPTION_VAL_LIGHT_GREEN_3_18_AR NULL
#define OPTION_VAL_LIGHT_CYAN_2_19_AR NULL
#define OPTION_VAL_LIGHT_RED_2_20_AR NULL
#define OPTION_VAL_MAGENTA_21_AR NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_AR NULL
#define OPTION_VAL_LIGHT_ORANGE_23_AR NULL
#define OPTION_VAL_ORANGE_24_AR NULL
#define OPTION_VAL_LIGHT_PURPLE_4_25_AR NULL
#define OPTION_VAL_LIGHT_YELLOW_26_AR NULL
#define OPTION_VAL_LIGHT_YELLOW_2_27_AR NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_AR NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_AR NULL
#define OPTION_VAL_40_AR NULL
#define OPTION_VAL_50_AR NULL
#define OPTION_VAL_60_AR NULL
#define OPTION_VAL_70_AR NULL
#define OPTION_VAL_80_AR NULL
#define OPTION_VAL_90_AR NULL
#define OPTION_VAL_100_AR NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_AR NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_AR NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_AR NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_AR NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_AR NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_AR NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_AR NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_AR NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_AR NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_AR NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_AR NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_AR NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_AR NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_AR NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_AR NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_AR NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_AR NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_AR NULL

struct retro_core_option_v2_category option_cats_ar[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_AR,
      CATEGORY_SYSTEM_INFO_0_AR
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_AR,
      CATEGORY_VIDEO_INFO_0_AR
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_AR,
      CATEGORY_PERFORMANCE_INFO_0_AR
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_AR,
      CATEGORY_HACKS_INFO_0_AR
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_AR,
      CATEGORY_INPUT_INFO_0_AR
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_AR,
      CATEGORY_VMU_INFO_0_AR
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_ar[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_AR,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_AR },
         { "USA",     OPTION_VAL_USA_AR },
         { "Europe",  OPTION_VAL_EUROPE_AR },
         { "Default", OPTION_VAL_DEFAULT_AR },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_AR,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_AR },
         { "English",  OPTION_VAL_ENGLISH_AR },
         { "German",   OPTION_VAL_GERMAN_AR },
         { "French",   OPTION_VAL_FRENCH_AR },
         { "Spanish",  OPTION_VAL_SPANISH_AR },
         { "Italian",  OPTION_VAL_ITALIAN_AR },
         { "Default",  OPTION_VAL_DEFAULT_AR },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_AR,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_boot_to_bios",
      CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_AR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_AR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_force_wince",
      CORE_OPTION_NAME_FORCE_WINCE_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_FORCE_WINCE_INFO_0_AR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_AR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_AR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_AR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_AR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_AR,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_AR },
         { "640x480",    OPTION_VAL_640X480_AR },
         { "800x600",    OPTION_VAL_800X600_AR },
         { "960x720",    OPTION_VAL_960X720_AR },
         { "1024x768",   OPTION_VAL_1024X768_AR },
         { "1280x960",   OPTION_VAL_1280X960_AR },
         { "1440x1080",  OPTION_VAL_1440X1080_AR },
         { "1600x1200",  OPTION_VAL_1600X1200_AR },
         { "1920x1440",  OPTION_VAL_1920X1440_AR },
         { "2560x1920",  OPTION_VAL_2560X1920_AR },
         { "2880x2160",  OPTION_VAL_2880X2160_AR },
         { "3200x2400",  OPTION_VAL_3200X2400_AR },
         { "3840x2880",  OPTION_VAL_3840X2880_AR },
         { "4480x3360",  OPTION_VAL_4480X3360_AR },
         { "5120x3840",  OPTION_VAL_5120X3840_AR },
         { "5760x4320",  OPTION_VAL_5760X4320_AR },
         { "6400x4800",  OPTION_VAL_6400X4800_AR },
         { "7040x5280",  OPTION_VAL_7040X5280_AR },
         { "7680x5760",  OPTION_VAL_7680X5760_AR },
         { "8320x6240",  OPTION_VAL_8320X6240_AR },
         { "8960x6720",  OPTION_VAL_8960X6720_AR },
         { "9600x7200",  OPTION_VAL_9600X7200_AR },
         { "10240x7680", OPTION_VAL_10240X7680_AR },
         { "10880x8160", OPTION_VAL_10880X8160_AR },
         { "11520x8640", OPTION_VAL_11520X8640_AR },
         { "12160x9120", OPTION_VAL_12160X9120_AR },
         { "12800x9600", OPTION_VAL_12800X9600_AR },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_AR,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_AR },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_AR },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_AR },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_AR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_AR },
         { "PAL",     OPTION_VAL_PAL_AR },
         { "PAL_N",   OPTION_VAL_PAL_N_AR },
         { "PAL_M",   OPTION_VAL_PAL_M_AR },
         { "Default", OPTION_VAL_DEFAULT_AR },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_AR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_AR },
         { "vertical",   OPTION_VAL_VERTICAL_AR },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_AR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_AR },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_AR },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_AR },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_AR,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_AR },
         { "1GB",   OPTION_VAL_1GB_AR },
         { "2GB",   OPTION_VAL_2GB_AR },
         { "4GB",   OPTION_VAL_4GB_AR },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_AR,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_AR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_AR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_AR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_AR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_AR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_AR,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_AR,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_AR },
         { "1",  OPTION_VAL_1_AR },
         { "2",  OPTION_VAL_2_AR },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_AR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_AR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_AR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_AR,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O27_AR },
         { "4",  OPTION_VAL_4_AR },
         { "6",  OPTION_VAL_6_AR },
         { NULL, NULL },
      },
      "1",
   },
   {/* TODO: needs clarification */
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_AR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_AR,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_AR,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_AR,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_AR,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_AR },
         { "more", OPTION_VAL_MORE_AR },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_AR,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_AR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_AR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_AR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_AR,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: probably needs explanation */
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_AR,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_AR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_AR },
         { "5%",  OPTION_VAL_5_AR },
         { "10%", OPTION_VAL_10_AR },
         { "15%", OPTION_VAL_15_AR },
         { "20%", OPTION_VAL_20_AR },
         { "25%", OPTION_VAL_25_AR },
         { "30%", OPTION_VAL_30_AR },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_AR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_AR },
         { "5%",  OPTION_VAL_5_AR },
         { "10%", OPTION_VAL_10_AR },
         { "15%", OPTION_VAL_15_AR },
         { "20%", OPTION_VAL_20_AR },
         { "25%", OPTION_VAL_25_AR },
         { "30%", OPTION_VAL_30_AR },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_AR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_purupuru",
      CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_AR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_AR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_AR,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_AR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_AR },
         { "Red",      OPTION_VAL_RED_AR },
         { "Green",    OPTION_VAL_GREEN_AR },
         { "Blue",     OPTION_VAL_BLUE_AR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_AR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_AR },
         { "Red",      OPTION_VAL_RED_AR },
         { "Green",    OPTION_VAL_GREEN_AR },
         { "Blue",     OPTION_VAL_BLUE_AR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_AR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_AR },
         { "Red",      OPTION_VAL_RED_AR },
         { "Green",    OPTION_VAL_GREEN_AR },
         { "Blue",     OPTION_VAL_BLUE_AR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_AR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_AR },
         { "Red",      OPTION_VAL_RED_AR },
         { "Green",    OPTION_VAL_GREEN_AR },
         { "Blue",     OPTION_VAL_BLUE_AR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_AR,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_AR,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_AR,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_AR },
         { "All VMUs", OPTION_VAL_ALL_VMUS_AR },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_AR,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_AR,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_AR,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_AR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_AR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_AR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_AR },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_AR },
         { "2x", OPTION_VAL_2_O27_AR },
         { "3x", OPTION_VAL_3X_AR },
         { "4x", OPTION_VAL_4_AR },
         { "5x", OPTION_VAL_5X_AR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_AR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_AR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_AR },
         { "BLUE 03",           OPTION_VAL_BLUE_AR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_AR },
         { "GREEN 05",          OPTION_VAL_GREEN_AR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_AR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_AR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_AR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_AR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_AR },
         { "RED 11",            OPTION_VAL_RED_AR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_AR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_AR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_AR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_AR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_AR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_AR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_AR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_AR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_AR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_AR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_AR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_AR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_AR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_AR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_AR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_AR },
         { "WHITE 28",          OPTION_VAL_WHITE_AR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_AR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_AR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_AR },
         { "BLUE 03",           OPTION_VAL_BLUE_AR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_AR },
         { "GREEN 05",          OPTION_VAL_GREEN_AR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_AR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_AR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_AR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_AR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_AR },
         { "RED 11",            OPTION_VAL_RED_AR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_AR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_AR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_AR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_AR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_AR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_AR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_AR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_AR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_AR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_AR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_AR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_AR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_AR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_AR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_AR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_AR },
         { "WHITE 28",          OPTION_VAL_WHITE_AR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_AR },
         { "20%",  OPTION_VAL_20_AR },
         { "30%",  OPTION_VAL_30_AR },
         { "40%",  OPTION_VAL_40_AR },
         { "50%",  OPTION_VAL_50_AR },
         { "60%",  OPTION_VAL_60_AR },
         { "70%",  OPTION_VAL_70_AR },
         { "80%",  OPTION_VAL_80_AR },
         { "90%",  OPTION_VAL_90_AR },
         { "100%", OPTION_VAL_100_AR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_AR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_AR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_AR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_AR },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_AR },
         { "2x", OPTION_VAL_2_O27_AR },
         { "3x", OPTION_VAL_3X_AR },
         { "4x", OPTION_VAL_4_AR },
         { "5x", OPTION_VAL_5X_AR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_AR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_AR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_AR },
         { "BLUE 03",           OPTION_VAL_BLUE_AR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_AR },
         { "GREEN 05",          OPTION_VAL_GREEN_AR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_AR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_AR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_AR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_AR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_AR },
         { "RED 11",            OPTION_VAL_RED_AR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_AR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_AR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_AR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_AR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_AR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_AR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_AR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_AR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_AR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_AR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_AR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_AR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_AR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_AR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_AR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_AR },
         { "WHITE 28",          OPTION_VAL_WHITE_AR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_AR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_AR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_AR },
         { "BLUE 03",           OPTION_VAL_BLUE_AR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_AR },
         { "GREEN 05",          OPTION_VAL_GREEN_AR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_AR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_AR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_AR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_AR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_AR },
         { "RED 11",            OPTION_VAL_RED_AR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_AR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_AR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_AR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_AR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_AR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_AR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_AR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_AR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_AR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_AR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_AR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_AR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_AR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_AR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_AR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_AR },
         { "WHITE 28",          OPTION_VAL_WHITE_AR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_AR },
         { "20%",  OPTION_VAL_20_AR },
         { "30%",  OPTION_VAL_30_AR },
         { "40%",  OPTION_VAL_40_AR },
         { "50%",  OPTION_VAL_50_AR },
         { "60%",  OPTION_VAL_60_AR },
         { "70%",  OPTION_VAL_70_AR },
         { "80%",  OPTION_VAL_80_AR },
         { "90%",  OPTION_VAL_90_AR },
         { "100%", OPTION_VAL_100_AR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_AR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_AR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_AR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_AR },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_AR },
         { "2x", OPTION_VAL_2_O27_AR },
         { "3x", OPTION_VAL_3X_AR },
         { "4x", OPTION_VAL_4_AR },
         { "5x", OPTION_VAL_5X_AR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_AR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_AR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_AR },
         { "BLUE 03",           OPTION_VAL_BLUE_AR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_AR },
         { "GREEN 05",          OPTION_VAL_GREEN_AR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_AR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_AR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_AR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_AR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_AR },
         { "RED 11",            OPTION_VAL_RED_AR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_AR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_AR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_AR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_AR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_AR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_AR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_AR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_AR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_AR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_AR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_AR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_AR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_AR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_AR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_AR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_AR },
         { "WHITE 28",          OPTION_VAL_WHITE_AR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_AR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_AR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_AR },
         { "BLUE 03",           OPTION_VAL_BLUE_AR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_AR },
         { "GREEN 05",          OPTION_VAL_GREEN_AR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_AR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_AR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_AR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_AR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_AR },
         { "RED 11",            OPTION_VAL_RED_AR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_AR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_AR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_AR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_AR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_AR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_AR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_AR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_AR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_AR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_AR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_AR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_AR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_AR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_AR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_AR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_AR },
         { "WHITE 28",          OPTION_VAL_WHITE_AR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_AR },
         { "20%",  OPTION_VAL_20_AR },
         { "30%",  OPTION_VAL_30_AR },
         { "40%",  OPTION_VAL_40_AR },
         { "50%",  OPTION_VAL_50_AR },
         { "60%",  OPTION_VAL_60_AR },
         { "70%",  OPTION_VAL_70_AR },
         { "80%",  OPTION_VAL_80_AR },
         { "90%",  OPTION_VAL_90_AR },
         { "100%", OPTION_VAL_100_AR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_AR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_AR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_AR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_AR },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_AR },
         { "2x", OPTION_VAL_2_O27_AR },
         { "3x", OPTION_VAL_3X_AR },
         { "4x", OPTION_VAL_4_AR },
         { "5x", OPTION_VAL_5X_AR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_AR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_AR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_AR },
         { "BLUE 03",           OPTION_VAL_BLUE_AR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_AR },
         { "GREEN 05",          OPTION_VAL_GREEN_AR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_AR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_AR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_AR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_AR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_AR },
         { "RED 11",            OPTION_VAL_RED_AR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_AR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_AR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_AR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_AR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_AR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_AR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_AR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_AR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_AR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_AR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_AR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_AR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_AR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_AR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_AR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_AR },
         { "WHITE 28",          OPTION_VAL_WHITE_AR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_AR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_AR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_AR },
         { "BLUE 03",           OPTION_VAL_BLUE_AR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_AR },
         { "GREEN 05",          OPTION_VAL_GREEN_AR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_AR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_AR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_AR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_AR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_AR },
         { "RED 11",            OPTION_VAL_RED_AR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_AR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_AR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_AR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_AR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_AR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_AR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_AR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_AR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_AR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_AR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_AR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_AR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_AR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_AR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_AR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_AR },
         { "WHITE 28",          OPTION_VAL_WHITE_AR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_AR },
         { "20%",  OPTION_VAL_20_AR },
         { "30%",  OPTION_VAL_30_AR },
         { "40%",  OPTION_VAL_40_AR },
         { "50%",  OPTION_VAL_50_AR },
         { "60%",  OPTION_VAL_60_AR },
         { "70%",  OPTION_VAL_70_AR },
         { "80%",  OPTION_VAL_80_AR },
         { "90%",  OPTION_VAL_90_AR },
         { "100%", OPTION_VAL_100_AR },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_ar = {
   option_cats_ar,
   option_defs_ar
};

/* RETRO_LANGUAGE_AST */

#define CATEGORY_SYSTEM_LABEL_AST "Sistema"
#define CATEGORY_SYSTEM_INFO_0_AST NULL
#define CATEGORY_VIDEO_LABEL_AST "Videu"
#define CATEGORY_VIDEO_INFO_0_AST NULL
#define CATEGORY_PERFORMANCE_LABEL_AST "Rindimientu"
#define CATEGORY_PERFORMANCE_INFO_0_AST NULL
#define CATEGORY_HACKS_LABEL_AST "Hacks de la emulación"
#define CATEGORY_HACKS_INFO_0_AST NULL
#define CATEGORY_INPUT_LABEL_AST "Entrada"
#define CATEGORY_INPUT_INFO_0_AST NULL
#define CATEGORY_VMU_LABEL_AST NULL
#define CATEGORY_VMU_INFO_0_AST NULL
#define CORE_OPTION_NAME_REGION_LABEL_AST "Rexón"
#define OPTION_VAL_JAPAN_AST "Xapón"
#define OPTION_VAL_USA_AST "Estaos Xuníos"
#define OPTION_VAL_EUROPE_AST "Europa"
#define OPTION_VAL_DEFAULT_AST NULL
#define CORE_OPTION_NAME_LANGUAGE_LABEL_AST "Llingua"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_AST NULL
#define OPTION_VAL_JAPANESE_AST "Xaponés"
#define OPTION_VAL_ENGLISH_AST "Inglés"
#define OPTION_VAL_GERMAN_AST "Alemán"
#define OPTION_VAL_FRENCH_AST "Francés"
#define OPTION_VAL_SPANISH_AST "Castellán"
#define OPTION_VAL_ITALIAN_AST "Italianu"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_AST NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_AST NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_AST NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_AST NULL
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_AST "Activar el DSP"
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_AST "Activa la emulación del procesador de señal dixital (DSP) d'audiu de Dreamcast. Ameyora la precisión del soníu xeneráu mas aumenta los requirimientos."
#define CORE_OPTION_NAME_FORCE_WINCE_LABEL_AST "Forciar el mou de Windows CE"
#define CORE_OPTION_NAME_FORCE_WINCE_INFO_0_AST "Activa la emulación completa de la unidá de xestión de memoria (MMU) y otros axustes pa los xuegos de Windows CE."
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_AST NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_AST NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_AST NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_AST NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_AST NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_AST NULL
#define CORE_OPTION_NAME_UPNP_LABEL_AST NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_AST NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_AST "Resolución interna"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_AST "Modifica la resolución del renderizáu."
#define OPTION_VAL_320X240_AST NULL
#define OPTION_VAL_640X480_AST NULL
#define OPTION_VAL_800X600_AST NULL
#define OPTION_VAL_960X720_AST NULL
#define OPTION_VAL_1024X768_AST NULL
#define OPTION_VAL_1280X960_AST NULL
#define OPTION_VAL_1440X1080_AST NULL
#define OPTION_VAL_1600X1200_AST NULL
#define OPTION_VAL_1920X1440_AST NULL
#define OPTION_VAL_2560X1920_AST NULL
#define OPTION_VAL_2880X2160_AST NULL
#define OPTION_VAL_3200X2400_AST NULL
#define OPTION_VAL_3840X2880_AST NULL
#define OPTION_VAL_4480X3360_AST NULL
#define OPTION_VAL_5120X3840_AST NULL
#define OPTION_VAL_5760X4320_AST NULL
#define OPTION_VAL_6400X4800_AST NULL
#define OPTION_VAL_7040X5280_AST NULL
#define OPTION_VAL_7680X5760_AST NULL
#define OPTION_VAL_8320X6240_AST NULL
#define OPTION_VAL_8960X6720_AST NULL
#define OPTION_VAL_9600X7200_AST NULL
#define OPTION_VAL_10240X7680_AST NULL
#define OPTION_VAL_10880X8160_AST NULL
#define OPTION_VAL_11520X8640_AST NULL
#define OPTION_VAL_12160X9120_AST NULL
#define OPTION_VAL_12800X9600_AST NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_AST "Tipu de cable"
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_AST NULL
#define OPTION_VAL_VGA_AST NULL
#define OPTION_VAL_TV_RGB_AST NULL
#define OPTION_VAL_TV_COMPOSITE_AST "TV (Compuestu)"
#define CORE_OPTION_NAME_BROADCAST_LABEL_AST NULL
#define OPTION_VAL_NTSC_AST NULL
#define OPTION_VAL_PAL_AST NULL
#define OPTION_VAL_PAL_N_AST "PAL-N (Arxentina, Paraguái, Uruguái)"
#define OPTION_VAL_PAL_M_AST "PAL-M (Brasil)"
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_AST "Orientación de la pantalla"
#define OPTION_VAL_HORIZONTAL_AST NULL
#define OPTION_VAL_VERTICAL_AST NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_AST NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_AST NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_AST NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_AST NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_AST NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_AST NULL
#define OPTION_VAL_512MB_AST NULL
#define OPTION_VAL_1GB_AST NULL
#define OPTION_VAL_2GB_AST NULL
#define OPTION_VAL_4GB_AST NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_AST NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_AST NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_AST NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_AST NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_AST NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_AST NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_AST NULL
#define CORE_OPTION_NAME_FOG_LABEL_AST "Efeutos de nublina"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_AST "Modificador del volume"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_AST NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_AST "Filtru anisotrópicu"
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_AST NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_AST NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_AST NULL
#define OPTION_VAL_1_AST NULL
#define OPTION_VAL_2_AST NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_AST NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_AST NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_AST NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_AST NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_AST "Filtru del posprocesamientu de PowerVR2"
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_AST NULL
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_AST NULL
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_AST NULL
#define OPTION_VAL_2_O27_AST "x2"
#define OPTION_VAL_4_AST "x4"
#define OPTION_VAL_6_AST "x6"
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_AST NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_AST NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_AST NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_AST NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_AST NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_AST NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_AST NULL
#define OPTION_VAL_SOME_AST NULL
#define OPTION_VAL_MORE_AST NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_AST NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_AST NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_AST NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_AST NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_AST NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_AST NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_AST NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_AST "Aumenta la velocidá de carga de los GD-ROMs."
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_AST NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_AST NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_AST NULL
#define OPTION_VAL_0_AST NULL
#define OPTION_VAL_5_AST NULL
#define OPTION_VAL_10_AST NULL
#define OPTION_VAL_15_AST NULL
#define OPTION_VAL_20_AST NULL
#define OPTION_VAL_25_AST NULL
#define OPTION_VAL_30_AST NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_AST NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_AST NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_AST NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_AST NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_AST NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_AST NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_AST NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_AST NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_AST NULL
#define OPTION_VAL_WHITE_AST "Blancu"
#define OPTION_VAL_RED_AST "Coloráu"
#define OPTION_VAL_GREEN_AST "Verde"
#define OPTION_VAL_BLUE_AST "Azul"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_AST NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_AST NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_AST NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_AST NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_AST NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_AST NULL
#define OPTION_VAL_VMU_A1_AST NULL
#define OPTION_VAL_ALL_VMUS_AST "Toles VMUs"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_AST NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_AST NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_AST NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_AST NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_AST NULL
#define OPTION_VAL_UPPER_LEFT_AST NULL
#define OPTION_VAL_UPPER_RIGHT_AST NULL
#define OPTION_VAL_LOWER_LEFT_AST NULL
#define OPTION_VAL_LOWER_RIGHT_AST NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_AST NULL
#define OPTION_VAL_1X_AST "x1"
#define OPTION_VAL_3X_AST "x3"
#define OPTION_VAL_5X_AST "x5"
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_AST NULL
#define OPTION_VAL_DEFAULT_ON_00_AST NULL
#define OPTION_VAL_DEFAULT_OFF_01_AST NULL
#define OPTION_VAL_BLACK_02_AST "Prietu"
#define OPTION_VAL_LIGHT_BLUE_04_AST "Azul claro"
#define OPTION_VAL_CYAN_06_AST "Cianu"
#define OPTION_VAL_CYAN_BLUE_07_AST "Azul ciano"
#define OPTION_VAL_LIGHT_GREEN_08_AST "Verde claro"
#define OPTION_VAL_CYAN_GREEN_09_AST "Verde ciano"
#define OPTION_VAL_LIGHT_CYAN_10_AST "Cianu claro"
#define OPTION_VAL_PURPLE_12_AST "Moráu"
#define OPTION_VAL_LIGHT_PURPLE_13_AST "Moráu claro"
#define OPTION_VAL_YELLOW_14_AST "Mariellu"
#define OPTION_VAL_GRAY_15_AST "Buxu"
#define OPTION_VAL_LIGHT_PURPLE_2_16_AST "Moráu claro (2)"
#define OPTION_VAL_LIGHT_GREEN_2_17_AST "Verde claro (2)"
#define OPTION_VAL_LIGHT_GREEN_3_18_AST "Verde claro (3)"
#define OPTION_VAL_LIGHT_CYAN_2_19_AST "Cianu claro (2)"
#define OPTION_VAL_LIGHT_RED_2_20_AST "Coloráu claro (2)"
#define OPTION_VAL_MAGENTA_21_AST "Maxenta"
#define OPTION_VAL_LIGHT_PURPLE_3_22_AST "Moráu claro (3)"
#define OPTION_VAL_LIGHT_ORANGE_23_AST "Naranxa claro"
#define OPTION_VAL_ORANGE_24_AST "Naranxa"
#define OPTION_VAL_LIGHT_PURPLE_4_25_AST "Moráu claro (4)"
#define OPTION_VAL_LIGHT_YELLOW_26_AST "Mariellu claro"
#define OPTION_VAL_LIGHT_YELLOW_2_27_AST "Mariellu claro (2)"
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_AST NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_AST NULL
#define OPTION_VAL_40_AST NULL
#define OPTION_VAL_50_AST NULL
#define OPTION_VAL_60_AST NULL
#define OPTION_VAL_70_AST NULL
#define OPTION_VAL_80_AST NULL
#define OPTION_VAL_90_AST NULL
#define OPTION_VAL_100_AST NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_AST NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_AST NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_AST NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_AST NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_AST NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_AST NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_AST NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_AST NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_AST NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_AST NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_AST NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_AST NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_AST NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_AST NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_AST NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_AST NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_AST NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_AST NULL

struct retro_core_option_v2_category option_cats_ast[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_AST,
      CATEGORY_SYSTEM_INFO_0_AST
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_AST,
      CATEGORY_VIDEO_INFO_0_AST
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_AST,
      CATEGORY_PERFORMANCE_INFO_0_AST
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_AST,
      CATEGORY_HACKS_INFO_0_AST
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_AST,
      CATEGORY_INPUT_INFO_0_AST
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_AST,
      CATEGORY_VMU_INFO_0_AST
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_ast[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_AST,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_AST },
         { "USA",     OPTION_VAL_USA_AST },
         { "Europe",  OPTION_VAL_EUROPE_AST },
         { "Default", OPTION_VAL_DEFAULT_AST },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_AST,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_AST },
         { "English",  OPTION_VAL_ENGLISH_AST },
         { "German",   OPTION_VAL_GERMAN_AST },
         { "French",   OPTION_VAL_FRENCH_AST },
         { "Spanish",  OPTION_VAL_SPANISH_AST },
         { "Italian",  OPTION_VAL_ITALIAN_AST },
         { "Default",  OPTION_VAL_DEFAULT_AST },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_AST,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_boot_to_bios",
      CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_AST,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_AST,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_force_wince",
      CORE_OPTION_NAME_FORCE_WINCE_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_FORCE_WINCE_INFO_0_AST,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_AST,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_AST,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_AST,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_AST,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_AST,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_AST },
         { "640x480",    OPTION_VAL_640X480_AST },
         { "800x600",    OPTION_VAL_800X600_AST },
         { "960x720",    OPTION_VAL_960X720_AST },
         { "1024x768",   OPTION_VAL_1024X768_AST },
         { "1280x960",   OPTION_VAL_1280X960_AST },
         { "1440x1080",  OPTION_VAL_1440X1080_AST },
         { "1600x1200",  OPTION_VAL_1600X1200_AST },
         { "1920x1440",  OPTION_VAL_1920X1440_AST },
         { "2560x1920",  OPTION_VAL_2560X1920_AST },
         { "2880x2160",  OPTION_VAL_2880X2160_AST },
         { "3200x2400",  OPTION_VAL_3200X2400_AST },
         { "3840x2880",  OPTION_VAL_3840X2880_AST },
         { "4480x3360",  OPTION_VAL_4480X3360_AST },
         { "5120x3840",  OPTION_VAL_5120X3840_AST },
         { "5760x4320",  OPTION_VAL_5760X4320_AST },
         { "6400x4800",  OPTION_VAL_6400X4800_AST },
         { "7040x5280",  OPTION_VAL_7040X5280_AST },
         { "7680x5760",  OPTION_VAL_7680X5760_AST },
         { "8320x6240",  OPTION_VAL_8320X6240_AST },
         { "8960x6720",  OPTION_VAL_8960X6720_AST },
         { "9600x7200",  OPTION_VAL_9600X7200_AST },
         { "10240x7680", OPTION_VAL_10240X7680_AST },
         { "10880x8160", OPTION_VAL_10880X8160_AST },
         { "11520x8640", OPTION_VAL_11520X8640_AST },
         { "12160x9120", OPTION_VAL_12160X9120_AST },
         { "12800x9600", OPTION_VAL_12800X9600_AST },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_AST,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_AST },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_AST },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_AST },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_AST,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_AST },
         { "PAL",     OPTION_VAL_PAL_AST },
         { "PAL_N",   OPTION_VAL_PAL_N_AST },
         { "PAL_M",   OPTION_VAL_PAL_M_AST },
         { "Default", OPTION_VAL_DEFAULT_AST },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_AST,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_AST },
         { "vertical",   OPTION_VAL_VERTICAL_AST },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_AST,
      NULL,
      "",
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_AST },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_AST },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_AST },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_AST,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_AST },
         { "1GB",   OPTION_VAL_1GB_AST },
         { "2GB",   OPTION_VAL_2GB_AST },
         { "4GB",   OPTION_VAL_4GB_AST },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_AST,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_AST,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_AST,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_AST,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_AST,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_AST,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_AST,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_AST,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_AST },
         { "1",  OPTION_VAL_1_AST },
         { "2",  OPTION_VAL_2_AST },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_AST,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_AST,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_AST,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_AST,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O27_AST },
         { "4",  OPTION_VAL_4_AST },
         { "6",  OPTION_VAL_6_AST },
         { NULL, NULL },
      },
      "1",
   },
   {/* TODO: needs clarification */
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_AST,
      NULL,
      "",
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_AST,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_AST,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_AST,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_AST,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_AST },
         { "more", OPTION_VAL_MORE_AST },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_AST,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_AST,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_AST,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_AST,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_AST,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: probably needs explanation */
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_AST,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_AST,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_AST },
         { "5%",  OPTION_VAL_5_AST },
         { "10%", OPTION_VAL_10_AST },
         { "15%", OPTION_VAL_15_AST },
         { "20%", OPTION_VAL_20_AST },
         { "25%", OPTION_VAL_25_AST },
         { "30%", OPTION_VAL_30_AST },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_AST,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_AST },
         { "5%",  OPTION_VAL_5_AST },
         { "10%", OPTION_VAL_10_AST },
         { "15%", OPTION_VAL_15_AST },
         { "20%", OPTION_VAL_20_AST },
         { "25%", OPTION_VAL_25_AST },
         { "30%", OPTION_VAL_30_AST },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_AST,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_purupuru",
      CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_AST,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_AST,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_AST,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_AST,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_AST },
         { "Red",      OPTION_VAL_RED_AST },
         { "Green",    OPTION_VAL_GREEN_AST },
         { "Blue",     OPTION_VAL_BLUE_AST },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_AST,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_AST },
         { "Red",      OPTION_VAL_RED_AST },
         { "Green",    OPTION_VAL_GREEN_AST },
         { "Blue",     OPTION_VAL_BLUE_AST },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_AST,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_AST },
         { "Red",      OPTION_VAL_RED_AST },
         { "Green",    OPTION_VAL_GREEN_AST },
         { "Blue",     OPTION_VAL_BLUE_AST },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_AST,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_AST },
         { "Red",      OPTION_VAL_RED_AST },
         { "Green",    OPTION_VAL_GREEN_AST },
         { "Blue",     OPTION_VAL_BLUE_AST },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_AST,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_AST,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_AST,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_AST },
         { "All VMUs", OPTION_VAL_ALL_VMUS_AST },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_AST,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_AST,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_AST,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_AST },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_AST },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_AST },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_AST },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_AST },
         { "2x", OPTION_VAL_2_O27_AST },
         { "3x", OPTION_VAL_3X_AST },
         { "4x", OPTION_VAL_4_AST },
         { "5x", OPTION_VAL_5X_AST },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_AST },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_AST },
         { "BLACK 02",          OPTION_VAL_BLACK_02_AST },
         { "BLUE 03",           OPTION_VAL_BLUE_AST },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_AST },
         { "GREEN 05",          OPTION_VAL_GREEN_AST },
         { "CYAN 06",           OPTION_VAL_CYAN_06_AST },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_AST },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_AST },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_AST },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_AST },
         { "RED 11",            OPTION_VAL_RED_AST },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_AST },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_AST },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_AST },
         { "GRAY 15",           OPTION_VAL_GRAY_15_AST },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_AST },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_AST },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_AST },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_AST },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_AST },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_AST },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_AST },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_AST },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_AST },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_AST },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_AST },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_AST },
         { "WHITE 28",          OPTION_VAL_WHITE_AST },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_AST },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_AST },
         { "BLACK 02",          OPTION_VAL_BLACK_02_AST },
         { "BLUE 03",           OPTION_VAL_BLUE_AST },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_AST },
         { "GREEN 05",          OPTION_VAL_GREEN_AST },
         { "CYAN 06",           OPTION_VAL_CYAN_06_AST },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_AST },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_AST },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_AST },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_AST },
         { "RED 11",            OPTION_VAL_RED_AST },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_AST },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_AST },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_AST },
         { "GRAY 15",           OPTION_VAL_GRAY_15_AST },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_AST },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_AST },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_AST },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_AST },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_AST },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_AST },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_AST },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_AST },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_AST },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_AST },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_AST },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_AST },
         { "WHITE 28",          OPTION_VAL_WHITE_AST },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_AST },
         { "20%",  OPTION_VAL_20_AST },
         { "30%",  OPTION_VAL_30_AST },
         { "40%",  OPTION_VAL_40_AST },
         { "50%",  OPTION_VAL_50_AST },
         { "60%",  OPTION_VAL_60_AST },
         { "70%",  OPTION_VAL_70_AST },
         { "80%",  OPTION_VAL_80_AST },
         { "90%",  OPTION_VAL_90_AST },
         { "100%", OPTION_VAL_100_AST },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_AST },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_AST },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_AST },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_AST },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_AST },
         { "2x", OPTION_VAL_2_O27_AST },
         { "3x", OPTION_VAL_3X_AST },
         { "4x", OPTION_VAL_4_AST },
         { "5x", OPTION_VAL_5X_AST },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_AST },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_AST },
         { "BLACK 02",          OPTION_VAL_BLACK_02_AST },
         { "BLUE 03",           OPTION_VAL_BLUE_AST },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_AST },
         { "GREEN 05",          OPTION_VAL_GREEN_AST },
         { "CYAN 06",           OPTION_VAL_CYAN_06_AST },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_AST },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_AST },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_AST },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_AST },
         { "RED 11",            OPTION_VAL_RED_AST },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_AST },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_AST },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_AST },
         { "GRAY 15",           OPTION_VAL_GRAY_15_AST },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_AST },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_AST },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_AST },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_AST },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_AST },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_AST },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_AST },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_AST },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_AST },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_AST },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_AST },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_AST },
         { "WHITE 28",          OPTION_VAL_WHITE_AST },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_AST },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_AST },
         { "BLACK 02",          OPTION_VAL_BLACK_02_AST },
         { "BLUE 03",           OPTION_VAL_BLUE_AST },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_AST },
         { "GREEN 05",          OPTION_VAL_GREEN_AST },
         { "CYAN 06",           OPTION_VAL_CYAN_06_AST },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_AST },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_AST },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_AST },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_AST },
         { "RED 11",            OPTION_VAL_RED_AST },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_AST },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_AST },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_AST },
         { "GRAY 15",           OPTION_VAL_GRAY_15_AST },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_AST },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_AST },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_AST },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_AST },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_AST },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_AST },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_AST },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_AST },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_AST },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_AST },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_AST },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_AST },
         { "WHITE 28",          OPTION_VAL_WHITE_AST },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_AST },
         { "20%",  OPTION_VAL_20_AST },
         { "30%",  OPTION_VAL_30_AST },
         { "40%",  OPTION_VAL_40_AST },
         { "50%",  OPTION_VAL_50_AST },
         { "60%",  OPTION_VAL_60_AST },
         { "70%",  OPTION_VAL_70_AST },
         { "80%",  OPTION_VAL_80_AST },
         { "90%",  OPTION_VAL_90_AST },
         { "100%", OPTION_VAL_100_AST },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_AST },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_AST },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_AST },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_AST },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_AST },
         { "2x", OPTION_VAL_2_O27_AST },
         { "3x", OPTION_VAL_3X_AST },
         { "4x", OPTION_VAL_4_AST },
         { "5x", OPTION_VAL_5X_AST },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_AST },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_AST },
         { "BLACK 02",          OPTION_VAL_BLACK_02_AST },
         { "BLUE 03",           OPTION_VAL_BLUE_AST },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_AST },
         { "GREEN 05",          OPTION_VAL_GREEN_AST },
         { "CYAN 06",           OPTION_VAL_CYAN_06_AST },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_AST },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_AST },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_AST },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_AST },
         { "RED 11",            OPTION_VAL_RED_AST },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_AST },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_AST },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_AST },
         { "GRAY 15",           OPTION_VAL_GRAY_15_AST },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_AST },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_AST },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_AST },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_AST },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_AST },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_AST },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_AST },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_AST },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_AST },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_AST },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_AST },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_AST },
         { "WHITE 28",          OPTION_VAL_WHITE_AST },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_AST },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_AST },
         { "BLACK 02",          OPTION_VAL_BLACK_02_AST },
         { "BLUE 03",           OPTION_VAL_BLUE_AST },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_AST },
         { "GREEN 05",          OPTION_VAL_GREEN_AST },
         { "CYAN 06",           OPTION_VAL_CYAN_06_AST },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_AST },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_AST },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_AST },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_AST },
         { "RED 11",            OPTION_VAL_RED_AST },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_AST },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_AST },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_AST },
         { "GRAY 15",           OPTION_VAL_GRAY_15_AST },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_AST },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_AST },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_AST },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_AST },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_AST },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_AST },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_AST },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_AST },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_AST },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_AST },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_AST },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_AST },
         { "WHITE 28",          OPTION_VAL_WHITE_AST },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_AST },
         { "20%",  OPTION_VAL_20_AST },
         { "30%",  OPTION_VAL_30_AST },
         { "40%",  OPTION_VAL_40_AST },
         { "50%",  OPTION_VAL_50_AST },
         { "60%",  OPTION_VAL_60_AST },
         { "70%",  OPTION_VAL_70_AST },
         { "80%",  OPTION_VAL_80_AST },
         { "90%",  OPTION_VAL_90_AST },
         { "100%", OPTION_VAL_100_AST },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_AST },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_AST },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_AST },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_AST },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_AST },
         { "2x", OPTION_VAL_2_O27_AST },
         { "3x", OPTION_VAL_3X_AST },
         { "4x", OPTION_VAL_4_AST },
         { "5x", OPTION_VAL_5X_AST },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_AST },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_AST },
         { "BLACK 02",          OPTION_VAL_BLACK_02_AST },
         { "BLUE 03",           OPTION_VAL_BLUE_AST },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_AST },
         { "GREEN 05",          OPTION_VAL_GREEN_AST },
         { "CYAN 06",           OPTION_VAL_CYAN_06_AST },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_AST },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_AST },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_AST },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_AST },
         { "RED 11",            OPTION_VAL_RED_AST },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_AST },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_AST },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_AST },
         { "GRAY 15",           OPTION_VAL_GRAY_15_AST },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_AST },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_AST },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_AST },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_AST },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_AST },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_AST },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_AST },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_AST },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_AST },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_AST },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_AST },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_AST },
         { "WHITE 28",          OPTION_VAL_WHITE_AST },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_AST },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_AST },
         { "BLACK 02",          OPTION_VAL_BLACK_02_AST },
         { "BLUE 03",           OPTION_VAL_BLUE_AST },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_AST },
         { "GREEN 05",          OPTION_VAL_GREEN_AST },
         { "CYAN 06",           OPTION_VAL_CYAN_06_AST },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_AST },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_AST },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_AST },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_AST },
         { "RED 11",            OPTION_VAL_RED_AST },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_AST },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_AST },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_AST },
         { "GRAY 15",           OPTION_VAL_GRAY_15_AST },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_AST },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_AST },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_AST },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_AST },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_AST },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_AST },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_AST },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_AST },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_AST },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_AST },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_AST },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_AST },
         { "WHITE 28",          OPTION_VAL_WHITE_AST },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_AST },
         { "20%",  OPTION_VAL_20_AST },
         { "30%",  OPTION_VAL_30_AST },
         { "40%",  OPTION_VAL_40_AST },
         { "50%",  OPTION_VAL_50_AST },
         { "60%",  OPTION_VAL_60_AST },
         { "70%",  OPTION_VAL_70_AST },
         { "80%",  OPTION_VAL_80_AST },
         { "90%",  OPTION_VAL_90_AST },
         { "100%", OPTION_VAL_100_AST },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_ast = {
   option_cats_ast,
   option_defs_ast
};

/* RETRO_LANGUAGE_CA */

#define CATEGORY_SYSTEM_LABEL_CA "Sistema"
#define CATEGORY_SYSTEM_INFO_0_CA NULL
#define CATEGORY_VIDEO_LABEL_CA "Vídeo"
#define CATEGORY_VIDEO_INFO_0_CA NULL
#define CATEGORY_PERFORMANCE_LABEL_CA "Rendiment"
#define CATEGORY_PERFORMANCE_INFO_0_CA NULL
#define CATEGORY_HACKS_LABEL_CA NULL
#define CATEGORY_HACKS_INFO_0_CA NULL
#define CATEGORY_INPUT_LABEL_CA "Dispositius de joc"
#define CATEGORY_INPUT_INFO_0_CA NULL
#define CATEGORY_VMU_LABEL_CA NULL
#define CATEGORY_VMU_INFO_0_CA NULL
#define CORE_OPTION_NAME_REGION_LABEL_CA "Regió"
#define OPTION_VAL_JAPAN_CA NULL
#define OPTION_VAL_USA_CA NULL
#define OPTION_VAL_EUROPE_CA NULL
#define OPTION_VAL_DEFAULT_CA NULL
#define CORE_OPTION_NAME_LANGUAGE_LABEL_CA "Llengua"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_CA NULL
#define OPTION_VAL_JAPANESE_CA NULL
#define OPTION_VAL_ENGLISH_CA NULL
#define OPTION_VAL_GERMAN_CA NULL
#define OPTION_VAL_FRENCH_CA NULL
#define OPTION_VAL_SPANISH_CA NULL
#define OPTION_VAL_ITALIAN_CA NULL
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_CA NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_CA NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_CA NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_CA NULL
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_CA NULL
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_CA NULL
#define CORE_OPTION_NAME_FORCE_WINCE_LABEL_CA NULL
#define CORE_OPTION_NAME_FORCE_WINCE_INFO_0_CA NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_CA NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_CA NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_CA NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_CA NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_CA NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_CA NULL
#define CORE_OPTION_NAME_UPNP_LABEL_CA NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_CA NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_CA "Resolució Interna"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_CA NULL
#define OPTION_VAL_320X240_CA "320×240"
#define OPTION_VAL_640X480_CA "640×480"
#define OPTION_VAL_800X600_CA "800×600"
#define OPTION_VAL_960X720_CA "960×720"
#define OPTION_VAL_1024X768_CA "1024×768"
#define OPTION_VAL_1280X960_CA "1280×960"
#define OPTION_VAL_1440X1080_CA "1440×1080"
#define OPTION_VAL_1600X1200_CA "1600×1200"
#define OPTION_VAL_1920X1440_CA "1920×1440"
#define OPTION_VAL_2560X1920_CA "2560×1920"
#define OPTION_VAL_2880X2160_CA "2880×2160"
#define OPTION_VAL_3200X2400_CA "3200×2400"
#define OPTION_VAL_3840X2880_CA "3840×2880"
#define OPTION_VAL_4480X3360_CA "4480×3360"
#define OPTION_VAL_5120X3840_CA "5120×3840"
#define OPTION_VAL_5760X4320_CA "5760×4320"
#define OPTION_VAL_6400X4800_CA "6400×4800"
#define OPTION_VAL_7040X5280_CA "7040×5280"
#define OPTION_VAL_7680X5760_CA "7680×5760"
#define OPTION_VAL_8320X6240_CA "8320×6240"
#define OPTION_VAL_8960X6720_CA "8960×6720"
#define OPTION_VAL_9600X7200_CA "9600×7200"
#define OPTION_VAL_10240X7680_CA "10240×7680"
#define OPTION_VAL_10880X8160_CA "10880×8160"
#define OPTION_VAL_11520X8640_CA "11520×8640"
#define OPTION_VAL_12160X9120_CA "12160×9120"
#define OPTION_VAL_12800X9600_CA "12800×9600"
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_CA NULL
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_CA NULL
#define OPTION_VAL_VGA_CA NULL
#define OPTION_VAL_TV_RGB_CA NULL
#define OPTION_VAL_TV_COMPOSITE_CA NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_CA NULL
#define OPTION_VAL_NTSC_CA NULL
#define OPTION_VAL_PAL_CA NULL
#define OPTION_VAL_PAL_N_CA NULL
#define OPTION_VAL_PAL_M_CA NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_CA "Orientació de la pantalla"
#define OPTION_VAL_HORIZONTAL_CA NULL
#define OPTION_VAL_VERTICAL_CA NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_CA NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_CA NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_CA NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_CA NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_CA NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_CA NULL
#define OPTION_VAL_512MB_CA NULL
#define OPTION_VAL_1GB_CA NULL
#define OPTION_VAL_2GB_CA NULL
#define OPTION_VAL_4GB_CA NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_CA NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_CA NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_CA NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_CA NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_CA NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_CA NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_CA NULL
#define CORE_OPTION_NAME_FOG_LABEL_CA NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_CA NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_CA NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_CA NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_CA NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_CA NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_CA NULL
#define OPTION_VAL_1_CA NULL
#define OPTION_VAL_2_CA NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_CA NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_CA NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_CA NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_CA NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_CA NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_CA NULL
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_CA NULL
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_CA NULL
#define OPTION_VAL_2_O27_CA NULL
#define OPTION_VAL_4_CA NULL
#define OPTION_VAL_6_CA NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_CA NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_CA NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_CA NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_CA NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_CA NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_CA NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_CA NULL
#define OPTION_VAL_SOME_CA NULL
#define OPTION_VAL_MORE_CA NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_CA NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_CA NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_CA NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_CA NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_CA NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_CA NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_CA NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_CA NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_CA NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_CA NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_CA NULL
#define OPTION_VAL_0_CA NULL
#define OPTION_VAL_5_CA NULL
#define OPTION_VAL_10_CA NULL
#define OPTION_VAL_15_CA NULL
#define OPTION_VAL_20_CA NULL
#define OPTION_VAL_25_CA NULL
#define OPTION_VAL_30_CA NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_CA NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_CA NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_CA NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_CA NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_CA NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_CA NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_CA NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_CA NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_CA NULL
#define OPTION_VAL_WHITE_CA NULL
#define OPTION_VAL_RED_CA NULL
#define OPTION_VAL_GREEN_CA NULL
#define OPTION_VAL_BLUE_CA NULL
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_CA NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_CA NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_CA NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CA NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_CA NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_CA NULL
#define OPTION_VAL_VMU_A1_CA NULL
#define OPTION_VAL_ALL_VMUS_CA NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CA NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_CA NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_CA NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_CA NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_CA NULL
#define OPTION_VAL_UPPER_LEFT_CA NULL
#define OPTION_VAL_UPPER_RIGHT_CA NULL
#define OPTION_VAL_LOWER_LEFT_CA NULL
#define OPTION_VAL_LOWER_RIGHT_CA NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_CA NULL
#define OPTION_VAL_1X_CA NULL
#define OPTION_VAL_3X_CA NULL
#define OPTION_VAL_5X_CA NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_CA NULL
#define OPTION_VAL_DEFAULT_ON_00_CA NULL
#define OPTION_VAL_DEFAULT_OFF_01_CA NULL
#define OPTION_VAL_BLACK_02_CA NULL
#define OPTION_VAL_LIGHT_BLUE_04_CA NULL
#define OPTION_VAL_CYAN_06_CA NULL
#define OPTION_VAL_CYAN_BLUE_07_CA NULL
#define OPTION_VAL_LIGHT_GREEN_08_CA NULL
#define OPTION_VAL_CYAN_GREEN_09_CA NULL
#define OPTION_VAL_LIGHT_CYAN_10_CA NULL
#define OPTION_VAL_PURPLE_12_CA NULL
#define OPTION_VAL_LIGHT_PURPLE_13_CA NULL
#define OPTION_VAL_YELLOW_14_CA NULL
#define OPTION_VAL_GRAY_15_CA NULL
#define OPTION_VAL_LIGHT_PURPLE_2_16_CA NULL
#define OPTION_VAL_LIGHT_GREEN_2_17_CA NULL
#define OPTION_VAL_LIGHT_GREEN_3_18_CA NULL
#define OPTION_VAL_LIGHT_CYAN_2_19_CA NULL
#define OPTION_VAL_LIGHT_RED_2_20_CA NULL
#define OPTION_VAL_MAGENTA_21_CA NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_CA NULL
#define OPTION_VAL_LIGHT_ORANGE_23_CA NULL
#define OPTION_VAL_ORANGE_24_CA NULL
#define OPTION_VAL_LIGHT_PURPLE_4_25_CA NULL
#define OPTION_VAL_LIGHT_YELLOW_26_CA NULL
#define OPTION_VAL_LIGHT_YELLOW_2_27_CA NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_CA NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_CA NULL
#define OPTION_VAL_40_CA NULL
#define OPTION_VAL_50_CA NULL
#define OPTION_VAL_60_CA NULL
#define OPTION_VAL_70_CA NULL
#define OPTION_VAL_80_CA NULL
#define OPTION_VAL_90_CA NULL
#define OPTION_VAL_100_CA NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_CA NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_CA NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_CA NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_CA NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_CA NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_CA NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_CA NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_CA NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_CA NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_CA NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_CA NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_CA NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_CA NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_CA NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_CA NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_CA NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_CA NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_CA NULL

struct retro_core_option_v2_category option_cats_ca[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_CA,
      CATEGORY_SYSTEM_INFO_0_CA
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_CA,
      CATEGORY_VIDEO_INFO_0_CA
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_CA,
      CATEGORY_PERFORMANCE_INFO_0_CA
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_CA,
      CATEGORY_HACKS_INFO_0_CA
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_CA,
      CATEGORY_INPUT_INFO_0_CA
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_CA,
      CATEGORY_VMU_INFO_0_CA
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_ca[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_CA,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_CA },
         { "USA",     OPTION_VAL_USA_CA },
         { "Europe",  OPTION_VAL_EUROPE_CA },
         { "Default", OPTION_VAL_DEFAULT_CA },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_CA,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_CA },
         { "English",  OPTION_VAL_ENGLISH_CA },
         { "German",   OPTION_VAL_GERMAN_CA },
         { "French",   OPTION_VAL_FRENCH_CA },
         { "Spanish",  OPTION_VAL_SPANISH_CA },
         { "Italian",  OPTION_VAL_ITALIAN_CA },
         { "Default",  OPTION_VAL_DEFAULT_CA },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_CA,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_boot_to_bios",
      CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_CA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_CA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_force_wince",
      CORE_OPTION_NAME_FORCE_WINCE_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_FORCE_WINCE_INFO_0_CA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_CA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_CA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_CA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_CA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_CA,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_CA },
         { "640x480",    OPTION_VAL_640X480_CA },
         { "800x600",    OPTION_VAL_800X600_CA },
         { "960x720",    OPTION_VAL_960X720_CA },
         { "1024x768",   OPTION_VAL_1024X768_CA },
         { "1280x960",   OPTION_VAL_1280X960_CA },
         { "1440x1080",  OPTION_VAL_1440X1080_CA },
         { "1600x1200",  OPTION_VAL_1600X1200_CA },
         { "1920x1440",  OPTION_VAL_1920X1440_CA },
         { "2560x1920",  OPTION_VAL_2560X1920_CA },
         { "2880x2160",  OPTION_VAL_2880X2160_CA },
         { "3200x2400",  OPTION_VAL_3200X2400_CA },
         { "3840x2880",  OPTION_VAL_3840X2880_CA },
         { "4480x3360",  OPTION_VAL_4480X3360_CA },
         { "5120x3840",  OPTION_VAL_5120X3840_CA },
         { "5760x4320",  OPTION_VAL_5760X4320_CA },
         { "6400x4800",  OPTION_VAL_6400X4800_CA },
         { "7040x5280",  OPTION_VAL_7040X5280_CA },
         { "7680x5760",  OPTION_VAL_7680X5760_CA },
         { "8320x6240",  OPTION_VAL_8320X6240_CA },
         { "8960x6720",  OPTION_VAL_8960X6720_CA },
         { "9600x7200",  OPTION_VAL_9600X7200_CA },
         { "10240x7680", OPTION_VAL_10240X7680_CA },
         { "10880x8160", OPTION_VAL_10880X8160_CA },
         { "11520x8640", OPTION_VAL_11520X8640_CA },
         { "12160x9120", OPTION_VAL_12160X9120_CA },
         { "12800x9600", OPTION_VAL_12800X9600_CA },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_CA,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_CA },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_CA },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_CA },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_CA,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_CA },
         { "PAL",     OPTION_VAL_PAL_CA },
         { "PAL_N",   OPTION_VAL_PAL_N_CA },
         { "PAL_M",   OPTION_VAL_PAL_M_CA },
         { "Default", OPTION_VAL_DEFAULT_CA },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_CA,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_CA },
         { "vertical",   OPTION_VAL_VERTICAL_CA },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_CA,
      NULL,
      "",
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_CA },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_CA },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_CA },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_CA,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_CA },
         { "1GB",   OPTION_VAL_1GB_CA },
         { "2GB",   OPTION_VAL_2GB_CA },
         { "4GB",   OPTION_VAL_4GB_CA },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_CA,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_CA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_CA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_CA,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_CA,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_CA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_CA,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_CA,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_CA },
         { "1",  OPTION_VAL_1_CA },
         { "2",  OPTION_VAL_2_CA },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_CA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_CA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_CA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_CA,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O27_CA },
         { "4",  OPTION_VAL_4_CA },
         { "6",  OPTION_VAL_6_CA },
         { NULL, NULL },
      },
      "1",
   },
   {/* TODO: needs clarification */
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_CA,
      NULL,
      "",
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_CA,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_CA,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_CA,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_CA,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_CA },
         { "more", OPTION_VAL_MORE_CA },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_CA,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_CA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_CA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_CA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_CA,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: probably needs explanation */
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_CA,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_CA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_CA },
         { "5%",  OPTION_VAL_5_CA },
         { "10%", OPTION_VAL_10_CA },
         { "15%", OPTION_VAL_15_CA },
         { "20%", OPTION_VAL_20_CA },
         { "25%", OPTION_VAL_25_CA },
         { "30%", OPTION_VAL_30_CA },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_CA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_CA },
         { "5%",  OPTION_VAL_5_CA },
         { "10%", OPTION_VAL_10_CA },
         { "15%", OPTION_VAL_15_CA },
         { "20%", OPTION_VAL_20_CA },
         { "25%", OPTION_VAL_25_CA },
         { "30%", OPTION_VAL_30_CA },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_CA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_purupuru",
      CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_CA,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_CA,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_CA,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_CA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_CA },
         { "Red",      OPTION_VAL_RED_CA },
         { "Green",    OPTION_VAL_GREEN_CA },
         { "Blue",     OPTION_VAL_BLUE_CA },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_CA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_CA },
         { "Red",      OPTION_VAL_RED_CA },
         { "Green",    OPTION_VAL_GREEN_CA },
         { "Blue",     OPTION_VAL_BLUE_CA },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_CA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_CA },
         { "Red",      OPTION_VAL_RED_CA },
         { "Green",    OPTION_VAL_GREEN_CA },
         { "Blue",     OPTION_VAL_BLUE_CA },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_CA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_CA },
         { "Red",      OPTION_VAL_RED_CA },
         { "Green",    OPTION_VAL_GREEN_CA },
         { "Blue",     OPTION_VAL_BLUE_CA },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CA,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_CA,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_CA,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_CA },
         { "All VMUs", OPTION_VAL_ALL_VMUS_CA },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CA,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_CA,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_CA,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_CA },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_CA },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_CA },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_CA },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_CA },
         { "2x", OPTION_VAL_2_O27_CA },
         { "3x", OPTION_VAL_3X_CA },
         { "4x", OPTION_VAL_4_CA },
         { "5x", OPTION_VAL_5X_CA },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CA },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CA },
         { "BLUE 03",           OPTION_VAL_BLUE_CA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CA },
         { "GREEN 05",          OPTION_VAL_GREEN_CA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CA },
         { "RED 11",            OPTION_VAL_RED_CA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CA },
         { "WHITE 28",          OPTION_VAL_WHITE_CA },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CA },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CA },
         { "BLUE 03",           OPTION_VAL_BLUE_CA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CA },
         { "GREEN 05",          OPTION_VAL_GREEN_CA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CA },
         { "RED 11",            OPTION_VAL_RED_CA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CA },
         { "WHITE 28",          OPTION_VAL_WHITE_CA },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_CA },
         { "20%",  OPTION_VAL_20_CA },
         { "30%",  OPTION_VAL_30_CA },
         { "40%",  OPTION_VAL_40_CA },
         { "50%",  OPTION_VAL_50_CA },
         { "60%",  OPTION_VAL_60_CA },
         { "70%",  OPTION_VAL_70_CA },
         { "80%",  OPTION_VAL_80_CA },
         { "90%",  OPTION_VAL_90_CA },
         { "100%", OPTION_VAL_100_CA },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_CA },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_CA },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_CA },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_CA },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_CA },
         { "2x", OPTION_VAL_2_O27_CA },
         { "3x", OPTION_VAL_3X_CA },
         { "4x", OPTION_VAL_4_CA },
         { "5x", OPTION_VAL_5X_CA },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CA },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CA },
         { "BLUE 03",           OPTION_VAL_BLUE_CA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CA },
         { "GREEN 05",          OPTION_VAL_GREEN_CA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CA },
         { "RED 11",            OPTION_VAL_RED_CA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CA },
         { "WHITE 28",          OPTION_VAL_WHITE_CA },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CA },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CA },
         { "BLUE 03",           OPTION_VAL_BLUE_CA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CA },
         { "GREEN 05",          OPTION_VAL_GREEN_CA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CA },
         { "RED 11",            OPTION_VAL_RED_CA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CA },
         { "WHITE 28",          OPTION_VAL_WHITE_CA },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_CA },
         { "20%",  OPTION_VAL_20_CA },
         { "30%",  OPTION_VAL_30_CA },
         { "40%",  OPTION_VAL_40_CA },
         { "50%",  OPTION_VAL_50_CA },
         { "60%",  OPTION_VAL_60_CA },
         { "70%",  OPTION_VAL_70_CA },
         { "80%",  OPTION_VAL_80_CA },
         { "90%",  OPTION_VAL_90_CA },
         { "100%", OPTION_VAL_100_CA },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_CA },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_CA },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_CA },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_CA },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_CA },
         { "2x", OPTION_VAL_2_O27_CA },
         { "3x", OPTION_VAL_3X_CA },
         { "4x", OPTION_VAL_4_CA },
         { "5x", OPTION_VAL_5X_CA },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CA },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CA },
         { "BLUE 03",           OPTION_VAL_BLUE_CA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CA },
         { "GREEN 05",          OPTION_VAL_GREEN_CA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CA },
         { "RED 11",            OPTION_VAL_RED_CA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CA },
         { "WHITE 28",          OPTION_VAL_WHITE_CA },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CA },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CA },
         { "BLUE 03",           OPTION_VAL_BLUE_CA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CA },
         { "GREEN 05",          OPTION_VAL_GREEN_CA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CA },
         { "RED 11",            OPTION_VAL_RED_CA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CA },
         { "WHITE 28",          OPTION_VAL_WHITE_CA },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_CA },
         { "20%",  OPTION_VAL_20_CA },
         { "30%",  OPTION_VAL_30_CA },
         { "40%",  OPTION_VAL_40_CA },
         { "50%",  OPTION_VAL_50_CA },
         { "60%",  OPTION_VAL_60_CA },
         { "70%",  OPTION_VAL_70_CA },
         { "80%",  OPTION_VAL_80_CA },
         { "90%",  OPTION_VAL_90_CA },
         { "100%", OPTION_VAL_100_CA },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_CA },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_CA },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_CA },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_CA },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_CA },
         { "2x", OPTION_VAL_2_O27_CA },
         { "3x", OPTION_VAL_3X_CA },
         { "4x", OPTION_VAL_4_CA },
         { "5x", OPTION_VAL_5X_CA },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CA },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CA },
         { "BLUE 03",           OPTION_VAL_BLUE_CA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CA },
         { "GREEN 05",          OPTION_VAL_GREEN_CA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CA },
         { "RED 11",            OPTION_VAL_RED_CA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CA },
         { "WHITE 28",          OPTION_VAL_WHITE_CA },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CA },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CA },
         { "BLUE 03",           OPTION_VAL_BLUE_CA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CA },
         { "GREEN 05",          OPTION_VAL_GREEN_CA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CA },
         { "RED 11",            OPTION_VAL_RED_CA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CA },
         { "WHITE 28",          OPTION_VAL_WHITE_CA },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_CA },
         { "20%",  OPTION_VAL_20_CA },
         { "30%",  OPTION_VAL_30_CA },
         { "40%",  OPTION_VAL_40_CA },
         { "50%",  OPTION_VAL_50_CA },
         { "60%",  OPTION_VAL_60_CA },
         { "70%",  OPTION_VAL_70_CA },
         { "80%",  OPTION_VAL_80_CA },
         { "90%",  OPTION_VAL_90_CA },
         { "100%", OPTION_VAL_100_CA },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_ca = {
   option_cats_ca,
   option_defs_ca
};

/* RETRO_LANGUAGE_CHS */

#define CATEGORY_SYSTEM_LABEL_CHS "系统"
#define CATEGORY_SYSTEM_INFO_0_CHS NULL
#define CATEGORY_VIDEO_LABEL_CHS "视频"
#define CATEGORY_VIDEO_INFO_0_CHS NULL
#define CATEGORY_PERFORMANCE_LABEL_CHS "性能"
#define CATEGORY_PERFORMANCE_INFO_0_CHS NULL
#define CATEGORY_HACKS_LABEL_CHS "模拟修改"
#define CATEGORY_HACKS_INFO_0_CHS NULL
#define CATEGORY_INPUT_LABEL_CHS "输入"
#define CATEGORY_INPUT_INFO_0_CHS NULL
#define CATEGORY_VMU_LABEL_CHS NULL
#define CATEGORY_VMU_INFO_0_CHS NULL
#define CORE_OPTION_NAME_REGION_LABEL_CHS "区域"
#define OPTION_VAL_JAPAN_CHS "日本"
#define OPTION_VAL_USA_CHS NULL
#define OPTION_VAL_EUROPE_CHS "欧洲"
#define OPTION_VAL_DEFAULT_CHS "默认值"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_CHS "语言"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_CHS NULL
#define OPTION_VAL_JAPANESE_CHS "日本历"
#define OPTION_VAL_ENGLISH_CHS "英语"
#define OPTION_VAL_GERMAN_CHS "德语"
#define OPTION_VAL_FRENCH_CHS "法语"
#define OPTION_VAL_SPANISH_CHS "西班牙语"
#define OPTION_VAL_ITALIAN_CHS "意大利语"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_CHS NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_CHS NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_CHS NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_CHS NULL
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_CHS NULL
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_CHS NULL
#define CORE_OPTION_NAME_FORCE_WINCE_LABEL_CHS NULL
#define CORE_OPTION_NAME_FORCE_WINCE_INFO_0_CHS NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_CHS NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_CHS NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_CHS NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_CHS NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_CHS NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_CHS NULL
#define CORE_OPTION_NAME_UPNP_LABEL_CHS NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_CHS NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_CHS "内部分辨率"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_CHS NULL
#define OPTION_VAL_320X240_CHS NULL
#define OPTION_VAL_640X480_CHS NULL
#define OPTION_VAL_800X600_CHS NULL
#define OPTION_VAL_960X720_CHS NULL
#define OPTION_VAL_1024X768_CHS NULL
#define OPTION_VAL_1280X960_CHS NULL
#define OPTION_VAL_1440X1080_CHS NULL
#define OPTION_VAL_1600X1200_CHS NULL
#define OPTION_VAL_1920X1440_CHS NULL
#define OPTION_VAL_2560X1920_CHS NULL
#define OPTION_VAL_2880X2160_CHS NULL
#define OPTION_VAL_3200X2400_CHS NULL
#define OPTION_VAL_3840X2880_CHS NULL
#define OPTION_VAL_4480X3360_CHS NULL
#define OPTION_VAL_5120X3840_CHS NULL
#define OPTION_VAL_5760X4320_CHS NULL
#define OPTION_VAL_6400X4800_CHS NULL
#define OPTION_VAL_7040X5280_CHS NULL
#define OPTION_VAL_7680X5760_CHS NULL
#define OPTION_VAL_8320X6240_CHS NULL
#define OPTION_VAL_8960X6720_CHS NULL
#define OPTION_VAL_9600X7200_CHS NULL
#define OPTION_VAL_10240X7680_CHS NULL
#define OPTION_VAL_10880X8160_CHS NULL
#define OPTION_VAL_11520X8640_CHS NULL
#define OPTION_VAL_12160X9120_CHS NULL
#define OPTION_VAL_12800X9600_CHS NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_CHS NULL
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_CHS NULL
#define OPTION_VAL_VGA_CHS NULL
#define OPTION_VAL_TV_RGB_CHS NULL
#define OPTION_VAL_TV_COMPOSITE_CHS NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_CHS NULL
#define OPTION_VAL_NTSC_CHS NULL
#define OPTION_VAL_PAL_CHS NULL
#define OPTION_VAL_PAL_N_CHS NULL
#define OPTION_VAL_PAL_M_CHS NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_CHS "屏幕方向"
#define OPTION_VAL_HORIZONTAL_CHS "水平"
#define OPTION_VAL_VERTICAL_CHS "垂直"
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_CHS NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_CHS NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_CHS NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_CHS NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_CHS NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_CHS NULL
#define OPTION_VAL_512MB_CHS NULL
#define OPTION_VAL_1GB_CHS NULL
#define OPTION_VAL_2GB_CHS NULL
#define OPTION_VAL_4GB_CHS NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_CHS NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_CHS NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_CHS NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_CHS NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_CHS NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_CHS NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_CHS NULL
#define CORE_OPTION_NAME_FOG_LABEL_CHS NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_CHS NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_CHS NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_CHS NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_CHS NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_CHS "纹理过滤"
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_CHS NULL
#define OPTION_VAL_1_CHS NULL
#define OPTION_VAL_2_CHS NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_CHS NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_CHS NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_CHS NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_CHS NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_CHS NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_CHS NULL
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_CHS NULL
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_CHS NULL
#define OPTION_VAL_2_O27_CHS "2 倍"
#define OPTION_VAL_4_CHS "4 倍"
#define OPTION_VAL_6_CHS NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_CHS NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_CHS NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_CHS NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_CHS "多线程渲染"
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_CHS NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_CHS NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_CHS NULL
#define OPTION_VAL_SOME_CHS "正常"
#define OPTION_VAL_MORE_CHS "最大"
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_CHS NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_CHS NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_CHS NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_CHS NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_CHS NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_CHS NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_CHS NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_CHS NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_CHS NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_CHS NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_CHS NULL
#define OPTION_VAL_0_CHS NULL
#define OPTION_VAL_5_CHS NULL
#define OPTION_VAL_10_CHS NULL
#define OPTION_VAL_15_CHS NULL
#define OPTION_VAL_20_CHS NULL
#define OPTION_VAL_25_CHS NULL
#define OPTION_VAL_30_CHS NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_CHS NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_CHS NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_CHS NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_CHS NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_CHS NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_CHS NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_CHS NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_CHS NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_CHS NULL
#define OPTION_VAL_WHITE_CHS "白色"
#define OPTION_VAL_RED_CHS "红色"
#define OPTION_VAL_GREEN_CHS "绿色"
#define OPTION_VAL_BLUE_CHS "蓝色"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_CHS NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_CHS NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_CHS NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CHS NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_CHS NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_CHS NULL
#define OPTION_VAL_VMU_A1_CHS NULL
#define OPTION_VAL_ALL_VMUS_CHS NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CHS NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_CHS NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_CHS NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_CHS NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_CHS NULL
#define OPTION_VAL_UPPER_LEFT_CHS NULL
#define OPTION_VAL_UPPER_RIGHT_CHS NULL
#define OPTION_VAL_LOWER_LEFT_CHS NULL
#define OPTION_VAL_LOWER_RIGHT_CHS NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_CHS NULL
#define OPTION_VAL_1X_CHS NULL
#define OPTION_VAL_3X_CHS NULL
#define OPTION_VAL_5X_CHS NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_CHS NULL
#define OPTION_VAL_DEFAULT_ON_00_CHS NULL
#define OPTION_VAL_DEFAULT_OFF_01_CHS NULL
#define OPTION_VAL_BLACK_02_CHS "浓体 (Black)"
#define OPTION_VAL_LIGHT_BLUE_04_CHS "浅蓝"
#define OPTION_VAL_CYAN_06_CHS "青色"
#define OPTION_VAL_CYAN_BLUE_07_CHS NULL
#define OPTION_VAL_LIGHT_GREEN_08_CHS "浅绿色"
#define OPTION_VAL_CYAN_GREEN_09_CHS NULL
#define OPTION_VAL_LIGHT_CYAN_10_CHS "浅青色"
#define OPTION_VAL_PURPLE_12_CHS "紫色"
#define OPTION_VAL_LIGHT_PURPLE_13_CHS NULL
#define OPTION_VAL_YELLOW_14_CHS "黄色"
#define OPTION_VAL_GRAY_15_CHS "灰"
#define OPTION_VAL_LIGHT_PURPLE_2_16_CHS NULL
#define OPTION_VAL_LIGHT_GREEN_2_17_CHS NULL
#define OPTION_VAL_LIGHT_GREEN_3_18_CHS NULL
#define OPTION_VAL_LIGHT_CYAN_2_19_CHS NULL
#define OPTION_VAL_LIGHT_RED_2_20_CHS NULL
#define OPTION_VAL_MAGENTA_21_CHS "品红"
#define OPTION_VAL_LIGHT_PURPLE_3_22_CHS NULL
#define OPTION_VAL_LIGHT_ORANGE_23_CHS "浅橙色"
#define OPTION_VAL_ORANGE_24_CHS "奥朗日"
#define OPTION_VAL_LIGHT_PURPLE_4_25_CHS NULL
#define OPTION_VAL_LIGHT_YELLOW_26_CHS "淡黄色"
#define OPTION_VAL_LIGHT_YELLOW_2_27_CHS NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_CHS NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_CHS NULL
#define OPTION_VAL_40_CHS NULL
#define OPTION_VAL_50_CHS NULL
#define OPTION_VAL_60_CHS NULL
#define OPTION_VAL_70_CHS NULL
#define OPTION_VAL_80_CHS NULL
#define OPTION_VAL_90_CHS NULL
#define OPTION_VAL_100_CHS NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_CHS NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_CHS NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_CHS NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_CHS NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_CHS NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_CHS NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_CHS NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_CHS NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_CHS NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_CHS NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_CHS NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_CHS NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_CHS NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_CHS NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_CHS NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_CHS NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_CHS NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_CHS NULL

struct retro_core_option_v2_category option_cats_chs[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_CHS,
      CATEGORY_SYSTEM_INFO_0_CHS
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_CHS,
      CATEGORY_VIDEO_INFO_0_CHS
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_CHS,
      CATEGORY_PERFORMANCE_INFO_0_CHS
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_CHS,
      CATEGORY_HACKS_INFO_0_CHS
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_CHS,
      CATEGORY_INPUT_INFO_0_CHS
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_CHS,
      CATEGORY_VMU_INFO_0_CHS
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_chs[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_CHS,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_CHS },
         { "USA",     OPTION_VAL_USA_CHS },
         { "Europe",  OPTION_VAL_EUROPE_CHS },
         { "Default", OPTION_VAL_DEFAULT_CHS },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_CHS,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_CHS },
         { "English",  OPTION_VAL_ENGLISH_CHS },
         { "German",   OPTION_VAL_GERMAN_CHS },
         { "French",   OPTION_VAL_FRENCH_CHS },
         { "Spanish",  OPTION_VAL_SPANISH_CHS },
         { "Italian",  OPTION_VAL_ITALIAN_CHS },
         { "Default",  OPTION_VAL_DEFAULT_CHS },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_CHS,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_boot_to_bios",
      CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_CHS,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_CHS,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_force_wince",
      CORE_OPTION_NAME_FORCE_WINCE_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_FORCE_WINCE_INFO_0_CHS,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_CHS,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_CHS,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_CHS,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_CHS,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_CHS,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_CHS },
         { "640x480",    OPTION_VAL_640X480_CHS },
         { "800x600",    OPTION_VAL_800X600_CHS },
         { "960x720",    OPTION_VAL_960X720_CHS },
         { "1024x768",   OPTION_VAL_1024X768_CHS },
         { "1280x960",   OPTION_VAL_1280X960_CHS },
         { "1440x1080",  OPTION_VAL_1440X1080_CHS },
         { "1600x1200",  OPTION_VAL_1600X1200_CHS },
         { "1920x1440",  OPTION_VAL_1920X1440_CHS },
         { "2560x1920",  OPTION_VAL_2560X1920_CHS },
         { "2880x2160",  OPTION_VAL_2880X2160_CHS },
         { "3200x2400",  OPTION_VAL_3200X2400_CHS },
         { "3840x2880",  OPTION_VAL_3840X2880_CHS },
         { "4480x3360",  OPTION_VAL_4480X3360_CHS },
         { "5120x3840",  OPTION_VAL_5120X3840_CHS },
         { "5760x4320",  OPTION_VAL_5760X4320_CHS },
         { "6400x4800",  OPTION_VAL_6400X4800_CHS },
         { "7040x5280",  OPTION_VAL_7040X5280_CHS },
         { "7680x5760",  OPTION_VAL_7680X5760_CHS },
         { "8320x6240",  OPTION_VAL_8320X6240_CHS },
         { "8960x6720",  OPTION_VAL_8960X6720_CHS },
         { "9600x7200",  OPTION_VAL_9600X7200_CHS },
         { "10240x7680", OPTION_VAL_10240X7680_CHS },
         { "10880x8160", OPTION_VAL_10880X8160_CHS },
         { "11520x8640", OPTION_VAL_11520X8640_CHS },
         { "12160x9120", OPTION_VAL_12160X9120_CHS },
         { "12800x9600", OPTION_VAL_12800X9600_CHS },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_CHS,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_CHS },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_CHS },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_CHS },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_CHS,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_CHS },
         { "PAL",     OPTION_VAL_PAL_CHS },
         { "PAL_N",   OPTION_VAL_PAL_N_CHS },
         { "PAL_M",   OPTION_VAL_PAL_M_CHS },
         { "Default", OPTION_VAL_DEFAULT_CHS },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_CHS,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_CHS },
         { "vertical",   OPTION_VAL_VERTICAL_CHS },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_CHS,
      NULL,
      "",
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_CHS },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_CHS },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_CHS },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_CHS,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_CHS },
         { "1GB",   OPTION_VAL_1GB_CHS },
         { "2GB",   OPTION_VAL_2GB_CHS },
         { "4GB",   OPTION_VAL_4GB_CHS },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_CHS,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_CHS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_CHS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_CHS,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_CHS,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_CHS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_CHS,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_CHS,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_CHS },
         { "1",  OPTION_VAL_1_CHS },
         { "2",  OPTION_VAL_2_CHS },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_CHS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_CHS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_CHS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_CHS,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O27_CHS },
         { "4",  OPTION_VAL_4_CHS },
         { "6",  OPTION_VAL_6_CHS },
         { NULL, NULL },
      },
      "1",
   },
   {/* TODO: needs clarification */
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_CHS,
      NULL,
      "",
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_CHS,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_CHS,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_CHS,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_CHS,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_CHS },
         { "more", OPTION_VAL_MORE_CHS },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_CHS,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_CHS,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_CHS,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_CHS,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_CHS,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: probably needs explanation */
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_CHS,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_CHS,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_CHS },
         { "5%",  OPTION_VAL_5_CHS },
         { "10%", OPTION_VAL_10_CHS },
         { "15%", OPTION_VAL_15_CHS },
         { "20%", OPTION_VAL_20_CHS },
         { "25%", OPTION_VAL_25_CHS },
         { "30%", OPTION_VAL_30_CHS },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_CHS,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_CHS },
         { "5%",  OPTION_VAL_5_CHS },
         { "10%", OPTION_VAL_10_CHS },
         { "15%", OPTION_VAL_15_CHS },
         { "20%", OPTION_VAL_20_CHS },
         { "25%", OPTION_VAL_25_CHS },
         { "30%", OPTION_VAL_30_CHS },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_CHS,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_purupuru",
      CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_CHS,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_CHS,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_CHS,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_CHS,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_CHS },
         { "Red",      OPTION_VAL_RED_CHS },
         { "Green",    OPTION_VAL_GREEN_CHS },
         { "Blue",     OPTION_VAL_BLUE_CHS },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_CHS,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_CHS },
         { "Red",      OPTION_VAL_RED_CHS },
         { "Green",    OPTION_VAL_GREEN_CHS },
         { "Blue",     OPTION_VAL_BLUE_CHS },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_CHS,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_CHS },
         { "Red",      OPTION_VAL_RED_CHS },
         { "Green",    OPTION_VAL_GREEN_CHS },
         { "Blue",     OPTION_VAL_BLUE_CHS },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_CHS,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_CHS },
         { "Red",      OPTION_VAL_RED_CHS },
         { "Green",    OPTION_VAL_GREEN_CHS },
         { "Blue",     OPTION_VAL_BLUE_CHS },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CHS,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_CHS,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_CHS,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_CHS },
         { "All VMUs", OPTION_VAL_ALL_VMUS_CHS },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CHS,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_CHS,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_CHS,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_CHS },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_CHS },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_CHS },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_CHS },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_CHS },
         { "2x", OPTION_VAL_2_O27_CHS },
         { "3x", OPTION_VAL_3X_CHS },
         { "4x", OPTION_VAL_4_CHS },
         { "5x", OPTION_VAL_5X_CHS },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CHS },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CHS },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CHS },
         { "BLUE 03",           OPTION_VAL_BLUE_CHS },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CHS },
         { "GREEN 05",          OPTION_VAL_GREEN_CHS },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CHS },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CHS },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CHS },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CHS },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CHS },
         { "RED 11",            OPTION_VAL_RED_CHS },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CHS },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CHS },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CHS },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CHS },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CHS },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CHS },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CHS },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CHS },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CHS },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CHS },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CHS },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CHS },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CHS },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CHS },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CHS },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CHS },
         { "WHITE 28",          OPTION_VAL_WHITE_CHS },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CHS },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CHS },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CHS },
         { "BLUE 03",           OPTION_VAL_BLUE_CHS },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CHS },
         { "GREEN 05",          OPTION_VAL_GREEN_CHS },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CHS },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CHS },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CHS },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CHS },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CHS },
         { "RED 11",            OPTION_VAL_RED_CHS },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CHS },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CHS },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CHS },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CHS },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CHS },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CHS },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CHS },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CHS },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CHS },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CHS },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CHS },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CHS },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CHS },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CHS },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CHS },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CHS },
         { "WHITE 28",          OPTION_VAL_WHITE_CHS },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_CHS },
         { "20%",  OPTION_VAL_20_CHS },
         { "30%",  OPTION_VAL_30_CHS },
         { "40%",  OPTION_VAL_40_CHS },
         { "50%",  OPTION_VAL_50_CHS },
         { "60%",  OPTION_VAL_60_CHS },
         { "70%",  OPTION_VAL_70_CHS },
         { "80%",  OPTION_VAL_80_CHS },
         { "90%",  OPTION_VAL_90_CHS },
         { "100%", OPTION_VAL_100_CHS },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_CHS },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_CHS },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_CHS },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_CHS },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_CHS },
         { "2x", OPTION_VAL_2_O27_CHS },
         { "3x", OPTION_VAL_3X_CHS },
         { "4x", OPTION_VAL_4_CHS },
         { "5x", OPTION_VAL_5X_CHS },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CHS },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CHS },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CHS },
         { "BLUE 03",           OPTION_VAL_BLUE_CHS },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CHS },
         { "GREEN 05",          OPTION_VAL_GREEN_CHS },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CHS },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CHS },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CHS },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CHS },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CHS },
         { "RED 11",            OPTION_VAL_RED_CHS },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CHS },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CHS },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CHS },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CHS },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CHS },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CHS },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CHS },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CHS },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CHS },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CHS },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CHS },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CHS },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CHS },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CHS },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CHS },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CHS },
         { "WHITE 28",          OPTION_VAL_WHITE_CHS },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CHS },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CHS },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CHS },
         { "BLUE 03",           OPTION_VAL_BLUE_CHS },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CHS },
         { "GREEN 05",          OPTION_VAL_GREEN_CHS },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CHS },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CHS },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CHS },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CHS },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CHS },
         { "RED 11",            OPTION_VAL_RED_CHS },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CHS },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CHS },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CHS },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CHS },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CHS },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CHS },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CHS },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CHS },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CHS },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CHS },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CHS },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CHS },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CHS },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CHS },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CHS },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CHS },
         { "WHITE 28",          OPTION_VAL_WHITE_CHS },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_CHS },
         { "20%",  OPTION_VAL_20_CHS },
         { "30%",  OPTION_VAL_30_CHS },
         { "40%",  OPTION_VAL_40_CHS },
         { "50%",  OPTION_VAL_50_CHS },
         { "60%",  OPTION_VAL_60_CHS },
         { "70%",  OPTION_VAL_70_CHS },
         { "80%",  OPTION_VAL_80_CHS },
         { "90%",  OPTION_VAL_90_CHS },
         { "100%", OPTION_VAL_100_CHS },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_CHS },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_CHS },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_CHS },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_CHS },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_CHS },
         { "2x", OPTION_VAL_2_O27_CHS },
         { "3x", OPTION_VAL_3X_CHS },
         { "4x", OPTION_VAL_4_CHS },
         { "5x", OPTION_VAL_5X_CHS },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CHS },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CHS },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CHS },
         { "BLUE 03",           OPTION_VAL_BLUE_CHS },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CHS },
         { "GREEN 05",          OPTION_VAL_GREEN_CHS },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CHS },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CHS },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CHS },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CHS },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CHS },
         { "RED 11",            OPTION_VAL_RED_CHS },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CHS },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CHS },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CHS },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CHS },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CHS },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CHS },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CHS },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CHS },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CHS },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CHS },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CHS },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CHS },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CHS },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CHS },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CHS },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CHS },
         { "WHITE 28",          OPTION_VAL_WHITE_CHS },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CHS },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CHS },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CHS },
         { "BLUE 03",           OPTION_VAL_BLUE_CHS },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CHS },
         { "GREEN 05",          OPTION_VAL_GREEN_CHS },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CHS },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CHS },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CHS },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CHS },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CHS },
         { "RED 11",            OPTION_VAL_RED_CHS },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CHS },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CHS },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CHS },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CHS },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CHS },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CHS },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CHS },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CHS },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CHS },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CHS },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CHS },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CHS },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CHS },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CHS },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CHS },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CHS },
         { "WHITE 28",          OPTION_VAL_WHITE_CHS },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_CHS },
         { "20%",  OPTION_VAL_20_CHS },
         { "30%",  OPTION_VAL_30_CHS },
         { "40%",  OPTION_VAL_40_CHS },
         { "50%",  OPTION_VAL_50_CHS },
         { "60%",  OPTION_VAL_60_CHS },
         { "70%",  OPTION_VAL_70_CHS },
         { "80%",  OPTION_VAL_80_CHS },
         { "90%",  OPTION_VAL_90_CHS },
         { "100%", OPTION_VAL_100_CHS },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_CHS },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_CHS },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_CHS },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_CHS },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_CHS },
         { "2x", OPTION_VAL_2_O27_CHS },
         { "3x", OPTION_VAL_3X_CHS },
         { "4x", OPTION_VAL_4_CHS },
         { "5x", OPTION_VAL_5X_CHS },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CHS },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CHS },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CHS },
         { "BLUE 03",           OPTION_VAL_BLUE_CHS },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CHS },
         { "GREEN 05",          OPTION_VAL_GREEN_CHS },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CHS },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CHS },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CHS },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CHS },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CHS },
         { "RED 11",            OPTION_VAL_RED_CHS },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CHS },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CHS },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CHS },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CHS },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CHS },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CHS },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CHS },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CHS },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CHS },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CHS },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CHS },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CHS },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CHS },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CHS },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CHS },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CHS },
         { "WHITE 28",          OPTION_VAL_WHITE_CHS },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CHS },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CHS },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CHS },
         { "BLUE 03",           OPTION_VAL_BLUE_CHS },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CHS },
         { "GREEN 05",          OPTION_VAL_GREEN_CHS },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CHS },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CHS },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CHS },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CHS },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CHS },
         { "RED 11",            OPTION_VAL_RED_CHS },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CHS },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CHS },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CHS },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CHS },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CHS },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CHS },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CHS },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CHS },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CHS },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CHS },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CHS },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CHS },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CHS },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CHS },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CHS },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CHS },
         { "WHITE 28",          OPTION_VAL_WHITE_CHS },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_CHS },
         { "20%",  OPTION_VAL_20_CHS },
         { "30%",  OPTION_VAL_30_CHS },
         { "40%",  OPTION_VAL_40_CHS },
         { "50%",  OPTION_VAL_50_CHS },
         { "60%",  OPTION_VAL_60_CHS },
         { "70%",  OPTION_VAL_70_CHS },
         { "80%",  OPTION_VAL_80_CHS },
         { "90%",  OPTION_VAL_90_CHS },
         { "100%", OPTION_VAL_100_CHS },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_chs = {
   option_cats_chs,
   option_defs_chs
};

/* RETRO_LANGUAGE_CHT */

#define CATEGORY_SYSTEM_LABEL_CHT "系統"
#define CATEGORY_SYSTEM_INFO_0_CHT NULL
#define CATEGORY_VIDEO_LABEL_CHT "顯示"
#define CATEGORY_VIDEO_INFO_0_CHT "變更解析度、渲染透明度和視覺效果的設定。"
#define CATEGORY_PERFORMANCE_LABEL_CHT "效能"
#define CATEGORY_PERFORMANCE_INFO_0_CHT NULL
#define CATEGORY_HACKS_LABEL_CHT "進階"
#define CATEGORY_HACKS_INFO_0_CHT NULL
#define CATEGORY_INPUT_LABEL_CHT "輸入"
#define CATEGORY_INPUT_INFO_0_CHT NULL
#define CATEGORY_VMU_LABEL_CHT NULL
#define CATEGORY_VMU_INFO_0_CHT NULL
#define CORE_OPTION_NAME_REGION_LABEL_CHT "地區"
#define OPTION_VAL_JAPAN_CHT NULL
#define OPTION_VAL_USA_CHT NULL
#define OPTION_VAL_EUROPE_CHT NULL
#define OPTION_VAL_DEFAULT_CHT "預設"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_CHT "語言"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_CHT NULL
#define OPTION_VAL_JAPANESE_CHT "日語"
#define OPTION_VAL_ENGLISH_CHT "英語"
#define OPTION_VAL_GERMAN_CHT "德語"
#define OPTION_VAL_FRENCH_CHT "法語"
#define OPTION_VAL_SPANISH_CHT "西班牙語"
#define OPTION_VAL_ITALIAN_CHT "意大利語"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_CHT NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_CHT NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_CHT NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_CHT NULL
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_CHT NULL
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_CHT NULL
#define CORE_OPTION_NAME_FORCE_WINCE_LABEL_CHT NULL
#define CORE_OPTION_NAME_FORCE_WINCE_INFO_0_CHT NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_CHT NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_CHT NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_CHT NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_CHT NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_CHT NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_CHT NULL
#define CORE_OPTION_NAME_UPNP_LABEL_CHT NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_CHT NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_CHT NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_CHT NULL
#define OPTION_VAL_320X240_CHT NULL
#define OPTION_VAL_640X480_CHT NULL
#define OPTION_VAL_800X600_CHT NULL
#define OPTION_VAL_960X720_CHT NULL
#define OPTION_VAL_1024X768_CHT NULL
#define OPTION_VAL_1280X960_CHT NULL
#define OPTION_VAL_1440X1080_CHT NULL
#define OPTION_VAL_1600X1200_CHT NULL
#define OPTION_VAL_1920X1440_CHT NULL
#define OPTION_VAL_2560X1920_CHT NULL
#define OPTION_VAL_2880X2160_CHT NULL
#define OPTION_VAL_3200X2400_CHT NULL
#define OPTION_VAL_3840X2880_CHT NULL
#define OPTION_VAL_4480X3360_CHT NULL
#define OPTION_VAL_5120X3840_CHT NULL
#define OPTION_VAL_5760X4320_CHT NULL
#define OPTION_VAL_6400X4800_CHT NULL
#define OPTION_VAL_7040X5280_CHT NULL
#define OPTION_VAL_7680X5760_CHT NULL
#define OPTION_VAL_8320X6240_CHT NULL
#define OPTION_VAL_8960X6720_CHT NULL
#define OPTION_VAL_9600X7200_CHT NULL
#define OPTION_VAL_10240X7680_CHT NULL
#define OPTION_VAL_10880X8160_CHT NULL
#define OPTION_VAL_11520X8640_CHT NULL
#define OPTION_VAL_12160X9120_CHT NULL
#define OPTION_VAL_12800X9600_CHT NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_CHT NULL
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_CHT NULL
#define OPTION_VAL_VGA_CHT NULL
#define OPTION_VAL_TV_RGB_CHT NULL
#define OPTION_VAL_TV_COMPOSITE_CHT NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_CHT NULL
#define OPTION_VAL_NTSC_CHT NULL
#define OPTION_VAL_PAL_CHT NULL
#define OPTION_VAL_PAL_N_CHT NULL
#define OPTION_VAL_PAL_M_CHT NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_CHT "螢幕顯示方向"
#define OPTION_VAL_HORIZONTAL_CHT NULL
#define OPTION_VAL_VERTICAL_CHT NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_CHT NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_CHT NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_CHT NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_CHT NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_CHT NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_CHT NULL
#define OPTION_VAL_512MB_CHT NULL
#define OPTION_VAL_1GB_CHT NULL
#define OPTION_VAL_2GB_CHT NULL
#define OPTION_VAL_4GB_CHT NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_CHT NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_CHT NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_CHT NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_CHT NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_CHT NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_CHT NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_CHT NULL
#define CORE_OPTION_NAME_FOG_LABEL_CHT NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_CHT NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_CHT NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_CHT NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_CHT NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_CHT "材質過濾"
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_CHT NULL
#define OPTION_VAL_1_CHT NULL
#define OPTION_VAL_2_CHT NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_CHT NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_CHT NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_CHT NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_CHT NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_CHT NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_CHT NULL
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_CHT NULL
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_CHT NULL
#define OPTION_VAL_2_O27_CHT "2倍"
#define OPTION_VAL_4_CHT "4倍"
#define OPTION_VAL_6_CHT "6倍"
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_CHT NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_CHT NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_CHT NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_CHT NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_CHT NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_CHT NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_CHT NULL
#define OPTION_VAL_SOME_CHT "正常"
#define OPTION_VAL_MORE_CHT NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_CHT NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_CHT NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_CHT NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_CHT NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_CHT NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_CHT NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_CHT NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_CHT NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_CHT NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_CHT NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_CHT NULL
#define OPTION_VAL_0_CHT NULL
#define OPTION_VAL_5_CHT NULL
#define OPTION_VAL_10_CHT NULL
#define OPTION_VAL_15_CHT NULL
#define OPTION_VAL_20_CHT NULL
#define OPTION_VAL_25_CHT NULL
#define OPTION_VAL_30_CHT NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_CHT NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_CHT NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_CHT NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_CHT NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_CHT NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_CHT NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_CHT NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_CHT NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_CHT NULL
#define OPTION_VAL_WHITE_CHT "白色"
#define OPTION_VAL_RED_CHT "紅色"
#define OPTION_VAL_GREEN_CHT "綠色"
#define OPTION_VAL_BLUE_CHT "藍色"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_CHT NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_CHT NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_CHT NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CHT NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_CHT NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_CHT NULL
#define OPTION_VAL_VMU_A1_CHT NULL
#define OPTION_VAL_ALL_VMUS_CHT NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CHT NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_CHT NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_CHT NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_CHT NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_CHT NULL
#define OPTION_VAL_UPPER_LEFT_CHT NULL
#define OPTION_VAL_UPPER_RIGHT_CHT NULL
#define OPTION_VAL_LOWER_LEFT_CHT NULL
#define OPTION_VAL_LOWER_RIGHT_CHT NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_CHT NULL
#define OPTION_VAL_1X_CHT "1倍"
#define OPTION_VAL_3X_CHT "3倍"
#define OPTION_VAL_5X_CHT "5倍"
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_CHT NULL
#define OPTION_VAL_DEFAULT_ON_00_CHT NULL
#define OPTION_VAL_DEFAULT_OFF_01_CHT NULL
#define OPTION_VAL_BLACK_02_CHT "黑色"
#define OPTION_VAL_LIGHT_BLUE_04_CHT "淺藍色"
#define OPTION_VAL_CYAN_06_CHT "青色"
#define OPTION_VAL_CYAN_BLUE_07_CHT "青藍色"
#define OPTION_VAL_LIGHT_GREEN_08_CHT "淺綠色"
#define OPTION_VAL_CYAN_GREEN_09_CHT "青綠色"
#define OPTION_VAL_LIGHT_CYAN_10_CHT "淺青色"
#define OPTION_VAL_PURPLE_12_CHT "紫色"
#define OPTION_VAL_LIGHT_PURPLE_13_CHT "淺紫色"
#define OPTION_VAL_YELLOW_14_CHT "黃色"
#define OPTION_VAL_GRAY_15_CHT "灰色"
#define OPTION_VAL_LIGHT_PURPLE_2_16_CHT "淺紫色 (2)"
#define OPTION_VAL_LIGHT_GREEN_2_17_CHT "淺綠色 (2)"
#define OPTION_VAL_LIGHT_GREEN_3_18_CHT "淺綠色 (3)"
#define OPTION_VAL_LIGHT_CYAN_2_19_CHT "淺青色 (2)"
#define OPTION_VAL_LIGHT_RED_2_20_CHT "淡紅色 (2)"
#define OPTION_VAL_MAGENTA_21_CHT "洋紅色"
#define OPTION_VAL_LIGHT_PURPLE_3_22_CHT "淺紫色 (3)"
#define OPTION_VAL_LIGHT_ORANGE_23_CHT "淺橘色"
#define OPTION_VAL_ORANGE_24_CHT "橙色"
#define OPTION_VAL_LIGHT_PURPLE_4_25_CHT "淺紫色 (4)"
#define OPTION_VAL_LIGHT_YELLOW_26_CHT "淡黃色"
#define OPTION_VAL_LIGHT_YELLOW_2_27_CHT "淡黃色 (2)"
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_CHT NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_CHT NULL
#define OPTION_VAL_40_CHT NULL
#define OPTION_VAL_50_CHT NULL
#define OPTION_VAL_60_CHT NULL
#define OPTION_VAL_70_CHT NULL
#define OPTION_VAL_80_CHT NULL
#define OPTION_VAL_90_CHT NULL
#define OPTION_VAL_100_CHT NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_CHT NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_CHT NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_CHT NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_CHT NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_CHT NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_CHT NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_CHT NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_CHT NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_CHT NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_CHT NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_CHT NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_CHT NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_CHT NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_CHT NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_CHT NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_CHT NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_CHT NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_CHT NULL

struct retro_core_option_v2_category option_cats_cht[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_CHT,
      CATEGORY_SYSTEM_INFO_0_CHT
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_CHT,
      CATEGORY_VIDEO_INFO_0_CHT
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_CHT,
      CATEGORY_PERFORMANCE_INFO_0_CHT
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_CHT,
      CATEGORY_HACKS_INFO_0_CHT
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_CHT,
      CATEGORY_INPUT_INFO_0_CHT
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_CHT,
      CATEGORY_VMU_INFO_0_CHT
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_cht[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_CHT,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_CHT },
         { "USA",     OPTION_VAL_USA_CHT },
         { "Europe",  OPTION_VAL_EUROPE_CHT },
         { "Default", OPTION_VAL_DEFAULT_CHT },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_CHT,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_CHT },
         { "English",  OPTION_VAL_ENGLISH_CHT },
         { "German",   OPTION_VAL_GERMAN_CHT },
         { "French",   OPTION_VAL_FRENCH_CHT },
         { "Spanish",  OPTION_VAL_SPANISH_CHT },
         { "Italian",  OPTION_VAL_ITALIAN_CHT },
         { "Default",  OPTION_VAL_DEFAULT_CHT },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_CHT,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_boot_to_bios",
      CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_CHT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_CHT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_force_wince",
      CORE_OPTION_NAME_FORCE_WINCE_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_FORCE_WINCE_INFO_0_CHT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_CHT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_CHT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_CHT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_CHT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_CHT,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_CHT },
         { "640x480",    OPTION_VAL_640X480_CHT },
         { "800x600",    OPTION_VAL_800X600_CHT },
         { "960x720",    OPTION_VAL_960X720_CHT },
         { "1024x768",   OPTION_VAL_1024X768_CHT },
         { "1280x960",   OPTION_VAL_1280X960_CHT },
         { "1440x1080",  OPTION_VAL_1440X1080_CHT },
         { "1600x1200",  OPTION_VAL_1600X1200_CHT },
         { "1920x1440",  OPTION_VAL_1920X1440_CHT },
         { "2560x1920",  OPTION_VAL_2560X1920_CHT },
         { "2880x2160",  OPTION_VAL_2880X2160_CHT },
         { "3200x2400",  OPTION_VAL_3200X2400_CHT },
         { "3840x2880",  OPTION_VAL_3840X2880_CHT },
         { "4480x3360",  OPTION_VAL_4480X3360_CHT },
         { "5120x3840",  OPTION_VAL_5120X3840_CHT },
         { "5760x4320",  OPTION_VAL_5760X4320_CHT },
         { "6400x4800",  OPTION_VAL_6400X4800_CHT },
         { "7040x5280",  OPTION_VAL_7040X5280_CHT },
         { "7680x5760",  OPTION_VAL_7680X5760_CHT },
         { "8320x6240",  OPTION_VAL_8320X6240_CHT },
         { "8960x6720",  OPTION_VAL_8960X6720_CHT },
         { "9600x7200",  OPTION_VAL_9600X7200_CHT },
         { "10240x7680", OPTION_VAL_10240X7680_CHT },
         { "10880x8160", OPTION_VAL_10880X8160_CHT },
         { "11520x8640", OPTION_VAL_11520X8640_CHT },
         { "12160x9120", OPTION_VAL_12160X9120_CHT },
         { "12800x9600", OPTION_VAL_12800X9600_CHT },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_CHT,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_CHT },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_CHT },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_CHT },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_CHT,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_CHT },
         { "PAL",     OPTION_VAL_PAL_CHT },
         { "PAL_N",   OPTION_VAL_PAL_N_CHT },
         { "PAL_M",   OPTION_VAL_PAL_M_CHT },
         { "Default", OPTION_VAL_DEFAULT_CHT },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_CHT,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_CHT },
         { "vertical",   OPTION_VAL_VERTICAL_CHT },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_CHT,
      NULL,
      "",
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_CHT },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_CHT },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_CHT },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_CHT,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_CHT },
         { "1GB",   OPTION_VAL_1GB_CHT },
         { "2GB",   OPTION_VAL_2GB_CHT },
         { "4GB",   OPTION_VAL_4GB_CHT },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_CHT,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_CHT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_CHT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_CHT,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_CHT,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_CHT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_CHT,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_CHT,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_CHT },
         { "1",  OPTION_VAL_1_CHT },
         { "2",  OPTION_VAL_2_CHT },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_CHT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_CHT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_CHT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_CHT,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O27_CHT },
         { "4",  OPTION_VAL_4_CHT },
         { "6",  OPTION_VAL_6_CHT },
         { NULL, NULL },
      },
      "1",
   },
   {/* TODO: needs clarification */
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_CHT,
      NULL,
      "",
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_CHT,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_CHT,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_CHT,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_CHT,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_CHT },
         { "more", OPTION_VAL_MORE_CHT },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_CHT,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_CHT,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_CHT,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_CHT,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_CHT,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: probably needs explanation */
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_CHT,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_CHT,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_CHT },
         { "5%",  OPTION_VAL_5_CHT },
         { "10%", OPTION_VAL_10_CHT },
         { "15%", OPTION_VAL_15_CHT },
         { "20%", OPTION_VAL_20_CHT },
         { "25%", OPTION_VAL_25_CHT },
         { "30%", OPTION_VAL_30_CHT },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_CHT,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_CHT },
         { "5%",  OPTION_VAL_5_CHT },
         { "10%", OPTION_VAL_10_CHT },
         { "15%", OPTION_VAL_15_CHT },
         { "20%", OPTION_VAL_20_CHT },
         { "25%", OPTION_VAL_25_CHT },
         { "30%", OPTION_VAL_30_CHT },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_CHT,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_purupuru",
      CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_CHT,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_CHT,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_CHT,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_CHT,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_CHT },
         { "Red",      OPTION_VAL_RED_CHT },
         { "Green",    OPTION_VAL_GREEN_CHT },
         { "Blue",     OPTION_VAL_BLUE_CHT },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_CHT,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_CHT },
         { "Red",      OPTION_VAL_RED_CHT },
         { "Green",    OPTION_VAL_GREEN_CHT },
         { "Blue",     OPTION_VAL_BLUE_CHT },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_CHT,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_CHT },
         { "Red",      OPTION_VAL_RED_CHT },
         { "Green",    OPTION_VAL_GREEN_CHT },
         { "Blue",     OPTION_VAL_BLUE_CHT },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_CHT,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_CHT },
         { "Red",      OPTION_VAL_RED_CHT },
         { "Green",    OPTION_VAL_GREEN_CHT },
         { "Blue",     OPTION_VAL_BLUE_CHT },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CHT,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_CHT,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_CHT,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_CHT },
         { "All VMUs", OPTION_VAL_ALL_VMUS_CHT },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CHT,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_CHT,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_CHT,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_CHT },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_CHT },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_CHT },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_CHT },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_CHT },
         { "2x", OPTION_VAL_2_O27_CHT },
         { "3x", OPTION_VAL_3X_CHT },
         { "4x", OPTION_VAL_4_CHT },
         { "5x", OPTION_VAL_5X_CHT },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CHT },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CHT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CHT },
         { "BLUE 03",           OPTION_VAL_BLUE_CHT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CHT },
         { "GREEN 05",          OPTION_VAL_GREEN_CHT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CHT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CHT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CHT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CHT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CHT },
         { "RED 11",            OPTION_VAL_RED_CHT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CHT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CHT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CHT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CHT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CHT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CHT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CHT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CHT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CHT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CHT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CHT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CHT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CHT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CHT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CHT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CHT },
         { "WHITE 28",          OPTION_VAL_WHITE_CHT },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CHT },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CHT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CHT },
         { "BLUE 03",           OPTION_VAL_BLUE_CHT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CHT },
         { "GREEN 05",          OPTION_VAL_GREEN_CHT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CHT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CHT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CHT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CHT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CHT },
         { "RED 11",            OPTION_VAL_RED_CHT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CHT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CHT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CHT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CHT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CHT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CHT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CHT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CHT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CHT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CHT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CHT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CHT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CHT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CHT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CHT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CHT },
         { "WHITE 28",          OPTION_VAL_WHITE_CHT },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_CHT },
         { "20%",  OPTION_VAL_20_CHT },
         { "30%",  OPTION_VAL_30_CHT },
         { "40%",  OPTION_VAL_40_CHT },
         { "50%",  OPTION_VAL_50_CHT },
         { "60%",  OPTION_VAL_60_CHT },
         { "70%",  OPTION_VAL_70_CHT },
         { "80%",  OPTION_VAL_80_CHT },
         { "90%",  OPTION_VAL_90_CHT },
         { "100%", OPTION_VAL_100_CHT },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_CHT },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_CHT },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_CHT },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_CHT },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_CHT },
         { "2x", OPTION_VAL_2_O27_CHT },
         { "3x", OPTION_VAL_3X_CHT },
         { "4x", OPTION_VAL_4_CHT },
         { "5x", OPTION_VAL_5X_CHT },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CHT },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CHT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CHT },
         { "BLUE 03",           OPTION_VAL_BLUE_CHT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CHT },
         { "GREEN 05",          OPTION_VAL_GREEN_CHT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CHT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CHT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CHT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CHT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CHT },
         { "RED 11",            OPTION_VAL_RED_CHT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CHT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CHT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CHT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CHT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CHT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CHT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CHT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CHT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CHT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CHT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CHT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CHT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CHT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CHT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CHT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CHT },
         { "WHITE 28",          OPTION_VAL_WHITE_CHT },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CHT },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CHT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CHT },
         { "BLUE 03",           OPTION_VAL_BLUE_CHT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CHT },
         { "GREEN 05",          OPTION_VAL_GREEN_CHT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CHT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CHT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CHT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CHT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CHT },
         { "RED 11",            OPTION_VAL_RED_CHT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CHT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CHT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CHT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CHT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CHT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CHT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CHT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CHT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CHT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CHT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CHT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CHT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CHT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CHT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CHT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CHT },
         { "WHITE 28",          OPTION_VAL_WHITE_CHT },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_CHT },
         { "20%",  OPTION_VAL_20_CHT },
         { "30%",  OPTION_VAL_30_CHT },
         { "40%",  OPTION_VAL_40_CHT },
         { "50%",  OPTION_VAL_50_CHT },
         { "60%",  OPTION_VAL_60_CHT },
         { "70%",  OPTION_VAL_70_CHT },
         { "80%",  OPTION_VAL_80_CHT },
         { "90%",  OPTION_VAL_90_CHT },
         { "100%", OPTION_VAL_100_CHT },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_CHT },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_CHT },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_CHT },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_CHT },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_CHT },
         { "2x", OPTION_VAL_2_O27_CHT },
         { "3x", OPTION_VAL_3X_CHT },
         { "4x", OPTION_VAL_4_CHT },
         { "5x", OPTION_VAL_5X_CHT },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CHT },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CHT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CHT },
         { "BLUE 03",           OPTION_VAL_BLUE_CHT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CHT },
         { "GREEN 05",          OPTION_VAL_GREEN_CHT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CHT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CHT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CHT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CHT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CHT },
         { "RED 11",            OPTION_VAL_RED_CHT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CHT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CHT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CHT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CHT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CHT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CHT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CHT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CHT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CHT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CHT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CHT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CHT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CHT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CHT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CHT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CHT },
         { "WHITE 28",          OPTION_VAL_WHITE_CHT },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CHT },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CHT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CHT },
         { "BLUE 03",           OPTION_VAL_BLUE_CHT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CHT },
         { "GREEN 05",          OPTION_VAL_GREEN_CHT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CHT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CHT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CHT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CHT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CHT },
         { "RED 11",            OPTION_VAL_RED_CHT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CHT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CHT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CHT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CHT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CHT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CHT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CHT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CHT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CHT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CHT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CHT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CHT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CHT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CHT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CHT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CHT },
         { "WHITE 28",          OPTION_VAL_WHITE_CHT },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_CHT },
         { "20%",  OPTION_VAL_20_CHT },
         { "30%",  OPTION_VAL_30_CHT },
         { "40%",  OPTION_VAL_40_CHT },
         { "50%",  OPTION_VAL_50_CHT },
         { "60%",  OPTION_VAL_60_CHT },
         { "70%",  OPTION_VAL_70_CHT },
         { "80%",  OPTION_VAL_80_CHT },
         { "90%",  OPTION_VAL_90_CHT },
         { "100%", OPTION_VAL_100_CHT },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_CHT },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_CHT },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_CHT },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_CHT },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_CHT },
         { "2x", OPTION_VAL_2_O27_CHT },
         { "3x", OPTION_VAL_3X_CHT },
         { "4x", OPTION_VAL_4_CHT },
         { "5x", OPTION_VAL_5X_CHT },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CHT },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CHT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CHT },
         { "BLUE 03",           OPTION_VAL_BLUE_CHT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CHT },
         { "GREEN 05",          OPTION_VAL_GREEN_CHT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CHT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CHT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CHT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CHT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CHT },
         { "RED 11",            OPTION_VAL_RED_CHT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CHT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CHT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CHT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CHT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CHT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CHT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CHT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CHT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CHT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CHT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CHT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CHT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CHT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CHT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CHT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CHT },
         { "WHITE 28",          OPTION_VAL_WHITE_CHT },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CHT },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CHT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CHT },
         { "BLUE 03",           OPTION_VAL_BLUE_CHT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CHT },
         { "GREEN 05",          OPTION_VAL_GREEN_CHT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CHT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CHT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CHT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CHT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CHT },
         { "RED 11",            OPTION_VAL_RED_CHT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CHT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CHT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CHT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CHT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CHT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CHT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CHT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CHT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CHT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CHT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CHT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CHT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CHT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CHT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CHT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CHT },
         { "WHITE 28",          OPTION_VAL_WHITE_CHT },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_CHT },
         { "20%",  OPTION_VAL_20_CHT },
         { "30%",  OPTION_VAL_30_CHT },
         { "40%",  OPTION_VAL_40_CHT },
         { "50%",  OPTION_VAL_50_CHT },
         { "60%",  OPTION_VAL_60_CHT },
         { "70%",  OPTION_VAL_70_CHT },
         { "80%",  OPTION_VAL_80_CHT },
         { "90%",  OPTION_VAL_90_CHT },
         { "100%", OPTION_VAL_100_CHT },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_cht = {
   option_cats_cht,
   option_defs_cht
};

/* RETRO_LANGUAGE_CS */

#define CATEGORY_SYSTEM_LABEL_CS "Systém"
#define CATEGORY_SYSTEM_INFO_0_CS "Konfigurace regionu, jazyka, systému BIOS a základního nastavení hardwaru."
#define CATEGORY_VIDEO_LABEL_CS NULL
#define CATEGORY_VIDEO_INFO_0_CS "Konfigurace rozlišení, nastavení průhlednosti a vizuálních efektů nezávislých na pořadí."
#define CATEGORY_PERFORMANCE_LABEL_CS "Výkon"
#define CATEGORY_PERFORMANCE_INFO_0_CS "Konfigurace nastavení vykreslování ve vláknech a přeskakování snímků."
#define CATEGORY_HACKS_LABEL_CS NULL
#define CATEGORY_HACKS_INFO_0_CS "Konfigurace přepisu širokoúhlého zobrazení, rychlosti načítání GD-ROM a nastavení výměny textur."
#define CATEGORY_INPUT_LABEL_CS "Vstup"
#define CATEGORY_INPUT_INFO_0_CS "Konfigurace nastavení gamepadu a světelné pistole."
#define CATEGORY_VMU_LABEL_CS "Jednotka Vizuální Paměti"
#define CATEGORY_VMU_INFO_0_CS "Konfigurace ukládacích souborů VMU pro jednotlivé hry a nastavení viditelnosti VMU na obrazovce."
#define CORE_OPTION_NAME_REGION_LABEL_CS NULL
#define OPTION_VAL_JAPAN_CS "Japonsko"
#define OPTION_VAL_USA_CS NULL
#define OPTION_VAL_EUROPE_CS "Evropa"
#define OPTION_VAL_DEFAULT_CS "Výchozí"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_CS "Jazyk"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_CS "Změní jazyk používaný systémem BIOS a všemi hrami, které obsahují více jazyků."
#define OPTION_VAL_JAPANESE_CS "Japonština"
#define OPTION_VAL_ENGLISH_CS "Angličtina"
#define OPTION_VAL_GERMAN_CS "Němčina"
#define OPTION_VAL_FRENCH_CS "Francouzština"
#define OPTION_VAL_SPANISH_CS "Španělština"
#define OPTION_VAL_ITALIAN_CS "Italský"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_CS "HLE BIOS (nutný restart)"
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_CS "Vynucení použití vysokoúrovňové emulace systému BIOS."
#define CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_CS "Nabootovat Bios (Nutný Restart)"
#define CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_CS "Spustit přímo do menu Dreamcast BIOS."
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_CS "Povolit DSP"
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_CS "Umožňuje emulaci zvukového DSP (digitálního signálového procesoru) konzole Dreamcast. Zlepšuje přesnost generovaného zvuku, ale zvyšuje nároky na výkon."
#define CORE_OPTION_NAME_FORCE_WINCE_LABEL_CS "Vynutit Režim Windows CE"
#define CORE_OPTION_NAME_FORCE_WINCE_INFO_0_CS "Povolte úplnou emulaci jednotky MMU (Memory Management Unit) a další nastavení pro hry pro Windows CE."
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_CS "Povolení servisních tlačítek NAOMI"
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_CS "Povolí tlačítko SERVIS pro NAOMI, abyste mohli vstoupit do nastavení skříně."
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_CS "Nastavení her NAOMI na volné hraní"
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_CS "Upravte nastavení mincí ve hře na volnou hru."
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_CS "Emulace širokopásmového adaptéru"
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_CS "Místo modemu emulujte širokopásmový adaptér ethernetu. (nutný restart)"
#define CORE_OPTION_NAME_UPNP_LABEL_CS "Povolit UPnP"
#define CORE_OPTION_NAME_UPNP_INFO_0_CS "Pomocí UPnP můžete automaticky nakonfigurovat svůj internetový směrovač pro online hry."
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_CS "Vnitřní Rozlišení"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_CS "Úprava Rozlišení Vykreslování."
#define OPTION_VAL_320X240_CS NULL
#define OPTION_VAL_640X480_CS NULL
#define OPTION_VAL_800X600_CS NULL
#define OPTION_VAL_960X720_CS NULL
#define OPTION_VAL_1024X768_CS NULL
#define OPTION_VAL_1280X960_CS NULL
#define OPTION_VAL_1440X1080_CS NULL
#define OPTION_VAL_1600X1200_CS NULL
#define OPTION_VAL_1920X1440_CS NULL
#define OPTION_VAL_2560X1920_CS NULL
#define OPTION_VAL_2880X2160_CS NULL
#define OPTION_VAL_3200X2400_CS NULL
#define OPTION_VAL_3840X2880_CS NULL
#define OPTION_VAL_4480X3360_CS NULL
#define OPTION_VAL_5120X3840_CS NULL
#define OPTION_VAL_5760X4320_CS NULL
#define OPTION_VAL_6400X4800_CS NULL
#define OPTION_VAL_7040X5280_CS NULL
#define OPTION_VAL_7680X5760_CS NULL
#define OPTION_VAL_8320X6240_CS NULL
#define OPTION_VAL_8960X6720_CS NULL
#define OPTION_VAL_9600X7200_CS NULL
#define OPTION_VAL_10240X7680_CS NULL
#define OPTION_VAL_10880X8160_CS NULL
#define OPTION_VAL_11520X8640_CS NULL
#define OPTION_VAL_12160X9120_CS NULL
#define OPTION_VAL_12800X9600_CS NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_CS "Typ Kabelu"
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_CS "Typ výstupního signálu. Nejvíce podporovaný je 'TV (Kompozitní)'."
#define OPTION_VAL_VGA_CS NULL
#define OPTION_VAL_TV_RGB_CS NULL
#define OPTION_VAL_TV_COMPOSITE_CS "TV (Kompozitní)"
#define CORE_OPTION_NAME_BROADCAST_LABEL_CS "Standardní Vysílání"
#define OPTION_VAL_NTSC_CS NULL
#define OPTION_VAL_PAL_CS "PAL (Svět)"
#define OPTION_VAL_PAL_N_CS NULL
#define OPTION_VAL_PAL_M_CS "PAL-M (Brazílie)"
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_CS "Orientace Obrazovky"
#define OPTION_VAL_HORIZONTAL_CS "Horizontální"
#define OPTION_VAL_VERTICAL_CS "Vertikální"
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_CS "Třídění Alpha"
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_CS "Per-Strip (rychlé, nejméně přesné)"
#define OPTION_VAL_PER_TRIANGLE_NORMAL_CS "Per-Triangle (normální)"
#define OPTION_VAL_PER_PIXEL_ACCURATE_CS "Per-Pixel (přesné, ale nejpomalejší)"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_CS "Velikost akumulační vyrovnávací paměti pixelů"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_CS "Pro správný výstup ve vyšších rozlišeních mohou být nutné vyšší hodnoty."
#define OPTION_VAL_512MB_CS NULL
#define OPTION_VAL_1GB_CS NULL
#define OPTION_VAL_2GB_CS NULL
#define OPTION_VAL_4GB_CS NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_CS "Maximální počet průhledných vrstev"
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_CS "U složitých scén mohou být nutné vyšší hodnoty."
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_CS "Úplná emulace framebufferu"
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_CS "Povolení úplné emulace framebufferu v paměti VRAM. To je užitečné pro hry, které přímo čtou nebo zapisují framebuffer do VRAM. Pokud je tato funkce povolena, interní rozlišení je vynuceno na 640x480 a může dojít k výraznému snížení výkonu."
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_CS "Povolení Vyrovnávací Paměti RTT (Render To Texture)"
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_CS "Kopírování vykreslených textur z GPU zpět do VRAM. Tato možnost je obvykle povolena u her, které ji vyžadují. Pokud je povolena, je vypnuto zvyšování rozlišení vykreslovaných textur a může dojít ke snížení výkonu."
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_CS "Mipmapování"
#define CORE_OPTION_NAME_FOG_LABEL_CS "Efekty Mlhy"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_CS "Modifikátor Hlasitosti"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_CS "Funkce grafického procesoru Dreamcast, která se obvykle používá ve hrách ke kreslení stínů objektů. Tato funkce by měla být normálně povolena - dopad na výkon je obvykle minimální až zanedbatelný."
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_CS "Anizotropní Filtrování"
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_CS "Vylepšete kvalitu textur na površích, které jsou vůči kameře pod šikmým úhlem pohledu. Vyšší hodnoty jsou náročnější na grafický procesor. Změny tohoto nastavení se uplatní až po restartu."
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_CS "Filtrování Textur"
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_CS "Režim filtrování textur, který se má použít. Tento parametr lze použít k vynucení určitého režimu filtrování textur na všech texturách, abyste dosáhli ostřejšího (nebo hladšího) vzhledu než ve výchozím nastavení. Jiné hodnoty než výchozí mohou způsobit různé problémy s vykreslováním. Změny tohoto nastavení se uplatní až po restartu."
#define OPTION_VAL_1_CS "Vynucený Nearest-Neighbor"
#define OPTION_VAL_2_CS "Vynucený lineární"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_CS "Výměna Zpožděných Snímků"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_CS "Je užitečné vyhnout se blikajícím obrazovkám nebo poruchovým videím. Nedoporučuje se na pomalých platformách."
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_CS "Detekce změn snímkové frekvence"
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_CS "Upozornit frontend při změně interní snímkové frekvence (např. z 60 fps na 30 fps). Zlepšuje snímkovou frekvenci ve hrách, které běží s uzamčenou snímkovou frekvencí 30 fps nebo 20 fps, ale měl by být vypnut pro hry s odemčenou (nestabilní) snímkovou frekvencí (např. Ecco the Dolphin, Unreal Tournament). Poznámka: Není k dispozici, pokud je povoleno 'Automatické přeskakování snímků'."
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_CS "Postprocesní Filtr PowerVR2"
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_CS "Následné zpracování vykresleného obrazu pro simulaci efektů specifických pro grafický procesor PowerVR2 a analogové video signály."
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_CS "Vylepšení Textur (xBRZ)"
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_CS "Vylepšete ručně kreslenou 2D pixelovou grafiku. Mělo by se používat pouze ve 2D pixelových hrách."
#define OPTION_VAL_2_O27_CS NULL
#define OPTION_VAL_4_CS NULL
#define OPTION_VAL_6_CS NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_CS "Zvýšení Rozlišení Textur Max. Filtrovaná Velikost"
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_CS "Nativní interpolace hloubky"
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_CS "Pomáhá řešit problémy s poškozením textur a hloubkou na grafických procesorech AMD. V některých případech může pomoci i GPU Intel."
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_CS "Vykreslování Vláken"
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_CS "Spouští GPU a CPU v různých vláknech. Důrazně doporučujeme."
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_CS "Automatické Přeskočení Snímků"
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_CS "Automatické přeskakování snímků při pomalém běhu emulátoru. Poznámka: Toto nastavení platí pouze tehdy, když je povoleno 'Vláknové Vykreslování'."
#define OPTION_VAL_SOME_CS "Normální"
#define OPTION_VAL_MORE_CS "Maximální"
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_CS "Přeskočení Snímků"
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_CS "Nastaví počet snímků, které se mají přeskočit mezi jednotlivými zobrazenými snímky."
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_CS "Širokoúhlé Cheaty (Nutný Restart)"
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_CS "Aktivuje cheaty, které umožňují zobrazení některých her v širokoúhlém formátu."
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_CS "Hack Širokoúhleho Obrazu"
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_CS "Kreslení geometrie mimo běžný poměr stran 4:3. Může způsobit grafické závady v odhalených oblastech."
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_CS "Rychlé Načítání GD-ROM (nepřesné)"
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_CS "Zrychluje načítání GD-ROM."
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_CS "Načíst Vlastní Textury"
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_CS "Skladiště Textur"
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_CS "Analogova Páčka Mrtvé Zóny"
#define OPTION_VAL_0_CS NULL
#define OPTION_VAL_5_CS NULL
#define OPTION_VAL_10_CS NULL
#define OPTION_VAL_15_CS NULL
#define OPTION_VAL_20_CS NULL
#define OPTION_VAL_25_CS NULL
#define OPTION_VAL_30_CS NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_CS "Spouštěč Mrtvé Zóny"
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_CS "Digitální Spouštěče"
#define CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_CS "Purupuru Pack/Vibrační Balíček"
#define CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_CS "Povoluje silovou zpětnou vazbu regulátoru."
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_CS "Digitální výstupy pro vysílání"
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_CS "Vysílání digitálních výstupů a stavu vynucene zpětné vazby na portu TCP 8000. Kompatibilní s volbou MAME \"-výstupní síť\"."
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_CS "Zobrazit Nastavení Světelné Pistole"
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_CS "Povolení konfigurace možností zobrazení zaměřovacího kříže světelné pistole. POZNÁMKA: Aby se toto nastavení projevilo, může být nutné přepnout rychlou nabídku."
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_CS "Zobrazení Kříže Zbraně 1"
#define OPTION_VAL_WHITE_CS "Bílá"
#define OPTION_VAL_RED_CS "Červený"
#define OPTION_VAL_GREEN_CS "Zelený"
#define OPTION_VAL_BLUE_CS "Modrý"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_CS "Zobrazení Kříže Zbraně 2"
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_CS "Zobrazení Kříže Zbraně 3"
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_CS "Zobrazení Kříže Zbraně 4"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CS "Vizuální Paměťové Jednotky/Systémy pro Jednotlivé Hry (VMU)"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_CS "VMUs Na-Hru"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_CS "Pokud je tato funkce vypnuta, všechny hry sdílejí 4 ukládací soubory VMU (A1, B1, C1, D1) umístěné v systémovém adresáři aplikace RetroArch. Nastavení 'VMU A1' vytvoří v adresáři ukládání aplikace RetroArch pro každou spuštěnou hru jedinečný soubor VMU 'A1'. Nastavení 'All VMUs' Všechny VMU vytvoří 4 jedinečné soubory VMU (A1, B1, C1, D1) pro každou spuštěnou hru."
#define OPTION_VAL_VMU_A1_CS NULL
#define OPTION_VAL_ALL_VMUS_CS NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CS "Zobrazit Nastavení Displeje Vizuální Paměťové Jednotky/Systému (VMU)"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_CS "Zobrazit Nastavení VMU Displeje"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_CS "Umožňuje konfiguraci viditelnosti, velikosti, polohy a barvy obrazovky LCD emulovaného VMU. POZNÁMKA: Může být nutné přepnout Rychlé Menu, aby se toto nastavení projevilo."
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_CS "Displej Obrazovky VMU 1"
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_CS "Pozice Obrazovky VMU 1"
#define OPTION_VAL_UPPER_LEFT_CS "Vlevo Nahoře"
#define OPTION_VAL_UPPER_RIGHT_CS "Vpravo Nahoře"
#define OPTION_VAL_LOWER_LEFT_CS "Vlevo Dole"
#define OPTION_VAL_LOWER_RIGHT_CS "Vpravo Dole"
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_CS "Velikost Obrazovky VMU 1"
#define OPTION_VAL_1X_CS NULL
#define OPTION_VAL_3X_CS NULL
#define OPTION_VAL_5X_CS NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_CS "Zapnout Barvu Pixelu Obrazovky VMU 1"
#define OPTION_VAL_DEFAULT_ON_00_CS "Výchozí ON"
#define OPTION_VAL_DEFAULT_OFF_01_CS "Výchozí OFF"
#define OPTION_VAL_BLACK_02_CS "Černá"
#define OPTION_VAL_LIGHT_BLUE_04_CS "Světle Modrá"
#define OPTION_VAL_CYAN_06_CS "Azurová"
#define OPTION_VAL_CYAN_BLUE_07_CS "Azurově Modrá"
#define OPTION_VAL_LIGHT_GREEN_08_CS "Světle Zelená"
#define OPTION_VAL_CYAN_GREEN_09_CS "Azurově Zelená"
#define OPTION_VAL_LIGHT_CYAN_10_CS "Světle Azurová"
#define OPTION_VAL_PURPLE_12_CS "Fialová"
#define OPTION_VAL_LIGHT_PURPLE_13_CS "Světle Fialová"
#define OPTION_VAL_YELLOW_14_CS "Žlutá"
#define OPTION_VAL_GRAY_15_CS "Šedá"
#define OPTION_VAL_LIGHT_PURPLE_2_16_CS "Světle Fialová (2)"
#define OPTION_VAL_LIGHT_GREEN_2_17_CS "Světle Zelená (2)"
#define OPTION_VAL_LIGHT_GREEN_3_18_CS "Světle Zelená (3)"
#define OPTION_VAL_LIGHT_CYAN_2_19_CS "Světle Azurová (2)"
#define OPTION_VAL_LIGHT_RED_2_20_CS "Světle Červená (2)"
#define OPTION_VAL_MAGENTA_21_CS "Purpurová"
#define OPTION_VAL_LIGHT_PURPLE_3_22_CS "Světle Fialová (3)"
#define OPTION_VAL_LIGHT_ORANGE_23_CS "Světle Oranžová"
#define OPTION_VAL_ORANGE_24_CS "Oranžová"
#define OPTION_VAL_LIGHT_PURPLE_4_25_CS "Světle Fialová (4)"
#define OPTION_VAL_LIGHT_YELLOW_26_CS "Světle Žlutá"
#define OPTION_VAL_LIGHT_YELLOW_2_27_CS "Světle Žlutá (2)"
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_CS "Vypnout Barvu Pixelu Obrazovky VMU 1"
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_CS "Průhlednost Obrazovky VMU 1"
#define OPTION_VAL_40_CS NULL
#define OPTION_VAL_50_CS NULL
#define OPTION_VAL_60_CS NULL
#define OPTION_VAL_70_CS NULL
#define OPTION_VAL_80_CS NULL
#define OPTION_VAL_90_CS NULL
#define OPTION_VAL_100_CS NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_CS "Displej Obrazovky VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_CS "Pozice Obrazovky VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_CS "Velikost Obrazovky VMU 2"
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_CS "Zapnout Barvu Pixelu Obrazovky VMU 2"
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_CS "Vypnout Barvu Pixelu Obrazovky VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_CS "Průhlednost Obrazovky VMU 2"
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_CS "Displej Obrazovky VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_CS "Pozice Obrazovky VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_CS "Velikost Obrazovky VMU 3"
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_CS "Zapnout Barvu Pixelu Obrazovky VMU 3"
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_CS "Vypnout Barvu Pixelu Obrazovky VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_CS "Průhlednost Obrazovky VMU 3"
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_CS "Displej Obrazovky VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_CS "Pozice Obrazovky VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_CS "Velikost Obrazovky VMU 4"
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_CS "Zapnout Barvu Pixelu Obrazovky VMU 4"
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_CS "Vypnout Barvu Pixelu Obrazovky VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_CS "Průhlednost Obrazovky VMU 4"

struct retro_core_option_v2_category option_cats_cs[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_CS,
      CATEGORY_SYSTEM_INFO_0_CS
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_CS,
      CATEGORY_VIDEO_INFO_0_CS
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_CS,
      CATEGORY_PERFORMANCE_INFO_0_CS
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_CS,
      CATEGORY_HACKS_INFO_0_CS
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_CS,
      CATEGORY_INPUT_INFO_0_CS
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_CS,
      CATEGORY_VMU_INFO_0_CS
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_cs[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_CS,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_CS },
         { "USA",     OPTION_VAL_USA_CS },
         { "Europe",  OPTION_VAL_EUROPE_CS },
         { "Default", OPTION_VAL_DEFAULT_CS },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_CS,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_CS },
         { "English",  OPTION_VAL_ENGLISH_CS },
         { "German",   OPTION_VAL_GERMAN_CS },
         { "French",   OPTION_VAL_FRENCH_CS },
         { "Spanish",  OPTION_VAL_SPANISH_CS },
         { "Italian",  OPTION_VAL_ITALIAN_CS },
         { "Default",  OPTION_VAL_DEFAULT_CS },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_CS,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_boot_to_bios",
      CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_CS,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_CS,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_force_wince",
      CORE_OPTION_NAME_FORCE_WINCE_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_FORCE_WINCE_INFO_0_CS,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_CS,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_CS,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_CS,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_CS,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_CS,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_CS },
         { "640x480",    OPTION_VAL_640X480_CS },
         { "800x600",    OPTION_VAL_800X600_CS },
         { "960x720",    OPTION_VAL_960X720_CS },
         { "1024x768",   OPTION_VAL_1024X768_CS },
         { "1280x960",   OPTION_VAL_1280X960_CS },
         { "1440x1080",  OPTION_VAL_1440X1080_CS },
         { "1600x1200",  OPTION_VAL_1600X1200_CS },
         { "1920x1440",  OPTION_VAL_1920X1440_CS },
         { "2560x1920",  OPTION_VAL_2560X1920_CS },
         { "2880x2160",  OPTION_VAL_2880X2160_CS },
         { "3200x2400",  OPTION_VAL_3200X2400_CS },
         { "3840x2880",  OPTION_VAL_3840X2880_CS },
         { "4480x3360",  OPTION_VAL_4480X3360_CS },
         { "5120x3840",  OPTION_VAL_5120X3840_CS },
         { "5760x4320",  OPTION_VAL_5760X4320_CS },
         { "6400x4800",  OPTION_VAL_6400X4800_CS },
         { "7040x5280",  OPTION_VAL_7040X5280_CS },
         { "7680x5760",  OPTION_VAL_7680X5760_CS },
         { "8320x6240",  OPTION_VAL_8320X6240_CS },
         { "8960x6720",  OPTION_VAL_8960X6720_CS },
         { "9600x7200",  OPTION_VAL_9600X7200_CS },
         { "10240x7680", OPTION_VAL_10240X7680_CS },
         { "10880x8160", OPTION_VAL_10880X8160_CS },
         { "11520x8640", OPTION_VAL_11520X8640_CS },
         { "12160x9120", OPTION_VAL_12160X9120_CS },
         { "12800x9600", OPTION_VAL_12800X9600_CS },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_CS,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_CS },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_CS },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_CS },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_CS,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_CS },
         { "PAL",     OPTION_VAL_PAL_CS },
         { "PAL_N",   OPTION_VAL_PAL_N_CS },
         { "PAL_M",   OPTION_VAL_PAL_M_CS },
         { "Default", OPTION_VAL_DEFAULT_CS },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_CS,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_CS },
         { "vertical",   OPTION_VAL_VERTICAL_CS },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_CS,
      NULL,
      "",
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_CS },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_CS },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_CS },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_CS,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_CS },
         { "1GB",   OPTION_VAL_1GB_CS },
         { "2GB",   OPTION_VAL_2GB_CS },
         { "4GB",   OPTION_VAL_4GB_CS },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_CS,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_CS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_CS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_CS,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_CS,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_CS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_CS,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_CS,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_CS },
         { "1",  OPTION_VAL_1_CS },
         { "2",  OPTION_VAL_2_CS },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_CS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_CS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_CS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_CS,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O27_CS },
         { "4",  OPTION_VAL_4_CS },
         { "6",  OPTION_VAL_6_CS },
         { NULL, NULL },
      },
      "1",
   },
   {/* TODO: needs clarification */
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_CS,
      NULL,
      "",
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_CS,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_CS,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_CS,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_CS,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_CS },
         { "more", OPTION_VAL_MORE_CS },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_CS,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_CS,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_CS,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_CS,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_CS,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: probably needs explanation */
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_CS,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_CS,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_CS },
         { "5%",  OPTION_VAL_5_CS },
         { "10%", OPTION_VAL_10_CS },
         { "15%", OPTION_VAL_15_CS },
         { "20%", OPTION_VAL_20_CS },
         { "25%", OPTION_VAL_25_CS },
         { "30%", OPTION_VAL_30_CS },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_CS,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_CS },
         { "5%",  OPTION_VAL_5_CS },
         { "10%", OPTION_VAL_10_CS },
         { "15%", OPTION_VAL_15_CS },
         { "20%", OPTION_VAL_20_CS },
         { "25%", OPTION_VAL_25_CS },
         { "30%", OPTION_VAL_30_CS },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_CS,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_purupuru",
      CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_CS,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_CS,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_CS,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_CS,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_CS },
         { "Red",      OPTION_VAL_RED_CS },
         { "Green",    OPTION_VAL_GREEN_CS },
         { "Blue",     OPTION_VAL_BLUE_CS },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_CS,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_CS },
         { "Red",      OPTION_VAL_RED_CS },
         { "Green",    OPTION_VAL_GREEN_CS },
         { "Blue",     OPTION_VAL_BLUE_CS },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_CS,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_CS },
         { "Red",      OPTION_VAL_RED_CS },
         { "Green",    OPTION_VAL_GREEN_CS },
         { "Blue",     OPTION_VAL_BLUE_CS },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_CS,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_CS },
         { "Red",      OPTION_VAL_RED_CS },
         { "Green",    OPTION_VAL_GREEN_CS },
         { "Blue",     OPTION_VAL_BLUE_CS },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CS,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_CS,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_CS,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_CS },
         { "All VMUs", OPTION_VAL_ALL_VMUS_CS },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CS,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_CS,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_CS,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_CS },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_CS },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_CS },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_CS },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_CS },
         { "2x", OPTION_VAL_2_O27_CS },
         { "3x", OPTION_VAL_3X_CS },
         { "4x", OPTION_VAL_4_CS },
         { "5x", OPTION_VAL_5X_CS },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CS },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CS },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CS },
         { "BLUE 03",           OPTION_VAL_BLUE_CS },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CS },
         { "GREEN 05",          OPTION_VAL_GREEN_CS },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CS },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CS },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CS },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CS },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CS },
         { "RED 11",            OPTION_VAL_RED_CS },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CS },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CS },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CS },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CS },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CS },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CS },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CS },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CS },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CS },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CS },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CS },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CS },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CS },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CS },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CS },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CS },
         { "WHITE 28",          OPTION_VAL_WHITE_CS },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CS },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CS },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CS },
         { "BLUE 03",           OPTION_VAL_BLUE_CS },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CS },
         { "GREEN 05",          OPTION_VAL_GREEN_CS },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CS },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CS },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CS },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CS },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CS },
         { "RED 11",            OPTION_VAL_RED_CS },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CS },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CS },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CS },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CS },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CS },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CS },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CS },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CS },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CS },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CS },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CS },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CS },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CS },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CS },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CS },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CS },
         { "WHITE 28",          OPTION_VAL_WHITE_CS },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_CS },
         { "20%",  OPTION_VAL_20_CS },
         { "30%",  OPTION_VAL_30_CS },
         { "40%",  OPTION_VAL_40_CS },
         { "50%",  OPTION_VAL_50_CS },
         { "60%",  OPTION_VAL_60_CS },
         { "70%",  OPTION_VAL_70_CS },
         { "80%",  OPTION_VAL_80_CS },
         { "90%",  OPTION_VAL_90_CS },
         { "100%", OPTION_VAL_100_CS },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_CS },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_CS },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_CS },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_CS },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_CS },
         { "2x", OPTION_VAL_2_O27_CS },
         { "3x", OPTION_VAL_3X_CS },
         { "4x", OPTION_VAL_4_CS },
         { "5x", OPTION_VAL_5X_CS },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CS },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CS },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CS },
         { "BLUE 03",           OPTION_VAL_BLUE_CS },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CS },
         { "GREEN 05",          OPTION_VAL_GREEN_CS },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CS },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CS },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CS },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CS },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CS },
         { "RED 11",            OPTION_VAL_RED_CS },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CS },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CS },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CS },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CS },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CS },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CS },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CS },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CS },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CS },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CS },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CS },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CS },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CS },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CS },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CS },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CS },
         { "WHITE 28",          OPTION_VAL_WHITE_CS },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CS },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CS },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CS },
         { "BLUE 03",           OPTION_VAL_BLUE_CS },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CS },
         { "GREEN 05",          OPTION_VAL_GREEN_CS },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CS },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CS },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CS },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CS },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CS },
         { "RED 11",            OPTION_VAL_RED_CS },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CS },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CS },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CS },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CS },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CS },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CS },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CS },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CS },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CS },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CS },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CS },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CS },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CS },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CS },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CS },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CS },
         { "WHITE 28",          OPTION_VAL_WHITE_CS },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_CS },
         { "20%",  OPTION_VAL_20_CS },
         { "30%",  OPTION_VAL_30_CS },
         { "40%",  OPTION_VAL_40_CS },
         { "50%",  OPTION_VAL_50_CS },
         { "60%",  OPTION_VAL_60_CS },
         { "70%",  OPTION_VAL_70_CS },
         { "80%",  OPTION_VAL_80_CS },
         { "90%",  OPTION_VAL_90_CS },
         { "100%", OPTION_VAL_100_CS },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_CS },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_CS },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_CS },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_CS },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_CS },
         { "2x", OPTION_VAL_2_O27_CS },
         { "3x", OPTION_VAL_3X_CS },
         { "4x", OPTION_VAL_4_CS },
         { "5x", OPTION_VAL_5X_CS },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CS },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CS },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CS },
         { "BLUE 03",           OPTION_VAL_BLUE_CS },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CS },
         { "GREEN 05",          OPTION_VAL_GREEN_CS },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CS },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CS },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CS },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CS },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CS },
         { "RED 11",            OPTION_VAL_RED_CS },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CS },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CS },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CS },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CS },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CS },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CS },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CS },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CS },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CS },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CS },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CS },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CS },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CS },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CS },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CS },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CS },
         { "WHITE 28",          OPTION_VAL_WHITE_CS },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CS },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CS },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CS },
         { "BLUE 03",           OPTION_VAL_BLUE_CS },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CS },
         { "GREEN 05",          OPTION_VAL_GREEN_CS },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CS },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CS },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CS },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CS },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CS },
         { "RED 11",            OPTION_VAL_RED_CS },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CS },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CS },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CS },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CS },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CS },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CS },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CS },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CS },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CS },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CS },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CS },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CS },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CS },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CS },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CS },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CS },
         { "WHITE 28",          OPTION_VAL_WHITE_CS },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_CS },
         { "20%",  OPTION_VAL_20_CS },
         { "30%",  OPTION_VAL_30_CS },
         { "40%",  OPTION_VAL_40_CS },
         { "50%",  OPTION_VAL_50_CS },
         { "60%",  OPTION_VAL_60_CS },
         { "70%",  OPTION_VAL_70_CS },
         { "80%",  OPTION_VAL_80_CS },
         { "90%",  OPTION_VAL_90_CS },
         { "100%", OPTION_VAL_100_CS },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_CS },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_CS },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_CS },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_CS },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_CS },
         { "2x", OPTION_VAL_2_O27_CS },
         { "3x", OPTION_VAL_3X_CS },
         { "4x", OPTION_VAL_4_CS },
         { "5x", OPTION_VAL_5X_CS },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CS },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CS },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CS },
         { "BLUE 03",           OPTION_VAL_BLUE_CS },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CS },
         { "GREEN 05",          OPTION_VAL_GREEN_CS },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CS },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CS },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CS },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CS },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CS },
         { "RED 11",            OPTION_VAL_RED_CS },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CS },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CS },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CS },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CS },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CS },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CS },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CS },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CS },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CS },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CS },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CS },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CS },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CS },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CS },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CS },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CS },
         { "WHITE 28",          OPTION_VAL_WHITE_CS },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CS },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CS },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CS },
         { "BLUE 03",           OPTION_VAL_BLUE_CS },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CS },
         { "GREEN 05",          OPTION_VAL_GREEN_CS },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CS },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CS },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CS },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CS },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CS },
         { "RED 11",            OPTION_VAL_RED_CS },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CS },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CS },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CS },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CS },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CS },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CS },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CS },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CS },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CS },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CS },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CS },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CS },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CS },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CS },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CS },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CS },
         { "WHITE 28",          OPTION_VAL_WHITE_CS },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_CS },
         { "20%",  OPTION_VAL_20_CS },
         { "30%",  OPTION_VAL_30_CS },
         { "40%",  OPTION_VAL_40_CS },
         { "50%",  OPTION_VAL_50_CS },
         { "60%",  OPTION_VAL_60_CS },
         { "70%",  OPTION_VAL_70_CS },
         { "80%",  OPTION_VAL_80_CS },
         { "90%",  OPTION_VAL_90_CS },
         { "100%", OPTION_VAL_100_CS },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_cs = {
   option_cats_cs,
   option_defs_cs
};

/* RETRO_LANGUAGE_CY */

#define CATEGORY_SYSTEM_LABEL_CY NULL
#define CATEGORY_SYSTEM_INFO_0_CY NULL
#define CATEGORY_VIDEO_LABEL_CY NULL
#define CATEGORY_VIDEO_INFO_0_CY NULL
#define CATEGORY_PERFORMANCE_LABEL_CY NULL
#define CATEGORY_PERFORMANCE_INFO_0_CY NULL
#define CATEGORY_HACKS_LABEL_CY NULL
#define CATEGORY_HACKS_INFO_0_CY NULL
#define CATEGORY_INPUT_LABEL_CY "Mewnbwn"
#define CATEGORY_INPUT_INFO_0_CY NULL
#define CATEGORY_VMU_LABEL_CY NULL
#define CATEGORY_VMU_INFO_0_CY NULL
#define CORE_OPTION_NAME_REGION_LABEL_CY NULL
#define OPTION_VAL_JAPAN_CY NULL
#define OPTION_VAL_USA_CY NULL
#define OPTION_VAL_EUROPE_CY NULL
#define OPTION_VAL_DEFAULT_CY NULL
#define CORE_OPTION_NAME_LANGUAGE_LABEL_CY NULL
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_CY NULL
#define OPTION_VAL_JAPANESE_CY NULL
#define OPTION_VAL_ENGLISH_CY NULL
#define OPTION_VAL_GERMAN_CY NULL
#define OPTION_VAL_FRENCH_CY NULL
#define OPTION_VAL_SPANISH_CY NULL
#define OPTION_VAL_ITALIAN_CY NULL
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_CY NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_CY NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_CY NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_CY NULL
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_CY NULL
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_CY NULL
#define CORE_OPTION_NAME_FORCE_WINCE_LABEL_CY NULL
#define CORE_OPTION_NAME_FORCE_WINCE_INFO_0_CY NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_CY NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_CY NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_CY NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_CY NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_CY NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_CY NULL
#define CORE_OPTION_NAME_UPNP_LABEL_CY NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_CY NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_CY NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_CY NULL
#define OPTION_VAL_320X240_CY NULL
#define OPTION_VAL_640X480_CY NULL
#define OPTION_VAL_800X600_CY NULL
#define OPTION_VAL_960X720_CY NULL
#define OPTION_VAL_1024X768_CY NULL
#define OPTION_VAL_1280X960_CY NULL
#define OPTION_VAL_1440X1080_CY NULL
#define OPTION_VAL_1600X1200_CY NULL
#define OPTION_VAL_1920X1440_CY NULL
#define OPTION_VAL_2560X1920_CY NULL
#define OPTION_VAL_2880X2160_CY NULL
#define OPTION_VAL_3200X2400_CY NULL
#define OPTION_VAL_3840X2880_CY NULL
#define OPTION_VAL_4480X3360_CY NULL
#define OPTION_VAL_5120X3840_CY NULL
#define OPTION_VAL_5760X4320_CY NULL
#define OPTION_VAL_6400X4800_CY NULL
#define OPTION_VAL_7040X5280_CY NULL
#define OPTION_VAL_7680X5760_CY NULL
#define OPTION_VAL_8320X6240_CY NULL
#define OPTION_VAL_8960X6720_CY NULL
#define OPTION_VAL_9600X7200_CY NULL
#define OPTION_VAL_10240X7680_CY NULL
#define OPTION_VAL_10880X8160_CY NULL
#define OPTION_VAL_11520X8640_CY NULL
#define OPTION_VAL_12160X9120_CY NULL
#define OPTION_VAL_12800X9600_CY NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_CY NULL
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_CY NULL
#define OPTION_VAL_VGA_CY NULL
#define OPTION_VAL_TV_RGB_CY NULL
#define OPTION_VAL_TV_COMPOSITE_CY NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_CY NULL
#define OPTION_VAL_NTSC_CY NULL
#define OPTION_VAL_PAL_CY NULL
#define OPTION_VAL_PAL_N_CY NULL
#define OPTION_VAL_PAL_M_CY NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_CY NULL
#define OPTION_VAL_HORIZONTAL_CY NULL
#define OPTION_VAL_VERTICAL_CY NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_CY NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_CY NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_CY NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_CY NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_CY NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_CY NULL
#define OPTION_VAL_512MB_CY NULL
#define OPTION_VAL_1GB_CY NULL
#define OPTION_VAL_2GB_CY NULL
#define OPTION_VAL_4GB_CY NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_CY NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_CY NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_CY NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_CY NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_CY NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_CY NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_CY NULL
#define CORE_OPTION_NAME_FOG_LABEL_CY NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_CY NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_CY NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_CY NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_CY NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_CY NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_CY NULL
#define OPTION_VAL_1_CY NULL
#define OPTION_VAL_2_CY NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_CY NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_CY NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_CY NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_CY NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_CY NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_CY NULL
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_CY NULL
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_CY NULL
#define OPTION_VAL_2_O27_CY NULL
#define OPTION_VAL_4_CY NULL
#define OPTION_VAL_6_CY NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_CY NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_CY NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_CY NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_CY NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_CY NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_CY NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_CY NULL
#define OPTION_VAL_SOME_CY NULL
#define OPTION_VAL_MORE_CY NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_CY NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_CY NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_CY NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_CY NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_CY NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_CY NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_CY NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_CY NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_CY NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_CY NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_CY NULL
#define OPTION_VAL_0_CY NULL
#define OPTION_VAL_5_CY NULL
#define OPTION_VAL_10_CY NULL
#define OPTION_VAL_15_CY NULL
#define OPTION_VAL_20_CY NULL
#define OPTION_VAL_25_CY NULL
#define OPTION_VAL_30_CY NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_CY NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_CY NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_CY NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_CY NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_CY NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_CY NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_CY NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_CY NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_CY NULL
#define OPTION_VAL_WHITE_CY NULL
#define OPTION_VAL_RED_CY "Coch"
#define OPTION_VAL_GREEN_CY NULL
#define OPTION_VAL_BLUE_CY NULL
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_CY NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_CY NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_CY NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CY NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_CY NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_CY NULL
#define OPTION_VAL_VMU_A1_CY NULL
#define OPTION_VAL_ALL_VMUS_CY NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CY NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_CY NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_CY NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_CY NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_CY NULL
#define OPTION_VAL_UPPER_LEFT_CY NULL
#define OPTION_VAL_UPPER_RIGHT_CY NULL
#define OPTION_VAL_LOWER_LEFT_CY NULL
#define OPTION_VAL_LOWER_RIGHT_CY NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_CY NULL
#define OPTION_VAL_1X_CY NULL
#define OPTION_VAL_3X_CY NULL
#define OPTION_VAL_5X_CY NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_CY NULL
#define OPTION_VAL_DEFAULT_ON_00_CY NULL
#define OPTION_VAL_DEFAULT_OFF_01_CY NULL
#define OPTION_VAL_BLACK_02_CY NULL
#define OPTION_VAL_LIGHT_BLUE_04_CY NULL
#define OPTION_VAL_CYAN_06_CY NULL
#define OPTION_VAL_CYAN_BLUE_07_CY NULL
#define OPTION_VAL_LIGHT_GREEN_08_CY NULL
#define OPTION_VAL_CYAN_GREEN_09_CY NULL
#define OPTION_VAL_LIGHT_CYAN_10_CY NULL
#define OPTION_VAL_PURPLE_12_CY NULL
#define OPTION_VAL_LIGHT_PURPLE_13_CY NULL
#define OPTION_VAL_YELLOW_14_CY NULL
#define OPTION_VAL_GRAY_15_CY NULL
#define OPTION_VAL_LIGHT_PURPLE_2_16_CY NULL
#define OPTION_VAL_LIGHT_GREEN_2_17_CY NULL
#define OPTION_VAL_LIGHT_GREEN_3_18_CY NULL
#define OPTION_VAL_LIGHT_CYAN_2_19_CY NULL
#define OPTION_VAL_LIGHT_RED_2_20_CY NULL
#define OPTION_VAL_MAGENTA_21_CY NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_CY NULL
#define OPTION_VAL_LIGHT_ORANGE_23_CY NULL
#define OPTION_VAL_ORANGE_24_CY NULL
#define OPTION_VAL_LIGHT_PURPLE_4_25_CY NULL
#define OPTION_VAL_LIGHT_YELLOW_26_CY NULL
#define OPTION_VAL_LIGHT_YELLOW_2_27_CY NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_CY NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_CY NULL
#define OPTION_VAL_40_CY NULL
#define OPTION_VAL_50_CY NULL
#define OPTION_VAL_60_CY NULL
#define OPTION_VAL_70_CY NULL
#define OPTION_VAL_80_CY NULL
#define OPTION_VAL_90_CY NULL
#define OPTION_VAL_100_CY NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_CY NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_CY NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_CY NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_CY NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_CY NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_CY NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_CY NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_CY NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_CY NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_CY NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_CY NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_CY NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_CY NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_CY NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_CY NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_CY NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_CY NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_CY NULL

struct retro_core_option_v2_category option_cats_cy[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_CY,
      CATEGORY_SYSTEM_INFO_0_CY
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_CY,
      CATEGORY_VIDEO_INFO_0_CY
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_CY,
      CATEGORY_PERFORMANCE_INFO_0_CY
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_CY,
      CATEGORY_HACKS_INFO_0_CY
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_CY,
      CATEGORY_INPUT_INFO_0_CY
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_CY,
      CATEGORY_VMU_INFO_0_CY
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_cy[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_CY,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_CY },
         { "USA",     OPTION_VAL_USA_CY },
         { "Europe",  OPTION_VAL_EUROPE_CY },
         { "Default", OPTION_VAL_DEFAULT_CY },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_CY,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_CY },
         { "English",  OPTION_VAL_ENGLISH_CY },
         { "German",   OPTION_VAL_GERMAN_CY },
         { "French",   OPTION_VAL_FRENCH_CY },
         { "Spanish",  OPTION_VAL_SPANISH_CY },
         { "Italian",  OPTION_VAL_ITALIAN_CY },
         { "Default",  OPTION_VAL_DEFAULT_CY },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_CY,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_boot_to_bios",
      CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_CY,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_CY,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_force_wince",
      CORE_OPTION_NAME_FORCE_WINCE_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_FORCE_WINCE_INFO_0_CY,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_CY,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_CY,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_CY,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_CY,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_CY,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_CY },
         { "640x480",    OPTION_VAL_640X480_CY },
         { "800x600",    OPTION_VAL_800X600_CY },
         { "960x720",    OPTION_VAL_960X720_CY },
         { "1024x768",   OPTION_VAL_1024X768_CY },
         { "1280x960",   OPTION_VAL_1280X960_CY },
         { "1440x1080",  OPTION_VAL_1440X1080_CY },
         { "1600x1200",  OPTION_VAL_1600X1200_CY },
         { "1920x1440",  OPTION_VAL_1920X1440_CY },
         { "2560x1920",  OPTION_VAL_2560X1920_CY },
         { "2880x2160",  OPTION_VAL_2880X2160_CY },
         { "3200x2400",  OPTION_VAL_3200X2400_CY },
         { "3840x2880",  OPTION_VAL_3840X2880_CY },
         { "4480x3360",  OPTION_VAL_4480X3360_CY },
         { "5120x3840",  OPTION_VAL_5120X3840_CY },
         { "5760x4320",  OPTION_VAL_5760X4320_CY },
         { "6400x4800",  OPTION_VAL_6400X4800_CY },
         { "7040x5280",  OPTION_VAL_7040X5280_CY },
         { "7680x5760",  OPTION_VAL_7680X5760_CY },
         { "8320x6240",  OPTION_VAL_8320X6240_CY },
         { "8960x6720",  OPTION_VAL_8960X6720_CY },
         { "9600x7200",  OPTION_VAL_9600X7200_CY },
         { "10240x7680", OPTION_VAL_10240X7680_CY },
         { "10880x8160", OPTION_VAL_10880X8160_CY },
         { "11520x8640", OPTION_VAL_11520X8640_CY },
         { "12160x9120", OPTION_VAL_12160X9120_CY },
         { "12800x9600", OPTION_VAL_12800X9600_CY },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_CY,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_CY },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_CY },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_CY },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_CY,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_CY },
         { "PAL",     OPTION_VAL_PAL_CY },
         { "PAL_N",   OPTION_VAL_PAL_N_CY },
         { "PAL_M",   OPTION_VAL_PAL_M_CY },
         { "Default", OPTION_VAL_DEFAULT_CY },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_CY,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_CY },
         { "vertical",   OPTION_VAL_VERTICAL_CY },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_CY,
      NULL,
      "",
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_CY },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_CY },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_CY },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_CY,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_CY },
         { "1GB",   OPTION_VAL_1GB_CY },
         { "2GB",   OPTION_VAL_2GB_CY },
         { "4GB",   OPTION_VAL_4GB_CY },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_CY,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_CY,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_CY,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_CY,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_CY,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_CY,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_CY,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_CY,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_CY },
         { "1",  OPTION_VAL_1_CY },
         { "2",  OPTION_VAL_2_CY },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_CY,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_CY,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_CY,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_CY,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O27_CY },
         { "4",  OPTION_VAL_4_CY },
         { "6",  OPTION_VAL_6_CY },
         { NULL, NULL },
      },
      "1",
   },
   {/* TODO: needs clarification */
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_CY,
      NULL,
      "",
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_CY,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_CY,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_CY,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_CY,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_CY },
         { "more", OPTION_VAL_MORE_CY },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_CY,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_CY,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_CY,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_CY,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_CY,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: probably needs explanation */
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_CY,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_CY,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_CY },
         { "5%",  OPTION_VAL_5_CY },
         { "10%", OPTION_VAL_10_CY },
         { "15%", OPTION_VAL_15_CY },
         { "20%", OPTION_VAL_20_CY },
         { "25%", OPTION_VAL_25_CY },
         { "30%", OPTION_VAL_30_CY },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_CY,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_CY },
         { "5%",  OPTION_VAL_5_CY },
         { "10%", OPTION_VAL_10_CY },
         { "15%", OPTION_VAL_15_CY },
         { "20%", OPTION_VAL_20_CY },
         { "25%", OPTION_VAL_25_CY },
         { "30%", OPTION_VAL_30_CY },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_CY,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_purupuru",
      CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_CY,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_CY,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_CY,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_CY,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_CY },
         { "Red",      OPTION_VAL_RED_CY },
         { "Green",    OPTION_VAL_GREEN_CY },
         { "Blue",     OPTION_VAL_BLUE_CY },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_CY,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_CY },
         { "Red",      OPTION_VAL_RED_CY },
         { "Green",    OPTION_VAL_GREEN_CY },
         { "Blue",     OPTION_VAL_BLUE_CY },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_CY,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_CY },
         { "Red",      OPTION_VAL_RED_CY },
         { "Green",    OPTION_VAL_GREEN_CY },
         { "Blue",     OPTION_VAL_BLUE_CY },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_CY,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_CY },
         { "Red",      OPTION_VAL_RED_CY },
         { "Green",    OPTION_VAL_GREEN_CY },
         { "Blue",     OPTION_VAL_BLUE_CY },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CY,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_CY,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_CY,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_CY },
         { "All VMUs", OPTION_VAL_ALL_VMUS_CY },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CY,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_CY,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_CY,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_CY },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_CY },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_CY },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_CY },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_CY },
         { "2x", OPTION_VAL_2_O27_CY },
         { "3x", OPTION_VAL_3X_CY },
         { "4x", OPTION_VAL_4_CY },
         { "5x", OPTION_VAL_5X_CY },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CY },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CY },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CY },
         { "BLUE 03",           OPTION_VAL_BLUE_CY },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CY },
         { "GREEN 05",          OPTION_VAL_GREEN_CY },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CY },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CY },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CY },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CY },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CY },
         { "RED 11",            OPTION_VAL_RED_CY },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CY },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CY },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CY },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CY },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CY },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CY },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CY },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CY },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CY },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CY },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CY },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CY },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CY },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CY },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CY },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CY },
         { "WHITE 28",          OPTION_VAL_WHITE_CY },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CY },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CY },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CY },
         { "BLUE 03",           OPTION_VAL_BLUE_CY },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CY },
         { "GREEN 05",          OPTION_VAL_GREEN_CY },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CY },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CY },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CY },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CY },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CY },
         { "RED 11",            OPTION_VAL_RED_CY },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CY },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CY },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CY },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CY },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CY },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CY },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CY },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CY },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CY },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CY },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CY },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CY },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CY },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CY },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CY },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CY },
         { "WHITE 28",          OPTION_VAL_WHITE_CY },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_CY },
         { "20%",  OPTION_VAL_20_CY },
         { "30%",  OPTION_VAL_30_CY },
         { "40%",  OPTION_VAL_40_CY },
         { "50%",  OPTION_VAL_50_CY },
         { "60%",  OPTION_VAL_60_CY },
         { "70%",  OPTION_VAL_70_CY },
         { "80%",  OPTION_VAL_80_CY },
         { "90%",  OPTION_VAL_90_CY },
         { "100%", OPTION_VAL_100_CY },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_CY },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_CY },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_CY },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_CY },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_CY },
         { "2x", OPTION_VAL_2_O27_CY },
         { "3x", OPTION_VAL_3X_CY },
         { "4x", OPTION_VAL_4_CY },
         { "5x", OPTION_VAL_5X_CY },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CY },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CY },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CY },
         { "BLUE 03",           OPTION_VAL_BLUE_CY },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CY },
         { "GREEN 05",          OPTION_VAL_GREEN_CY },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CY },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CY },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CY },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CY },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CY },
         { "RED 11",            OPTION_VAL_RED_CY },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CY },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CY },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CY },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CY },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CY },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CY },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CY },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CY },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CY },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CY },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CY },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CY },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CY },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CY },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CY },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CY },
         { "WHITE 28",          OPTION_VAL_WHITE_CY },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CY },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CY },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CY },
         { "BLUE 03",           OPTION_VAL_BLUE_CY },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CY },
         { "GREEN 05",          OPTION_VAL_GREEN_CY },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CY },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CY },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CY },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CY },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CY },
         { "RED 11",            OPTION_VAL_RED_CY },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CY },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CY },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CY },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CY },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CY },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CY },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CY },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CY },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CY },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CY },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CY },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CY },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CY },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CY },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CY },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CY },
         { "WHITE 28",          OPTION_VAL_WHITE_CY },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_CY },
         { "20%",  OPTION_VAL_20_CY },
         { "30%",  OPTION_VAL_30_CY },
         { "40%",  OPTION_VAL_40_CY },
         { "50%",  OPTION_VAL_50_CY },
         { "60%",  OPTION_VAL_60_CY },
         { "70%",  OPTION_VAL_70_CY },
         { "80%",  OPTION_VAL_80_CY },
         { "90%",  OPTION_VAL_90_CY },
         { "100%", OPTION_VAL_100_CY },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_CY },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_CY },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_CY },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_CY },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_CY },
         { "2x", OPTION_VAL_2_O27_CY },
         { "3x", OPTION_VAL_3X_CY },
         { "4x", OPTION_VAL_4_CY },
         { "5x", OPTION_VAL_5X_CY },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CY },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CY },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CY },
         { "BLUE 03",           OPTION_VAL_BLUE_CY },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CY },
         { "GREEN 05",          OPTION_VAL_GREEN_CY },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CY },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CY },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CY },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CY },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CY },
         { "RED 11",            OPTION_VAL_RED_CY },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CY },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CY },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CY },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CY },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CY },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CY },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CY },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CY },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CY },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CY },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CY },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CY },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CY },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CY },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CY },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CY },
         { "WHITE 28",          OPTION_VAL_WHITE_CY },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CY },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CY },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CY },
         { "BLUE 03",           OPTION_VAL_BLUE_CY },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CY },
         { "GREEN 05",          OPTION_VAL_GREEN_CY },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CY },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CY },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CY },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CY },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CY },
         { "RED 11",            OPTION_VAL_RED_CY },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CY },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CY },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CY },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CY },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CY },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CY },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CY },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CY },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CY },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CY },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CY },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CY },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CY },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CY },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CY },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CY },
         { "WHITE 28",          OPTION_VAL_WHITE_CY },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_CY },
         { "20%",  OPTION_VAL_20_CY },
         { "30%",  OPTION_VAL_30_CY },
         { "40%",  OPTION_VAL_40_CY },
         { "50%",  OPTION_VAL_50_CY },
         { "60%",  OPTION_VAL_60_CY },
         { "70%",  OPTION_VAL_70_CY },
         { "80%",  OPTION_VAL_80_CY },
         { "90%",  OPTION_VAL_90_CY },
         { "100%", OPTION_VAL_100_CY },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_CY },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_CY },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_CY },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_CY },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_CY },
         { "2x", OPTION_VAL_2_O27_CY },
         { "3x", OPTION_VAL_3X_CY },
         { "4x", OPTION_VAL_4_CY },
         { "5x", OPTION_VAL_5X_CY },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CY },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CY },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CY },
         { "BLUE 03",           OPTION_VAL_BLUE_CY },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CY },
         { "GREEN 05",          OPTION_VAL_GREEN_CY },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CY },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CY },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CY },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CY },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CY },
         { "RED 11",            OPTION_VAL_RED_CY },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CY },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CY },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CY },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CY },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CY },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CY },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CY },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CY },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CY },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CY },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CY },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CY },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CY },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CY },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CY },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CY },
         { "WHITE 28",          OPTION_VAL_WHITE_CY },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CY },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CY },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CY },
         { "BLUE 03",           OPTION_VAL_BLUE_CY },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CY },
         { "GREEN 05",          OPTION_VAL_GREEN_CY },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CY },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CY },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CY },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CY },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CY },
         { "RED 11",            OPTION_VAL_RED_CY },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CY },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CY },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CY },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CY },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CY },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CY },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CY },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CY },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CY },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CY },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CY },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CY },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CY },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CY },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CY },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CY },
         { "WHITE 28",          OPTION_VAL_WHITE_CY },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_CY },
         { "20%",  OPTION_VAL_20_CY },
         { "30%",  OPTION_VAL_30_CY },
         { "40%",  OPTION_VAL_40_CY },
         { "50%",  OPTION_VAL_50_CY },
         { "60%",  OPTION_VAL_60_CY },
         { "70%",  OPTION_VAL_70_CY },
         { "80%",  OPTION_VAL_80_CY },
         { "90%",  OPTION_VAL_90_CY },
         { "100%", OPTION_VAL_100_CY },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_cy = {
   option_cats_cy,
   option_defs_cy
};

/* RETRO_LANGUAGE_DA */

#define CATEGORY_SYSTEM_LABEL_DA NULL
#define CATEGORY_SYSTEM_INFO_0_DA NULL
#define CATEGORY_VIDEO_LABEL_DA NULL
#define CATEGORY_VIDEO_INFO_0_DA NULL
#define CATEGORY_PERFORMANCE_LABEL_DA NULL
#define CATEGORY_PERFORMANCE_INFO_0_DA NULL
#define CATEGORY_HACKS_LABEL_DA NULL
#define CATEGORY_HACKS_INFO_0_DA NULL
#define CATEGORY_INPUT_LABEL_DA NULL
#define CATEGORY_INPUT_INFO_0_DA NULL
#define CATEGORY_VMU_LABEL_DA NULL
#define CATEGORY_VMU_INFO_0_DA NULL
#define CORE_OPTION_NAME_REGION_LABEL_DA NULL
#define OPTION_VAL_JAPAN_DA NULL
#define OPTION_VAL_USA_DA NULL
#define OPTION_VAL_EUROPE_DA NULL
#define OPTION_VAL_DEFAULT_DA NULL
#define CORE_OPTION_NAME_LANGUAGE_LABEL_DA NULL
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_DA NULL
#define OPTION_VAL_JAPANESE_DA NULL
#define OPTION_VAL_ENGLISH_DA NULL
#define OPTION_VAL_GERMAN_DA NULL
#define OPTION_VAL_FRENCH_DA NULL
#define OPTION_VAL_SPANISH_DA NULL
#define OPTION_VAL_ITALIAN_DA NULL
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_DA NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_DA NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_DA NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_DA NULL
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_DA NULL
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_DA NULL
#define CORE_OPTION_NAME_FORCE_WINCE_LABEL_DA NULL
#define CORE_OPTION_NAME_FORCE_WINCE_INFO_0_DA NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_DA NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_DA NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_DA NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_DA NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_DA NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_DA NULL
#define CORE_OPTION_NAME_UPNP_LABEL_DA NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_DA NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_DA NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_DA NULL
#define OPTION_VAL_320X240_DA NULL
#define OPTION_VAL_640X480_DA NULL
#define OPTION_VAL_800X600_DA NULL
#define OPTION_VAL_960X720_DA NULL
#define OPTION_VAL_1024X768_DA NULL
#define OPTION_VAL_1280X960_DA NULL
#define OPTION_VAL_1440X1080_DA NULL
#define OPTION_VAL_1600X1200_DA NULL
#define OPTION_VAL_1920X1440_DA NULL
#define OPTION_VAL_2560X1920_DA NULL
#define OPTION_VAL_2880X2160_DA NULL
#define OPTION_VAL_3200X2400_DA NULL
#define OPTION_VAL_3840X2880_DA NULL
#define OPTION_VAL_4480X3360_DA NULL
#define OPTION_VAL_5120X3840_DA NULL
#define OPTION_VAL_5760X4320_DA NULL
#define OPTION_VAL_6400X4800_DA NULL
#define OPTION_VAL_7040X5280_DA NULL
#define OPTION_VAL_7680X5760_DA NULL
#define OPTION_VAL_8320X6240_DA NULL
#define OPTION_VAL_8960X6720_DA NULL
#define OPTION_VAL_9600X7200_DA NULL
#define OPTION_VAL_10240X7680_DA NULL
#define OPTION_VAL_10880X8160_DA NULL
#define OPTION_VAL_11520X8640_DA NULL
#define OPTION_VAL_12160X9120_DA NULL
#define OPTION_VAL_12800X9600_DA NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_DA NULL
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_DA NULL
#define OPTION_VAL_VGA_DA NULL
#define OPTION_VAL_TV_RGB_DA NULL
#define OPTION_VAL_TV_COMPOSITE_DA NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_DA NULL
#define OPTION_VAL_NTSC_DA NULL
#define OPTION_VAL_PAL_DA NULL
#define OPTION_VAL_PAL_N_DA NULL
#define OPTION_VAL_PAL_M_DA NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_DA NULL
#define OPTION_VAL_HORIZONTAL_DA NULL
#define OPTION_VAL_VERTICAL_DA NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_DA NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_DA NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_DA NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_DA NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_DA NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_DA NULL
#define OPTION_VAL_512MB_DA NULL
#define OPTION_VAL_1GB_DA NULL
#define OPTION_VAL_2GB_DA NULL
#define OPTION_VAL_4GB_DA NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_DA NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_DA NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_DA NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_DA NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_DA NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_DA NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_DA NULL
#define CORE_OPTION_NAME_FOG_LABEL_DA NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_DA NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_DA NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_DA NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_DA NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_DA NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_DA NULL
#define OPTION_VAL_1_DA NULL
#define OPTION_VAL_2_DA NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_DA NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_DA NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_DA NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_DA NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_DA NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_DA NULL
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_DA NULL
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_DA NULL
#define OPTION_VAL_2_O27_DA NULL
#define OPTION_VAL_4_DA NULL
#define OPTION_VAL_6_DA NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_DA NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_DA NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_DA NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_DA NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_DA NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_DA NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_DA NULL
#define OPTION_VAL_SOME_DA NULL
#define OPTION_VAL_MORE_DA NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_DA NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_DA NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_DA NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_DA NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_DA NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_DA NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_DA NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_DA NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_DA NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_DA NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_DA NULL
#define OPTION_VAL_0_DA NULL
#define OPTION_VAL_5_DA NULL
#define OPTION_VAL_10_DA NULL
#define OPTION_VAL_15_DA NULL
#define OPTION_VAL_20_DA NULL
#define OPTION_VAL_25_DA NULL
#define OPTION_VAL_30_DA NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_DA NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_DA NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_DA NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_DA NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_DA NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_DA NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_DA NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_DA NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_DA NULL
#define OPTION_VAL_WHITE_DA NULL
#define OPTION_VAL_RED_DA NULL
#define OPTION_VAL_GREEN_DA NULL
#define OPTION_VAL_BLUE_DA NULL
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_DA NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_DA NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_DA NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_DA NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_DA NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_DA NULL
#define OPTION_VAL_VMU_A1_DA NULL
#define OPTION_VAL_ALL_VMUS_DA NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_DA NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_DA NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_DA NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_DA NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_DA NULL
#define OPTION_VAL_UPPER_LEFT_DA NULL
#define OPTION_VAL_UPPER_RIGHT_DA NULL
#define OPTION_VAL_LOWER_LEFT_DA NULL
#define OPTION_VAL_LOWER_RIGHT_DA NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_DA NULL
#define OPTION_VAL_1X_DA NULL
#define OPTION_VAL_3X_DA NULL
#define OPTION_VAL_5X_DA NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_DA NULL
#define OPTION_VAL_DEFAULT_ON_00_DA NULL
#define OPTION_VAL_DEFAULT_OFF_01_DA NULL
#define OPTION_VAL_BLACK_02_DA NULL
#define OPTION_VAL_LIGHT_BLUE_04_DA NULL
#define OPTION_VAL_CYAN_06_DA NULL
#define OPTION_VAL_CYAN_BLUE_07_DA NULL
#define OPTION_VAL_LIGHT_GREEN_08_DA NULL
#define OPTION_VAL_CYAN_GREEN_09_DA NULL
#define OPTION_VAL_LIGHT_CYAN_10_DA NULL
#define OPTION_VAL_PURPLE_12_DA NULL
#define OPTION_VAL_LIGHT_PURPLE_13_DA NULL
#define OPTION_VAL_YELLOW_14_DA NULL
#define OPTION_VAL_GRAY_15_DA NULL
#define OPTION_VAL_LIGHT_PURPLE_2_16_DA NULL
#define OPTION_VAL_LIGHT_GREEN_2_17_DA NULL
#define OPTION_VAL_LIGHT_GREEN_3_18_DA NULL
#define OPTION_VAL_LIGHT_CYAN_2_19_DA NULL
#define OPTION_VAL_LIGHT_RED_2_20_DA NULL
#define OPTION_VAL_MAGENTA_21_DA NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_DA NULL
#define OPTION_VAL_LIGHT_ORANGE_23_DA NULL
#define OPTION_VAL_ORANGE_24_DA NULL
#define OPTION_VAL_LIGHT_PURPLE_4_25_DA NULL
#define OPTION_VAL_LIGHT_YELLOW_26_DA NULL
#define OPTION_VAL_LIGHT_YELLOW_2_27_DA NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_DA NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_DA NULL
#define OPTION_VAL_40_DA NULL
#define OPTION_VAL_50_DA NULL
#define OPTION_VAL_60_DA NULL
#define OPTION_VAL_70_DA NULL
#define OPTION_VAL_80_DA NULL
#define OPTION_VAL_90_DA NULL
#define OPTION_VAL_100_DA NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_DA NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_DA NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_DA NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_DA NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_DA NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_DA NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_DA NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_DA NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_DA NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_DA NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_DA NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_DA NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_DA NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_DA NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_DA NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_DA NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_DA NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_DA NULL

struct retro_core_option_v2_category option_cats_da[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_DA,
      CATEGORY_SYSTEM_INFO_0_DA
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_DA,
      CATEGORY_VIDEO_INFO_0_DA
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_DA,
      CATEGORY_PERFORMANCE_INFO_0_DA
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_DA,
      CATEGORY_HACKS_INFO_0_DA
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_DA,
      CATEGORY_INPUT_INFO_0_DA
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_DA,
      CATEGORY_VMU_INFO_0_DA
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_da[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_DA,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_DA },
         { "USA",     OPTION_VAL_USA_DA },
         { "Europe",  OPTION_VAL_EUROPE_DA },
         { "Default", OPTION_VAL_DEFAULT_DA },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_DA,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_DA },
         { "English",  OPTION_VAL_ENGLISH_DA },
         { "German",   OPTION_VAL_GERMAN_DA },
         { "French",   OPTION_VAL_FRENCH_DA },
         { "Spanish",  OPTION_VAL_SPANISH_DA },
         { "Italian",  OPTION_VAL_ITALIAN_DA },
         { "Default",  OPTION_VAL_DEFAULT_DA },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_DA,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_boot_to_bios",
      CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_DA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_DA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_force_wince",
      CORE_OPTION_NAME_FORCE_WINCE_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_FORCE_WINCE_INFO_0_DA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_DA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_DA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_DA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_DA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_DA,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_DA },
         { "640x480",    OPTION_VAL_640X480_DA },
         { "800x600",    OPTION_VAL_800X600_DA },
         { "960x720",    OPTION_VAL_960X720_DA },
         { "1024x768",   OPTION_VAL_1024X768_DA },
         { "1280x960",   OPTION_VAL_1280X960_DA },
         { "1440x1080",  OPTION_VAL_1440X1080_DA },
         { "1600x1200",  OPTION_VAL_1600X1200_DA },
         { "1920x1440",  OPTION_VAL_1920X1440_DA },
         { "2560x1920",  OPTION_VAL_2560X1920_DA },
         { "2880x2160",  OPTION_VAL_2880X2160_DA },
         { "3200x2400",  OPTION_VAL_3200X2400_DA },
         { "3840x2880",  OPTION_VAL_3840X2880_DA },
         { "4480x3360",  OPTION_VAL_4480X3360_DA },
         { "5120x3840",  OPTION_VAL_5120X3840_DA },
         { "5760x4320",  OPTION_VAL_5760X4320_DA },
         { "6400x4800",  OPTION_VAL_6400X4800_DA },
         { "7040x5280",  OPTION_VAL_7040X5280_DA },
         { "7680x5760",  OPTION_VAL_7680X5760_DA },
         { "8320x6240",  OPTION_VAL_8320X6240_DA },
         { "8960x6720",  OPTION_VAL_8960X6720_DA },
         { "9600x7200",  OPTION_VAL_9600X7200_DA },
         { "10240x7680", OPTION_VAL_10240X7680_DA },
         { "10880x8160", OPTION_VAL_10880X8160_DA },
         { "11520x8640", OPTION_VAL_11520X8640_DA },
         { "12160x9120", OPTION_VAL_12160X9120_DA },
         { "12800x9600", OPTION_VAL_12800X9600_DA },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_DA,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_DA },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_DA },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_DA },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_DA,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_DA },
         { "PAL",     OPTION_VAL_PAL_DA },
         { "PAL_N",   OPTION_VAL_PAL_N_DA },
         { "PAL_M",   OPTION_VAL_PAL_M_DA },
         { "Default", OPTION_VAL_DEFAULT_DA },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_DA,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_DA },
         { "vertical",   OPTION_VAL_VERTICAL_DA },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_DA,
      NULL,
      "",
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_DA },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_DA },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_DA },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_DA,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_DA },
         { "1GB",   OPTION_VAL_1GB_DA },
         { "2GB",   OPTION_VAL_2GB_DA },
         { "4GB",   OPTION_VAL_4GB_DA },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_DA,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_DA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_DA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_DA,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_DA,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_DA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_DA,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_DA,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_DA },
         { "1",  OPTION_VAL_1_DA },
         { "2",  OPTION_VAL_2_DA },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_DA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_DA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_DA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_DA,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O27_DA },
         { "4",  OPTION_VAL_4_DA },
         { "6",  OPTION_VAL_6_DA },
         { NULL, NULL },
      },
      "1",
   },
   {/* TODO: needs clarification */
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_DA,
      NULL,
      "",
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_DA,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_DA,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_DA,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_DA,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_DA },
         { "more", OPTION_VAL_MORE_DA },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_DA,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_DA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_DA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_DA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_DA,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: probably needs explanation */
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_DA,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_DA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_DA },
         { "5%",  OPTION_VAL_5_DA },
         { "10%", OPTION_VAL_10_DA },
         { "15%", OPTION_VAL_15_DA },
         { "20%", OPTION_VAL_20_DA },
         { "25%", OPTION_VAL_25_DA },
         { "30%", OPTION_VAL_30_DA },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_DA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_DA },
         { "5%",  OPTION_VAL_5_DA },
         { "10%", OPTION_VAL_10_DA },
         { "15%", OPTION_VAL_15_DA },
         { "20%", OPTION_VAL_20_DA },
         { "25%", OPTION_VAL_25_DA },
         { "30%", OPTION_VAL_30_DA },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_DA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_purupuru",
      CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_DA,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_DA,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_DA,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_DA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_DA },
         { "Red",      OPTION_VAL_RED_DA },
         { "Green",    OPTION_VAL_GREEN_DA },
         { "Blue",     OPTION_VAL_BLUE_DA },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_DA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_DA },
         { "Red",      OPTION_VAL_RED_DA },
         { "Green",    OPTION_VAL_GREEN_DA },
         { "Blue",     OPTION_VAL_BLUE_DA },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_DA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_DA },
         { "Red",      OPTION_VAL_RED_DA },
         { "Green",    OPTION_VAL_GREEN_DA },
         { "Blue",     OPTION_VAL_BLUE_DA },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_DA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_DA },
         { "Red",      OPTION_VAL_RED_DA },
         { "Green",    OPTION_VAL_GREEN_DA },
         { "Blue",     OPTION_VAL_BLUE_DA },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_DA,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_DA,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_DA,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_DA },
         { "All VMUs", OPTION_VAL_ALL_VMUS_DA },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_DA,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_DA,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_DA,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_DA },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_DA },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_DA },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_DA },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_DA },
         { "2x", OPTION_VAL_2_O27_DA },
         { "3x", OPTION_VAL_3X_DA },
         { "4x", OPTION_VAL_4_DA },
         { "5x", OPTION_VAL_5X_DA },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_DA },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_DA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_DA },
         { "BLUE 03",           OPTION_VAL_BLUE_DA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_DA },
         { "GREEN 05",          OPTION_VAL_GREEN_DA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_DA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_DA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_DA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_DA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_DA },
         { "RED 11",            OPTION_VAL_RED_DA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_DA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_DA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_DA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_DA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_DA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_DA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_DA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_DA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_DA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_DA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_DA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_DA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_DA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_DA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_DA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_DA },
         { "WHITE 28",          OPTION_VAL_WHITE_DA },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_DA },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_DA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_DA },
         { "BLUE 03",           OPTION_VAL_BLUE_DA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_DA },
         { "GREEN 05",          OPTION_VAL_GREEN_DA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_DA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_DA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_DA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_DA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_DA },
         { "RED 11",            OPTION_VAL_RED_DA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_DA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_DA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_DA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_DA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_DA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_DA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_DA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_DA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_DA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_DA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_DA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_DA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_DA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_DA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_DA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_DA },
         { "WHITE 28",          OPTION_VAL_WHITE_DA },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_DA },
         { "20%",  OPTION_VAL_20_DA },
         { "30%",  OPTION_VAL_30_DA },
         { "40%",  OPTION_VAL_40_DA },
         { "50%",  OPTION_VAL_50_DA },
         { "60%",  OPTION_VAL_60_DA },
         { "70%",  OPTION_VAL_70_DA },
         { "80%",  OPTION_VAL_80_DA },
         { "90%",  OPTION_VAL_90_DA },
         { "100%", OPTION_VAL_100_DA },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_DA },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_DA },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_DA },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_DA },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_DA },
         { "2x", OPTION_VAL_2_O27_DA },
         { "3x", OPTION_VAL_3X_DA },
         { "4x", OPTION_VAL_4_DA },
         { "5x", OPTION_VAL_5X_DA },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_DA },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_DA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_DA },
         { "BLUE 03",           OPTION_VAL_BLUE_DA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_DA },
         { "GREEN 05",          OPTION_VAL_GREEN_DA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_DA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_DA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_DA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_DA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_DA },
         { "RED 11",            OPTION_VAL_RED_DA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_DA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_DA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_DA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_DA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_DA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_DA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_DA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_DA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_DA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_DA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_DA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_DA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_DA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_DA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_DA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_DA },
         { "WHITE 28",          OPTION_VAL_WHITE_DA },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_DA },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_DA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_DA },
         { "BLUE 03",           OPTION_VAL_BLUE_DA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_DA },
         { "GREEN 05",          OPTION_VAL_GREEN_DA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_DA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_DA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_DA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_DA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_DA },
         { "RED 11",            OPTION_VAL_RED_DA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_DA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_DA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_DA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_DA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_DA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_DA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_DA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_DA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_DA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_DA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_DA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_DA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_DA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_DA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_DA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_DA },
         { "WHITE 28",          OPTION_VAL_WHITE_DA },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_DA },
         { "20%",  OPTION_VAL_20_DA },
         { "30%",  OPTION_VAL_30_DA },
         { "40%",  OPTION_VAL_40_DA },
         { "50%",  OPTION_VAL_50_DA },
         { "60%",  OPTION_VAL_60_DA },
         { "70%",  OPTION_VAL_70_DA },
         { "80%",  OPTION_VAL_80_DA },
         { "90%",  OPTION_VAL_90_DA },
         { "100%", OPTION_VAL_100_DA },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_DA },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_DA },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_DA },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_DA },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_DA },
         { "2x", OPTION_VAL_2_O27_DA },
         { "3x", OPTION_VAL_3X_DA },
         { "4x", OPTION_VAL_4_DA },
         { "5x", OPTION_VAL_5X_DA },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_DA },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_DA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_DA },
         { "BLUE 03",           OPTION_VAL_BLUE_DA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_DA },
         { "GREEN 05",          OPTION_VAL_GREEN_DA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_DA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_DA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_DA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_DA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_DA },
         { "RED 11",            OPTION_VAL_RED_DA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_DA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_DA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_DA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_DA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_DA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_DA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_DA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_DA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_DA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_DA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_DA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_DA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_DA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_DA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_DA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_DA },
         { "WHITE 28",          OPTION_VAL_WHITE_DA },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_DA },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_DA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_DA },
         { "BLUE 03",           OPTION_VAL_BLUE_DA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_DA },
         { "GREEN 05",          OPTION_VAL_GREEN_DA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_DA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_DA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_DA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_DA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_DA },
         { "RED 11",            OPTION_VAL_RED_DA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_DA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_DA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_DA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_DA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_DA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_DA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_DA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_DA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_DA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_DA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_DA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_DA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_DA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_DA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_DA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_DA },
         { "WHITE 28",          OPTION_VAL_WHITE_DA },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_DA },
         { "20%",  OPTION_VAL_20_DA },
         { "30%",  OPTION_VAL_30_DA },
         { "40%",  OPTION_VAL_40_DA },
         { "50%",  OPTION_VAL_50_DA },
         { "60%",  OPTION_VAL_60_DA },
         { "70%",  OPTION_VAL_70_DA },
         { "80%",  OPTION_VAL_80_DA },
         { "90%",  OPTION_VAL_90_DA },
         { "100%", OPTION_VAL_100_DA },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_DA },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_DA },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_DA },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_DA },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_DA },
         { "2x", OPTION_VAL_2_O27_DA },
         { "3x", OPTION_VAL_3X_DA },
         { "4x", OPTION_VAL_4_DA },
         { "5x", OPTION_VAL_5X_DA },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_DA },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_DA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_DA },
         { "BLUE 03",           OPTION_VAL_BLUE_DA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_DA },
         { "GREEN 05",          OPTION_VAL_GREEN_DA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_DA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_DA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_DA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_DA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_DA },
         { "RED 11",            OPTION_VAL_RED_DA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_DA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_DA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_DA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_DA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_DA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_DA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_DA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_DA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_DA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_DA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_DA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_DA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_DA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_DA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_DA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_DA },
         { "WHITE 28",          OPTION_VAL_WHITE_DA },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_DA },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_DA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_DA },
         { "BLUE 03",           OPTION_VAL_BLUE_DA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_DA },
         { "GREEN 05",          OPTION_VAL_GREEN_DA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_DA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_DA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_DA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_DA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_DA },
         { "RED 11",            OPTION_VAL_RED_DA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_DA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_DA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_DA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_DA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_DA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_DA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_DA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_DA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_DA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_DA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_DA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_DA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_DA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_DA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_DA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_DA },
         { "WHITE 28",          OPTION_VAL_WHITE_DA },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_DA },
         { "20%",  OPTION_VAL_20_DA },
         { "30%",  OPTION_VAL_30_DA },
         { "40%",  OPTION_VAL_40_DA },
         { "50%",  OPTION_VAL_50_DA },
         { "60%",  OPTION_VAL_60_DA },
         { "70%",  OPTION_VAL_70_DA },
         { "80%",  OPTION_VAL_80_DA },
         { "90%",  OPTION_VAL_90_DA },
         { "100%", OPTION_VAL_100_DA },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_da = {
   option_cats_da,
   option_defs_da
};

/* RETRO_LANGUAGE_DE */

#define CATEGORY_SYSTEM_LABEL_DE NULL
#define CATEGORY_SYSTEM_INFO_0_DE "Region, Sprache, BIOS und Basis-Hardware Einstellungen konfigurieren."
#define CATEGORY_VIDEO_LABEL_DE NULL
#define CATEGORY_VIDEO_INFO_0_DE "Auflösung, sortierungsunabhängige Transparenz und visuelle Effekt-Einstellungen konfigurieren."
#define CATEGORY_PERFORMANCE_LABEL_DE "Leistung"
#define CATEGORY_PERFORMANCE_INFO_0_DE "Einstellungen für Mehrgängiges Rendern und Bilder überspringen ändern."
#define CATEGORY_HACKS_LABEL_DE "Emulations-Hacks"
#define CATEGORY_HACKS_INFO_0_DE "Breitbild-Überschreibungen, GD-ROM Ladegeschwindigkeit und Textur-Austausch-Einstellungen konfigurieren."
#define CATEGORY_INPUT_LABEL_DE "Eingabe"
#define CATEGORY_INPUT_INFO_0_DE "Gamepad- und Pistoleneinstellungen konfigurieren."
#define CATEGORY_VMU_LABEL_DE NULL
#define CATEGORY_VMU_INFO_0_DE "VMU Dateien pro Spiel und VMU Anzeigeeinstellungen konfigurieren."
#define CORE_OPTION_NAME_REGION_LABEL_DE NULL
#define OPTION_VAL_JAPAN_DE NULL
#define OPTION_VAL_USA_DE NULL
#define OPTION_VAL_EUROPE_DE "Europa"
#define OPTION_VAL_DEFAULT_DE "Standard"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_DE "Sprache"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_DE "Ändert die Sprache, die vom BIOS und von Spielen verwendet wird, die mehrere Sprachen enthalten."
#define OPTION_VAL_JAPANESE_DE "Japanisch"
#define OPTION_VAL_ENGLISH_DE "Englisch"
#define OPTION_VAL_GERMAN_DE "Deutsch"
#define OPTION_VAL_FRENCH_DE "Französisch"
#define OPTION_VAL_SPANISH_DE "Spanisch"
#define OPTION_VAL_ITALIAN_DE "Italienisch"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_DE "HLE BIOS (Neustart erforderlich)"
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_DE "Erzwingt Verwendung von High-Level-Emulation BIOS."
#define CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_DE "BIOS starten (Neustart erforderlich)"
#define CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_DE "Direkt ins Dreamcast BIOS Menü starten."
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_DE "DSP aktivieren"
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_DE "Aktiviert Emulation von DSP des Dreamcasts (digitaler Signalprozessor). Verbessert die Genauigkeit des erzeugten Klanges, erhöht aber die Leistungsanforderungen."
#define CORE_OPTION_NAME_FORCE_WINCE_LABEL_DE "Windows CE-Modus erzwingen"
#define CORE_OPTION_NAME_FORCE_WINCE_INFO_0_DE "Aktiviert vollständige MMU (Memory Management Unit) Emulation und andere Einstellungen für Windows CE Spiele."
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_DE "NAOMI-Servicetasten erlauben"
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_DE "Aktiviert den SERVICE-Knopf für NAOMI, um die Automateneinstellungen aufzurufen."
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_DE "NAOMI Spiele auf kostenloses Spiel setzen"
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_DE "Ändert die Münzeinstellungen des Spiels zu kostenlos."
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_DE "Breitbandadapter-Emulation"
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_DE "Den Ethernet-Breitbandadapter anstelle des Modems emulieren. (Neustart erforderlich)"
#define CORE_OPTION_NAME_UPNP_LABEL_DE "UPnP aktivieren"
#define CORE_OPTION_NAME_UPNP_INFO_0_DE "Verwendet UPnP, um den Internetrouter automatisch für Onlinespiele zu konfigurieren."
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_DE "Interne Auflösung"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_DE "Rendering-Auflösung ändern."
#define OPTION_VAL_320X240_DE "320 x 240"
#define OPTION_VAL_640X480_DE "640 x 480"
#define OPTION_VAL_800X600_DE "800 x 600"
#define OPTION_VAL_960X720_DE "960 x 720"
#define OPTION_VAL_1024X768_DE "1024 x 768"
#define OPTION_VAL_1280X960_DE "1280 x 960"
#define OPTION_VAL_1440X1080_DE "1440 x 1080"
#define OPTION_VAL_1600X1200_DE "1600 x 1200"
#define OPTION_VAL_1920X1440_DE "1920 x 1440"
#define OPTION_VAL_2560X1920_DE "2560 x 1920"
#define OPTION_VAL_2880X2160_DE "2880 x 2160"
#define OPTION_VAL_3200X2400_DE "3200 x 2400"
#define OPTION_VAL_3840X2880_DE "3840 x 2880"
#define OPTION_VAL_4480X3360_DE "4480 x 3360"
#define OPTION_VAL_5120X3840_DE "5120 x 3840"
#define OPTION_VAL_5760X4320_DE "5760 x 4320"
#define OPTION_VAL_6400X4800_DE "6400 x 4800"
#define OPTION_VAL_7040X5280_DE "7040 x 5280"
#define OPTION_VAL_7680X5760_DE "7680 x 5760"
#define OPTION_VAL_8320X6240_DE "8320 x 6240"
#define OPTION_VAL_8960X6720_DE "8960 x 6720"
#define OPTION_VAL_9600X7200_DE "9600 x 7200"
#define OPTION_VAL_10240X7680_DE "10240 x 7680"
#define OPTION_VAL_10880X8160_DE "10880 x 8160"
#define OPTION_VAL_11520X8640_DE "11520 x 8640"
#define OPTION_VAL_12160X9120_DE "12160 x 9120"
#define OPTION_VAL_12800X9600_DE "12800 x 9600"
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_DE "Kabeltyp"
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_DE "Das Ausgabesignalkabel: 'TV (Composite)' ist das am weitesten unterstützte."
#define OPTION_VAL_VGA_DE NULL
#define OPTION_VAL_TV_RGB_DE NULL
#define OPTION_VAL_TV_COMPOSITE_DE NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_DE "Ausgabe-Regions-Standard"
#define OPTION_VAL_NTSC_DE NULL
#define OPTION_VAL_PAL_DE "PAL (Welt)"
#define OPTION_VAL_PAL_N_DE "PAL-N (Argentinien, Paraguay, Uruguay)"
#define OPTION_VAL_PAL_M_DE "PAL-M (Brasilien)"
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_DE "Bildschirmausrichtung"
#define OPTION_VAL_HORIZONTAL_DE NULL
#define OPTION_VAL_VERTICAL_DE "Vertikal"
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_DE "Alpha-Sortierung"
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_DE "Pro-Streifen (schnell, am wenigsten genau)"
#define OPTION_VAL_PER_TRIANGLE_NORMAL_DE "Pro-Dreieck (normal)"
#define OPTION_VAL_PER_PIXEL_ACCURATE_DE "Pro-Pixel (genau, aber langsamer)"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_DE "Anhäufung Pixel Puffergröße"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_DE "Für eine korrekte Ausgabe höherer Auflösungen sind möglicherweise höhere Werte erforderlich."
#define OPTION_VAL_512MB_DE "512 MB"
#define OPTION_VAL_1GB_DE "1 GB"
#define OPTION_VAL_2GB_DE "2 GB"
#define OPTION_VAL_4GB_DE "4 GB"
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_DE "Maximale transparente Ebenen"
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_DE "Höhere Werte könnten erforderlich sein für komplexe Szenen."
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_DE "Vollständige Bildspeicheremulation"
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_DE "Vollständige Bildspeicheremulation in VRAM aktivieren. Dies ist für Spiele nützlich, die den Bildspeicher direkt in VRAM lesen oder schreiben. Wenn aktiviert, wird die interne Auflösung auf 640×480 erzwungen und die Leistung kann stark beeinträchtigt werden."
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_DE "RTT-Puffer (Render-To-Texture) erlauben"
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_DE "Kopiert gerenderte Texturen von der GPU in VRAM zurück. Diese Option ist normalerweise für Spiele aktiviert, die dies erfordern. Wenn aktiviert, ist das Hochskalieren von Texturen deaktiviert und die Leistung kann beeinträchtigt werden."
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_DE NULL
#define CORE_OPTION_NAME_FOG_LABEL_DE "Nebeleffekte"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_DE "Volumen-Modifizierer"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_DE "Eine Dreamcast GPU-Funktion, die normalerweise von Spielen benutzt wird, um Objektschatten zu zeichnen. Normalerweise sollte dies aktiviert werden - der Performance-Verlust ist in der Regel minimal bis vernachlässigbar."
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_DE "Anisotrope Filterung"
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_DE "Verbessert die Qualität der Texturen auf Oberflächen, die in schrägen Blickwinkeln gegenüber der Kamera liegen. Höhere Werte sind anspruchsvoller für die GPU. Änderungen an dieser Einstellung gelten erst nach einem Neustart."
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_DE "Texturfilterung"
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_DE "Der zu verwendende Texturfiltermodus. Dies kann verwendet werden, um einen bestimmten Texturfiltermodus für alle Texturen zu erzwingen, um ein klareres (oder glatteres) Erscheinungsbild als Standard zu erhalten. Andere Werte als Standardwerte können zu verschiedenen Rendering-Problemen führen. Änderungen an dieser Einstellung werden nur nach einem Neustart angewendet."
#define OPTION_VAL_1_DE "Nächste-Nachbarn erzwingen"
#define OPTION_VAL_2_DE "Linear erzwingen"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_DE "Bildaustausch verzögern"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_DE "Nützlich, um flackernden Bildschirm oder fehlerhafte Videos zu vermeiden. Nicht empfohlen auf langsamen Plattformen."
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_DE "Änderungen der Bildrate erkennen"
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_DE "Benachrichtigt das Frontend, wenn sich die interne Bildrate ändert (z.B. von 60 fps auf 30 fps). Verbessert das Bildtempo in Spielen, die mit gesperrten 30 bps oder 20 bps laufen, aber sollte deaktiviert sein für Spiele mit entsperrten (instabilen) Bildraten (z. B. Ecco the Dolphin, Unreal Tournament). Hinweis: Nicht verfügbar, wenn 'Automatisches Bilder überspringen' aktiviert ist."
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_DE "PowerVR2 Nachbearbeitungs-Filter"
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_DE "Nachbearbeitung des gerenderten Bildes zur Simulation von Effekten speziell für PowerVR2 GPU und analoge Videosignale."
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_DE "Texturen hochskalieren (xBRZ)"
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_DE "Erweitert handgezeichnete 2D Pixel Kunstgrafiken. Darf nur mit 2D verpixelten Spielen verwendet werden."
#define OPTION_VAL_2_O27_DE NULL
#define OPTION_VAL_4_DE NULL
#define OPTION_VAL_6_DE "6 x"
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_DE "Textur-Hochskalierung Max. Gefilterte Größe"
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_DE "Native Tiefeninterpolation"
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_DE "Hilft bei Texturfehlern und tieferen Problemen mit AMD GPUs. Kann auch Intel GPUs in einigen Fällen helfen."
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_DE "Mehrgängiges Rendern"
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_DE "Führt die GPU und CPU in verschiedenen Prozessen aus. Sehr empfohlen."
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_DE "Automatisches Bilder überspringen"
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_DE "Überspringt Bilder automatisch, wenn der Emulator langsam läuft. Hinweis: Diese Einstellung gilt nur, wenn 'Mehrgängiges Rendern' aktiviert ist."
#define OPTION_VAL_SOME_DE NULL
#define OPTION_VAL_MORE_DE NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_DE "Bilder überspringen"
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_DE "Legt die Anzahl der Bilder fest, die zwischen jedem angezeigten Bild übersprungen werden sollen."
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_DE "Breitbild-Cheats (Neustart erforderlich)"
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_DE "Aktiviert Cheats, mit denen bestimmte Spiele im Breitbild-Format angezeigt werden können."
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_DE "Breitbildmodus-Hack"
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_DE "Zeichnet Geometrie außerhalb des normalen 4:3 Seitenverhältnisses. Kann grafische Fehler in den aufgedeckten Bereichen erzeugen."
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_DE "GD-ROM schnelles Laden (ungenau)"
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_DE "Beschleunigt das Laden von GD-ROM."
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_DE "Eigene Texturen laden"
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_DE "Texturen dumpen"
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_DE "Analog-Stick Totzone"
#define OPTION_VAL_0_DE "0 %"
#define OPTION_VAL_5_DE "5 %"
#define OPTION_VAL_10_DE "10 %"
#define OPTION_VAL_15_DE "15 %"
#define OPTION_VAL_20_DE "20 %"
#define OPTION_VAL_25_DE "25 %"
#define OPTION_VAL_30_DE "30 %"
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_DE "Trigger Totzone"
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_DE "Digitale Trigger"
#define CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_DE NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_DE "Aktiviert das Vibrieren des Controllers."
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_DE "Digitale Ausgänge übertragen"
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_DE "Digitale Ausgänge und Force-Feedback-Status auf TCP Port 8000 übertragen. Kompatibel mit der \"-output network\" MAME Option."
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_DE "Pistolen-Einstellungen anzeigen"
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_DE "Aktiviert die Konfiguration der Pistolen Fadenkreuz Anzeigeoptionen. HINWEIS: Schnellmenü muss eingeschaltet sein, damit diese Einstellung wirksam wird."
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_DE "Pistolen Fadenkreuz 1 anzeigen"
#define OPTION_VAL_WHITE_DE "Weiß"
#define OPTION_VAL_RED_DE "Rot"
#define OPTION_VAL_GREEN_DE "Grün"
#define OPTION_VAL_BLUE_DE "Blau"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_DE "Pistolen Fadenkreuz 2 anzeigen"
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_DE "Pistolen Fadenkreuz 3 anzeigen"
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_DE "Pistolen Fadenkreuz 4 anzeigen"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_DE "Visual Memory Units/Systeme (VMU) pro Spiel"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_DE "VMUs pro Spiel"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_DE "Wenn deaktiviert, teilen alle Spiele 4 VMU Speicherdateien (A1, B1, C1, D1) im Systemverzeichnis von RetroArch. Die Einstellung 'VMU A1' erzeugt für jedes Spiel, das gestartet wird, eine einzigartige VMU 'A1' Datei im Speicherverzeichnis von RetroArch. Die Einstellung 'Alle VMUs' erzeugt 4 einzigartige VMU-Dateien (A1, B1, C1, D1) für jedes Spiel, das gestartet wird."
#define OPTION_VAL_VMU_A1_DE NULL
#define OPTION_VAL_ALL_VMUS_DE "Alle VMUs"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_DE "Visual Memory Unit/System (VMU) Anzeigeeinstellungen anzeigen"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_DE "VMU Bildschirmeinstellungen anzeigen"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_DE "Konfiguration von emulierten VMU LCD-Bildschirmen über Sichtbarkeit, Größe, Position und Farbe aktivieren. HINWEIS: Das Schnellmenü muss aktiviert sein, damit diese Einstellung wirksam wird."
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_DE "VMU Bildschirm 1 anzeigen"
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_DE "VMU Bildschirm 1 Position"
#define OPTION_VAL_UPPER_LEFT_DE "Oben links"
#define OPTION_VAL_UPPER_RIGHT_DE "Oben rechts"
#define OPTION_VAL_LOWER_LEFT_DE "Unten links"
#define OPTION_VAL_LOWER_RIGHT_DE "Unten rechts"
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_DE "VMU Bildschirm 1 Grösse"
#define OPTION_VAL_1X_DE "1 x"
#define OPTION_VAL_3X_DE "3 x"
#define OPTION_VAL_5X_DE "5 x"
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_DE "VMU Bildschirm 1 Pixel an Farbe"
#define OPTION_VAL_DEFAULT_ON_00_DE "Standard EIN"
#define OPTION_VAL_DEFAULT_OFF_01_DE "Standard AUS"
#define OPTION_VAL_BLACK_02_DE "Schwarz"
#define OPTION_VAL_LIGHT_BLUE_04_DE "Hellblau"
#define OPTION_VAL_CYAN_06_DE NULL
#define OPTION_VAL_CYAN_BLUE_07_DE "Zyanblau"
#define OPTION_VAL_LIGHT_GREEN_08_DE "Hellgrün"
#define OPTION_VAL_CYAN_GREEN_09_DE "Zyangrün"
#define OPTION_VAL_LIGHT_CYAN_10_DE "Hellzyan"
#define OPTION_VAL_PURPLE_12_DE "Lila"
#define OPTION_VAL_LIGHT_PURPLE_13_DE "Hellviolett"
#define OPTION_VAL_YELLOW_14_DE "Gelb"
#define OPTION_VAL_GRAY_15_DE "Grau"
#define OPTION_VAL_LIGHT_PURPLE_2_16_DE "Hellviolett (2)"
#define OPTION_VAL_LIGHT_GREEN_2_17_DE "Hellgrün (2)"
#define OPTION_VAL_LIGHT_GREEN_3_18_DE "Hellgrün (3)"
#define OPTION_VAL_LIGHT_CYAN_2_19_DE "Hellzyan (2)"
#define OPTION_VAL_LIGHT_RED_2_20_DE "Hellrot (2)"
#define OPTION_VAL_MAGENTA_21_DE NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_DE "Hellviolett (3)"
#define OPTION_VAL_LIGHT_ORANGE_23_DE "Hellorange"
#define OPTION_VAL_ORANGE_24_DE NULL
#define OPTION_VAL_LIGHT_PURPLE_4_25_DE "Hellviolett (4)"
#define OPTION_VAL_LIGHT_YELLOW_26_DE "Hellgelb"
#define OPTION_VAL_LIGHT_YELLOW_2_27_DE "Hellgelb (2)"
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_DE "VMU Bildschirm 1 Pixel aus Farbe"
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_DE "VMU Bildschirm 1 Deckkraft"
#define OPTION_VAL_40_DE "40 %"
#define OPTION_VAL_50_DE "50 %"
#define OPTION_VAL_60_DE "60 %"
#define OPTION_VAL_70_DE "70 %"
#define OPTION_VAL_80_DE "80 %"
#define OPTION_VAL_90_DE "90 %"
#define OPTION_VAL_100_DE "100 %"
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_DE "VMU Bildschirm 2 anzeigen"
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_DE "VMU Bildschirm 2 Position"
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_DE "VMU Bildschirm 2 Grösse"
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_DE "VMU Bildschirm 2 Pixel an Farbe"
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_DE "VMU Bildschirm 2 Pixel aus Farbe"
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_DE "VMU Bildschirm 2 Deckkraft"
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_DE "VMU Bildschirm 3 anzeigen"
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_DE "VMU Bildschirm 3 Position"
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_DE "VMU Bildschirm 3 Grösse"
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_DE "VMU Bildschirm 3 Pixel an Farbe"
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_DE "VMU Bildschirm 3 Pixel aus Farbe"
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_DE "VMU Bildschirm 3 Deckkraft"
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_DE "VMU Bildschirm 4 anzeigen"
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_DE "VMU Bildschirm 4 Position"
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_DE "VMU Bildschirm 4 Grösse"
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_DE "VMU Bildschirm 4 Pixel an Farbe"
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_DE "VMU Bildschirm 4 Pixel aus Farbe"
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_DE "VMU Bildschirm 4 Deckkraft"

struct retro_core_option_v2_category option_cats_de[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_DE,
      CATEGORY_SYSTEM_INFO_0_DE
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_DE,
      CATEGORY_VIDEO_INFO_0_DE
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_DE,
      CATEGORY_PERFORMANCE_INFO_0_DE
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_DE,
      CATEGORY_HACKS_INFO_0_DE
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_DE,
      CATEGORY_INPUT_INFO_0_DE
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_DE,
      CATEGORY_VMU_INFO_0_DE
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_de[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_DE,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_DE },
         { "USA",     OPTION_VAL_USA_DE },
         { "Europe",  OPTION_VAL_EUROPE_DE },
         { "Default", OPTION_VAL_DEFAULT_DE },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_DE,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_DE },
         { "English",  OPTION_VAL_ENGLISH_DE },
         { "German",   OPTION_VAL_GERMAN_DE },
         { "French",   OPTION_VAL_FRENCH_DE },
         { "Spanish",  OPTION_VAL_SPANISH_DE },
         { "Italian",  OPTION_VAL_ITALIAN_DE },
         { "Default",  OPTION_VAL_DEFAULT_DE },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_DE,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_boot_to_bios",
      CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_DE,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_DE,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_force_wince",
      CORE_OPTION_NAME_FORCE_WINCE_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_FORCE_WINCE_INFO_0_DE,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_DE,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_DE,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_DE,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_DE,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_DE,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_DE },
         { "640x480",    OPTION_VAL_640X480_DE },
         { "800x600",    OPTION_VAL_800X600_DE },
         { "960x720",    OPTION_VAL_960X720_DE },
         { "1024x768",   OPTION_VAL_1024X768_DE },
         { "1280x960",   OPTION_VAL_1280X960_DE },
         { "1440x1080",  OPTION_VAL_1440X1080_DE },
         { "1600x1200",  OPTION_VAL_1600X1200_DE },
         { "1920x1440",  OPTION_VAL_1920X1440_DE },
         { "2560x1920",  OPTION_VAL_2560X1920_DE },
         { "2880x2160",  OPTION_VAL_2880X2160_DE },
         { "3200x2400",  OPTION_VAL_3200X2400_DE },
         { "3840x2880",  OPTION_VAL_3840X2880_DE },
         { "4480x3360",  OPTION_VAL_4480X3360_DE },
         { "5120x3840",  OPTION_VAL_5120X3840_DE },
         { "5760x4320",  OPTION_VAL_5760X4320_DE },
         { "6400x4800",  OPTION_VAL_6400X4800_DE },
         { "7040x5280",  OPTION_VAL_7040X5280_DE },
         { "7680x5760",  OPTION_VAL_7680X5760_DE },
         { "8320x6240",  OPTION_VAL_8320X6240_DE },
         { "8960x6720",  OPTION_VAL_8960X6720_DE },
         { "9600x7200",  OPTION_VAL_9600X7200_DE },
         { "10240x7680", OPTION_VAL_10240X7680_DE },
         { "10880x8160", OPTION_VAL_10880X8160_DE },
         { "11520x8640", OPTION_VAL_11520X8640_DE },
         { "12160x9120", OPTION_VAL_12160X9120_DE },
         { "12800x9600", OPTION_VAL_12800X9600_DE },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_DE,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_DE },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_DE },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_DE },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_DE,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_DE },
         { "PAL",     OPTION_VAL_PAL_DE },
         { "PAL_N",   OPTION_VAL_PAL_N_DE },
         { "PAL_M",   OPTION_VAL_PAL_M_DE },
         { "Default", OPTION_VAL_DEFAULT_DE },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_DE,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_DE },
         { "vertical",   OPTION_VAL_VERTICAL_DE },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_DE,
      NULL,
      "",
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_DE },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_DE },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_DE },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_DE,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_DE },
         { "1GB",   OPTION_VAL_1GB_DE },
         { "2GB",   OPTION_VAL_2GB_DE },
         { "4GB",   OPTION_VAL_4GB_DE },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_DE,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_DE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_DE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_DE,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_DE,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_DE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_DE,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_DE,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_DE },
         { "1",  OPTION_VAL_1_DE },
         { "2",  OPTION_VAL_2_DE },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_DE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_DE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_DE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_DE,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O27_DE },
         { "4",  OPTION_VAL_4_DE },
         { "6",  OPTION_VAL_6_DE },
         { NULL, NULL },
      },
      "1",
   },
   {/* TODO: needs clarification */
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_DE,
      NULL,
      "",
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_DE,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_DE,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_DE,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_DE,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_DE },
         { "more", OPTION_VAL_MORE_DE },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_DE,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_DE,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_DE,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_DE,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_DE,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: probably needs explanation */
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_DE,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_DE,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_DE },
         { "5%",  OPTION_VAL_5_DE },
         { "10%", OPTION_VAL_10_DE },
         { "15%", OPTION_VAL_15_DE },
         { "20%", OPTION_VAL_20_DE },
         { "25%", OPTION_VAL_25_DE },
         { "30%", OPTION_VAL_30_DE },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_DE,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_DE },
         { "5%",  OPTION_VAL_5_DE },
         { "10%", OPTION_VAL_10_DE },
         { "15%", OPTION_VAL_15_DE },
         { "20%", OPTION_VAL_20_DE },
         { "25%", OPTION_VAL_25_DE },
         { "30%", OPTION_VAL_30_DE },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_DE,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_purupuru",
      CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_DE,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_DE,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_DE,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_DE,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_DE },
         { "Red",      OPTION_VAL_RED_DE },
         { "Green",    OPTION_VAL_GREEN_DE },
         { "Blue",     OPTION_VAL_BLUE_DE },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_DE,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_DE },
         { "Red",      OPTION_VAL_RED_DE },
         { "Green",    OPTION_VAL_GREEN_DE },
         { "Blue",     OPTION_VAL_BLUE_DE },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_DE,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_DE },
         { "Red",      OPTION_VAL_RED_DE },
         { "Green",    OPTION_VAL_GREEN_DE },
         { "Blue",     OPTION_VAL_BLUE_DE },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_DE,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_DE },
         { "Red",      OPTION_VAL_RED_DE },
         { "Green",    OPTION_VAL_GREEN_DE },
         { "Blue",     OPTION_VAL_BLUE_DE },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_DE,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_DE,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_DE,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_DE },
         { "All VMUs", OPTION_VAL_ALL_VMUS_DE },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_DE,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_DE,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_DE,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_DE },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_DE },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_DE },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_DE },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_DE },
         { "2x", OPTION_VAL_2_O27_DE },
         { "3x", OPTION_VAL_3X_DE },
         { "4x", OPTION_VAL_4_DE },
         { "5x", OPTION_VAL_5X_DE },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_DE },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_DE },
         { "BLACK 02",          OPTION_VAL_BLACK_02_DE },
         { "BLUE 03",           OPTION_VAL_BLUE_DE },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_DE },
         { "GREEN 05",          OPTION_VAL_GREEN_DE },
         { "CYAN 06",           OPTION_VAL_CYAN_06_DE },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_DE },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_DE },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_DE },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_DE },
         { "RED 11",            OPTION_VAL_RED_DE },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_DE },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_DE },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_DE },
         { "GRAY 15",           OPTION_VAL_GRAY_15_DE },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_DE },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_DE },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_DE },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_DE },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_DE },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_DE },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_DE },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_DE },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_DE },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_DE },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_DE },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_DE },
         { "WHITE 28",          OPTION_VAL_WHITE_DE },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_DE },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_DE },
         { "BLACK 02",          OPTION_VAL_BLACK_02_DE },
         { "BLUE 03",           OPTION_VAL_BLUE_DE },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_DE },
         { "GREEN 05",          OPTION_VAL_GREEN_DE },
         { "CYAN 06",           OPTION_VAL_CYAN_06_DE },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_DE },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_DE },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_DE },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_DE },
         { "RED 11",            OPTION_VAL_RED_DE },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_DE },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_DE },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_DE },
         { "GRAY 15",           OPTION_VAL_GRAY_15_DE },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_DE },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_DE },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_DE },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_DE },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_DE },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_DE },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_DE },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_DE },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_DE },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_DE },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_DE },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_DE },
         { "WHITE 28",          OPTION_VAL_WHITE_DE },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_DE },
         { "20%",  OPTION_VAL_20_DE },
         { "30%",  OPTION_VAL_30_DE },
         { "40%",  OPTION_VAL_40_DE },
         { "50%",  OPTION_VAL_50_DE },
         { "60%",  OPTION_VAL_60_DE },
         { "70%",  OPTION_VAL_70_DE },
         { "80%",  OPTION_VAL_80_DE },
         { "90%",  OPTION_VAL_90_DE },
         { "100%", OPTION_VAL_100_DE },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_DE },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_DE },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_DE },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_DE },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_DE },
         { "2x", OPTION_VAL_2_O27_DE },
         { "3x", OPTION_VAL_3X_DE },
         { "4x", OPTION_VAL_4_DE },
         { "5x", OPTION_VAL_5X_DE },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_DE },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_DE },
         { "BLACK 02",          OPTION_VAL_BLACK_02_DE },
         { "BLUE 03",           OPTION_VAL_BLUE_DE },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_DE },
         { "GREEN 05",          OPTION_VAL_GREEN_DE },
         { "CYAN 06",           OPTION_VAL_CYAN_06_DE },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_DE },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_DE },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_DE },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_DE },
         { "RED 11",            OPTION_VAL_RED_DE },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_DE },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_DE },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_DE },
         { "GRAY 15",           OPTION_VAL_GRAY_15_DE },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_DE },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_DE },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_DE },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_DE },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_DE },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_DE },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_DE },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_DE },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_DE },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_DE },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_DE },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_DE },
         { "WHITE 28",          OPTION_VAL_WHITE_DE },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_DE },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_DE },
         { "BLACK 02",          OPTION_VAL_BLACK_02_DE },
         { "BLUE 03",           OPTION_VAL_BLUE_DE },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_DE },
         { "GREEN 05",          OPTION_VAL_GREEN_DE },
         { "CYAN 06",           OPTION_VAL_CYAN_06_DE },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_DE },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_DE },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_DE },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_DE },
         { "RED 11",            OPTION_VAL_RED_DE },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_DE },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_DE },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_DE },
         { "GRAY 15",           OPTION_VAL_GRAY_15_DE },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_DE },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_DE },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_DE },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_DE },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_DE },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_DE },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_DE },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_DE },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_DE },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_DE },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_DE },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_DE },
         { "WHITE 28",          OPTION_VAL_WHITE_DE },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_DE },
         { "20%",  OPTION_VAL_20_DE },
         { "30%",  OPTION_VAL_30_DE },
         { "40%",  OPTION_VAL_40_DE },
         { "50%",  OPTION_VAL_50_DE },
         { "60%",  OPTION_VAL_60_DE },
         { "70%",  OPTION_VAL_70_DE },
         { "80%",  OPTION_VAL_80_DE },
         { "90%",  OPTION_VAL_90_DE },
         { "100%", OPTION_VAL_100_DE },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_DE },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_DE },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_DE },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_DE },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_DE },
         { "2x", OPTION_VAL_2_O27_DE },
         { "3x", OPTION_VAL_3X_DE },
         { "4x", OPTION_VAL_4_DE },
         { "5x", OPTION_VAL_5X_DE },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_DE },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_DE },
         { "BLACK 02",          OPTION_VAL_BLACK_02_DE },
         { "BLUE 03",           OPTION_VAL_BLUE_DE },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_DE },
         { "GREEN 05",          OPTION_VAL_GREEN_DE },
         { "CYAN 06",           OPTION_VAL_CYAN_06_DE },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_DE },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_DE },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_DE },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_DE },
         { "RED 11",            OPTION_VAL_RED_DE },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_DE },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_DE },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_DE },
         { "GRAY 15",           OPTION_VAL_GRAY_15_DE },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_DE },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_DE },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_DE },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_DE },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_DE },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_DE },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_DE },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_DE },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_DE },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_DE },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_DE },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_DE },
         { "WHITE 28",          OPTION_VAL_WHITE_DE },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_DE },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_DE },
         { "BLACK 02",          OPTION_VAL_BLACK_02_DE },
         { "BLUE 03",           OPTION_VAL_BLUE_DE },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_DE },
         { "GREEN 05",          OPTION_VAL_GREEN_DE },
         { "CYAN 06",           OPTION_VAL_CYAN_06_DE },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_DE },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_DE },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_DE },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_DE },
         { "RED 11",            OPTION_VAL_RED_DE },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_DE },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_DE },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_DE },
         { "GRAY 15",           OPTION_VAL_GRAY_15_DE },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_DE },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_DE },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_DE },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_DE },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_DE },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_DE },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_DE },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_DE },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_DE },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_DE },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_DE },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_DE },
         { "WHITE 28",          OPTION_VAL_WHITE_DE },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_DE },
         { "20%",  OPTION_VAL_20_DE },
         { "30%",  OPTION_VAL_30_DE },
         { "40%",  OPTION_VAL_40_DE },
         { "50%",  OPTION_VAL_50_DE },
         { "60%",  OPTION_VAL_60_DE },
         { "70%",  OPTION_VAL_70_DE },
         { "80%",  OPTION_VAL_80_DE },
         { "90%",  OPTION_VAL_90_DE },
         { "100%", OPTION_VAL_100_DE },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_DE },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_DE },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_DE },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_DE },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_DE },
         { "2x", OPTION_VAL_2_O27_DE },
         { "3x", OPTION_VAL_3X_DE },
         { "4x", OPTION_VAL_4_DE },
         { "5x", OPTION_VAL_5X_DE },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_DE },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_DE },
         { "BLACK 02",          OPTION_VAL_BLACK_02_DE },
         { "BLUE 03",           OPTION_VAL_BLUE_DE },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_DE },
         { "GREEN 05",          OPTION_VAL_GREEN_DE },
         { "CYAN 06",           OPTION_VAL_CYAN_06_DE },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_DE },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_DE },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_DE },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_DE },
         { "RED 11",            OPTION_VAL_RED_DE },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_DE },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_DE },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_DE },
         { "GRAY 15",           OPTION_VAL_GRAY_15_DE },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_DE },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_DE },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_DE },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_DE },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_DE },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_DE },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_DE },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_DE },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_DE },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_DE },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_DE },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_DE },
         { "WHITE 28",          OPTION_VAL_WHITE_DE },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_DE },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_DE },
         { "BLACK 02",          OPTION_VAL_BLACK_02_DE },
         { "BLUE 03",           OPTION_VAL_BLUE_DE },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_DE },
         { "GREEN 05",          OPTION_VAL_GREEN_DE },
         { "CYAN 06",           OPTION_VAL_CYAN_06_DE },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_DE },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_DE },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_DE },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_DE },
         { "RED 11",            OPTION_VAL_RED_DE },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_DE },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_DE },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_DE },
         { "GRAY 15",           OPTION_VAL_GRAY_15_DE },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_DE },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_DE },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_DE },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_DE },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_DE },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_DE },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_DE },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_DE },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_DE },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_DE },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_DE },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_DE },
         { "WHITE 28",          OPTION_VAL_WHITE_DE },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_DE },
         { "20%",  OPTION_VAL_20_DE },
         { "30%",  OPTION_VAL_30_DE },
         { "40%",  OPTION_VAL_40_DE },
         { "50%",  OPTION_VAL_50_DE },
         { "60%",  OPTION_VAL_60_DE },
         { "70%",  OPTION_VAL_70_DE },
         { "80%",  OPTION_VAL_80_DE },
         { "90%",  OPTION_VAL_90_DE },
         { "100%", OPTION_VAL_100_DE },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_de = {
   option_cats_de,
   option_defs_de
};

/* RETRO_LANGUAGE_EL */

#define CATEGORY_SYSTEM_LABEL_EL "Σύστημα"
#define CATEGORY_SYSTEM_INFO_0_EL NULL
#define CATEGORY_VIDEO_LABEL_EL "Οδηγός Βίντεο"
#define CATEGORY_VIDEO_INFO_0_EL NULL
#define CATEGORY_PERFORMANCE_LABEL_EL "Επίδοση"
#define CATEGORY_PERFORMANCE_INFO_0_EL NULL
#define CATEGORY_HACKS_LABEL_EL NULL
#define CATEGORY_HACKS_INFO_0_EL NULL
#define CATEGORY_INPUT_LABEL_EL "Οδηγός Εισαγωγής"
#define CATEGORY_INPUT_INFO_0_EL NULL
#define CATEGORY_VMU_LABEL_EL NULL
#define CATEGORY_VMU_INFO_0_EL NULL
#define CORE_OPTION_NAME_REGION_LABEL_EL "Περιοχή"
#define OPTION_VAL_JAPAN_EL NULL
#define OPTION_VAL_USA_EL NULL
#define OPTION_VAL_EUROPE_EL NULL
#define OPTION_VAL_DEFAULT_EL "Προκαθορισμένο"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_EL "Γλώσσα"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_EL NULL
#define OPTION_VAL_JAPANESE_EL "Ιαπωνικά"
#define OPTION_VAL_ENGLISH_EL "Αγγλικά"
#define OPTION_VAL_GERMAN_EL "Γερμανός"
#define OPTION_VAL_FRENCH_EL "Γαλλική γλώσσα"
#define OPTION_VAL_SPANISH_EL "Ισπανικά"
#define OPTION_VAL_ITALIAN_EL "Ιταλικά"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_EL NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_EL NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_EL NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_EL NULL
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_EL NULL
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_EL NULL
#define CORE_OPTION_NAME_FORCE_WINCE_LABEL_EL NULL
#define CORE_OPTION_NAME_FORCE_WINCE_INFO_0_EL NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_EL NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_EL NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_EL NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_EL NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_EL NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_EL NULL
#define CORE_OPTION_NAME_UPNP_LABEL_EL NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_EL NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_EL NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_EL NULL
#define OPTION_VAL_320X240_EL NULL
#define OPTION_VAL_640X480_EL NULL
#define OPTION_VAL_800X600_EL NULL
#define OPTION_VAL_960X720_EL NULL
#define OPTION_VAL_1024X768_EL NULL
#define OPTION_VAL_1280X960_EL NULL
#define OPTION_VAL_1440X1080_EL NULL
#define OPTION_VAL_1600X1200_EL NULL
#define OPTION_VAL_1920X1440_EL NULL
#define OPTION_VAL_2560X1920_EL NULL
#define OPTION_VAL_2880X2160_EL NULL
#define OPTION_VAL_3200X2400_EL NULL
#define OPTION_VAL_3840X2880_EL NULL
#define OPTION_VAL_4480X3360_EL NULL
#define OPTION_VAL_5120X3840_EL NULL
#define OPTION_VAL_5760X4320_EL NULL
#define OPTION_VAL_6400X4800_EL NULL
#define OPTION_VAL_7040X5280_EL NULL
#define OPTION_VAL_7680X5760_EL NULL
#define OPTION_VAL_8320X6240_EL NULL
#define OPTION_VAL_8960X6720_EL NULL
#define OPTION_VAL_9600X7200_EL NULL
#define OPTION_VAL_10240X7680_EL NULL
#define OPTION_VAL_10880X8160_EL NULL
#define OPTION_VAL_11520X8640_EL NULL
#define OPTION_VAL_12160X9120_EL NULL
#define OPTION_VAL_12800X9600_EL NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_EL NULL
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_EL NULL
#define OPTION_VAL_VGA_EL NULL
#define OPTION_VAL_TV_RGB_EL NULL
#define OPTION_VAL_TV_COMPOSITE_EL NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_EL NULL
#define OPTION_VAL_NTSC_EL NULL
#define OPTION_VAL_PAL_EL NULL
#define OPTION_VAL_PAL_N_EL NULL
#define OPTION_VAL_PAL_M_EL NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_EL "Περιστροφή οθόνης"
#define OPTION_VAL_HORIZONTAL_EL NULL
#define OPTION_VAL_VERTICAL_EL NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_EL NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_EL NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_EL NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_EL NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_EL NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_EL NULL
#define OPTION_VAL_512MB_EL NULL
#define OPTION_VAL_1GB_EL NULL
#define OPTION_VAL_2GB_EL NULL
#define OPTION_VAL_4GB_EL NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_EL NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_EL NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_EL NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_EL NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_EL NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_EL NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_EL NULL
#define CORE_OPTION_NAME_FOG_LABEL_EL NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_EL NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_EL NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_EL NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_EL NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_EL NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_EL NULL
#define OPTION_VAL_1_EL NULL
#define OPTION_VAL_2_EL NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_EL NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_EL NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_EL NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_EL NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_EL NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_EL NULL
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_EL NULL
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_EL NULL
#define OPTION_VAL_2_O27_EL NULL
#define OPTION_VAL_4_EL NULL
#define OPTION_VAL_6_EL NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_EL NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_EL NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_EL NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_EL NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_EL NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_EL NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_EL NULL
#define OPTION_VAL_SOME_EL "Φυσιολογικά"
#define OPTION_VAL_MORE_EL NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_EL NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_EL NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_EL NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_EL NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_EL NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_EL NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_EL NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_EL NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_EL NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_EL NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_EL NULL
#define OPTION_VAL_0_EL NULL
#define OPTION_VAL_5_EL NULL
#define OPTION_VAL_10_EL NULL
#define OPTION_VAL_15_EL NULL
#define OPTION_VAL_20_EL NULL
#define OPTION_VAL_25_EL NULL
#define OPTION_VAL_30_EL NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_EL NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_EL NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_EL NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_EL NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_EL NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_EL NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_EL NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_EL NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_EL NULL
#define OPTION_VAL_WHITE_EL NULL
#define OPTION_VAL_RED_EL "Κόκκινο"
#define OPTION_VAL_GREEN_EL "Πράσινο"
#define OPTION_VAL_BLUE_EL "Μπλε"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_EL NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_EL NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_EL NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_EL NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_EL NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_EL NULL
#define OPTION_VAL_VMU_A1_EL NULL
#define OPTION_VAL_ALL_VMUS_EL NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_EL NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_EL NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_EL NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_EL NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_EL NULL
#define OPTION_VAL_UPPER_LEFT_EL NULL
#define OPTION_VAL_UPPER_RIGHT_EL NULL
#define OPTION_VAL_LOWER_LEFT_EL NULL
#define OPTION_VAL_LOWER_RIGHT_EL NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_EL NULL
#define OPTION_VAL_1X_EL NULL
#define OPTION_VAL_3X_EL NULL
#define OPTION_VAL_5X_EL NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_EL NULL
#define OPTION_VAL_DEFAULT_ON_00_EL NULL
#define OPTION_VAL_DEFAULT_OFF_01_EL NULL
#define OPTION_VAL_BLACK_02_EL NULL
#define OPTION_VAL_LIGHT_BLUE_04_EL NULL
#define OPTION_VAL_CYAN_06_EL NULL
#define OPTION_VAL_CYAN_BLUE_07_EL NULL
#define OPTION_VAL_LIGHT_GREEN_08_EL NULL
#define OPTION_VAL_CYAN_GREEN_09_EL NULL
#define OPTION_VAL_LIGHT_CYAN_10_EL NULL
#define OPTION_VAL_PURPLE_12_EL NULL
#define OPTION_VAL_LIGHT_PURPLE_13_EL NULL
#define OPTION_VAL_YELLOW_14_EL "Κίτρινο"
#define OPTION_VAL_GRAY_15_EL NULL
#define OPTION_VAL_LIGHT_PURPLE_2_16_EL NULL
#define OPTION_VAL_LIGHT_GREEN_2_17_EL NULL
#define OPTION_VAL_LIGHT_GREEN_3_18_EL NULL
#define OPTION_VAL_LIGHT_CYAN_2_19_EL NULL
#define OPTION_VAL_LIGHT_RED_2_20_EL NULL
#define OPTION_VAL_MAGENTA_21_EL NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_EL NULL
#define OPTION_VAL_LIGHT_ORANGE_23_EL NULL
#define OPTION_VAL_ORANGE_24_EL NULL
#define OPTION_VAL_LIGHT_PURPLE_4_25_EL NULL
#define OPTION_VAL_LIGHT_YELLOW_26_EL NULL
#define OPTION_VAL_LIGHT_YELLOW_2_27_EL NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_EL NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_EL NULL
#define OPTION_VAL_40_EL NULL
#define OPTION_VAL_50_EL NULL
#define OPTION_VAL_60_EL NULL
#define OPTION_VAL_70_EL NULL
#define OPTION_VAL_80_EL "το 80%"
#define OPTION_VAL_90_EL NULL
#define OPTION_VAL_100_EL NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_EL NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_EL NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_EL NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_EL NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_EL NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_EL NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_EL NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_EL NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_EL NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_EL NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_EL NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_EL NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_EL NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_EL NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_EL NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_EL NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_EL NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_EL NULL

struct retro_core_option_v2_category option_cats_el[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_EL,
      CATEGORY_SYSTEM_INFO_0_EL
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_EL,
      CATEGORY_VIDEO_INFO_0_EL
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_EL,
      CATEGORY_PERFORMANCE_INFO_0_EL
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_EL,
      CATEGORY_HACKS_INFO_0_EL
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_EL,
      CATEGORY_INPUT_INFO_0_EL
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_EL,
      CATEGORY_VMU_INFO_0_EL
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_el[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_EL,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_EL },
         { "USA",     OPTION_VAL_USA_EL },
         { "Europe",  OPTION_VAL_EUROPE_EL },
         { "Default", OPTION_VAL_DEFAULT_EL },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_EL,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_EL },
         { "English",  OPTION_VAL_ENGLISH_EL },
         { "German",   OPTION_VAL_GERMAN_EL },
         { "French",   OPTION_VAL_FRENCH_EL },
         { "Spanish",  OPTION_VAL_SPANISH_EL },
         { "Italian",  OPTION_VAL_ITALIAN_EL },
         { "Default",  OPTION_VAL_DEFAULT_EL },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_EL,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_boot_to_bios",
      CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_EL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_EL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_force_wince",
      CORE_OPTION_NAME_FORCE_WINCE_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_FORCE_WINCE_INFO_0_EL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_EL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_EL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_EL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_EL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_EL,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_EL },
         { "640x480",    OPTION_VAL_640X480_EL },
         { "800x600",    OPTION_VAL_800X600_EL },
         { "960x720",    OPTION_VAL_960X720_EL },
         { "1024x768",   OPTION_VAL_1024X768_EL },
         { "1280x960",   OPTION_VAL_1280X960_EL },
         { "1440x1080",  OPTION_VAL_1440X1080_EL },
         { "1600x1200",  OPTION_VAL_1600X1200_EL },
         { "1920x1440",  OPTION_VAL_1920X1440_EL },
         { "2560x1920",  OPTION_VAL_2560X1920_EL },
         { "2880x2160",  OPTION_VAL_2880X2160_EL },
         { "3200x2400",  OPTION_VAL_3200X2400_EL },
         { "3840x2880",  OPTION_VAL_3840X2880_EL },
         { "4480x3360",  OPTION_VAL_4480X3360_EL },
         { "5120x3840",  OPTION_VAL_5120X3840_EL },
         { "5760x4320",  OPTION_VAL_5760X4320_EL },
         { "6400x4800",  OPTION_VAL_6400X4800_EL },
         { "7040x5280",  OPTION_VAL_7040X5280_EL },
         { "7680x5760",  OPTION_VAL_7680X5760_EL },
         { "8320x6240",  OPTION_VAL_8320X6240_EL },
         { "8960x6720",  OPTION_VAL_8960X6720_EL },
         { "9600x7200",  OPTION_VAL_9600X7200_EL },
         { "10240x7680", OPTION_VAL_10240X7680_EL },
         { "10880x8160", OPTION_VAL_10880X8160_EL },
         { "11520x8640", OPTION_VAL_11520X8640_EL },
         { "12160x9120", OPTION_VAL_12160X9120_EL },
         { "12800x9600", OPTION_VAL_12800X9600_EL },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_EL,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_EL },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_EL },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_EL },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_EL,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_EL },
         { "PAL",     OPTION_VAL_PAL_EL },
         { "PAL_N",   OPTION_VAL_PAL_N_EL },
         { "PAL_M",   OPTION_VAL_PAL_M_EL },
         { "Default", OPTION_VAL_DEFAULT_EL },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_EL,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_EL },
         { "vertical",   OPTION_VAL_VERTICAL_EL },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_EL,
      NULL,
      "",
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_EL },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_EL },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_EL },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_EL,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_EL },
         { "1GB",   OPTION_VAL_1GB_EL },
         { "2GB",   OPTION_VAL_2GB_EL },
         { "4GB",   OPTION_VAL_4GB_EL },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_EL,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_EL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_EL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_EL,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_EL,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_EL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_EL,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_EL,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_EL },
         { "1",  OPTION_VAL_1_EL },
         { "2",  OPTION_VAL_2_EL },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_EL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_EL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_EL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_EL,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O27_EL },
         { "4",  OPTION_VAL_4_EL },
         { "6",  OPTION_VAL_6_EL },
         { NULL, NULL },
      },
      "1",
   },
   {/* TODO: needs clarification */
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_EL,
      NULL,
      "",
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_EL,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_EL,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_EL,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_EL,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_EL },
         { "more", OPTION_VAL_MORE_EL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_EL,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_EL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_EL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_EL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_EL,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: probably needs explanation */
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_EL,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_EL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_EL },
         { "5%",  OPTION_VAL_5_EL },
         { "10%", OPTION_VAL_10_EL },
         { "15%", OPTION_VAL_15_EL },
         { "20%", OPTION_VAL_20_EL },
         { "25%", OPTION_VAL_25_EL },
         { "30%", OPTION_VAL_30_EL },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_EL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_EL },
         { "5%",  OPTION_VAL_5_EL },
         { "10%", OPTION_VAL_10_EL },
         { "15%", OPTION_VAL_15_EL },
         { "20%", OPTION_VAL_20_EL },
         { "25%", OPTION_VAL_25_EL },
         { "30%", OPTION_VAL_30_EL },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_EL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_purupuru",
      CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_EL,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_EL,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_EL,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_EL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_EL },
         { "Red",      OPTION_VAL_RED_EL },
         { "Green",    OPTION_VAL_GREEN_EL },
         { "Blue",     OPTION_VAL_BLUE_EL },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_EL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_EL },
         { "Red",      OPTION_VAL_RED_EL },
         { "Green",    OPTION_VAL_GREEN_EL },
         { "Blue",     OPTION_VAL_BLUE_EL },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_EL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_EL },
         { "Red",      OPTION_VAL_RED_EL },
         { "Green",    OPTION_VAL_GREEN_EL },
         { "Blue",     OPTION_VAL_BLUE_EL },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_EL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_EL },
         { "Red",      OPTION_VAL_RED_EL },
         { "Green",    OPTION_VAL_GREEN_EL },
         { "Blue",     OPTION_VAL_BLUE_EL },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_EL,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_EL,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_EL,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_EL },
         { "All VMUs", OPTION_VAL_ALL_VMUS_EL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_EL,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_EL,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_EL,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_EL },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_EL },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_EL },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_EL },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_EL },
         { "2x", OPTION_VAL_2_O27_EL },
         { "3x", OPTION_VAL_3X_EL },
         { "4x", OPTION_VAL_4_EL },
         { "5x", OPTION_VAL_5X_EL },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EL },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EL },
         { "BLUE 03",           OPTION_VAL_BLUE_EL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EL },
         { "GREEN 05",          OPTION_VAL_GREEN_EL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EL },
         { "RED 11",            OPTION_VAL_RED_EL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EL },
         { "WHITE 28",          OPTION_VAL_WHITE_EL },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EL },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EL },
         { "BLUE 03",           OPTION_VAL_BLUE_EL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EL },
         { "GREEN 05",          OPTION_VAL_GREEN_EL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EL },
         { "RED 11",            OPTION_VAL_RED_EL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EL },
         { "WHITE 28",          OPTION_VAL_WHITE_EL },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_EL },
         { "20%",  OPTION_VAL_20_EL },
         { "30%",  OPTION_VAL_30_EL },
         { "40%",  OPTION_VAL_40_EL },
         { "50%",  OPTION_VAL_50_EL },
         { "60%",  OPTION_VAL_60_EL },
         { "70%",  OPTION_VAL_70_EL },
         { "80%",  OPTION_VAL_80_EL },
         { "90%",  OPTION_VAL_90_EL },
         { "100%", OPTION_VAL_100_EL },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_EL },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_EL },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_EL },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_EL },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_EL },
         { "2x", OPTION_VAL_2_O27_EL },
         { "3x", OPTION_VAL_3X_EL },
         { "4x", OPTION_VAL_4_EL },
         { "5x", OPTION_VAL_5X_EL },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EL },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EL },
         { "BLUE 03",           OPTION_VAL_BLUE_EL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EL },
         { "GREEN 05",          OPTION_VAL_GREEN_EL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EL },
         { "RED 11",            OPTION_VAL_RED_EL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EL },
         { "WHITE 28",          OPTION_VAL_WHITE_EL },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EL },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EL },
         { "BLUE 03",           OPTION_VAL_BLUE_EL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EL },
         { "GREEN 05",          OPTION_VAL_GREEN_EL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EL },
         { "RED 11",            OPTION_VAL_RED_EL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EL },
         { "WHITE 28",          OPTION_VAL_WHITE_EL },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_EL },
         { "20%",  OPTION_VAL_20_EL },
         { "30%",  OPTION_VAL_30_EL },
         { "40%",  OPTION_VAL_40_EL },
         { "50%",  OPTION_VAL_50_EL },
         { "60%",  OPTION_VAL_60_EL },
         { "70%",  OPTION_VAL_70_EL },
         { "80%",  OPTION_VAL_80_EL },
         { "90%",  OPTION_VAL_90_EL },
         { "100%", OPTION_VAL_100_EL },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_EL },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_EL },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_EL },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_EL },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_EL },
         { "2x", OPTION_VAL_2_O27_EL },
         { "3x", OPTION_VAL_3X_EL },
         { "4x", OPTION_VAL_4_EL },
         { "5x", OPTION_VAL_5X_EL },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EL },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EL },
         { "BLUE 03",           OPTION_VAL_BLUE_EL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EL },
         { "GREEN 05",          OPTION_VAL_GREEN_EL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EL },
         { "RED 11",            OPTION_VAL_RED_EL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EL },
         { "WHITE 28",          OPTION_VAL_WHITE_EL },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EL },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EL },
         { "BLUE 03",           OPTION_VAL_BLUE_EL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EL },
         { "GREEN 05",          OPTION_VAL_GREEN_EL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EL },
         { "RED 11",            OPTION_VAL_RED_EL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EL },
         { "WHITE 28",          OPTION_VAL_WHITE_EL },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_EL },
         { "20%",  OPTION_VAL_20_EL },
         { "30%",  OPTION_VAL_30_EL },
         { "40%",  OPTION_VAL_40_EL },
         { "50%",  OPTION_VAL_50_EL },
         { "60%",  OPTION_VAL_60_EL },
         { "70%",  OPTION_VAL_70_EL },
         { "80%",  OPTION_VAL_80_EL },
         { "90%",  OPTION_VAL_90_EL },
         { "100%", OPTION_VAL_100_EL },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_EL },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_EL },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_EL },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_EL },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_EL },
         { "2x", OPTION_VAL_2_O27_EL },
         { "3x", OPTION_VAL_3X_EL },
         { "4x", OPTION_VAL_4_EL },
         { "5x", OPTION_VAL_5X_EL },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EL },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EL },
         { "BLUE 03",           OPTION_VAL_BLUE_EL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EL },
         { "GREEN 05",          OPTION_VAL_GREEN_EL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EL },
         { "RED 11",            OPTION_VAL_RED_EL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EL },
         { "WHITE 28",          OPTION_VAL_WHITE_EL },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EL },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EL },
         { "BLUE 03",           OPTION_VAL_BLUE_EL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EL },
         { "GREEN 05",          OPTION_VAL_GREEN_EL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EL },
         { "RED 11",            OPTION_VAL_RED_EL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EL },
         { "WHITE 28",          OPTION_VAL_WHITE_EL },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_EL },
         { "20%",  OPTION_VAL_20_EL },
         { "30%",  OPTION_VAL_30_EL },
         { "40%",  OPTION_VAL_40_EL },
         { "50%",  OPTION_VAL_50_EL },
         { "60%",  OPTION_VAL_60_EL },
         { "70%",  OPTION_VAL_70_EL },
         { "80%",  OPTION_VAL_80_EL },
         { "90%",  OPTION_VAL_90_EL },
         { "100%", OPTION_VAL_100_EL },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_el = {
   option_cats_el,
   option_defs_el
};

/* RETRO_LANGUAGE_EN */

#define CATEGORY_SYSTEM_LABEL_EN NULL
#define CATEGORY_SYSTEM_INFO_0_EN NULL
#define CATEGORY_VIDEO_LABEL_EN NULL
#define CATEGORY_VIDEO_INFO_0_EN NULL
#define CATEGORY_PERFORMANCE_LABEL_EN NULL
#define CATEGORY_PERFORMANCE_INFO_0_EN NULL
#define CATEGORY_HACKS_LABEL_EN NULL
#define CATEGORY_HACKS_INFO_0_EN NULL
#define CATEGORY_INPUT_LABEL_EN NULL
#define CATEGORY_INPUT_INFO_0_EN NULL
#define CATEGORY_VMU_LABEL_EN NULL
#define CATEGORY_VMU_INFO_0_EN NULL
#define CORE_OPTION_NAME_REGION_LABEL_EN NULL
#define OPTION_VAL_JAPAN_EN NULL
#define OPTION_VAL_USA_EN NULL
#define OPTION_VAL_EUROPE_EN NULL
#define OPTION_VAL_DEFAULT_EN NULL
#define CORE_OPTION_NAME_LANGUAGE_LABEL_EN NULL
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_EN NULL
#define OPTION_VAL_JAPANESE_EN NULL
#define OPTION_VAL_ENGLISH_EN NULL
#define OPTION_VAL_GERMAN_EN NULL
#define OPTION_VAL_FRENCH_EN NULL
#define OPTION_VAL_SPANISH_EN NULL
#define OPTION_VAL_ITALIAN_EN NULL
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_EN NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_EN NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_EN NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_EN NULL
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_EN NULL
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_EN NULL
#define CORE_OPTION_NAME_FORCE_WINCE_LABEL_EN NULL
#define CORE_OPTION_NAME_FORCE_WINCE_INFO_0_EN NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_EN NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_EN NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_EN NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_EN NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_EN NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_EN NULL
#define CORE_OPTION_NAME_UPNP_LABEL_EN NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_EN NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_EN NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_EN NULL
#define OPTION_VAL_320X240_EN NULL
#define OPTION_VAL_640X480_EN NULL
#define OPTION_VAL_800X600_EN NULL
#define OPTION_VAL_960X720_EN NULL
#define OPTION_VAL_1024X768_EN NULL
#define OPTION_VAL_1280X960_EN NULL
#define OPTION_VAL_1440X1080_EN NULL
#define OPTION_VAL_1600X1200_EN NULL
#define OPTION_VAL_1920X1440_EN NULL
#define OPTION_VAL_2560X1920_EN NULL
#define OPTION_VAL_2880X2160_EN NULL
#define OPTION_VAL_3200X2400_EN NULL
#define OPTION_VAL_3840X2880_EN NULL
#define OPTION_VAL_4480X3360_EN NULL
#define OPTION_VAL_5120X3840_EN NULL
#define OPTION_VAL_5760X4320_EN NULL
#define OPTION_VAL_6400X4800_EN NULL
#define OPTION_VAL_7040X5280_EN NULL
#define OPTION_VAL_7680X5760_EN NULL
#define OPTION_VAL_8320X6240_EN NULL
#define OPTION_VAL_8960X6720_EN NULL
#define OPTION_VAL_9600X7200_EN NULL
#define OPTION_VAL_10240X7680_EN NULL
#define OPTION_VAL_10880X8160_EN NULL
#define OPTION_VAL_11520X8640_EN NULL
#define OPTION_VAL_12160X9120_EN NULL
#define OPTION_VAL_12800X9600_EN NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_EN NULL
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_EN NULL
#define OPTION_VAL_VGA_EN NULL
#define OPTION_VAL_TV_RGB_EN NULL
#define OPTION_VAL_TV_COMPOSITE_EN NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_EN NULL
#define OPTION_VAL_NTSC_EN NULL
#define OPTION_VAL_PAL_EN NULL
#define OPTION_VAL_PAL_N_EN NULL
#define OPTION_VAL_PAL_M_EN NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_EN NULL
#define OPTION_VAL_HORIZONTAL_EN NULL
#define OPTION_VAL_VERTICAL_EN NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_EN NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_EN NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_EN NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_EN NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_EN NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_EN NULL
#define OPTION_VAL_512MB_EN NULL
#define OPTION_VAL_1GB_EN NULL
#define OPTION_VAL_2GB_EN NULL
#define OPTION_VAL_4GB_EN NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_EN NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_EN NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_EN "Full frame-buffer emulation"
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_EN "Enable full frame-buffer emulation in VRAM. This is useful for games that directly read or write the frame-buffer in VRAM. When enabled, Internal Resolution is forced to 640x480 and performance may be severely impacted."
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_EN NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_EN NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_EN NULL
#define CORE_OPTION_NAME_FOG_LABEL_EN NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_EN NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_EN NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_EN NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_EN NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_EN NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_EN NULL
#define OPTION_VAL_1_EN "Force Nearest-Neighbour"
#define OPTION_VAL_2_EN NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_EN NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_EN NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_EN NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_EN "Notify front-end when internal frame rate changes (e.g. from 60 fps to 30 fps). Improves frame pacing in games that run at a locked 30 fps or 20 fps, but should be disabled for games with unlocked (unstable) frame rates (e.g. Ecco the Dolphin, Unreal Tournament). Note: Unavailable when 'Auto Skip Frame' is enabled."
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_EN NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_EN "Post-process the rendered image to simulate effects specific to the PowerVR2 GPU and analogue video signals."
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_EN NULL
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_EN NULL
#define OPTION_VAL_2_O27_EN NULL
#define OPTION_VAL_4_EN NULL
#define OPTION_VAL_6_EN NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_EN "Texture Upscaling Maximum Filtered Size"
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_EN NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_EN NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_EN NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_EN NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_EN NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_EN NULL
#define OPTION_VAL_SOME_EN NULL
#define OPTION_VAL_MORE_EN NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_EN NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_EN NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_EN NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_EN NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_EN NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_EN NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_EN NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_EN NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_EN NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_EN NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_EN "Analogue Stick Dead-zone"
#define OPTION_VAL_0_EN NULL
#define OPTION_VAL_5_EN NULL
#define OPTION_VAL_10_EN NULL
#define OPTION_VAL_15_EN NULL
#define OPTION_VAL_20_EN NULL
#define OPTION_VAL_25_EN NULL
#define OPTION_VAL_30_EN NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_EN "Trigger Dead-zone"
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_EN NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_EN NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_EN NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_EN NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_EN NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_EN NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_EN "Enable configuration of light gun cross-hair display options. NOTE: Quick Menu may need to be toggled for this setting to take effect."
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_EN NULL
#define OPTION_VAL_WHITE_EN NULL
#define OPTION_VAL_RED_EN NULL
#define OPTION_VAL_GREEN_EN NULL
#define OPTION_VAL_BLUE_EN NULL
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_EN NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_EN NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_EN NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_EN NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_EN NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_EN NULL
#define OPTION_VAL_VMU_A1_EN NULL
#define OPTION_VAL_ALL_VMUS_EN NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_EN NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_EN NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_EN "Enable configuration of emulated VMU LCD screen visibility, size, position and colour. NOTE: Quick Menu may need to be toggled for this setting to take effect."
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_EN NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_EN NULL
#define OPTION_VAL_UPPER_LEFT_EN NULL
#define OPTION_VAL_UPPER_RIGHT_EN NULL
#define OPTION_VAL_LOWER_LEFT_EN NULL
#define OPTION_VAL_LOWER_RIGHT_EN NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_EN NULL
#define OPTION_VAL_1X_EN NULL
#define OPTION_VAL_3X_EN NULL
#define OPTION_VAL_5X_EN NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_EN NULL
#define OPTION_VAL_DEFAULT_ON_00_EN NULL
#define OPTION_VAL_DEFAULT_OFF_01_EN NULL
#define OPTION_VAL_BLACK_02_EN NULL
#define OPTION_VAL_LIGHT_BLUE_04_EN NULL
#define OPTION_VAL_CYAN_06_EN NULL
#define OPTION_VAL_CYAN_BLUE_07_EN NULL
#define OPTION_VAL_LIGHT_GREEN_08_EN NULL
#define OPTION_VAL_CYAN_GREEN_09_EN NULL
#define OPTION_VAL_LIGHT_CYAN_10_EN NULL
#define OPTION_VAL_PURPLE_12_EN NULL
#define OPTION_VAL_LIGHT_PURPLE_13_EN NULL
#define OPTION_VAL_YELLOW_14_EN NULL
#define OPTION_VAL_GRAY_15_EN "Grey"
#define OPTION_VAL_LIGHT_PURPLE_2_16_EN NULL
#define OPTION_VAL_LIGHT_GREEN_2_17_EN NULL
#define OPTION_VAL_LIGHT_GREEN_3_18_EN NULL
#define OPTION_VAL_LIGHT_CYAN_2_19_EN NULL
#define OPTION_VAL_LIGHT_RED_2_20_EN NULL
#define OPTION_VAL_MAGENTA_21_EN NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_EN NULL
#define OPTION_VAL_LIGHT_ORANGE_23_EN NULL
#define OPTION_VAL_ORANGE_24_EN NULL
#define OPTION_VAL_LIGHT_PURPLE_4_25_EN NULL
#define OPTION_VAL_LIGHT_YELLOW_26_EN NULL
#define OPTION_VAL_LIGHT_YELLOW_2_27_EN NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_EN NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_EN NULL
#define OPTION_VAL_40_EN NULL
#define OPTION_VAL_50_EN NULL
#define OPTION_VAL_60_EN NULL
#define OPTION_VAL_70_EN NULL
#define OPTION_VAL_80_EN NULL
#define OPTION_VAL_90_EN NULL
#define OPTION_VAL_100_EN NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_EN NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_EN NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_EN NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_EN NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_EN NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_EN NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_EN NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_EN NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_EN NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_EN NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_EN NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_EN NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_EN NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_EN NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_EN NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_EN NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_EN NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_EN NULL

struct retro_core_option_v2_category option_cats_en[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_EN,
      CATEGORY_SYSTEM_INFO_0_EN
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_EN,
      CATEGORY_VIDEO_INFO_0_EN
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_EN,
      CATEGORY_PERFORMANCE_INFO_0_EN
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_EN,
      CATEGORY_HACKS_INFO_0_EN
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_EN,
      CATEGORY_INPUT_INFO_0_EN
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_EN,
      CATEGORY_VMU_INFO_0_EN
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_en[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_EN,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_EN },
         { "USA",     OPTION_VAL_USA_EN },
         { "Europe",  OPTION_VAL_EUROPE_EN },
         { "Default", OPTION_VAL_DEFAULT_EN },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_EN,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_EN },
         { "English",  OPTION_VAL_ENGLISH_EN },
         { "German",   OPTION_VAL_GERMAN_EN },
         { "French",   OPTION_VAL_FRENCH_EN },
         { "Spanish",  OPTION_VAL_SPANISH_EN },
         { "Italian",  OPTION_VAL_ITALIAN_EN },
         { "Default",  OPTION_VAL_DEFAULT_EN },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_EN,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_boot_to_bios",
      CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_EN,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_EN,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_force_wince",
      CORE_OPTION_NAME_FORCE_WINCE_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_FORCE_WINCE_INFO_0_EN,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_EN,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_EN,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_EN,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_EN,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_EN,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_EN },
         { "640x480",    OPTION_VAL_640X480_EN },
         { "800x600",    OPTION_VAL_800X600_EN },
         { "960x720",    OPTION_VAL_960X720_EN },
         { "1024x768",   OPTION_VAL_1024X768_EN },
         { "1280x960",   OPTION_VAL_1280X960_EN },
         { "1440x1080",  OPTION_VAL_1440X1080_EN },
         { "1600x1200",  OPTION_VAL_1600X1200_EN },
         { "1920x1440",  OPTION_VAL_1920X1440_EN },
         { "2560x1920",  OPTION_VAL_2560X1920_EN },
         { "2880x2160",  OPTION_VAL_2880X2160_EN },
         { "3200x2400",  OPTION_VAL_3200X2400_EN },
         { "3840x2880",  OPTION_VAL_3840X2880_EN },
         { "4480x3360",  OPTION_VAL_4480X3360_EN },
         { "5120x3840",  OPTION_VAL_5120X3840_EN },
         { "5760x4320",  OPTION_VAL_5760X4320_EN },
         { "6400x4800",  OPTION_VAL_6400X4800_EN },
         { "7040x5280",  OPTION_VAL_7040X5280_EN },
         { "7680x5760",  OPTION_VAL_7680X5760_EN },
         { "8320x6240",  OPTION_VAL_8320X6240_EN },
         { "8960x6720",  OPTION_VAL_8960X6720_EN },
         { "9600x7200",  OPTION_VAL_9600X7200_EN },
         { "10240x7680", OPTION_VAL_10240X7680_EN },
         { "10880x8160", OPTION_VAL_10880X8160_EN },
         { "11520x8640", OPTION_VAL_11520X8640_EN },
         { "12160x9120", OPTION_VAL_12160X9120_EN },
         { "12800x9600", OPTION_VAL_12800X9600_EN },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_EN,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_EN },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_EN },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_EN },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_EN,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_EN },
         { "PAL",     OPTION_VAL_PAL_EN },
         { "PAL_N",   OPTION_VAL_PAL_N_EN },
         { "PAL_M",   OPTION_VAL_PAL_M_EN },
         { "Default", OPTION_VAL_DEFAULT_EN },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_EN,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_EN },
         { "vertical",   OPTION_VAL_VERTICAL_EN },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_EN,
      NULL,
      "",
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_EN },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_EN },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_EN },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_EN,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_EN },
         { "1GB",   OPTION_VAL_1GB_EN },
         { "2GB",   OPTION_VAL_2GB_EN },
         { "4GB",   OPTION_VAL_4GB_EN },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_EN,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_EN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_EN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_EN,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_EN,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_EN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_EN,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_EN,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_EN },
         { "1",  OPTION_VAL_1_EN },
         { "2",  OPTION_VAL_2_EN },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_EN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_EN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_EN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_EN,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O27_EN },
         { "4",  OPTION_VAL_4_EN },
         { "6",  OPTION_VAL_6_EN },
         { NULL, NULL },
      },
      "1",
   },
   {/* TODO: needs clarification */
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_EN,
      NULL,
      "",
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_EN,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_EN,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_EN,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_EN,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_EN },
         { "more", OPTION_VAL_MORE_EN },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_EN,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_EN,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_EN,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_EN,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_EN,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: probably needs explanation */
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_EN,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_EN,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_EN },
         { "5%",  OPTION_VAL_5_EN },
         { "10%", OPTION_VAL_10_EN },
         { "15%", OPTION_VAL_15_EN },
         { "20%", OPTION_VAL_20_EN },
         { "25%", OPTION_VAL_25_EN },
         { "30%", OPTION_VAL_30_EN },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_EN,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_EN },
         { "5%",  OPTION_VAL_5_EN },
         { "10%", OPTION_VAL_10_EN },
         { "15%", OPTION_VAL_15_EN },
         { "20%", OPTION_VAL_20_EN },
         { "25%", OPTION_VAL_25_EN },
         { "30%", OPTION_VAL_30_EN },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_EN,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_purupuru",
      CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_EN,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_EN,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_EN,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_EN,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_EN },
         { "Red",      OPTION_VAL_RED_EN },
         { "Green",    OPTION_VAL_GREEN_EN },
         { "Blue",     OPTION_VAL_BLUE_EN },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_EN,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_EN },
         { "Red",      OPTION_VAL_RED_EN },
         { "Green",    OPTION_VAL_GREEN_EN },
         { "Blue",     OPTION_VAL_BLUE_EN },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_EN,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_EN },
         { "Red",      OPTION_VAL_RED_EN },
         { "Green",    OPTION_VAL_GREEN_EN },
         { "Blue",     OPTION_VAL_BLUE_EN },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_EN,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_EN },
         { "Red",      OPTION_VAL_RED_EN },
         { "Green",    OPTION_VAL_GREEN_EN },
         { "Blue",     OPTION_VAL_BLUE_EN },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_EN,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_EN,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_EN,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_EN },
         { "All VMUs", OPTION_VAL_ALL_VMUS_EN },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_EN,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_EN,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_EN,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_EN },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_EN },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_EN },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_EN },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_EN },
         { "2x", OPTION_VAL_2_O27_EN },
         { "3x", OPTION_VAL_3X_EN },
         { "4x", OPTION_VAL_4_EN },
         { "5x", OPTION_VAL_5X_EN },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EN },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EN },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EN },
         { "BLUE 03",           OPTION_VAL_BLUE_EN },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EN },
         { "GREEN 05",          OPTION_VAL_GREEN_EN },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EN },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EN },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EN },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EN },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EN },
         { "RED 11",            OPTION_VAL_RED_EN },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EN },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EN },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EN },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EN },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EN },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EN },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EN },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EN },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EN },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EN },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EN },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EN },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EN },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EN },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EN },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EN },
         { "WHITE 28",          OPTION_VAL_WHITE_EN },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EN },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EN },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EN },
         { "BLUE 03",           OPTION_VAL_BLUE_EN },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EN },
         { "GREEN 05",          OPTION_VAL_GREEN_EN },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EN },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EN },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EN },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EN },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EN },
         { "RED 11",            OPTION_VAL_RED_EN },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EN },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EN },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EN },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EN },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EN },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EN },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EN },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EN },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EN },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EN },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EN },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EN },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EN },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EN },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EN },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EN },
         { "WHITE 28",          OPTION_VAL_WHITE_EN },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_EN },
         { "20%",  OPTION_VAL_20_EN },
         { "30%",  OPTION_VAL_30_EN },
         { "40%",  OPTION_VAL_40_EN },
         { "50%",  OPTION_VAL_50_EN },
         { "60%",  OPTION_VAL_60_EN },
         { "70%",  OPTION_VAL_70_EN },
         { "80%",  OPTION_VAL_80_EN },
         { "90%",  OPTION_VAL_90_EN },
         { "100%", OPTION_VAL_100_EN },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_EN },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_EN },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_EN },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_EN },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_EN },
         { "2x", OPTION_VAL_2_O27_EN },
         { "3x", OPTION_VAL_3X_EN },
         { "4x", OPTION_VAL_4_EN },
         { "5x", OPTION_VAL_5X_EN },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EN },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EN },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EN },
         { "BLUE 03",           OPTION_VAL_BLUE_EN },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EN },
         { "GREEN 05",          OPTION_VAL_GREEN_EN },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EN },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EN },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EN },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EN },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EN },
         { "RED 11",            OPTION_VAL_RED_EN },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EN },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EN },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EN },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EN },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EN },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EN },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EN },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EN },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EN },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EN },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EN },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EN },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EN },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EN },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EN },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EN },
         { "WHITE 28",          OPTION_VAL_WHITE_EN },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EN },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EN },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EN },
         { "BLUE 03",           OPTION_VAL_BLUE_EN },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EN },
         { "GREEN 05",          OPTION_VAL_GREEN_EN },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EN },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EN },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EN },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EN },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EN },
         { "RED 11",            OPTION_VAL_RED_EN },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EN },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EN },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EN },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EN },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EN },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EN },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EN },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EN },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EN },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EN },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EN },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EN },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EN },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EN },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EN },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EN },
         { "WHITE 28",          OPTION_VAL_WHITE_EN },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_EN },
         { "20%",  OPTION_VAL_20_EN },
         { "30%",  OPTION_VAL_30_EN },
         { "40%",  OPTION_VAL_40_EN },
         { "50%",  OPTION_VAL_50_EN },
         { "60%",  OPTION_VAL_60_EN },
         { "70%",  OPTION_VAL_70_EN },
         { "80%",  OPTION_VAL_80_EN },
         { "90%",  OPTION_VAL_90_EN },
         { "100%", OPTION_VAL_100_EN },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_EN },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_EN },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_EN },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_EN },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_EN },
         { "2x", OPTION_VAL_2_O27_EN },
         { "3x", OPTION_VAL_3X_EN },
         { "4x", OPTION_VAL_4_EN },
         { "5x", OPTION_VAL_5X_EN },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EN },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EN },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EN },
         { "BLUE 03",           OPTION_VAL_BLUE_EN },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EN },
         { "GREEN 05",          OPTION_VAL_GREEN_EN },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EN },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EN },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EN },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EN },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EN },
         { "RED 11",            OPTION_VAL_RED_EN },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EN },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EN },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EN },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EN },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EN },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EN },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EN },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EN },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EN },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EN },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EN },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EN },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EN },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EN },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EN },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EN },
         { "WHITE 28",          OPTION_VAL_WHITE_EN },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EN },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EN },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EN },
         { "BLUE 03",           OPTION_VAL_BLUE_EN },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EN },
         { "GREEN 05",          OPTION_VAL_GREEN_EN },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EN },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EN },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EN },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EN },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EN },
         { "RED 11",            OPTION_VAL_RED_EN },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EN },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EN },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EN },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EN },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EN },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EN },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EN },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EN },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EN },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EN },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EN },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EN },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EN },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EN },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EN },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EN },
         { "WHITE 28",          OPTION_VAL_WHITE_EN },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_EN },
         { "20%",  OPTION_VAL_20_EN },
         { "30%",  OPTION_VAL_30_EN },
         { "40%",  OPTION_VAL_40_EN },
         { "50%",  OPTION_VAL_50_EN },
         { "60%",  OPTION_VAL_60_EN },
         { "70%",  OPTION_VAL_70_EN },
         { "80%",  OPTION_VAL_80_EN },
         { "90%",  OPTION_VAL_90_EN },
         { "100%", OPTION_VAL_100_EN },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_EN },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_EN },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_EN },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_EN },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_EN },
         { "2x", OPTION_VAL_2_O27_EN },
         { "3x", OPTION_VAL_3X_EN },
         { "4x", OPTION_VAL_4_EN },
         { "5x", OPTION_VAL_5X_EN },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EN },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EN },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EN },
         { "BLUE 03",           OPTION_VAL_BLUE_EN },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EN },
         { "GREEN 05",          OPTION_VAL_GREEN_EN },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EN },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EN },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EN },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EN },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EN },
         { "RED 11",            OPTION_VAL_RED_EN },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EN },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EN },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EN },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EN },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EN },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EN },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EN },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EN },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EN },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EN },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EN },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EN },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EN },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EN },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EN },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EN },
         { "WHITE 28",          OPTION_VAL_WHITE_EN },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EN },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EN },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EN },
         { "BLUE 03",           OPTION_VAL_BLUE_EN },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EN },
         { "GREEN 05",          OPTION_VAL_GREEN_EN },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EN },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EN },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EN },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EN },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EN },
         { "RED 11",            OPTION_VAL_RED_EN },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EN },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EN },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EN },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EN },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EN },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EN },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EN },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EN },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EN },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EN },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EN },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EN },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EN },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EN },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EN },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EN },
         { "WHITE 28",          OPTION_VAL_WHITE_EN },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_EN },
         { "20%",  OPTION_VAL_20_EN },
         { "30%",  OPTION_VAL_30_EN },
         { "40%",  OPTION_VAL_40_EN },
         { "50%",  OPTION_VAL_50_EN },
         { "60%",  OPTION_VAL_60_EN },
         { "70%",  OPTION_VAL_70_EN },
         { "80%",  OPTION_VAL_80_EN },
         { "90%",  OPTION_VAL_90_EN },
         { "100%", OPTION_VAL_100_EN },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_en = {
   option_cats_en,
   option_defs_en
};

/* RETRO_LANGUAGE_EO */

#define CATEGORY_SYSTEM_LABEL_EO NULL
#define CATEGORY_SYSTEM_INFO_0_EO NULL
#define CATEGORY_VIDEO_LABEL_EO "Video Driver"
#define CATEGORY_VIDEO_INFO_0_EO NULL
#define CATEGORY_PERFORMANCE_LABEL_EO NULL
#define CATEGORY_PERFORMANCE_INFO_0_EO NULL
#define CATEGORY_HACKS_LABEL_EO NULL
#define CATEGORY_HACKS_INFO_0_EO NULL
#define CATEGORY_INPUT_LABEL_EO "Input Driver"
#define CATEGORY_INPUT_INFO_0_EO NULL
#define CATEGORY_VMU_LABEL_EO NULL
#define CATEGORY_VMU_INFO_0_EO NULL
#define CORE_OPTION_NAME_REGION_LABEL_EO NULL
#define OPTION_VAL_JAPAN_EO NULL
#define OPTION_VAL_USA_EO NULL
#define OPTION_VAL_EUROPE_EO NULL
#define OPTION_VAL_DEFAULT_EO NULL
#define CORE_OPTION_NAME_LANGUAGE_LABEL_EO NULL
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_EO NULL
#define OPTION_VAL_JAPANESE_EO NULL
#define OPTION_VAL_ENGLISH_EO NULL
#define OPTION_VAL_GERMAN_EO NULL
#define OPTION_VAL_FRENCH_EO NULL
#define OPTION_VAL_SPANISH_EO NULL
#define OPTION_VAL_ITALIAN_EO NULL
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_EO NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_EO NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_EO NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_EO NULL
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_EO NULL
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_EO NULL
#define CORE_OPTION_NAME_FORCE_WINCE_LABEL_EO NULL
#define CORE_OPTION_NAME_FORCE_WINCE_INFO_0_EO NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_EO NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_EO NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_EO NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_EO NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_EO NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_EO NULL
#define CORE_OPTION_NAME_UPNP_LABEL_EO NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_EO NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_EO "Interna Rezolucio"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_EO NULL
#define OPTION_VAL_320X240_EO NULL
#define OPTION_VAL_640X480_EO NULL
#define OPTION_VAL_800X600_EO NULL
#define OPTION_VAL_960X720_EO NULL
#define OPTION_VAL_1024X768_EO NULL
#define OPTION_VAL_1280X960_EO NULL
#define OPTION_VAL_1440X1080_EO NULL
#define OPTION_VAL_1600X1200_EO NULL
#define OPTION_VAL_1920X1440_EO NULL
#define OPTION_VAL_2560X1920_EO NULL
#define OPTION_VAL_2880X2160_EO NULL
#define OPTION_VAL_3200X2400_EO NULL
#define OPTION_VAL_3840X2880_EO NULL
#define OPTION_VAL_4480X3360_EO NULL
#define OPTION_VAL_5120X3840_EO NULL
#define OPTION_VAL_5760X4320_EO NULL
#define OPTION_VAL_6400X4800_EO NULL
#define OPTION_VAL_7040X5280_EO NULL
#define OPTION_VAL_7680X5760_EO NULL
#define OPTION_VAL_8320X6240_EO NULL
#define OPTION_VAL_8960X6720_EO NULL
#define OPTION_VAL_9600X7200_EO NULL
#define OPTION_VAL_10240X7680_EO NULL
#define OPTION_VAL_10880X8160_EO NULL
#define OPTION_VAL_11520X8640_EO NULL
#define OPTION_VAL_12160X9120_EO NULL
#define OPTION_VAL_12800X9600_EO NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_EO NULL
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_EO NULL
#define OPTION_VAL_VGA_EO NULL
#define OPTION_VAL_TV_RGB_EO NULL
#define OPTION_VAL_TV_COMPOSITE_EO NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_EO NULL
#define OPTION_VAL_NTSC_EO NULL
#define OPTION_VAL_PAL_EO NULL
#define OPTION_VAL_PAL_N_EO NULL
#define OPTION_VAL_PAL_M_EO NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_EO NULL
#define OPTION_VAL_HORIZONTAL_EO NULL
#define OPTION_VAL_VERTICAL_EO NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_EO NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_EO NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_EO NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_EO NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_EO NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_EO NULL
#define OPTION_VAL_512MB_EO NULL
#define OPTION_VAL_1GB_EO NULL
#define OPTION_VAL_2GB_EO NULL
#define OPTION_VAL_4GB_EO NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_EO NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_EO NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_EO NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_EO NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_EO NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_EO NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_EO NULL
#define CORE_OPTION_NAME_FOG_LABEL_EO NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_EO NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_EO NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_EO NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_EO NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_EO NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_EO NULL
#define OPTION_VAL_1_EO NULL
#define OPTION_VAL_2_EO NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_EO NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_EO NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_EO NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_EO NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_EO NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_EO NULL
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_EO NULL
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_EO NULL
#define OPTION_VAL_2_O27_EO NULL
#define OPTION_VAL_4_EO NULL
#define OPTION_VAL_6_EO NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_EO NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_EO NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_EO NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_EO NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_EO NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_EO NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_EO NULL
#define OPTION_VAL_SOME_EO NULL
#define OPTION_VAL_MORE_EO NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_EO NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_EO NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_EO NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_EO NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_EO NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_EO NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_EO NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_EO NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_EO NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_EO NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_EO NULL
#define OPTION_VAL_0_EO NULL
#define OPTION_VAL_5_EO NULL
#define OPTION_VAL_10_EO NULL
#define OPTION_VAL_15_EO NULL
#define OPTION_VAL_20_EO NULL
#define OPTION_VAL_25_EO NULL
#define OPTION_VAL_30_EO NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_EO NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_EO NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_EO NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_EO NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_EO NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_EO NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_EO NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_EO NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_EO NULL
#define OPTION_VAL_WHITE_EO NULL
#define OPTION_VAL_RED_EO NULL
#define OPTION_VAL_GREEN_EO NULL
#define OPTION_VAL_BLUE_EO NULL
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_EO NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_EO NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_EO NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_EO NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_EO NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_EO NULL
#define OPTION_VAL_VMU_A1_EO NULL
#define OPTION_VAL_ALL_VMUS_EO NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_EO NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_EO NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_EO NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_EO NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_EO NULL
#define OPTION_VAL_UPPER_LEFT_EO NULL
#define OPTION_VAL_UPPER_RIGHT_EO NULL
#define OPTION_VAL_LOWER_LEFT_EO NULL
#define OPTION_VAL_LOWER_RIGHT_EO NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_EO NULL
#define OPTION_VAL_1X_EO NULL
#define OPTION_VAL_3X_EO NULL
#define OPTION_VAL_5X_EO NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_EO NULL
#define OPTION_VAL_DEFAULT_ON_00_EO NULL
#define OPTION_VAL_DEFAULT_OFF_01_EO NULL
#define OPTION_VAL_BLACK_02_EO NULL
#define OPTION_VAL_LIGHT_BLUE_04_EO NULL
#define OPTION_VAL_CYAN_06_EO NULL
#define OPTION_VAL_CYAN_BLUE_07_EO NULL
#define OPTION_VAL_LIGHT_GREEN_08_EO NULL
#define OPTION_VAL_CYAN_GREEN_09_EO NULL
#define OPTION_VAL_LIGHT_CYAN_10_EO NULL
#define OPTION_VAL_PURPLE_12_EO NULL
#define OPTION_VAL_LIGHT_PURPLE_13_EO NULL
#define OPTION_VAL_YELLOW_14_EO NULL
#define OPTION_VAL_GRAY_15_EO NULL
#define OPTION_VAL_LIGHT_PURPLE_2_16_EO NULL
#define OPTION_VAL_LIGHT_GREEN_2_17_EO NULL
#define OPTION_VAL_LIGHT_GREEN_3_18_EO NULL
#define OPTION_VAL_LIGHT_CYAN_2_19_EO NULL
#define OPTION_VAL_LIGHT_RED_2_20_EO NULL
#define OPTION_VAL_MAGENTA_21_EO NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_EO NULL
#define OPTION_VAL_LIGHT_ORANGE_23_EO NULL
#define OPTION_VAL_ORANGE_24_EO NULL
#define OPTION_VAL_LIGHT_PURPLE_4_25_EO NULL
#define OPTION_VAL_LIGHT_YELLOW_26_EO NULL
#define OPTION_VAL_LIGHT_YELLOW_2_27_EO NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_EO NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_EO NULL
#define OPTION_VAL_40_EO NULL
#define OPTION_VAL_50_EO NULL
#define OPTION_VAL_60_EO NULL
#define OPTION_VAL_70_EO NULL
#define OPTION_VAL_80_EO NULL
#define OPTION_VAL_90_EO NULL
#define OPTION_VAL_100_EO NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_EO NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_EO NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_EO NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_EO NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_EO NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_EO NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_EO NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_EO NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_EO NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_EO NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_EO NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_EO NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_EO NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_EO NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_EO NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_EO NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_EO NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_EO NULL

struct retro_core_option_v2_category option_cats_eo[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_EO,
      CATEGORY_SYSTEM_INFO_0_EO
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_EO,
      CATEGORY_VIDEO_INFO_0_EO
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_EO,
      CATEGORY_PERFORMANCE_INFO_0_EO
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_EO,
      CATEGORY_HACKS_INFO_0_EO
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_EO,
      CATEGORY_INPUT_INFO_0_EO
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_EO,
      CATEGORY_VMU_INFO_0_EO
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_eo[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_EO,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_EO },
         { "USA",     OPTION_VAL_USA_EO },
         { "Europe",  OPTION_VAL_EUROPE_EO },
         { "Default", OPTION_VAL_DEFAULT_EO },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_EO,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_EO },
         { "English",  OPTION_VAL_ENGLISH_EO },
         { "German",   OPTION_VAL_GERMAN_EO },
         { "French",   OPTION_VAL_FRENCH_EO },
         { "Spanish",  OPTION_VAL_SPANISH_EO },
         { "Italian",  OPTION_VAL_ITALIAN_EO },
         { "Default",  OPTION_VAL_DEFAULT_EO },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_EO,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_boot_to_bios",
      CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_EO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_EO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_force_wince",
      CORE_OPTION_NAME_FORCE_WINCE_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_FORCE_WINCE_INFO_0_EO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_EO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_EO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_EO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_EO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_EO,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_EO },
         { "640x480",    OPTION_VAL_640X480_EO },
         { "800x600",    OPTION_VAL_800X600_EO },
         { "960x720",    OPTION_VAL_960X720_EO },
         { "1024x768",   OPTION_VAL_1024X768_EO },
         { "1280x960",   OPTION_VAL_1280X960_EO },
         { "1440x1080",  OPTION_VAL_1440X1080_EO },
         { "1600x1200",  OPTION_VAL_1600X1200_EO },
         { "1920x1440",  OPTION_VAL_1920X1440_EO },
         { "2560x1920",  OPTION_VAL_2560X1920_EO },
         { "2880x2160",  OPTION_VAL_2880X2160_EO },
         { "3200x2400",  OPTION_VAL_3200X2400_EO },
         { "3840x2880",  OPTION_VAL_3840X2880_EO },
         { "4480x3360",  OPTION_VAL_4480X3360_EO },
         { "5120x3840",  OPTION_VAL_5120X3840_EO },
         { "5760x4320",  OPTION_VAL_5760X4320_EO },
         { "6400x4800",  OPTION_VAL_6400X4800_EO },
         { "7040x5280",  OPTION_VAL_7040X5280_EO },
         { "7680x5760",  OPTION_VAL_7680X5760_EO },
         { "8320x6240",  OPTION_VAL_8320X6240_EO },
         { "8960x6720",  OPTION_VAL_8960X6720_EO },
         { "9600x7200",  OPTION_VAL_9600X7200_EO },
         { "10240x7680", OPTION_VAL_10240X7680_EO },
         { "10880x8160", OPTION_VAL_10880X8160_EO },
         { "11520x8640", OPTION_VAL_11520X8640_EO },
         { "12160x9120", OPTION_VAL_12160X9120_EO },
         { "12800x9600", OPTION_VAL_12800X9600_EO },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_EO,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_EO },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_EO },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_EO },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_EO,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_EO },
         { "PAL",     OPTION_VAL_PAL_EO },
         { "PAL_N",   OPTION_VAL_PAL_N_EO },
         { "PAL_M",   OPTION_VAL_PAL_M_EO },
         { "Default", OPTION_VAL_DEFAULT_EO },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_EO,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_EO },
         { "vertical",   OPTION_VAL_VERTICAL_EO },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_EO,
      NULL,
      "",
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_EO },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_EO },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_EO },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_EO,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_EO },
         { "1GB",   OPTION_VAL_1GB_EO },
         { "2GB",   OPTION_VAL_2GB_EO },
         { "4GB",   OPTION_VAL_4GB_EO },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_EO,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_EO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_EO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_EO,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_EO,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_EO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_EO,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_EO,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_EO },
         { "1",  OPTION_VAL_1_EO },
         { "2",  OPTION_VAL_2_EO },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_EO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_EO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_EO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_EO,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O27_EO },
         { "4",  OPTION_VAL_4_EO },
         { "6",  OPTION_VAL_6_EO },
         { NULL, NULL },
      },
      "1",
   },
   {/* TODO: needs clarification */
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_EO,
      NULL,
      "",
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_EO,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_EO,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_EO,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_EO,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_EO },
         { "more", OPTION_VAL_MORE_EO },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_EO,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_EO,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_EO,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_EO,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_EO,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: probably needs explanation */
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_EO,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_EO,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_EO },
         { "5%",  OPTION_VAL_5_EO },
         { "10%", OPTION_VAL_10_EO },
         { "15%", OPTION_VAL_15_EO },
         { "20%", OPTION_VAL_20_EO },
         { "25%", OPTION_VAL_25_EO },
         { "30%", OPTION_VAL_30_EO },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_EO,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_EO },
         { "5%",  OPTION_VAL_5_EO },
         { "10%", OPTION_VAL_10_EO },
         { "15%", OPTION_VAL_15_EO },
         { "20%", OPTION_VAL_20_EO },
         { "25%", OPTION_VAL_25_EO },
         { "30%", OPTION_VAL_30_EO },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_EO,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_purupuru",
      CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_EO,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_EO,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_EO,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_EO,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_EO },
         { "Red",      OPTION_VAL_RED_EO },
         { "Green",    OPTION_VAL_GREEN_EO },
         { "Blue",     OPTION_VAL_BLUE_EO },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_EO,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_EO },
         { "Red",      OPTION_VAL_RED_EO },
         { "Green",    OPTION_VAL_GREEN_EO },
         { "Blue",     OPTION_VAL_BLUE_EO },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_EO,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_EO },
         { "Red",      OPTION_VAL_RED_EO },
         { "Green",    OPTION_VAL_GREEN_EO },
         { "Blue",     OPTION_VAL_BLUE_EO },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_EO,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_EO },
         { "Red",      OPTION_VAL_RED_EO },
         { "Green",    OPTION_VAL_GREEN_EO },
         { "Blue",     OPTION_VAL_BLUE_EO },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_EO,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_EO,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_EO,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_EO },
         { "All VMUs", OPTION_VAL_ALL_VMUS_EO },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_EO,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_EO,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_EO,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_EO },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_EO },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_EO },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_EO },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_EO },
         { "2x", OPTION_VAL_2_O27_EO },
         { "3x", OPTION_VAL_3X_EO },
         { "4x", OPTION_VAL_4_EO },
         { "5x", OPTION_VAL_5X_EO },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EO },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EO },
         { "BLUE 03",           OPTION_VAL_BLUE_EO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EO },
         { "GREEN 05",          OPTION_VAL_GREEN_EO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EO },
         { "RED 11",            OPTION_VAL_RED_EO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EO },
         { "WHITE 28",          OPTION_VAL_WHITE_EO },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EO },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EO },
         { "BLUE 03",           OPTION_VAL_BLUE_EO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EO },
         { "GREEN 05",          OPTION_VAL_GREEN_EO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EO },
         { "RED 11",            OPTION_VAL_RED_EO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EO },
         { "WHITE 28",          OPTION_VAL_WHITE_EO },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_EO },
         { "20%",  OPTION_VAL_20_EO },
         { "30%",  OPTION_VAL_30_EO },
         { "40%",  OPTION_VAL_40_EO },
         { "50%",  OPTION_VAL_50_EO },
         { "60%",  OPTION_VAL_60_EO },
         { "70%",  OPTION_VAL_70_EO },
         { "80%",  OPTION_VAL_80_EO },
         { "90%",  OPTION_VAL_90_EO },
         { "100%", OPTION_VAL_100_EO },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_EO },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_EO },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_EO },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_EO },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_EO },
         { "2x", OPTION_VAL_2_O27_EO },
         { "3x", OPTION_VAL_3X_EO },
         { "4x", OPTION_VAL_4_EO },
         { "5x", OPTION_VAL_5X_EO },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EO },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EO },
         { "BLUE 03",           OPTION_VAL_BLUE_EO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EO },
         { "GREEN 05",          OPTION_VAL_GREEN_EO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EO },
         { "RED 11",            OPTION_VAL_RED_EO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EO },
         { "WHITE 28",          OPTION_VAL_WHITE_EO },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EO },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EO },
         { "BLUE 03",           OPTION_VAL_BLUE_EO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EO },
         { "GREEN 05",          OPTION_VAL_GREEN_EO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EO },
         { "RED 11",            OPTION_VAL_RED_EO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EO },
         { "WHITE 28",          OPTION_VAL_WHITE_EO },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_EO },
         { "20%",  OPTION_VAL_20_EO },
         { "30%",  OPTION_VAL_30_EO },
         { "40%",  OPTION_VAL_40_EO },
         { "50%",  OPTION_VAL_50_EO },
         { "60%",  OPTION_VAL_60_EO },
         { "70%",  OPTION_VAL_70_EO },
         { "80%",  OPTION_VAL_80_EO },
         { "90%",  OPTION_VAL_90_EO },
         { "100%", OPTION_VAL_100_EO },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_EO },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_EO },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_EO },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_EO },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_EO },
         { "2x", OPTION_VAL_2_O27_EO },
         { "3x", OPTION_VAL_3X_EO },
         { "4x", OPTION_VAL_4_EO },
         { "5x", OPTION_VAL_5X_EO },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EO },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EO },
         { "BLUE 03",           OPTION_VAL_BLUE_EO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EO },
         { "GREEN 05",          OPTION_VAL_GREEN_EO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EO },
         { "RED 11",            OPTION_VAL_RED_EO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EO },
         { "WHITE 28",          OPTION_VAL_WHITE_EO },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EO },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EO },
         { "BLUE 03",           OPTION_VAL_BLUE_EO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EO },
         { "GREEN 05",          OPTION_VAL_GREEN_EO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EO },
         { "RED 11",            OPTION_VAL_RED_EO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EO },
         { "WHITE 28",          OPTION_VAL_WHITE_EO },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_EO },
         { "20%",  OPTION_VAL_20_EO },
         { "30%",  OPTION_VAL_30_EO },
         { "40%",  OPTION_VAL_40_EO },
         { "50%",  OPTION_VAL_50_EO },
         { "60%",  OPTION_VAL_60_EO },
         { "70%",  OPTION_VAL_70_EO },
         { "80%",  OPTION_VAL_80_EO },
         { "90%",  OPTION_VAL_90_EO },
         { "100%", OPTION_VAL_100_EO },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_EO },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_EO },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_EO },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_EO },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_EO },
         { "2x", OPTION_VAL_2_O27_EO },
         { "3x", OPTION_VAL_3X_EO },
         { "4x", OPTION_VAL_4_EO },
         { "5x", OPTION_VAL_5X_EO },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EO },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EO },
         { "BLUE 03",           OPTION_VAL_BLUE_EO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EO },
         { "GREEN 05",          OPTION_VAL_GREEN_EO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EO },
         { "RED 11",            OPTION_VAL_RED_EO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EO },
         { "WHITE 28",          OPTION_VAL_WHITE_EO },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EO },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EO },
         { "BLUE 03",           OPTION_VAL_BLUE_EO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EO },
         { "GREEN 05",          OPTION_VAL_GREEN_EO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EO },
         { "RED 11",            OPTION_VAL_RED_EO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EO },
         { "WHITE 28",          OPTION_VAL_WHITE_EO },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_EO },
         { "20%",  OPTION_VAL_20_EO },
         { "30%",  OPTION_VAL_30_EO },
         { "40%",  OPTION_VAL_40_EO },
         { "50%",  OPTION_VAL_50_EO },
         { "60%",  OPTION_VAL_60_EO },
         { "70%",  OPTION_VAL_70_EO },
         { "80%",  OPTION_VAL_80_EO },
         { "90%",  OPTION_VAL_90_EO },
         { "100%", OPTION_VAL_100_EO },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_eo = {
   option_cats_eo,
   option_defs_eo
};

/* RETRO_LANGUAGE_ES */

#define CATEGORY_SYSTEM_LABEL_ES "Sistema"
#define CATEGORY_SYSTEM_INFO_0_ES "Cambia los ajustes de región, idioma, BIOS y del hardware base."
#define CATEGORY_VIDEO_LABEL_ES "Vídeo"
#define CATEGORY_VIDEO_INFO_0_ES "Cambia los ajustes de resolución, las transparencias independientes del orden y los efectos visuales."
#define CATEGORY_PERFORMANCE_LABEL_ES "Rendimiento"
#define CATEGORY_PERFORMANCE_INFO_0_ES "Configura los ajustes del renderizado multihilo y la omisión de fotogramas."
#define CATEGORY_HACKS_LABEL_ES "Arreglos de emulación"
#define CATEGORY_HACKS_INFO_0_ES "Cambia los ajustes de forzado de imagen panorámica, la velocidad de carga de GD-ROM y la sustitución de texturas."
#define CATEGORY_INPUT_LABEL_ES "Entrada"
#define CATEGORY_INPUT_INFO_0_ES "Cambia los ajustes de los mandos y las pistolas de luz."
#define CATEGORY_VMU_LABEL_ES "Visual Memory"
#define CATEGORY_VMU_INFO_0_ES "Cambia los ajustes de los archivos de VMU individuales para cada juego y la visibilidad en pantalla de las mismas."
#define CORE_OPTION_NAME_REGION_LABEL_ES "Región"
#define OPTION_VAL_JAPAN_ES "Japón"
#define OPTION_VAL_USA_ES "Estados Unidos"
#define OPTION_VAL_EUROPE_ES "Europa"
#define OPTION_VAL_DEFAULT_ES "Valor predeterminado"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_ES "Idioma"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_ES "Cambia el idioma que utilizará la BIOS y cualquier juego que contenga varios idiomas."
#define OPTION_VAL_JAPANESE_ES "Japonés"
#define OPTION_VAL_ENGLISH_ES "Inglés"
#define OPTION_VAL_GERMAN_ES "Alemán"
#define OPTION_VAL_FRENCH_ES "Francés"
#define OPTION_VAL_SPANISH_ES "Español"
#define OPTION_VAL_ITALIAN_ES "Italiano"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_ES "BIOS HLE (es necesario reiniciar)"
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_ES "Fuerza el uso de una BIOS creada con emulación de alto nivel."
#define CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_ES "Arrancar desde la BIOS (es necesario reiniciar)"
#define CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_ES "Arranca directamente el menú de la BIOS de Dreamcast."
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_ES "Activar DSP"
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_ES "Habilita la emulación del DSP (procesador de señales digitales) de audio de Dreamcast. Mejora la precisión del sonido generado a costa de aumentar los requisitos."
#define CORE_OPTION_NAME_FORCE_WINCE_LABEL_ES "Forzar el modo Windows CE"
#define CORE_OPTION_NAME_FORCE_WINCE_INFO_0_ES "Permite emular por completo la MMU (unidad de gestión de memoria) y otros elementos necesarios para los juegos compatibles con Windows CE."
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_ES "Activar botones de servicio de NAOMI"
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_ES "Activa el botón SERVICE de la placa NAOMI para acceder a la configuración del mueble."
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_ES "Partidas gratuitas en NAOMI"
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_ES "Modifica la configuración de monedas del juego a partidas gratuitas."
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_ES "Emulación del Broadband Adapter (BBA)"
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_ES "Emula el adaptador de banda ancha por Ethernet en lugar del módem (es necesario reiniciar)."
#define CORE_OPTION_NAME_UPNP_LABEL_ES "Activar UPnP"
#define CORE_OPTION_NAME_UPNP_INFO_0_ES "Utiliza UPnP para configurar automáticamente tu router/enrutador de Internet para poder conectarte a juegos online."
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_ES "Resolución interna"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_ES "Modifica la resolución de renderizado."
#define OPTION_VAL_320X240_ES "320 × 240"
#define OPTION_VAL_640X480_ES "640 × 480"
#define OPTION_VAL_800X600_ES "800 × 600"
#define OPTION_VAL_960X720_ES "960 × 720"
#define OPTION_VAL_1024X768_ES "1024 × 768"
#define OPTION_VAL_1280X960_ES "1280 × 960"
#define OPTION_VAL_1440X1080_ES "1440 × 1080"
#define OPTION_VAL_1600X1200_ES "1600 × 1200"
#define OPTION_VAL_1920X1440_ES "1920 × 1440"
#define OPTION_VAL_2560X1920_ES "2560 × 1920"
#define OPTION_VAL_2880X2160_ES "2880 × 2160"
#define OPTION_VAL_3200X2400_ES "3200 × 2400"
#define OPTION_VAL_3840X2880_ES "3840 × 2880"
#define OPTION_VAL_4480X3360_ES "4480 × 3360"
#define OPTION_VAL_5120X3840_ES "5120 × 3840"
#define OPTION_VAL_5760X4320_ES "5760 × 4320"
#define OPTION_VAL_6400X4800_ES "6400 × 4800"
#define OPTION_VAL_7040X5280_ES "7040 × 5280"
#define OPTION_VAL_7680X5760_ES "7680 × 5760"
#define OPTION_VAL_8320X6240_ES "8320 × 6240"
#define OPTION_VAL_8960X6720_ES "8960 × 6720"
#define OPTION_VAL_9600X7200_ES "9600 × 7200"
#define OPTION_VAL_10240X7680_ES "10240 × 7680"
#define OPTION_VAL_10880X8160_ES "10880 × 8160"
#define OPTION_VAL_11520X8640_ES "11520 × 8640"
#define OPTION_VAL_12160X9120_ES "12160 × 9120"
#define OPTION_VAL_12800X9600_ES "12800 × 9600"
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_ES "Tipo de cable"
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_ES "El tipo de señal de salida de la imagen. «TV (Compuesto)» es el formato más compatible de todos."
#define OPTION_VAL_VGA_ES NULL
#define OPTION_VAL_TV_RGB_ES NULL
#define OPTION_VAL_TV_COMPOSITE_ES "TV (Compuesto)"
#define CORE_OPTION_NAME_BROADCAST_LABEL_ES "Estándar de transmisión"
#define OPTION_VAL_NTSC_ES NULL
#define OPTION_VAL_PAL_ES "PAL (Resto del mundo)"
#define OPTION_VAL_PAL_N_ES NULL
#define OPTION_VAL_PAL_M_ES "PAL-M (Brasil)"
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_ES "Orientación de pantalla"
#define OPTION_VAL_HORIZONTAL_ES NULL
#define OPTION_VAL_VERTICAL_ES NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_ES "Ordenación de transparencias"
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_ES "Por bandas (rápida, menos precisa)"
#define OPTION_VAL_PER_TRIANGLE_NORMAL_ES "Por triángulos (normal)"
#define OPTION_VAL_PER_PIXEL_ACCURATE_ES "Por píxeles (precisa, pero lenta)"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_ES "Tamaño del búfer de acumulación de píxeles"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_ES "Es posible que sea necesario un valor alto para que la imagen se muestre bien a resoluciones superiores a la nativa."
#define OPTION_VAL_512MB_ES "512 MB"
#define OPTION_VAL_1GB_ES "1 GB"
#define OPTION_VAL_2GB_ES "2 GB"
#define OPTION_VAL_4GB_ES "4 GB"
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_ES "N.º máximo de capas transparentes"
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_ES "Las escenas más complejas podrían necesitar un valor más alto."
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_ES "Emulación completa del framebuffer"
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_ES "Activa la emulación completa del framebuffer dentro de la VRAM. Ideal para aquellos juegos que lean o escriban directamente al framebuffer en la VRAM. Al activar esta opción, se forzará una resolución interna de 640 × 480 y el rendimiento podría verse gravemente afectado."
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_ES "Activar el búfer RTT (renderizado a texturas)"
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_ES "Copia las texturas renderizadas en la GPU de vuelta a la VRAM. Esta opción suele dejarse activada para aquellos juegos que la requieran. Al activar esta opción, se desactivará el escalado de texturas y el rendimiento podría verse afectado."
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_ES "«Mipmapping»"
#define CORE_OPTION_NAME_FOG_LABEL_ES "Efectos de niebla"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_ES "Modificador de volúmenes"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_ES "Una característica de la GPU de Dreamcast que suelen utilizar los juegos para dibujar las sombras de objetos. Esta opción debe activarse por norma general; apenas afecta al rendimiento."
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_ES "Filtrado anisotrópico"
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_ES "Mejora la calidad de las texturas en aquellas superficies que se encuentren en un ángulo oblicuo respecto a la cámara. Un valor muy alto necesitará de mayores recursos de la GPU. Es necesario reiniciar para aplicar los cambios en este ajuste."
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_ES "Filtrado de texturas"
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_ES "Indica el modo de filtrado de texturas que se utilizará. Puede utilizarse para forzar un modo de filtrado concreto en todas las texturas con el fin de conseguir un aspecto más definido (o suavizado) que el valor predeterminado. Cualquier valor que no sea el predeterminado puede provocar problemas de renderizado. Es necesario reiniciar para aplicar los cambios en este ajuste."
#define OPTION_VAL_1_ES "Forzar vecino más cercano"
#define OPTION_VAL_2_ES "Forzar lineal"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_ES "Retrasar cambio de fotogramas"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_ES "Ideal para evitar parpadeos constantes o vídeos con fallos gráficos. No se recomienda su uso en plataformas lentas."
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_ES "Detectar cambios en la frecuencia de actualización"
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_ES "Informa en el front-end cuando cambie la frecuencia de actualización interna (p. ej.: de 60 a 30 FPS). Mejora la cadencia de fotogramas en aquellos juegos que se ejecuten a 30 o 20 FPS fijos, pero esta opción debería desactivarse en aquellos juegos con frecuencias de fotogramas desbloqueadas (es decir, inestables, p. ej.: Ecco the Dolphin, Unreal Tournament...). NOTA: esta opción no estará disponible si se activa Omisión automática de fotogramas."
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_ES "Filtro de posprocesamiento del PowerVR2"
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_ES "Posprocesa la imagen renderizada para simular los efectos dedicados de la GPU PowerVR2 y las señales de vídeo analógico."
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_ES "Reescalado de texturas (xBRZ)"
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_ES "Mejora los gráficos 2D de corte «pixel art» dibujado a mano. Utilizar solo con juegos en 2D pixelado."
#define OPTION_VAL_2_O27_ES NULL
#define OPTION_VAL_4_ES NULL
#define OPTION_VAL_6_ES NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_ES "Tamaño máximo del escalado de texturas"
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_ES "Interpolación nativa de profundidad"
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_ES "Mitiga la corrupción de texturas y los problemas de profundidad en GPU de AMD. En algunos casos podría ayudar también en GPU de Intel."
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_ES "Renderizado multihilo"
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_ES "Ejecuta la GPU y la CPU en hilos distintos. Muy recomendado."
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_ES "Omisión automática de fotogramas"
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_ES "Omite fotogramas de forma automática cuando el emulador se esté ejecutando lentamente. Nota: este ajuste solo surtirá efecto cuando la opción «Renderizado multihilo» esté activada."
#define OPTION_VAL_SOME_ES NULL
#define OPTION_VAL_MORE_ES "Valor máximo"
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_ES "Omisión de fotogramas"
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_ES "Establece la cantidad de fotogramas que se omitirán tras cada fotograma que se muestre."
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_ES "Trucos para pantallas panorámicas (es necesario reiniciar)"
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_ES "Activa los trucos que permiten a ciertos juegos mostrarse en formato panorámico."
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_ES "Arreglo para pantallas panorámicas"
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_ES "Dibuja la geometría más allá de la relación de aspecto normal de 4:3. Puede producir fallos gráficos en las zonas reveladas."
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_ES "Lectura rápida de GD-ROM (inexacta)"
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_ES "Acelera la carga de los GD-ROM."
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_ES "Cargar texturas personalizadas"
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_ES "Volcar texturas"
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_ES "Zona muerta del stick analógico"
#define OPTION_VAL_0_ES "0 %"
#define OPTION_VAL_5_ES "5 %"
#define OPTION_VAL_10_ES "10 %"
#define OPTION_VAL_15_ES "15 %"
#define OPTION_VAL_20_ES "20 %"
#define OPTION_VAL_25_ES "25 %"
#define OPTION_VAL_30_ES "30 %"
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_ES "Zona muerta de los gatillos"
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_ES "Gatillos digitales"
#define CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_ES "Puru Puru/Vibration Pack (Vibración)"
#define CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_ES "Activa la vibración del mando."
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_ES "Transmitir salidas digitales"
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_ES "Transmite las salidas digitales y el estado de la vibración al puerto TCP 8000. Compatible con la opción de MAME «-output network»."
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_ES "Mostrar ajustes de las pistolas de luz"
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_ES "Permite configurar los puntos de mira de las pistolas de luz. NOTA: es necesario salir y volver a entrar del menú rápido para que el cambio surta efecto."
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_ES "Mostrar punto de mira del arma 1"
#define OPTION_VAL_WHITE_ES "Blanco"
#define OPTION_VAL_RED_ES "Rojo"
#define OPTION_VAL_GREEN_ES "Verde"
#define OPTION_VAL_BLUE_ES "Azul"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_ES "Mostrar punto de mira del arma 2"
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_ES "Mostrar punto de mira del arma 3"
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_ES "Mostrar punto de mira del arma 4"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_ES "Separar Visual Memory (VMU) por juego"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_ES "VMUs individuales para cada juego"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_ES "Al desactivar esta opción, todos los juegos compartirán cuatro archivos de guardado de VMU (A1, B1, C1 y D1) que se encontrarán en el directorio del sistema de RetroArch. El ajuste «VMU A1» creará un archivo único de la VMU A1 en el directorio de guardados de RetroArch para cada juego que se ejecute. «Todas las VMU» creará cuatro archivos VMU únicos (A1, B1, C1 y D1) para cada juego que se ejecute."
#define OPTION_VAL_VMU_A1_ES NULL
#define OPTION_VAL_ALL_VMUS_ES "Todas las VMUs"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_ES "Mostrar ajustes de visualización de las Visual Memory (VMU)"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_ES "Mostrar ajustes de visualización de las VMU"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_ES "Permite configurar la visibilidad, tamaño, posición y color de las pantallas LCD emuladas de las VMU. NOTA: es necesario salir y volver a entrar del menú rápido para que el cambio surta efecto."
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_ES "Pantalla de la VMU 1"
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_ES "Posición de la pantalla de la VMU 1"
#define OPTION_VAL_UPPER_LEFT_ES "Esquina superior izquierda"
#define OPTION_VAL_UPPER_RIGHT_ES "Esquina superior derecha"
#define OPTION_VAL_LOWER_LEFT_ES "Esquina inferior izquierda"
#define OPTION_VAL_LOWER_RIGHT_ES "Esquina interior derecha"
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_ES "Tamaño de la pantalla de la VMU 1"
#define OPTION_VAL_1X_ES NULL
#define OPTION_VAL_3X_ES NULL
#define OPTION_VAL_5X_ES NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_ES "Color de píxeles encendidos de la VMU 1"
#define OPTION_VAL_DEFAULT_ON_00_ES "Predeterminado (encendido)"
#define OPTION_VAL_DEFAULT_OFF_01_ES "Predeterminado (apagado)"
#define OPTION_VAL_BLACK_02_ES "Negro"
#define OPTION_VAL_LIGHT_BLUE_04_ES "Azul claro"
#define OPTION_VAL_CYAN_06_ES "Cian"
#define OPTION_VAL_CYAN_BLUE_07_ES "Azul celeste"
#define OPTION_VAL_LIGHT_GREEN_08_ES "Verde claro"
#define OPTION_VAL_CYAN_GREEN_09_ES "Verde cian"
#define OPTION_VAL_LIGHT_CYAN_10_ES "Cian claro"
#define OPTION_VAL_PURPLE_12_ES "Violeta"
#define OPTION_VAL_LIGHT_PURPLE_13_ES "Violeta claro"
#define OPTION_VAL_YELLOW_14_ES "Amarillo"
#define OPTION_VAL_GRAY_15_ES "Gris"
#define OPTION_VAL_LIGHT_PURPLE_2_16_ES "Violeta claro (2)"
#define OPTION_VAL_LIGHT_GREEN_2_17_ES "Verde claro (2)"
#define OPTION_VAL_LIGHT_GREEN_3_18_ES "Verde claro (3)"
#define OPTION_VAL_LIGHT_CYAN_2_19_ES "Cian claro (2)"
#define OPTION_VAL_LIGHT_RED_2_20_ES "Rojo claro (2)"
#define OPTION_VAL_MAGENTA_21_ES NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_ES "Violeta claro (3)"
#define OPTION_VAL_LIGHT_ORANGE_23_ES "Naranja claro"
#define OPTION_VAL_ORANGE_24_ES "Naranja"
#define OPTION_VAL_LIGHT_PURPLE_4_25_ES "Violeta claro (4)"
#define OPTION_VAL_LIGHT_YELLOW_26_ES "Amarillo claro"
#define OPTION_VAL_LIGHT_YELLOW_2_27_ES "Amarillo claro (2)"
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_ES "Color de píxeles apagados de la VMU 1"
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_ES "Opacidad de la pantalla de la VMU 1"
#define OPTION_VAL_40_ES "40 %"
#define OPTION_VAL_50_ES "50 %"
#define OPTION_VAL_60_ES "60 %"
#define OPTION_VAL_70_ES "70 %"
#define OPTION_VAL_80_ES "80 %"
#define OPTION_VAL_90_ES "90 %"
#define OPTION_VAL_100_ES "100 %"
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_ES "Pantalla de la VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_ES "Posición de la pantalla de la VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_ES "Tamaño de la pantalla de la VMU 2"
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_ES "Color de píxeles encendidos de la VMU 2"
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_ES "Color de píxeles apagados de la VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_ES "Opacidad de la pantalla de la VMU 2"
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_ES "Pantalla de la VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_ES "Posición de la pantalla de la VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_ES "Tamaño de la pantalla de la VMU 3"
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_ES "Color de píxeles encendidos de la VMU 3"
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_ES "Color de píxeles apagados de la VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_ES "Opacidad de la pantalla de la VMU 3"
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_ES "Pantalla de la VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_ES "Posición de la pantalla de la VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_ES "Tamaño de la pantalla de la VMU 4"
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_ES "Color de píxeles encendidos de la VMU 4"
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_ES "Color de píxeles apagados de la VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_ES "Opacidad de la pantalla de la VMU 4"

struct retro_core_option_v2_category option_cats_es[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_ES,
      CATEGORY_SYSTEM_INFO_0_ES
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_ES,
      CATEGORY_VIDEO_INFO_0_ES
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_ES,
      CATEGORY_PERFORMANCE_INFO_0_ES
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_ES,
      CATEGORY_HACKS_INFO_0_ES
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_ES,
      CATEGORY_INPUT_INFO_0_ES
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_ES,
      CATEGORY_VMU_INFO_0_ES
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_es[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_ES,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_ES },
         { "USA",     OPTION_VAL_USA_ES },
         { "Europe",  OPTION_VAL_EUROPE_ES },
         { "Default", OPTION_VAL_DEFAULT_ES },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_ES,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_ES },
         { "English",  OPTION_VAL_ENGLISH_ES },
         { "German",   OPTION_VAL_GERMAN_ES },
         { "French",   OPTION_VAL_FRENCH_ES },
         { "Spanish",  OPTION_VAL_SPANISH_ES },
         { "Italian",  OPTION_VAL_ITALIAN_ES },
         { "Default",  OPTION_VAL_DEFAULT_ES },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_ES,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_boot_to_bios",
      CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_ES,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_ES,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_force_wince",
      CORE_OPTION_NAME_FORCE_WINCE_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_FORCE_WINCE_INFO_0_ES,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_ES,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_ES,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_ES,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_ES,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_ES,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_ES },
         { "640x480",    OPTION_VAL_640X480_ES },
         { "800x600",    OPTION_VAL_800X600_ES },
         { "960x720",    OPTION_VAL_960X720_ES },
         { "1024x768",   OPTION_VAL_1024X768_ES },
         { "1280x960",   OPTION_VAL_1280X960_ES },
         { "1440x1080",  OPTION_VAL_1440X1080_ES },
         { "1600x1200",  OPTION_VAL_1600X1200_ES },
         { "1920x1440",  OPTION_VAL_1920X1440_ES },
         { "2560x1920",  OPTION_VAL_2560X1920_ES },
         { "2880x2160",  OPTION_VAL_2880X2160_ES },
         { "3200x2400",  OPTION_VAL_3200X2400_ES },
         { "3840x2880",  OPTION_VAL_3840X2880_ES },
         { "4480x3360",  OPTION_VAL_4480X3360_ES },
         { "5120x3840",  OPTION_VAL_5120X3840_ES },
         { "5760x4320",  OPTION_VAL_5760X4320_ES },
         { "6400x4800",  OPTION_VAL_6400X4800_ES },
         { "7040x5280",  OPTION_VAL_7040X5280_ES },
         { "7680x5760",  OPTION_VAL_7680X5760_ES },
         { "8320x6240",  OPTION_VAL_8320X6240_ES },
         { "8960x6720",  OPTION_VAL_8960X6720_ES },
         { "9600x7200",  OPTION_VAL_9600X7200_ES },
         { "10240x7680", OPTION_VAL_10240X7680_ES },
         { "10880x8160", OPTION_VAL_10880X8160_ES },
         { "11520x8640", OPTION_VAL_11520X8640_ES },
         { "12160x9120", OPTION_VAL_12160X9120_ES },
         { "12800x9600", OPTION_VAL_12800X9600_ES },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_ES,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_ES },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_ES },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_ES },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_ES,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_ES },
         { "PAL",     OPTION_VAL_PAL_ES },
         { "PAL_N",   OPTION_VAL_PAL_N_ES },
         { "PAL_M",   OPTION_VAL_PAL_M_ES },
         { "Default", OPTION_VAL_DEFAULT_ES },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_ES,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_ES },
         { "vertical",   OPTION_VAL_VERTICAL_ES },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_ES,
      NULL,
      "",
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_ES },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_ES },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_ES },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_ES,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_ES },
         { "1GB",   OPTION_VAL_1GB_ES },
         { "2GB",   OPTION_VAL_2GB_ES },
         { "4GB",   OPTION_VAL_4GB_ES },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_ES,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_ES,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_ES,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_ES,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_ES,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_ES,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_ES,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_ES,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_ES },
         { "1",  OPTION_VAL_1_ES },
         { "2",  OPTION_VAL_2_ES },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_ES,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_ES,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_ES,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_ES,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O27_ES },
         { "4",  OPTION_VAL_4_ES },
         { "6",  OPTION_VAL_6_ES },
         { NULL, NULL },
      },
      "1",
   },
   {/* TODO: needs clarification */
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_ES,
      NULL,
      "",
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_ES,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_ES,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_ES,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_ES,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_ES },
         { "more", OPTION_VAL_MORE_ES },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_ES,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_ES,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_ES,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_ES,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_ES,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: probably needs explanation */
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_ES,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_ES,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_ES },
         { "5%",  OPTION_VAL_5_ES },
         { "10%", OPTION_VAL_10_ES },
         { "15%", OPTION_VAL_15_ES },
         { "20%", OPTION_VAL_20_ES },
         { "25%", OPTION_VAL_25_ES },
         { "30%", OPTION_VAL_30_ES },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_ES,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_ES },
         { "5%",  OPTION_VAL_5_ES },
         { "10%", OPTION_VAL_10_ES },
         { "15%", OPTION_VAL_15_ES },
         { "20%", OPTION_VAL_20_ES },
         { "25%", OPTION_VAL_25_ES },
         { "30%", OPTION_VAL_30_ES },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_ES,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_purupuru",
      CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_ES,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_ES,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_ES,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_ES,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_ES },
         { "Red",      OPTION_VAL_RED_ES },
         { "Green",    OPTION_VAL_GREEN_ES },
         { "Blue",     OPTION_VAL_BLUE_ES },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_ES,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_ES },
         { "Red",      OPTION_VAL_RED_ES },
         { "Green",    OPTION_VAL_GREEN_ES },
         { "Blue",     OPTION_VAL_BLUE_ES },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_ES,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_ES },
         { "Red",      OPTION_VAL_RED_ES },
         { "Green",    OPTION_VAL_GREEN_ES },
         { "Blue",     OPTION_VAL_BLUE_ES },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_ES,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_ES },
         { "Red",      OPTION_VAL_RED_ES },
         { "Green",    OPTION_VAL_GREEN_ES },
         { "Blue",     OPTION_VAL_BLUE_ES },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_ES,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_ES,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_ES,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_ES },
         { "All VMUs", OPTION_VAL_ALL_VMUS_ES },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_ES,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_ES,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_ES,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_ES },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_ES },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_ES },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_ES },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_ES },
         { "2x", OPTION_VAL_2_O27_ES },
         { "3x", OPTION_VAL_3X_ES },
         { "4x", OPTION_VAL_4_ES },
         { "5x", OPTION_VAL_5X_ES },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_ES },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_ES },
         { "BLACK 02",          OPTION_VAL_BLACK_02_ES },
         { "BLUE 03",           OPTION_VAL_BLUE_ES },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_ES },
         { "GREEN 05",          OPTION_VAL_GREEN_ES },
         { "CYAN 06",           OPTION_VAL_CYAN_06_ES },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_ES },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_ES },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_ES },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_ES },
         { "RED 11",            OPTION_VAL_RED_ES },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_ES },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_ES },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_ES },
         { "GRAY 15",           OPTION_VAL_GRAY_15_ES },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_ES },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_ES },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_ES },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_ES },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_ES },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_ES },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_ES },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_ES },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_ES },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_ES },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_ES },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_ES },
         { "WHITE 28",          OPTION_VAL_WHITE_ES },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_ES },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_ES },
         { "BLACK 02",          OPTION_VAL_BLACK_02_ES },
         { "BLUE 03",           OPTION_VAL_BLUE_ES },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_ES },
         { "GREEN 05",          OPTION_VAL_GREEN_ES },
         { "CYAN 06",           OPTION_VAL_CYAN_06_ES },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_ES },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_ES },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_ES },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_ES },
         { "RED 11",            OPTION_VAL_RED_ES },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_ES },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_ES },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_ES },
         { "GRAY 15",           OPTION_VAL_GRAY_15_ES },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_ES },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_ES },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_ES },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_ES },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_ES },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_ES },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_ES },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_ES },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_ES },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_ES },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_ES },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_ES },
         { "WHITE 28",          OPTION_VAL_WHITE_ES },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_ES },
         { "20%",  OPTION_VAL_20_ES },
         { "30%",  OPTION_VAL_30_ES },
         { "40%",  OPTION_VAL_40_ES },
         { "50%",  OPTION_VAL_50_ES },
         { "60%",  OPTION_VAL_60_ES },
         { "70%",  OPTION_VAL_70_ES },
         { "80%",  OPTION_VAL_80_ES },
         { "90%",  OPTION_VAL_90_ES },
         { "100%", OPTION_VAL_100_ES },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_ES },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_ES },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_ES },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_ES },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_ES },
         { "2x", OPTION_VAL_2_O27_ES },
         { "3x", OPTION_VAL_3X_ES },
         { "4x", OPTION_VAL_4_ES },
         { "5x", OPTION_VAL_5X_ES },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_ES },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_ES },
         { "BLACK 02",          OPTION_VAL_BLACK_02_ES },
         { "BLUE 03",           OPTION_VAL_BLUE_ES },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_ES },
         { "GREEN 05",          OPTION_VAL_GREEN_ES },
         { "CYAN 06",           OPTION_VAL_CYAN_06_ES },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_ES },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_ES },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_ES },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_ES },
         { "RED 11",            OPTION_VAL_RED_ES },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_ES },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_ES },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_ES },
         { "GRAY 15",           OPTION_VAL_GRAY_15_ES },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_ES },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_ES },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_ES },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_ES },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_ES },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_ES },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_ES },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_ES },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_ES },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_ES },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_ES },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_ES },
         { "WHITE 28",          OPTION_VAL_WHITE_ES },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_ES },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_ES },
         { "BLACK 02",          OPTION_VAL_BLACK_02_ES },
         { "BLUE 03",           OPTION_VAL_BLUE_ES },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_ES },
         { "GREEN 05",          OPTION_VAL_GREEN_ES },
         { "CYAN 06",           OPTION_VAL_CYAN_06_ES },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_ES },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_ES },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_ES },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_ES },
         { "RED 11",            OPTION_VAL_RED_ES },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_ES },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_ES },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_ES },
         { "GRAY 15",           OPTION_VAL_GRAY_15_ES },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_ES },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_ES },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_ES },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_ES },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_ES },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_ES },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_ES },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_ES },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_ES },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_ES },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_ES },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_ES },
         { "WHITE 28",          OPTION_VAL_WHITE_ES },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_ES },
         { "20%",  OPTION_VAL_20_ES },
         { "30%",  OPTION_VAL_30_ES },
         { "40%",  OPTION_VAL_40_ES },
         { "50%",  OPTION_VAL_50_ES },
         { "60%",  OPTION_VAL_60_ES },
         { "70%",  OPTION_VAL_70_ES },
         { "80%",  OPTION_VAL_80_ES },
         { "90%",  OPTION_VAL_90_ES },
         { "100%", OPTION_VAL_100_ES },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_ES },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_ES },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_ES },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_ES },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_ES },
         { "2x", OPTION_VAL_2_O27_ES },
         { "3x", OPTION_VAL_3X_ES },
         { "4x", OPTION_VAL_4_ES },
         { "5x", OPTION_VAL_5X_ES },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_ES },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_ES },
         { "BLACK 02",          OPTION_VAL_BLACK_02_ES },
         { "BLUE 03",           OPTION_VAL_BLUE_ES },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_ES },
         { "GREEN 05",          OPTION_VAL_GREEN_ES },
         { "CYAN 06",           OPTION_VAL_CYAN_06_ES },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_ES },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_ES },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_ES },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_ES },
         { "RED 11",            OPTION_VAL_RED_ES },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_ES },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_ES },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_ES },
         { "GRAY 15",           OPTION_VAL_GRAY_15_ES },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_ES },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_ES },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_ES },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_ES },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_ES },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_ES },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_ES },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_ES },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_ES },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_ES },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_ES },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_ES },
         { "WHITE 28",          OPTION_VAL_WHITE_ES },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_ES },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_ES },
         { "BLACK 02",          OPTION_VAL_BLACK_02_ES },
         { "BLUE 03",           OPTION_VAL_BLUE_ES },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_ES },
         { "GREEN 05",          OPTION_VAL_GREEN_ES },
         { "CYAN 06",           OPTION_VAL_CYAN_06_ES },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_ES },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_ES },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_ES },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_ES },
         { "RED 11",            OPTION_VAL_RED_ES },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_ES },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_ES },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_ES },
         { "GRAY 15",           OPTION_VAL_GRAY_15_ES },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_ES },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_ES },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_ES },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_ES },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_ES },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_ES },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_ES },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_ES },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_ES },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_ES },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_ES },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_ES },
         { "WHITE 28",          OPTION_VAL_WHITE_ES },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_ES },
         { "20%",  OPTION_VAL_20_ES },
         { "30%",  OPTION_VAL_30_ES },
         { "40%",  OPTION_VAL_40_ES },
         { "50%",  OPTION_VAL_50_ES },
         { "60%",  OPTION_VAL_60_ES },
         { "70%",  OPTION_VAL_70_ES },
         { "80%",  OPTION_VAL_80_ES },
         { "90%",  OPTION_VAL_90_ES },
         { "100%", OPTION_VAL_100_ES },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_ES },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_ES },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_ES },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_ES },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_ES },
         { "2x", OPTION_VAL_2_O27_ES },
         { "3x", OPTION_VAL_3X_ES },
         { "4x", OPTION_VAL_4_ES },
         { "5x", OPTION_VAL_5X_ES },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_ES },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_ES },
         { "BLACK 02",          OPTION_VAL_BLACK_02_ES },
         { "BLUE 03",           OPTION_VAL_BLUE_ES },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_ES },
         { "GREEN 05",          OPTION_VAL_GREEN_ES },
         { "CYAN 06",           OPTION_VAL_CYAN_06_ES },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_ES },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_ES },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_ES },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_ES },
         { "RED 11",            OPTION_VAL_RED_ES },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_ES },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_ES },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_ES },
         { "GRAY 15",           OPTION_VAL_GRAY_15_ES },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_ES },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_ES },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_ES },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_ES },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_ES },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_ES },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_ES },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_ES },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_ES },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_ES },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_ES },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_ES },
         { "WHITE 28",          OPTION_VAL_WHITE_ES },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_ES },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_ES },
         { "BLACK 02",          OPTION_VAL_BLACK_02_ES },
         { "BLUE 03",           OPTION_VAL_BLUE_ES },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_ES },
         { "GREEN 05",          OPTION_VAL_GREEN_ES },
         { "CYAN 06",           OPTION_VAL_CYAN_06_ES },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_ES },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_ES },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_ES },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_ES },
         { "RED 11",            OPTION_VAL_RED_ES },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_ES },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_ES },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_ES },
         { "GRAY 15",           OPTION_VAL_GRAY_15_ES },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_ES },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_ES },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_ES },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_ES },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_ES },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_ES },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_ES },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_ES },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_ES },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_ES },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_ES },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_ES },
         { "WHITE 28",          OPTION_VAL_WHITE_ES },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_ES },
         { "20%",  OPTION_VAL_20_ES },
         { "30%",  OPTION_VAL_30_ES },
         { "40%",  OPTION_VAL_40_ES },
         { "50%",  OPTION_VAL_50_ES },
         { "60%",  OPTION_VAL_60_ES },
         { "70%",  OPTION_VAL_70_ES },
         { "80%",  OPTION_VAL_80_ES },
         { "90%",  OPTION_VAL_90_ES },
         { "100%", OPTION_VAL_100_ES },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_es = {
   option_cats_es,
   option_defs_es
};

/* RETRO_LANGUAGE_FA */

#define CATEGORY_SYSTEM_LABEL_FA "سیستم"
#define CATEGORY_SYSTEM_INFO_0_FA NULL
#define CATEGORY_VIDEO_LABEL_FA "ویدیو"
#define CATEGORY_VIDEO_INFO_0_FA NULL
#define CATEGORY_PERFORMANCE_LABEL_FA "عملکرد"
#define CATEGORY_PERFORMANCE_INFO_0_FA NULL
#define CATEGORY_HACKS_LABEL_FA NULL
#define CATEGORY_HACKS_INFO_0_FA NULL
#define CATEGORY_INPUT_LABEL_FA "داده"
#define CATEGORY_INPUT_INFO_0_FA NULL
#define CATEGORY_VMU_LABEL_FA NULL
#define CATEGORY_VMU_INFO_0_FA NULL
#define CORE_OPTION_NAME_REGION_LABEL_FA "منطقه"
#define OPTION_VAL_JAPAN_FA NULL
#define OPTION_VAL_USA_FA NULL
#define OPTION_VAL_EUROPE_FA NULL
#define OPTION_VAL_DEFAULT_FA NULL
#define CORE_OPTION_NAME_LANGUAGE_LABEL_FA "زبان"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_FA NULL
#define OPTION_VAL_JAPANESE_FA NULL
#define OPTION_VAL_ENGLISH_FA NULL
#define OPTION_VAL_GERMAN_FA NULL
#define OPTION_VAL_FRENCH_FA NULL
#define OPTION_VAL_SPANISH_FA NULL
#define OPTION_VAL_ITALIAN_FA NULL
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_FA NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_FA NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_FA NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_FA NULL
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_FA NULL
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_FA NULL
#define CORE_OPTION_NAME_FORCE_WINCE_LABEL_FA NULL
#define CORE_OPTION_NAME_FORCE_WINCE_INFO_0_FA NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_FA NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_FA NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_FA NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_FA NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_FA NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_FA NULL
#define CORE_OPTION_NAME_UPNP_LABEL_FA NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_FA NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_FA NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_FA NULL
#define OPTION_VAL_320X240_FA NULL
#define OPTION_VAL_640X480_FA NULL
#define OPTION_VAL_800X600_FA NULL
#define OPTION_VAL_960X720_FA NULL
#define OPTION_VAL_1024X768_FA NULL
#define OPTION_VAL_1280X960_FA NULL
#define OPTION_VAL_1440X1080_FA NULL
#define OPTION_VAL_1600X1200_FA NULL
#define OPTION_VAL_1920X1440_FA NULL
#define OPTION_VAL_2560X1920_FA NULL
#define OPTION_VAL_2880X2160_FA NULL
#define OPTION_VAL_3200X2400_FA NULL
#define OPTION_VAL_3840X2880_FA NULL
#define OPTION_VAL_4480X3360_FA NULL
#define OPTION_VAL_5120X3840_FA NULL
#define OPTION_VAL_5760X4320_FA NULL
#define OPTION_VAL_6400X4800_FA NULL
#define OPTION_VAL_7040X5280_FA NULL
#define OPTION_VAL_7680X5760_FA NULL
#define OPTION_VAL_8320X6240_FA NULL
#define OPTION_VAL_8960X6720_FA NULL
#define OPTION_VAL_9600X7200_FA NULL
#define OPTION_VAL_10240X7680_FA NULL
#define OPTION_VAL_10880X8160_FA NULL
#define OPTION_VAL_11520X8640_FA NULL
#define OPTION_VAL_12160X9120_FA NULL
#define OPTION_VAL_12800X9600_FA NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_FA NULL
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_FA NULL
#define OPTION_VAL_VGA_FA NULL
#define OPTION_VAL_TV_RGB_FA NULL
#define OPTION_VAL_TV_COMPOSITE_FA NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_FA NULL
#define OPTION_VAL_NTSC_FA NULL
#define OPTION_VAL_PAL_FA NULL
#define OPTION_VAL_PAL_N_FA NULL
#define OPTION_VAL_PAL_M_FA NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_FA "جهت صفحه نمایش"
#define OPTION_VAL_HORIZONTAL_FA NULL
#define OPTION_VAL_VERTICAL_FA NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_FA NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_FA NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_FA NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_FA NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_FA NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_FA NULL
#define OPTION_VAL_512MB_FA NULL
#define OPTION_VAL_1GB_FA NULL
#define OPTION_VAL_2GB_FA NULL
#define OPTION_VAL_4GB_FA NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_FA NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_FA NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_FA NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_FA NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_FA NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_FA NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_FA NULL
#define CORE_OPTION_NAME_FOG_LABEL_FA NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_FA NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_FA NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_FA NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_FA NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_FA NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_FA NULL
#define OPTION_VAL_1_FA NULL
#define OPTION_VAL_2_FA NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_FA NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_FA NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_FA NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_FA NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_FA NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_FA NULL
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_FA NULL
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_FA NULL
#define OPTION_VAL_2_O27_FA NULL
#define OPTION_VAL_4_FA NULL
#define OPTION_VAL_6_FA NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_FA NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_FA NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_FA NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_FA NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_FA NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_FA NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_FA NULL
#define OPTION_VAL_SOME_FA NULL
#define OPTION_VAL_MORE_FA NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_FA NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_FA NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_FA NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_FA NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_FA NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_FA NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_FA NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_FA NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_FA NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_FA NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_FA NULL
#define OPTION_VAL_0_FA NULL
#define OPTION_VAL_5_FA NULL
#define OPTION_VAL_10_FA NULL
#define OPTION_VAL_15_FA NULL
#define OPTION_VAL_20_FA NULL
#define OPTION_VAL_25_FA NULL
#define OPTION_VAL_30_FA NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_FA NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_FA NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_FA NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_FA NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_FA NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_FA NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_FA NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_FA NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_FA NULL
#define OPTION_VAL_WHITE_FA NULL
#define OPTION_VAL_RED_FA NULL
#define OPTION_VAL_GREEN_FA NULL
#define OPTION_VAL_BLUE_FA NULL
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_FA NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_FA NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_FA NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_FA NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_FA NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_FA NULL
#define OPTION_VAL_VMU_A1_FA NULL
#define OPTION_VAL_ALL_VMUS_FA NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_FA NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_FA NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_FA NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_FA NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_FA NULL
#define OPTION_VAL_UPPER_LEFT_FA NULL
#define OPTION_VAL_UPPER_RIGHT_FA NULL
#define OPTION_VAL_LOWER_LEFT_FA NULL
#define OPTION_VAL_LOWER_RIGHT_FA NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_FA NULL
#define OPTION_VAL_1X_FA NULL
#define OPTION_VAL_3X_FA NULL
#define OPTION_VAL_5X_FA NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_FA NULL
#define OPTION_VAL_DEFAULT_ON_00_FA NULL
#define OPTION_VAL_DEFAULT_OFF_01_FA NULL
#define OPTION_VAL_BLACK_02_FA NULL
#define OPTION_VAL_LIGHT_BLUE_04_FA NULL
#define OPTION_VAL_CYAN_06_FA NULL
#define OPTION_VAL_CYAN_BLUE_07_FA NULL
#define OPTION_VAL_LIGHT_GREEN_08_FA NULL
#define OPTION_VAL_CYAN_GREEN_09_FA NULL
#define OPTION_VAL_LIGHT_CYAN_10_FA NULL
#define OPTION_VAL_PURPLE_12_FA NULL
#define OPTION_VAL_LIGHT_PURPLE_13_FA NULL
#define OPTION_VAL_YELLOW_14_FA NULL
#define OPTION_VAL_GRAY_15_FA NULL
#define OPTION_VAL_LIGHT_PURPLE_2_16_FA NULL
#define OPTION_VAL_LIGHT_GREEN_2_17_FA NULL
#define OPTION_VAL_LIGHT_GREEN_3_18_FA NULL
#define OPTION_VAL_LIGHT_CYAN_2_19_FA NULL
#define OPTION_VAL_LIGHT_RED_2_20_FA NULL
#define OPTION_VAL_MAGENTA_21_FA NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_FA NULL
#define OPTION_VAL_LIGHT_ORANGE_23_FA NULL
#define OPTION_VAL_ORANGE_24_FA NULL
#define OPTION_VAL_LIGHT_PURPLE_4_25_FA NULL
#define OPTION_VAL_LIGHT_YELLOW_26_FA NULL
#define OPTION_VAL_LIGHT_YELLOW_2_27_FA NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_FA NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_FA NULL
#define OPTION_VAL_40_FA NULL
#define OPTION_VAL_50_FA NULL
#define OPTION_VAL_60_FA NULL
#define OPTION_VAL_70_FA NULL
#define OPTION_VAL_80_FA NULL
#define OPTION_VAL_90_FA NULL
#define OPTION_VAL_100_FA NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_FA NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_FA NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_FA NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_FA NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_FA NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_FA NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_FA NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_FA NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_FA NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_FA NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_FA NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_FA NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_FA NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_FA NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_FA NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_FA NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_FA NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_FA NULL

struct retro_core_option_v2_category option_cats_fa[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_FA,
      CATEGORY_SYSTEM_INFO_0_FA
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_FA,
      CATEGORY_VIDEO_INFO_0_FA
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_FA,
      CATEGORY_PERFORMANCE_INFO_0_FA
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_FA,
      CATEGORY_HACKS_INFO_0_FA
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_FA,
      CATEGORY_INPUT_INFO_0_FA
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_FA,
      CATEGORY_VMU_INFO_0_FA
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_fa[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_FA,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_FA },
         { "USA",     OPTION_VAL_USA_FA },
         { "Europe",  OPTION_VAL_EUROPE_FA },
         { "Default", OPTION_VAL_DEFAULT_FA },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_FA,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_FA },
         { "English",  OPTION_VAL_ENGLISH_FA },
         { "German",   OPTION_VAL_GERMAN_FA },
         { "French",   OPTION_VAL_FRENCH_FA },
         { "Spanish",  OPTION_VAL_SPANISH_FA },
         { "Italian",  OPTION_VAL_ITALIAN_FA },
         { "Default",  OPTION_VAL_DEFAULT_FA },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_FA,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_boot_to_bios",
      CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_FA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_FA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_force_wince",
      CORE_OPTION_NAME_FORCE_WINCE_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_FORCE_WINCE_INFO_0_FA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_FA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_FA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_FA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_FA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_FA,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_FA },
         { "640x480",    OPTION_VAL_640X480_FA },
         { "800x600",    OPTION_VAL_800X600_FA },
         { "960x720",    OPTION_VAL_960X720_FA },
         { "1024x768",   OPTION_VAL_1024X768_FA },
         { "1280x960",   OPTION_VAL_1280X960_FA },
         { "1440x1080",  OPTION_VAL_1440X1080_FA },
         { "1600x1200",  OPTION_VAL_1600X1200_FA },
         { "1920x1440",  OPTION_VAL_1920X1440_FA },
         { "2560x1920",  OPTION_VAL_2560X1920_FA },
         { "2880x2160",  OPTION_VAL_2880X2160_FA },
         { "3200x2400",  OPTION_VAL_3200X2400_FA },
         { "3840x2880",  OPTION_VAL_3840X2880_FA },
         { "4480x3360",  OPTION_VAL_4480X3360_FA },
         { "5120x3840",  OPTION_VAL_5120X3840_FA },
         { "5760x4320",  OPTION_VAL_5760X4320_FA },
         { "6400x4800",  OPTION_VAL_6400X4800_FA },
         { "7040x5280",  OPTION_VAL_7040X5280_FA },
         { "7680x5760",  OPTION_VAL_7680X5760_FA },
         { "8320x6240",  OPTION_VAL_8320X6240_FA },
         { "8960x6720",  OPTION_VAL_8960X6720_FA },
         { "9600x7200",  OPTION_VAL_9600X7200_FA },
         { "10240x7680", OPTION_VAL_10240X7680_FA },
         { "10880x8160", OPTION_VAL_10880X8160_FA },
         { "11520x8640", OPTION_VAL_11520X8640_FA },
         { "12160x9120", OPTION_VAL_12160X9120_FA },
         { "12800x9600", OPTION_VAL_12800X9600_FA },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_FA,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_FA },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_FA },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_FA },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_FA,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_FA },
         { "PAL",     OPTION_VAL_PAL_FA },
         { "PAL_N",   OPTION_VAL_PAL_N_FA },
         { "PAL_M",   OPTION_VAL_PAL_M_FA },
         { "Default", OPTION_VAL_DEFAULT_FA },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_FA,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_FA },
         { "vertical",   OPTION_VAL_VERTICAL_FA },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_FA,
      NULL,
      "",
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_FA },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_FA },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_FA },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_FA,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_FA },
         { "1GB",   OPTION_VAL_1GB_FA },
         { "2GB",   OPTION_VAL_2GB_FA },
         { "4GB",   OPTION_VAL_4GB_FA },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_FA,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_FA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_FA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_FA,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_FA,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_FA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_FA,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_FA,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_FA },
         { "1",  OPTION_VAL_1_FA },
         { "2",  OPTION_VAL_2_FA },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_FA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_FA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_FA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_FA,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O27_FA },
         { "4",  OPTION_VAL_4_FA },
         { "6",  OPTION_VAL_6_FA },
         { NULL, NULL },
      },
      "1",
   },
   {/* TODO: needs clarification */
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_FA,
      NULL,
      "",
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_FA,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_FA,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_FA,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_FA,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_FA },
         { "more", OPTION_VAL_MORE_FA },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_FA,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_FA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_FA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_FA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_FA,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: probably needs explanation */
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_FA,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_FA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_FA },
         { "5%",  OPTION_VAL_5_FA },
         { "10%", OPTION_VAL_10_FA },
         { "15%", OPTION_VAL_15_FA },
         { "20%", OPTION_VAL_20_FA },
         { "25%", OPTION_VAL_25_FA },
         { "30%", OPTION_VAL_30_FA },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_FA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_FA },
         { "5%",  OPTION_VAL_5_FA },
         { "10%", OPTION_VAL_10_FA },
         { "15%", OPTION_VAL_15_FA },
         { "20%", OPTION_VAL_20_FA },
         { "25%", OPTION_VAL_25_FA },
         { "30%", OPTION_VAL_30_FA },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_FA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_purupuru",
      CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_FA,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_FA,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_FA,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_FA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_FA },
         { "Red",      OPTION_VAL_RED_FA },
         { "Green",    OPTION_VAL_GREEN_FA },
         { "Blue",     OPTION_VAL_BLUE_FA },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_FA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_FA },
         { "Red",      OPTION_VAL_RED_FA },
         { "Green",    OPTION_VAL_GREEN_FA },
         { "Blue",     OPTION_VAL_BLUE_FA },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_FA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_FA },
         { "Red",      OPTION_VAL_RED_FA },
         { "Green",    OPTION_VAL_GREEN_FA },
         { "Blue",     OPTION_VAL_BLUE_FA },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_FA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_FA },
         { "Red",      OPTION_VAL_RED_FA },
         { "Green",    OPTION_VAL_GREEN_FA },
         { "Blue",     OPTION_VAL_BLUE_FA },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_FA,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_FA,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_FA,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_FA },
         { "All VMUs", OPTION_VAL_ALL_VMUS_FA },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_FA,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_FA,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_FA,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_FA },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_FA },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_FA },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_FA },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_FA },
         { "2x", OPTION_VAL_2_O27_FA },
         { "3x", OPTION_VAL_3X_FA },
         { "4x", OPTION_VAL_4_FA },
         { "5x", OPTION_VAL_5X_FA },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FA },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FA },
         { "BLUE 03",           OPTION_VAL_BLUE_FA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FA },
         { "GREEN 05",          OPTION_VAL_GREEN_FA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FA },
         { "RED 11",            OPTION_VAL_RED_FA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FA },
         { "WHITE 28",          OPTION_VAL_WHITE_FA },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FA },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FA },
         { "BLUE 03",           OPTION_VAL_BLUE_FA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FA },
         { "GREEN 05",          OPTION_VAL_GREEN_FA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FA },
         { "RED 11",            OPTION_VAL_RED_FA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FA },
         { "WHITE 28",          OPTION_VAL_WHITE_FA },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_FA },
         { "20%",  OPTION_VAL_20_FA },
         { "30%",  OPTION_VAL_30_FA },
         { "40%",  OPTION_VAL_40_FA },
         { "50%",  OPTION_VAL_50_FA },
         { "60%",  OPTION_VAL_60_FA },
         { "70%",  OPTION_VAL_70_FA },
         { "80%",  OPTION_VAL_80_FA },
         { "90%",  OPTION_VAL_90_FA },
         { "100%", OPTION_VAL_100_FA },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_FA },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_FA },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_FA },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_FA },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_FA },
         { "2x", OPTION_VAL_2_O27_FA },
         { "3x", OPTION_VAL_3X_FA },
         { "4x", OPTION_VAL_4_FA },
         { "5x", OPTION_VAL_5X_FA },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FA },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FA },
         { "BLUE 03",           OPTION_VAL_BLUE_FA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FA },
         { "GREEN 05",          OPTION_VAL_GREEN_FA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FA },
         { "RED 11",            OPTION_VAL_RED_FA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FA },
         { "WHITE 28",          OPTION_VAL_WHITE_FA },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FA },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FA },
         { "BLUE 03",           OPTION_VAL_BLUE_FA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FA },
         { "GREEN 05",          OPTION_VAL_GREEN_FA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FA },
         { "RED 11",            OPTION_VAL_RED_FA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FA },
         { "WHITE 28",          OPTION_VAL_WHITE_FA },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_FA },
         { "20%",  OPTION_VAL_20_FA },
         { "30%",  OPTION_VAL_30_FA },
         { "40%",  OPTION_VAL_40_FA },
         { "50%",  OPTION_VAL_50_FA },
         { "60%",  OPTION_VAL_60_FA },
         { "70%",  OPTION_VAL_70_FA },
         { "80%",  OPTION_VAL_80_FA },
         { "90%",  OPTION_VAL_90_FA },
         { "100%", OPTION_VAL_100_FA },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_FA },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_FA },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_FA },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_FA },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_FA },
         { "2x", OPTION_VAL_2_O27_FA },
         { "3x", OPTION_VAL_3X_FA },
         { "4x", OPTION_VAL_4_FA },
         { "5x", OPTION_VAL_5X_FA },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FA },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FA },
         { "BLUE 03",           OPTION_VAL_BLUE_FA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FA },
         { "GREEN 05",          OPTION_VAL_GREEN_FA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FA },
         { "RED 11",            OPTION_VAL_RED_FA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FA },
         { "WHITE 28",          OPTION_VAL_WHITE_FA },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FA },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FA },
         { "BLUE 03",           OPTION_VAL_BLUE_FA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FA },
         { "GREEN 05",          OPTION_VAL_GREEN_FA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FA },
         { "RED 11",            OPTION_VAL_RED_FA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FA },
         { "WHITE 28",          OPTION_VAL_WHITE_FA },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_FA },
         { "20%",  OPTION_VAL_20_FA },
         { "30%",  OPTION_VAL_30_FA },
         { "40%",  OPTION_VAL_40_FA },
         { "50%",  OPTION_VAL_50_FA },
         { "60%",  OPTION_VAL_60_FA },
         { "70%",  OPTION_VAL_70_FA },
         { "80%",  OPTION_VAL_80_FA },
         { "90%",  OPTION_VAL_90_FA },
         { "100%", OPTION_VAL_100_FA },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_FA },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_FA },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_FA },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_FA },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_FA },
         { "2x", OPTION_VAL_2_O27_FA },
         { "3x", OPTION_VAL_3X_FA },
         { "4x", OPTION_VAL_4_FA },
         { "5x", OPTION_VAL_5X_FA },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FA },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FA },
         { "BLUE 03",           OPTION_VAL_BLUE_FA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FA },
         { "GREEN 05",          OPTION_VAL_GREEN_FA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FA },
         { "RED 11",            OPTION_VAL_RED_FA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FA },
         { "WHITE 28",          OPTION_VAL_WHITE_FA },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FA },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FA },
         { "BLUE 03",           OPTION_VAL_BLUE_FA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FA },
         { "GREEN 05",          OPTION_VAL_GREEN_FA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FA },
         { "RED 11",            OPTION_VAL_RED_FA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FA },
         { "WHITE 28",          OPTION_VAL_WHITE_FA },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_FA },
         { "20%",  OPTION_VAL_20_FA },
         { "30%",  OPTION_VAL_30_FA },
         { "40%",  OPTION_VAL_40_FA },
         { "50%",  OPTION_VAL_50_FA },
         { "60%",  OPTION_VAL_60_FA },
         { "70%",  OPTION_VAL_70_FA },
         { "80%",  OPTION_VAL_80_FA },
         { "90%",  OPTION_VAL_90_FA },
         { "100%", OPTION_VAL_100_FA },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_fa = {
   option_cats_fa,
   option_defs_fa
};

/* RETRO_LANGUAGE_FI */

#define CATEGORY_SYSTEM_LABEL_FI "Järjestelmä"
#define CATEGORY_SYSTEM_INFO_0_FI NULL
#define CATEGORY_VIDEO_LABEL_FI NULL
#define CATEGORY_VIDEO_INFO_0_FI NULL
#define CATEGORY_PERFORMANCE_LABEL_FI "Suorituskyky"
#define CATEGORY_PERFORMANCE_INFO_0_FI NULL
#define CATEGORY_HACKS_LABEL_FI "Emulointikikat"
#define CATEGORY_HACKS_INFO_0_FI NULL
#define CATEGORY_INPUT_LABEL_FI "Syöte"
#define CATEGORY_INPUT_INFO_0_FI NULL
#define CATEGORY_VMU_LABEL_FI NULL
#define CATEGORY_VMU_INFO_0_FI NULL
#define CORE_OPTION_NAME_REGION_LABEL_FI "Alue"
#define OPTION_VAL_JAPAN_FI NULL
#define OPTION_VAL_USA_FI NULL
#define OPTION_VAL_EUROPE_FI "Eurooppa"
#define OPTION_VAL_DEFAULT_FI "Oletus"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_FI "Kieli"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_FI NULL
#define OPTION_VAL_JAPANESE_FI NULL
#define OPTION_VAL_ENGLISH_FI NULL
#define OPTION_VAL_GERMAN_FI NULL
#define OPTION_VAL_FRENCH_FI NULL
#define OPTION_VAL_SPANISH_FI NULL
#define OPTION_VAL_ITALIAN_FI NULL
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_FI NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_FI NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_FI "Käynnistä BIOS:iin (Uudelleenkäynnistys vaaditaan)"
#define CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_FI "Käynnistä suoraan Dreamcastin BIOS-valikkoon."
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_FI "Käytä DSP:tä"
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_FI NULL
#define CORE_OPTION_NAME_FORCE_WINCE_LABEL_FI "Pakota Windows CE -tila"
#define CORE_OPTION_NAME_FORCE_WINCE_INFO_0_FI NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_FI NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_FI NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_FI NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_FI NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_FI "Laajakaista-adapterin emulaatio"
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_FI NULL
#define CORE_OPTION_NAME_UPNP_LABEL_FI "Käytä UPnP:tä"
#define CORE_OPTION_NAME_UPNP_INFO_0_FI "Käytä UPnP:tä Internet-reitittimen määrittämiseksi verkkopelejä varten."
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_FI "Sisäinen resoluutio"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_FI NULL
#define OPTION_VAL_320X240_FI NULL
#define OPTION_VAL_640X480_FI NULL
#define OPTION_VAL_800X600_FI NULL
#define OPTION_VAL_960X720_FI NULL
#define OPTION_VAL_1024X768_FI NULL
#define OPTION_VAL_1280X960_FI NULL
#define OPTION_VAL_1440X1080_FI NULL
#define OPTION_VAL_1600X1200_FI NULL
#define OPTION_VAL_1920X1440_FI NULL
#define OPTION_VAL_2560X1920_FI NULL
#define OPTION_VAL_2880X2160_FI NULL
#define OPTION_VAL_3200X2400_FI NULL
#define OPTION_VAL_3840X2880_FI NULL
#define OPTION_VAL_4480X3360_FI NULL
#define OPTION_VAL_5120X3840_FI NULL
#define OPTION_VAL_5760X4320_FI NULL
#define OPTION_VAL_6400X4800_FI NULL
#define OPTION_VAL_7040X5280_FI NULL
#define OPTION_VAL_7680X5760_FI NULL
#define OPTION_VAL_8320X6240_FI NULL
#define OPTION_VAL_8960X6720_FI NULL
#define OPTION_VAL_9600X7200_FI NULL
#define OPTION_VAL_10240X7680_FI NULL
#define OPTION_VAL_10880X8160_FI NULL
#define OPTION_VAL_11520X8640_FI NULL
#define OPTION_VAL_12160X9120_FI NULL
#define OPTION_VAL_12800X9600_FI NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_FI "Kaapelin tyyppi"
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_FI NULL
#define OPTION_VAL_VGA_FI NULL
#define OPTION_VAL_TV_RGB_FI NULL
#define OPTION_VAL_TV_COMPOSITE_FI NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_FI NULL
#define OPTION_VAL_NTSC_FI NULL
#define OPTION_VAL_PAL_FI NULL
#define OPTION_VAL_PAL_N_FI "PAL-N (Argentiina, Paraguay, Uruguay)"
#define OPTION_VAL_PAL_M_FI "PAL-M (Brasilia)"
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_FI "Näytön suunta"
#define OPTION_VAL_HORIZONTAL_FI NULL
#define OPTION_VAL_VERTICAL_FI NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_FI NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_FI NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_FI NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_FI NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_FI NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_FI NULL
#define OPTION_VAL_512MB_FI "512 Mt"
#define OPTION_VAL_1GB_FI "1 Gt"
#define OPTION_VAL_2GB_FI "2 Gt"
#define OPTION_VAL_4GB_FI "4 Gt"
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_FI NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_FI NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_FI NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_FI NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_FI NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_FI NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_FI NULL
#define CORE_OPTION_NAME_FOG_LABEL_FI "Sumutehosteet"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_FI NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_FI NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_FI "Anisotrooppinen suodatus"
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_FI NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_FI "Tekstuurisuodatus"
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_FI NULL
#define OPTION_VAL_1_FI NULL
#define OPTION_VAL_2_FI NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_FI NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_FI NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_FI NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_FI NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_FI "PowerVR2-jälkikäsittelysuodatin"
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_FI NULL
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_FI "Tekstuurin yliskaalaus (xBRZ)"
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_FI NULL
#define OPTION_VAL_2_O27_FI NULL
#define OPTION_VAL_4_FI NULL
#define OPTION_VAL_6_FI NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_FI NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_FI NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_FI NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_FI "Säikeistetty renderöinti"
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_FI NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_FI NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_FI NULL
#define OPTION_VAL_SOME_FI "Normaali"
#define OPTION_VAL_MORE_FI NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_FI NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_FI NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_FI NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_FI NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_FI NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_FI NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_FI NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_FI NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_FI NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_FI NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_FI NULL
#define OPTION_VAL_0_FI "0 %"
#define OPTION_VAL_5_FI "5 %"
#define OPTION_VAL_10_FI "10 %"
#define OPTION_VAL_15_FI "15 %"
#define OPTION_VAL_20_FI "20 %"
#define OPTION_VAL_25_FI "25 %"
#define OPTION_VAL_30_FI "30 %"
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_FI NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_FI NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_FI NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_FI NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_FI NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_FI NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_FI "Näytä valoaseen asetukset"
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_FI NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_FI NULL
#define OPTION_VAL_WHITE_FI "Valkoinen"
#define OPTION_VAL_RED_FI "Punainen"
#define OPTION_VAL_GREEN_FI "Vihreä"
#define OPTION_VAL_BLUE_FI "Sininen"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_FI NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_FI NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_FI NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_FI NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_FI NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_FI NULL
#define OPTION_VAL_VMU_A1_FI NULL
#define OPTION_VAL_ALL_VMUS_FI NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_FI NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_FI NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_FI NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_FI NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_FI NULL
#define OPTION_VAL_UPPER_LEFT_FI "Ylävasen"
#define OPTION_VAL_UPPER_RIGHT_FI "Yläoikea"
#define OPTION_VAL_LOWER_LEFT_FI "Alavasen"
#define OPTION_VAL_LOWER_RIGHT_FI "Alaoikea"
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_FI NULL
#define OPTION_VAL_1X_FI NULL
#define OPTION_VAL_3X_FI NULL
#define OPTION_VAL_5X_FI NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_FI NULL
#define OPTION_VAL_DEFAULT_ON_00_FI NULL
#define OPTION_VAL_DEFAULT_OFF_01_FI NULL
#define OPTION_VAL_BLACK_02_FI "Musta"
#define OPTION_VAL_LIGHT_BLUE_04_FI "Vaaleansininen"
#define OPTION_VAL_CYAN_06_FI "Syaani"
#define OPTION_VAL_CYAN_BLUE_07_FI "Turkoosi sininen"
#define OPTION_VAL_LIGHT_GREEN_08_FI "Vaaleanvihreä"
#define OPTION_VAL_CYAN_GREEN_09_FI "Syaani vihreä"
#define OPTION_VAL_LIGHT_CYAN_10_FI NULL
#define OPTION_VAL_PURPLE_12_FI "Violetti"
#define OPTION_VAL_LIGHT_PURPLE_13_FI NULL
#define OPTION_VAL_YELLOW_14_FI "Keltainen"
#define OPTION_VAL_GRAY_15_FI "Harmaa"
#define OPTION_VAL_LIGHT_PURPLE_2_16_FI NULL
#define OPTION_VAL_LIGHT_GREEN_2_17_FI NULL
#define OPTION_VAL_LIGHT_GREEN_3_18_FI NULL
#define OPTION_VAL_LIGHT_CYAN_2_19_FI NULL
#define OPTION_VAL_LIGHT_RED_2_20_FI NULL
#define OPTION_VAL_MAGENTA_21_FI NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_FI NULL
#define OPTION_VAL_LIGHT_ORANGE_23_FI NULL
#define OPTION_VAL_ORANGE_24_FI "Oranssi"
#define OPTION_VAL_LIGHT_PURPLE_4_25_FI NULL
#define OPTION_VAL_LIGHT_YELLOW_26_FI NULL
#define OPTION_VAL_LIGHT_YELLOW_2_27_FI NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_FI NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_FI NULL
#define OPTION_VAL_40_FI "40 %"
#define OPTION_VAL_50_FI "50 %"
#define OPTION_VAL_60_FI "60 %"
#define OPTION_VAL_70_FI "70 %"
#define OPTION_VAL_80_FI "80 %"
#define OPTION_VAL_90_FI "90 %"
#define OPTION_VAL_100_FI "100 %"
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_FI NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_FI NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_FI NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_FI NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_FI NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_FI NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_FI NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_FI NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_FI NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_FI NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_FI NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_FI NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_FI NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_FI NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_FI NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_FI NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_FI NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_FI NULL

struct retro_core_option_v2_category option_cats_fi[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_FI,
      CATEGORY_SYSTEM_INFO_0_FI
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_FI,
      CATEGORY_VIDEO_INFO_0_FI
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_FI,
      CATEGORY_PERFORMANCE_INFO_0_FI
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_FI,
      CATEGORY_HACKS_INFO_0_FI
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_FI,
      CATEGORY_INPUT_INFO_0_FI
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_FI,
      CATEGORY_VMU_INFO_0_FI
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_fi[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_FI,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_FI },
         { "USA",     OPTION_VAL_USA_FI },
         { "Europe",  OPTION_VAL_EUROPE_FI },
         { "Default", OPTION_VAL_DEFAULT_FI },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_FI,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_FI },
         { "English",  OPTION_VAL_ENGLISH_FI },
         { "German",   OPTION_VAL_GERMAN_FI },
         { "French",   OPTION_VAL_FRENCH_FI },
         { "Spanish",  OPTION_VAL_SPANISH_FI },
         { "Italian",  OPTION_VAL_ITALIAN_FI },
         { "Default",  OPTION_VAL_DEFAULT_FI },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_FI,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_boot_to_bios",
      CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_FI,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_FI,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_force_wince",
      CORE_OPTION_NAME_FORCE_WINCE_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_FORCE_WINCE_INFO_0_FI,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_FI,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_FI,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_FI,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_FI,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_FI,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_FI },
         { "640x480",    OPTION_VAL_640X480_FI },
         { "800x600",    OPTION_VAL_800X600_FI },
         { "960x720",    OPTION_VAL_960X720_FI },
         { "1024x768",   OPTION_VAL_1024X768_FI },
         { "1280x960",   OPTION_VAL_1280X960_FI },
         { "1440x1080",  OPTION_VAL_1440X1080_FI },
         { "1600x1200",  OPTION_VAL_1600X1200_FI },
         { "1920x1440",  OPTION_VAL_1920X1440_FI },
         { "2560x1920",  OPTION_VAL_2560X1920_FI },
         { "2880x2160",  OPTION_VAL_2880X2160_FI },
         { "3200x2400",  OPTION_VAL_3200X2400_FI },
         { "3840x2880",  OPTION_VAL_3840X2880_FI },
         { "4480x3360",  OPTION_VAL_4480X3360_FI },
         { "5120x3840",  OPTION_VAL_5120X3840_FI },
         { "5760x4320",  OPTION_VAL_5760X4320_FI },
         { "6400x4800",  OPTION_VAL_6400X4800_FI },
         { "7040x5280",  OPTION_VAL_7040X5280_FI },
         { "7680x5760",  OPTION_VAL_7680X5760_FI },
         { "8320x6240",  OPTION_VAL_8320X6240_FI },
         { "8960x6720",  OPTION_VAL_8960X6720_FI },
         { "9600x7200",  OPTION_VAL_9600X7200_FI },
         { "10240x7680", OPTION_VAL_10240X7680_FI },
         { "10880x8160", OPTION_VAL_10880X8160_FI },
         { "11520x8640", OPTION_VAL_11520X8640_FI },
         { "12160x9120", OPTION_VAL_12160X9120_FI },
         { "12800x9600", OPTION_VAL_12800X9600_FI },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_FI,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_FI },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_FI },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_FI },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_FI,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_FI },
         { "PAL",     OPTION_VAL_PAL_FI },
         { "PAL_N",   OPTION_VAL_PAL_N_FI },
         { "PAL_M",   OPTION_VAL_PAL_M_FI },
         { "Default", OPTION_VAL_DEFAULT_FI },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_FI,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_FI },
         { "vertical",   OPTION_VAL_VERTICAL_FI },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_FI,
      NULL,
      "",
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_FI },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_FI },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_FI },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_FI,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_FI },
         { "1GB",   OPTION_VAL_1GB_FI },
         { "2GB",   OPTION_VAL_2GB_FI },
         { "4GB",   OPTION_VAL_4GB_FI },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_FI,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_FI,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_FI,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_FI,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_FI,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_FI,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_FI,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_FI,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_FI },
         { "1",  OPTION_VAL_1_FI },
         { "2",  OPTION_VAL_2_FI },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_FI,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_FI,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_FI,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_FI,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O27_FI },
         { "4",  OPTION_VAL_4_FI },
         { "6",  OPTION_VAL_6_FI },
         { NULL, NULL },
      },
      "1",
   },
   {/* TODO: needs clarification */
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_FI,
      NULL,
      "",
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_FI,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_FI,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_FI,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_FI,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_FI },
         { "more", OPTION_VAL_MORE_FI },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_FI,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_FI,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_FI,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_FI,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_FI,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: probably needs explanation */
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_FI,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_FI,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_FI },
         { "5%",  OPTION_VAL_5_FI },
         { "10%", OPTION_VAL_10_FI },
         { "15%", OPTION_VAL_15_FI },
         { "20%", OPTION_VAL_20_FI },
         { "25%", OPTION_VAL_25_FI },
         { "30%", OPTION_VAL_30_FI },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_FI,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_FI },
         { "5%",  OPTION_VAL_5_FI },
         { "10%", OPTION_VAL_10_FI },
         { "15%", OPTION_VAL_15_FI },
         { "20%", OPTION_VAL_20_FI },
         { "25%", OPTION_VAL_25_FI },
         { "30%", OPTION_VAL_30_FI },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_FI,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_purupuru",
      CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_FI,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_FI,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_FI,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_FI,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_FI },
         { "Red",      OPTION_VAL_RED_FI },
         { "Green",    OPTION_VAL_GREEN_FI },
         { "Blue",     OPTION_VAL_BLUE_FI },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_FI,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_FI },
         { "Red",      OPTION_VAL_RED_FI },
         { "Green",    OPTION_VAL_GREEN_FI },
         { "Blue",     OPTION_VAL_BLUE_FI },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_FI,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_FI },
         { "Red",      OPTION_VAL_RED_FI },
         { "Green",    OPTION_VAL_GREEN_FI },
         { "Blue",     OPTION_VAL_BLUE_FI },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_FI,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_FI },
         { "Red",      OPTION_VAL_RED_FI },
         { "Green",    OPTION_VAL_GREEN_FI },
         { "Blue",     OPTION_VAL_BLUE_FI },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_FI,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_FI,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_FI,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_FI },
         { "All VMUs", OPTION_VAL_ALL_VMUS_FI },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_FI,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_FI,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_FI,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_FI },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_FI },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_FI },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_FI },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_FI },
         { "2x", OPTION_VAL_2_O27_FI },
         { "3x", OPTION_VAL_3X_FI },
         { "4x", OPTION_VAL_4_FI },
         { "5x", OPTION_VAL_5X_FI },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FI },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FI },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FI },
         { "BLUE 03",           OPTION_VAL_BLUE_FI },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FI },
         { "GREEN 05",          OPTION_VAL_GREEN_FI },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FI },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FI },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FI },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FI },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FI },
         { "RED 11",            OPTION_VAL_RED_FI },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FI },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FI },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FI },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FI },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FI },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FI },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FI },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FI },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FI },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FI },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FI },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FI },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FI },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FI },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FI },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FI },
         { "WHITE 28",          OPTION_VAL_WHITE_FI },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FI },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FI },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FI },
         { "BLUE 03",           OPTION_VAL_BLUE_FI },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FI },
         { "GREEN 05",          OPTION_VAL_GREEN_FI },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FI },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FI },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FI },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FI },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FI },
         { "RED 11",            OPTION_VAL_RED_FI },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FI },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FI },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FI },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FI },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FI },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FI },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FI },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FI },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FI },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FI },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FI },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FI },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FI },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FI },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FI },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FI },
         { "WHITE 28",          OPTION_VAL_WHITE_FI },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_FI },
         { "20%",  OPTION_VAL_20_FI },
         { "30%",  OPTION_VAL_30_FI },
         { "40%",  OPTION_VAL_40_FI },
         { "50%",  OPTION_VAL_50_FI },
         { "60%",  OPTION_VAL_60_FI },
         { "70%",  OPTION_VAL_70_FI },
         { "80%",  OPTION_VAL_80_FI },
         { "90%",  OPTION_VAL_90_FI },
         { "100%", OPTION_VAL_100_FI },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_FI },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_FI },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_FI },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_FI },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_FI },
         { "2x", OPTION_VAL_2_O27_FI },
         { "3x", OPTION_VAL_3X_FI },
         { "4x", OPTION_VAL_4_FI },
         { "5x", OPTION_VAL_5X_FI },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FI },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FI },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FI },
         { "BLUE 03",           OPTION_VAL_BLUE_FI },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FI },
         { "GREEN 05",          OPTION_VAL_GREEN_FI },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FI },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FI },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FI },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FI },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FI },
         { "RED 11",            OPTION_VAL_RED_FI },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FI },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FI },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FI },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FI },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FI },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FI },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FI },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FI },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FI },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FI },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FI },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FI },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FI },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FI },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FI },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FI },
         { "WHITE 28",          OPTION_VAL_WHITE_FI },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FI },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FI },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FI },
         { "BLUE 03",           OPTION_VAL_BLUE_FI },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FI },
         { "GREEN 05",          OPTION_VAL_GREEN_FI },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FI },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FI },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FI },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FI },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FI },
         { "RED 11",            OPTION_VAL_RED_FI },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FI },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FI },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FI },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FI },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FI },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FI },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FI },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FI },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FI },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FI },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FI },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FI },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FI },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FI },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FI },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FI },
         { "WHITE 28",          OPTION_VAL_WHITE_FI },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_FI },
         { "20%",  OPTION_VAL_20_FI },
         { "30%",  OPTION_VAL_30_FI },
         { "40%",  OPTION_VAL_40_FI },
         { "50%",  OPTION_VAL_50_FI },
         { "60%",  OPTION_VAL_60_FI },
         { "70%",  OPTION_VAL_70_FI },
         { "80%",  OPTION_VAL_80_FI },
         { "90%",  OPTION_VAL_90_FI },
         { "100%", OPTION_VAL_100_FI },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_FI },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_FI },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_FI },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_FI },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_FI },
         { "2x", OPTION_VAL_2_O27_FI },
         { "3x", OPTION_VAL_3X_FI },
         { "4x", OPTION_VAL_4_FI },
         { "5x", OPTION_VAL_5X_FI },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FI },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FI },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FI },
         { "BLUE 03",           OPTION_VAL_BLUE_FI },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FI },
         { "GREEN 05",          OPTION_VAL_GREEN_FI },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FI },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FI },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FI },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FI },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FI },
         { "RED 11",            OPTION_VAL_RED_FI },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FI },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FI },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FI },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FI },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FI },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FI },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FI },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FI },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FI },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FI },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FI },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FI },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FI },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FI },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FI },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FI },
         { "WHITE 28",          OPTION_VAL_WHITE_FI },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FI },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FI },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FI },
         { "BLUE 03",           OPTION_VAL_BLUE_FI },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FI },
         { "GREEN 05",          OPTION_VAL_GREEN_FI },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FI },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FI },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FI },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FI },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FI },
         { "RED 11",            OPTION_VAL_RED_FI },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FI },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FI },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FI },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FI },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FI },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FI },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FI },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FI },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FI },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FI },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FI },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FI },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FI },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FI },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FI },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FI },
         { "WHITE 28",          OPTION_VAL_WHITE_FI },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_FI },
         { "20%",  OPTION_VAL_20_FI },
         { "30%",  OPTION_VAL_30_FI },
         { "40%",  OPTION_VAL_40_FI },
         { "50%",  OPTION_VAL_50_FI },
         { "60%",  OPTION_VAL_60_FI },
         { "70%",  OPTION_VAL_70_FI },
         { "80%",  OPTION_VAL_80_FI },
         { "90%",  OPTION_VAL_90_FI },
         { "100%", OPTION_VAL_100_FI },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_FI },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_FI },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_FI },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_FI },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_FI },
         { "2x", OPTION_VAL_2_O27_FI },
         { "3x", OPTION_VAL_3X_FI },
         { "4x", OPTION_VAL_4_FI },
         { "5x", OPTION_VAL_5X_FI },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FI },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FI },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FI },
         { "BLUE 03",           OPTION_VAL_BLUE_FI },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FI },
         { "GREEN 05",          OPTION_VAL_GREEN_FI },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FI },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FI },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FI },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FI },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FI },
         { "RED 11",            OPTION_VAL_RED_FI },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FI },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FI },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FI },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FI },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FI },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FI },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FI },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FI },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FI },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FI },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FI },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FI },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FI },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FI },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FI },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FI },
         { "WHITE 28",          OPTION_VAL_WHITE_FI },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FI },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FI },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FI },
         { "BLUE 03",           OPTION_VAL_BLUE_FI },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FI },
         { "GREEN 05",          OPTION_VAL_GREEN_FI },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FI },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FI },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FI },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FI },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FI },
         { "RED 11",            OPTION_VAL_RED_FI },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FI },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FI },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FI },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FI },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FI },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FI },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FI },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FI },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FI },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FI },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FI },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FI },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FI },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FI },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FI },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FI },
         { "WHITE 28",          OPTION_VAL_WHITE_FI },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_FI },
         { "20%",  OPTION_VAL_20_FI },
         { "30%",  OPTION_VAL_30_FI },
         { "40%",  OPTION_VAL_40_FI },
         { "50%",  OPTION_VAL_50_FI },
         { "60%",  OPTION_VAL_60_FI },
         { "70%",  OPTION_VAL_70_FI },
         { "80%",  OPTION_VAL_80_FI },
         { "90%",  OPTION_VAL_90_FI },
         { "100%", OPTION_VAL_100_FI },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_fi = {
   option_cats_fi,
   option_defs_fi
};

/* RETRO_LANGUAGE_FR */

#define CATEGORY_SYSTEM_LABEL_FR "Système"
#define CATEGORY_SYSTEM_INFO_0_FR "Configurer les réglages de région, de langue, du BIOS et du matériel de base."
#define CATEGORY_VIDEO_LABEL_FR "Vidéo"
#define CATEGORY_VIDEO_INFO_0_FR "Configurez les réglages de résolution, de transparence indépendante de l'ordre et des effets visuels."
#define CATEGORY_PERFORMANCE_LABEL_FR "Performances"
#define CATEGORY_PERFORMANCE_INFO_0_FR "Configurer les réglages de rendu et de saut d'images."
#define CATEGORY_HACKS_LABEL_FR "Hacks d'émulation"
#define CATEGORY_HACKS_INFO_0_FR "Configurez les réglages de remplacements pour écrans larges, de vitesse de chargement des GD-ROM et de remplacement de textures."
#define CATEGORY_INPUT_LABEL_FR "Entrées"
#define CATEGORY_INPUT_INFO_0_FR "Configurer les réglages des manettes et pistolets."
#define CATEGORY_VMU_LABEL_FR "Unité de mémoire visuelle (VMU)"
#define CATEGORY_VMU_INFO_0_FR "Configurer les réglages de fichiers de sauvegarde VMU par jeu et de visibilité des VMU à l'écran."
#define CORE_OPTION_NAME_REGION_LABEL_FR "Région"
#define OPTION_VAL_JAPAN_FR "Japon"
#define OPTION_VAL_USA_FR "États-Unis"
#define OPTION_VAL_EUROPE_FR NULL
#define OPTION_VAL_DEFAULT_FR "Par défaut"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_FR "Langue "
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_FR "Modifie la langue utilisée par le BIOS et par tous les jeux qui contiennent plusieurs langues."
#define OPTION_VAL_JAPANESE_FR "Japonais"
#define OPTION_VAL_ENGLISH_FR "Anglais"
#define OPTION_VAL_GERMAN_FR "Allemand"
#define OPTION_VAL_FRENCH_FR "Français"
#define OPTION_VAL_SPANISH_FR "Espagnol"
#define OPTION_VAL_ITALIAN_FR "Italien"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_FR "BIOS HLE (Redémarrage requis)"
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_FR "Forcer l'utilisation d'un BIOS en émulation de haut niveau."
#define CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_FR "Démarrer sur le BIOS (Redémarrage requis)"
#define CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_FR "Démarrer directement sur le menu BIOS Dreamcast."
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_FR "Activer le DSP"
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_FR "Active l'émulation du DSP audio de la Dreamcast (processeur de signal numérique). Améliore la précision du son généré, mais augmente les exigences de performances."
#define CORE_OPTION_NAME_FORCE_WINCE_LABEL_FR "Forcer le mode Windows CE"
#define CORE_OPTION_NAME_FORCE_WINCE_INFO_0_FR "Activer l'émulation MMU (unité de gestion mémoire) complète et d'autres réglages pour les jeux Windows CE."
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_FR "Autoriser les boutons de service NAOMI"
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_FR "Active le bouton SERVICE pour NAOMI, pour entrer dans les paramètres de la borne."
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_FR "Définir les jeux NAOMI en Free Play"
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_FR "Modifier les réglages de pièces du jeu en mode gratuit."
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_FR "Émulation de l'adaptateur haut débit"
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_FR "Émuler l'adaptateur haut débit Ethernet au lieu du modem. (Redémarrage requis)"
#define CORE_OPTION_NAME_UPNP_LABEL_FR "Activer l'UPnP"
#define CORE_OPTION_NAME_UPNP_INFO_0_FR "Utilisez l'UPnP pour configurer automatiquement votre routeur Internet pour les jeux en ligne."
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_FR "Résolution interne"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_FR "Modifier la résolution de rendu."
#define OPTION_VAL_320X240_FR NULL
#define OPTION_VAL_640X480_FR NULL
#define OPTION_VAL_800X600_FR NULL
#define OPTION_VAL_960X720_FR NULL
#define OPTION_VAL_1024X768_FR NULL
#define OPTION_VAL_1280X960_FR NULL
#define OPTION_VAL_1440X1080_FR NULL
#define OPTION_VAL_1600X1200_FR NULL
#define OPTION_VAL_1920X1440_FR NULL
#define OPTION_VAL_2560X1920_FR NULL
#define OPTION_VAL_2880X2160_FR NULL
#define OPTION_VAL_3200X2400_FR NULL
#define OPTION_VAL_3840X2880_FR NULL
#define OPTION_VAL_4480X3360_FR NULL
#define OPTION_VAL_5120X3840_FR NULL
#define OPTION_VAL_5760X4320_FR NULL
#define OPTION_VAL_6400X4800_FR NULL
#define OPTION_VAL_7040X5280_FR NULL
#define OPTION_VAL_7680X5760_FR NULL
#define OPTION_VAL_8320X6240_FR NULL
#define OPTION_VAL_8960X6720_FR NULL
#define OPTION_VAL_9600X7200_FR NULL
#define OPTION_VAL_10240X7680_FR NULL
#define OPTION_VAL_10880X8160_FR NULL
#define OPTION_VAL_11520X8640_FR NULL
#define OPTION_VAL_12160X9120_FR NULL
#define OPTION_VAL_12800X9600_FR NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_FR "Type de câble"
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_FR "Le type de signal de sortie. 'TV (Composite)' est le plus largement pris en charge."
#define OPTION_VAL_VGA_FR NULL
#define OPTION_VAL_TV_RGB_FR "TV (RVB)"
#define OPTION_VAL_TV_COMPOSITE_FR NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_FR "Standard de diffusion"
#define OPTION_VAL_NTSC_FR NULL
#define OPTION_VAL_PAL_FR "PAL (Monde)"
#define OPTION_VAL_PAL_N_FR "PAL-N (Argentine, Paraguay, Uruguay)"
#define OPTION_VAL_PAL_M_FR "PAL-M (Brésil)"
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_FR "Orientation de l'écran"
#define OPTION_VAL_HORIZONTAL_FR "Horizontale"
#define OPTION_VAL_VERTICAL_FR "Verticale"
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_FR "Tri alpha"
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_FR "Par bande (rapide, moins précis)"
#define OPTION_VAL_PER_TRIANGLE_NORMAL_FR "Par triangle (normal)"
#define OPTION_VAL_PER_PIXEL_ACCURATE_FR "Par pixel (précis, mais le plus lent)"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_FR "Taille du tampon d'accumulation des pixels"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_FR "Des valeurs plus élevées peuvent être requises pour que les résolutions plus élevées s'affichent correctement."
#define OPTION_VAL_512MB_FR "512 Mo"
#define OPTION_VAL_1GB_FR "1 Go"
#define OPTION_VAL_2GB_FR "2 Go"
#define OPTION_VAL_4GB_FR "4 Go"
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_FR "Calques transparents maximum"
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_FR "Des valeurs plus élevées peuvent être requises pour les scènes complexes."
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_FR "Émulation complète du framebuffer"
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_FR "Active l'émulation complète du framebuffer dans la VRAM. Ceci est utile pour les jeux qui lisent ou écrivent directement le framebuffer dans la VRAM. Lorsque cette option est activée, la résolution interne est forcée à 640x480 et les performances peuvent être sévèrement affectées."
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_FR "Activer la mémoire tampon du RTT (rendu vers texture)"
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_FR "Copier les textures rendues depuis le processeur graphique vers la VRAM. Cette option est normalement activée pour les jeux qui en ont besoin. Lorsque cette option est activée, la mise à l'échelle du rendu des textures est désactivée et les performances peuvent être impactées."
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_FR NULL
#define CORE_OPTION_NAME_FOG_LABEL_FR "Effets de brouillard"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_FR "Modificateur du volume"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_FR "Une fonctionnalité du processeur graphique de la Dreamcast qui est généralement utilisée par les jeux pour le rendu des ombres d'objets. Cette option devrait normalement être activée — l'impact sur les performances va généralement de minimal à négligeable."
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_FR "Filtrage anisotrope"
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_FR "Améliore la qualité des textures sur les surfaces qui sont à des angles de vue obliques par rapport à la caméra. Des valeurs plus élevées sont plus exigeantes sur le processeur graphique. Les modifications de ce réglage ne sont appliqués qu'après redémarrage."
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_FR "Filtrage des textures"
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_FR "Le mode de filtrage des textures à utiliser. Cela peut être utilisé pour forcer un certain mode de filtrage de texture sur toutes les textures pour obtenir une apparence plus nette (ou plus lisse) que par défaut. Les valeurs autres que celles par défaut peuvent causer divers problèmes de rendu. Les modifications de ce réglage ne sont appliquées qu'après redémarrage."
#define OPTION_VAL_1_FR "Forcer au plus proche"
#define OPTION_VAL_2_FR "Forcer linéaire"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_FR "Retarder l'échange d'images"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_FR "Utile pour éviter les écrans clignotants ou les vidéos buggées. Non recommandé sur les plates-formes lentes."
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_FR "Détecter les changements de fréquence d'images"
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_FR "Notifier l'interface lorsque la fréquence d’images interne change (par exemple, de 60 i/s à 30 i/s). Améliore la cadence des images dans les jeux qui fonctionnent à 30 i/s ou 20 i/s, mais devrait être désactivé pour les jeux avec des fréquences d'images (non stables) déverrouillées (Ecco the Dolphin, Unreal Tournament par exemple). Remarque : Indisponible lorsque 'Saut d'images automatique' est activé."
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_FR "Filtre de post-traitement PowerVR2"
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_FR "Post-traitement de l'image rendue pour simuler les effets spécifiques au processeur graphique PowerVR2 et aux signaux vidéo analogiques."
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_FR "Agrandissement des textures (xBRZ)"
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_FR "Améliorer les graphismes en pixels 2D dessinés à la main. Devrait être utilisé uniquement avec les jeux en 2D pixelisés."
#define OPTION_VAL_2_O27_FR "x2"
#define OPTION_VAL_4_FR "x4"
#define OPTION_VAL_6_FR "x6"
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_FR "Facteur maximal du filtre d'agrandissement de textures"
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_FR "Interpolation de profondeur native"
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_FR "Aide pour la corruption de textures et les problèmes de profondeur sur les processeurs graphiques AMD. Peut aussi aider les processeurs graphiques Intel dans certains cas."
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_FR "Rendu sur plusieurs fils d'exécution"
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_FR "Fais tourner le processeur graphique et le processeur sur différents fils d'exécution. Fortement recommandé."
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_FR "Saut d'images automatique"
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_FR "Sauter automatiquement des images lorsque l'émulateur s'exécute lentement. Remarque : Cette option ne s'applique que lorsque 'Rendu sur plusieurs fils d'exécution' est activé."
#define OPTION_VAL_SOME_FR NULL
#define OPTION_VAL_MORE_FR "Maximal"
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_FR "Saut d'images"
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_FR "Définit le nombre d'images à sauter entre chaque image affichée."
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_FR "Cheats d'écran large (Redémarrage requis)"
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_FR "Active les cheats qui permettent à certains jeux de s'afficher au format écran large."
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_FR "Hack d'écran large"
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_FR "Afficher un rendu de la géométrie en dehors du rapport d'aspect normal de 4:3. Peut produire des bugs graphiques dans les zones révélées."
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_FR "Chargement rapide des GD-ROM (inexact)"
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_FR "Accélère le chargement des GD-ROM."
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_FR "Charger les textures personnalisées"
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_FR "Importer les textures"
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_FR "Deadzone du stick analogique"
#define OPTION_VAL_0_FR NULL
#define OPTION_VAL_5_FR NULL
#define OPTION_VAL_10_FR NULL
#define OPTION_VAL_15_FR NULL
#define OPTION_VAL_20_FR NULL
#define OPTION_VAL_25_FR NULL
#define OPTION_VAL_30_FR NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_FR "Deadzone des gâchettes"
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_FR "Gâchettes numériques"
#define CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_FR "Pack Puru Puru/vibration"
#define CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_FR "Active le retour de force de la manette."
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_FR "Diffuser les sorties numériques"
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_FR "Diffuser les sorties numériques et l'état du retour de force sur le port TCP 8000. Compatible avec l'option de MAME \"-output network\" ."
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_FR "Afficher les réglages du pistolet"
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_FR "Activer la configuration des options d'affichage du viseur des pistolets. REMARQUE : le menu rapide devrait être réouvert pour que ce paramètre prenne effet."
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_FR "Afficher le viseur du pistolet 1"
#define OPTION_VAL_WHITE_FR "Blanc"
#define OPTION_VAL_RED_FR "Rouge"
#define OPTION_VAL_GREEN_FR "Vert"
#define OPTION_VAL_BLUE_FR "Bleu"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_FR "Afficher le viseur du pistolet 2"
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_FR "Afficher le viseur du pistolet 3"
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_FR "Afficher le viseur du pistolet 4"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_FR "Unités/systèmes de mémoire visuelle par jeu (VMU)"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_FR "VMU par jeu"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_FR "Lorsque cette option est désactivée, tous les jeux partagent 4 fichiers de sauvegarde VMU (A1, B1, C1, D1) situés dans le dossier system de RetroArch. Le réglage 'VMU A1' crée un fichier VMU 'A1' unique dans le dossier de sauvegarde de RetroArch pour chaque jeu qui est lancé. Le paramètre 'Toutes les VMU' crée 4 fichiers VMU uniques (A1, B1, C1, D1) pour chaque jeu qui est lancé."
#define OPTION_VAL_VMU_A1_FR NULL
#define OPTION_VAL_ALL_VMUS_FR "Tous les VMU"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_FR "Afficher les réglage d'affichage de l'unité/système de mémoire visuelle (VMU)"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_FR "Afficher les réglages d'affichage du VMU"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_FR "Activer la configuration de la visibilité, de la taille, de la position et de la couleur de l'écran du VMU émulé. REMARQUE : Le menu rapide devrait être réouvert pour que ce réglage prenne effet."
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_FR "Affichage de l'écran du VMU 1"
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_FR "Position de l'écran du VMU 1"
#define OPTION_VAL_UPPER_LEFT_FR "En haut à gauche"
#define OPTION_VAL_UPPER_RIGHT_FR "En haut à droite"
#define OPTION_VAL_LOWER_LEFT_FR "En bas à gauche"
#define OPTION_VAL_LOWER_RIGHT_FR "En bas à droite"
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_FR "Taille de l'écran du VMU 1"
#define OPTION_VAL_1X_FR "x1"
#define OPTION_VAL_3X_FR "x3"
#define OPTION_VAL_5X_FR "x5"
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_FR "Couleur des pixels allumés de l'écran du VMU 1"
#define OPTION_VAL_DEFAULT_ON_00_FR "Par défaut activé"
#define OPTION_VAL_DEFAULT_OFF_01_FR "Par défaut désactivé"
#define OPTION_VAL_BLACK_02_FR "Noire"
#define OPTION_VAL_LIGHT_BLUE_04_FR "Bleu clair"
#define OPTION_VAL_CYAN_06_FR NULL
#define OPTION_VAL_CYAN_BLUE_07_FR "Bleu cyan"
#define OPTION_VAL_LIGHT_GREEN_08_FR "Vert clair"
#define OPTION_VAL_CYAN_GREEN_09_FR "Vert cyan"
#define OPTION_VAL_LIGHT_CYAN_10_FR "Cyan clair"
#define OPTION_VAL_PURPLE_12_FR "Violette"
#define OPTION_VAL_LIGHT_PURPLE_13_FR "Violet clair"
#define OPTION_VAL_YELLOW_14_FR "Jaune"
#define OPTION_VAL_GRAY_15_FR "Grise"
#define OPTION_VAL_LIGHT_PURPLE_2_16_FR "Violet clair (2)"
#define OPTION_VAL_LIGHT_GREEN_2_17_FR "Vert clair (2)"
#define OPTION_VAL_LIGHT_GREEN_3_18_FR "Vert clair (3)"
#define OPTION_VAL_LIGHT_CYAN_2_19_FR "Cyan clair (2)"
#define OPTION_VAL_LIGHT_RED_2_20_FR "Rouge clair (2)"
#define OPTION_VAL_MAGENTA_21_FR NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_FR "Violet clair (3)"
#define OPTION_VAL_LIGHT_ORANGE_23_FR "Orange clair"
#define OPTION_VAL_ORANGE_24_FR NULL
#define OPTION_VAL_LIGHT_PURPLE_4_25_FR "Violet clair (4)"
#define OPTION_VAL_LIGHT_YELLOW_26_FR "Jaune clair"
#define OPTION_VAL_LIGHT_YELLOW_2_27_FR "Jaune clair (2)"
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_FR "Couleur des pixels éteints de l'écran du VMU 1"
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_FR "Opacité de l'écran du VMU 1"
#define OPTION_VAL_40_FR NULL
#define OPTION_VAL_50_FR NULL
#define OPTION_VAL_60_FR NULL
#define OPTION_VAL_70_FR NULL
#define OPTION_VAL_80_FR NULL
#define OPTION_VAL_90_FR NULL
#define OPTION_VAL_100_FR NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_FR "Affichage de l'écran du VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_FR "Position de l'écran du VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_FR "Taille de l'écran du VMU 2"
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_FR "Couleur des pixels allumés de l'écran du VMU 2"
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_FR "Couleur des pixels éteints de l'écran du VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_FR "Opacité de l'écran du VMU 2"
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_FR "Affichage de l'écran du VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_FR "Position de l'écran du VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_FR "Taille de l'écran du VMU 3"
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_FR "Couleur des pixels allumés de l'écran du VMU 3"
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_FR "Couleur des pixels éteints de l'écran du VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_FR "Opacité de l'écran du VMU 3"
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_FR "Affichage de l'écran du VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_FR "Position de l'écran du VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_FR "Taille de l'écran du VMU 4"
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_FR "Couleur des pixels allumés de l'écran du VMU 4"
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_FR "Couleur des pixels éteints de l'écran du VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_FR "Opacité de l'écran du VMU 4"

struct retro_core_option_v2_category option_cats_fr[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_FR,
      CATEGORY_SYSTEM_INFO_0_FR
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_FR,
      CATEGORY_VIDEO_INFO_0_FR
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_FR,
      CATEGORY_PERFORMANCE_INFO_0_FR
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_FR,
      CATEGORY_HACKS_INFO_0_FR
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_FR,
      CATEGORY_INPUT_INFO_0_FR
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_FR,
      CATEGORY_VMU_INFO_0_FR
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_fr[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_FR,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_FR },
         { "USA",     OPTION_VAL_USA_FR },
         { "Europe",  OPTION_VAL_EUROPE_FR },
         { "Default", OPTION_VAL_DEFAULT_FR },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_FR,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_FR },
         { "English",  OPTION_VAL_ENGLISH_FR },
         { "German",   OPTION_VAL_GERMAN_FR },
         { "French",   OPTION_VAL_FRENCH_FR },
         { "Spanish",  OPTION_VAL_SPANISH_FR },
         { "Italian",  OPTION_VAL_ITALIAN_FR },
         { "Default",  OPTION_VAL_DEFAULT_FR },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_FR,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_boot_to_bios",
      CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_FR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_FR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_force_wince",
      CORE_OPTION_NAME_FORCE_WINCE_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_FORCE_WINCE_INFO_0_FR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_FR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_FR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_FR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_FR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_FR,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_FR },
         { "640x480",    OPTION_VAL_640X480_FR },
         { "800x600",    OPTION_VAL_800X600_FR },
         { "960x720",    OPTION_VAL_960X720_FR },
         { "1024x768",   OPTION_VAL_1024X768_FR },
         { "1280x960",   OPTION_VAL_1280X960_FR },
         { "1440x1080",  OPTION_VAL_1440X1080_FR },
         { "1600x1200",  OPTION_VAL_1600X1200_FR },
         { "1920x1440",  OPTION_VAL_1920X1440_FR },
         { "2560x1920",  OPTION_VAL_2560X1920_FR },
         { "2880x2160",  OPTION_VAL_2880X2160_FR },
         { "3200x2400",  OPTION_VAL_3200X2400_FR },
         { "3840x2880",  OPTION_VAL_3840X2880_FR },
         { "4480x3360",  OPTION_VAL_4480X3360_FR },
         { "5120x3840",  OPTION_VAL_5120X3840_FR },
         { "5760x4320",  OPTION_VAL_5760X4320_FR },
         { "6400x4800",  OPTION_VAL_6400X4800_FR },
         { "7040x5280",  OPTION_VAL_7040X5280_FR },
         { "7680x5760",  OPTION_VAL_7680X5760_FR },
         { "8320x6240",  OPTION_VAL_8320X6240_FR },
         { "8960x6720",  OPTION_VAL_8960X6720_FR },
         { "9600x7200",  OPTION_VAL_9600X7200_FR },
         { "10240x7680", OPTION_VAL_10240X7680_FR },
         { "10880x8160", OPTION_VAL_10880X8160_FR },
         { "11520x8640", OPTION_VAL_11520X8640_FR },
         { "12160x9120", OPTION_VAL_12160X9120_FR },
         { "12800x9600", OPTION_VAL_12800X9600_FR },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_FR,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_FR },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_FR },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_FR },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_FR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_FR },
         { "PAL",     OPTION_VAL_PAL_FR },
         { "PAL_N",   OPTION_VAL_PAL_N_FR },
         { "PAL_M",   OPTION_VAL_PAL_M_FR },
         { "Default", OPTION_VAL_DEFAULT_FR },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_FR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_FR },
         { "vertical",   OPTION_VAL_VERTICAL_FR },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_FR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_FR },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_FR },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_FR },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_FR,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_FR },
         { "1GB",   OPTION_VAL_1GB_FR },
         { "2GB",   OPTION_VAL_2GB_FR },
         { "4GB",   OPTION_VAL_4GB_FR },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_FR,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_FR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_FR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_FR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_FR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_FR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_FR,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_FR,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_FR },
         { "1",  OPTION_VAL_1_FR },
         { "2",  OPTION_VAL_2_FR },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_FR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_FR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_FR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_FR,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O27_FR },
         { "4",  OPTION_VAL_4_FR },
         { "6",  OPTION_VAL_6_FR },
         { NULL, NULL },
      },
      "1",
   },
   {/* TODO: needs clarification */
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_FR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_FR,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_FR,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_FR,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_FR,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_FR },
         { "more", OPTION_VAL_MORE_FR },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_FR,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_FR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_FR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_FR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_FR,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: probably needs explanation */
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_FR,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_FR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_FR },
         { "5%",  OPTION_VAL_5_FR },
         { "10%", OPTION_VAL_10_FR },
         { "15%", OPTION_VAL_15_FR },
         { "20%", OPTION_VAL_20_FR },
         { "25%", OPTION_VAL_25_FR },
         { "30%", OPTION_VAL_30_FR },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_FR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_FR },
         { "5%",  OPTION_VAL_5_FR },
         { "10%", OPTION_VAL_10_FR },
         { "15%", OPTION_VAL_15_FR },
         { "20%", OPTION_VAL_20_FR },
         { "25%", OPTION_VAL_25_FR },
         { "30%", OPTION_VAL_30_FR },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_FR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_purupuru",
      CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_FR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_FR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_FR,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_FR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_FR },
         { "Red",      OPTION_VAL_RED_FR },
         { "Green",    OPTION_VAL_GREEN_FR },
         { "Blue",     OPTION_VAL_BLUE_FR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_FR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_FR },
         { "Red",      OPTION_VAL_RED_FR },
         { "Green",    OPTION_VAL_GREEN_FR },
         { "Blue",     OPTION_VAL_BLUE_FR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_FR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_FR },
         { "Red",      OPTION_VAL_RED_FR },
         { "Green",    OPTION_VAL_GREEN_FR },
         { "Blue",     OPTION_VAL_BLUE_FR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_FR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_FR },
         { "Red",      OPTION_VAL_RED_FR },
         { "Green",    OPTION_VAL_GREEN_FR },
         { "Blue",     OPTION_VAL_BLUE_FR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_FR,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_FR,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_FR,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_FR },
         { "All VMUs", OPTION_VAL_ALL_VMUS_FR },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_FR,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_FR,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_FR,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_FR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_FR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_FR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_FR },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_FR },
         { "2x", OPTION_VAL_2_O27_FR },
         { "3x", OPTION_VAL_3X_FR },
         { "4x", OPTION_VAL_4_FR },
         { "5x", OPTION_VAL_5X_FR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FR },
         { "BLUE 03",           OPTION_VAL_BLUE_FR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FR },
         { "GREEN 05",          OPTION_VAL_GREEN_FR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FR },
         { "RED 11",            OPTION_VAL_RED_FR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FR },
         { "WHITE 28",          OPTION_VAL_WHITE_FR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FR },
         { "BLUE 03",           OPTION_VAL_BLUE_FR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FR },
         { "GREEN 05",          OPTION_VAL_GREEN_FR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FR },
         { "RED 11",            OPTION_VAL_RED_FR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FR },
         { "WHITE 28",          OPTION_VAL_WHITE_FR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_FR },
         { "20%",  OPTION_VAL_20_FR },
         { "30%",  OPTION_VAL_30_FR },
         { "40%",  OPTION_VAL_40_FR },
         { "50%",  OPTION_VAL_50_FR },
         { "60%",  OPTION_VAL_60_FR },
         { "70%",  OPTION_VAL_70_FR },
         { "80%",  OPTION_VAL_80_FR },
         { "90%",  OPTION_VAL_90_FR },
         { "100%", OPTION_VAL_100_FR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_FR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_FR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_FR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_FR },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_FR },
         { "2x", OPTION_VAL_2_O27_FR },
         { "3x", OPTION_VAL_3X_FR },
         { "4x", OPTION_VAL_4_FR },
         { "5x", OPTION_VAL_5X_FR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FR },
         { "BLUE 03",           OPTION_VAL_BLUE_FR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FR },
         { "GREEN 05",          OPTION_VAL_GREEN_FR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FR },
         { "RED 11",            OPTION_VAL_RED_FR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FR },
         { "WHITE 28",          OPTION_VAL_WHITE_FR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FR },
         { "BLUE 03",           OPTION_VAL_BLUE_FR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FR },
         { "GREEN 05",          OPTION_VAL_GREEN_FR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FR },
         { "RED 11",            OPTION_VAL_RED_FR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FR },
         { "WHITE 28",          OPTION_VAL_WHITE_FR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_FR },
         { "20%",  OPTION_VAL_20_FR },
         { "30%",  OPTION_VAL_30_FR },
         { "40%",  OPTION_VAL_40_FR },
         { "50%",  OPTION_VAL_50_FR },
         { "60%",  OPTION_VAL_60_FR },
         { "70%",  OPTION_VAL_70_FR },
         { "80%",  OPTION_VAL_80_FR },
         { "90%",  OPTION_VAL_90_FR },
         { "100%", OPTION_VAL_100_FR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_FR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_FR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_FR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_FR },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_FR },
         { "2x", OPTION_VAL_2_O27_FR },
         { "3x", OPTION_VAL_3X_FR },
         { "4x", OPTION_VAL_4_FR },
         { "5x", OPTION_VAL_5X_FR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FR },
         { "BLUE 03",           OPTION_VAL_BLUE_FR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FR },
         { "GREEN 05",          OPTION_VAL_GREEN_FR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FR },
         { "RED 11",            OPTION_VAL_RED_FR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FR },
         { "WHITE 28",          OPTION_VAL_WHITE_FR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FR },
         { "BLUE 03",           OPTION_VAL_BLUE_FR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FR },
         { "GREEN 05",          OPTION_VAL_GREEN_FR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FR },
         { "RED 11",            OPTION_VAL_RED_FR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FR },
         { "WHITE 28",          OPTION_VAL_WHITE_FR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_FR },
         { "20%",  OPTION_VAL_20_FR },
         { "30%",  OPTION_VAL_30_FR },
         { "40%",  OPTION_VAL_40_FR },
         { "50%",  OPTION_VAL_50_FR },
         { "60%",  OPTION_VAL_60_FR },
         { "70%",  OPTION_VAL_70_FR },
         { "80%",  OPTION_VAL_80_FR },
         { "90%",  OPTION_VAL_90_FR },
         { "100%", OPTION_VAL_100_FR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_FR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_FR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_FR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_FR },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_FR },
         { "2x", OPTION_VAL_2_O27_FR },
         { "3x", OPTION_VAL_3X_FR },
         { "4x", OPTION_VAL_4_FR },
         { "5x", OPTION_VAL_5X_FR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FR },
         { "BLUE 03",           OPTION_VAL_BLUE_FR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FR },
         { "GREEN 05",          OPTION_VAL_GREEN_FR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FR },
         { "RED 11",            OPTION_VAL_RED_FR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FR },
         { "WHITE 28",          OPTION_VAL_WHITE_FR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FR },
         { "BLUE 03",           OPTION_VAL_BLUE_FR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FR },
         { "GREEN 05",          OPTION_VAL_GREEN_FR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FR },
         { "RED 11",            OPTION_VAL_RED_FR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FR },
         { "WHITE 28",          OPTION_VAL_WHITE_FR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_FR },
         { "20%",  OPTION_VAL_20_FR },
         { "30%",  OPTION_VAL_30_FR },
         { "40%",  OPTION_VAL_40_FR },
         { "50%",  OPTION_VAL_50_FR },
         { "60%",  OPTION_VAL_60_FR },
         { "70%",  OPTION_VAL_70_FR },
         { "80%",  OPTION_VAL_80_FR },
         { "90%",  OPTION_VAL_90_FR },
         { "100%", OPTION_VAL_100_FR },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_fr = {
   option_cats_fr,
   option_defs_fr
};

/* RETRO_LANGUAGE_GL */

#define CATEGORY_SYSTEM_LABEL_GL NULL
#define CATEGORY_SYSTEM_INFO_0_GL NULL
#define CATEGORY_VIDEO_LABEL_GL "Vídeo"
#define CATEGORY_VIDEO_INFO_0_GL NULL
#define CATEGORY_PERFORMANCE_LABEL_GL NULL
#define CATEGORY_PERFORMANCE_INFO_0_GL NULL
#define CATEGORY_HACKS_LABEL_GL "Hacks de emulación"
#define CATEGORY_HACKS_INFO_0_GL NULL
#define CATEGORY_INPUT_LABEL_GL "Entrada"
#define CATEGORY_INPUT_INFO_0_GL NULL
#define CATEGORY_VMU_LABEL_GL NULL
#define CATEGORY_VMU_INFO_0_GL NULL
#define CORE_OPTION_NAME_REGION_LABEL_GL NULL
#define OPTION_VAL_JAPAN_GL NULL
#define OPTION_VAL_USA_GL NULL
#define OPTION_VAL_EUROPE_GL NULL
#define OPTION_VAL_DEFAULT_GL NULL
#define CORE_OPTION_NAME_LANGUAGE_LABEL_GL NULL
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_GL NULL
#define OPTION_VAL_JAPANESE_GL NULL
#define OPTION_VAL_ENGLISH_GL NULL
#define OPTION_VAL_GERMAN_GL NULL
#define OPTION_VAL_FRENCH_GL NULL
#define OPTION_VAL_SPANISH_GL NULL
#define OPTION_VAL_ITALIAN_GL NULL
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_GL NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_GL "Forza o uso de emulación BIOS de alto nivel."
#define CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_GL "Arranca na BIOS (requírese reiniciar)"
#define CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_GL "Arranca directamente no menú da BIOS de Dreamcast."
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_GL NULL
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_GL NULL
#define CORE_OPTION_NAME_FORCE_WINCE_LABEL_GL NULL
#define CORE_OPTION_NAME_FORCE_WINCE_INFO_0_GL NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_GL "Permite os botóns de servizo Naomi"
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_GL "Activa o botón SERVIZO para NAOMI, para entrar nas opcións da recreativa."
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_GL NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_GL NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_GL NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_GL NULL
#define CORE_OPTION_NAME_UPNP_LABEL_GL NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_GL NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_GL "Resolución interna"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_GL "Modificar a resolución do renderizado."
#define OPTION_VAL_320X240_GL NULL
#define OPTION_VAL_640X480_GL NULL
#define OPTION_VAL_800X600_GL NULL
#define OPTION_VAL_960X720_GL NULL
#define OPTION_VAL_1024X768_GL NULL
#define OPTION_VAL_1280X960_GL NULL
#define OPTION_VAL_1440X1080_GL NULL
#define OPTION_VAL_1600X1200_GL NULL
#define OPTION_VAL_1920X1440_GL NULL
#define OPTION_VAL_2560X1920_GL NULL
#define OPTION_VAL_2880X2160_GL NULL
#define OPTION_VAL_3200X2400_GL NULL
#define OPTION_VAL_3840X2880_GL NULL
#define OPTION_VAL_4480X3360_GL NULL
#define OPTION_VAL_5120X3840_GL NULL
#define OPTION_VAL_5760X4320_GL NULL
#define OPTION_VAL_6400X4800_GL NULL
#define OPTION_VAL_7040X5280_GL NULL
#define OPTION_VAL_7680X5760_GL NULL
#define OPTION_VAL_8320X6240_GL NULL
#define OPTION_VAL_8960X6720_GL NULL
#define OPTION_VAL_9600X7200_GL NULL
#define OPTION_VAL_10240X7680_GL NULL
#define OPTION_VAL_10880X8160_GL NULL
#define OPTION_VAL_11520X8640_GL NULL
#define OPTION_VAL_12160X9120_GL NULL
#define OPTION_VAL_12800X9600_GL NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_GL NULL
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_GL NULL
#define OPTION_VAL_VGA_GL NULL
#define OPTION_VAL_TV_RGB_GL NULL
#define OPTION_VAL_TV_COMPOSITE_GL NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_GL NULL
#define OPTION_VAL_NTSC_GL NULL
#define OPTION_VAL_PAL_GL NULL
#define OPTION_VAL_PAL_N_GL NULL
#define OPTION_VAL_PAL_M_GL NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_GL "Orientación da Pantalla"
#define OPTION_VAL_HORIZONTAL_GL NULL
#define OPTION_VAL_VERTICAL_GL NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_GL "Clasificación Alpha"
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_GL "Por franxa (rápido, menos preciso)"
#define OPTION_VAL_PER_TRIANGLE_NORMAL_GL "Por triángulo (normal)"
#define OPTION_VAL_PER_PIXEL_ACCURATE_GL NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_GL NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_GL "É posible que se precisen valores máis altos para que as resolucións máis altas saian correctamente."
#define OPTION_VAL_512MB_GL NULL
#define OPTION_VAL_1GB_GL NULL
#define OPTION_VAL_2GB_GL NULL
#define OPTION_VAL_4GB_GL NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_GL NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_GL NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_GL NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_GL NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_GL "Activa o buffer RTT (Render To Texture)"
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_GL NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_GL NULL
#define CORE_OPTION_NAME_FOG_LABEL_GL NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_GL NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_GL NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_GL "Filtrado anisotrópico"
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_GL NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_GL "Filtrado de Texturas"
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_GL NULL
#define OPTION_VAL_1_GL NULL
#define OPTION_VAL_2_GL NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_GL "Intercambio de fotogramas con retardo"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_GL "Útil para evitar o parpadeo nas pantallas ou os vídeos con fallos. Non recomendado en plataformas lentas."
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_GL NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_GL NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_GL "Filtrado de postprocesado PowerVR2"
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_GL "Postproceso da imaxe renderizada para simular efectos específicos da PowerVR2 GPU e os sinais analóxicos de vídeo."
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_GL "Aumento das texturas (xBRZ)"
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_GL "Millora dos gráficos 2D pixel art debuxados a man. Só se debería usar con xogos pixelados en 2D."
#define OPTION_VAL_2_O27_GL NULL
#define OPTION_VAL_4_GL NULL
#define OPTION_VAL_6_GL NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_GL "Aumento da escala das texturas. Tamaño filtrado"
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_GL NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_GL NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_GL "Renderizado de fíos"
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_GL "Fai funcionar a GPU e a CPU en fíos diferentes. Altamente recomendado."
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_GL "Salto de fotograma automático"
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_GL "Automáticamente salta fotogramas cando o emulador está correndo lento. Nota: Esta opción só se aplica cando 'Renderizado por fíos' está activado."
#define OPTION_VAL_SOME_GL NULL
#define OPTION_VAL_MORE_GL "Máximo"
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_GL "Salto de fotogramas"
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_GL "Especifica o número de fotogramas para saltar entre cada fotograma que é mostrado."
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_GL NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_GL NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_GL NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_GL NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_GL NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_GL NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_GL "Carga as texturas personalizadas"
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_GL "Texturas de volcado"
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_GL NULL
#define OPTION_VAL_0_GL NULL
#define OPTION_VAL_5_GL NULL
#define OPTION_VAL_10_GL NULL
#define OPTION_VAL_15_GL NULL
#define OPTION_VAL_20_GL NULL
#define OPTION_VAL_25_GL NULL
#define OPTION_VAL_30_GL NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_GL NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_GL NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_GL "Pack Purupuru/Pack de vibración"
#define CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_GL "Activa a retroalimentación de forza do controlador."
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_GL NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_GL NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_GL NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_GL NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_GL NULL
#define OPTION_VAL_WHITE_GL "Branco"
#define OPTION_VAL_RED_GL "Vermello"
#define OPTION_VAL_GREEN_GL "Verde"
#define OPTION_VAL_BLUE_GL "Azul"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_GL NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_GL NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_GL NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_GL "Unidades/sistemas de memoria visual (VMU) por xogo"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_GL "VMUs por xogo"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_GL "Cando se desactiva, todos os xogos comparten 4 arquivos de gardado VMU (A1, B1, C1, D1) localizados no directorio do sistema de RetroArch. O axuste \"VMU A1\" crea un único arquivo VMU \"A1\" no directorio de gardado de RetroArch para cada xogo que se inicie. A opción \"Todas as VMU\" crea 4 arquivos VMU únicos (A1, B1, C1, D1) para cada xogo que se inicie."
#define OPTION_VAL_VMU_A1_GL NULL
#define OPTION_VAL_ALL_VMUS_GL "Todas as VMU"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_GL "Amosa as opcións de pantalla da Unidade/Sistema de Memoria visual (VMU)"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_GL "Amosa as opcións de pantalla da VMU"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_GL NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_GL "Pantalla 1 da VMU"
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_GL "Posición da pantalla 1 VMU"
#define OPTION_VAL_UPPER_LEFT_GL "Superior esquerda"
#define OPTION_VAL_UPPER_RIGHT_GL "Superior dereita"
#define OPTION_VAL_LOWER_LEFT_GL "Inferior esquerda"
#define OPTION_VAL_LOWER_RIGHT_GL "Inferior dereita"
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_GL "Tamaño da pantalla 1 VMU"
#define OPTION_VAL_1X_GL NULL
#define OPTION_VAL_3X_GL NULL
#define OPTION_VAL_5X_GL NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_GL "Activar pixel en cor na pantalla 1 VMU"
#define OPTION_VAL_DEFAULT_ON_00_GL "Por defecto activado"
#define OPTION_VAL_DEFAULT_OFF_01_GL "Por defecto desactivado"
#define OPTION_VAL_BLACK_02_GL "Negro"
#define OPTION_VAL_LIGHT_BLUE_04_GL "Azul claro"
#define OPTION_VAL_CYAN_06_GL "Azul ceo"
#define OPTION_VAL_CYAN_BLUE_07_GL "Cián"
#define OPTION_VAL_LIGHT_GREEN_08_GL "Verde claro"
#define OPTION_VAL_CYAN_GREEN_09_GL "Cián Verde"
#define OPTION_VAL_LIGHT_CYAN_10_GL "Cián claro"
#define OPTION_VAL_PURPLE_12_GL "Malva"
#define OPTION_VAL_LIGHT_PURPLE_13_GL "Malva claro"
#define OPTION_VAL_YELLOW_14_GL "Amarelo"
#define OPTION_VAL_GRAY_15_GL "Gris"
#define OPTION_VAL_LIGHT_PURPLE_2_16_GL "Malva claro (2)"
#define OPTION_VAL_LIGHT_GREEN_2_17_GL "Verde claro (2)"
#define OPTION_VAL_LIGHT_GREEN_3_18_GL "Verde claro (3)"
#define OPTION_VAL_LIGHT_CYAN_2_19_GL "Cián claro (2)"
#define OPTION_VAL_LIGHT_RED_2_20_GL "Vermello claro (2)"
#define OPTION_VAL_MAGENTA_21_GL "Maxenta"
#define OPTION_VAL_LIGHT_PURPLE_3_22_GL "Malva claro (3)"
#define OPTION_VAL_LIGHT_ORANGE_23_GL "Laranxa claro"
#define OPTION_VAL_ORANGE_24_GL "Laranxa"
#define OPTION_VAL_LIGHT_PURPLE_4_25_GL "Malva claro (4)"
#define OPTION_VAL_LIGHT_YELLOW_26_GL "Amarelo claro"
#define OPTION_VAL_LIGHT_YELLOW_2_27_GL "Amarelo claro (2)"
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_GL "Desactivar píxel en cor na pantalla 1 VMU"
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_GL "Opacidade da pantalla 1 VMU"
#define OPTION_VAL_40_GL NULL
#define OPTION_VAL_50_GL NULL
#define OPTION_VAL_60_GL NULL
#define OPTION_VAL_70_GL NULL
#define OPTION_VAL_80_GL NULL
#define OPTION_VAL_90_GL NULL
#define OPTION_VAL_100_GL NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_GL "Pantalla 2 da VMU"
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_GL "Posición da pantalla 2 VMU"
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_GL "Tamaño da pantalla 2 VMU"
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_GL "Activar pixel en cor na pantalla 2 VMU"
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_GL "Desactivar píxel en cor na pantalla 2 VMU"
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_GL "Opacidade da pantalla 2 VMU"
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_GL "Pantalla 3 da VMU"
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_GL "Posición da pantalla 3 VMU"
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_GL "Tamaño da pantalla 3 VMU"
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_GL "Activar pixel en cor na pantalla 3 VMU"
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_GL "Desactivar píxel en cor na pantalla 3 VMU"
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_GL "Opacidade da pantalla 3 VMU"
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_GL "Pantalla 4 da VMU"
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_GL "Posición da pantalla 4 VMU"
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_GL "Tamaño da pantalla 4 VMU"
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_GL "Activar pixel en cor na pantalla 4 VMU"
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_GL "Desactivar píxel en cor na pantalla 4 VMU"
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_GL "Opacidade da pantalla 4 VMU"

struct retro_core_option_v2_category option_cats_gl[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_GL,
      CATEGORY_SYSTEM_INFO_0_GL
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_GL,
      CATEGORY_VIDEO_INFO_0_GL
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_GL,
      CATEGORY_PERFORMANCE_INFO_0_GL
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_GL,
      CATEGORY_HACKS_INFO_0_GL
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_GL,
      CATEGORY_INPUT_INFO_0_GL
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_GL,
      CATEGORY_VMU_INFO_0_GL
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_gl[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_GL,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_GL },
         { "USA",     OPTION_VAL_USA_GL },
         { "Europe",  OPTION_VAL_EUROPE_GL },
         { "Default", OPTION_VAL_DEFAULT_GL },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_GL,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_GL },
         { "English",  OPTION_VAL_ENGLISH_GL },
         { "German",   OPTION_VAL_GERMAN_GL },
         { "French",   OPTION_VAL_FRENCH_GL },
         { "Spanish",  OPTION_VAL_SPANISH_GL },
         { "Italian",  OPTION_VAL_ITALIAN_GL },
         { "Default",  OPTION_VAL_DEFAULT_GL },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_GL,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_boot_to_bios",
      CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_GL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_GL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_force_wince",
      CORE_OPTION_NAME_FORCE_WINCE_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_FORCE_WINCE_INFO_0_GL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_GL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_GL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_GL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_GL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_GL,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_GL },
         { "640x480",    OPTION_VAL_640X480_GL },
         { "800x600",    OPTION_VAL_800X600_GL },
         { "960x720",    OPTION_VAL_960X720_GL },
         { "1024x768",   OPTION_VAL_1024X768_GL },
         { "1280x960",   OPTION_VAL_1280X960_GL },
         { "1440x1080",  OPTION_VAL_1440X1080_GL },
         { "1600x1200",  OPTION_VAL_1600X1200_GL },
         { "1920x1440",  OPTION_VAL_1920X1440_GL },
         { "2560x1920",  OPTION_VAL_2560X1920_GL },
         { "2880x2160",  OPTION_VAL_2880X2160_GL },
         { "3200x2400",  OPTION_VAL_3200X2400_GL },
         { "3840x2880",  OPTION_VAL_3840X2880_GL },
         { "4480x3360",  OPTION_VAL_4480X3360_GL },
         { "5120x3840",  OPTION_VAL_5120X3840_GL },
         { "5760x4320",  OPTION_VAL_5760X4320_GL },
         { "6400x4800",  OPTION_VAL_6400X4800_GL },
         { "7040x5280",  OPTION_VAL_7040X5280_GL },
         { "7680x5760",  OPTION_VAL_7680X5760_GL },
         { "8320x6240",  OPTION_VAL_8320X6240_GL },
         { "8960x6720",  OPTION_VAL_8960X6720_GL },
         { "9600x7200",  OPTION_VAL_9600X7200_GL },
         { "10240x7680", OPTION_VAL_10240X7680_GL },
         { "10880x8160", OPTION_VAL_10880X8160_GL },
         { "11520x8640", OPTION_VAL_11520X8640_GL },
         { "12160x9120", OPTION_VAL_12160X9120_GL },
         { "12800x9600", OPTION_VAL_12800X9600_GL },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_GL,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_GL },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_GL },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_GL },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_GL,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_GL },
         { "PAL",     OPTION_VAL_PAL_GL },
         { "PAL_N",   OPTION_VAL_PAL_N_GL },
         { "PAL_M",   OPTION_VAL_PAL_M_GL },
         { "Default", OPTION_VAL_DEFAULT_GL },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_GL,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_GL },
         { "vertical",   OPTION_VAL_VERTICAL_GL },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_GL,
      NULL,
      "",
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_GL },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_GL },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_GL },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_GL,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_GL },
         { "1GB",   OPTION_VAL_1GB_GL },
         { "2GB",   OPTION_VAL_2GB_GL },
         { "4GB",   OPTION_VAL_4GB_GL },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_GL,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_GL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_GL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_GL,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_GL,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_GL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_GL,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_GL,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_GL },
         { "1",  OPTION_VAL_1_GL },
         { "2",  OPTION_VAL_2_GL },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_GL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_GL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_GL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_GL,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O27_GL },
         { "4",  OPTION_VAL_4_GL },
         { "6",  OPTION_VAL_6_GL },
         { NULL, NULL },
      },
      "1",
   },
   {/* TODO: needs clarification */
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_GL,
      NULL,
      "",
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_GL,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_GL,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_GL,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_GL,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_GL },
         { "more", OPTION_VAL_MORE_GL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_GL,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_GL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_GL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_GL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_GL,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: probably needs explanation */
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_GL,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_GL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_GL },
         { "5%",  OPTION_VAL_5_GL },
         { "10%", OPTION_VAL_10_GL },
         { "15%", OPTION_VAL_15_GL },
         { "20%", OPTION_VAL_20_GL },
         { "25%", OPTION_VAL_25_GL },
         { "30%", OPTION_VAL_30_GL },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_GL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_GL },
         { "5%",  OPTION_VAL_5_GL },
         { "10%", OPTION_VAL_10_GL },
         { "15%", OPTION_VAL_15_GL },
         { "20%", OPTION_VAL_20_GL },
         { "25%", OPTION_VAL_25_GL },
         { "30%", OPTION_VAL_30_GL },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_GL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_purupuru",
      CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_GL,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_GL,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_GL,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_GL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_GL },
         { "Red",      OPTION_VAL_RED_GL },
         { "Green",    OPTION_VAL_GREEN_GL },
         { "Blue",     OPTION_VAL_BLUE_GL },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_GL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_GL },
         { "Red",      OPTION_VAL_RED_GL },
         { "Green",    OPTION_VAL_GREEN_GL },
         { "Blue",     OPTION_VAL_BLUE_GL },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_GL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_GL },
         { "Red",      OPTION_VAL_RED_GL },
         { "Green",    OPTION_VAL_GREEN_GL },
         { "Blue",     OPTION_VAL_BLUE_GL },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_GL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_GL },
         { "Red",      OPTION_VAL_RED_GL },
         { "Green",    OPTION_VAL_GREEN_GL },
         { "Blue",     OPTION_VAL_BLUE_GL },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_GL,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_GL,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_GL,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_GL },
         { "All VMUs", OPTION_VAL_ALL_VMUS_GL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_GL,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_GL,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_GL,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_GL },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_GL },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_GL },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_GL },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_GL },
         { "2x", OPTION_VAL_2_O27_GL },
         { "3x", OPTION_VAL_3X_GL },
         { "4x", OPTION_VAL_4_GL },
         { "5x", OPTION_VAL_5X_GL },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_GL },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_GL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_GL },
         { "BLUE 03",           OPTION_VAL_BLUE_GL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_GL },
         { "GREEN 05",          OPTION_VAL_GREEN_GL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_GL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_GL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_GL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_GL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_GL },
         { "RED 11",            OPTION_VAL_RED_GL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_GL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_GL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_GL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_GL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_GL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_GL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_GL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_GL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_GL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_GL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_GL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_GL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_GL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_GL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_GL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_GL },
         { "WHITE 28",          OPTION_VAL_WHITE_GL },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_GL },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_GL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_GL },
         { "BLUE 03",           OPTION_VAL_BLUE_GL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_GL },
         { "GREEN 05",          OPTION_VAL_GREEN_GL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_GL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_GL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_GL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_GL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_GL },
         { "RED 11",            OPTION_VAL_RED_GL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_GL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_GL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_GL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_GL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_GL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_GL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_GL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_GL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_GL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_GL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_GL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_GL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_GL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_GL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_GL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_GL },
         { "WHITE 28",          OPTION_VAL_WHITE_GL },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_GL },
         { "20%",  OPTION_VAL_20_GL },
         { "30%",  OPTION_VAL_30_GL },
         { "40%",  OPTION_VAL_40_GL },
         { "50%",  OPTION_VAL_50_GL },
         { "60%",  OPTION_VAL_60_GL },
         { "70%",  OPTION_VAL_70_GL },
         { "80%",  OPTION_VAL_80_GL },
         { "90%",  OPTION_VAL_90_GL },
         { "100%", OPTION_VAL_100_GL },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_GL },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_GL },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_GL },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_GL },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_GL },
         { "2x", OPTION_VAL_2_O27_GL },
         { "3x", OPTION_VAL_3X_GL },
         { "4x", OPTION_VAL_4_GL },
         { "5x", OPTION_VAL_5X_GL },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_GL },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_GL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_GL },
         { "BLUE 03",           OPTION_VAL_BLUE_GL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_GL },
         { "GREEN 05",          OPTION_VAL_GREEN_GL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_GL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_GL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_GL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_GL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_GL },
         { "RED 11",            OPTION_VAL_RED_GL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_GL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_GL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_GL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_GL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_GL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_GL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_GL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_GL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_GL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_GL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_GL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_GL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_GL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_GL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_GL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_GL },
         { "WHITE 28",          OPTION_VAL_WHITE_GL },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_GL },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_GL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_GL },
         { "BLUE 03",           OPTION_VAL_BLUE_GL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_GL },
         { "GREEN 05",          OPTION_VAL_GREEN_GL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_GL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_GL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_GL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_GL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_GL },
         { "RED 11",            OPTION_VAL_RED_GL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_GL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_GL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_GL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_GL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_GL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_GL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_GL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_GL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_GL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_GL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_GL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_GL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_GL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_GL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_GL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_GL },
         { "WHITE 28",          OPTION_VAL_WHITE_GL },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_GL },
         { "20%",  OPTION_VAL_20_GL },
         { "30%",  OPTION_VAL_30_GL },
         { "40%",  OPTION_VAL_40_GL },
         { "50%",  OPTION_VAL_50_GL },
         { "60%",  OPTION_VAL_60_GL },
         { "70%",  OPTION_VAL_70_GL },
         { "80%",  OPTION_VAL_80_GL },
         { "90%",  OPTION_VAL_90_GL },
         { "100%", OPTION_VAL_100_GL },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_GL },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_GL },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_GL },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_GL },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_GL },
         { "2x", OPTION_VAL_2_O27_GL },
         { "3x", OPTION_VAL_3X_GL },
         { "4x", OPTION_VAL_4_GL },
         { "5x", OPTION_VAL_5X_GL },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_GL },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_GL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_GL },
         { "BLUE 03",           OPTION_VAL_BLUE_GL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_GL },
         { "GREEN 05",          OPTION_VAL_GREEN_GL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_GL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_GL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_GL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_GL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_GL },
         { "RED 11",            OPTION_VAL_RED_GL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_GL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_GL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_GL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_GL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_GL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_GL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_GL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_GL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_GL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_GL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_GL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_GL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_GL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_GL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_GL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_GL },
         { "WHITE 28",          OPTION_VAL_WHITE_GL },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_GL },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_GL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_GL },
         { "BLUE 03",           OPTION_VAL_BLUE_GL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_GL },
         { "GREEN 05",          OPTION_VAL_GREEN_GL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_GL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_GL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_GL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_GL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_GL },
         { "RED 11",            OPTION_VAL_RED_GL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_GL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_GL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_GL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_GL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_GL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_GL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_GL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_GL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_GL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_GL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_GL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_GL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_GL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_GL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_GL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_GL },
         { "WHITE 28",          OPTION_VAL_WHITE_GL },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_GL },
         { "20%",  OPTION_VAL_20_GL },
         { "30%",  OPTION_VAL_30_GL },
         { "40%",  OPTION_VAL_40_GL },
         { "50%",  OPTION_VAL_50_GL },
         { "60%",  OPTION_VAL_60_GL },
         { "70%",  OPTION_VAL_70_GL },
         { "80%",  OPTION_VAL_80_GL },
         { "90%",  OPTION_VAL_90_GL },
         { "100%", OPTION_VAL_100_GL },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_GL },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_GL },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_GL },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_GL },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_GL },
         { "2x", OPTION_VAL_2_O27_GL },
         { "3x", OPTION_VAL_3X_GL },
         { "4x", OPTION_VAL_4_GL },
         { "5x", OPTION_VAL_5X_GL },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_GL },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_GL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_GL },
         { "BLUE 03",           OPTION_VAL_BLUE_GL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_GL },
         { "GREEN 05",          OPTION_VAL_GREEN_GL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_GL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_GL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_GL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_GL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_GL },
         { "RED 11",            OPTION_VAL_RED_GL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_GL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_GL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_GL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_GL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_GL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_GL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_GL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_GL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_GL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_GL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_GL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_GL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_GL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_GL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_GL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_GL },
         { "WHITE 28",          OPTION_VAL_WHITE_GL },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_GL },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_GL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_GL },
         { "BLUE 03",           OPTION_VAL_BLUE_GL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_GL },
         { "GREEN 05",          OPTION_VAL_GREEN_GL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_GL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_GL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_GL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_GL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_GL },
         { "RED 11",            OPTION_VAL_RED_GL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_GL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_GL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_GL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_GL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_GL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_GL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_GL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_GL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_GL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_GL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_GL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_GL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_GL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_GL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_GL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_GL },
         { "WHITE 28",          OPTION_VAL_WHITE_GL },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_GL },
         { "20%",  OPTION_VAL_20_GL },
         { "30%",  OPTION_VAL_30_GL },
         { "40%",  OPTION_VAL_40_GL },
         { "50%",  OPTION_VAL_50_GL },
         { "60%",  OPTION_VAL_60_GL },
         { "70%",  OPTION_VAL_70_GL },
         { "80%",  OPTION_VAL_80_GL },
         { "90%",  OPTION_VAL_90_GL },
         { "100%", OPTION_VAL_100_GL },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_gl = {
   option_cats_gl,
   option_defs_gl
};

/* RETRO_LANGUAGE_HE */

#define CATEGORY_SYSTEM_LABEL_HE NULL
#define CATEGORY_SYSTEM_INFO_0_HE NULL
#define CATEGORY_VIDEO_LABEL_HE "וידאו"
#define CATEGORY_VIDEO_INFO_0_HE NULL
#define CATEGORY_PERFORMANCE_LABEL_HE NULL
#define CATEGORY_PERFORMANCE_INFO_0_HE NULL
#define CATEGORY_HACKS_LABEL_HE NULL
#define CATEGORY_HACKS_INFO_0_HE NULL
#define CATEGORY_INPUT_LABEL_HE "קלט"
#define CATEGORY_INPUT_INFO_0_HE NULL
#define CATEGORY_VMU_LABEL_HE NULL
#define CATEGORY_VMU_INFO_0_HE NULL
#define CORE_OPTION_NAME_REGION_LABEL_HE NULL
#define OPTION_VAL_JAPAN_HE NULL
#define OPTION_VAL_USA_HE NULL
#define OPTION_VAL_EUROPE_HE NULL
#define OPTION_VAL_DEFAULT_HE NULL
#define CORE_OPTION_NAME_LANGUAGE_LABEL_HE NULL
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_HE NULL
#define OPTION_VAL_JAPANESE_HE NULL
#define OPTION_VAL_ENGLISH_HE NULL
#define OPTION_VAL_GERMAN_HE NULL
#define OPTION_VAL_FRENCH_HE NULL
#define OPTION_VAL_SPANISH_HE NULL
#define OPTION_VAL_ITALIAN_HE NULL
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_HE NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_HE NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_HE NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_HE NULL
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_HE NULL
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_HE NULL
#define CORE_OPTION_NAME_FORCE_WINCE_LABEL_HE NULL
#define CORE_OPTION_NAME_FORCE_WINCE_INFO_0_HE NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_HE NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_HE NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_HE NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_HE NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_HE NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_HE NULL
#define CORE_OPTION_NAME_UPNP_LABEL_HE NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_HE NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_HE NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_HE NULL
#define OPTION_VAL_320X240_HE NULL
#define OPTION_VAL_640X480_HE NULL
#define OPTION_VAL_800X600_HE NULL
#define OPTION_VAL_960X720_HE NULL
#define OPTION_VAL_1024X768_HE NULL
#define OPTION_VAL_1280X960_HE NULL
#define OPTION_VAL_1440X1080_HE NULL
#define OPTION_VAL_1600X1200_HE NULL
#define OPTION_VAL_1920X1440_HE NULL
#define OPTION_VAL_2560X1920_HE NULL
#define OPTION_VAL_2880X2160_HE NULL
#define OPTION_VAL_3200X2400_HE NULL
#define OPTION_VAL_3840X2880_HE NULL
#define OPTION_VAL_4480X3360_HE NULL
#define OPTION_VAL_5120X3840_HE NULL
#define OPTION_VAL_5760X4320_HE NULL
#define OPTION_VAL_6400X4800_HE NULL
#define OPTION_VAL_7040X5280_HE NULL
#define OPTION_VAL_7680X5760_HE NULL
#define OPTION_VAL_8320X6240_HE NULL
#define OPTION_VAL_8960X6720_HE NULL
#define OPTION_VAL_9600X7200_HE NULL
#define OPTION_VAL_10240X7680_HE NULL
#define OPTION_VAL_10880X8160_HE NULL
#define OPTION_VAL_11520X8640_HE NULL
#define OPTION_VAL_12160X9120_HE NULL
#define OPTION_VAL_12800X9600_HE NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_HE NULL
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_HE NULL
#define OPTION_VAL_VGA_HE NULL
#define OPTION_VAL_TV_RGB_HE NULL
#define OPTION_VAL_TV_COMPOSITE_HE NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_HE NULL
#define OPTION_VAL_NTSC_HE NULL
#define OPTION_VAL_PAL_HE NULL
#define OPTION_VAL_PAL_N_HE NULL
#define OPTION_VAL_PAL_M_HE NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_HE NULL
#define OPTION_VAL_HORIZONTAL_HE NULL
#define OPTION_VAL_VERTICAL_HE NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_HE NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_HE NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_HE NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_HE NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_HE NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_HE NULL
#define OPTION_VAL_512MB_HE NULL
#define OPTION_VAL_1GB_HE NULL
#define OPTION_VAL_2GB_HE NULL
#define OPTION_VAL_4GB_HE NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_HE NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_HE NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_HE NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_HE NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_HE NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_HE NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_HE NULL
#define CORE_OPTION_NAME_FOG_LABEL_HE NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_HE NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_HE NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_HE NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_HE NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_HE NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_HE NULL
#define OPTION_VAL_1_HE NULL
#define OPTION_VAL_2_HE NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_HE NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_HE NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_HE NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_HE NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_HE NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_HE NULL
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_HE NULL
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_HE NULL
#define OPTION_VAL_2_O27_HE NULL
#define OPTION_VAL_4_HE NULL
#define OPTION_VAL_6_HE NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_HE NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_HE NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_HE NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_HE NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_HE NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_HE NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_HE NULL
#define OPTION_VAL_SOME_HE NULL
#define OPTION_VAL_MORE_HE NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_HE NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_HE NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_HE NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_HE NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_HE NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_HE NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_HE NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_HE NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_HE NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_HE NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_HE NULL
#define OPTION_VAL_0_HE NULL
#define OPTION_VAL_5_HE NULL
#define OPTION_VAL_10_HE NULL
#define OPTION_VAL_15_HE NULL
#define OPTION_VAL_20_HE NULL
#define OPTION_VAL_25_HE NULL
#define OPTION_VAL_30_HE NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_HE NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_HE NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_HE NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_HE NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_HE NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_HE NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_HE NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_HE NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_HE NULL
#define OPTION_VAL_WHITE_HE NULL
#define OPTION_VAL_RED_HE NULL
#define OPTION_VAL_GREEN_HE NULL
#define OPTION_VAL_BLUE_HE NULL
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_HE NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_HE NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_HE NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_HE NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_HE NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_HE NULL
#define OPTION_VAL_VMU_A1_HE NULL
#define OPTION_VAL_ALL_VMUS_HE NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_HE NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_HE NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_HE NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_HE NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_HE NULL
#define OPTION_VAL_UPPER_LEFT_HE NULL
#define OPTION_VAL_UPPER_RIGHT_HE NULL
#define OPTION_VAL_LOWER_LEFT_HE NULL
#define OPTION_VAL_LOWER_RIGHT_HE NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_HE NULL
#define OPTION_VAL_1X_HE NULL
#define OPTION_VAL_3X_HE NULL
#define OPTION_VAL_5X_HE NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_HE NULL
#define OPTION_VAL_DEFAULT_ON_00_HE NULL
#define OPTION_VAL_DEFAULT_OFF_01_HE NULL
#define OPTION_VAL_BLACK_02_HE NULL
#define OPTION_VAL_LIGHT_BLUE_04_HE NULL
#define OPTION_VAL_CYAN_06_HE NULL
#define OPTION_VAL_CYAN_BLUE_07_HE NULL
#define OPTION_VAL_LIGHT_GREEN_08_HE NULL
#define OPTION_VAL_CYAN_GREEN_09_HE NULL
#define OPTION_VAL_LIGHT_CYAN_10_HE NULL
#define OPTION_VAL_PURPLE_12_HE NULL
#define OPTION_VAL_LIGHT_PURPLE_13_HE NULL
#define OPTION_VAL_YELLOW_14_HE NULL
#define OPTION_VAL_GRAY_15_HE NULL
#define OPTION_VAL_LIGHT_PURPLE_2_16_HE NULL
#define OPTION_VAL_LIGHT_GREEN_2_17_HE NULL
#define OPTION_VAL_LIGHT_GREEN_3_18_HE NULL
#define OPTION_VAL_LIGHT_CYAN_2_19_HE NULL
#define OPTION_VAL_LIGHT_RED_2_20_HE NULL
#define OPTION_VAL_MAGENTA_21_HE NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_HE NULL
#define OPTION_VAL_LIGHT_ORANGE_23_HE NULL
#define OPTION_VAL_ORANGE_24_HE NULL
#define OPTION_VAL_LIGHT_PURPLE_4_25_HE NULL
#define OPTION_VAL_LIGHT_YELLOW_26_HE NULL
#define OPTION_VAL_LIGHT_YELLOW_2_27_HE NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_HE NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_HE NULL
#define OPTION_VAL_40_HE NULL
#define OPTION_VAL_50_HE NULL
#define OPTION_VAL_60_HE NULL
#define OPTION_VAL_70_HE NULL
#define OPTION_VAL_80_HE NULL
#define OPTION_VAL_90_HE NULL
#define OPTION_VAL_100_HE NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_HE NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_HE NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_HE NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_HE NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_HE NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_HE NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_HE NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_HE NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_HE NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_HE NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_HE NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_HE NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_HE NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_HE NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_HE NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_HE NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_HE NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_HE NULL

struct retro_core_option_v2_category option_cats_he[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_HE,
      CATEGORY_SYSTEM_INFO_0_HE
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_HE,
      CATEGORY_VIDEO_INFO_0_HE
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_HE,
      CATEGORY_PERFORMANCE_INFO_0_HE
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_HE,
      CATEGORY_HACKS_INFO_0_HE
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_HE,
      CATEGORY_INPUT_INFO_0_HE
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_HE,
      CATEGORY_VMU_INFO_0_HE
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_he[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_HE,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_HE },
         { "USA",     OPTION_VAL_USA_HE },
         { "Europe",  OPTION_VAL_EUROPE_HE },
         { "Default", OPTION_VAL_DEFAULT_HE },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_HE,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_HE },
         { "English",  OPTION_VAL_ENGLISH_HE },
         { "German",   OPTION_VAL_GERMAN_HE },
         { "French",   OPTION_VAL_FRENCH_HE },
         { "Spanish",  OPTION_VAL_SPANISH_HE },
         { "Italian",  OPTION_VAL_ITALIAN_HE },
         { "Default",  OPTION_VAL_DEFAULT_HE },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_HE,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_boot_to_bios",
      CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_HE,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_HE,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_force_wince",
      CORE_OPTION_NAME_FORCE_WINCE_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_FORCE_WINCE_INFO_0_HE,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_HE,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_HE,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_HE,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_HE,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_HE,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_HE },
         { "640x480",    OPTION_VAL_640X480_HE },
         { "800x600",    OPTION_VAL_800X600_HE },
         { "960x720",    OPTION_VAL_960X720_HE },
         { "1024x768",   OPTION_VAL_1024X768_HE },
         { "1280x960",   OPTION_VAL_1280X960_HE },
         { "1440x1080",  OPTION_VAL_1440X1080_HE },
         { "1600x1200",  OPTION_VAL_1600X1200_HE },
         { "1920x1440",  OPTION_VAL_1920X1440_HE },
         { "2560x1920",  OPTION_VAL_2560X1920_HE },
         { "2880x2160",  OPTION_VAL_2880X2160_HE },
         { "3200x2400",  OPTION_VAL_3200X2400_HE },
         { "3840x2880",  OPTION_VAL_3840X2880_HE },
         { "4480x3360",  OPTION_VAL_4480X3360_HE },
         { "5120x3840",  OPTION_VAL_5120X3840_HE },
         { "5760x4320",  OPTION_VAL_5760X4320_HE },
         { "6400x4800",  OPTION_VAL_6400X4800_HE },
         { "7040x5280",  OPTION_VAL_7040X5280_HE },
         { "7680x5760",  OPTION_VAL_7680X5760_HE },
         { "8320x6240",  OPTION_VAL_8320X6240_HE },
         { "8960x6720",  OPTION_VAL_8960X6720_HE },
         { "9600x7200",  OPTION_VAL_9600X7200_HE },
         { "10240x7680", OPTION_VAL_10240X7680_HE },
         { "10880x8160", OPTION_VAL_10880X8160_HE },
         { "11520x8640", OPTION_VAL_11520X8640_HE },
         { "12160x9120", OPTION_VAL_12160X9120_HE },
         { "12800x9600", OPTION_VAL_12800X9600_HE },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_HE,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_HE },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_HE },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_HE },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_HE,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_HE },
         { "PAL",     OPTION_VAL_PAL_HE },
         { "PAL_N",   OPTION_VAL_PAL_N_HE },
         { "PAL_M",   OPTION_VAL_PAL_M_HE },
         { "Default", OPTION_VAL_DEFAULT_HE },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_HE,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_HE },
         { "vertical",   OPTION_VAL_VERTICAL_HE },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_HE,
      NULL,
      "",
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_HE },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_HE },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_HE },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_HE,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_HE },
         { "1GB",   OPTION_VAL_1GB_HE },
         { "2GB",   OPTION_VAL_2GB_HE },
         { "4GB",   OPTION_VAL_4GB_HE },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_HE,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_HE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_HE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_HE,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_HE,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_HE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_HE,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_HE,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_HE },
         { "1",  OPTION_VAL_1_HE },
         { "2",  OPTION_VAL_2_HE },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_HE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_HE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_HE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_HE,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O27_HE },
         { "4",  OPTION_VAL_4_HE },
         { "6",  OPTION_VAL_6_HE },
         { NULL, NULL },
      },
      "1",
   },
   {/* TODO: needs clarification */
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_HE,
      NULL,
      "",
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_HE,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_HE,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_HE,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_HE,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_HE },
         { "more", OPTION_VAL_MORE_HE },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_HE,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_HE,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_HE,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_HE,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_HE,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: probably needs explanation */
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_HE,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_HE,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_HE },
         { "5%",  OPTION_VAL_5_HE },
         { "10%", OPTION_VAL_10_HE },
         { "15%", OPTION_VAL_15_HE },
         { "20%", OPTION_VAL_20_HE },
         { "25%", OPTION_VAL_25_HE },
         { "30%", OPTION_VAL_30_HE },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_HE,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_HE },
         { "5%",  OPTION_VAL_5_HE },
         { "10%", OPTION_VAL_10_HE },
         { "15%", OPTION_VAL_15_HE },
         { "20%", OPTION_VAL_20_HE },
         { "25%", OPTION_VAL_25_HE },
         { "30%", OPTION_VAL_30_HE },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_HE,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_purupuru",
      CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_HE,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_HE,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_HE,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_HE,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_HE },
         { "Red",      OPTION_VAL_RED_HE },
         { "Green",    OPTION_VAL_GREEN_HE },
         { "Blue",     OPTION_VAL_BLUE_HE },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_HE,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_HE },
         { "Red",      OPTION_VAL_RED_HE },
         { "Green",    OPTION_VAL_GREEN_HE },
         { "Blue",     OPTION_VAL_BLUE_HE },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_HE,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_HE },
         { "Red",      OPTION_VAL_RED_HE },
         { "Green",    OPTION_VAL_GREEN_HE },
         { "Blue",     OPTION_VAL_BLUE_HE },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_HE,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_HE },
         { "Red",      OPTION_VAL_RED_HE },
         { "Green",    OPTION_VAL_GREEN_HE },
         { "Blue",     OPTION_VAL_BLUE_HE },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_HE,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_HE,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_HE,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_HE },
         { "All VMUs", OPTION_VAL_ALL_VMUS_HE },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_HE,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_HE,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_HE,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_HE },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_HE },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_HE },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_HE },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_HE },
         { "2x", OPTION_VAL_2_O27_HE },
         { "3x", OPTION_VAL_3X_HE },
         { "4x", OPTION_VAL_4_HE },
         { "5x", OPTION_VAL_5X_HE },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HE },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HE },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HE },
         { "BLUE 03",           OPTION_VAL_BLUE_HE },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HE },
         { "GREEN 05",          OPTION_VAL_GREEN_HE },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HE },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HE },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HE },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HE },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HE },
         { "RED 11",            OPTION_VAL_RED_HE },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HE },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HE },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HE },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HE },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HE },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HE },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HE },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HE },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HE },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HE },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HE },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HE },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HE },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HE },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HE },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HE },
         { "WHITE 28",          OPTION_VAL_WHITE_HE },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HE },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HE },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HE },
         { "BLUE 03",           OPTION_VAL_BLUE_HE },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HE },
         { "GREEN 05",          OPTION_VAL_GREEN_HE },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HE },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HE },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HE },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HE },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HE },
         { "RED 11",            OPTION_VAL_RED_HE },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HE },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HE },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HE },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HE },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HE },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HE },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HE },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HE },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HE },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HE },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HE },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HE },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HE },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HE },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HE },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HE },
         { "WHITE 28",          OPTION_VAL_WHITE_HE },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_HE },
         { "20%",  OPTION_VAL_20_HE },
         { "30%",  OPTION_VAL_30_HE },
         { "40%",  OPTION_VAL_40_HE },
         { "50%",  OPTION_VAL_50_HE },
         { "60%",  OPTION_VAL_60_HE },
         { "70%",  OPTION_VAL_70_HE },
         { "80%",  OPTION_VAL_80_HE },
         { "90%",  OPTION_VAL_90_HE },
         { "100%", OPTION_VAL_100_HE },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_HE },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_HE },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_HE },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_HE },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_HE },
         { "2x", OPTION_VAL_2_O27_HE },
         { "3x", OPTION_VAL_3X_HE },
         { "4x", OPTION_VAL_4_HE },
         { "5x", OPTION_VAL_5X_HE },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HE },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HE },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HE },
         { "BLUE 03",           OPTION_VAL_BLUE_HE },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HE },
         { "GREEN 05",          OPTION_VAL_GREEN_HE },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HE },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HE },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HE },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HE },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HE },
         { "RED 11",            OPTION_VAL_RED_HE },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HE },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HE },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HE },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HE },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HE },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HE },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HE },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HE },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HE },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HE },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HE },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HE },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HE },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HE },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HE },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HE },
         { "WHITE 28",          OPTION_VAL_WHITE_HE },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HE },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HE },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HE },
         { "BLUE 03",           OPTION_VAL_BLUE_HE },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HE },
         { "GREEN 05",          OPTION_VAL_GREEN_HE },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HE },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HE },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HE },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HE },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HE },
         { "RED 11",            OPTION_VAL_RED_HE },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HE },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HE },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HE },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HE },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HE },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HE },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HE },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HE },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HE },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HE },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HE },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HE },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HE },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HE },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HE },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HE },
         { "WHITE 28",          OPTION_VAL_WHITE_HE },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_HE },
         { "20%",  OPTION_VAL_20_HE },
         { "30%",  OPTION_VAL_30_HE },
         { "40%",  OPTION_VAL_40_HE },
         { "50%",  OPTION_VAL_50_HE },
         { "60%",  OPTION_VAL_60_HE },
         { "70%",  OPTION_VAL_70_HE },
         { "80%",  OPTION_VAL_80_HE },
         { "90%",  OPTION_VAL_90_HE },
         { "100%", OPTION_VAL_100_HE },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_HE },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_HE },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_HE },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_HE },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_HE },
         { "2x", OPTION_VAL_2_O27_HE },
         { "3x", OPTION_VAL_3X_HE },
         { "4x", OPTION_VAL_4_HE },
         { "5x", OPTION_VAL_5X_HE },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HE },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HE },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HE },
         { "BLUE 03",           OPTION_VAL_BLUE_HE },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HE },
         { "GREEN 05",          OPTION_VAL_GREEN_HE },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HE },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HE },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HE },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HE },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HE },
         { "RED 11",            OPTION_VAL_RED_HE },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HE },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HE },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HE },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HE },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HE },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HE },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HE },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HE },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HE },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HE },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HE },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HE },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HE },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HE },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HE },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HE },
         { "WHITE 28",          OPTION_VAL_WHITE_HE },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HE },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HE },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HE },
         { "BLUE 03",           OPTION_VAL_BLUE_HE },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HE },
         { "GREEN 05",          OPTION_VAL_GREEN_HE },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HE },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HE },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HE },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HE },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HE },
         { "RED 11",            OPTION_VAL_RED_HE },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HE },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HE },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HE },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HE },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HE },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HE },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HE },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HE },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HE },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HE },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HE },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HE },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HE },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HE },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HE },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HE },
         { "WHITE 28",          OPTION_VAL_WHITE_HE },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_HE },
         { "20%",  OPTION_VAL_20_HE },
         { "30%",  OPTION_VAL_30_HE },
         { "40%",  OPTION_VAL_40_HE },
         { "50%",  OPTION_VAL_50_HE },
         { "60%",  OPTION_VAL_60_HE },
         { "70%",  OPTION_VAL_70_HE },
         { "80%",  OPTION_VAL_80_HE },
         { "90%",  OPTION_VAL_90_HE },
         { "100%", OPTION_VAL_100_HE },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_HE },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_HE },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_HE },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_HE },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_HE },
         { "2x", OPTION_VAL_2_O27_HE },
         { "3x", OPTION_VAL_3X_HE },
         { "4x", OPTION_VAL_4_HE },
         { "5x", OPTION_VAL_5X_HE },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HE },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HE },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HE },
         { "BLUE 03",           OPTION_VAL_BLUE_HE },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HE },
         { "GREEN 05",          OPTION_VAL_GREEN_HE },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HE },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HE },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HE },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HE },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HE },
         { "RED 11",            OPTION_VAL_RED_HE },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HE },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HE },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HE },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HE },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HE },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HE },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HE },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HE },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HE },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HE },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HE },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HE },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HE },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HE },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HE },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HE },
         { "WHITE 28",          OPTION_VAL_WHITE_HE },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HE },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HE },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HE },
         { "BLUE 03",           OPTION_VAL_BLUE_HE },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HE },
         { "GREEN 05",          OPTION_VAL_GREEN_HE },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HE },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HE },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HE },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HE },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HE },
         { "RED 11",            OPTION_VAL_RED_HE },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HE },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HE },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HE },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HE },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HE },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HE },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HE },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HE },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HE },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HE },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HE },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HE },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HE },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HE },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HE },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HE },
         { "WHITE 28",          OPTION_VAL_WHITE_HE },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_HE },
         { "20%",  OPTION_VAL_20_HE },
         { "30%",  OPTION_VAL_30_HE },
         { "40%",  OPTION_VAL_40_HE },
         { "50%",  OPTION_VAL_50_HE },
         { "60%",  OPTION_VAL_60_HE },
         { "70%",  OPTION_VAL_70_HE },
         { "80%",  OPTION_VAL_80_HE },
         { "90%",  OPTION_VAL_90_HE },
         { "100%", OPTION_VAL_100_HE },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_he = {
   option_cats_he,
   option_defs_he
};

/* RETRO_LANGUAGE_HR */

#define CATEGORY_SYSTEM_LABEL_HR NULL
#define CATEGORY_SYSTEM_INFO_0_HR NULL
#define CATEGORY_VIDEO_LABEL_HR NULL
#define CATEGORY_VIDEO_INFO_0_HR NULL
#define CATEGORY_PERFORMANCE_LABEL_HR NULL
#define CATEGORY_PERFORMANCE_INFO_0_HR NULL
#define CATEGORY_HACKS_LABEL_HR NULL
#define CATEGORY_HACKS_INFO_0_HR NULL
#define CATEGORY_INPUT_LABEL_HR NULL
#define CATEGORY_INPUT_INFO_0_HR NULL
#define CATEGORY_VMU_LABEL_HR NULL
#define CATEGORY_VMU_INFO_0_HR NULL
#define CORE_OPTION_NAME_REGION_LABEL_HR "Regija"
#define OPTION_VAL_JAPAN_HR NULL
#define OPTION_VAL_USA_HR NULL
#define OPTION_VAL_EUROPE_HR NULL
#define OPTION_VAL_DEFAULT_HR NULL
#define CORE_OPTION_NAME_LANGUAGE_LABEL_HR "Jezik"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_HR NULL
#define OPTION_VAL_JAPANESE_HR NULL
#define OPTION_VAL_ENGLISH_HR NULL
#define OPTION_VAL_GERMAN_HR NULL
#define OPTION_VAL_FRENCH_HR NULL
#define OPTION_VAL_SPANISH_HR NULL
#define OPTION_VAL_ITALIAN_HR NULL
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_HR NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_HR NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_HR NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_HR NULL
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_HR NULL
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_HR NULL
#define CORE_OPTION_NAME_FORCE_WINCE_LABEL_HR NULL
#define CORE_OPTION_NAME_FORCE_WINCE_INFO_0_HR NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_HR NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_HR NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_HR NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_HR NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_HR NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_HR NULL
#define CORE_OPTION_NAME_UPNP_LABEL_HR NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_HR NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_HR NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_HR NULL
#define OPTION_VAL_320X240_HR NULL
#define OPTION_VAL_640X480_HR NULL
#define OPTION_VAL_800X600_HR NULL
#define OPTION_VAL_960X720_HR NULL
#define OPTION_VAL_1024X768_HR NULL
#define OPTION_VAL_1280X960_HR NULL
#define OPTION_VAL_1440X1080_HR NULL
#define OPTION_VAL_1600X1200_HR NULL
#define OPTION_VAL_1920X1440_HR NULL
#define OPTION_VAL_2560X1920_HR NULL
#define OPTION_VAL_2880X2160_HR NULL
#define OPTION_VAL_3200X2400_HR NULL
#define OPTION_VAL_3840X2880_HR NULL
#define OPTION_VAL_4480X3360_HR NULL
#define OPTION_VAL_5120X3840_HR NULL
#define OPTION_VAL_5760X4320_HR NULL
#define OPTION_VAL_6400X4800_HR NULL
#define OPTION_VAL_7040X5280_HR NULL
#define OPTION_VAL_7680X5760_HR NULL
#define OPTION_VAL_8320X6240_HR NULL
#define OPTION_VAL_8960X6720_HR NULL
#define OPTION_VAL_9600X7200_HR NULL
#define OPTION_VAL_10240X7680_HR NULL
#define OPTION_VAL_10880X8160_HR NULL
#define OPTION_VAL_11520X8640_HR NULL
#define OPTION_VAL_12160X9120_HR NULL
#define OPTION_VAL_12800X9600_HR NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_HR NULL
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_HR NULL
#define OPTION_VAL_VGA_HR NULL
#define OPTION_VAL_TV_RGB_HR NULL
#define OPTION_VAL_TV_COMPOSITE_HR NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_HR NULL
#define OPTION_VAL_NTSC_HR NULL
#define OPTION_VAL_PAL_HR NULL
#define OPTION_VAL_PAL_N_HR NULL
#define OPTION_VAL_PAL_M_HR NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_HR NULL
#define OPTION_VAL_HORIZONTAL_HR NULL
#define OPTION_VAL_VERTICAL_HR NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_HR NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_HR NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_HR NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_HR NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_HR NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_HR NULL
#define OPTION_VAL_512MB_HR NULL
#define OPTION_VAL_1GB_HR NULL
#define OPTION_VAL_2GB_HR NULL
#define OPTION_VAL_4GB_HR NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_HR NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_HR NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_HR NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_HR NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_HR NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_HR NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_HR NULL
#define CORE_OPTION_NAME_FOG_LABEL_HR NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_HR NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_HR NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_HR NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_HR NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_HR NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_HR NULL
#define OPTION_VAL_1_HR NULL
#define OPTION_VAL_2_HR NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_HR NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_HR NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_HR NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_HR NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_HR NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_HR NULL
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_HR NULL
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_HR NULL
#define OPTION_VAL_2_O27_HR NULL
#define OPTION_VAL_4_HR NULL
#define OPTION_VAL_6_HR NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_HR NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_HR NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_HR NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_HR NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_HR NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_HR NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_HR NULL
#define OPTION_VAL_SOME_HR NULL
#define OPTION_VAL_MORE_HR NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_HR NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_HR NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_HR NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_HR NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_HR NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_HR NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_HR NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_HR NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_HR NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_HR NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_HR NULL
#define OPTION_VAL_0_HR NULL
#define OPTION_VAL_5_HR NULL
#define OPTION_VAL_10_HR NULL
#define OPTION_VAL_15_HR NULL
#define OPTION_VAL_20_HR NULL
#define OPTION_VAL_25_HR NULL
#define OPTION_VAL_30_HR NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_HR NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_HR NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_HR NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_HR NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_HR NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_HR NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_HR NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_HR NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_HR NULL
#define OPTION_VAL_WHITE_HR NULL
#define OPTION_VAL_RED_HR NULL
#define OPTION_VAL_GREEN_HR NULL
#define OPTION_VAL_BLUE_HR NULL
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_HR NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_HR NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_HR NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_HR NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_HR NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_HR NULL
#define OPTION_VAL_VMU_A1_HR NULL
#define OPTION_VAL_ALL_VMUS_HR NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_HR NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_HR NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_HR NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_HR NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_HR NULL
#define OPTION_VAL_UPPER_LEFT_HR NULL
#define OPTION_VAL_UPPER_RIGHT_HR NULL
#define OPTION_VAL_LOWER_LEFT_HR NULL
#define OPTION_VAL_LOWER_RIGHT_HR NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_HR NULL
#define OPTION_VAL_1X_HR NULL
#define OPTION_VAL_3X_HR NULL
#define OPTION_VAL_5X_HR NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_HR NULL
#define OPTION_VAL_DEFAULT_ON_00_HR NULL
#define OPTION_VAL_DEFAULT_OFF_01_HR NULL
#define OPTION_VAL_BLACK_02_HR NULL
#define OPTION_VAL_LIGHT_BLUE_04_HR NULL
#define OPTION_VAL_CYAN_06_HR NULL
#define OPTION_VAL_CYAN_BLUE_07_HR NULL
#define OPTION_VAL_LIGHT_GREEN_08_HR NULL
#define OPTION_VAL_CYAN_GREEN_09_HR NULL
#define OPTION_VAL_LIGHT_CYAN_10_HR NULL
#define OPTION_VAL_PURPLE_12_HR NULL
#define OPTION_VAL_LIGHT_PURPLE_13_HR NULL
#define OPTION_VAL_YELLOW_14_HR NULL
#define OPTION_VAL_GRAY_15_HR NULL
#define OPTION_VAL_LIGHT_PURPLE_2_16_HR NULL
#define OPTION_VAL_LIGHT_GREEN_2_17_HR NULL
#define OPTION_VAL_LIGHT_GREEN_3_18_HR NULL
#define OPTION_VAL_LIGHT_CYAN_2_19_HR NULL
#define OPTION_VAL_LIGHT_RED_2_20_HR NULL
#define OPTION_VAL_MAGENTA_21_HR NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_HR NULL
#define OPTION_VAL_LIGHT_ORANGE_23_HR NULL
#define OPTION_VAL_ORANGE_24_HR NULL
#define OPTION_VAL_LIGHT_PURPLE_4_25_HR NULL
#define OPTION_VAL_LIGHT_YELLOW_26_HR NULL
#define OPTION_VAL_LIGHT_YELLOW_2_27_HR NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_HR NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_HR NULL
#define OPTION_VAL_40_HR NULL
#define OPTION_VAL_50_HR NULL
#define OPTION_VAL_60_HR NULL
#define OPTION_VAL_70_HR NULL
#define OPTION_VAL_80_HR NULL
#define OPTION_VAL_90_HR NULL
#define OPTION_VAL_100_HR NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_HR NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_HR NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_HR NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_HR NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_HR NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_HR NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_HR NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_HR NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_HR NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_HR NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_HR NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_HR NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_HR NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_HR NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_HR NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_HR NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_HR NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_HR NULL

struct retro_core_option_v2_category option_cats_hr[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_HR,
      CATEGORY_SYSTEM_INFO_0_HR
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_HR,
      CATEGORY_VIDEO_INFO_0_HR
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_HR,
      CATEGORY_PERFORMANCE_INFO_0_HR
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_HR,
      CATEGORY_HACKS_INFO_0_HR
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_HR,
      CATEGORY_INPUT_INFO_0_HR
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_HR,
      CATEGORY_VMU_INFO_0_HR
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_hr[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_HR,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_HR },
         { "USA",     OPTION_VAL_USA_HR },
         { "Europe",  OPTION_VAL_EUROPE_HR },
         { "Default", OPTION_VAL_DEFAULT_HR },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_HR,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_HR },
         { "English",  OPTION_VAL_ENGLISH_HR },
         { "German",   OPTION_VAL_GERMAN_HR },
         { "French",   OPTION_VAL_FRENCH_HR },
         { "Spanish",  OPTION_VAL_SPANISH_HR },
         { "Italian",  OPTION_VAL_ITALIAN_HR },
         { "Default",  OPTION_VAL_DEFAULT_HR },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_HR,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_boot_to_bios",
      CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_HR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_HR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_force_wince",
      CORE_OPTION_NAME_FORCE_WINCE_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_FORCE_WINCE_INFO_0_HR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_HR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_HR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_HR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_HR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_HR,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_HR },
         { "640x480",    OPTION_VAL_640X480_HR },
         { "800x600",    OPTION_VAL_800X600_HR },
         { "960x720",    OPTION_VAL_960X720_HR },
         { "1024x768",   OPTION_VAL_1024X768_HR },
         { "1280x960",   OPTION_VAL_1280X960_HR },
         { "1440x1080",  OPTION_VAL_1440X1080_HR },
         { "1600x1200",  OPTION_VAL_1600X1200_HR },
         { "1920x1440",  OPTION_VAL_1920X1440_HR },
         { "2560x1920",  OPTION_VAL_2560X1920_HR },
         { "2880x2160",  OPTION_VAL_2880X2160_HR },
         { "3200x2400",  OPTION_VAL_3200X2400_HR },
         { "3840x2880",  OPTION_VAL_3840X2880_HR },
         { "4480x3360",  OPTION_VAL_4480X3360_HR },
         { "5120x3840",  OPTION_VAL_5120X3840_HR },
         { "5760x4320",  OPTION_VAL_5760X4320_HR },
         { "6400x4800",  OPTION_VAL_6400X4800_HR },
         { "7040x5280",  OPTION_VAL_7040X5280_HR },
         { "7680x5760",  OPTION_VAL_7680X5760_HR },
         { "8320x6240",  OPTION_VAL_8320X6240_HR },
         { "8960x6720",  OPTION_VAL_8960X6720_HR },
         { "9600x7200",  OPTION_VAL_9600X7200_HR },
         { "10240x7680", OPTION_VAL_10240X7680_HR },
         { "10880x8160", OPTION_VAL_10880X8160_HR },
         { "11520x8640", OPTION_VAL_11520X8640_HR },
         { "12160x9120", OPTION_VAL_12160X9120_HR },
         { "12800x9600", OPTION_VAL_12800X9600_HR },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_HR,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_HR },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_HR },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_HR },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_HR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_HR },
         { "PAL",     OPTION_VAL_PAL_HR },
         { "PAL_N",   OPTION_VAL_PAL_N_HR },
         { "PAL_M",   OPTION_VAL_PAL_M_HR },
         { "Default", OPTION_VAL_DEFAULT_HR },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_HR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_HR },
         { "vertical",   OPTION_VAL_VERTICAL_HR },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_HR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_HR },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_HR },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_HR },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_HR,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_HR },
         { "1GB",   OPTION_VAL_1GB_HR },
         { "2GB",   OPTION_VAL_2GB_HR },
         { "4GB",   OPTION_VAL_4GB_HR },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_HR,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_HR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_HR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_HR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_HR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_HR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_HR,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_HR,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_HR },
         { "1",  OPTION_VAL_1_HR },
         { "2",  OPTION_VAL_2_HR },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_HR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_HR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_HR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_HR,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O27_HR },
         { "4",  OPTION_VAL_4_HR },
         { "6",  OPTION_VAL_6_HR },
         { NULL, NULL },
      },
      "1",
   },
   {/* TODO: needs clarification */
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_HR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_HR,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_HR,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_HR,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_HR,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_HR },
         { "more", OPTION_VAL_MORE_HR },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_HR,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_HR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_HR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_HR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_HR,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: probably needs explanation */
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_HR,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_HR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_HR },
         { "5%",  OPTION_VAL_5_HR },
         { "10%", OPTION_VAL_10_HR },
         { "15%", OPTION_VAL_15_HR },
         { "20%", OPTION_VAL_20_HR },
         { "25%", OPTION_VAL_25_HR },
         { "30%", OPTION_VAL_30_HR },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_HR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_HR },
         { "5%",  OPTION_VAL_5_HR },
         { "10%", OPTION_VAL_10_HR },
         { "15%", OPTION_VAL_15_HR },
         { "20%", OPTION_VAL_20_HR },
         { "25%", OPTION_VAL_25_HR },
         { "30%", OPTION_VAL_30_HR },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_HR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_purupuru",
      CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_HR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_HR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_HR,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_HR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_HR },
         { "Red",      OPTION_VAL_RED_HR },
         { "Green",    OPTION_VAL_GREEN_HR },
         { "Blue",     OPTION_VAL_BLUE_HR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_HR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_HR },
         { "Red",      OPTION_VAL_RED_HR },
         { "Green",    OPTION_VAL_GREEN_HR },
         { "Blue",     OPTION_VAL_BLUE_HR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_HR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_HR },
         { "Red",      OPTION_VAL_RED_HR },
         { "Green",    OPTION_VAL_GREEN_HR },
         { "Blue",     OPTION_VAL_BLUE_HR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_HR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_HR },
         { "Red",      OPTION_VAL_RED_HR },
         { "Green",    OPTION_VAL_GREEN_HR },
         { "Blue",     OPTION_VAL_BLUE_HR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_HR,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_HR,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_HR,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_HR },
         { "All VMUs", OPTION_VAL_ALL_VMUS_HR },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_HR,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_HR,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_HR,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_HR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_HR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_HR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_HR },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_HR },
         { "2x", OPTION_VAL_2_O27_HR },
         { "3x", OPTION_VAL_3X_HR },
         { "4x", OPTION_VAL_4_HR },
         { "5x", OPTION_VAL_5X_HR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HR },
         { "BLUE 03",           OPTION_VAL_BLUE_HR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HR },
         { "GREEN 05",          OPTION_VAL_GREEN_HR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HR },
         { "RED 11",            OPTION_VAL_RED_HR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HR },
         { "WHITE 28",          OPTION_VAL_WHITE_HR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HR },
         { "BLUE 03",           OPTION_VAL_BLUE_HR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HR },
         { "GREEN 05",          OPTION_VAL_GREEN_HR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HR },
         { "RED 11",            OPTION_VAL_RED_HR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HR },
         { "WHITE 28",          OPTION_VAL_WHITE_HR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_HR },
         { "20%",  OPTION_VAL_20_HR },
         { "30%",  OPTION_VAL_30_HR },
         { "40%",  OPTION_VAL_40_HR },
         { "50%",  OPTION_VAL_50_HR },
         { "60%",  OPTION_VAL_60_HR },
         { "70%",  OPTION_VAL_70_HR },
         { "80%",  OPTION_VAL_80_HR },
         { "90%",  OPTION_VAL_90_HR },
         { "100%", OPTION_VAL_100_HR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_HR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_HR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_HR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_HR },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_HR },
         { "2x", OPTION_VAL_2_O27_HR },
         { "3x", OPTION_VAL_3X_HR },
         { "4x", OPTION_VAL_4_HR },
         { "5x", OPTION_VAL_5X_HR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HR },
         { "BLUE 03",           OPTION_VAL_BLUE_HR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HR },
         { "GREEN 05",          OPTION_VAL_GREEN_HR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HR },
         { "RED 11",            OPTION_VAL_RED_HR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HR },
         { "WHITE 28",          OPTION_VAL_WHITE_HR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HR },
         { "BLUE 03",           OPTION_VAL_BLUE_HR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HR },
         { "GREEN 05",          OPTION_VAL_GREEN_HR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HR },
         { "RED 11",            OPTION_VAL_RED_HR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HR },
         { "WHITE 28",          OPTION_VAL_WHITE_HR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_HR },
         { "20%",  OPTION_VAL_20_HR },
         { "30%",  OPTION_VAL_30_HR },
         { "40%",  OPTION_VAL_40_HR },
         { "50%",  OPTION_VAL_50_HR },
         { "60%",  OPTION_VAL_60_HR },
         { "70%",  OPTION_VAL_70_HR },
         { "80%",  OPTION_VAL_80_HR },
         { "90%",  OPTION_VAL_90_HR },
         { "100%", OPTION_VAL_100_HR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_HR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_HR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_HR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_HR },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_HR },
         { "2x", OPTION_VAL_2_O27_HR },
         { "3x", OPTION_VAL_3X_HR },
         { "4x", OPTION_VAL_4_HR },
         { "5x", OPTION_VAL_5X_HR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HR },
         { "BLUE 03",           OPTION_VAL_BLUE_HR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HR },
         { "GREEN 05",          OPTION_VAL_GREEN_HR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HR },
         { "RED 11",            OPTION_VAL_RED_HR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HR },
         { "WHITE 28",          OPTION_VAL_WHITE_HR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HR },
         { "BLUE 03",           OPTION_VAL_BLUE_HR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HR },
         { "GREEN 05",          OPTION_VAL_GREEN_HR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HR },
         { "RED 11",            OPTION_VAL_RED_HR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HR },
         { "WHITE 28",          OPTION_VAL_WHITE_HR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_HR },
         { "20%",  OPTION_VAL_20_HR },
         { "30%",  OPTION_VAL_30_HR },
         { "40%",  OPTION_VAL_40_HR },
         { "50%",  OPTION_VAL_50_HR },
         { "60%",  OPTION_VAL_60_HR },
         { "70%",  OPTION_VAL_70_HR },
         { "80%",  OPTION_VAL_80_HR },
         { "90%",  OPTION_VAL_90_HR },
         { "100%", OPTION_VAL_100_HR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_HR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_HR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_HR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_HR },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_HR },
         { "2x", OPTION_VAL_2_O27_HR },
         { "3x", OPTION_VAL_3X_HR },
         { "4x", OPTION_VAL_4_HR },
         { "5x", OPTION_VAL_5X_HR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HR },
         { "BLUE 03",           OPTION_VAL_BLUE_HR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HR },
         { "GREEN 05",          OPTION_VAL_GREEN_HR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HR },
         { "RED 11",            OPTION_VAL_RED_HR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HR },
         { "WHITE 28",          OPTION_VAL_WHITE_HR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HR },
         { "BLUE 03",           OPTION_VAL_BLUE_HR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HR },
         { "GREEN 05",          OPTION_VAL_GREEN_HR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HR },
         { "RED 11",            OPTION_VAL_RED_HR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HR },
         { "WHITE 28",          OPTION_VAL_WHITE_HR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_HR },
         { "20%",  OPTION_VAL_20_HR },
         { "30%",  OPTION_VAL_30_HR },
         { "40%",  OPTION_VAL_40_HR },
         { "50%",  OPTION_VAL_50_HR },
         { "60%",  OPTION_VAL_60_HR },
         { "70%",  OPTION_VAL_70_HR },
         { "80%",  OPTION_VAL_80_HR },
         { "90%",  OPTION_VAL_90_HR },
         { "100%", OPTION_VAL_100_HR },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_hr = {
   option_cats_hr,
   option_defs_hr
};

/* RETRO_LANGUAGE_HU */

#define CATEGORY_SYSTEM_LABEL_HU "Rendszer"
#define CATEGORY_SYSTEM_INFO_0_HU "A régió, nyelv, BIOS, és a hardveralap beállításai."
#define CATEGORY_VIDEO_LABEL_HU "Kép"
#define CATEGORY_VIDEO_INFO_0_HU "A felbontás, sorrendfüggetlen átlátszóság, és vizuális effektek beállításai."
#define CATEGORY_PERFORMANCE_LABEL_HU "Teljesítmény"
#define CATEGORY_PERFORMANCE_INFO_0_HU "A többszálas megjelenítés és képkockakihagyás beállításai."
#define CATEGORY_HACKS_LABEL_HU "Emulációs trükkök"
#define CATEGORY_HACKS_INFO_0_HU "A szélesvásznú módosítások, GD-ROM betöltés sebessége, és a textúracserék beállításai."
#define CATEGORY_INPUT_LABEL_HU "Bemenet"
#define CATEGORY_INPUT_INFO_0_HU "A kontroller és fénypisztoly beállításai."
#define CATEGORY_VMU_LABEL_HU NULL
#define CATEGORY_VMU_INFO_0_HU "A játékonkénti VMU mentések és a képernyőn látható VMU beállításai."
#define CORE_OPTION_NAME_REGION_LABEL_HU "Régió"
#define OPTION_VAL_JAPAN_HU "Japán"
#define OPTION_VAL_USA_HU NULL
#define OPTION_VAL_EUROPE_HU "Európa"
#define OPTION_VAL_DEFAULT_HU "Alapértelmezett"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_HU "Nyelv"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_HU "A nyelv, amelyet a BIOS és a több nyelvet tartalmazó játékok használnak."
#define OPTION_VAL_JAPANESE_HU "japán"
#define OPTION_VAL_ENGLISH_HU "angol"
#define OPTION_VAL_GERMAN_HU "német"
#define OPTION_VAL_FRENCH_HU "francia"
#define OPTION_VAL_SPANISH_HU "spanyol"
#define OPTION_VAL_ITALIAN_HU "olasz"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_HU "HLE BIOS (újraindítás szükséges)"
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_HU "Mindenképp a magas szinten emulált BIOS használata."
#define CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_HU "Indítás BIOS-ban (újraindítás szükséges)"
#define CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_HU "Belépés a Dreamcast BIOS menüjébe rögtön Indításkor."
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_HU "DSP engedélyezése"
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_HU "A Dreamcast DSP (digitális jelfeldolgozó) emulálásának engedélyezése. Javítja a generált hang pontosságát, de megnöveli a teljesítményigényt."
#define CORE_OPTION_NAME_FORCE_WINCE_LABEL_HU "Mindenképp Windows CE mód"
#define CORE_OPTION_NAME_FORCE_WINCE_INFO_0_HU "A teljes MMU (Memory Management Unit) emulálása és egyéb beállítások Windows CE játékokhoz."
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_HU "NAOMI szervizgombok engedélyezése"
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_HU "A NAOMI szervizgombjának engedélyezése, a kabinet beállításaihoz."
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_HU "Ingyen játék a NAOMI játékokhoz"
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_HU "Az érmebeállítások ingyen játékra állítása."
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_HU "Szélessávú adapter emulálása"
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_HU "Modem helyett a szélessávú Ethernet adapter emulálása. (Újraindítás szükséges.)"
#define CORE_OPTION_NAME_UPNP_LABEL_HU "UPnP engedélyezése"
#define CORE_OPTION_NAME_UPNP_INFO_0_HU "Az internetre csatlakozó router automatikus konfigurálása az online játékokhoz UPnP-n keresztül."
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_HU "Belső felbontás"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_HU "A képelőállítás felbontása."
#define OPTION_VAL_320X240_HU NULL
#define OPTION_VAL_640X480_HU NULL
#define OPTION_VAL_800X600_HU NULL
#define OPTION_VAL_960X720_HU NULL
#define OPTION_VAL_1024X768_HU NULL
#define OPTION_VAL_1280X960_HU NULL
#define OPTION_VAL_1440X1080_HU NULL
#define OPTION_VAL_1600X1200_HU NULL
#define OPTION_VAL_1920X1440_HU NULL
#define OPTION_VAL_2560X1920_HU NULL
#define OPTION_VAL_2880X2160_HU NULL
#define OPTION_VAL_3200X2400_HU NULL
#define OPTION_VAL_3840X2880_HU NULL
#define OPTION_VAL_4480X3360_HU NULL
#define OPTION_VAL_5120X3840_HU NULL
#define OPTION_VAL_5760X4320_HU NULL
#define OPTION_VAL_6400X4800_HU NULL
#define OPTION_VAL_7040X5280_HU NULL
#define OPTION_VAL_7680X5760_HU NULL
#define OPTION_VAL_8320X6240_HU NULL
#define OPTION_VAL_8960X6720_HU NULL
#define OPTION_VAL_9600X7200_HU NULL
#define OPTION_VAL_10240X7680_HU NULL
#define OPTION_VAL_10880X8160_HU NULL
#define OPTION_VAL_11520X8640_HU NULL
#define OPTION_VAL_12160X9120_HU NULL
#define OPTION_VAL_12800X9600_HU NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_HU "Kábel típusa"
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_HU "A kimenő jel típusa. A \"TV (kompozit)\" támogatása a legelterjedtebb."
#define OPTION_VAL_VGA_HU NULL
#define OPTION_VAL_TV_RGB_HU NULL
#define OPTION_VAL_TV_COMPOSITE_HU "TV (kompozit)"
#define CORE_OPTION_NAME_BROADCAST_LABEL_HU "Műsorszórási szabvány"
#define OPTION_VAL_NTSC_HU NULL
#define OPTION_VAL_PAL_HU "PAL (Világszerte)"
#define OPTION_VAL_PAL_N_HU NULL
#define OPTION_VAL_PAL_M_HU NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_HU "Képernyő tájolása"
#define OPTION_VAL_HORIZONTAL_HU "Vízszintes"
#define OPTION_VAL_VERTICAL_HU "Függőleges"
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_HU "Alfa rendezés"
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_HU "Per-strip (gyors, legkevésbé pontos)"
#define OPTION_VAL_PER_TRIANGLE_NORMAL_HU "Per-triangle (normál)"
#define OPTION_VAL_PER_PIXEL_ACCURATE_HU "Per-pixel (pontos, de a leglassabb)"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_HU "Pixel gyűjtőbuffer mérete"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_HU "Magasabb felbontások helyes visszaadásához nagyobb értékek kellhetnek."
#define OPTION_VAL_512MB_HU NULL
#define OPTION_VAL_1GB_HU NULL
#define OPTION_VAL_2GB_HU NULL
#define OPTION_VAL_4GB_HU NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_HU "Átlátszó rétegek maximális száma"
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_HU "Összetett jelenetekhez nagyobb értékekre lehet szükség."
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_HU "Teljes framebuffer emuláció"
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_HU "A VRAM framebuffer teljes emulálásának engedélyezése. Olyan játékokhoz hasznos, amelyek közvetlenül a VRAM-ban írják vagy olvassák a framebuffert. Bekapcsolva a belső felbontás mindenképpen 640x480 lesz és a teljesítmény erősen visszaeshet."
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_HU "RTT (Render to texture) puffer engedélyezése"
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_HU "Az előállított textúrák visszamásolása a GPU-ból a VRAM-ba. Ez a lehetőség rendszerint be van kapcsolva azoknál a játékoknál, amelyek igénylik. Bekapcsolt állapotban a textúra előállítás felméretezése kikapcsol, és hatással lehet a teljesítményre."
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_HU NULL
#define CORE_OPTION_NAME_FOG_LABEL_HU "Köd effekt"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_HU "Térfogatmódosító"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_HU "A Dreamcast GPU-jának egy szolgáltatása, amit a játékok rendszerint tárgyak árnyékának megrajzolásához használnak. Normál esetben legyen bekapcsolva, hatása a teljesítményre rendszerint minimális vagy elhanyagolható."
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_HU "Anizotróp szûrés"
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_HU "Minőségjavítás azokon a felületi textúrákon, amik a kamerához képest rézsútosan helyezkednek el. Nagyobb érték jobban terheli a GPU-t. A változtatás csak újraindítás után lép életbe."
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_HU "Textúraszűrés"
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_HU "A textúraszűrés módja. Ezzel lehet egy bizonyos textúraszűrési módot rákényszeríteni minden textúrára, hogy az alapértelmezésnél élesebb (vagy simább) megjelenésük legyen. Az alapértelmezettől eltérő értékek különféle megjelenítési hibákat okozhatnak. A változások csak újraindítás után lépnek életbe."
#define OPTION_VAL_1_HU "Mindenképp Nearest-Neighbor"
#define OPTION_VAL_2_HU "Mindenképp lineáris"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_HU "Képkockacsere késleltetése"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_HU "A villódzó képernyőt vagy hibás mozgóképet segít elkerülni. Lassú platformokon nem ajánlott."
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_HU "Képkockasebesség változásainak érzékelése"
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_HU "A frontend értesítése, amikor a belső képsebesség megváltozik (pl. 60 fps-ről 30 fps-re). Javítja a megjelenítés ütemét olyan játékoknál, amelyek fix 20 vagy 30 fps-sel futnak, de legyen kikapcsolva nem rögzített (instabil) képsebességű játékoknál (pl. Ecco the Dolphin, Unreal Tournament). Figyelem: nem elérhető, ha az automatikus képkocka kihagyás be van kapcsolva."
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_HU "PowerVR2 utófeldolgozó szűrő"
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_HU "Az előállított kép utófeldolgozása a PowerVR2 GPU és az analóg videojelek egyedi effektjeinek szimulálásához."
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_HU "Textúra felméretezés (xBRZ)"
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_HU "A kézzel rajzolt 2D pixel art grafikák feljavítása. Csak a 2D pixelgrafikus játékokhoz ajánlott."
#define OPTION_VAL_2_O27_HU NULL
#define OPTION_VAL_4_HU NULL
#define OPTION_VAL_6_HU NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_HU "Textúra felméretezés maximális szűrt mérete"
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_HU "Natív mélységi interpoláció"
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_HU "Segíthet a textúrahibákon és mélységi problémákon AMD GPU-k, vagy néhány esetben Intel GPU-k használatakor is."
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_HU "Többszálas képelőállítás"
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_HU "A CPU és GPU külön szálon futnak. Erősen ajánlott."
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_HU "Automatikus képkocka kihagyás"
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_HU "Képkockák automatikus kihagyása, ha az emulátor lassan fut. Figyelem: csak akkor érvényes, ha a többszálas képelőállítás engedélyezett."
#define OPTION_VAL_SOME_HU "Normál"
#define OPTION_VAL_MORE_HU "Maximális"
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_HU "Képkocka kihagyás"
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_HU "Két megjelenített képkocka közt ennyi képkocka eldobása."
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_HU "Szélesvásznú csalások (újraindítás szükséges)"
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_HU "Csalások alkalmazása, amelyekkel bizonyos játékok szélesvásznú formátumban jelennek meg."
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_HU "Szélesvásznú trükk"
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_HU "A normál 4:3-as képarányon kívüli geometria felrajzolása. Grafikai hibákat tartalmazhat az így megjelenített területeken."
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_HU "GD-ROM gyorstöltés (pontatlan)"
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_HU "A GD-ROM betöltésének felgyorsítása."
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_HU "Egyedi textúrák betöltése"
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_HU "Textúrák lementése"
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_HU "Analóg kar holtsávja"
#define OPTION_VAL_0_HU NULL
#define OPTION_VAL_5_HU NULL
#define OPTION_VAL_10_HU NULL
#define OPTION_VAL_15_HU NULL
#define OPTION_VAL_20_HU NULL
#define OPTION_VAL_25_HU NULL
#define OPTION_VAL_30_HU NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_HU "Ravasz holtsávja"
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_HU "Digitális ravasz"
#define CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_HU "Purupuru csomag / Vibrációs csomag"
#define CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_HU "A kontroller erővisszacsatolásának engedélyezése."
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_HU "Digitális kimenetek sugárzása"
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_HU "A digitális kimenetek és az erővisszacsatolás állapotának közzététele a 8000-es TCP porton. Kompatibilis a MAME \"-output network\" opciójával."
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_HU "Fénypisztoly beállítások megjelenítése"
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_HU "A fénypisztoly célkereszt megjelenítés beállításai. Figyelem: a változtatás életbelépéséhez újra be kell lépni a gyorsmenübe."
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_HU "1. fénypisztoly célkeresztjének megjelenítése"
#define OPTION_VAL_WHITE_HU "Fehér"
#define OPTION_VAL_RED_HU "Vörös"
#define OPTION_VAL_GREEN_HU "Zöld"
#define OPTION_VAL_BLUE_HU "Kék"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_HU "2. fénypisztoly célkeresztjének megjelenítése"
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_HU "3. fénypisztoly célkeresztjének megjelenítése"
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_HU "4. fénypisztoly célkeresztjének megjelenítése"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_HU "Játékonként külön Visual Memory Unit/System (VMU)"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_HU "Játékonként külön VMU"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_HU "Kikapcsolva minden játék ugyanazt a 4 VMU mentés fájlt (A1, B1, C1, D1) használja a RetroArch rendszerkönyvtárban. \"VMU A1\" beállításnál minden elindított játéknak külön VMU \"A1\" fájlt hoz létre. \"Minden VMU\" beállításnál minden elindított játéknak négy külön VMU fájlt (A1, B1, C1, D1) hoz létre."
#define OPTION_VAL_VMU_A1_HU NULL
#define OPTION_VAL_ALL_VMUS_HU "Minden VMU"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_HU "Visual Memory Unit/System (VMU) megjelenítési beállítások jelenjenek meg"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_HU "VMU megjelenítési beállítások jelenjenek meg"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_HU "Az emulált VMU LCD képernyő láthatósági, méret, helyzet, és szín beállításainak engedélyezése. Figyelem: a gyorsmenüt ki-be kell kapcsolni, hogy ez a beállítás érvényre jusson."
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_HU "VMU 1 képernyő megjelenítése"
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_HU "VMU 1 képernyő helyzete"
#define OPTION_VAL_UPPER_LEFT_HU "Bal felső"
#define OPTION_VAL_UPPER_RIGHT_HU "Jobb felső"
#define OPTION_VAL_LOWER_LEFT_HU "Bal alsó"
#define OPTION_VAL_LOWER_RIGHT_HU "Jobb alsó"
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_HU "VMU 1 képernyő mérete"
#define OPTION_VAL_1X_HU NULL
#define OPTION_VAL_3X_HU NULL
#define OPTION_VAL_5X_HU NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_HU "VMU 1 képernyő bekapcsolt pixel színe"
#define OPTION_VAL_DEFAULT_ON_00_HU "Alapértelmezésben bekapcsolva"
#define OPTION_VAL_DEFAULT_OFF_01_HU "Alapértelmezésben kikapcsolva"
#define OPTION_VAL_BLACK_02_HU "Fekete"
#define OPTION_VAL_LIGHT_BLUE_04_HU "Világoskék"
#define OPTION_VAL_CYAN_06_HU "Ciánkék"
#define OPTION_VAL_CYAN_BLUE_07_HU "Ciánkék"
#define OPTION_VAL_LIGHT_GREEN_08_HU "Világoszöld"
#define OPTION_VAL_CYAN_GREEN_09_HU "Ciánzöld"
#define OPTION_VAL_LIGHT_CYAN_10_HU "Világos ciánkék"
#define OPTION_VAL_PURPLE_12_HU "Lila"
#define OPTION_VAL_LIGHT_PURPLE_13_HU "Világoslila"
#define OPTION_VAL_YELLOW_14_HU "Sárga"
#define OPTION_VAL_GRAY_15_HU "Szürke"
#define OPTION_VAL_LIGHT_PURPLE_2_16_HU "Világoslila (2)"
#define OPTION_VAL_LIGHT_GREEN_2_17_HU "Világoszöld (2)"
#define OPTION_VAL_LIGHT_GREEN_3_18_HU "Világoszöld (3)"
#define OPTION_VAL_LIGHT_CYAN_2_19_HU "Világos ciánkék (2)"
#define OPTION_VAL_LIGHT_RED_2_20_HU "Világospiros (2)"
#define OPTION_VAL_MAGENTA_21_HU "Bíbor"
#define OPTION_VAL_LIGHT_PURPLE_3_22_HU "Világoslila (3)"
#define OPTION_VAL_LIGHT_ORANGE_23_HU "Világos narancs"
#define OPTION_VAL_ORANGE_24_HU "Narancssárga"
#define OPTION_VAL_LIGHT_PURPLE_4_25_HU "Világoslila (4)"
#define OPTION_VAL_LIGHT_YELLOW_26_HU "Halványsárga"
#define OPTION_VAL_LIGHT_YELLOW_2_27_HU "Halványsárga (2)"
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_HU "VMU 1 képernyő kikapcsolt pixel színe"
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_HU "VMU 1 képernyő átlátszatlansága"
#define OPTION_VAL_40_HU NULL
#define OPTION_VAL_50_HU NULL
#define OPTION_VAL_60_HU NULL
#define OPTION_VAL_70_HU NULL
#define OPTION_VAL_80_HU NULL
#define OPTION_VAL_90_HU NULL
#define OPTION_VAL_100_HU NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_HU "VMU 2 képernyő megjelenítése"
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_HU "VMU 2 képernyő helyzete"
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_HU "VMU 2 képernyő mérete"
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_HU "VMU 2 képernyő bekapcsolt pixel színe"
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_HU "VMU 2 képernyő kikapcsolt pixel színe"
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_HU "VMU 2 képernyő átlátszatlansága"
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_HU "VMU 3 képernyő megjelenítése"
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_HU "VMU 3 képernyő helyzete"
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_HU "VMU 3 képernyő mérete"
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_HU "VMU 3 képernyő bekapcsolt pixel színe"
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_HU "VMU 3 képernyő kikapcsolt pixel színe"
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_HU "VMU 3 képernyő átlátszatlansága"
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_HU "VMU 4 képernyő megjelenítése"
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_HU "VMU 4 képernyő helyzete"
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_HU "VMU 4 képernyő mérete"
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_HU "VMU 4 képernyő bekapcsolt pixel színe"
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_HU "VMU 4 képernyő kikapcsolt pixel színe"
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_HU "VMU 4 képernyő átlátszatlansága"

struct retro_core_option_v2_category option_cats_hu[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_HU,
      CATEGORY_SYSTEM_INFO_0_HU
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_HU,
      CATEGORY_VIDEO_INFO_0_HU
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_HU,
      CATEGORY_PERFORMANCE_INFO_0_HU
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_HU,
      CATEGORY_HACKS_INFO_0_HU
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_HU,
      CATEGORY_INPUT_INFO_0_HU
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_HU,
      CATEGORY_VMU_INFO_0_HU
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_hu[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_HU,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_HU },
         { "USA",     OPTION_VAL_USA_HU },
         { "Europe",  OPTION_VAL_EUROPE_HU },
         { "Default", OPTION_VAL_DEFAULT_HU },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_HU,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_HU },
         { "English",  OPTION_VAL_ENGLISH_HU },
         { "German",   OPTION_VAL_GERMAN_HU },
         { "French",   OPTION_VAL_FRENCH_HU },
         { "Spanish",  OPTION_VAL_SPANISH_HU },
         { "Italian",  OPTION_VAL_ITALIAN_HU },
         { "Default",  OPTION_VAL_DEFAULT_HU },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_HU,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_boot_to_bios",
      CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_HU,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_HU,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_force_wince",
      CORE_OPTION_NAME_FORCE_WINCE_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_FORCE_WINCE_INFO_0_HU,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_HU,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_HU,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_HU,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_HU,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_HU,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_HU },
         { "640x480",    OPTION_VAL_640X480_HU },
         { "800x600",    OPTION_VAL_800X600_HU },
         { "960x720",    OPTION_VAL_960X720_HU },
         { "1024x768",   OPTION_VAL_1024X768_HU },
         { "1280x960",   OPTION_VAL_1280X960_HU },
         { "1440x1080",  OPTION_VAL_1440X1080_HU },
         { "1600x1200",  OPTION_VAL_1600X1200_HU },
         { "1920x1440",  OPTION_VAL_1920X1440_HU },
         { "2560x1920",  OPTION_VAL_2560X1920_HU },
         { "2880x2160",  OPTION_VAL_2880X2160_HU },
         { "3200x2400",  OPTION_VAL_3200X2400_HU },
         { "3840x2880",  OPTION_VAL_3840X2880_HU },
         { "4480x3360",  OPTION_VAL_4480X3360_HU },
         { "5120x3840",  OPTION_VAL_5120X3840_HU },
         { "5760x4320",  OPTION_VAL_5760X4320_HU },
         { "6400x4800",  OPTION_VAL_6400X4800_HU },
         { "7040x5280",  OPTION_VAL_7040X5280_HU },
         { "7680x5760",  OPTION_VAL_7680X5760_HU },
         { "8320x6240",  OPTION_VAL_8320X6240_HU },
         { "8960x6720",  OPTION_VAL_8960X6720_HU },
         { "9600x7200",  OPTION_VAL_9600X7200_HU },
         { "10240x7680", OPTION_VAL_10240X7680_HU },
         { "10880x8160", OPTION_VAL_10880X8160_HU },
         { "11520x8640", OPTION_VAL_11520X8640_HU },
         { "12160x9120", OPTION_VAL_12160X9120_HU },
         { "12800x9600", OPTION_VAL_12800X9600_HU },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_HU,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_HU },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_HU },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_HU },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_HU,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_HU },
         { "PAL",     OPTION_VAL_PAL_HU },
         { "PAL_N",   OPTION_VAL_PAL_N_HU },
         { "PAL_M",   OPTION_VAL_PAL_M_HU },
         { "Default", OPTION_VAL_DEFAULT_HU },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_HU,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_HU },
         { "vertical",   OPTION_VAL_VERTICAL_HU },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_HU,
      NULL,
      "",
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_HU },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_HU },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_HU },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_HU,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_HU },
         { "1GB",   OPTION_VAL_1GB_HU },
         { "2GB",   OPTION_VAL_2GB_HU },
         { "4GB",   OPTION_VAL_4GB_HU },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_HU,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_HU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_HU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_HU,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_HU,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_HU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_HU,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_HU,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_HU },
         { "1",  OPTION_VAL_1_HU },
         { "2",  OPTION_VAL_2_HU },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_HU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_HU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_HU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_HU,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O27_HU },
         { "4",  OPTION_VAL_4_HU },
         { "6",  OPTION_VAL_6_HU },
         { NULL, NULL },
      },
      "1",
   },
   {/* TODO: needs clarification */
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_HU,
      NULL,
      "",
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_HU,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_HU,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_HU,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_HU,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_HU },
         { "more", OPTION_VAL_MORE_HU },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_HU,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_HU,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_HU,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_HU,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_HU,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: probably needs explanation */
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_HU,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_HU,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_HU },
         { "5%",  OPTION_VAL_5_HU },
         { "10%", OPTION_VAL_10_HU },
         { "15%", OPTION_VAL_15_HU },
         { "20%", OPTION_VAL_20_HU },
         { "25%", OPTION_VAL_25_HU },
         { "30%", OPTION_VAL_30_HU },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_HU,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_HU },
         { "5%",  OPTION_VAL_5_HU },
         { "10%", OPTION_VAL_10_HU },
         { "15%", OPTION_VAL_15_HU },
         { "20%", OPTION_VAL_20_HU },
         { "25%", OPTION_VAL_25_HU },
         { "30%", OPTION_VAL_30_HU },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_HU,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_purupuru",
      CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_HU,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_HU,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_HU,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_HU,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_HU },
         { "Red",      OPTION_VAL_RED_HU },
         { "Green",    OPTION_VAL_GREEN_HU },
         { "Blue",     OPTION_VAL_BLUE_HU },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_HU,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_HU },
         { "Red",      OPTION_VAL_RED_HU },
         { "Green",    OPTION_VAL_GREEN_HU },
         { "Blue",     OPTION_VAL_BLUE_HU },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_HU,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_HU },
         { "Red",      OPTION_VAL_RED_HU },
         { "Green",    OPTION_VAL_GREEN_HU },
         { "Blue",     OPTION_VAL_BLUE_HU },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_HU,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_HU },
         { "Red",      OPTION_VAL_RED_HU },
         { "Green",    OPTION_VAL_GREEN_HU },
         { "Blue",     OPTION_VAL_BLUE_HU },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_HU,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_HU,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_HU,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_HU },
         { "All VMUs", OPTION_VAL_ALL_VMUS_HU },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_HU,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_HU,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_HU,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_HU },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_HU },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_HU },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_HU },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_HU },
         { "2x", OPTION_VAL_2_O27_HU },
         { "3x", OPTION_VAL_3X_HU },
         { "4x", OPTION_VAL_4_HU },
         { "5x", OPTION_VAL_5X_HU },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HU },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HU },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HU },
         { "BLUE 03",           OPTION_VAL_BLUE_HU },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HU },
         { "GREEN 05",          OPTION_VAL_GREEN_HU },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HU },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HU },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HU },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HU },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HU },
         { "RED 11",            OPTION_VAL_RED_HU },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HU },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HU },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HU },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HU },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HU },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HU },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HU },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HU },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HU },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HU },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HU },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HU },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HU },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HU },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HU },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HU },
         { "WHITE 28",          OPTION_VAL_WHITE_HU },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HU },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HU },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HU },
         { "BLUE 03",           OPTION_VAL_BLUE_HU },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HU },
         { "GREEN 05",          OPTION_VAL_GREEN_HU },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HU },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HU },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HU },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HU },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HU },
         { "RED 11",            OPTION_VAL_RED_HU },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HU },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HU },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HU },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HU },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HU },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HU },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HU },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HU },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HU },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HU },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HU },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HU },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HU },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HU },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HU },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HU },
         { "WHITE 28",          OPTION_VAL_WHITE_HU },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_HU },
         { "20%",  OPTION_VAL_20_HU },
         { "30%",  OPTION_VAL_30_HU },
         { "40%",  OPTION_VAL_40_HU },
         { "50%",  OPTION_VAL_50_HU },
         { "60%",  OPTION_VAL_60_HU },
         { "70%",  OPTION_VAL_70_HU },
         { "80%",  OPTION_VAL_80_HU },
         { "90%",  OPTION_VAL_90_HU },
         { "100%", OPTION_VAL_100_HU },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_HU },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_HU },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_HU },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_HU },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_HU },
         { "2x", OPTION_VAL_2_O27_HU },
         { "3x", OPTION_VAL_3X_HU },
         { "4x", OPTION_VAL_4_HU },
         { "5x", OPTION_VAL_5X_HU },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HU },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HU },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HU },
         { "BLUE 03",           OPTION_VAL_BLUE_HU },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HU },
         { "GREEN 05",          OPTION_VAL_GREEN_HU },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HU },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HU },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HU },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HU },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HU },
         { "RED 11",            OPTION_VAL_RED_HU },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HU },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HU },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HU },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HU },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HU },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HU },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HU },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HU },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HU },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HU },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HU },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HU },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HU },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HU },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HU },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HU },
         { "WHITE 28",          OPTION_VAL_WHITE_HU },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HU },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HU },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HU },
         { "BLUE 03",           OPTION_VAL_BLUE_HU },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HU },
         { "GREEN 05",          OPTION_VAL_GREEN_HU },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HU },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HU },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HU },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HU },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HU },
         { "RED 11",            OPTION_VAL_RED_HU },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HU },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HU },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HU },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HU },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HU },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HU },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HU },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HU },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HU },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HU },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HU },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HU },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HU },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HU },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HU },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HU },
         { "WHITE 28",          OPTION_VAL_WHITE_HU },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_HU },
         { "20%",  OPTION_VAL_20_HU },
         { "30%",  OPTION_VAL_30_HU },
         { "40%",  OPTION_VAL_40_HU },
         { "50%",  OPTION_VAL_50_HU },
         { "60%",  OPTION_VAL_60_HU },
         { "70%",  OPTION_VAL_70_HU },
         { "80%",  OPTION_VAL_80_HU },
         { "90%",  OPTION_VAL_90_HU },
         { "100%", OPTION_VAL_100_HU },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_HU },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_HU },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_HU },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_HU },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_HU },
         { "2x", OPTION_VAL_2_O27_HU },
         { "3x", OPTION_VAL_3X_HU },
         { "4x", OPTION_VAL_4_HU },
         { "5x", OPTION_VAL_5X_HU },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HU },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HU },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HU },
         { "BLUE 03",           OPTION_VAL_BLUE_HU },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HU },
         { "GREEN 05",          OPTION_VAL_GREEN_HU },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HU },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HU },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HU },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HU },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HU },
         { "RED 11",            OPTION_VAL_RED_HU },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HU },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HU },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HU },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HU },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HU },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HU },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HU },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HU },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HU },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HU },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HU },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HU },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HU },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HU },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HU },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HU },
         { "WHITE 28",          OPTION_VAL_WHITE_HU },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HU },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HU },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HU },
         { "BLUE 03",           OPTION_VAL_BLUE_HU },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HU },
         { "GREEN 05",          OPTION_VAL_GREEN_HU },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HU },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HU },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HU },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HU },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HU },
         { "RED 11",            OPTION_VAL_RED_HU },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HU },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HU },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HU },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HU },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HU },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HU },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HU },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HU },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HU },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HU },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HU },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HU },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HU },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HU },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HU },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HU },
         { "WHITE 28",          OPTION_VAL_WHITE_HU },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_HU },
         { "20%",  OPTION_VAL_20_HU },
         { "30%",  OPTION_VAL_30_HU },
         { "40%",  OPTION_VAL_40_HU },
         { "50%",  OPTION_VAL_50_HU },
         { "60%",  OPTION_VAL_60_HU },
         { "70%",  OPTION_VAL_70_HU },
         { "80%",  OPTION_VAL_80_HU },
         { "90%",  OPTION_VAL_90_HU },
         { "100%", OPTION_VAL_100_HU },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_HU },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_HU },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_HU },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_HU },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_HU },
         { "2x", OPTION_VAL_2_O27_HU },
         { "3x", OPTION_VAL_3X_HU },
         { "4x", OPTION_VAL_4_HU },
         { "5x", OPTION_VAL_5X_HU },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HU },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HU },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HU },
         { "BLUE 03",           OPTION_VAL_BLUE_HU },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HU },
         { "GREEN 05",          OPTION_VAL_GREEN_HU },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HU },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HU },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HU },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HU },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HU },
         { "RED 11",            OPTION_VAL_RED_HU },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HU },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HU },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HU },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HU },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HU },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HU },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HU },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HU },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HU },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HU },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HU },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HU },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HU },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HU },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HU },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HU },
         { "WHITE 28",          OPTION_VAL_WHITE_HU },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HU },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HU },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HU },
         { "BLUE 03",           OPTION_VAL_BLUE_HU },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HU },
         { "GREEN 05",          OPTION_VAL_GREEN_HU },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HU },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HU },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HU },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HU },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HU },
         { "RED 11",            OPTION_VAL_RED_HU },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HU },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HU },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HU },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HU },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HU },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HU },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HU },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HU },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HU },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HU },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HU },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HU },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HU },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HU },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HU },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HU },
         { "WHITE 28",          OPTION_VAL_WHITE_HU },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_HU },
         { "20%",  OPTION_VAL_20_HU },
         { "30%",  OPTION_VAL_30_HU },
         { "40%",  OPTION_VAL_40_HU },
         { "50%",  OPTION_VAL_50_HU },
         { "60%",  OPTION_VAL_60_HU },
         { "70%",  OPTION_VAL_70_HU },
         { "80%",  OPTION_VAL_80_HU },
         { "90%",  OPTION_VAL_90_HU },
         { "100%", OPTION_VAL_100_HU },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_hu = {
   option_cats_hu,
   option_defs_hu
};

/* RETRO_LANGUAGE_ID */

#define CATEGORY_SYSTEM_LABEL_ID "Sistem"
#define CATEGORY_SYSTEM_INFO_0_ID NULL
#define CATEGORY_VIDEO_LABEL_ID NULL
#define CATEGORY_VIDEO_INFO_0_ID NULL
#define CATEGORY_PERFORMANCE_LABEL_ID "Kinerja"
#define CATEGORY_PERFORMANCE_INFO_0_ID NULL
#define CATEGORY_HACKS_LABEL_ID "Peretasan Emulasi"
#define CATEGORY_HACKS_INFO_0_ID NULL
#define CATEGORY_INPUT_LABEL_ID "Masukan"
#define CATEGORY_INPUT_INFO_0_ID NULL
#define CATEGORY_VMU_LABEL_ID NULL
#define CATEGORY_VMU_INFO_0_ID NULL
#define CORE_OPTION_NAME_REGION_LABEL_ID "Daerah"
#define OPTION_VAL_JAPAN_ID NULL
#define OPTION_VAL_USA_ID NULL
#define OPTION_VAL_EUROPE_ID NULL
#define OPTION_VAL_DEFAULT_ID NULL
#define CORE_OPTION_NAME_LANGUAGE_LABEL_ID "Bahasa"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_ID NULL
#define OPTION_VAL_JAPANESE_ID NULL
#define OPTION_VAL_ENGLISH_ID NULL
#define OPTION_VAL_GERMAN_ID NULL
#define OPTION_VAL_FRENCH_ID NULL
#define OPTION_VAL_SPANISH_ID NULL
#define OPTION_VAL_ITALIAN_ID NULL
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_ID NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_ID "Paksa penggunaan BIOS emulasi tingkat tinggi."
#define CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_ID "Boot ke BIOS (Perlu Dimulai Ulang)"
#define CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_ID "Boot langsung ke menu BIOS Dreamcast."
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_ID "Aktifkan DSP"
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_ID "Aktifkan emulasi DSP audio Dreamcast (pengolah sinyal digital). Perbaiki akurasi suara yang dihasilakan, tetapi tingkatkan persyaratan kinerja."
#define CORE_OPTION_NAME_FORCE_WINCE_LABEL_ID NULL
#define CORE_OPTION_NAME_FORCE_WINCE_INFO_0_ID NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_ID NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_ID NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_ID NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_ID NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_ID NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_ID NULL
#define CORE_OPTION_NAME_UPNP_LABEL_ID NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_ID NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_ID NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_ID "Modifikasi resolusi pelukisan."
#define OPTION_VAL_320X240_ID NULL
#define OPTION_VAL_640X480_ID NULL
#define OPTION_VAL_800X600_ID NULL
#define OPTION_VAL_960X720_ID NULL
#define OPTION_VAL_1024X768_ID NULL
#define OPTION_VAL_1280X960_ID NULL
#define OPTION_VAL_1440X1080_ID NULL
#define OPTION_VAL_1600X1200_ID NULL
#define OPTION_VAL_1920X1440_ID NULL
#define OPTION_VAL_2560X1920_ID NULL
#define OPTION_VAL_2880X2160_ID NULL
#define OPTION_VAL_3200X2400_ID NULL
#define OPTION_VAL_3840X2880_ID NULL
#define OPTION_VAL_4480X3360_ID NULL
#define OPTION_VAL_5120X3840_ID NULL
#define OPTION_VAL_5760X4320_ID NULL
#define OPTION_VAL_6400X4800_ID NULL
#define OPTION_VAL_7040X5280_ID NULL
#define OPTION_VAL_7680X5760_ID NULL
#define OPTION_VAL_8320X6240_ID NULL
#define OPTION_VAL_8960X6720_ID NULL
#define OPTION_VAL_9600X7200_ID NULL
#define OPTION_VAL_10240X7680_ID NULL
#define OPTION_VAL_10880X8160_ID NULL
#define OPTION_VAL_11520X8640_ID NULL
#define OPTION_VAL_12160X9120_ID NULL
#define OPTION_VAL_12800X9600_ID NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_ID NULL
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_ID NULL
#define OPTION_VAL_VGA_ID NULL
#define OPTION_VAL_TV_RGB_ID NULL
#define OPTION_VAL_TV_COMPOSITE_ID NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_ID NULL
#define OPTION_VAL_NTSC_ID NULL
#define OPTION_VAL_PAL_ID NULL
#define OPTION_VAL_PAL_N_ID NULL
#define OPTION_VAL_PAL_M_ID NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_ID "Orientasi Layar"
#define OPTION_VAL_HORIZONTAL_ID NULL
#define OPTION_VAL_VERTICAL_ID NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_ID NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_ID NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_ID NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_ID NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_ID NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_ID "Mungkin perlu nilai yang lebih besar agar resolusi yang lebih besar terlihat benar."
#define OPTION_VAL_512MB_ID NULL
#define OPTION_VAL_1GB_ID NULL
#define OPTION_VAL_2GB_ID NULL
#define OPTION_VAL_4GB_ID NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_ID NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_ID NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_ID NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_ID NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_ID "Mengaktifkan Penyangga RTT (Lukis Ke Tekstur)"
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_ID NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_ID NULL
#define CORE_OPTION_NAME_FOG_LABEL_ID NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_ID NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_ID NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_ID "Pemfilteran Anisotropik"
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_ID NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_ID NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_ID NULL
#define OPTION_VAL_1_ID NULL
#define OPTION_VAL_2_ID NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_ID NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_ID NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_ID NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_ID NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_ID "Filter Pasca-pengolahan PowerVR2"
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_ID "Pasca-olah gambar yang dilukiskan untuk menyimulasikan efek dengan spesifik ke GPU PowerVR2 dan sinyal video analog."
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_ID "Penskalaan Naik Tekstur (xBRZ)"
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_ID "Tingkatkan grafis seni piksel 2D tulisan tangan. Hanya bisa digunakan di pemainan terpikselisasi 2D."
#define OPTION_VAL_2_O27_ID NULL
#define OPTION_VAL_4_ID NULL
#define OPTION_VAL_6_ID NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_ID "Ukuran Terfilter Maks. Penskalaan Naik"
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_ID NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_ID NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_ID NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_ID "Jalankan GPU dan CPU pada Thread berbeda. Sangat direkomendasikan."
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_ID NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_ID NULL
#define OPTION_VAL_SOME_ID NULL
#define OPTION_VAL_MORE_ID NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_ID NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_ID NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_ID NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_ID NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_ID NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_ID NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_ID NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_ID NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_ID NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_ID NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_ID NULL
#define OPTION_VAL_0_ID NULL
#define OPTION_VAL_5_ID NULL
#define OPTION_VAL_10_ID NULL
#define OPTION_VAL_15_ID NULL
#define OPTION_VAL_20_ID NULL
#define OPTION_VAL_25_ID NULL
#define OPTION_VAL_30_ID NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_ID "Pemicu Deadzone"
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_ID "Pemicu Digital"
#define CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_ID NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_ID NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_ID NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_ID NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_ID NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_ID NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_ID NULL
#define OPTION_VAL_WHITE_ID NULL
#define OPTION_VAL_RED_ID NULL
#define OPTION_VAL_GREEN_ID NULL
#define OPTION_VAL_BLUE_ID NULL
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_ID NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_ID NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_ID NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_ID NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_ID NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_ID NULL
#define OPTION_VAL_VMU_A1_ID NULL
#define OPTION_VAL_ALL_VMUS_ID NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_ID NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_ID NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_ID NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_ID NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_ID NULL
#define OPTION_VAL_UPPER_LEFT_ID NULL
#define OPTION_VAL_UPPER_RIGHT_ID NULL
#define OPTION_VAL_LOWER_LEFT_ID NULL
#define OPTION_VAL_LOWER_RIGHT_ID NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_ID NULL
#define OPTION_VAL_1X_ID NULL
#define OPTION_VAL_3X_ID NULL
#define OPTION_VAL_5X_ID NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_ID NULL
#define OPTION_VAL_DEFAULT_ON_00_ID NULL
#define OPTION_VAL_DEFAULT_OFF_01_ID NULL
#define OPTION_VAL_BLACK_02_ID NULL
#define OPTION_VAL_LIGHT_BLUE_04_ID NULL
#define OPTION_VAL_CYAN_06_ID NULL
#define OPTION_VAL_CYAN_BLUE_07_ID NULL
#define OPTION_VAL_LIGHT_GREEN_08_ID NULL
#define OPTION_VAL_CYAN_GREEN_09_ID NULL
#define OPTION_VAL_LIGHT_CYAN_10_ID NULL
#define OPTION_VAL_PURPLE_12_ID NULL
#define OPTION_VAL_LIGHT_PURPLE_13_ID NULL
#define OPTION_VAL_YELLOW_14_ID NULL
#define OPTION_VAL_GRAY_15_ID NULL
#define OPTION_VAL_LIGHT_PURPLE_2_16_ID NULL
#define OPTION_VAL_LIGHT_GREEN_2_17_ID NULL
#define OPTION_VAL_LIGHT_GREEN_3_18_ID NULL
#define OPTION_VAL_LIGHT_CYAN_2_19_ID NULL
#define OPTION_VAL_LIGHT_RED_2_20_ID NULL
#define OPTION_VAL_MAGENTA_21_ID NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_ID NULL
#define OPTION_VAL_LIGHT_ORANGE_23_ID NULL
#define OPTION_VAL_ORANGE_24_ID NULL
#define OPTION_VAL_LIGHT_PURPLE_4_25_ID NULL
#define OPTION_VAL_LIGHT_YELLOW_26_ID NULL
#define OPTION_VAL_LIGHT_YELLOW_2_27_ID NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_ID NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_ID NULL
#define OPTION_VAL_40_ID NULL
#define OPTION_VAL_50_ID NULL
#define OPTION_VAL_60_ID NULL
#define OPTION_VAL_70_ID NULL
#define OPTION_VAL_80_ID NULL
#define OPTION_VAL_90_ID NULL
#define OPTION_VAL_100_ID NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_ID NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_ID NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_ID NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_ID NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_ID NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_ID NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_ID NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_ID NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_ID NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_ID NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_ID NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_ID NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_ID NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_ID NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_ID NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_ID NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_ID NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_ID NULL

struct retro_core_option_v2_category option_cats_id[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_ID,
      CATEGORY_SYSTEM_INFO_0_ID
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_ID,
      CATEGORY_VIDEO_INFO_0_ID
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_ID,
      CATEGORY_PERFORMANCE_INFO_0_ID
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_ID,
      CATEGORY_HACKS_INFO_0_ID
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_ID,
      CATEGORY_INPUT_INFO_0_ID
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_ID,
      CATEGORY_VMU_INFO_0_ID
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_id[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_ID,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_ID },
         { "USA",     OPTION_VAL_USA_ID },
         { "Europe",  OPTION_VAL_EUROPE_ID },
         { "Default", OPTION_VAL_DEFAULT_ID },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_ID,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_ID },
         { "English",  OPTION_VAL_ENGLISH_ID },
         { "German",   OPTION_VAL_GERMAN_ID },
         { "French",   OPTION_VAL_FRENCH_ID },
         { "Spanish",  OPTION_VAL_SPANISH_ID },
         { "Italian",  OPTION_VAL_ITALIAN_ID },
         { "Default",  OPTION_VAL_DEFAULT_ID },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_ID,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_boot_to_bios",
      CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_ID,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_ID,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_force_wince",
      CORE_OPTION_NAME_FORCE_WINCE_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_FORCE_WINCE_INFO_0_ID,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_ID,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_ID,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_ID,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_ID,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_ID,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_ID },
         { "640x480",    OPTION_VAL_640X480_ID },
         { "800x600",    OPTION_VAL_800X600_ID },
         { "960x720",    OPTION_VAL_960X720_ID },
         { "1024x768",   OPTION_VAL_1024X768_ID },
         { "1280x960",   OPTION_VAL_1280X960_ID },
         { "1440x1080",  OPTION_VAL_1440X1080_ID },
         { "1600x1200",  OPTION_VAL_1600X1200_ID },
         { "1920x1440",  OPTION_VAL_1920X1440_ID },
         { "2560x1920",  OPTION_VAL_2560X1920_ID },
         { "2880x2160",  OPTION_VAL_2880X2160_ID },
         { "3200x2400",  OPTION_VAL_3200X2400_ID },
         { "3840x2880",  OPTION_VAL_3840X2880_ID },
         { "4480x3360",  OPTION_VAL_4480X3360_ID },
         { "5120x3840",  OPTION_VAL_5120X3840_ID },
         { "5760x4320",  OPTION_VAL_5760X4320_ID },
         { "6400x4800",  OPTION_VAL_6400X4800_ID },
         { "7040x5280",  OPTION_VAL_7040X5280_ID },
         { "7680x5760",  OPTION_VAL_7680X5760_ID },
         { "8320x6240",  OPTION_VAL_8320X6240_ID },
         { "8960x6720",  OPTION_VAL_8960X6720_ID },
         { "9600x7200",  OPTION_VAL_9600X7200_ID },
         { "10240x7680", OPTION_VAL_10240X7680_ID },
         { "10880x8160", OPTION_VAL_10880X8160_ID },
         { "11520x8640", OPTION_VAL_11520X8640_ID },
         { "12160x9120", OPTION_VAL_12160X9120_ID },
         { "12800x9600", OPTION_VAL_12800X9600_ID },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_ID,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_ID },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_ID },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_ID },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_ID,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_ID },
         { "PAL",     OPTION_VAL_PAL_ID },
         { "PAL_N",   OPTION_VAL_PAL_N_ID },
         { "PAL_M",   OPTION_VAL_PAL_M_ID },
         { "Default", OPTION_VAL_DEFAULT_ID },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_ID,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_ID },
         { "vertical",   OPTION_VAL_VERTICAL_ID },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_ID,
      NULL,
      "",
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_ID },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_ID },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_ID },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_ID,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_ID },
         { "1GB",   OPTION_VAL_1GB_ID },
         { "2GB",   OPTION_VAL_2GB_ID },
         { "4GB",   OPTION_VAL_4GB_ID },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_ID,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_ID,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_ID,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_ID,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_ID,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_ID,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_ID,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_ID,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_ID },
         { "1",  OPTION_VAL_1_ID },
         { "2",  OPTION_VAL_2_ID },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_ID,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_ID,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_ID,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_ID,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O27_ID },
         { "4",  OPTION_VAL_4_ID },
         { "6",  OPTION_VAL_6_ID },
         { NULL, NULL },
      },
      "1",
   },
   {/* TODO: needs clarification */
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_ID,
      NULL,
      "",
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_ID,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_ID,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_ID,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_ID,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_ID },
         { "more", OPTION_VAL_MORE_ID },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_ID,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_ID,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_ID,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_ID,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_ID,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: probably needs explanation */
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_ID,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_ID,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_ID },
         { "5%",  OPTION_VAL_5_ID },
         { "10%", OPTION_VAL_10_ID },
         { "15%", OPTION_VAL_15_ID },
         { "20%", OPTION_VAL_20_ID },
         { "25%", OPTION_VAL_25_ID },
         { "30%", OPTION_VAL_30_ID },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_ID,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_ID },
         { "5%",  OPTION_VAL_5_ID },
         { "10%", OPTION_VAL_10_ID },
         { "15%", OPTION_VAL_15_ID },
         { "20%", OPTION_VAL_20_ID },
         { "25%", OPTION_VAL_25_ID },
         { "30%", OPTION_VAL_30_ID },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_ID,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_purupuru",
      CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_ID,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_ID,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_ID,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_ID,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_ID },
         { "Red",      OPTION_VAL_RED_ID },
         { "Green",    OPTION_VAL_GREEN_ID },
         { "Blue",     OPTION_VAL_BLUE_ID },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_ID,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_ID },
         { "Red",      OPTION_VAL_RED_ID },
         { "Green",    OPTION_VAL_GREEN_ID },
         { "Blue",     OPTION_VAL_BLUE_ID },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_ID,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_ID },
         { "Red",      OPTION_VAL_RED_ID },
         { "Green",    OPTION_VAL_GREEN_ID },
         { "Blue",     OPTION_VAL_BLUE_ID },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_ID,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_ID },
         { "Red",      OPTION_VAL_RED_ID },
         { "Green",    OPTION_VAL_GREEN_ID },
         { "Blue",     OPTION_VAL_BLUE_ID },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_ID,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_ID,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_ID,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_ID },
         { "All VMUs", OPTION_VAL_ALL_VMUS_ID },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_ID,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_ID,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_ID,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_ID },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_ID },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_ID },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_ID },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_ID },
         { "2x", OPTION_VAL_2_O27_ID },
         { "3x", OPTION_VAL_3X_ID },
         { "4x", OPTION_VAL_4_ID },
         { "5x", OPTION_VAL_5X_ID },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_ID },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_ID },
         { "BLACK 02",          OPTION_VAL_BLACK_02_ID },
         { "BLUE 03",           OPTION_VAL_BLUE_ID },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_ID },
         { "GREEN 05",          OPTION_VAL_GREEN_ID },
         { "CYAN 06",           OPTION_VAL_CYAN_06_ID },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_ID },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_ID },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_ID },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_ID },
         { "RED 11",            OPTION_VAL_RED_ID },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_ID },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_ID },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_ID },
         { "GRAY 15",           OPTION_VAL_GRAY_15_ID },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_ID },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_ID },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_ID },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_ID },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_ID },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_ID },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_ID },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_ID },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_ID },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_ID },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_ID },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_ID },
         { "WHITE 28",          OPTION_VAL_WHITE_ID },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_ID },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_ID },
         { "BLACK 02",          OPTION_VAL_BLACK_02_ID },
         { "BLUE 03",           OPTION_VAL_BLUE_ID },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_ID },
         { "GREEN 05",          OPTION_VAL_GREEN_ID },
         { "CYAN 06",           OPTION_VAL_CYAN_06_ID },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_ID },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_ID },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_ID },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_ID },
         { "RED 11",            OPTION_VAL_RED_ID },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_ID },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_ID },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_ID },
         { "GRAY 15",           OPTION_VAL_GRAY_15_ID },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_ID },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_ID },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_ID },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_ID },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_ID },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_ID },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_ID },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_ID },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_ID },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_ID },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_ID },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_ID },
         { "WHITE 28",          OPTION_VAL_WHITE_ID },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_ID },
         { "20%",  OPTION_VAL_20_ID },
         { "30%",  OPTION_VAL_30_ID },
         { "40%",  OPTION_VAL_40_ID },
         { "50%",  OPTION_VAL_50_ID },
         { "60%",  OPTION_VAL_60_ID },
         { "70%",  OPTION_VAL_70_ID },
         { "80%",  OPTION_VAL_80_ID },
         { "90%",  OPTION_VAL_90_ID },
         { "100%", OPTION_VAL_100_ID },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_ID },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_ID },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_ID },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_ID },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_ID },
         { "2x", OPTION_VAL_2_O27_ID },
         { "3x", OPTION_VAL_3X_ID },
         { "4x", OPTION_VAL_4_ID },
         { "5x", OPTION_VAL_5X_ID },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_ID },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_ID },
         { "BLACK 02",          OPTION_VAL_BLACK_02_ID },
         { "BLUE 03",           OPTION_VAL_BLUE_ID },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_ID },
         { "GREEN 05",          OPTION_VAL_GREEN_ID },
         { "CYAN 06",           OPTION_VAL_CYAN_06_ID },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_ID },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_ID },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_ID },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_ID },
         { "RED 11",            OPTION_VAL_RED_ID },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_ID },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_ID },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_ID },
         { "GRAY 15",           OPTION_VAL_GRAY_15_ID },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_ID },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_ID },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_ID },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_ID },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_ID },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_ID },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_ID },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_ID },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_ID },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_ID },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_ID },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_ID },
         { "WHITE 28",          OPTION_VAL_WHITE_ID },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_ID },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_ID },
         { "BLACK 02",          OPTION_VAL_BLACK_02_ID },
         { "BLUE 03",           OPTION_VAL_BLUE_ID },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_ID },
         { "GREEN 05",          OPTION_VAL_GREEN_ID },
         { "CYAN 06",           OPTION_VAL_CYAN_06_ID },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_ID },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_ID },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_ID },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_ID },
         { "RED 11",            OPTION_VAL_RED_ID },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_ID },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_ID },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_ID },
         { "GRAY 15",           OPTION_VAL_GRAY_15_ID },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_ID },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_ID },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_ID },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_ID },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_ID },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_ID },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_ID },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_ID },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_ID },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_ID },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_ID },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_ID },
         { "WHITE 28",          OPTION_VAL_WHITE_ID },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_ID },
         { "20%",  OPTION_VAL_20_ID },
         { "30%",  OPTION_VAL_30_ID },
         { "40%",  OPTION_VAL_40_ID },
         { "50%",  OPTION_VAL_50_ID },
         { "60%",  OPTION_VAL_60_ID },
         { "70%",  OPTION_VAL_70_ID },
         { "80%",  OPTION_VAL_80_ID },
         { "90%",  OPTION_VAL_90_ID },
         { "100%", OPTION_VAL_100_ID },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_ID },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_ID },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_ID },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_ID },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_ID },
         { "2x", OPTION_VAL_2_O27_ID },
         { "3x", OPTION_VAL_3X_ID },
         { "4x", OPTION_VAL_4_ID },
         { "5x", OPTION_VAL_5X_ID },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_ID },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_ID },
         { "BLACK 02",          OPTION_VAL_BLACK_02_ID },
         { "BLUE 03",           OPTION_VAL_BLUE_ID },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_ID },
         { "GREEN 05",          OPTION_VAL_GREEN_ID },
         { "CYAN 06",           OPTION_VAL_CYAN_06_ID },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_ID },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_ID },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_ID },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_ID },
         { "RED 11",            OPTION_VAL_RED_ID },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_ID },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_ID },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_ID },
         { "GRAY 15",           OPTION_VAL_GRAY_15_ID },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_ID },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_ID },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_ID },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_ID },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_ID },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_ID },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_ID },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_ID },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_ID },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_ID },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_ID },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_ID },
         { "WHITE 28",          OPTION_VAL_WHITE_ID },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_ID },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_ID },
         { "BLACK 02",          OPTION_VAL_BLACK_02_ID },
         { "BLUE 03",           OPTION_VAL_BLUE_ID },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_ID },
         { "GREEN 05",          OPTION_VAL_GREEN_ID },
         { "CYAN 06",           OPTION_VAL_CYAN_06_ID },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_ID },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_ID },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_ID },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_ID },
         { "RED 11",            OPTION_VAL_RED_ID },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_ID },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_ID },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_ID },
         { "GRAY 15",           OPTION_VAL_GRAY_15_ID },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_ID },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_ID },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_ID },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_ID },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_ID },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_ID },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_ID },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_ID },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_ID },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_ID },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_ID },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_ID },
         { "WHITE 28",          OPTION_VAL_WHITE_ID },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_ID },
         { "20%",  OPTION_VAL_20_ID },
         { "30%",  OPTION_VAL_30_ID },
         { "40%",  OPTION_VAL_40_ID },
         { "50%",  OPTION_VAL_50_ID },
         { "60%",  OPTION_VAL_60_ID },
         { "70%",  OPTION_VAL_70_ID },
         { "80%",  OPTION_VAL_80_ID },
         { "90%",  OPTION_VAL_90_ID },
         { "100%", OPTION_VAL_100_ID },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_ID },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_ID },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_ID },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_ID },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_ID },
         { "2x", OPTION_VAL_2_O27_ID },
         { "3x", OPTION_VAL_3X_ID },
         { "4x", OPTION_VAL_4_ID },
         { "5x", OPTION_VAL_5X_ID },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_ID },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_ID },
         { "BLACK 02",          OPTION_VAL_BLACK_02_ID },
         { "BLUE 03",           OPTION_VAL_BLUE_ID },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_ID },
         { "GREEN 05",          OPTION_VAL_GREEN_ID },
         { "CYAN 06",           OPTION_VAL_CYAN_06_ID },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_ID },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_ID },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_ID },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_ID },
         { "RED 11",            OPTION_VAL_RED_ID },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_ID },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_ID },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_ID },
         { "GRAY 15",           OPTION_VAL_GRAY_15_ID },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_ID },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_ID },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_ID },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_ID },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_ID },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_ID },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_ID },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_ID },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_ID },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_ID },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_ID },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_ID },
         { "WHITE 28",          OPTION_VAL_WHITE_ID },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_ID },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_ID },
         { "BLACK 02",          OPTION_VAL_BLACK_02_ID },
         { "BLUE 03",           OPTION_VAL_BLUE_ID },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_ID },
         { "GREEN 05",          OPTION_VAL_GREEN_ID },
         { "CYAN 06",           OPTION_VAL_CYAN_06_ID },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_ID },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_ID },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_ID },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_ID },
         { "RED 11",            OPTION_VAL_RED_ID },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_ID },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_ID },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_ID },
         { "GRAY 15",           OPTION_VAL_GRAY_15_ID },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_ID },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_ID },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_ID },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_ID },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_ID },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_ID },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_ID },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_ID },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_ID },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_ID },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_ID },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_ID },
         { "WHITE 28",          OPTION_VAL_WHITE_ID },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_ID },
         { "20%",  OPTION_VAL_20_ID },
         { "30%",  OPTION_VAL_30_ID },
         { "40%",  OPTION_VAL_40_ID },
         { "50%",  OPTION_VAL_50_ID },
         { "60%",  OPTION_VAL_60_ID },
         { "70%",  OPTION_VAL_70_ID },
         { "80%",  OPTION_VAL_80_ID },
         { "90%",  OPTION_VAL_90_ID },
         { "100%", OPTION_VAL_100_ID },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_id = {
   option_cats_id,
   option_defs_id
};

/* RETRO_LANGUAGE_IT */

#define CATEGORY_SYSTEM_LABEL_IT "Sistema"
#define CATEGORY_SYSTEM_INFO_0_IT "Configurare la regione, la lingua, il BIOS e le impostazioni hardware di base."
#define CATEGORY_VIDEO_LABEL_IT NULL
#define CATEGORY_VIDEO_INFO_0_IT "Configurare la risoluzione, la trasparenza indipendente dall'ordine e le impostazioni degli effetti visivi."
#define CATEGORY_PERFORMANCE_LABEL_IT "Prestazioni"
#define CATEGORY_PERFORMANCE_INFO_0_IT "Configurare il rendering filettato e le impostazioni del salto di fotogramma."
#define CATEGORY_HACKS_LABEL_IT "Hack di Emulazione"
#define CATEGORY_HACKS_INFO_0_IT "Configura override widescreen, velocità di caricamento GD-ROM e impostazioni di sostituzione texture."
#define CATEGORY_INPUT_LABEL_IT NULL
#define CATEGORY_INPUT_INFO_0_IT "Configura impostazioni gamepad e pistola ottica."
#define CATEGORY_VMU_LABEL_IT "Unità Memoria Visiva"
#define CATEGORY_VMU_INFO_0_IT "Configura i file VMU per gioco e le impostazioni di visibilità VMU sullo schermo."
#define CORE_OPTION_NAME_REGION_LABEL_IT "Regione"
#define OPTION_VAL_JAPAN_IT "Giappone"
#define OPTION_VAL_USA_IT NULL
#define OPTION_VAL_EUROPE_IT "Europa"
#define OPTION_VAL_DEFAULT_IT "Predefinito"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_IT "Lingua"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_IT "Cambia la lingua usata dal BIOS e da qualsiasi gioco che contenga più lingue."
#define OPTION_VAL_JAPANESE_IT "Giapponese"
#define OPTION_VAL_ENGLISH_IT "Inglese"
#define OPTION_VAL_GERMAN_IT "Tedesco"
#define OPTION_VAL_FRENCH_IT "Francese"
#define OPTION_VAL_SPANISH_IT "Spagnolo"
#define OPTION_VAL_ITALIAN_IT "Italiano"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_IT "HLE BIOS (Riavvio Richiesto)"
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_IT "Forza l'uso di BIOS di emulazione di alto livello."
#define CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_IT "Avvio in BIOS (Riavvio richiesto)"
#define CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_IT "Avviare direttamente nel menu BIOS Dreamcast."
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_IT "Abilita DSP"
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_IT "Abilita l'emulazione del DSP audio di Dreamcast (processore di segnale digitale). Migliora l'accuratezza del suono generato, ma aumenta i requisiti di prestazioni."
#define CORE_OPTION_NAME_FORCE_WINCE_LABEL_IT "Forza Modalità Windows Ce"
#define CORE_OPTION_NAME_FORCE_WINCE_INFO_0_IT "Abilita l'emulazione di MMU (Memory Management Unit) e altre impostazioni per i giochi Windows CE."
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_IT "Consenti I Pulsanti Di Servizio NAOMI"
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_IT "Abilita il pulsante SERVICE per NAOMI, per inserire le impostazioni del cabinato."
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_IT "Imposta giochi NAOMI a Free Play"
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_IT "Modificare le impostazioni del gioco per free play."
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_IT "Emulazione Adattatore A banda larga"
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_IT "Emulare l'adattatore a banda larga ethernet invece del modem. (Riavvio richiesto)"
#define CORE_OPTION_NAME_UPNP_LABEL_IT "Abilita UPnP"
#define CORE_OPTION_NAME_UPNP_INFO_0_IT "Usa UPnP per configurare automaticamente il router Internet per i giochi online."
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_IT "Risoluzione Interna"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_IT "Modificare la risoluzione di rendering."
#define OPTION_VAL_320X240_IT NULL
#define OPTION_VAL_640X480_IT NULL
#define OPTION_VAL_800X600_IT NULL
#define OPTION_VAL_960X720_IT NULL
#define OPTION_VAL_1024X768_IT NULL
#define OPTION_VAL_1280X960_IT NULL
#define OPTION_VAL_1440X1080_IT NULL
#define OPTION_VAL_1600X1200_IT NULL
#define OPTION_VAL_1920X1440_IT NULL
#define OPTION_VAL_2560X1920_IT NULL
#define OPTION_VAL_2880X2160_IT NULL
#define OPTION_VAL_3200X2400_IT NULL
#define OPTION_VAL_3840X2880_IT NULL
#define OPTION_VAL_4480X3360_IT NULL
#define OPTION_VAL_5120X3840_IT NULL
#define OPTION_VAL_5760X4320_IT NULL
#define OPTION_VAL_6400X4800_IT NULL
#define OPTION_VAL_7040X5280_IT NULL
#define OPTION_VAL_7680X5760_IT NULL
#define OPTION_VAL_8320X6240_IT NULL
#define OPTION_VAL_8960X6720_IT NULL
#define OPTION_VAL_9600X7200_IT NULL
#define OPTION_VAL_10240X7680_IT NULL
#define OPTION_VAL_10880X8160_IT NULL
#define OPTION_VAL_11520X8640_IT NULL
#define OPTION_VAL_12160X9120_IT NULL
#define OPTION_VAL_12800X9600_IT NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_IT "Tipo Di Cavo"
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_IT "Il tipo di segnale di uscita. 'TV (Composito)' è il più ampiamente supportato."
#define OPTION_VAL_VGA_IT NULL
#define OPTION_VAL_TV_RGB_IT NULL
#define OPTION_VAL_TV_COMPOSITE_IT "TV (Composto)"
#define CORE_OPTION_NAME_BROADCAST_LABEL_IT "Standard di trasmissione"
#define OPTION_VAL_NTSC_IT NULL
#define OPTION_VAL_PAL_IT "PAL (Mondo)"
#define OPTION_VAL_PAL_N_IT NULL
#define OPTION_VAL_PAL_M_IT "PAL-M (Brasile)"
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_IT "Orientamento Schermo"
#define OPTION_VAL_HORIZONTAL_IT "Orizzontale"
#define OPTION_VAL_VERTICAL_IT "Verticale"
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_IT "Ordinamento Alfa"
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_IT "Per-Strip (veloce, meno preciso)"
#define OPTION_VAL_PER_TRIANGLE_NORMAL_IT "Per-Triangolo (normale)"
#define OPTION_VAL_PER_PIXEL_ACCURATE_IT "Per-Pixel (preciso, ma più lento)"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_IT "Dimensione di accumulo del Buffer Pixel"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_IT "Potrebbero essere necessari valori più alti per ottenere risoluzioni più elevate in uscita correttamente."
#define OPTION_VAL_512MB_IT NULL
#define OPTION_VAL_1GB_IT NULL
#define OPTION_VAL_2GB_IT NULL
#define OPTION_VAL_4GB_IT NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_IT "Strati trasparenti massimi"
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_IT "Potrebbero essere necessari valori più elevati per scene complesse."
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_IT "Emulazione framebuffer completa"
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_IT "Abilita l'emulazione completa del framebuffer in VRAM. Ciò è utile per i giochi che leggono o scrivono direttamente il framebuffer in VRAM. Se abilitata, la risoluzione interna è costretta a 640x480 e le prestazioni possono essere gravemente colpite."
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_IT "Abilita Buffer RTT (Render To Texture)"
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_IT "Copia le texture renderizzate dalla GPU a VRAM. Questa opzione è normalmente abilitata per i giochi che lo richiedono. Se abilitata, il rendering delle texture è disabilitato e le prestazioni potrebbero essere influenzate."
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_IT NULL
#define CORE_OPTION_NAME_FOG_LABEL_IT "Effetti Nebbia"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_IT "Modificatore del Volume"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_IT "Una funzionalità della GPU Dreamcast che viene tipicamente utilizzata dai giochi per disegnare ombre degli oggetti. Questo dovrebbe normalmente essere abilitato - l'impatto delle prestazioni di solito è minimo o trascurabile."
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_IT "Filtro Anisotropico"
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_IT "Migliorare la qualità delle texture su superfici che sono ad angoli obliqui rispetto alla fotocamera. Valori più alti sono più impegnativi per la GPU. Le modifiche a questa impostazione si applicano solo dopo il riavvio."
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_IT "Filtro Texture"
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_IT "La modalità di filtraggio delle texture da usare. Questo può essere usato per forzare una certa modalità di filtraggio delle texture su tutte le texture per ottenere un aspetto più nitido (o più liscio) di quello predefinito. Valori diversi da quelli predefiniti possono causare vari problemi di rendering. Le modifiche a questa impostazione si applicano solo dopo il riavvio."
#define OPTION_VAL_1_IT "Forza il Vicino più Vicino"
#define OPTION_VAL_2_IT "Forza Lineare"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_IT "Ritardo Scambio fotogrammi"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_IT "Utile per evitare di lampeggiare schermi o video glitchati. Sconsigliato su piattaforme lente."
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_IT "Rileva Modifiche Frame Rate"
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_IT "Notificare il frontend quando il frame rate interno cambia (ad esempio da 60 fps a 30 fps). Migliora il ritmo dei fotogrammi nei giochi che girano a un bloccato 30 fps o 20 fps, ma dovrebbe essere disabilitato per i giochi con frame rate sbloccati (instabili) (e.. Ecco il delfino, torneo irreale). Nota: non disponibile quando 'Auto Salta Frame' è abilitato."
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_IT "Filtro Di Post-Elaborazione PowerVR2"
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_IT "Post-elaborare l'immagine renderizzata per simulare effetti specifici alla GPU PowerVR2 e segnali video analogici."
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_IT "Miglioramento Della Texture (xbrz)"
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_IT "Migliora la grafica grafica grafica a pixel 2D disegnata a mano. Dovrebbe essere usata solo con giochi in pixel 2D."
#define OPTION_VAL_2_O27_IT NULL
#define OPTION_VAL_4_IT NULL
#define OPTION_VAL_6_IT NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_IT "Miglioramento Texture Dimensione Massima Filtrata"
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_IT "Interpolazione Di Profondità Nativa"
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_IT "Aiuta con la corruzione della texture e problemi di profondità sulle GPU AMD. Può anche aiutare le GPU Intel in alcuni casi."
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_IT "Rendering Filettato"
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_IT "Esegue la GPU e la CPU su diversi thread. Altamente raccomandato."
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_IT "Auto Salto Fotogrammi"
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_IT "Salta automaticamente i fotogrammi quando l'emulatore è lento. Nota: questa impostazione si applica solo quando il Rendering Threaded è abilitato."
#define OPTION_VAL_SOME_IT "Normale"
#define OPTION_VAL_MORE_IT "Massimo"
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_IT "Salto fotogrammi"
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_IT "Imposta il numero di fotogrammi da saltare tra ogni fotogramma visualizzato."
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_IT "Trucchi Widescreen (Riavvio Richiesto)"
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_IT "Attiva i trucchi che permettono ad alcuni giochi di visualizzare in formato widescreen."
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_IT "Hack Widescreen"
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_IT "Disegnare la geometria al di fuori del normale rapporto di aspetto 4:3. Può produrre difetti grafici nelle aree rivelate."
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_IT "Caricamento Veloce GD-ROM (inaccurato)"
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_IT "Accelera il caricamento di GD-ROM."
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_IT "Carica Texture Personalizzate"
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_IT "Scarica Texture"
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_IT "Zona morta dello sticl analogico"
#define OPTION_VAL_0_IT NULL
#define OPTION_VAL_5_IT NULL
#define OPTION_VAL_10_IT NULL
#define OPTION_VAL_15_IT NULL
#define OPTION_VAL_20_IT NULL
#define OPTION_VAL_25_IT NULL
#define OPTION_VAL_30_IT NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_IT "Zona morta dei trigger"
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_IT "Trigger Digitali"
#define CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_IT "Pacchetto/Vibrazione Purupuru"
#define CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_IT "Attiva al vibrazione del controller."
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_IT "Uscite di trasmissione digitale"
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_IT "Trasmissione di uscite digitali e stato di forza-feedback sulla porta TCP 8000. Compatibile con l'opzione MAME \"-output network\"."
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_IT "Mostra Impostazioni Pistola Ottica"
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_IT "Abilita la configurazione delle opzioni di visualizzazione di una pistola leggera. NOTA: potrebbe essere necessario attivare il menu rapido perché questa impostazione abbia effetto."
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_IT "Mostra mirino della Pistola 1"
#define OPTION_VAL_WHITE_IT "Bianco"
#define OPTION_VAL_RED_IT "Rosso"
#define OPTION_VAL_GREEN_IT "Verde"
#define OPTION_VAL_BLUE_IT "Blu"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_IT "Mostra mirino della Pistola 2"
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_IT "Mostra mirino della Pistola 3"
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_IT "Mostra mirino della Pistola 4"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_IT "Visual Memory Units/Sistemi per Gioco (VMU)"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_IT "VMU per gioco"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_IT "Quando disabilitato, tutti i giochi condividono 4 file VMU salvati (A1, B1, C1, D1) situati nella directory di sistema di RetroArch. L'impostazione 'VMU A1' crea un unico file VMU 'A1' nella directory di salvataggio di RetroArch, per ogni gioco che viene lanciato. L'impostazione 'All VMUs' crea 4 file VMU unici (A1, B1, C1, D1) per ogni gioco che viene lanciato."
#define OPTION_VAL_VMU_A1_IT NULL
#define OPTION_VAL_ALL_VMUS_IT "Tutte Le VMU"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_IT "Mostra Le Impostazioni Di Visualizzazione della Visual Memory Unit/System (VMU)"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_IT "Mostra Impostazioni Di Visualizzazione VMU"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_IT "Abilita la configurazione dello schermo LCD VMU emulato, la dimensione, la posizione e il colore. NOTA: potrebbe essere necessario attivare il menu rapido affinché questa impostazione abbia effetto."
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_IT "Visualizzazione Schermo VMU 1"
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_IT "Posizione VMU Schermo 1"
#define OPTION_VAL_UPPER_LEFT_IT "In Alto a Sinistra"
#define OPTION_VAL_UPPER_RIGHT_IT "In Alto a Destra"
#define OPTION_VAL_LOWER_LEFT_IT "In Basso a Sinistra"
#define OPTION_VAL_LOWER_RIGHT_IT "In Basso a Destra"
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_IT "Dimensione VMU Schermo 1"
#define OPTION_VAL_1X_IT NULL
#define OPTION_VAL_3X_IT NULL
#define OPTION_VAL_5X_IT NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_IT "Schermo VMU 1 Pixel Su Colore"
#define OPTION_VAL_DEFAULT_ON_00_IT "Predefinito ON"
#define OPTION_VAL_DEFAULT_OFF_01_IT "Predefinito OFF"
#define OPTION_VAL_BLACK_02_IT "Nero"
#define OPTION_VAL_LIGHT_BLUE_04_IT "Azzurro"
#define OPTION_VAL_CYAN_06_IT "Ciano"
#define OPTION_VAL_CYAN_BLUE_07_IT "Blu Ciano"
#define OPTION_VAL_LIGHT_GREEN_08_IT "Verde Chiaro"
#define OPTION_VAL_CYAN_GREEN_09_IT "Verde Ciano"
#define OPTION_VAL_LIGHT_CYAN_10_IT "Ciano Chiaro"
#define OPTION_VAL_PURPLE_12_IT "Viola"
#define OPTION_VAL_LIGHT_PURPLE_13_IT "Viola Chiaro"
#define OPTION_VAL_YELLOW_14_IT "Giallo"
#define OPTION_VAL_GRAY_15_IT "Grigio"
#define OPTION_VAL_LIGHT_PURPLE_2_16_IT "Viola Chiaro (2)"
#define OPTION_VAL_LIGHT_GREEN_2_17_IT "Verde Chiaro (2)"
#define OPTION_VAL_LIGHT_GREEN_3_18_IT "Verde Chiaro (3)"
#define OPTION_VAL_LIGHT_CYAN_2_19_IT "Ciano Chiaro (2)"
#define OPTION_VAL_LIGHT_RED_2_20_IT "Rosso Chiaro (2)"
#define OPTION_VAL_MAGENTA_21_IT NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_IT "Viola Chiaro (3)"
#define OPTION_VAL_LIGHT_ORANGE_23_IT "Arancione Chiaro"
#define OPTION_VAL_ORANGE_24_IT "Arancione"
#define OPTION_VAL_LIGHT_PURPLE_4_25_IT "Viola Chiaro (4)"
#define OPTION_VAL_LIGHT_YELLOW_26_IT "Giallo Chiaro"
#define OPTION_VAL_LIGHT_YELLOW_2_27_IT "Giallo Chiaro (2)"
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_IT "Colore Vmu Schermo 1 Pixel Off"
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_IT "Opacità VMU Screen 1"
#define OPTION_VAL_40_IT NULL
#define OPTION_VAL_50_IT NULL
#define OPTION_VAL_60_IT NULL
#define OPTION_VAL_70_IT NULL
#define OPTION_VAL_80_IT NULL
#define OPTION_VAL_90_IT NULL
#define OPTION_VAL_100_IT NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_IT "Visualizzazione Schermo VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_IT "Posizione VMU Schermo 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_IT "Dimensione VMU Schermo 2"
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_IT "Colore VMU Schermo 2 Pixel On"
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_IT "Colore VMU Schermo 2 Pixel Off"
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_IT "Opacità VMU Schermo 2"
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_IT "Visualizzazione Schermo 3 VMU"
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_IT "Posizione Schermo 3 VMU"
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_IT "Dimensione Schermo 3 VMU"
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_IT "Colore su Pixel Schermo VMU 3"
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_IT "Colore Pixel Off Schermo VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_IT "Opacità Schermo VMU 3"
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_IT "Visualizzazione Schermo VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_IT "Posizione Schermo VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_IT "Dimensione Schermo VMU 4"
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_IT "Colore Pixel On Schermo VMU 4"
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_IT "Colore Pixel Off Schermo VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_IT "Opacità Schermo VMU 4"

struct retro_core_option_v2_category option_cats_it[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_IT,
      CATEGORY_SYSTEM_INFO_0_IT
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_IT,
      CATEGORY_VIDEO_INFO_0_IT
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_IT,
      CATEGORY_PERFORMANCE_INFO_0_IT
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_IT,
      CATEGORY_HACKS_INFO_0_IT
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_IT,
      CATEGORY_INPUT_INFO_0_IT
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_IT,
      CATEGORY_VMU_INFO_0_IT
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_it[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_IT,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_IT },
         { "USA",     OPTION_VAL_USA_IT },
         { "Europe",  OPTION_VAL_EUROPE_IT },
         { "Default", OPTION_VAL_DEFAULT_IT },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_IT,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_IT },
         { "English",  OPTION_VAL_ENGLISH_IT },
         { "German",   OPTION_VAL_GERMAN_IT },
         { "French",   OPTION_VAL_FRENCH_IT },
         { "Spanish",  OPTION_VAL_SPANISH_IT },
         { "Italian",  OPTION_VAL_ITALIAN_IT },
         { "Default",  OPTION_VAL_DEFAULT_IT },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_IT,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_boot_to_bios",
      CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_IT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_IT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_force_wince",
      CORE_OPTION_NAME_FORCE_WINCE_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_FORCE_WINCE_INFO_0_IT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_IT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_IT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_IT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_IT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_IT,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_IT },
         { "640x480",    OPTION_VAL_640X480_IT },
         { "800x600",    OPTION_VAL_800X600_IT },
         { "960x720",    OPTION_VAL_960X720_IT },
         { "1024x768",   OPTION_VAL_1024X768_IT },
         { "1280x960",   OPTION_VAL_1280X960_IT },
         { "1440x1080",  OPTION_VAL_1440X1080_IT },
         { "1600x1200",  OPTION_VAL_1600X1200_IT },
         { "1920x1440",  OPTION_VAL_1920X1440_IT },
         { "2560x1920",  OPTION_VAL_2560X1920_IT },
         { "2880x2160",  OPTION_VAL_2880X2160_IT },
         { "3200x2400",  OPTION_VAL_3200X2400_IT },
         { "3840x2880",  OPTION_VAL_3840X2880_IT },
         { "4480x3360",  OPTION_VAL_4480X3360_IT },
         { "5120x3840",  OPTION_VAL_5120X3840_IT },
         { "5760x4320",  OPTION_VAL_5760X4320_IT },
         { "6400x4800",  OPTION_VAL_6400X4800_IT },
         { "7040x5280",  OPTION_VAL_7040X5280_IT },
         { "7680x5760",  OPTION_VAL_7680X5760_IT },
         { "8320x6240",  OPTION_VAL_8320X6240_IT },
         { "8960x6720",  OPTION_VAL_8960X6720_IT },
         { "9600x7200",  OPTION_VAL_9600X7200_IT },
         { "10240x7680", OPTION_VAL_10240X7680_IT },
         { "10880x8160", OPTION_VAL_10880X8160_IT },
         { "11520x8640", OPTION_VAL_11520X8640_IT },
         { "12160x9120", OPTION_VAL_12160X9120_IT },
         { "12800x9600", OPTION_VAL_12800X9600_IT },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_IT,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_IT },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_IT },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_IT },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_IT,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_IT },
         { "PAL",     OPTION_VAL_PAL_IT },
         { "PAL_N",   OPTION_VAL_PAL_N_IT },
         { "PAL_M",   OPTION_VAL_PAL_M_IT },
         { "Default", OPTION_VAL_DEFAULT_IT },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_IT,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_IT },
         { "vertical",   OPTION_VAL_VERTICAL_IT },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_IT,
      NULL,
      "",
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_IT },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_IT },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_IT },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_IT,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_IT },
         { "1GB",   OPTION_VAL_1GB_IT },
         { "2GB",   OPTION_VAL_2GB_IT },
         { "4GB",   OPTION_VAL_4GB_IT },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_IT,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_IT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_IT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_IT,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_IT,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_IT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_IT,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_IT,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_IT },
         { "1",  OPTION_VAL_1_IT },
         { "2",  OPTION_VAL_2_IT },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_IT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_IT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_IT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_IT,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O27_IT },
         { "4",  OPTION_VAL_4_IT },
         { "6",  OPTION_VAL_6_IT },
         { NULL, NULL },
      },
      "1",
   },
   {/* TODO: needs clarification */
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_IT,
      NULL,
      "",
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_IT,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_IT,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_IT,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_IT,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_IT },
         { "more", OPTION_VAL_MORE_IT },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_IT,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_IT,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_IT,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_IT,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_IT,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: probably needs explanation */
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_IT,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_IT,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_IT },
         { "5%",  OPTION_VAL_5_IT },
         { "10%", OPTION_VAL_10_IT },
         { "15%", OPTION_VAL_15_IT },
         { "20%", OPTION_VAL_20_IT },
         { "25%", OPTION_VAL_25_IT },
         { "30%", OPTION_VAL_30_IT },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_IT,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_IT },
         { "5%",  OPTION_VAL_5_IT },
         { "10%", OPTION_VAL_10_IT },
         { "15%", OPTION_VAL_15_IT },
         { "20%", OPTION_VAL_20_IT },
         { "25%", OPTION_VAL_25_IT },
         { "30%", OPTION_VAL_30_IT },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_IT,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_purupuru",
      CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_IT,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_IT,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_IT,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_IT,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_IT },
         { "Red",      OPTION_VAL_RED_IT },
         { "Green",    OPTION_VAL_GREEN_IT },
         { "Blue",     OPTION_VAL_BLUE_IT },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_IT,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_IT },
         { "Red",      OPTION_VAL_RED_IT },
         { "Green",    OPTION_VAL_GREEN_IT },
         { "Blue",     OPTION_VAL_BLUE_IT },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_IT,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_IT },
         { "Red",      OPTION_VAL_RED_IT },
         { "Green",    OPTION_VAL_GREEN_IT },
         { "Blue",     OPTION_VAL_BLUE_IT },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_IT,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_IT },
         { "Red",      OPTION_VAL_RED_IT },
         { "Green",    OPTION_VAL_GREEN_IT },
         { "Blue",     OPTION_VAL_BLUE_IT },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_IT,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_IT,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_IT,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_IT },
         { "All VMUs", OPTION_VAL_ALL_VMUS_IT },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_IT,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_IT,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_IT,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_IT },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_IT },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_IT },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_IT },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_IT },
         { "2x", OPTION_VAL_2_O27_IT },
         { "3x", OPTION_VAL_3X_IT },
         { "4x", OPTION_VAL_4_IT },
         { "5x", OPTION_VAL_5X_IT },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_IT },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_IT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_IT },
         { "BLUE 03",           OPTION_VAL_BLUE_IT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_IT },
         { "GREEN 05",          OPTION_VAL_GREEN_IT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_IT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_IT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_IT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_IT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_IT },
         { "RED 11",            OPTION_VAL_RED_IT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_IT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_IT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_IT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_IT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_IT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_IT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_IT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_IT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_IT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_IT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_IT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_IT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_IT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_IT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_IT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_IT },
         { "WHITE 28",          OPTION_VAL_WHITE_IT },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_IT },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_IT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_IT },
         { "BLUE 03",           OPTION_VAL_BLUE_IT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_IT },
         { "GREEN 05",          OPTION_VAL_GREEN_IT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_IT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_IT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_IT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_IT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_IT },
         { "RED 11",            OPTION_VAL_RED_IT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_IT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_IT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_IT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_IT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_IT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_IT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_IT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_IT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_IT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_IT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_IT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_IT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_IT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_IT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_IT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_IT },
         { "WHITE 28",          OPTION_VAL_WHITE_IT },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_IT },
         { "20%",  OPTION_VAL_20_IT },
         { "30%",  OPTION_VAL_30_IT },
         { "40%",  OPTION_VAL_40_IT },
         { "50%",  OPTION_VAL_50_IT },
         { "60%",  OPTION_VAL_60_IT },
         { "70%",  OPTION_VAL_70_IT },
         { "80%",  OPTION_VAL_80_IT },
         { "90%",  OPTION_VAL_90_IT },
         { "100%", OPTION_VAL_100_IT },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_IT },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_IT },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_IT },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_IT },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_IT },
         { "2x", OPTION_VAL_2_O27_IT },
         { "3x", OPTION_VAL_3X_IT },
         { "4x", OPTION_VAL_4_IT },
         { "5x", OPTION_VAL_5X_IT },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_IT },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_IT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_IT },
         { "BLUE 03",           OPTION_VAL_BLUE_IT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_IT },
         { "GREEN 05",          OPTION_VAL_GREEN_IT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_IT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_IT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_IT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_IT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_IT },
         { "RED 11",            OPTION_VAL_RED_IT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_IT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_IT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_IT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_IT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_IT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_IT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_IT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_IT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_IT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_IT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_IT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_IT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_IT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_IT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_IT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_IT },
         { "WHITE 28",          OPTION_VAL_WHITE_IT },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_IT },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_IT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_IT },
         { "BLUE 03",           OPTION_VAL_BLUE_IT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_IT },
         { "GREEN 05",          OPTION_VAL_GREEN_IT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_IT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_IT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_IT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_IT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_IT },
         { "RED 11",            OPTION_VAL_RED_IT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_IT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_IT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_IT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_IT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_IT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_IT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_IT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_IT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_IT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_IT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_IT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_IT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_IT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_IT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_IT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_IT },
         { "WHITE 28",          OPTION_VAL_WHITE_IT },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_IT },
         { "20%",  OPTION_VAL_20_IT },
         { "30%",  OPTION_VAL_30_IT },
         { "40%",  OPTION_VAL_40_IT },
         { "50%",  OPTION_VAL_50_IT },
         { "60%",  OPTION_VAL_60_IT },
         { "70%",  OPTION_VAL_70_IT },
         { "80%",  OPTION_VAL_80_IT },
         { "90%",  OPTION_VAL_90_IT },
         { "100%", OPTION_VAL_100_IT },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_IT },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_IT },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_IT },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_IT },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_IT },
         { "2x", OPTION_VAL_2_O27_IT },
         { "3x", OPTION_VAL_3X_IT },
         { "4x", OPTION_VAL_4_IT },
         { "5x", OPTION_VAL_5X_IT },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_IT },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_IT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_IT },
         { "BLUE 03",           OPTION_VAL_BLUE_IT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_IT },
         { "GREEN 05",          OPTION_VAL_GREEN_IT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_IT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_IT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_IT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_IT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_IT },
         { "RED 11",            OPTION_VAL_RED_IT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_IT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_IT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_IT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_IT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_IT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_IT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_IT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_IT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_IT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_IT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_IT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_IT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_IT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_IT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_IT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_IT },
         { "WHITE 28",          OPTION_VAL_WHITE_IT },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_IT },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_IT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_IT },
         { "BLUE 03",           OPTION_VAL_BLUE_IT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_IT },
         { "GREEN 05",          OPTION_VAL_GREEN_IT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_IT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_IT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_IT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_IT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_IT },
         { "RED 11",            OPTION_VAL_RED_IT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_IT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_IT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_IT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_IT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_IT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_IT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_IT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_IT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_IT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_IT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_IT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_IT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_IT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_IT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_IT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_IT },
         { "WHITE 28",          OPTION_VAL_WHITE_IT },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_IT },
         { "20%",  OPTION_VAL_20_IT },
         { "30%",  OPTION_VAL_30_IT },
         { "40%",  OPTION_VAL_40_IT },
         { "50%",  OPTION_VAL_50_IT },
         { "60%",  OPTION_VAL_60_IT },
         { "70%",  OPTION_VAL_70_IT },
         { "80%",  OPTION_VAL_80_IT },
         { "90%",  OPTION_VAL_90_IT },
         { "100%", OPTION_VAL_100_IT },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_IT },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_IT },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_IT },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_IT },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_IT },
         { "2x", OPTION_VAL_2_O27_IT },
         { "3x", OPTION_VAL_3X_IT },
         { "4x", OPTION_VAL_4_IT },
         { "5x", OPTION_VAL_5X_IT },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_IT },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_IT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_IT },
         { "BLUE 03",           OPTION_VAL_BLUE_IT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_IT },
         { "GREEN 05",          OPTION_VAL_GREEN_IT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_IT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_IT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_IT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_IT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_IT },
         { "RED 11",            OPTION_VAL_RED_IT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_IT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_IT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_IT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_IT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_IT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_IT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_IT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_IT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_IT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_IT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_IT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_IT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_IT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_IT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_IT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_IT },
         { "WHITE 28",          OPTION_VAL_WHITE_IT },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_IT },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_IT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_IT },
         { "BLUE 03",           OPTION_VAL_BLUE_IT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_IT },
         { "GREEN 05",          OPTION_VAL_GREEN_IT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_IT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_IT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_IT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_IT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_IT },
         { "RED 11",            OPTION_VAL_RED_IT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_IT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_IT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_IT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_IT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_IT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_IT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_IT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_IT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_IT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_IT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_IT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_IT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_IT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_IT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_IT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_IT },
         { "WHITE 28",          OPTION_VAL_WHITE_IT },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_IT },
         { "20%",  OPTION_VAL_20_IT },
         { "30%",  OPTION_VAL_30_IT },
         { "40%",  OPTION_VAL_40_IT },
         { "50%",  OPTION_VAL_50_IT },
         { "60%",  OPTION_VAL_60_IT },
         { "70%",  OPTION_VAL_70_IT },
         { "80%",  OPTION_VAL_80_IT },
         { "90%",  OPTION_VAL_90_IT },
         { "100%", OPTION_VAL_100_IT },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_it = {
   option_cats_it,
   option_defs_it
};

/* RETRO_LANGUAGE_JA */

#define CATEGORY_SYSTEM_LABEL_JA "システム"
#define CATEGORY_SYSTEM_INFO_0_JA NULL
#define CATEGORY_VIDEO_LABEL_JA "ビデオのドライバ"
#define CATEGORY_VIDEO_INFO_0_JA NULL
#define CATEGORY_PERFORMANCE_LABEL_JA "パフォーマンス"
#define CATEGORY_PERFORMANCE_INFO_0_JA NULL
#define CATEGORY_HACKS_LABEL_JA "エミュレーションハック"
#define CATEGORY_HACKS_INFO_0_JA NULL
#define CATEGORY_INPUT_LABEL_JA "入力"
#define CATEGORY_INPUT_INFO_0_JA NULL
#define CATEGORY_VMU_LABEL_JA NULL
#define CATEGORY_VMU_INFO_0_JA NULL
#define CORE_OPTION_NAME_REGION_LABEL_JA "地域"
#define OPTION_VAL_JAPAN_JA NULL
#define OPTION_VAL_USA_JA NULL
#define OPTION_VAL_EUROPE_JA NULL
#define OPTION_VAL_DEFAULT_JA "デフォルト"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_JA "言語"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_JA NULL
#define OPTION_VAL_JAPANESE_JA "Japanese - 日本語"
#define OPTION_VAL_ENGLISH_JA "英語"
#define OPTION_VAL_GERMAN_JA "German - ドイツ語"
#define OPTION_VAL_FRENCH_JA "French - フランス語"
#define OPTION_VAL_SPANISH_JA "Spanish - スペイン語"
#define OPTION_VAL_ITALIAN_JA "Italian - イタリア語"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_JA NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_JA NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_JA NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_JA NULL
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_JA NULL
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_JA NULL
#define CORE_OPTION_NAME_FORCE_WINCE_LABEL_JA NULL
#define CORE_OPTION_NAME_FORCE_WINCE_INFO_0_JA NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_JA NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_JA NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_JA NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_JA NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_JA NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_JA NULL
#define CORE_OPTION_NAME_UPNP_LABEL_JA NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_JA NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_JA NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_JA NULL
#define OPTION_VAL_320X240_JA NULL
#define OPTION_VAL_640X480_JA NULL
#define OPTION_VAL_800X600_JA NULL
#define OPTION_VAL_960X720_JA NULL
#define OPTION_VAL_1024X768_JA NULL
#define OPTION_VAL_1280X960_JA NULL
#define OPTION_VAL_1440X1080_JA NULL
#define OPTION_VAL_1600X1200_JA NULL
#define OPTION_VAL_1920X1440_JA NULL
#define OPTION_VAL_2560X1920_JA NULL
#define OPTION_VAL_2880X2160_JA NULL
#define OPTION_VAL_3200X2400_JA NULL
#define OPTION_VAL_3840X2880_JA NULL
#define OPTION_VAL_4480X3360_JA NULL
#define OPTION_VAL_5120X3840_JA NULL
#define OPTION_VAL_5760X4320_JA NULL
#define OPTION_VAL_6400X4800_JA NULL
#define OPTION_VAL_7040X5280_JA NULL
#define OPTION_VAL_7680X5760_JA NULL
#define OPTION_VAL_8320X6240_JA NULL
#define OPTION_VAL_8960X6720_JA NULL
#define OPTION_VAL_9600X7200_JA NULL
#define OPTION_VAL_10240X7680_JA NULL
#define OPTION_VAL_10880X8160_JA NULL
#define OPTION_VAL_11520X8640_JA NULL
#define OPTION_VAL_12160X9120_JA NULL
#define OPTION_VAL_12800X9600_JA NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_JA NULL
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_JA NULL
#define OPTION_VAL_VGA_JA NULL
#define OPTION_VAL_TV_RGB_JA NULL
#define OPTION_VAL_TV_COMPOSITE_JA NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_JA NULL
#define OPTION_VAL_NTSC_JA NULL
#define OPTION_VAL_PAL_JA NULL
#define OPTION_VAL_PAL_N_JA NULL
#define OPTION_VAL_PAL_M_JA NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_JA "スクリーンの回転"
#define OPTION_VAL_HORIZONTAL_JA NULL
#define OPTION_VAL_VERTICAL_JA NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_JA NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_JA NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_JA NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_JA NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_JA NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_JA NULL
#define OPTION_VAL_512MB_JA NULL
#define OPTION_VAL_1GB_JA NULL
#define OPTION_VAL_2GB_JA NULL
#define OPTION_VAL_4GB_JA NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_JA NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_JA NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_JA NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_JA NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_JA NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_JA NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_JA NULL
#define CORE_OPTION_NAME_FOG_LABEL_JA NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_JA NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_JA NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_JA NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_JA NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_JA NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_JA NULL
#define OPTION_VAL_1_JA NULL
#define OPTION_VAL_2_JA NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_JA NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_JA NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_JA NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_JA NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_JA NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_JA NULL
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_JA NULL
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_JA NULL
#define OPTION_VAL_2_O27_JA NULL
#define OPTION_VAL_4_JA NULL
#define OPTION_VAL_6_JA NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_JA NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_JA NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_JA NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_JA NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_JA NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_JA NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_JA NULL
#define OPTION_VAL_SOME_JA "普通"
#define OPTION_VAL_MORE_JA NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_JA NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_JA NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_JA NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_JA NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_JA NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_JA NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_JA NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_JA NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_JA NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_JA NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_JA NULL
#define OPTION_VAL_0_JA NULL
#define OPTION_VAL_5_JA NULL
#define OPTION_VAL_10_JA NULL
#define OPTION_VAL_15_JA NULL
#define OPTION_VAL_20_JA NULL
#define OPTION_VAL_25_JA NULL
#define OPTION_VAL_30_JA NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_JA NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_JA NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_JA NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_JA NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_JA NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_JA NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_JA NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_JA NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_JA NULL
#define OPTION_VAL_WHITE_JA NULL
#define OPTION_VAL_RED_JA "レッド"
#define OPTION_VAL_GREEN_JA "グリーン"
#define OPTION_VAL_BLUE_JA "ブルー"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_JA NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_JA NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_JA NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_JA NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_JA NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_JA NULL
#define OPTION_VAL_VMU_A1_JA NULL
#define OPTION_VAL_ALL_VMUS_JA NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_JA NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_JA NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_JA NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_JA NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_JA NULL
#define OPTION_VAL_UPPER_LEFT_JA NULL
#define OPTION_VAL_UPPER_RIGHT_JA NULL
#define OPTION_VAL_LOWER_LEFT_JA NULL
#define OPTION_VAL_LOWER_RIGHT_JA NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_JA NULL
#define OPTION_VAL_1X_JA NULL
#define OPTION_VAL_3X_JA NULL
#define OPTION_VAL_5X_JA NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_JA NULL
#define OPTION_VAL_DEFAULT_ON_00_JA NULL
#define OPTION_VAL_DEFAULT_OFF_01_JA NULL
#define OPTION_VAL_BLACK_02_JA NULL
#define OPTION_VAL_LIGHT_BLUE_04_JA NULL
#define OPTION_VAL_CYAN_06_JA NULL
#define OPTION_VAL_CYAN_BLUE_07_JA NULL
#define OPTION_VAL_LIGHT_GREEN_08_JA NULL
#define OPTION_VAL_CYAN_GREEN_09_JA NULL
#define OPTION_VAL_LIGHT_CYAN_10_JA NULL
#define OPTION_VAL_PURPLE_12_JA NULL
#define OPTION_VAL_LIGHT_PURPLE_13_JA NULL
#define OPTION_VAL_YELLOW_14_JA "イエロー"
#define OPTION_VAL_GRAY_15_JA NULL
#define OPTION_VAL_LIGHT_PURPLE_2_16_JA NULL
#define OPTION_VAL_LIGHT_GREEN_2_17_JA NULL
#define OPTION_VAL_LIGHT_GREEN_3_18_JA NULL
#define OPTION_VAL_LIGHT_CYAN_2_19_JA NULL
#define OPTION_VAL_LIGHT_RED_2_20_JA NULL
#define OPTION_VAL_MAGENTA_21_JA NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_JA NULL
#define OPTION_VAL_LIGHT_ORANGE_23_JA NULL
#define OPTION_VAL_ORANGE_24_JA NULL
#define OPTION_VAL_LIGHT_PURPLE_4_25_JA NULL
#define OPTION_VAL_LIGHT_YELLOW_26_JA NULL
#define OPTION_VAL_LIGHT_YELLOW_2_27_JA NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_JA NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_JA NULL
#define OPTION_VAL_40_JA NULL
#define OPTION_VAL_50_JA NULL
#define OPTION_VAL_60_JA NULL
#define OPTION_VAL_70_JA NULL
#define OPTION_VAL_80_JA NULL
#define OPTION_VAL_90_JA NULL
#define OPTION_VAL_100_JA NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_JA NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_JA NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_JA NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_JA NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_JA NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_JA NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_JA NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_JA NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_JA NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_JA NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_JA NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_JA NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_JA NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_JA NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_JA NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_JA NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_JA NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_JA NULL

struct retro_core_option_v2_category option_cats_ja[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_JA,
      CATEGORY_SYSTEM_INFO_0_JA
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_JA,
      CATEGORY_VIDEO_INFO_0_JA
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_JA,
      CATEGORY_PERFORMANCE_INFO_0_JA
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_JA,
      CATEGORY_HACKS_INFO_0_JA
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_JA,
      CATEGORY_INPUT_INFO_0_JA
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_JA,
      CATEGORY_VMU_INFO_0_JA
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_ja[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_JA,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_JA },
         { "USA",     OPTION_VAL_USA_JA },
         { "Europe",  OPTION_VAL_EUROPE_JA },
         { "Default", OPTION_VAL_DEFAULT_JA },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_JA,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_JA },
         { "English",  OPTION_VAL_ENGLISH_JA },
         { "German",   OPTION_VAL_GERMAN_JA },
         { "French",   OPTION_VAL_FRENCH_JA },
         { "Spanish",  OPTION_VAL_SPANISH_JA },
         { "Italian",  OPTION_VAL_ITALIAN_JA },
         { "Default",  OPTION_VAL_DEFAULT_JA },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_JA,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_boot_to_bios",
      CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_JA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_JA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_force_wince",
      CORE_OPTION_NAME_FORCE_WINCE_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_FORCE_WINCE_INFO_0_JA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_JA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_JA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_JA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_JA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_JA,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_JA },
         { "640x480",    OPTION_VAL_640X480_JA },
         { "800x600",    OPTION_VAL_800X600_JA },
         { "960x720",    OPTION_VAL_960X720_JA },
         { "1024x768",   OPTION_VAL_1024X768_JA },
         { "1280x960",   OPTION_VAL_1280X960_JA },
         { "1440x1080",  OPTION_VAL_1440X1080_JA },
         { "1600x1200",  OPTION_VAL_1600X1200_JA },
         { "1920x1440",  OPTION_VAL_1920X1440_JA },
         { "2560x1920",  OPTION_VAL_2560X1920_JA },
         { "2880x2160",  OPTION_VAL_2880X2160_JA },
         { "3200x2400",  OPTION_VAL_3200X2400_JA },
         { "3840x2880",  OPTION_VAL_3840X2880_JA },
         { "4480x3360",  OPTION_VAL_4480X3360_JA },
         { "5120x3840",  OPTION_VAL_5120X3840_JA },
         { "5760x4320",  OPTION_VAL_5760X4320_JA },
         { "6400x4800",  OPTION_VAL_6400X4800_JA },
         { "7040x5280",  OPTION_VAL_7040X5280_JA },
         { "7680x5760",  OPTION_VAL_7680X5760_JA },
         { "8320x6240",  OPTION_VAL_8320X6240_JA },
         { "8960x6720",  OPTION_VAL_8960X6720_JA },
         { "9600x7200",  OPTION_VAL_9600X7200_JA },
         { "10240x7680", OPTION_VAL_10240X7680_JA },
         { "10880x8160", OPTION_VAL_10880X8160_JA },
         { "11520x8640", OPTION_VAL_11520X8640_JA },
         { "12160x9120", OPTION_VAL_12160X9120_JA },
         { "12800x9600", OPTION_VAL_12800X9600_JA },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_JA,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_JA },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_JA },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_JA },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_JA,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_JA },
         { "PAL",     OPTION_VAL_PAL_JA },
         { "PAL_N",   OPTION_VAL_PAL_N_JA },
         { "PAL_M",   OPTION_VAL_PAL_M_JA },
         { "Default", OPTION_VAL_DEFAULT_JA },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_JA,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_JA },
         { "vertical",   OPTION_VAL_VERTICAL_JA },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_JA,
      NULL,
      "",
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_JA },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_JA },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_JA },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_JA,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_JA },
         { "1GB",   OPTION_VAL_1GB_JA },
         { "2GB",   OPTION_VAL_2GB_JA },
         { "4GB",   OPTION_VAL_4GB_JA },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_JA,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_JA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_JA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_JA,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_JA,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_JA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_JA,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_JA,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_JA },
         { "1",  OPTION_VAL_1_JA },
         { "2",  OPTION_VAL_2_JA },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_JA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_JA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_JA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_JA,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O27_JA },
         { "4",  OPTION_VAL_4_JA },
         { "6",  OPTION_VAL_6_JA },
         { NULL, NULL },
      },
      "1",
   },
   {/* TODO: needs clarification */
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_JA,
      NULL,
      "",
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_JA,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_JA,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_JA,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_JA,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_JA },
         { "more", OPTION_VAL_MORE_JA },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_JA,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_JA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_JA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_JA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_JA,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: probably needs explanation */
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_JA,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_JA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_JA },
         { "5%",  OPTION_VAL_5_JA },
         { "10%", OPTION_VAL_10_JA },
         { "15%", OPTION_VAL_15_JA },
         { "20%", OPTION_VAL_20_JA },
         { "25%", OPTION_VAL_25_JA },
         { "30%", OPTION_VAL_30_JA },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_JA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_JA },
         { "5%",  OPTION_VAL_5_JA },
         { "10%", OPTION_VAL_10_JA },
         { "15%", OPTION_VAL_15_JA },
         { "20%", OPTION_VAL_20_JA },
         { "25%", OPTION_VAL_25_JA },
         { "30%", OPTION_VAL_30_JA },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_JA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_purupuru",
      CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_JA,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_JA,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_JA,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_JA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_JA },
         { "Red",      OPTION_VAL_RED_JA },
         { "Green",    OPTION_VAL_GREEN_JA },
         { "Blue",     OPTION_VAL_BLUE_JA },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_JA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_JA },
         { "Red",      OPTION_VAL_RED_JA },
         { "Green",    OPTION_VAL_GREEN_JA },
         { "Blue",     OPTION_VAL_BLUE_JA },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_JA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_JA },
         { "Red",      OPTION_VAL_RED_JA },
         { "Green",    OPTION_VAL_GREEN_JA },
         { "Blue",     OPTION_VAL_BLUE_JA },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_JA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_JA },
         { "Red",      OPTION_VAL_RED_JA },
         { "Green",    OPTION_VAL_GREEN_JA },
         { "Blue",     OPTION_VAL_BLUE_JA },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_JA,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_JA,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_JA,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_JA },
         { "All VMUs", OPTION_VAL_ALL_VMUS_JA },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_JA,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_JA,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_JA,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_JA },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_JA },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_JA },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_JA },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_JA },
         { "2x", OPTION_VAL_2_O27_JA },
         { "3x", OPTION_VAL_3X_JA },
         { "4x", OPTION_VAL_4_JA },
         { "5x", OPTION_VAL_5X_JA },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_JA },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_JA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_JA },
         { "BLUE 03",           OPTION_VAL_BLUE_JA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_JA },
         { "GREEN 05",          OPTION_VAL_GREEN_JA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_JA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_JA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_JA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_JA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_JA },
         { "RED 11",            OPTION_VAL_RED_JA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_JA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_JA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_JA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_JA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_JA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_JA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_JA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_JA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_JA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_JA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_JA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_JA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_JA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_JA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_JA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_JA },
         { "WHITE 28",          OPTION_VAL_WHITE_JA },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_JA },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_JA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_JA },
         { "BLUE 03",           OPTION_VAL_BLUE_JA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_JA },
         { "GREEN 05",          OPTION_VAL_GREEN_JA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_JA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_JA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_JA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_JA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_JA },
         { "RED 11",            OPTION_VAL_RED_JA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_JA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_JA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_JA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_JA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_JA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_JA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_JA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_JA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_JA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_JA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_JA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_JA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_JA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_JA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_JA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_JA },
         { "WHITE 28",          OPTION_VAL_WHITE_JA },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_JA },
         { "20%",  OPTION_VAL_20_JA },
         { "30%",  OPTION_VAL_30_JA },
         { "40%",  OPTION_VAL_40_JA },
         { "50%",  OPTION_VAL_50_JA },
         { "60%",  OPTION_VAL_60_JA },
         { "70%",  OPTION_VAL_70_JA },
         { "80%",  OPTION_VAL_80_JA },
         { "90%",  OPTION_VAL_90_JA },
         { "100%", OPTION_VAL_100_JA },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_JA },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_JA },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_JA },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_JA },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_JA },
         { "2x", OPTION_VAL_2_O27_JA },
         { "3x", OPTION_VAL_3X_JA },
         { "4x", OPTION_VAL_4_JA },
         { "5x", OPTION_VAL_5X_JA },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_JA },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_JA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_JA },
         { "BLUE 03",           OPTION_VAL_BLUE_JA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_JA },
         { "GREEN 05",          OPTION_VAL_GREEN_JA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_JA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_JA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_JA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_JA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_JA },
         { "RED 11",            OPTION_VAL_RED_JA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_JA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_JA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_JA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_JA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_JA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_JA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_JA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_JA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_JA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_JA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_JA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_JA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_JA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_JA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_JA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_JA },
         { "WHITE 28",          OPTION_VAL_WHITE_JA },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_JA },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_JA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_JA },
         { "BLUE 03",           OPTION_VAL_BLUE_JA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_JA },
         { "GREEN 05",          OPTION_VAL_GREEN_JA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_JA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_JA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_JA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_JA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_JA },
         { "RED 11",            OPTION_VAL_RED_JA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_JA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_JA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_JA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_JA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_JA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_JA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_JA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_JA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_JA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_JA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_JA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_JA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_JA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_JA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_JA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_JA },
         { "WHITE 28",          OPTION_VAL_WHITE_JA },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_JA },
         { "20%",  OPTION_VAL_20_JA },
         { "30%",  OPTION_VAL_30_JA },
         { "40%",  OPTION_VAL_40_JA },
         { "50%",  OPTION_VAL_50_JA },
         { "60%",  OPTION_VAL_60_JA },
         { "70%",  OPTION_VAL_70_JA },
         { "80%",  OPTION_VAL_80_JA },
         { "90%",  OPTION_VAL_90_JA },
         { "100%", OPTION_VAL_100_JA },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_JA },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_JA },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_JA },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_JA },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_JA },
         { "2x", OPTION_VAL_2_O27_JA },
         { "3x", OPTION_VAL_3X_JA },
         { "4x", OPTION_VAL_4_JA },
         { "5x", OPTION_VAL_5X_JA },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_JA },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_JA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_JA },
         { "BLUE 03",           OPTION_VAL_BLUE_JA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_JA },
         { "GREEN 05",          OPTION_VAL_GREEN_JA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_JA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_JA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_JA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_JA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_JA },
         { "RED 11",            OPTION_VAL_RED_JA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_JA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_JA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_JA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_JA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_JA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_JA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_JA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_JA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_JA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_JA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_JA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_JA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_JA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_JA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_JA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_JA },
         { "WHITE 28",          OPTION_VAL_WHITE_JA },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_JA },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_JA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_JA },
         { "BLUE 03",           OPTION_VAL_BLUE_JA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_JA },
         { "GREEN 05",          OPTION_VAL_GREEN_JA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_JA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_JA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_JA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_JA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_JA },
         { "RED 11",            OPTION_VAL_RED_JA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_JA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_JA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_JA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_JA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_JA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_JA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_JA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_JA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_JA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_JA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_JA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_JA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_JA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_JA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_JA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_JA },
         { "WHITE 28",          OPTION_VAL_WHITE_JA },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_JA },
         { "20%",  OPTION_VAL_20_JA },
         { "30%",  OPTION_VAL_30_JA },
         { "40%",  OPTION_VAL_40_JA },
         { "50%",  OPTION_VAL_50_JA },
         { "60%",  OPTION_VAL_60_JA },
         { "70%",  OPTION_VAL_70_JA },
         { "80%",  OPTION_VAL_80_JA },
         { "90%",  OPTION_VAL_90_JA },
         { "100%", OPTION_VAL_100_JA },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_JA },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_JA },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_JA },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_JA },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_JA },
         { "2x", OPTION_VAL_2_O27_JA },
         { "3x", OPTION_VAL_3X_JA },
         { "4x", OPTION_VAL_4_JA },
         { "5x", OPTION_VAL_5X_JA },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_JA },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_JA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_JA },
         { "BLUE 03",           OPTION_VAL_BLUE_JA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_JA },
         { "GREEN 05",          OPTION_VAL_GREEN_JA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_JA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_JA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_JA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_JA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_JA },
         { "RED 11",            OPTION_VAL_RED_JA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_JA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_JA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_JA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_JA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_JA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_JA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_JA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_JA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_JA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_JA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_JA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_JA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_JA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_JA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_JA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_JA },
         { "WHITE 28",          OPTION_VAL_WHITE_JA },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_JA },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_JA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_JA },
         { "BLUE 03",           OPTION_VAL_BLUE_JA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_JA },
         { "GREEN 05",          OPTION_VAL_GREEN_JA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_JA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_JA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_JA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_JA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_JA },
         { "RED 11",            OPTION_VAL_RED_JA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_JA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_JA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_JA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_JA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_JA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_JA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_JA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_JA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_JA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_JA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_JA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_JA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_JA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_JA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_JA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_JA },
         { "WHITE 28",          OPTION_VAL_WHITE_JA },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_JA },
         { "20%",  OPTION_VAL_20_JA },
         { "30%",  OPTION_VAL_30_JA },
         { "40%",  OPTION_VAL_40_JA },
         { "50%",  OPTION_VAL_50_JA },
         { "60%",  OPTION_VAL_60_JA },
         { "70%",  OPTION_VAL_70_JA },
         { "80%",  OPTION_VAL_80_JA },
         { "90%",  OPTION_VAL_90_JA },
         { "100%", OPTION_VAL_100_JA },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_ja = {
   option_cats_ja,
   option_defs_ja
};

/* RETRO_LANGUAGE_KO */

#define CATEGORY_SYSTEM_LABEL_KO "시스템"
#define CATEGORY_SYSTEM_INFO_0_KO "지역, 언어, BIOS 및 하드웨어 설정을 구성합니다."
#define CATEGORY_VIDEO_LABEL_KO "비디오"
#define CATEGORY_VIDEO_INFO_0_KO "해상도, 투명도 및 화면 효과 설정을 구성합니다."
#define CATEGORY_PERFORMANCE_LABEL_KO "성능"
#define CATEGORY_PERFORMANCE_INFO_0_KO "스레드 렌더링 및 프레임 스킵 설정을 구성합니다."
#define CATEGORY_HACKS_LABEL_KO "에뮬레이션 핵"
#define CATEGORY_HACKS_INFO_0_KO "와이드스크린 핵, GD롬 로딩 속도 조절 및 텍스쳐 변경 설정등을 구성합니다."
#define CATEGORY_INPUT_LABEL_KO "입력"
#define CATEGORY_INPUT_INFO_0_KO "게임패드 및 라이트 건 설정을 구성합니다."
#define CATEGORY_VMU_LABEL_KO "비쥬얼 메모리 유닛"
#define CATEGORY_VMU_INFO_0_KO "게임 별 VMU 저장 파일 및 VMU OSD 설정을 구성합니다."
#define CORE_OPTION_NAME_REGION_LABEL_KO "지역"
#define OPTION_VAL_JAPAN_KO "일본"
#define OPTION_VAL_USA_KO "미국"
#define OPTION_VAL_EUROPE_KO "유럽"
#define OPTION_VAL_DEFAULT_KO "기본"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_KO "언어"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_KO "BIOS 및 다국어 게임에서 사용되는 언어 설정을 변경합니다"
#define OPTION_VAL_JAPANESE_KO "일본어"
#define OPTION_VAL_ENGLISH_KO "영어"
#define OPTION_VAL_GERMAN_KO "독일어"
#define OPTION_VAL_FRENCH_KO "프랑스어"
#define OPTION_VAL_SPANISH_KO "스페인어"
#define OPTION_VAL_ITALIAN_KO "이탈리아어"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_KO "HLE BIOS (재시작 필요)"
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_KO "HLE BIOS 사용을 강제합니다."
#define CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_KO "BIOS 로 부트 (재시작 필요)"
#define CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_KO "드림캐스트 BIOS 메뉴로 부팅합니다."
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_KO "HDR 활성화"
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_KO "Dreamcast의 오디오 DSP(디지털 신호 프로세서)의 에뮬레이션을 활성화합니다. 사운드의 정확도는 향상되지만 성능 요구 사항이 증가합니다."
#define CORE_OPTION_NAME_FORCE_WINCE_LABEL_KO "Windows CE 모드 강제 사용"
#define CORE_OPTION_NAME_FORCE_WINCE_INFO_0_KO "Windows CE 게임을 위한 전체 MMU(메모리 관리 유닛) 에뮬레이션 및 기타 설정을 활성화합니다."
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_KO "NAOMI 서비스 버튼 허용"
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_KO "캐비넷 설정에 들어가기위한 NAOMI 서비스 버튼을 활성화합니다."
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_KO "NAOMI 게임을 무료 플레이로 설정"
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_KO "무료 플레이를 위해 게임의 코인 설정을 수정합니다."
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_KO "브로드밴드 어댑터 에뮬레이션"
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_KO "모뎀 대신 이더넷 브로드밴드 어댑터를 에뮬레이트합니다. (재시작 필요)"
#define CORE_OPTION_NAME_UPNP_LABEL_KO "UPnP 활성화"
#define CORE_OPTION_NAME_UPNP_INFO_0_KO "온라인 게임에서 UPnP를 사용해 자동으로 인터넷 라우터를 설정합니다."
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_KO "내부 해상도"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_KO "렌더링 해상도를 변경합니다."
#define OPTION_VAL_320X240_KO NULL
#define OPTION_VAL_640X480_KO NULL
#define OPTION_VAL_800X600_KO NULL
#define OPTION_VAL_960X720_KO NULL
#define OPTION_VAL_1024X768_KO NULL
#define OPTION_VAL_1280X960_KO NULL
#define OPTION_VAL_1440X1080_KO NULL
#define OPTION_VAL_1600X1200_KO NULL
#define OPTION_VAL_1920X1440_KO NULL
#define OPTION_VAL_2560X1920_KO NULL
#define OPTION_VAL_2880X2160_KO NULL
#define OPTION_VAL_3200X2400_KO NULL
#define OPTION_VAL_3840X2880_KO NULL
#define OPTION_VAL_4480X3360_KO NULL
#define OPTION_VAL_5120X3840_KO NULL
#define OPTION_VAL_5760X4320_KO NULL
#define OPTION_VAL_6400X4800_KO NULL
#define OPTION_VAL_7040X5280_KO NULL
#define OPTION_VAL_7680X5760_KO NULL
#define OPTION_VAL_8320X6240_KO NULL
#define OPTION_VAL_8960X6720_KO NULL
#define OPTION_VAL_9600X7200_KO NULL
#define OPTION_VAL_10240X7680_KO NULL
#define OPTION_VAL_10880X8160_KO NULL
#define OPTION_VAL_11520X8640_KO NULL
#define OPTION_VAL_12160X9120_KO NULL
#define OPTION_VAL_12800X9600_KO NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_KO "케이블 타입"
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_KO "외부 출력 형식. 'TV (컴포지트)'가 가장 많이 지원됨."
#define OPTION_VAL_VGA_KO NULL
#define OPTION_VAL_TV_RGB_KO NULL
#define OPTION_VAL_TV_COMPOSITE_KO "TV (컴포지트)"
#define CORE_OPTION_NAME_BROADCAST_LABEL_KO "방송 표준"
#define OPTION_VAL_NTSC_KO NULL
#define OPTION_VAL_PAL_KO "PAL (국제)"
#define OPTION_VAL_PAL_N_KO "PAL-N (아르헨티나, 파라과이, 우르과이)"
#define OPTION_VAL_PAL_M_KO "PAL-M (브라질)"
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_KO "화면 방향"
#define OPTION_VAL_HORIZONTAL_KO "가로"
#define OPTION_VAL_VERTICAL_KO "세로"
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_KO "알파 정렬"
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_KO "줄당 (빠름, 가장 부정확)"
#define OPTION_VAL_PER_TRIANGLE_NORMAL_KO "삼각형당 (중간)"
#define OPTION_VAL_PER_PIXEL_ACCURATE_KO "픽셀당 (정확, 가장 느림)"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_KO "누산 픽셀 버퍼 크기"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_KO "높은 해상도 출력을 올바르게 표시하려면 이 값을 높은 값으로 설정해야 할 수도 있습니다."
#define OPTION_VAL_512MB_KO NULL
#define OPTION_VAL_1GB_KO NULL
#define OPTION_VAL_2GB_KO NULL
#define OPTION_VAL_4GB_KO NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_KO "최대 투명 레이어"
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_KO "복잡한 장면을 그릴 때 높은 값이 필요할 수 있습니다."
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_KO "전체 프레임버퍼 에뮬레이션"
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_KO "VRAM 상 전체 프레임버퍼 에뮬레이션을 활성화합니다. VRAM 상의 프레임버퍼에 직접 읽기 또는 쓰기를 수행하는 게임을 실행할 때 유용합니다. 활성화할 경우, 내부 해상도가 640x480으로 고정되며 성능에 큰 영향을 끼칠 수 있습니다."
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_KO "RTT (Render To Texture) 버퍼 사용"
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_KO "렌더된 텍스쳐를 GPU에서 VRAM으로 복사합니다. 이 옵션은 이 옵션을 필요로 하는 게임에서는 기본적으로 활성화됩니다. 활성화할 경우, 텍스쳐 업스케일링이 비활성화되며 성능에 큰 영향을 끼칠 수 있습니다."
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_KO "밉맵"
#define CORE_OPTION_NAME_FOG_LABEL_KO "안개 효과"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_KO "볼륨 모디파이어"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_KO "드림캐스트 GPU의 기능으로 보통 개체의 그림자를 그리는 데 사용됩니다. 이 기능은 보통 활성화해두어야 하며 성능에 미치는 영향은 아주 적습니다."
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_KO "비등방성 필터링"
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_KO "카메라와 비스듬하게 배치된 표면의 텍스쳐 품질을 향상합니다. 높은 값으로 설정하면 GPU 부하가 높아집니다. 변경 사항은 코어를 다시 시작해야 적용됩니다."
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_KO "텍스쳐 필터링"
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_KO "사용할 텍스쳐 필터링 모드를 선택합니다. 모든 텍스쳐에 강제로 특정 필터링 모드를 적용하여 더 날카로운(또는 부드러운) 비주얼을 얻을 수 있습니다. 기본 외의 다른 값으로 설정 시 다양한 렌더링 문제가 발생할 수 있습니다. 변경 사항은 코어를 다시 시작해야 적용됩니다."
#define OPTION_VAL_1_KO "강제 최근접이웃"
#define OPTION_VAL_2_KO "강제 선형"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_KO "프레임 스왑 지연"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_KO "화면이 번쩍이거나 깨지는 것을 방지할 수 있습니다. 느린 플랫폼에서는 사용을 권장하지 않습니다."
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_KO "프레임 레이트 변화 감지"
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_KO "내부 프레임 레이트가 변경될 때(예: 60FPS에서 30FPS로) 프론트엔드에게 알림을 보냅니다. 30FPS 또는 20FPS 고정으로 실행되는 게임에서는 프레임 페이스를 개선하지만, 무제한(불안정) 프레임 레이트로 실행되는 게임(예: Ecco the Dolphin, Unreal Tournament)을 실행할 때는 비활성화되어야 합니다. 주의: '자동 프레임 건너뛰기'가 활성화되어있으면 작동하지 않습니다."
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_KO "PowerVR2 포스트프로세싱 필터"
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_KO "렌더된 이미지에 PowerVR2 GPU와 아날로그 비디오 신호에서 나타나는 효과를 후처리로 구현합니다."
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_KO "텍스쳐 업스케일링 (xBRZ)"
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_KO "손으로 그려진 2D 픽셀아트 그래픽을 개선합니다. 2D 픽셀아트 게임에서만 사용되어야 합니다."
#define OPTION_VAL_2_O27_KO NULL
#define OPTION_VAL_4_KO NULL
#define OPTION_VAL_6_KO NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_KO "텍스쳐 업스케일링 필터 최대 크기"
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_KO "네이티브 깊이 보간"
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_KO "AMD GPU에서 깊이 문제와 텍스쳐 망가짐 증상을 개선하는 데 도움이 됩니다. 일부 상황에서는 인텔 GPU에서도 도움이 될 수 있습니다."
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_KO "스레드 렌더링"
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_KO "GPU와 CPU를 별도의 스레드에서 실행합니다. 사용하는 것이 매우 권장됩니다."
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_KO "자동 프레임 건너뛰기"
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_KO "에뮬레이터 실행이 느려질 경우 자동으로 프레임을 건너뜁니다. 주의: 이 설정은 '스레드 렌더링'이 활성화되어 있어야 적용됩니다."
#define OPTION_VAL_SOME_KO "보통"
#define OPTION_VAL_MORE_KO "최대"
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_KO "프레임 건너뛰기"
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_KO "표시되는 각 프레임 사이에서 건너뛸 프레임 수를 설정합니다."
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_KO "와이드스크린 핵 (재시작 필요)"
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_KO "일부 게임이 와이드스크린 형식으로 출력되게 하는 치트를 활성화합니다."
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_KO "와이드스크린 핵"
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_KO "일반적인 4:3 화면비 밖에도 개체를 그립니다. 드러난 부분에서 그래픽 오류가 발생할 수도 있습니다."
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_KO "GD-ROM 빠른 로딩 (정확하지 않음)"
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_KO "GD-ROM 로딩 속도를 높입니다."
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_KO "사용자 텍스처 불러오기"
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_KO "텍스쳐 덤프하기"
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_KO "아날로그 스틱 데드존"
#define OPTION_VAL_0_KO NULL
#define OPTION_VAL_5_KO NULL
#define OPTION_VAL_10_KO NULL
#define OPTION_VAL_15_KO NULL
#define OPTION_VAL_20_KO NULL
#define OPTION_VAL_25_KO NULL
#define OPTION_VAL_30_KO NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_KO "트리거 데드존"
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_KO "디지털 트리거"
#define CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_KO "부루부루 팩/진동 팩"
#define CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_KO "컨트롤러 포스 피드백을 활성합니다."
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_KO "디지털 출력 송출하기"
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_KO "디지털 출력과 포스 피드백 상태를 TCP 포트 8000번으로 송출합니다. MAME의 \"-output network\" 옵션과 호환됩니다."
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_KO "라이트 건 설정 표시"
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_KO "라이트건 십자 조준선 옵션 구성을 활성화합니다. 참고: 이 설정을 적용하려면 빠른 메뉴를 다시 표시해야 할 수 있습니다."
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_KO "라이트건 조준선 1 표시"
#define OPTION_VAL_WHITE_KO "흰색"
#define OPTION_VAL_RED_KO "빨강"
#define OPTION_VAL_GREEN_KO "초록"
#define OPTION_VAL_BLUE_KO "파랑"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_KO "라이트건 조준선 2 표시"
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_KO "라이트건 조준선 3 표시"
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_KO "라이트건 조준선 4 표시"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_KO "게임당 별도 Visual Memory Units/Systems(VMU) 사용"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_KO "게임당 별도 VMU 사용"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_KO "비활성화할 경우, 모든 게임이 RetroArch의 시스템 디렉토리에 저장된 4개의 VMU 저장 파일(A1, B1, C1, D1)을 공유합니다. 'VMU A1' 설정은 RetroArch의 세이브 디렉토리에 게임 하나 당 하나의 VMU 'A1' 파일을 생성합니다. '모든 VMU'는 게임 하나 당 4개의 VMU 파일(A1, B1, C1, D1)을 생성합니다."
#define OPTION_VAL_VMU_A1_KO NULL
#define OPTION_VAL_ALL_VMUS_KO "전체 VMU"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_KO "Visual Memory Unit/System(VMU) 디스플레이 설정 표시"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_KO "VMU 디스플레이 설정 표시"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_KO "에뮬레이트되는 VMU LCD 화면의 표시 여부, 크기, 위치, 색상 설정을 표시합니다. 주의: 변경 사항이 적용되려면 빠른 메뉴를 껐다 켜야 할 수도 있습니다."
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_KO "VMU 화면 1 표시"
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_KO "VMU 화면 1 위치"
#define OPTION_VAL_UPPER_LEFT_KO "왼쪽 위"
#define OPTION_VAL_UPPER_RIGHT_KO "오른쪽 위"
#define OPTION_VAL_LOWER_LEFT_KO "왼쪽 아래"
#define OPTION_VAL_LOWER_RIGHT_KO "오른쪽 아래"
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_KO "VMU 화면 1 크기"
#define OPTION_VAL_1X_KO NULL
#define OPTION_VAL_3X_KO NULL
#define OPTION_VAL_5X_KO NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_KO "VMU 화면 1 켜진 픽셀 색상"
#define OPTION_VAL_DEFAULT_ON_00_KO "기본설정 켜짐"
#define OPTION_VAL_DEFAULT_OFF_01_KO "기본설정 꺼짐"
#define OPTION_VAL_BLACK_02_KO "검정"
#define OPTION_VAL_LIGHT_BLUE_04_KO "밝은 파랑"
#define OPTION_VAL_CYAN_06_KO "청록"
#define OPTION_VAL_CYAN_BLUE_07_KO "청록 파랑"
#define OPTION_VAL_LIGHT_GREEN_08_KO "밝은 녹색"
#define OPTION_VAL_CYAN_GREEN_09_KO "청록 초록"
#define OPTION_VAL_LIGHT_CYAN_10_KO "밝은 청록"
#define OPTION_VAL_PURPLE_12_KO "보라"
#define OPTION_VAL_LIGHT_PURPLE_13_KO "밝은 보라"
#define OPTION_VAL_YELLOW_14_KO "노란색"
#define OPTION_VAL_GRAY_15_KO "회색"
#define OPTION_VAL_LIGHT_PURPLE_2_16_KO "연보라색 (2)"
#define OPTION_VAL_LIGHT_GREEN_2_17_KO "밝은 녹색 (2)"
#define OPTION_VAL_LIGHT_GREEN_3_18_KO "밝은 녹색 (3)"
#define OPTION_VAL_LIGHT_CYAN_2_19_KO "밝은 청록 (2)"
#define OPTION_VAL_LIGHT_RED_2_20_KO "밝은 빨강 (2)"
#define OPTION_VAL_MAGENTA_21_KO "자홍색"
#define OPTION_VAL_LIGHT_PURPLE_3_22_KO "연보라색 (3)"
#define OPTION_VAL_LIGHT_ORANGE_23_KO "밝은 주황색"
#define OPTION_VAL_ORANGE_24_KO "오렌지"
#define OPTION_VAL_LIGHT_PURPLE_4_25_KO "연보라색 (4)"
#define OPTION_VAL_LIGHT_YELLOW_26_KO "연한 노란색"
#define OPTION_VAL_LIGHT_YELLOW_2_27_KO "연한 노란색 (2)"
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_KO "VMU 화면 1 꺼진 픽셀 색상"
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_KO "VMU 화면 1 투명도"
#define OPTION_VAL_40_KO NULL
#define OPTION_VAL_50_KO NULL
#define OPTION_VAL_60_KO NULL
#define OPTION_VAL_70_KO NULL
#define OPTION_VAL_80_KO NULL
#define OPTION_VAL_90_KO NULL
#define OPTION_VAL_100_KO NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_KO "VMU 화면 2 표시"
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_KO "VMU 화면 2 위치"
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_KO "VMU 화면 2 크기"
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_KO "VMU 화면 2 켜진 픽셀 색상"
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_KO "VMU 화면 2 꺼진 픽셀 색상"
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_KO "VMU 화면 2 투명도"
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_KO "VMU 화면 3 표시"
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_KO "VMU 화면 3 위치"
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_KO "VMU 화면 3 크기"
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_KO "VMU 화면 3 켜진 픽셀 색상"
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_KO "VMU 화면 3 꺼진 픽셀 색상"
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_KO "VMU 화면 3 투명도"
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_KO "VMU 화면 4 표시"
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_KO "VMU 화면 4 위치"
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_KO "VMU 화면 4 크기"
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_KO "VMU 화면 4 켜진 픽셀 색상"
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_KO "VMU 화면 4 꺼진 픽셀 색상"
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_KO "VMU 화면 4 투명도"

struct retro_core_option_v2_category option_cats_ko[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_KO,
      CATEGORY_SYSTEM_INFO_0_KO
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_KO,
      CATEGORY_VIDEO_INFO_0_KO
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_KO,
      CATEGORY_PERFORMANCE_INFO_0_KO
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_KO,
      CATEGORY_HACKS_INFO_0_KO
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_KO,
      CATEGORY_INPUT_INFO_0_KO
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_KO,
      CATEGORY_VMU_INFO_0_KO
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_ko[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_KO,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_KO },
         { "USA",     OPTION_VAL_USA_KO },
         { "Europe",  OPTION_VAL_EUROPE_KO },
         { "Default", OPTION_VAL_DEFAULT_KO },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_KO,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_KO },
         { "English",  OPTION_VAL_ENGLISH_KO },
         { "German",   OPTION_VAL_GERMAN_KO },
         { "French",   OPTION_VAL_FRENCH_KO },
         { "Spanish",  OPTION_VAL_SPANISH_KO },
         { "Italian",  OPTION_VAL_ITALIAN_KO },
         { "Default",  OPTION_VAL_DEFAULT_KO },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_KO,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_boot_to_bios",
      CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_KO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_KO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_force_wince",
      CORE_OPTION_NAME_FORCE_WINCE_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_FORCE_WINCE_INFO_0_KO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_KO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_KO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_KO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_KO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_KO,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_KO },
         { "640x480",    OPTION_VAL_640X480_KO },
         { "800x600",    OPTION_VAL_800X600_KO },
         { "960x720",    OPTION_VAL_960X720_KO },
         { "1024x768",   OPTION_VAL_1024X768_KO },
         { "1280x960",   OPTION_VAL_1280X960_KO },
         { "1440x1080",  OPTION_VAL_1440X1080_KO },
         { "1600x1200",  OPTION_VAL_1600X1200_KO },
         { "1920x1440",  OPTION_VAL_1920X1440_KO },
         { "2560x1920",  OPTION_VAL_2560X1920_KO },
         { "2880x2160",  OPTION_VAL_2880X2160_KO },
         { "3200x2400",  OPTION_VAL_3200X2400_KO },
         { "3840x2880",  OPTION_VAL_3840X2880_KO },
         { "4480x3360",  OPTION_VAL_4480X3360_KO },
         { "5120x3840",  OPTION_VAL_5120X3840_KO },
         { "5760x4320",  OPTION_VAL_5760X4320_KO },
         { "6400x4800",  OPTION_VAL_6400X4800_KO },
         { "7040x5280",  OPTION_VAL_7040X5280_KO },
         { "7680x5760",  OPTION_VAL_7680X5760_KO },
         { "8320x6240",  OPTION_VAL_8320X6240_KO },
         { "8960x6720",  OPTION_VAL_8960X6720_KO },
         { "9600x7200",  OPTION_VAL_9600X7200_KO },
         { "10240x7680", OPTION_VAL_10240X7680_KO },
         { "10880x8160", OPTION_VAL_10880X8160_KO },
         { "11520x8640", OPTION_VAL_11520X8640_KO },
         { "12160x9120", OPTION_VAL_12160X9120_KO },
         { "12800x9600", OPTION_VAL_12800X9600_KO },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_KO,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_KO },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_KO },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_KO },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_KO,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_KO },
         { "PAL",     OPTION_VAL_PAL_KO },
         { "PAL_N",   OPTION_VAL_PAL_N_KO },
         { "PAL_M",   OPTION_VAL_PAL_M_KO },
         { "Default", OPTION_VAL_DEFAULT_KO },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_KO,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_KO },
         { "vertical",   OPTION_VAL_VERTICAL_KO },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_KO,
      NULL,
      "",
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_KO },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_KO },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_KO },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_KO,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_KO },
         { "1GB",   OPTION_VAL_1GB_KO },
         { "2GB",   OPTION_VAL_2GB_KO },
         { "4GB",   OPTION_VAL_4GB_KO },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_KO,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_KO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_KO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_KO,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_KO,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_KO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_KO,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_KO,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_KO },
         { "1",  OPTION_VAL_1_KO },
         { "2",  OPTION_VAL_2_KO },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_KO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_KO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_KO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_KO,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O27_KO },
         { "4",  OPTION_VAL_4_KO },
         { "6",  OPTION_VAL_6_KO },
         { NULL, NULL },
      },
      "1",
   },
   {/* TODO: needs clarification */
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_KO,
      NULL,
      "",
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_KO,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_KO,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_KO,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_KO,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_KO },
         { "more", OPTION_VAL_MORE_KO },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_KO,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_KO,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_KO,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_KO,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_KO,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: probably needs explanation */
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_KO,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_KO,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_KO },
         { "5%",  OPTION_VAL_5_KO },
         { "10%", OPTION_VAL_10_KO },
         { "15%", OPTION_VAL_15_KO },
         { "20%", OPTION_VAL_20_KO },
         { "25%", OPTION_VAL_25_KO },
         { "30%", OPTION_VAL_30_KO },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_KO,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_KO },
         { "5%",  OPTION_VAL_5_KO },
         { "10%", OPTION_VAL_10_KO },
         { "15%", OPTION_VAL_15_KO },
         { "20%", OPTION_VAL_20_KO },
         { "25%", OPTION_VAL_25_KO },
         { "30%", OPTION_VAL_30_KO },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_KO,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_purupuru",
      CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_KO,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_KO,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_KO,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_KO,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_KO },
         { "Red",      OPTION_VAL_RED_KO },
         { "Green",    OPTION_VAL_GREEN_KO },
         { "Blue",     OPTION_VAL_BLUE_KO },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_KO,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_KO },
         { "Red",      OPTION_VAL_RED_KO },
         { "Green",    OPTION_VAL_GREEN_KO },
         { "Blue",     OPTION_VAL_BLUE_KO },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_KO,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_KO },
         { "Red",      OPTION_VAL_RED_KO },
         { "Green",    OPTION_VAL_GREEN_KO },
         { "Blue",     OPTION_VAL_BLUE_KO },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_KO,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_KO },
         { "Red",      OPTION_VAL_RED_KO },
         { "Green",    OPTION_VAL_GREEN_KO },
         { "Blue",     OPTION_VAL_BLUE_KO },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_KO,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_KO,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_KO,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_KO },
         { "All VMUs", OPTION_VAL_ALL_VMUS_KO },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_KO,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_KO,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_KO,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_KO },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_KO },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_KO },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_KO },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_KO },
         { "2x", OPTION_VAL_2_O27_KO },
         { "3x", OPTION_VAL_3X_KO },
         { "4x", OPTION_VAL_4_KO },
         { "5x", OPTION_VAL_5X_KO },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_KO },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_KO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_KO },
         { "BLUE 03",           OPTION_VAL_BLUE_KO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_KO },
         { "GREEN 05",          OPTION_VAL_GREEN_KO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_KO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_KO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_KO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_KO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_KO },
         { "RED 11",            OPTION_VAL_RED_KO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_KO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_KO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_KO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_KO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_KO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_KO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_KO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_KO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_KO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_KO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_KO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_KO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_KO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_KO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_KO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_KO },
         { "WHITE 28",          OPTION_VAL_WHITE_KO },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_KO },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_KO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_KO },
         { "BLUE 03",           OPTION_VAL_BLUE_KO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_KO },
         { "GREEN 05",          OPTION_VAL_GREEN_KO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_KO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_KO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_KO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_KO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_KO },
         { "RED 11",            OPTION_VAL_RED_KO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_KO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_KO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_KO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_KO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_KO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_KO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_KO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_KO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_KO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_KO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_KO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_KO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_KO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_KO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_KO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_KO },
         { "WHITE 28",          OPTION_VAL_WHITE_KO },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_KO },
         { "20%",  OPTION_VAL_20_KO },
         { "30%",  OPTION_VAL_30_KO },
         { "40%",  OPTION_VAL_40_KO },
         { "50%",  OPTION_VAL_50_KO },
         { "60%",  OPTION_VAL_60_KO },
         { "70%",  OPTION_VAL_70_KO },
         { "80%",  OPTION_VAL_80_KO },
         { "90%",  OPTION_VAL_90_KO },
         { "100%", OPTION_VAL_100_KO },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_KO },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_KO },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_KO },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_KO },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_KO },
         { "2x", OPTION_VAL_2_O27_KO },
         { "3x", OPTION_VAL_3X_KO },
         { "4x", OPTION_VAL_4_KO },
         { "5x", OPTION_VAL_5X_KO },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_KO },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_KO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_KO },
         { "BLUE 03",           OPTION_VAL_BLUE_KO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_KO },
         { "GREEN 05",          OPTION_VAL_GREEN_KO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_KO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_KO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_KO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_KO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_KO },
         { "RED 11",            OPTION_VAL_RED_KO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_KO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_KO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_KO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_KO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_KO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_KO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_KO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_KO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_KO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_KO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_KO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_KO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_KO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_KO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_KO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_KO },
         { "WHITE 28",          OPTION_VAL_WHITE_KO },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_KO },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_KO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_KO },
         { "BLUE 03",           OPTION_VAL_BLUE_KO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_KO },
         { "GREEN 05",          OPTION_VAL_GREEN_KO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_KO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_KO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_KO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_KO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_KO },
         { "RED 11",            OPTION_VAL_RED_KO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_KO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_KO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_KO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_KO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_KO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_KO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_KO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_KO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_KO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_KO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_KO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_KO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_KO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_KO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_KO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_KO },
         { "WHITE 28",          OPTION_VAL_WHITE_KO },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_KO },
         { "20%",  OPTION_VAL_20_KO },
         { "30%",  OPTION_VAL_30_KO },
         { "40%",  OPTION_VAL_40_KO },
         { "50%",  OPTION_VAL_50_KO },
         { "60%",  OPTION_VAL_60_KO },
         { "70%",  OPTION_VAL_70_KO },
         { "80%",  OPTION_VAL_80_KO },
         { "90%",  OPTION_VAL_90_KO },
         { "100%", OPTION_VAL_100_KO },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_KO },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_KO },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_KO },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_KO },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_KO },
         { "2x", OPTION_VAL_2_O27_KO },
         { "3x", OPTION_VAL_3X_KO },
         { "4x", OPTION_VAL_4_KO },
         { "5x", OPTION_VAL_5X_KO },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_KO },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_KO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_KO },
         { "BLUE 03",           OPTION_VAL_BLUE_KO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_KO },
         { "GREEN 05",          OPTION_VAL_GREEN_KO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_KO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_KO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_KO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_KO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_KO },
         { "RED 11",            OPTION_VAL_RED_KO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_KO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_KO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_KO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_KO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_KO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_KO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_KO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_KO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_KO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_KO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_KO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_KO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_KO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_KO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_KO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_KO },
         { "WHITE 28",          OPTION_VAL_WHITE_KO },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_KO },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_KO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_KO },
         { "BLUE 03",           OPTION_VAL_BLUE_KO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_KO },
         { "GREEN 05",          OPTION_VAL_GREEN_KO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_KO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_KO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_KO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_KO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_KO },
         { "RED 11",            OPTION_VAL_RED_KO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_KO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_KO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_KO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_KO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_KO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_KO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_KO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_KO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_KO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_KO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_KO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_KO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_KO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_KO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_KO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_KO },
         { "WHITE 28",          OPTION_VAL_WHITE_KO },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_KO },
         { "20%",  OPTION_VAL_20_KO },
         { "30%",  OPTION_VAL_30_KO },
         { "40%",  OPTION_VAL_40_KO },
         { "50%",  OPTION_VAL_50_KO },
         { "60%",  OPTION_VAL_60_KO },
         { "70%",  OPTION_VAL_70_KO },
         { "80%",  OPTION_VAL_80_KO },
         { "90%",  OPTION_VAL_90_KO },
         { "100%", OPTION_VAL_100_KO },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_KO },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_KO },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_KO },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_KO },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_KO },
         { "2x", OPTION_VAL_2_O27_KO },
         { "3x", OPTION_VAL_3X_KO },
         { "4x", OPTION_VAL_4_KO },
         { "5x", OPTION_VAL_5X_KO },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_KO },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_KO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_KO },
         { "BLUE 03",           OPTION_VAL_BLUE_KO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_KO },
         { "GREEN 05",          OPTION_VAL_GREEN_KO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_KO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_KO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_KO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_KO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_KO },
         { "RED 11",            OPTION_VAL_RED_KO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_KO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_KO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_KO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_KO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_KO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_KO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_KO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_KO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_KO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_KO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_KO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_KO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_KO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_KO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_KO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_KO },
         { "WHITE 28",          OPTION_VAL_WHITE_KO },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_KO },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_KO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_KO },
         { "BLUE 03",           OPTION_VAL_BLUE_KO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_KO },
         { "GREEN 05",          OPTION_VAL_GREEN_KO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_KO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_KO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_KO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_KO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_KO },
         { "RED 11",            OPTION_VAL_RED_KO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_KO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_KO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_KO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_KO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_KO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_KO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_KO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_KO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_KO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_KO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_KO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_KO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_KO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_KO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_KO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_KO },
         { "WHITE 28",          OPTION_VAL_WHITE_KO },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_KO },
         { "20%",  OPTION_VAL_20_KO },
         { "30%",  OPTION_VAL_30_KO },
         { "40%",  OPTION_VAL_40_KO },
         { "50%",  OPTION_VAL_50_KO },
         { "60%",  OPTION_VAL_60_KO },
         { "70%",  OPTION_VAL_70_KO },
         { "80%",  OPTION_VAL_80_KO },
         { "90%",  OPTION_VAL_90_KO },
         { "100%", OPTION_VAL_100_KO },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_ko = {
   option_cats_ko,
   option_defs_ko
};

/* RETRO_LANGUAGE_NL */

#define CATEGORY_SYSTEM_LABEL_NL "Systeem"
#define CATEGORY_SYSTEM_INFO_0_NL NULL
#define CATEGORY_VIDEO_LABEL_NL NULL
#define CATEGORY_VIDEO_INFO_0_NL NULL
#define CATEGORY_PERFORMANCE_LABEL_NL NULL
#define CATEGORY_PERFORMANCE_INFO_0_NL NULL
#define CATEGORY_HACKS_LABEL_NL NULL
#define CATEGORY_HACKS_INFO_0_NL NULL
#define CATEGORY_INPUT_LABEL_NL "Invoer"
#define CATEGORY_INPUT_INFO_0_NL NULL
#define CATEGORY_VMU_LABEL_NL NULL
#define CATEGORY_VMU_INFO_0_NL NULL
#define CORE_OPTION_NAME_REGION_LABEL_NL "Regio"
#define OPTION_VAL_JAPAN_NL NULL
#define OPTION_VAL_USA_NL NULL
#define OPTION_VAL_EUROPE_NL NULL
#define OPTION_VAL_DEFAULT_NL NULL
#define CORE_OPTION_NAME_LANGUAGE_LABEL_NL "Taal"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_NL NULL
#define OPTION_VAL_JAPANESE_NL "Japans"
#define OPTION_VAL_ENGLISH_NL "Engels"
#define OPTION_VAL_GERMAN_NL "Duits"
#define OPTION_VAL_FRENCH_NL "Frans"
#define OPTION_VAL_SPANISH_NL "Spaans"
#define OPTION_VAL_ITALIAN_NL "Italiaans"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_NL NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_NL NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_NL NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_NL NULL
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_NL NULL
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_NL NULL
#define CORE_OPTION_NAME_FORCE_WINCE_LABEL_NL NULL
#define CORE_OPTION_NAME_FORCE_WINCE_INFO_0_NL NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_NL NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_NL NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_NL NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_NL NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_NL NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_NL NULL
#define CORE_OPTION_NAME_UPNP_LABEL_NL NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_NL NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_NL NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_NL NULL
#define OPTION_VAL_320X240_NL NULL
#define OPTION_VAL_640X480_NL NULL
#define OPTION_VAL_800X600_NL NULL
#define OPTION_VAL_960X720_NL NULL
#define OPTION_VAL_1024X768_NL NULL
#define OPTION_VAL_1280X960_NL NULL
#define OPTION_VAL_1440X1080_NL NULL
#define OPTION_VAL_1600X1200_NL NULL
#define OPTION_VAL_1920X1440_NL NULL
#define OPTION_VAL_2560X1920_NL NULL
#define OPTION_VAL_2880X2160_NL NULL
#define OPTION_VAL_3200X2400_NL NULL
#define OPTION_VAL_3840X2880_NL NULL
#define OPTION_VAL_4480X3360_NL NULL
#define OPTION_VAL_5120X3840_NL NULL
#define OPTION_VAL_5760X4320_NL NULL
#define OPTION_VAL_6400X4800_NL NULL
#define OPTION_VAL_7040X5280_NL NULL
#define OPTION_VAL_7680X5760_NL NULL
#define OPTION_VAL_8320X6240_NL NULL
#define OPTION_VAL_8960X6720_NL NULL
#define OPTION_VAL_9600X7200_NL NULL
#define OPTION_VAL_10240X7680_NL NULL
#define OPTION_VAL_10880X8160_NL NULL
#define OPTION_VAL_11520X8640_NL NULL
#define OPTION_VAL_12160X9120_NL NULL
#define OPTION_VAL_12800X9600_NL NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_NL NULL
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_NL NULL
#define OPTION_VAL_VGA_NL NULL
#define OPTION_VAL_TV_RGB_NL NULL
#define OPTION_VAL_TV_COMPOSITE_NL NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_NL NULL
#define OPTION_VAL_NTSC_NL NULL
#define OPTION_VAL_PAL_NL NULL
#define OPTION_VAL_PAL_N_NL NULL
#define OPTION_VAL_PAL_M_NL NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_NL "Schermoriëntatie"
#define OPTION_VAL_HORIZONTAL_NL NULL
#define OPTION_VAL_VERTICAL_NL NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_NL NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_NL NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_NL NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_NL NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_NL NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_NL NULL
#define OPTION_VAL_512MB_NL NULL
#define OPTION_VAL_1GB_NL NULL
#define OPTION_VAL_2GB_NL NULL
#define OPTION_VAL_4GB_NL NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_NL NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_NL NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_NL NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_NL NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_NL NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_NL NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_NL NULL
#define CORE_OPTION_NAME_FOG_LABEL_NL NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_NL NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_NL NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_NL NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_NL NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_NL NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_NL NULL
#define OPTION_VAL_1_NL NULL
#define OPTION_VAL_2_NL NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_NL NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_NL NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_NL NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_NL NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_NL NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_NL NULL
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_NL NULL
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_NL NULL
#define OPTION_VAL_2_O27_NL NULL
#define OPTION_VAL_4_NL NULL
#define OPTION_VAL_6_NL NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_NL NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_NL NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_NL NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_NL "Gedraaide Rendering"
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_NL "Voert de GPU en CPU uit op verschillende threads. Zeer aanbevolen."
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_NL NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_NL "Automatisch frames overslaan wanneer de emulator traag draait. Deze instelling is alleen van toepassing wanneer 'Gedraaide Rendering' is ingeschakeld."
#define OPTION_VAL_SOME_NL "Normaal"
#define OPTION_VAL_MORE_NL NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_NL NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_NL NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_NL NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_NL NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_NL NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_NL NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_NL NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_NL NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_NL NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_NL NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_NL NULL
#define OPTION_VAL_0_NL NULL
#define OPTION_VAL_5_NL NULL
#define OPTION_VAL_10_NL NULL
#define OPTION_VAL_15_NL NULL
#define OPTION_VAL_20_NL NULL
#define OPTION_VAL_25_NL NULL
#define OPTION_VAL_30_NL NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_NL NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_NL NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_NL NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_NL NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_NL NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_NL NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_NL NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_NL NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_NL NULL
#define OPTION_VAL_WHITE_NL NULL
#define OPTION_VAL_RED_NL "Rood"
#define OPTION_VAL_GREEN_NL "Groen"
#define OPTION_VAL_BLUE_NL "Blauw"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_NL NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_NL NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_NL NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_NL NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_NL NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_NL NULL
#define OPTION_VAL_VMU_A1_NL NULL
#define OPTION_VAL_ALL_VMUS_NL NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_NL NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_NL NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_NL NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_NL NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_NL NULL
#define OPTION_VAL_UPPER_LEFT_NL NULL
#define OPTION_VAL_UPPER_RIGHT_NL NULL
#define OPTION_VAL_LOWER_LEFT_NL NULL
#define OPTION_VAL_LOWER_RIGHT_NL NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_NL NULL
#define OPTION_VAL_1X_NL NULL
#define OPTION_VAL_3X_NL NULL
#define OPTION_VAL_5X_NL NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_NL NULL
#define OPTION_VAL_DEFAULT_ON_00_NL NULL
#define OPTION_VAL_DEFAULT_OFF_01_NL NULL
#define OPTION_VAL_BLACK_02_NL NULL
#define OPTION_VAL_LIGHT_BLUE_04_NL NULL
#define OPTION_VAL_CYAN_06_NL NULL
#define OPTION_VAL_CYAN_BLUE_07_NL NULL
#define OPTION_VAL_LIGHT_GREEN_08_NL NULL
#define OPTION_VAL_CYAN_GREEN_09_NL NULL
#define OPTION_VAL_LIGHT_CYAN_10_NL NULL
#define OPTION_VAL_PURPLE_12_NL NULL
#define OPTION_VAL_LIGHT_PURPLE_13_NL NULL
#define OPTION_VAL_YELLOW_14_NL "Geel"
#define OPTION_VAL_GRAY_15_NL NULL
#define OPTION_VAL_LIGHT_PURPLE_2_16_NL NULL
#define OPTION_VAL_LIGHT_GREEN_2_17_NL NULL
#define OPTION_VAL_LIGHT_GREEN_3_18_NL NULL
#define OPTION_VAL_LIGHT_CYAN_2_19_NL NULL
#define OPTION_VAL_LIGHT_RED_2_20_NL NULL
#define OPTION_VAL_MAGENTA_21_NL NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_NL NULL
#define OPTION_VAL_LIGHT_ORANGE_23_NL NULL
#define OPTION_VAL_ORANGE_24_NL NULL
#define OPTION_VAL_LIGHT_PURPLE_4_25_NL NULL
#define OPTION_VAL_LIGHT_YELLOW_26_NL NULL
#define OPTION_VAL_LIGHT_YELLOW_2_27_NL NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_NL NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_NL NULL
#define OPTION_VAL_40_NL NULL
#define OPTION_VAL_50_NL NULL
#define OPTION_VAL_60_NL NULL
#define OPTION_VAL_70_NL NULL
#define OPTION_VAL_80_NL NULL
#define OPTION_VAL_90_NL NULL
#define OPTION_VAL_100_NL NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_NL NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_NL NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_NL NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_NL NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_NL NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_NL NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_NL NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_NL NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_NL NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_NL NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_NL NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_NL NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_NL NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_NL NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_NL NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_NL NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_NL NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_NL NULL

struct retro_core_option_v2_category option_cats_nl[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_NL,
      CATEGORY_SYSTEM_INFO_0_NL
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_NL,
      CATEGORY_VIDEO_INFO_0_NL
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_NL,
      CATEGORY_PERFORMANCE_INFO_0_NL
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_NL,
      CATEGORY_HACKS_INFO_0_NL
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_NL,
      CATEGORY_INPUT_INFO_0_NL
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_NL,
      CATEGORY_VMU_INFO_0_NL
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_nl[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_NL,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_NL },
         { "USA",     OPTION_VAL_USA_NL },
         { "Europe",  OPTION_VAL_EUROPE_NL },
         { "Default", OPTION_VAL_DEFAULT_NL },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_NL,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_NL },
         { "English",  OPTION_VAL_ENGLISH_NL },
         { "German",   OPTION_VAL_GERMAN_NL },
         { "French",   OPTION_VAL_FRENCH_NL },
         { "Spanish",  OPTION_VAL_SPANISH_NL },
         { "Italian",  OPTION_VAL_ITALIAN_NL },
         { "Default",  OPTION_VAL_DEFAULT_NL },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_NL,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_boot_to_bios",
      CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_NL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_NL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_force_wince",
      CORE_OPTION_NAME_FORCE_WINCE_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_FORCE_WINCE_INFO_0_NL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_NL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_NL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_NL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_NL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_NL,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_NL },
         { "640x480",    OPTION_VAL_640X480_NL },
         { "800x600",    OPTION_VAL_800X600_NL },
         { "960x720",    OPTION_VAL_960X720_NL },
         { "1024x768",   OPTION_VAL_1024X768_NL },
         { "1280x960",   OPTION_VAL_1280X960_NL },
         { "1440x1080",  OPTION_VAL_1440X1080_NL },
         { "1600x1200",  OPTION_VAL_1600X1200_NL },
         { "1920x1440",  OPTION_VAL_1920X1440_NL },
         { "2560x1920",  OPTION_VAL_2560X1920_NL },
         { "2880x2160",  OPTION_VAL_2880X2160_NL },
         { "3200x2400",  OPTION_VAL_3200X2400_NL },
         { "3840x2880",  OPTION_VAL_3840X2880_NL },
         { "4480x3360",  OPTION_VAL_4480X3360_NL },
         { "5120x3840",  OPTION_VAL_5120X3840_NL },
         { "5760x4320",  OPTION_VAL_5760X4320_NL },
         { "6400x4800",  OPTION_VAL_6400X4800_NL },
         { "7040x5280",  OPTION_VAL_7040X5280_NL },
         { "7680x5760",  OPTION_VAL_7680X5760_NL },
         { "8320x6240",  OPTION_VAL_8320X6240_NL },
         { "8960x6720",  OPTION_VAL_8960X6720_NL },
         { "9600x7200",  OPTION_VAL_9600X7200_NL },
         { "10240x7680", OPTION_VAL_10240X7680_NL },
         { "10880x8160", OPTION_VAL_10880X8160_NL },
         { "11520x8640", OPTION_VAL_11520X8640_NL },
         { "12160x9120", OPTION_VAL_12160X9120_NL },
         { "12800x9600", OPTION_VAL_12800X9600_NL },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_NL,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_NL },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_NL },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_NL },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_NL,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_NL },
         { "PAL",     OPTION_VAL_PAL_NL },
         { "PAL_N",   OPTION_VAL_PAL_N_NL },
         { "PAL_M",   OPTION_VAL_PAL_M_NL },
         { "Default", OPTION_VAL_DEFAULT_NL },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_NL,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_NL },
         { "vertical",   OPTION_VAL_VERTICAL_NL },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_NL,
      NULL,
      "",
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_NL },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_NL },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_NL },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_NL,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_NL },
         { "1GB",   OPTION_VAL_1GB_NL },
         { "2GB",   OPTION_VAL_2GB_NL },
         { "4GB",   OPTION_VAL_4GB_NL },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_NL,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_NL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_NL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_NL,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_NL,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_NL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_NL,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_NL,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_NL },
         { "1",  OPTION_VAL_1_NL },
         { "2",  OPTION_VAL_2_NL },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_NL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_NL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_NL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_NL,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O27_NL },
         { "4",  OPTION_VAL_4_NL },
         { "6",  OPTION_VAL_6_NL },
         { NULL, NULL },
      },
      "1",
   },
   {/* TODO: needs clarification */
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_NL,
      NULL,
      "",
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_NL,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_NL,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_NL,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_NL,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_NL },
         { "more", OPTION_VAL_MORE_NL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_NL,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_NL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_NL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_NL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_NL,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: probably needs explanation */
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_NL,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_NL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_NL },
         { "5%",  OPTION_VAL_5_NL },
         { "10%", OPTION_VAL_10_NL },
         { "15%", OPTION_VAL_15_NL },
         { "20%", OPTION_VAL_20_NL },
         { "25%", OPTION_VAL_25_NL },
         { "30%", OPTION_VAL_30_NL },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_NL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_NL },
         { "5%",  OPTION_VAL_5_NL },
         { "10%", OPTION_VAL_10_NL },
         { "15%", OPTION_VAL_15_NL },
         { "20%", OPTION_VAL_20_NL },
         { "25%", OPTION_VAL_25_NL },
         { "30%", OPTION_VAL_30_NL },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_NL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_purupuru",
      CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_NL,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_NL,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_NL,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_NL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_NL },
         { "Red",      OPTION_VAL_RED_NL },
         { "Green",    OPTION_VAL_GREEN_NL },
         { "Blue",     OPTION_VAL_BLUE_NL },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_NL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_NL },
         { "Red",      OPTION_VAL_RED_NL },
         { "Green",    OPTION_VAL_GREEN_NL },
         { "Blue",     OPTION_VAL_BLUE_NL },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_NL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_NL },
         { "Red",      OPTION_VAL_RED_NL },
         { "Green",    OPTION_VAL_GREEN_NL },
         { "Blue",     OPTION_VAL_BLUE_NL },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_NL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_NL },
         { "Red",      OPTION_VAL_RED_NL },
         { "Green",    OPTION_VAL_GREEN_NL },
         { "Blue",     OPTION_VAL_BLUE_NL },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_NL,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_NL,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_NL,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_NL },
         { "All VMUs", OPTION_VAL_ALL_VMUS_NL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_NL,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_NL,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_NL,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_NL },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_NL },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_NL },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_NL },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_NL },
         { "2x", OPTION_VAL_2_O27_NL },
         { "3x", OPTION_VAL_3X_NL },
         { "4x", OPTION_VAL_4_NL },
         { "5x", OPTION_VAL_5X_NL },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_NL },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_NL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_NL },
         { "BLUE 03",           OPTION_VAL_BLUE_NL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_NL },
         { "GREEN 05",          OPTION_VAL_GREEN_NL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_NL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_NL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_NL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_NL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_NL },
         { "RED 11",            OPTION_VAL_RED_NL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_NL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_NL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_NL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_NL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_NL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_NL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_NL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_NL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_NL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_NL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_NL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_NL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_NL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_NL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_NL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_NL },
         { "WHITE 28",          OPTION_VAL_WHITE_NL },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_NL },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_NL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_NL },
         { "BLUE 03",           OPTION_VAL_BLUE_NL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_NL },
         { "GREEN 05",          OPTION_VAL_GREEN_NL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_NL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_NL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_NL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_NL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_NL },
         { "RED 11",            OPTION_VAL_RED_NL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_NL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_NL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_NL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_NL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_NL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_NL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_NL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_NL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_NL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_NL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_NL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_NL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_NL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_NL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_NL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_NL },
         { "WHITE 28",          OPTION_VAL_WHITE_NL },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_NL },
         { "20%",  OPTION_VAL_20_NL },
         { "30%",  OPTION_VAL_30_NL },
         { "40%",  OPTION_VAL_40_NL },
         { "50%",  OPTION_VAL_50_NL },
         { "60%",  OPTION_VAL_60_NL },
         { "70%",  OPTION_VAL_70_NL },
         { "80%",  OPTION_VAL_80_NL },
         { "90%",  OPTION_VAL_90_NL },
         { "100%", OPTION_VAL_100_NL },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_NL },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_NL },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_NL },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_NL },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_NL },
         { "2x", OPTION_VAL_2_O27_NL },
         { "3x", OPTION_VAL_3X_NL },
         { "4x", OPTION_VAL_4_NL },
         { "5x", OPTION_VAL_5X_NL },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_NL },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_NL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_NL },
         { "BLUE 03",           OPTION_VAL_BLUE_NL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_NL },
         { "GREEN 05",          OPTION_VAL_GREEN_NL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_NL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_NL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_NL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_NL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_NL },
         { "RED 11",            OPTION_VAL_RED_NL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_NL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_NL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_NL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_NL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_NL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_NL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_NL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_NL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_NL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_NL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_NL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_NL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_NL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_NL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_NL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_NL },
         { "WHITE 28",          OPTION_VAL_WHITE_NL },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_NL },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_NL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_NL },
         { "BLUE 03",           OPTION_VAL_BLUE_NL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_NL },
         { "GREEN 05",          OPTION_VAL_GREEN_NL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_NL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_NL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_NL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_NL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_NL },
         { "RED 11",            OPTION_VAL_RED_NL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_NL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_NL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_NL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_NL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_NL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_NL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_NL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_NL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_NL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_NL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_NL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_NL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_NL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_NL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_NL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_NL },
         { "WHITE 28",          OPTION_VAL_WHITE_NL },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_NL },
         { "20%",  OPTION_VAL_20_NL },
         { "30%",  OPTION_VAL_30_NL },
         { "40%",  OPTION_VAL_40_NL },
         { "50%",  OPTION_VAL_50_NL },
         { "60%",  OPTION_VAL_60_NL },
         { "70%",  OPTION_VAL_70_NL },
         { "80%",  OPTION_VAL_80_NL },
         { "90%",  OPTION_VAL_90_NL },
         { "100%", OPTION_VAL_100_NL },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_NL },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_NL },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_NL },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_NL },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_NL },
         { "2x", OPTION_VAL_2_O27_NL },
         { "3x", OPTION_VAL_3X_NL },
         { "4x", OPTION_VAL_4_NL },
         { "5x", OPTION_VAL_5X_NL },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_NL },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_NL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_NL },
         { "BLUE 03",           OPTION_VAL_BLUE_NL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_NL },
         { "GREEN 05",          OPTION_VAL_GREEN_NL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_NL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_NL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_NL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_NL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_NL },
         { "RED 11",            OPTION_VAL_RED_NL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_NL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_NL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_NL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_NL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_NL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_NL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_NL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_NL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_NL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_NL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_NL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_NL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_NL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_NL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_NL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_NL },
         { "WHITE 28",          OPTION_VAL_WHITE_NL },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_NL },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_NL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_NL },
         { "BLUE 03",           OPTION_VAL_BLUE_NL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_NL },
         { "GREEN 05",          OPTION_VAL_GREEN_NL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_NL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_NL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_NL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_NL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_NL },
         { "RED 11",            OPTION_VAL_RED_NL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_NL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_NL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_NL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_NL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_NL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_NL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_NL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_NL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_NL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_NL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_NL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_NL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_NL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_NL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_NL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_NL },
         { "WHITE 28",          OPTION_VAL_WHITE_NL },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_NL },
         { "20%",  OPTION_VAL_20_NL },
         { "30%",  OPTION_VAL_30_NL },
         { "40%",  OPTION_VAL_40_NL },
         { "50%",  OPTION_VAL_50_NL },
         { "60%",  OPTION_VAL_60_NL },
         { "70%",  OPTION_VAL_70_NL },
         { "80%",  OPTION_VAL_80_NL },
         { "90%",  OPTION_VAL_90_NL },
         { "100%", OPTION_VAL_100_NL },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_NL },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_NL },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_NL },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_NL },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_NL },
         { "2x", OPTION_VAL_2_O27_NL },
         { "3x", OPTION_VAL_3X_NL },
         { "4x", OPTION_VAL_4_NL },
         { "5x", OPTION_VAL_5X_NL },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_NL },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_NL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_NL },
         { "BLUE 03",           OPTION_VAL_BLUE_NL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_NL },
         { "GREEN 05",          OPTION_VAL_GREEN_NL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_NL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_NL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_NL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_NL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_NL },
         { "RED 11",            OPTION_VAL_RED_NL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_NL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_NL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_NL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_NL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_NL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_NL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_NL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_NL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_NL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_NL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_NL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_NL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_NL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_NL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_NL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_NL },
         { "WHITE 28",          OPTION_VAL_WHITE_NL },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_NL },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_NL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_NL },
         { "BLUE 03",           OPTION_VAL_BLUE_NL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_NL },
         { "GREEN 05",          OPTION_VAL_GREEN_NL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_NL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_NL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_NL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_NL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_NL },
         { "RED 11",            OPTION_VAL_RED_NL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_NL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_NL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_NL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_NL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_NL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_NL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_NL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_NL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_NL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_NL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_NL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_NL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_NL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_NL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_NL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_NL },
         { "WHITE 28",          OPTION_VAL_WHITE_NL },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_NL },
         { "20%",  OPTION_VAL_20_NL },
         { "30%",  OPTION_VAL_30_NL },
         { "40%",  OPTION_VAL_40_NL },
         { "50%",  OPTION_VAL_50_NL },
         { "60%",  OPTION_VAL_60_NL },
         { "70%",  OPTION_VAL_70_NL },
         { "80%",  OPTION_VAL_80_NL },
         { "90%",  OPTION_VAL_90_NL },
         { "100%", OPTION_VAL_100_NL },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_nl = {
   option_cats_nl,
   option_defs_nl
};

/* RETRO_LANGUAGE_NO */

#define CATEGORY_SYSTEM_LABEL_NO NULL
#define CATEGORY_SYSTEM_INFO_0_NO NULL
#define CATEGORY_VIDEO_LABEL_NO NULL
#define CATEGORY_VIDEO_INFO_0_NO NULL
#define CATEGORY_PERFORMANCE_LABEL_NO NULL
#define CATEGORY_PERFORMANCE_INFO_0_NO NULL
#define CATEGORY_HACKS_LABEL_NO NULL
#define CATEGORY_HACKS_INFO_0_NO NULL
#define CATEGORY_INPUT_LABEL_NO "Inndata"
#define CATEGORY_INPUT_INFO_0_NO NULL
#define CATEGORY_VMU_LABEL_NO NULL
#define CATEGORY_VMU_INFO_0_NO NULL
#define CORE_OPTION_NAME_REGION_LABEL_NO NULL
#define OPTION_VAL_JAPAN_NO NULL
#define OPTION_VAL_USA_NO NULL
#define OPTION_VAL_EUROPE_NO NULL
#define OPTION_VAL_DEFAULT_NO NULL
#define CORE_OPTION_NAME_LANGUAGE_LABEL_NO NULL
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_NO NULL
#define OPTION_VAL_JAPANESE_NO NULL
#define OPTION_VAL_ENGLISH_NO NULL
#define OPTION_VAL_GERMAN_NO NULL
#define OPTION_VAL_FRENCH_NO NULL
#define OPTION_VAL_SPANISH_NO NULL
#define OPTION_VAL_ITALIAN_NO NULL
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_NO NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_NO NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_NO NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_NO NULL
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_NO NULL
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_NO NULL
#define CORE_OPTION_NAME_FORCE_WINCE_LABEL_NO NULL
#define CORE_OPTION_NAME_FORCE_WINCE_INFO_0_NO NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_NO NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_NO NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_NO NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_NO NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_NO NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_NO NULL
#define CORE_OPTION_NAME_UPNP_LABEL_NO NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_NO NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_NO "Intern oppløsning"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_NO NULL
#define OPTION_VAL_320X240_NO NULL
#define OPTION_VAL_640X480_NO NULL
#define OPTION_VAL_800X600_NO NULL
#define OPTION_VAL_960X720_NO NULL
#define OPTION_VAL_1024X768_NO NULL
#define OPTION_VAL_1280X960_NO NULL
#define OPTION_VAL_1440X1080_NO NULL
#define OPTION_VAL_1600X1200_NO NULL
#define OPTION_VAL_1920X1440_NO NULL
#define OPTION_VAL_2560X1920_NO NULL
#define OPTION_VAL_2880X2160_NO NULL
#define OPTION_VAL_3200X2400_NO NULL
#define OPTION_VAL_3840X2880_NO NULL
#define OPTION_VAL_4480X3360_NO NULL
#define OPTION_VAL_5120X3840_NO NULL
#define OPTION_VAL_5760X4320_NO NULL
#define OPTION_VAL_6400X4800_NO NULL
#define OPTION_VAL_7040X5280_NO NULL
#define OPTION_VAL_7680X5760_NO NULL
#define OPTION_VAL_8320X6240_NO NULL
#define OPTION_VAL_8960X6720_NO NULL
#define OPTION_VAL_9600X7200_NO NULL
#define OPTION_VAL_10240X7680_NO NULL
#define OPTION_VAL_10880X8160_NO NULL
#define OPTION_VAL_11520X8640_NO NULL
#define OPTION_VAL_12160X9120_NO NULL
#define OPTION_VAL_12800X9600_NO NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_NO NULL
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_NO NULL
#define OPTION_VAL_VGA_NO NULL
#define OPTION_VAL_TV_RGB_NO NULL
#define OPTION_VAL_TV_COMPOSITE_NO NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_NO NULL
#define OPTION_VAL_NTSC_NO NULL
#define OPTION_VAL_PAL_NO NULL
#define OPTION_VAL_PAL_N_NO NULL
#define OPTION_VAL_PAL_M_NO NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_NO "Skjermorientering"
#define OPTION_VAL_HORIZONTAL_NO NULL
#define OPTION_VAL_VERTICAL_NO NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_NO NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_NO NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_NO NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_NO NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_NO NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_NO NULL
#define OPTION_VAL_512MB_NO NULL
#define OPTION_VAL_1GB_NO NULL
#define OPTION_VAL_2GB_NO NULL
#define OPTION_VAL_4GB_NO NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_NO NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_NO NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_NO NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_NO NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_NO NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_NO NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_NO NULL
#define CORE_OPTION_NAME_FOG_LABEL_NO NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_NO NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_NO NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_NO NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_NO NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_NO NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_NO NULL
#define OPTION_VAL_1_NO NULL
#define OPTION_VAL_2_NO NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_NO NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_NO NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_NO NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_NO NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_NO NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_NO NULL
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_NO NULL
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_NO NULL
#define OPTION_VAL_2_O27_NO NULL
#define OPTION_VAL_4_NO NULL
#define OPTION_VAL_6_NO NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_NO NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_NO NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_NO NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_NO NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_NO NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_NO NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_NO NULL
#define OPTION_VAL_SOME_NO NULL
#define OPTION_VAL_MORE_NO NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_NO NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_NO NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_NO NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_NO NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_NO NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_NO NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_NO NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_NO NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_NO NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_NO NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_NO NULL
#define OPTION_VAL_0_NO NULL
#define OPTION_VAL_5_NO NULL
#define OPTION_VAL_10_NO NULL
#define OPTION_VAL_15_NO NULL
#define OPTION_VAL_20_NO NULL
#define OPTION_VAL_25_NO NULL
#define OPTION_VAL_30_NO NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_NO NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_NO NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_NO NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_NO NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_NO NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_NO NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_NO NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_NO NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_NO NULL
#define OPTION_VAL_WHITE_NO NULL
#define OPTION_VAL_RED_NO NULL
#define OPTION_VAL_GREEN_NO NULL
#define OPTION_VAL_BLUE_NO NULL
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_NO NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_NO NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_NO NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_NO NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_NO NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_NO NULL
#define OPTION_VAL_VMU_A1_NO NULL
#define OPTION_VAL_ALL_VMUS_NO NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_NO NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_NO NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_NO NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_NO NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_NO NULL
#define OPTION_VAL_UPPER_LEFT_NO NULL
#define OPTION_VAL_UPPER_RIGHT_NO NULL
#define OPTION_VAL_LOWER_LEFT_NO NULL
#define OPTION_VAL_LOWER_RIGHT_NO NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_NO NULL
#define OPTION_VAL_1X_NO NULL
#define OPTION_VAL_3X_NO NULL
#define OPTION_VAL_5X_NO NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_NO NULL
#define OPTION_VAL_DEFAULT_ON_00_NO NULL
#define OPTION_VAL_DEFAULT_OFF_01_NO NULL
#define OPTION_VAL_BLACK_02_NO NULL
#define OPTION_VAL_LIGHT_BLUE_04_NO NULL
#define OPTION_VAL_CYAN_06_NO NULL
#define OPTION_VAL_CYAN_BLUE_07_NO NULL
#define OPTION_VAL_LIGHT_GREEN_08_NO NULL
#define OPTION_VAL_CYAN_GREEN_09_NO NULL
#define OPTION_VAL_LIGHT_CYAN_10_NO NULL
#define OPTION_VAL_PURPLE_12_NO NULL
#define OPTION_VAL_LIGHT_PURPLE_13_NO NULL
#define OPTION_VAL_YELLOW_14_NO NULL
#define OPTION_VAL_GRAY_15_NO NULL
#define OPTION_VAL_LIGHT_PURPLE_2_16_NO NULL
#define OPTION_VAL_LIGHT_GREEN_2_17_NO NULL
#define OPTION_VAL_LIGHT_GREEN_3_18_NO NULL
#define OPTION_VAL_LIGHT_CYAN_2_19_NO NULL
#define OPTION_VAL_LIGHT_RED_2_20_NO NULL
#define OPTION_VAL_MAGENTA_21_NO NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_NO NULL
#define OPTION_VAL_LIGHT_ORANGE_23_NO NULL
#define OPTION_VAL_ORANGE_24_NO NULL
#define OPTION_VAL_LIGHT_PURPLE_4_25_NO NULL
#define OPTION_VAL_LIGHT_YELLOW_26_NO NULL
#define OPTION_VAL_LIGHT_YELLOW_2_27_NO NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_NO NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_NO NULL
#define OPTION_VAL_40_NO NULL
#define OPTION_VAL_50_NO NULL
#define OPTION_VAL_60_NO NULL
#define OPTION_VAL_70_NO NULL
#define OPTION_VAL_80_NO NULL
#define OPTION_VAL_90_NO NULL
#define OPTION_VAL_100_NO NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_NO NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_NO NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_NO NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_NO NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_NO NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_NO NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_NO NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_NO NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_NO NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_NO NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_NO NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_NO NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_NO NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_NO NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_NO NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_NO NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_NO NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_NO NULL

struct retro_core_option_v2_category option_cats_no[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_NO,
      CATEGORY_SYSTEM_INFO_0_NO
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_NO,
      CATEGORY_VIDEO_INFO_0_NO
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_NO,
      CATEGORY_PERFORMANCE_INFO_0_NO
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_NO,
      CATEGORY_HACKS_INFO_0_NO
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_NO,
      CATEGORY_INPUT_INFO_0_NO
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_NO,
      CATEGORY_VMU_INFO_0_NO
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_no[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_NO,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_NO },
         { "USA",     OPTION_VAL_USA_NO },
         { "Europe",  OPTION_VAL_EUROPE_NO },
         { "Default", OPTION_VAL_DEFAULT_NO },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_NO,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_NO },
         { "English",  OPTION_VAL_ENGLISH_NO },
         { "German",   OPTION_VAL_GERMAN_NO },
         { "French",   OPTION_VAL_FRENCH_NO },
         { "Spanish",  OPTION_VAL_SPANISH_NO },
         { "Italian",  OPTION_VAL_ITALIAN_NO },
         { "Default",  OPTION_VAL_DEFAULT_NO },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_NO,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_boot_to_bios",
      CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_NO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_NO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_force_wince",
      CORE_OPTION_NAME_FORCE_WINCE_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_FORCE_WINCE_INFO_0_NO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_NO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_NO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_NO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_NO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_NO,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_NO },
         { "640x480",    OPTION_VAL_640X480_NO },
         { "800x600",    OPTION_VAL_800X600_NO },
         { "960x720",    OPTION_VAL_960X720_NO },
         { "1024x768",   OPTION_VAL_1024X768_NO },
         { "1280x960",   OPTION_VAL_1280X960_NO },
         { "1440x1080",  OPTION_VAL_1440X1080_NO },
         { "1600x1200",  OPTION_VAL_1600X1200_NO },
         { "1920x1440",  OPTION_VAL_1920X1440_NO },
         { "2560x1920",  OPTION_VAL_2560X1920_NO },
         { "2880x2160",  OPTION_VAL_2880X2160_NO },
         { "3200x2400",  OPTION_VAL_3200X2400_NO },
         { "3840x2880",  OPTION_VAL_3840X2880_NO },
         { "4480x3360",  OPTION_VAL_4480X3360_NO },
         { "5120x3840",  OPTION_VAL_5120X3840_NO },
         { "5760x4320",  OPTION_VAL_5760X4320_NO },
         { "6400x4800",  OPTION_VAL_6400X4800_NO },
         { "7040x5280",  OPTION_VAL_7040X5280_NO },
         { "7680x5760",  OPTION_VAL_7680X5760_NO },
         { "8320x6240",  OPTION_VAL_8320X6240_NO },
         { "8960x6720",  OPTION_VAL_8960X6720_NO },
         { "9600x7200",  OPTION_VAL_9600X7200_NO },
         { "10240x7680", OPTION_VAL_10240X7680_NO },
         { "10880x8160", OPTION_VAL_10880X8160_NO },
         { "11520x8640", OPTION_VAL_11520X8640_NO },
         { "12160x9120", OPTION_VAL_12160X9120_NO },
         { "12800x9600", OPTION_VAL_12800X9600_NO },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_NO,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_NO },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_NO },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_NO },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_NO,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_NO },
         { "PAL",     OPTION_VAL_PAL_NO },
         { "PAL_N",   OPTION_VAL_PAL_N_NO },
         { "PAL_M",   OPTION_VAL_PAL_M_NO },
         { "Default", OPTION_VAL_DEFAULT_NO },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_NO,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_NO },
         { "vertical",   OPTION_VAL_VERTICAL_NO },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_NO,
      NULL,
      "",
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_NO },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_NO },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_NO },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_NO,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_NO },
         { "1GB",   OPTION_VAL_1GB_NO },
         { "2GB",   OPTION_VAL_2GB_NO },
         { "4GB",   OPTION_VAL_4GB_NO },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_NO,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_NO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_NO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_NO,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_NO,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_NO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_NO,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_NO,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_NO },
         { "1",  OPTION_VAL_1_NO },
         { "2",  OPTION_VAL_2_NO },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_NO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_NO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_NO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_NO,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O27_NO },
         { "4",  OPTION_VAL_4_NO },
         { "6",  OPTION_VAL_6_NO },
         { NULL, NULL },
      },
      "1",
   },
   {/* TODO: needs clarification */
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_NO,
      NULL,
      "",
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_NO,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_NO,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_NO,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_NO,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_NO },
         { "more", OPTION_VAL_MORE_NO },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_NO,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_NO,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_NO,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_NO,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_NO,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: probably needs explanation */
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_NO,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_NO,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_NO },
         { "5%",  OPTION_VAL_5_NO },
         { "10%", OPTION_VAL_10_NO },
         { "15%", OPTION_VAL_15_NO },
         { "20%", OPTION_VAL_20_NO },
         { "25%", OPTION_VAL_25_NO },
         { "30%", OPTION_VAL_30_NO },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_NO,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_NO },
         { "5%",  OPTION_VAL_5_NO },
         { "10%", OPTION_VAL_10_NO },
         { "15%", OPTION_VAL_15_NO },
         { "20%", OPTION_VAL_20_NO },
         { "25%", OPTION_VAL_25_NO },
         { "30%", OPTION_VAL_30_NO },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_NO,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_purupuru",
      CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_NO,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_NO,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_NO,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_NO,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_NO },
         { "Red",      OPTION_VAL_RED_NO },
         { "Green",    OPTION_VAL_GREEN_NO },
         { "Blue",     OPTION_VAL_BLUE_NO },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_NO,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_NO },
         { "Red",      OPTION_VAL_RED_NO },
         { "Green",    OPTION_VAL_GREEN_NO },
         { "Blue",     OPTION_VAL_BLUE_NO },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_NO,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_NO },
         { "Red",      OPTION_VAL_RED_NO },
         { "Green",    OPTION_VAL_GREEN_NO },
         { "Blue",     OPTION_VAL_BLUE_NO },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_NO,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_NO },
         { "Red",      OPTION_VAL_RED_NO },
         { "Green",    OPTION_VAL_GREEN_NO },
         { "Blue",     OPTION_VAL_BLUE_NO },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_NO,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_NO,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_NO,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_NO },
         { "All VMUs", OPTION_VAL_ALL_VMUS_NO },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_NO,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_NO,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_NO,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_NO },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_NO },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_NO },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_NO },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_NO },
         { "2x", OPTION_VAL_2_O27_NO },
         { "3x", OPTION_VAL_3X_NO },
         { "4x", OPTION_VAL_4_NO },
         { "5x", OPTION_VAL_5X_NO },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_NO },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_NO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_NO },
         { "BLUE 03",           OPTION_VAL_BLUE_NO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_NO },
         { "GREEN 05",          OPTION_VAL_GREEN_NO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_NO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_NO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_NO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_NO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_NO },
         { "RED 11",            OPTION_VAL_RED_NO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_NO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_NO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_NO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_NO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_NO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_NO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_NO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_NO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_NO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_NO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_NO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_NO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_NO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_NO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_NO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_NO },
         { "WHITE 28",          OPTION_VAL_WHITE_NO },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_NO },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_NO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_NO },
         { "BLUE 03",           OPTION_VAL_BLUE_NO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_NO },
         { "GREEN 05",          OPTION_VAL_GREEN_NO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_NO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_NO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_NO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_NO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_NO },
         { "RED 11",            OPTION_VAL_RED_NO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_NO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_NO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_NO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_NO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_NO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_NO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_NO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_NO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_NO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_NO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_NO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_NO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_NO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_NO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_NO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_NO },
         { "WHITE 28",          OPTION_VAL_WHITE_NO },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_NO },
         { "20%",  OPTION_VAL_20_NO },
         { "30%",  OPTION_VAL_30_NO },
         { "40%",  OPTION_VAL_40_NO },
         { "50%",  OPTION_VAL_50_NO },
         { "60%",  OPTION_VAL_60_NO },
         { "70%",  OPTION_VAL_70_NO },
         { "80%",  OPTION_VAL_80_NO },
         { "90%",  OPTION_VAL_90_NO },
         { "100%", OPTION_VAL_100_NO },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_NO },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_NO },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_NO },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_NO },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_NO },
         { "2x", OPTION_VAL_2_O27_NO },
         { "3x", OPTION_VAL_3X_NO },
         { "4x", OPTION_VAL_4_NO },
         { "5x", OPTION_VAL_5X_NO },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_NO },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_NO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_NO },
         { "BLUE 03",           OPTION_VAL_BLUE_NO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_NO },
         { "GREEN 05",          OPTION_VAL_GREEN_NO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_NO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_NO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_NO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_NO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_NO },
         { "RED 11",            OPTION_VAL_RED_NO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_NO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_NO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_NO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_NO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_NO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_NO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_NO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_NO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_NO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_NO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_NO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_NO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_NO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_NO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_NO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_NO },
         { "WHITE 28",          OPTION_VAL_WHITE_NO },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_NO },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_NO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_NO },
         { "BLUE 03",           OPTION_VAL_BLUE_NO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_NO },
         { "GREEN 05",          OPTION_VAL_GREEN_NO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_NO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_NO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_NO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_NO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_NO },
         { "RED 11",            OPTION_VAL_RED_NO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_NO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_NO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_NO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_NO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_NO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_NO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_NO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_NO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_NO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_NO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_NO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_NO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_NO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_NO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_NO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_NO },
         { "WHITE 28",          OPTION_VAL_WHITE_NO },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_NO },
         { "20%",  OPTION_VAL_20_NO },
         { "30%",  OPTION_VAL_30_NO },
         { "40%",  OPTION_VAL_40_NO },
         { "50%",  OPTION_VAL_50_NO },
         { "60%",  OPTION_VAL_60_NO },
         { "70%",  OPTION_VAL_70_NO },
         { "80%",  OPTION_VAL_80_NO },
         { "90%",  OPTION_VAL_90_NO },
         { "100%", OPTION_VAL_100_NO },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_NO },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_NO },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_NO },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_NO },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_NO },
         { "2x", OPTION_VAL_2_O27_NO },
         { "3x", OPTION_VAL_3X_NO },
         { "4x", OPTION_VAL_4_NO },
         { "5x", OPTION_VAL_5X_NO },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_NO },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_NO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_NO },
         { "BLUE 03",           OPTION_VAL_BLUE_NO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_NO },
         { "GREEN 05",          OPTION_VAL_GREEN_NO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_NO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_NO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_NO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_NO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_NO },
         { "RED 11",            OPTION_VAL_RED_NO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_NO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_NO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_NO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_NO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_NO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_NO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_NO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_NO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_NO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_NO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_NO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_NO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_NO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_NO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_NO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_NO },
         { "WHITE 28",          OPTION_VAL_WHITE_NO },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_NO },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_NO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_NO },
         { "BLUE 03",           OPTION_VAL_BLUE_NO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_NO },
         { "GREEN 05",          OPTION_VAL_GREEN_NO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_NO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_NO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_NO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_NO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_NO },
         { "RED 11",            OPTION_VAL_RED_NO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_NO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_NO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_NO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_NO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_NO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_NO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_NO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_NO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_NO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_NO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_NO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_NO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_NO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_NO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_NO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_NO },
         { "WHITE 28",          OPTION_VAL_WHITE_NO },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_NO },
         { "20%",  OPTION_VAL_20_NO },
         { "30%",  OPTION_VAL_30_NO },
         { "40%",  OPTION_VAL_40_NO },
         { "50%",  OPTION_VAL_50_NO },
         { "60%",  OPTION_VAL_60_NO },
         { "70%",  OPTION_VAL_70_NO },
         { "80%",  OPTION_VAL_80_NO },
         { "90%",  OPTION_VAL_90_NO },
         { "100%", OPTION_VAL_100_NO },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_NO },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_NO },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_NO },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_NO },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_NO },
         { "2x", OPTION_VAL_2_O27_NO },
         { "3x", OPTION_VAL_3X_NO },
         { "4x", OPTION_VAL_4_NO },
         { "5x", OPTION_VAL_5X_NO },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_NO },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_NO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_NO },
         { "BLUE 03",           OPTION_VAL_BLUE_NO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_NO },
         { "GREEN 05",          OPTION_VAL_GREEN_NO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_NO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_NO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_NO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_NO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_NO },
         { "RED 11",            OPTION_VAL_RED_NO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_NO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_NO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_NO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_NO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_NO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_NO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_NO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_NO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_NO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_NO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_NO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_NO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_NO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_NO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_NO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_NO },
         { "WHITE 28",          OPTION_VAL_WHITE_NO },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_NO },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_NO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_NO },
         { "BLUE 03",           OPTION_VAL_BLUE_NO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_NO },
         { "GREEN 05",          OPTION_VAL_GREEN_NO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_NO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_NO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_NO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_NO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_NO },
         { "RED 11",            OPTION_VAL_RED_NO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_NO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_NO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_NO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_NO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_NO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_NO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_NO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_NO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_NO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_NO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_NO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_NO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_NO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_NO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_NO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_NO },
         { "WHITE 28",          OPTION_VAL_WHITE_NO },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_NO },
         { "20%",  OPTION_VAL_20_NO },
         { "30%",  OPTION_VAL_30_NO },
         { "40%",  OPTION_VAL_40_NO },
         { "50%",  OPTION_VAL_50_NO },
         { "60%",  OPTION_VAL_60_NO },
         { "70%",  OPTION_VAL_70_NO },
         { "80%",  OPTION_VAL_80_NO },
         { "90%",  OPTION_VAL_90_NO },
         { "100%", OPTION_VAL_100_NO },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_no = {
   option_cats_no,
   option_defs_no
};

/* RETRO_LANGUAGE_PL */

#define CATEGORY_SYSTEM_LABEL_PL NULL
#define CATEGORY_SYSTEM_INFO_0_PL "Skonfiguruj ustawienia regionu, języka, BIOS i podstawowego sprzętu."
#define CATEGORY_VIDEO_LABEL_PL "Wideo"
#define CATEGORY_VIDEO_INFO_0_PL NULL
#define CATEGORY_PERFORMANCE_LABEL_PL "Wydajność"
#define CATEGORY_PERFORMANCE_INFO_0_PL NULL
#define CATEGORY_HACKS_LABEL_PL "Hacki emulacyjne"
#define CATEGORY_HACKS_INFO_0_PL NULL
#define CATEGORY_INPUT_LABEL_PL "Wejście"
#define CATEGORY_INPUT_INFO_0_PL NULL
#define CATEGORY_VMU_LABEL_PL "Jednostka pamięci wizualnej"
#define CATEGORY_VMU_INFO_0_PL NULL
#define CORE_OPTION_NAME_REGION_LABEL_PL NULL
#define OPTION_VAL_JAPAN_PL NULL
#define OPTION_VAL_USA_PL NULL
#define OPTION_VAL_EUROPE_PL NULL
#define OPTION_VAL_DEFAULT_PL "Domyślny"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_PL "Język"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_PL "Zmienia język używany przez BIOS i przez gry zawierające wiele języków."
#define OPTION_VAL_JAPANESE_PL "japoński"
#define OPTION_VAL_ENGLISH_PL "angielski"
#define OPTION_VAL_GERMAN_PL "niemiecki"
#define OPTION_VAL_FRENCH_PL "francuski"
#define OPTION_VAL_SPANISH_PL "hiszpański"
#define OPTION_VAL_ITALIAN_PL "włoski"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_PL "HLE BIOS (wymagany restart)"
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_PL NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_PL NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_PL NULL
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_PL NULL
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_PL NULL
#define CORE_OPTION_NAME_FORCE_WINCE_LABEL_PL NULL
#define CORE_OPTION_NAME_FORCE_WINCE_INFO_0_PL NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_PL NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_PL NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_PL "Ustaw gry NAOMI jako Free Play"
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_PL NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_PL "Emulacja adaptera szerokopasmowego"
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_PL NULL
#define CORE_OPTION_NAME_UPNP_LABEL_PL "Włącz UPnP"
#define CORE_OPTION_NAME_UPNP_INFO_0_PL "Użyj UPnP, aby automatycznie skonfigurować router internetowy dla gier online."
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_PL "Rozdzielczość wewnętrzna"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_PL NULL
#define OPTION_VAL_320X240_PL NULL
#define OPTION_VAL_640X480_PL NULL
#define OPTION_VAL_800X600_PL NULL
#define OPTION_VAL_960X720_PL NULL
#define OPTION_VAL_1024X768_PL NULL
#define OPTION_VAL_1280X960_PL NULL
#define OPTION_VAL_1440X1080_PL NULL
#define OPTION_VAL_1600X1200_PL NULL
#define OPTION_VAL_1920X1440_PL NULL
#define OPTION_VAL_2560X1920_PL NULL
#define OPTION_VAL_2880X2160_PL NULL
#define OPTION_VAL_3200X2400_PL NULL
#define OPTION_VAL_3840X2880_PL NULL
#define OPTION_VAL_4480X3360_PL NULL
#define OPTION_VAL_5120X3840_PL NULL
#define OPTION_VAL_5760X4320_PL NULL
#define OPTION_VAL_6400X4800_PL NULL
#define OPTION_VAL_7040X5280_PL NULL
#define OPTION_VAL_7680X5760_PL NULL
#define OPTION_VAL_8320X6240_PL NULL
#define OPTION_VAL_8960X6720_PL NULL
#define OPTION_VAL_9600X7200_PL NULL
#define OPTION_VAL_10240X7680_PL NULL
#define OPTION_VAL_10880X8160_PL NULL
#define OPTION_VAL_11520X8640_PL NULL
#define OPTION_VAL_12160X9120_PL NULL
#define OPTION_VAL_12800X9600_PL NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_PL NULL
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_PL NULL
#define OPTION_VAL_VGA_PL NULL
#define OPTION_VAL_TV_RGB_PL NULL
#define OPTION_VAL_TV_COMPOSITE_PL NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_PL NULL
#define OPTION_VAL_NTSC_PL NULL
#define OPTION_VAL_PAL_PL NULL
#define OPTION_VAL_PAL_N_PL NULL
#define OPTION_VAL_PAL_M_PL NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_PL "Orientacja ekranu"
#define OPTION_VAL_HORIZONTAL_PL NULL
#define OPTION_VAL_VERTICAL_PL NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_PL NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_PL NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_PL NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_PL NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_PL NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_PL NULL
#define OPTION_VAL_512MB_PL NULL
#define OPTION_VAL_1GB_PL NULL
#define OPTION_VAL_2GB_PL NULL
#define OPTION_VAL_4GB_PL NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_PL NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_PL NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_PL "Emulacja pełnego bufora ramki"
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_PL NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_PL NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_PL NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_PL NULL
#define CORE_OPTION_NAME_FOG_LABEL_PL NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_PL NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_PL NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_PL NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_PL NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_PL "Filtrowanie tekstur"
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_PL NULL
#define OPTION_VAL_1_PL NULL
#define OPTION_VAL_2_PL "Wymuś liniowy"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_PL NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_PL NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_PL "Wykrywanie zmian częstotliwości klatek"
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_PL NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_PL NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_PL NULL
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_PL NULL
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_PL NULL
#define OPTION_VAL_2_O27_PL NULL
#define OPTION_VAL_4_PL NULL
#define OPTION_VAL_6_PL NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_PL NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_PL NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_PL NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_PL NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_PL NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_PL NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_PL NULL
#define OPTION_VAL_SOME_PL "Normalny"
#define OPTION_VAL_MORE_PL NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_PL NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_PL NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_PL NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_PL NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_PL NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_PL NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_PL NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_PL NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_PL NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_PL NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_PL NULL
#define OPTION_VAL_0_PL NULL
#define OPTION_VAL_5_PL NULL
#define OPTION_VAL_10_PL NULL
#define OPTION_VAL_15_PL NULL
#define OPTION_VAL_20_PL NULL
#define OPTION_VAL_25_PL NULL
#define OPTION_VAL_30_PL "30 %"
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_PL NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_PL NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_PL NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_PL NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_PL NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_PL NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_PL NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_PL NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_PL NULL
#define OPTION_VAL_WHITE_PL NULL
#define OPTION_VAL_RED_PL "Czerwony"
#define OPTION_VAL_GREEN_PL "Zielony"
#define OPTION_VAL_BLUE_PL "Niebieski"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_PL NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_PL NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_PL NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_PL NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_PL NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_PL NULL
#define OPTION_VAL_VMU_A1_PL NULL
#define OPTION_VAL_ALL_VMUS_PL NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_PL NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_PL NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_PL NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_PL NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_PL NULL
#define OPTION_VAL_UPPER_LEFT_PL NULL
#define OPTION_VAL_UPPER_RIGHT_PL NULL
#define OPTION_VAL_LOWER_LEFT_PL NULL
#define OPTION_VAL_LOWER_RIGHT_PL NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_PL NULL
#define OPTION_VAL_1X_PL NULL
#define OPTION_VAL_3X_PL NULL
#define OPTION_VAL_5X_PL NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_PL NULL
#define OPTION_VAL_DEFAULT_ON_00_PL NULL
#define OPTION_VAL_DEFAULT_OFF_01_PL NULL
#define OPTION_VAL_BLACK_02_PL NULL
#define OPTION_VAL_LIGHT_BLUE_04_PL NULL
#define OPTION_VAL_CYAN_06_PL NULL
#define OPTION_VAL_CYAN_BLUE_07_PL NULL
#define OPTION_VAL_LIGHT_GREEN_08_PL NULL
#define OPTION_VAL_CYAN_GREEN_09_PL NULL
#define OPTION_VAL_LIGHT_CYAN_10_PL NULL
#define OPTION_VAL_PURPLE_12_PL NULL
#define OPTION_VAL_LIGHT_PURPLE_13_PL NULL
#define OPTION_VAL_YELLOW_14_PL "Żółty"
#define OPTION_VAL_GRAY_15_PL NULL
#define OPTION_VAL_LIGHT_PURPLE_2_16_PL NULL
#define OPTION_VAL_LIGHT_GREEN_2_17_PL NULL
#define OPTION_VAL_LIGHT_GREEN_3_18_PL NULL
#define OPTION_VAL_LIGHT_CYAN_2_19_PL NULL
#define OPTION_VAL_LIGHT_RED_2_20_PL NULL
#define OPTION_VAL_MAGENTA_21_PL NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_PL NULL
#define OPTION_VAL_LIGHT_ORANGE_23_PL NULL
#define OPTION_VAL_ORANGE_24_PL NULL
#define OPTION_VAL_LIGHT_PURPLE_4_25_PL NULL
#define OPTION_VAL_LIGHT_YELLOW_26_PL NULL
#define OPTION_VAL_LIGHT_YELLOW_2_27_PL NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_PL NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_PL NULL
#define OPTION_VAL_40_PL NULL
#define OPTION_VAL_50_PL "50 %"
#define OPTION_VAL_60_PL "60 %"
#define OPTION_VAL_70_PL NULL
#define OPTION_VAL_80_PL "80 %"
#define OPTION_VAL_90_PL NULL
#define OPTION_VAL_100_PL NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_PL NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_PL NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_PL NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_PL NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_PL NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_PL NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_PL NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_PL NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_PL NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_PL NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_PL NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_PL NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_PL NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_PL NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_PL NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_PL NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_PL NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_PL NULL

struct retro_core_option_v2_category option_cats_pl[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_PL,
      CATEGORY_SYSTEM_INFO_0_PL
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_PL,
      CATEGORY_VIDEO_INFO_0_PL
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_PL,
      CATEGORY_PERFORMANCE_INFO_0_PL
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_PL,
      CATEGORY_HACKS_INFO_0_PL
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_PL,
      CATEGORY_INPUT_INFO_0_PL
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_PL,
      CATEGORY_VMU_INFO_0_PL
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_pl[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_PL,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_PL },
         { "USA",     OPTION_VAL_USA_PL },
         { "Europe",  OPTION_VAL_EUROPE_PL },
         { "Default", OPTION_VAL_DEFAULT_PL },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_PL,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_PL },
         { "English",  OPTION_VAL_ENGLISH_PL },
         { "German",   OPTION_VAL_GERMAN_PL },
         { "French",   OPTION_VAL_FRENCH_PL },
         { "Spanish",  OPTION_VAL_SPANISH_PL },
         { "Italian",  OPTION_VAL_ITALIAN_PL },
         { "Default",  OPTION_VAL_DEFAULT_PL },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_PL,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_boot_to_bios",
      CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_PL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_PL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_force_wince",
      CORE_OPTION_NAME_FORCE_WINCE_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_FORCE_WINCE_INFO_0_PL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_PL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_PL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_PL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_PL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_PL,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_PL },
         { "640x480",    OPTION_VAL_640X480_PL },
         { "800x600",    OPTION_VAL_800X600_PL },
         { "960x720",    OPTION_VAL_960X720_PL },
         { "1024x768",   OPTION_VAL_1024X768_PL },
         { "1280x960",   OPTION_VAL_1280X960_PL },
         { "1440x1080",  OPTION_VAL_1440X1080_PL },
         { "1600x1200",  OPTION_VAL_1600X1200_PL },
         { "1920x1440",  OPTION_VAL_1920X1440_PL },
         { "2560x1920",  OPTION_VAL_2560X1920_PL },
         { "2880x2160",  OPTION_VAL_2880X2160_PL },
         { "3200x2400",  OPTION_VAL_3200X2400_PL },
         { "3840x2880",  OPTION_VAL_3840X2880_PL },
         { "4480x3360",  OPTION_VAL_4480X3360_PL },
         { "5120x3840",  OPTION_VAL_5120X3840_PL },
         { "5760x4320",  OPTION_VAL_5760X4320_PL },
         { "6400x4800",  OPTION_VAL_6400X4800_PL },
         { "7040x5280",  OPTION_VAL_7040X5280_PL },
         { "7680x5760",  OPTION_VAL_7680X5760_PL },
         { "8320x6240",  OPTION_VAL_8320X6240_PL },
         { "8960x6720",  OPTION_VAL_8960X6720_PL },
         { "9600x7200",  OPTION_VAL_9600X7200_PL },
         { "10240x7680", OPTION_VAL_10240X7680_PL },
         { "10880x8160", OPTION_VAL_10880X8160_PL },
         { "11520x8640", OPTION_VAL_11520X8640_PL },
         { "12160x9120", OPTION_VAL_12160X9120_PL },
         { "12800x9600", OPTION_VAL_12800X9600_PL },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_PL,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_PL },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_PL },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_PL },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_PL,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_PL },
         { "PAL",     OPTION_VAL_PAL_PL },
         { "PAL_N",   OPTION_VAL_PAL_N_PL },
         { "PAL_M",   OPTION_VAL_PAL_M_PL },
         { "Default", OPTION_VAL_DEFAULT_PL },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_PL,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_PL },
         { "vertical",   OPTION_VAL_VERTICAL_PL },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_PL,
      NULL,
      "",
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_PL },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_PL },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_PL },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_PL,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_PL },
         { "1GB",   OPTION_VAL_1GB_PL },
         { "2GB",   OPTION_VAL_2GB_PL },
         { "4GB",   OPTION_VAL_4GB_PL },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_PL,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_PL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_PL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_PL,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_PL,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_PL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_PL,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_PL,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_PL },
         { "1",  OPTION_VAL_1_PL },
         { "2",  OPTION_VAL_2_PL },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_PL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_PL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_PL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_PL,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O27_PL },
         { "4",  OPTION_VAL_4_PL },
         { "6",  OPTION_VAL_6_PL },
         { NULL, NULL },
      },
      "1",
   },
   {/* TODO: needs clarification */
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_PL,
      NULL,
      "",
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_PL,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_PL,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_PL,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_PL,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_PL },
         { "more", OPTION_VAL_MORE_PL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_PL,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_PL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_PL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_PL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_PL,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: probably needs explanation */
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_PL,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_PL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_PL },
         { "5%",  OPTION_VAL_5_PL },
         { "10%", OPTION_VAL_10_PL },
         { "15%", OPTION_VAL_15_PL },
         { "20%", OPTION_VAL_20_PL },
         { "25%", OPTION_VAL_25_PL },
         { "30%", OPTION_VAL_30_PL },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_PL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_PL },
         { "5%",  OPTION_VAL_5_PL },
         { "10%", OPTION_VAL_10_PL },
         { "15%", OPTION_VAL_15_PL },
         { "20%", OPTION_VAL_20_PL },
         { "25%", OPTION_VAL_25_PL },
         { "30%", OPTION_VAL_30_PL },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_PL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_purupuru",
      CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_PL,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_PL,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_PL,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_PL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_PL },
         { "Red",      OPTION_VAL_RED_PL },
         { "Green",    OPTION_VAL_GREEN_PL },
         { "Blue",     OPTION_VAL_BLUE_PL },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_PL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_PL },
         { "Red",      OPTION_VAL_RED_PL },
         { "Green",    OPTION_VAL_GREEN_PL },
         { "Blue",     OPTION_VAL_BLUE_PL },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_PL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_PL },
         { "Red",      OPTION_VAL_RED_PL },
         { "Green",    OPTION_VAL_GREEN_PL },
         { "Blue",     OPTION_VAL_BLUE_PL },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_PL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_PL },
         { "Red",      OPTION_VAL_RED_PL },
         { "Green",    OPTION_VAL_GREEN_PL },
         { "Blue",     OPTION_VAL_BLUE_PL },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_PL,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_PL,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_PL,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_PL },
         { "All VMUs", OPTION_VAL_ALL_VMUS_PL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_PL,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_PL,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_PL,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_PL },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_PL },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_PL },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_PL },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_PL },
         { "2x", OPTION_VAL_2_O27_PL },
         { "3x", OPTION_VAL_3X_PL },
         { "4x", OPTION_VAL_4_PL },
         { "5x", OPTION_VAL_5X_PL },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PL },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PL },
         { "BLUE 03",           OPTION_VAL_BLUE_PL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PL },
         { "GREEN 05",          OPTION_VAL_GREEN_PL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PL },
         { "RED 11",            OPTION_VAL_RED_PL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PL },
         { "WHITE 28",          OPTION_VAL_WHITE_PL },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PL },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PL },
         { "BLUE 03",           OPTION_VAL_BLUE_PL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PL },
         { "GREEN 05",          OPTION_VAL_GREEN_PL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PL },
         { "RED 11",            OPTION_VAL_RED_PL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PL },
         { "WHITE 28",          OPTION_VAL_WHITE_PL },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_PL },
         { "20%",  OPTION_VAL_20_PL },
         { "30%",  OPTION_VAL_30_PL },
         { "40%",  OPTION_VAL_40_PL },
         { "50%",  OPTION_VAL_50_PL },
         { "60%",  OPTION_VAL_60_PL },
         { "70%",  OPTION_VAL_70_PL },
         { "80%",  OPTION_VAL_80_PL },
         { "90%",  OPTION_VAL_90_PL },
         { "100%", OPTION_VAL_100_PL },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_PL },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_PL },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_PL },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_PL },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_PL },
         { "2x", OPTION_VAL_2_O27_PL },
         { "3x", OPTION_VAL_3X_PL },
         { "4x", OPTION_VAL_4_PL },
         { "5x", OPTION_VAL_5X_PL },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PL },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PL },
         { "BLUE 03",           OPTION_VAL_BLUE_PL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PL },
         { "GREEN 05",          OPTION_VAL_GREEN_PL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PL },
         { "RED 11",            OPTION_VAL_RED_PL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PL },
         { "WHITE 28",          OPTION_VAL_WHITE_PL },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PL },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PL },
         { "BLUE 03",           OPTION_VAL_BLUE_PL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PL },
         { "GREEN 05",          OPTION_VAL_GREEN_PL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PL },
         { "RED 11",            OPTION_VAL_RED_PL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PL },
         { "WHITE 28",          OPTION_VAL_WHITE_PL },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_PL },
         { "20%",  OPTION_VAL_20_PL },
         { "30%",  OPTION_VAL_30_PL },
         { "40%",  OPTION_VAL_40_PL },
         { "50%",  OPTION_VAL_50_PL },
         { "60%",  OPTION_VAL_60_PL },
         { "70%",  OPTION_VAL_70_PL },
         { "80%",  OPTION_VAL_80_PL },
         { "90%",  OPTION_VAL_90_PL },
         { "100%", OPTION_VAL_100_PL },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_PL },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_PL },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_PL },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_PL },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_PL },
         { "2x", OPTION_VAL_2_O27_PL },
         { "3x", OPTION_VAL_3X_PL },
         { "4x", OPTION_VAL_4_PL },
         { "5x", OPTION_VAL_5X_PL },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PL },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PL },
         { "BLUE 03",           OPTION_VAL_BLUE_PL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PL },
         { "GREEN 05",          OPTION_VAL_GREEN_PL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PL },
         { "RED 11",            OPTION_VAL_RED_PL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PL },
         { "WHITE 28",          OPTION_VAL_WHITE_PL },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PL },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PL },
         { "BLUE 03",           OPTION_VAL_BLUE_PL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PL },
         { "GREEN 05",          OPTION_VAL_GREEN_PL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PL },
         { "RED 11",            OPTION_VAL_RED_PL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PL },
         { "WHITE 28",          OPTION_VAL_WHITE_PL },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_PL },
         { "20%",  OPTION_VAL_20_PL },
         { "30%",  OPTION_VAL_30_PL },
         { "40%",  OPTION_VAL_40_PL },
         { "50%",  OPTION_VAL_50_PL },
         { "60%",  OPTION_VAL_60_PL },
         { "70%",  OPTION_VAL_70_PL },
         { "80%",  OPTION_VAL_80_PL },
         { "90%",  OPTION_VAL_90_PL },
         { "100%", OPTION_VAL_100_PL },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_PL },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_PL },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_PL },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_PL },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_PL },
         { "2x", OPTION_VAL_2_O27_PL },
         { "3x", OPTION_VAL_3X_PL },
         { "4x", OPTION_VAL_4_PL },
         { "5x", OPTION_VAL_5X_PL },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PL },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PL },
         { "BLUE 03",           OPTION_VAL_BLUE_PL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PL },
         { "GREEN 05",          OPTION_VAL_GREEN_PL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PL },
         { "RED 11",            OPTION_VAL_RED_PL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PL },
         { "WHITE 28",          OPTION_VAL_WHITE_PL },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PL },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PL },
         { "BLUE 03",           OPTION_VAL_BLUE_PL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PL },
         { "GREEN 05",          OPTION_VAL_GREEN_PL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PL },
         { "RED 11",            OPTION_VAL_RED_PL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PL },
         { "WHITE 28",          OPTION_VAL_WHITE_PL },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_PL },
         { "20%",  OPTION_VAL_20_PL },
         { "30%",  OPTION_VAL_30_PL },
         { "40%",  OPTION_VAL_40_PL },
         { "50%",  OPTION_VAL_50_PL },
         { "60%",  OPTION_VAL_60_PL },
         { "70%",  OPTION_VAL_70_PL },
         { "80%",  OPTION_VAL_80_PL },
         { "90%",  OPTION_VAL_90_PL },
         { "100%", OPTION_VAL_100_PL },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_pl = {
   option_cats_pl,
   option_defs_pl
};

/* RETRO_LANGUAGE_PT_BR */

#define CATEGORY_SYSTEM_LABEL_PT_BR "Sistema"
#define CATEGORY_SYSTEM_INFO_0_PT_BR "Configure a região, o idioma, a BIOS e as configurações básicas do hardware."
#define CATEGORY_VIDEO_LABEL_PT_BR "Vídeo"
#define CATEGORY_VIDEO_INFO_0_PT_BR "Configure a resolução, a transparência independente da ordem e dos efeitos visuais."
#define CATEGORY_PERFORMANCE_LABEL_PT_BR "Desempenho"
#define CATEGORY_PERFORMANCE_INFO_0_PT_BR NULL
#define CATEGORY_HACKS_LABEL_PT_BR "Hacks de emulação"
#define CATEGORY_HACKS_INFO_0_PT_BR "Configura as configurações da substituição da tela widescreen, da velocidade do carregamento da GD-ROM e da textura."
#define CATEGORY_INPUT_LABEL_PT_BR "Entrada"
#define CATEGORY_INPUT_INFO_0_PT_BR "Configura os ajustes do controle e da pistola de luz."
#define CATEGORY_VMU_LABEL_PT_BR "Unidade de memória visual"
#define CATEGORY_VMU_INFO_0_PT_BR "Configura os arquivos de salvamento VMU por jogo e as configurações da visibilidade do VMU na tela."
#define CORE_OPTION_NAME_REGION_LABEL_PT_BR "Região"
#define OPTION_VAL_JAPAN_PT_BR "Japão"
#define OPTION_VAL_USA_PT_BR "EUA"
#define OPTION_VAL_EUROPE_PT_BR "Europa"
#define OPTION_VAL_DEFAULT_PT_BR "Padrão"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_PT_BR "Idioma"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_PT_BR "Altera o idioma usado pela BIOS e por quaisquer jogos que contenham vários idiomas."
#define OPTION_VAL_JAPANESE_PT_BR "Japonês"
#define OPTION_VAL_ENGLISH_PT_BR "Inglês"
#define OPTION_VAL_GERMAN_PT_BR "Alemão"
#define OPTION_VAL_FRENCH_PT_BR "Francês"
#define OPTION_VAL_SPANISH_PT_BR "Espanhol"
#define OPTION_VAL_ITALIAN_PT_BR "Italiano"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_PT_BR NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_PT_BR "Força o uso de uma BIOS criada com emulação de alto nível."
#define CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_PT_BR "Inicialização a partir da BIOS (requer reinício)"
#define CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_PT_BR "Inicializa diretamente no menu da BIOS do Dreamcast."
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_PT_BR "Ativar DSP"
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_PT_BR "Ativa a emulação Dreamcast do DSP (processador de sinal digital) de áudio. Melhora a precisão do som gerado ao custo de requisitos maiores."
#define CORE_OPTION_NAME_FORCE_WINCE_LABEL_PT_BR "Forçar o modo Windows CE"
#define CORE_OPTION_NAME_FORCE_WINCE_INFO_0_PT_BR "Ativa a emulação completa da MMU (Unidade de Gerenciamento de Memória) e outras configurações para jogos compatíveis com o Windows CE."
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_PT_BR "Ativar botões de serviço da NAOMI"
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_PT_BR "Ativa o botão SERVICE da placa NAOMI para acessar a configuração do gabinete."
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_PT_BR NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_PT_BR NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_PT_BR NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_PT_BR NULL
#define CORE_OPTION_NAME_UPNP_LABEL_PT_BR NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_PT_BR NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_PT_BR "Resolução interna"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_PT_BR "Modificar resolução de renderização."
#define OPTION_VAL_320X240_PT_BR NULL
#define OPTION_VAL_640X480_PT_BR NULL
#define OPTION_VAL_800X600_PT_BR NULL
#define OPTION_VAL_960X720_PT_BR NULL
#define OPTION_VAL_1024X768_PT_BR NULL
#define OPTION_VAL_1280X960_PT_BR NULL
#define OPTION_VAL_1440X1080_PT_BR NULL
#define OPTION_VAL_1600X1200_PT_BR NULL
#define OPTION_VAL_1920X1440_PT_BR NULL
#define OPTION_VAL_2560X1920_PT_BR NULL
#define OPTION_VAL_2880X2160_PT_BR NULL
#define OPTION_VAL_3200X2400_PT_BR NULL
#define OPTION_VAL_3840X2880_PT_BR NULL
#define OPTION_VAL_4480X3360_PT_BR NULL
#define OPTION_VAL_5120X3840_PT_BR NULL
#define OPTION_VAL_5760X4320_PT_BR NULL
#define OPTION_VAL_6400X4800_PT_BR NULL
#define OPTION_VAL_7040X5280_PT_BR NULL
#define OPTION_VAL_7680X5760_PT_BR NULL
#define OPTION_VAL_8320X6240_PT_BR NULL
#define OPTION_VAL_8960X6720_PT_BR NULL
#define OPTION_VAL_9600X7200_PT_BR NULL
#define OPTION_VAL_10240X7680_PT_BR NULL
#define OPTION_VAL_10880X8160_PT_BR NULL
#define OPTION_VAL_11520X8640_PT_BR NULL
#define OPTION_VAL_12160X9120_PT_BR NULL
#define OPTION_VAL_12800X9600_PT_BR NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_PT_BR "Tipo de cabo"
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_PT_BR "O tipo de sinal de saída. 'TV (Composto)' é o mais amplamente suportado."
#define OPTION_VAL_VGA_PT_BR NULL
#define OPTION_VAL_TV_RGB_PT_BR NULL
#define OPTION_VAL_TV_COMPOSITE_PT_BR "TV (Composto)"
#define CORE_OPTION_NAME_BROADCAST_LABEL_PT_BR "Padrão de transmissão"
#define OPTION_VAL_NTSC_PT_BR NULL
#define OPTION_VAL_PAL_PT_BR "PAL (Resto do mundo)"
#define OPTION_VAL_PAL_N_PT_BR "PAL-N (Argentina, Paraguai, Uruguai)"
#define OPTION_VAL_PAL_M_PT_BR "PAL-M (Brasil)"
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_PT_BR "Orientação da tela"
#define OPTION_VAL_HORIZONTAL_PT_BR NULL
#define OPTION_VAL_VERTICAL_PT_BR NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_PT_BR "Classificação de transparências"
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_PT_BR "Por faixa (rápido, menos preciso)"
#define OPTION_VAL_PER_TRIANGLE_NORMAL_PT_BR "Por triângulo (normal)"
#define OPTION_VAL_PER_PIXEL_ACCURATE_PT_BR "Por pixel (preciso, mas lento)"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_PT_BR NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_PT_BR "Um alto valor pode ser necessário para que a imagem seja bem exibida em resoluções superiores às nativas."
#define OPTION_VAL_512MB_PT_BR NULL
#define OPTION_VAL_1GB_PT_BR NULL
#define OPTION_VAL_2GB_PT_BR NULL
#define OPTION_VAL_4GB_PT_BR NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_PT_BR NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_PT_BR NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_PT_BR NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_PT_BR NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_PT_BR "Ativar o buffer RTT (renderização para texturas)"
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_PT_BR NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_PT_BR NULL
#define CORE_OPTION_NAME_FOG_LABEL_PT_BR "Efeitos de nevoeiro"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_PT_BR "Modificador de volume"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_PT_BR "Uma característica da GPU do Dreamcast que é frequentemente usada por jogos para desenhar sombras em objetos. Esta opção deve geralmente ser ativada; tem pouco efeito sobre o desempenho."
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_PT_BR "Filtro anisotrópico"
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_PT_BR NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_PT_BR "Filtragem de textura"
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_PT_BR NULL
#define OPTION_VAL_1_PT_BR NULL
#define OPTION_VAL_2_PT_BR NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_PT_BR "Atrasar troca de quadro"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_PT_BR "Útil para evitar tremulações da imagem ou defeitos de vídeo. Não recomendado para uso em plataformas lentas."
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_PT_BR NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_PT_BR NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_PT_BR "Filtro de pós-processamento do PowerVR2"
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_PT_BR "Pós-processa a imagem renderizada para simular efeitos específicos da GPU PowerVR2 e os sinais de vídeo analógicos."
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_PT_BR "Redimensionamento de textura (xBRZ)"
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_PT_BR "Melhora os gráficos 2D de 'pixel arte' desenhados à mão. Usar somente com jogos em 2D pixelado."
#define OPTION_VAL_2_O27_PT_BR NULL
#define OPTION_VAL_4_PT_BR NULL
#define OPTION_VAL_6_PT_BR NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_PT_BR "Tamanho máximo do redimensionamento de texturas"
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_PT_BR NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_PT_BR NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_PT_BR "Renderização multi-thread"
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_PT_BR "Executa a GPU e a CPU em diferentes threads. Altamente recomendado."
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_PT_BR "Pulo de quadros automático"
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_PT_BR "Pula automaticamente quadros quando o emulador está lento. Nota: Esta configuração só será aplicada quando a 'Renderização multi-thread' estiver ativada."
#define OPTION_VAL_SOME_PT_BR NULL
#define OPTION_VAL_MORE_PT_BR "Máximo"
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_PT_BR "Pulando quadro"
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_PT_BR "Define o número de quadros a serem pulados após cada quadro ser exibido."
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_PT_BR "Hack de trapaças (requer reinício)"
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_PT_BR "Ativa trapaças que permitem a exibição de certos jogos em formato widescreen."
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_PT_BR "Hack de widescreen"
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_PT_BR "Desenha a geometria além da proporção de tela normal de 4:3. Pode produzir falhas gráficas em áreas reveladas."
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_PT_BR "Carregamento rápido do GD-ROM (impreciso)"
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_PT_BR "Acelera o carregamento do GD-ROM."
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_PT_BR "Carregar texturas personalizadas"
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_PT_BR "Despejo de texturas"
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_PT_BR "Zona morta do direcional analógico"
#define OPTION_VAL_0_PT_BR NULL
#define OPTION_VAL_5_PT_BR NULL
#define OPTION_VAL_10_PT_BR NULL
#define OPTION_VAL_15_PT_BR NULL
#define OPTION_VAL_20_PT_BR NULL
#define OPTION_VAL_25_PT_BR NULL
#define OPTION_VAL_30_PT_BR NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_PT_BR "Zona morta dos gatilhos"
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_PT_BR "Gatilhos digitais"
#define CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_PT_BR "Purupuru Pack (vibração)"
#define CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_PT_BR "Habilita a vibração do controle."
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_PT_BR NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_PT_BR NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_PT_BR "Mostrar configurações das pistolas de luz"
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_PT_BR "Ativa a configuração do da mira da pistola de luz na tela. OBSERVAÇÃO: O menu rápido pode precisar ser alternado para que esta configuração surta efeito."
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_PT_BR "Mostrar ponto de mira da arma 1"
#define OPTION_VAL_WHITE_PT_BR "Branco"
#define OPTION_VAL_RED_PT_BR "Vermelho"
#define OPTION_VAL_GREEN_PT_BR "Verde"
#define OPTION_VAL_BLUE_PT_BR "Azul"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_PT_BR "Mostrar ponto de mira da arma 2"
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_PT_BR "Mostrar ponto de mira da arma 3"
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_PT_BR "Mostrar ponto de mira da arma 4"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_PT_BR "Separar Visual Memory (VMU) por jogo"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_PT_BR "VMUs por jogo"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_PT_BR "Ao desativar esta opção, todos os jogos compartilharão quatro arquivos de salvamento da VMU (A1, B1, C1, D1) que serão localizados no diretório de sistema do RetroArch. A configuração 'VMU A1' criará um arquivo único da VMU A1 no diretório de salvamento do RetroArch para cada jogo que for executado. A configuração 'Todas as VMUs' criará quatro arquivos únicos VMU (A1, B1, C1, D1) para cada jogo que for executado."
#define OPTION_VAL_VMU_A1_PT_BR NULL
#define OPTION_VAL_ALL_VMUS_PT_BR "Todas as VMUs"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_PT_BR "Mostrar configurações de visualização das Visual Memory (VMU)"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_PT_BR "Mostrar configurações da exibição da VMU"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_PT_BR "Ativa a configuração da visibilidade, do tamanho, da posição e da cor das telas de LCD emulados da VMU. OBSERVAÇÃO: O menu rápido pode precisar ser alternado para que esta configuração surta efeito."
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_PT_BR "Tela da VMU 1"
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_PT_BR "Posição da tela da VMU 1"
#define OPTION_VAL_UPPER_LEFT_PT_BR "Canto superior esquerdo"
#define OPTION_VAL_UPPER_RIGHT_PT_BR "Canto superior direito"
#define OPTION_VAL_LOWER_LEFT_PT_BR "Canto inferior esquerdo"
#define OPTION_VAL_LOWER_RIGHT_PT_BR "Canto inferior direito"
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_PT_BR "Tamanho da tela da VMU 1"
#define OPTION_VAL_1X_PT_BR NULL
#define OPTION_VAL_3X_PT_BR NULL
#define OPTION_VAL_5X_PT_BR NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_PT_BR "Cor dos pixels acesos da VMU 1"
#define OPTION_VAL_DEFAULT_ON_00_PT_BR "Ligar por padrão"
#define OPTION_VAL_DEFAULT_OFF_01_PT_BR "Desligar por padrão"
#define OPTION_VAL_BLACK_02_PT_BR "Preta"
#define OPTION_VAL_LIGHT_BLUE_04_PT_BR "Azul claro"
#define OPTION_VAL_CYAN_06_PT_BR "Azul-marinho"
#define OPTION_VAL_CYAN_BLUE_07_PT_BR "Azul ciano"
#define OPTION_VAL_LIGHT_GREEN_08_PT_BR "Verde claro"
#define OPTION_VAL_CYAN_GREEN_09_PT_BR "Verde ciano"
#define OPTION_VAL_LIGHT_CYAN_10_PT_BR "Ciano claro"
#define OPTION_VAL_PURPLE_12_PT_BR "Violeta"
#define OPTION_VAL_LIGHT_PURPLE_13_PT_BR "Roxo claro"
#define OPTION_VAL_YELLOW_14_PT_BR "Amarelo"
#define OPTION_VAL_GRAY_15_PT_BR "Cinza"
#define OPTION_VAL_LIGHT_PURPLE_2_16_PT_BR "Roxo claro (2)"
#define OPTION_VAL_LIGHT_GREEN_2_17_PT_BR "Verde claro (2)"
#define OPTION_VAL_LIGHT_GREEN_3_18_PT_BR "Verde claro (3)"
#define OPTION_VAL_LIGHT_CYAN_2_19_PT_BR "Ciano claro (2)"
#define OPTION_VAL_LIGHT_RED_2_20_PT_BR "Vermelho claro (2)"
#define OPTION_VAL_MAGENTA_21_PT_BR NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_PT_BR "Roxo claro (3)"
#define OPTION_VAL_LIGHT_ORANGE_23_PT_BR "Laranja claro"
#define OPTION_VAL_ORANGE_24_PT_BR "Laranja"
#define OPTION_VAL_LIGHT_PURPLE_4_25_PT_BR "Roxo claro (4)"
#define OPTION_VAL_LIGHT_YELLOW_26_PT_BR "Amarelo claro"
#define OPTION_VAL_LIGHT_YELLOW_2_27_PT_BR "Amarelo claro (2)"
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_PT_BR "Cor dos pixels apagados da VMU 1"
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_PT_BR "Opacidade da tela da VMU 1"
#define OPTION_VAL_40_PT_BR NULL
#define OPTION_VAL_50_PT_BR NULL
#define OPTION_VAL_60_PT_BR NULL
#define OPTION_VAL_70_PT_BR NULL
#define OPTION_VAL_80_PT_BR NULL
#define OPTION_VAL_90_PT_BR NULL
#define OPTION_VAL_100_PT_BR NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_PT_BR "Tela da VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_PT_BR "Posição da tela da VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_PT_BR "Tamanho da tela da VMU 2"
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_PT_BR "Cor dos pixels acesos da VMU 2"
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_PT_BR "Cor dos pixels apagados da VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_PT_BR "Opacidade da tela da VMU 2"
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_PT_BR "Tela da VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_PT_BR "Posição da tela da VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_PT_BR "Tamanho da tela da VMU 3"
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_PT_BR "Cor dos pixels acesos da VMU 3"
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_PT_BR "Cor dos pixels apagados da VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_PT_BR "Opacidade da tela da VMU 3"
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_PT_BR "Tela da VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_PT_BR "Posição da tela da VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_PT_BR "Tamanho da tela da VMU 4"
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_PT_BR "Cor dos pixels acesos da VMU 4"
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_PT_BR "Cor dos pixels apagados da VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_PT_BR "Opacidade da tela da VMU 4"

struct retro_core_option_v2_category option_cats_pt_br[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_PT_BR,
      CATEGORY_SYSTEM_INFO_0_PT_BR
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_PT_BR,
      CATEGORY_VIDEO_INFO_0_PT_BR
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_PT_BR,
      CATEGORY_PERFORMANCE_INFO_0_PT_BR
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_PT_BR,
      CATEGORY_HACKS_INFO_0_PT_BR
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_PT_BR,
      CATEGORY_INPUT_INFO_0_PT_BR
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_PT_BR,
      CATEGORY_VMU_INFO_0_PT_BR
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_pt_br[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_PT_BR },
         { "USA",     OPTION_VAL_USA_PT_BR },
         { "Europe",  OPTION_VAL_EUROPE_PT_BR },
         { "Default", OPTION_VAL_DEFAULT_PT_BR },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_PT_BR,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_PT_BR },
         { "English",  OPTION_VAL_ENGLISH_PT_BR },
         { "German",   OPTION_VAL_GERMAN_PT_BR },
         { "French",   OPTION_VAL_FRENCH_PT_BR },
         { "Spanish",  OPTION_VAL_SPANISH_PT_BR },
         { "Italian",  OPTION_VAL_ITALIAN_PT_BR },
         { "Default",  OPTION_VAL_DEFAULT_PT_BR },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_PT_BR,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_boot_to_bios",
      CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_PT_BR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_PT_BR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_force_wince",
      CORE_OPTION_NAME_FORCE_WINCE_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_FORCE_WINCE_INFO_0_PT_BR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_PT_BR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_PT_BR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_PT_BR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_PT_BR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_PT_BR },
         { "640x480",    OPTION_VAL_640X480_PT_BR },
         { "800x600",    OPTION_VAL_800X600_PT_BR },
         { "960x720",    OPTION_VAL_960X720_PT_BR },
         { "1024x768",   OPTION_VAL_1024X768_PT_BR },
         { "1280x960",   OPTION_VAL_1280X960_PT_BR },
         { "1440x1080",  OPTION_VAL_1440X1080_PT_BR },
         { "1600x1200",  OPTION_VAL_1600X1200_PT_BR },
         { "1920x1440",  OPTION_VAL_1920X1440_PT_BR },
         { "2560x1920",  OPTION_VAL_2560X1920_PT_BR },
         { "2880x2160",  OPTION_VAL_2880X2160_PT_BR },
         { "3200x2400",  OPTION_VAL_3200X2400_PT_BR },
         { "3840x2880",  OPTION_VAL_3840X2880_PT_BR },
         { "4480x3360",  OPTION_VAL_4480X3360_PT_BR },
         { "5120x3840",  OPTION_VAL_5120X3840_PT_BR },
         { "5760x4320",  OPTION_VAL_5760X4320_PT_BR },
         { "6400x4800",  OPTION_VAL_6400X4800_PT_BR },
         { "7040x5280",  OPTION_VAL_7040X5280_PT_BR },
         { "7680x5760",  OPTION_VAL_7680X5760_PT_BR },
         { "8320x6240",  OPTION_VAL_8320X6240_PT_BR },
         { "8960x6720",  OPTION_VAL_8960X6720_PT_BR },
         { "9600x7200",  OPTION_VAL_9600X7200_PT_BR },
         { "10240x7680", OPTION_VAL_10240X7680_PT_BR },
         { "10880x8160", OPTION_VAL_10880X8160_PT_BR },
         { "11520x8640", OPTION_VAL_11520X8640_PT_BR },
         { "12160x9120", OPTION_VAL_12160X9120_PT_BR },
         { "12800x9600", OPTION_VAL_12800X9600_PT_BR },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_PT_BR },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_PT_BR },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_PT_BR },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_PT_BR },
         { "PAL",     OPTION_VAL_PAL_PT_BR },
         { "PAL_N",   OPTION_VAL_PAL_N_PT_BR },
         { "PAL_M",   OPTION_VAL_PAL_M_PT_BR },
         { "Default", OPTION_VAL_DEFAULT_PT_BR },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_PT_BR },
         { "vertical",   OPTION_VAL_VERTICAL_PT_BR },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_PT_BR },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_PT_BR },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_PT_BR },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_PT_BR },
         { "1GB",   OPTION_VAL_1GB_PT_BR },
         { "2GB",   OPTION_VAL_2GB_PT_BR },
         { "4GB",   OPTION_VAL_4GB_PT_BR },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_PT_BR },
         { "1",  OPTION_VAL_1_PT_BR },
         { "2",  OPTION_VAL_2_PT_BR },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O27_PT_BR },
         { "4",  OPTION_VAL_4_PT_BR },
         { "6",  OPTION_VAL_6_PT_BR },
         { NULL, NULL },
      },
      "1",
   },
   {/* TODO: needs clarification */
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_PT_BR,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_PT_BR,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_PT_BR,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_PT_BR,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_PT_BR },
         { "more", OPTION_VAL_MORE_PT_BR },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_PT_BR,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_PT_BR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_PT_BR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_PT_BR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: probably needs explanation */
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_PT_BR },
         { "5%",  OPTION_VAL_5_PT_BR },
         { "10%", OPTION_VAL_10_PT_BR },
         { "15%", OPTION_VAL_15_PT_BR },
         { "20%", OPTION_VAL_20_PT_BR },
         { "25%", OPTION_VAL_25_PT_BR },
         { "30%", OPTION_VAL_30_PT_BR },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_PT_BR },
         { "5%",  OPTION_VAL_5_PT_BR },
         { "10%", OPTION_VAL_10_PT_BR },
         { "15%", OPTION_VAL_15_PT_BR },
         { "20%", OPTION_VAL_20_PT_BR },
         { "25%", OPTION_VAL_25_PT_BR },
         { "30%", OPTION_VAL_30_PT_BR },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_purupuru",
      CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_PT_BR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_PT_BR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_PT_BR,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_PT_BR },
         { "Red",      OPTION_VAL_RED_PT_BR },
         { "Green",    OPTION_VAL_GREEN_PT_BR },
         { "Blue",     OPTION_VAL_BLUE_PT_BR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_PT_BR },
         { "Red",      OPTION_VAL_RED_PT_BR },
         { "Green",    OPTION_VAL_GREEN_PT_BR },
         { "Blue",     OPTION_VAL_BLUE_PT_BR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_PT_BR },
         { "Red",      OPTION_VAL_RED_PT_BR },
         { "Green",    OPTION_VAL_GREEN_PT_BR },
         { "Blue",     OPTION_VAL_BLUE_PT_BR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_PT_BR },
         { "Red",      OPTION_VAL_RED_PT_BR },
         { "Green",    OPTION_VAL_GREEN_PT_BR },
         { "Blue",     OPTION_VAL_BLUE_PT_BR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_PT_BR,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_PT_BR,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_PT_BR,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_PT_BR },
         { "All VMUs", OPTION_VAL_ALL_VMUS_PT_BR },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_PT_BR,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_PT_BR,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_PT_BR,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_PT_BR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_PT_BR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_PT_BR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_PT_BR },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_PT_BR },
         { "2x", OPTION_VAL_2_O27_PT_BR },
         { "3x", OPTION_VAL_3X_PT_BR },
         { "4x", OPTION_VAL_4_PT_BR },
         { "5x", OPTION_VAL_5X_PT_BR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PT_BR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PT_BR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PT_BR },
         { "BLUE 03",           OPTION_VAL_BLUE_PT_BR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PT_BR },
         { "GREEN 05",          OPTION_VAL_GREEN_PT_BR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PT_BR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PT_BR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PT_BR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PT_BR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PT_BR },
         { "RED 11",            OPTION_VAL_RED_PT_BR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PT_BR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PT_BR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PT_BR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PT_BR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PT_BR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PT_BR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PT_BR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PT_BR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PT_BR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PT_BR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PT_BR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PT_BR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PT_BR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PT_BR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PT_BR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PT_BR },
         { "WHITE 28",          OPTION_VAL_WHITE_PT_BR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PT_BR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PT_BR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PT_BR },
         { "BLUE 03",           OPTION_VAL_BLUE_PT_BR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PT_BR },
         { "GREEN 05",          OPTION_VAL_GREEN_PT_BR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PT_BR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PT_BR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PT_BR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PT_BR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PT_BR },
         { "RED 11",            OPTION_VAL_RED_PT_BR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PT_BR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PT_BR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PT_BR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PT_BR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PT_BR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PT_BR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PT_BR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PT_BR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PT_BR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PT_BR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PT_BR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PT_BR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PT_BR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PT_BR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PT_BR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PT_BR },
         { "WHITE 28",          OPTION_VAL_WHITE_PT_BR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_PT_BR },
         { "20%",  OPTION_VAL_20_PT_BR },
         { "30%",  OPTION_VAL_30_PT_BR },
         { "40%",  OPTION_VAL_40_PT_BR },
         { "50%",  OPTION_VAL_50_PT_BR },
         { "60%",  OPTION_VAL_60_PT_BR },
         { "70%",  OPTION_VAL_70_PT_BR },
         { "80%",  OPTION_VAL_80_PT_BR },
         { "90%",  OPTION_VAL_90_PT_BR },
         { "100%", OPTION_VAL_100_PT_BR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_PT_BR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_PT_BR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_PT_BR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_PT_BR },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_PT_BR },
         { "2x", OPTION_VAL_2_O27_PT_BR },
         { "3x", OPTION_VAL_3X_PT_BR },
         { "4x", OPTION_VAL_4_PT_BR },
         { "5x", OPTION_VAL_5X_PT_BR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PT_BR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PT_BR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PT_BR },
         { "BLUE 03",           OPTION_VAL_BLUE_PT_BR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PT_BR },
         { "GREEN 05",          OPTION_VAL_GREEN_PT_BR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PT_BR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PT_BR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PT_BR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PT_BR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PT_BR },
         { "RED 11",            OPTION_VAL_RED_PT_BR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PT_BR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PT_BR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PT_BR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PT_BR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PT_BR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PT_BR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PT_BR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PT_BR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PT_BR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PT_BR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PT_BR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PT_BR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PT_BR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PT_BR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PT_BR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PT_BR },
         { "WHITE 28",          OPTION_VAL_WHITE_PT_BR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PT_BR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PT_BR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PT_BR },
         { "BLUE 03",           OPTION_VAL_BLUE_PT_BR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PT_BR },
         { "GREEN 05",          OPTION_VAL_GREEN_PT_BR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PT_BR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PT_BR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PT_BR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PT_BR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PT_BR },
         { "RED 11",            OPTION_VAL_RED_PT_BR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PT_BR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PT_BR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PT_BR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PT_BR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PT_BR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PT_BR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PT_BR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PT_BR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PT_BR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PT_BR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PT_BR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PT_BR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PT_BR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PT_BR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PT_BR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PT_BR },
         { "WHITE 28",          OPTION_VAL_WHITE_PT_BR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_PT_BR },
         { "20%",  OPTION_VAL_20_PT_BR },
         { "30%",  OPTION_VAL_30_PT_BR },
         { "40%",  OPTION_VAL_40_PT_BR },
         { "50%",  OPTION_VAL_50_PT_BR },
         { "60%",  OPTION_VAL_60_PT_BR },
         { "70%",  OPTION_VAL_70_PT_BR },
         { "80%",  OPTION_VAL_80_PT_BR },
         { "90%",  OPTION_VAL_90_PT_BR },
         { "100%", OPTION_VAL_100_PT_BR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_PT_BR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_PT_BR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_PT_BR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_PT_BR },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_PT_BR },
         { "2x", OPTION_VAL_2_O27_PT_BR },
         { "3x", OPTION_VAL_3X_PT_BR },
         { "4x", OPTION_VAL_4_PT_BR },
         { "5x", OPTION_VAL_5X_PT_BR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PT_BR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PT_BR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PT_BR },
         { "BLUE 03",           OPTION_VAL_BLUE_PT_BR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PT_BR },
         { "GREEN 05",          OPTION_VAL_GREEN_PT_BR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PT_BR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PT_BR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PT_BR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PT_BR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PT_BR },
         { "RED 11",            OPTION_VAL_RED_PT_BR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PT_BR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PT_BR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PT_BR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PT_BR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PT_BR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PT_BR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PT_BR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PT_BR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PT_BR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PT_BR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PT_BR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PT_BR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PT_BR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PT_BR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PT_BR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PT_BR },
         { "WHITE 28",          OPTION_VAL_WHITE_PT_BR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PT_BR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PT_BR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PT_BR },
         { "BLUE 03",           OPTION_VAL_BLUE_PT_BR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PT_BR },
         { "GREEN 05",          OPTION_VAL_GREEN_PT_BR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PT_BR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PT_BR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PT_BR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PT_BR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PT_BR },
         { "RED 11",            OPTION_VAL_RED_PT_BR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PT_BR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PT_BR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PT_BR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PT_BR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PT_BR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PT_BR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PT_BR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PT_BR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PT_BR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PT_BR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PT_BR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PT_BR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PT_BR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PT_BR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PT_BR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PT_BR },
         { "WHITE 28",          OPTION_VAL_WHITE_PT_BR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_PT_BR },
         { "20%",  OPTION_VAL_20_PT_BR },
         { "30%",  OPTION_VAL_30_PT_BR },
         { "40%",  OPTION_VAL_40_PT_BR },
         { "50%",  OPTION_VAL_50_PT_BR },
         { "60%",  OPTION_VAL_60_PT_BR },
         { "70%",  OPTION_VAL_70_PT_BR },
         { "80%",  OPTION_VAL_80_PT_BR },
         { "90%",  OPTION_VAL_90_PT_BR },
         { "100%", OPTION_VAL_100_PT_BR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_PT_BR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_PT_BR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_PT_BR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_PT_BR },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_PT_BR },
         { "2x", OPTION_VAL_2_O27_PT_BR },
         { "3x", OPTION_VAL_3X_PT_BR },
         { "4x", OPTION_VAL_4_PT_BR },
         { "5x", OPTION_VAL_5X_PT_BR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PT_BR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PT_BR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PT_BR },
         { "BLUE 03",           OPTION_VAL_BLUE_PT_BR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PT_BR },
         { "GREEN 05",          OPTION_VAL_GREEN_PT_BR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PT_BR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PT_BR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PT_BR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PT_BR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PT_BR },
         { "RED 11",            OPTION_VAL_RED_PT_BR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PT_BR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PT_BR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PT_BR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PT_BR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PT_BR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PT_BR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PT_BR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PT_BR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PT_BR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PT_BR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PT_BR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PT_BR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PT_BR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PT_BR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PT_BR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PT_BR },
         { "WHITE 28",          OPTION_VAL_WHITE_PT_BR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PT_BR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PT_BR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PT_BR },
         { "BLUE 03",           OPTION_VAL_BLUE_PT_BR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PT_BR },
         { "GREEN 05",          OPTION_VAL_GREEN_PT_BR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PT_BR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PT_BR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PT_BR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PT_BR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PT_BR },
         { "RED 11",            OPTION_VAL_RED_PT_BR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PT_BR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PT_BR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PT_BR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PT_BR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PT_BR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PT_BR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PT_BR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PT_BR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PT_BR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PT_BR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PT_BR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PT_BR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PT_BR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PT_BR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PT_BR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PT_BR },
         { "WHITE 28",          OPTION_VAL_WHITE_PT_BR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_PT_BR },
         { "20%",  OPTION_VAL_20_PT_BR },
         { "30%",  OPTION_VAL_30_PT_BR },
         { "40%",  OPTION_VAL_40_PT_BR },
         { "50%",  OPTION_VAL_50_PT_BR },
         { "60%",  OPTION_VAL_60_PT_BR },
         { "70%",  OPTION_VAL_70_PT_BR },
         { "80%",  OPTION_VAL_80_PT_BR },
         { "90%",  OPTION_VAL_90_PT_BR },
         { "100%", OPTION_VAL_100_PT_BR },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_pt_br = {
   option_cats_pt_br,
   option_defs_pt_br
};

/* RETRO_LANGUAGE_PT_PT */

#define CATEGORY_SYSTEM_LABEL_PT_PT NULL
#define CATEGORY_SYSTEM_INFO_0_PT_PT NULL
#define CATEGORY_VIDEO_LABEL_PT_PT "Vídeo"
#define CATEGORY_VIDEO_INFO_0_PT_PT NULL
#define CATEGORY_PERFORMANCE_LABEL_PT_PT NULL
#define CATEGORY_PERFORMANCE_INFO_0_PT_PT NULL
#define CATEGORY_HACKS_LABEL_PT_PT NULL
#define CATEGORY_HACKS_INFO_0_PT_PT NULL
#define CATEGORY_INPUT_LABEL_PT_PT "Entrada"
#define CATEGORY_INPUT_INFO_0_PT_PT NULL
#define CATEGORY_VMU_LABEL_PT_PT NULL
#define CATEGORY_VMU_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_REGION_LABEL_PT_PT NULL
#define OPTION_VAL_JAPAN_PT_PT NULL
#define OPTION_VAL_USA_PT_PT NULL
#define OPTION_VAL_EUROPE_PT_PT NULL
#define OPTION_VAL_DEFAULT_PT_PT NULL
#define CORE_OPTION_NAME_LANGUAGE_LABEL_PT_PT "Idioma"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_PT_PT NULL
#define OPTION_VAL_JAPANESE_PT_PT "Japonês"
#define OPTION_VAL_ENGLISH_PT_PT "Inglês"
#define OPTION_VAL_GERMAN_PT_PT "Alemão"
#define OPTION_VAL_FRENCH_PT_PT "Francês"
#define OPTION_VAL_SPANISH_PT_PT "Espanhol"
#define OPTION_VAL_ITALIAN_PT_PT "Italiano"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_FORCE_WINCE_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_FORCE_WINCE_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_UPNP_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_PT_PT NULL
#define OPTION_VAL_320X240_PT_PT NULL
#define OPTION_VAL_640X480_PT_PT NULL
#define OPTION_VAL_800X600_PT_PT NULL
#define OPTION_VAL_960X720_PT_PT NULL
#define OPTION_VAL_1024X768_PT_PT NULL
#define OPTION_VAL_1280X960_PT_PT NULL
#define OPTION_VAL_1440X1080_PT_PT NULL
#define OPTION_VAL_1600X1200_PT_PT NULL
#define OPTION_VAL_1920X1440_PT_PT NULL
#define OPTION_VAL_2560X1920_PT_PT NULL
#define OPTION_VAL_2880X2160_PT_PT NULL
#define OPTION_VAL_3200X2400_PT_PT NULL
#define OPTION_VAL_3840X2880_PT_PT NULL
#define OPTION_VAL_4480X3360_PT_PT NULL
#define OPTION_VAL_5120X3840_PT_PT NULL
#define OPTION_VAL_5760X4320_PT_PT NULL
#define OPTION_VAL_6400X4800_PT_PT NULL
#define OPTION_VAL_7040X5280_PT_PT NULL
#define OPTION_VAL_7680X5760_PT_PT NULL
#define OPTION_VAL_8320X6240_PT_PT NULL
#define OPTION_VAL_8960X6720_PT_PT NULL
#define OPTION_VAL_9600X7200_PT_PT NULL
#define OPTION_VAL_10240X7680_PT_PT NULL
#define OPTION_VAL_10880X8160_PT_PT NULL
#define OPTION_VAL_11520X8640_PT_PT NULL
#define OPTION_VAL_12160X9120_PT_PT NULL
#define OPTION_VAL_12800X9600_PT_PT NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_PT_PT NULL
#define OPTION_VAL_VGA_PT_PT NULL
#define OPTION_VAL_TV_RGB_PT_PT NULL
#define OPTION_VAL_TV_COMPOSITE_PT_PT NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_PT_PT NULL
#define OPTION_VAL_NTSC_PT_PT NULL
#define OPTION_VAL_PAL_PT_PT NULL
#define OPTION_VAL_PAL_N_PT_PT NULL
#define OPTION_VAL_PAL_M_PT_PT NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_PT_PT "Orientação do ecrã"
#define OPTION_VAL_HORIZONTAL_PT_PT NULL
#define OPTION_VAL_VERTICAL_PT_PT NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_PT_PT NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_PT_PT NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_PT_PT NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_PT_PT NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_PT_PT NULL
#define OPTION_VAL_512MB_PT_PT NULL
#define OPTION_VAL_1GB_PT_PT NULL
#define OPTION_VAL_2GB_PT_PT NULL
#define OPTION_VAL_4GB_PT_PT NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_FOG_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_PT_PT NULL
#define OPTION_VAL_1_PT_PT NULL
#define OPTION_VAL_2_PT_PT NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_PT_PT NULL
#define OPTION_VAL_2_O27_PT_PT NULL
#define OPTION_VAL_4_PT_PT NULL
#define OPTION_VAL_6_PT_PT NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_PT_PT NULL
#define OPTION_VAL_SOME_PT_PT NULL
#define OPTION_VAL_MORE_PT_PT NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_PT_PT NULL
#define OPTION_VAL_0_PT_PT NULL
#define OPTION_VAL_5_PT_PT NULL
#define OPTION_VAL_10_PT_PT NULL
#define OPTION_VAL_15_PT_PT NULL
#define OPTION_VAL_20_PT_PT NULL
#define OPTION_VAL_25_PT_PT NULL
#define OPTION_VAL_30_PT_PT NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_PT_PT NULL
#define OPTION_VAL_WHITE_PT_PT NULL
#define OPTION_VAL_RED_PT_PT "Vermelho"
#define OPTION_VAL_GREEN_PT_PT "Verde"
#define OPTION_VAL_BLUE_PT_PT "Azul"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_PT_PT NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_PT_PT NULL
#define OPTION_VAL_VMU_A1_PT_PT NULL
#define OPTION_VAL_ALL_VMUS_PT_PT NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_PT_PT NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_PT_PT NULL
#define OPTION_VAL_UPPER_LEFT_PT_PT NULL
#define OPTION_VAL_UPPER_RIGHT_PT_PT NULL
#define OPTION_VAL_LOWER_LEFT_PT_PT NULL
#define OPTION_VAL_LOWER_RIGHT_PT_PT NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_PT_PT NULL
#define OPTION_VAL_1X_PT_PT NULL
#define OPTION_VAL_3X_PT_PT NULL
#define OPTION_VAL_5X_PT_PT NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_PT_PT NULL
#define OPTION_VAL_DEFAULT_ON_00_PT_PT NULL
#define OPTION_VAL_DEFAULT_OFF_01_PT_PT NULL
#define OPTION_VAL_BLACK_02_PT_PT NULL
#define OPTION_VAL_LIGHT_BLUE_04_PT_PT NULL
#define OPTION_VAL_CYAN_06_PT_PT NULL
#define OPTION_VAL_CYAN_BLUE_07_PT_PT NULL
#define OPTION_VAL_LIGHT_GREEN_08_PT_PT NULL
#define OPTION_VAL_CYAN_GREEN_09_PT_PT NULL
#define OPTION_VAL_LIGHT_CYAN_10_PT_PT NULL
#define OPTION_VAL_PURPLE_12_PT_PT NULL
#define OPTION_VAL_LIGHT_PURPLE_13_PT_PT NULL
#define OPTION_VAL_YELLOW_14_PT_PT "Amarelo"
#define OPTION_VAL_GRAY_15_PT_PT NULL
#define OPTION_VAL_LIGHT_PURPLE_2_16_PT_PT NULL
#define OPTION_VAL_LIGHT_GREEN_2_17_PT_PT NULL
#define OPTION_VAL_LIGHT_GREEN_3_18_PT_PT NULL
#define OPTION_VAL_LIGHT_CYAN_2_19_PT_PT NULL
#define OPTION_VAL_LIGHT_RED_2_20_PT_PT NULL
#define OPTION_VAL_MAGENTA_21_PT_PT NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_PT_PT NULL
#define OPTION_VAL_LIGHT_ORANGE_23_PT_PT NULL
#define OPTION_VAL_ORANGE_24_PT_PT NULL
#define OPTION_VAL_LIGHT_PURPLE_4_25_PT_PT NULL
#define OPTION_VAL_LIGHT_YELLOW_26_PT_PT NULL
#define OPTION_VAL_LIGHT_YELLOW_2_27_PT_PT NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_PT_PT NULL
#define OPTION_VAL_40_PT_PT NULL
#define OPTION_VAL_50_PT_PT NULL
#define OPTION_VAL_60_PT_PT NULL
#define OPTION_VAL_70_PT_PT NULL
#define OPTION_VAL_80_PT_PT NULL
#define OPTION_VAL_90_PT_PT NULL
#define OPTION_VAL_100_PT_PT NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_PT_PT NULL

struct retro_core_option_v2_category option_cats_pt_pt[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_PT_PT,
      CATEGORY_SYSTEM_INFO_0_PT_PT
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_PT_PT,
      CATEGORY_VIDEO_INFO_0_PT_PT
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_PT_PT,
      CATEGORY_PERFORMANCE_INFO_0_PT_PT
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_PT_PT,
      CATEGORY_HACKS_INFO_0_PT_PT
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_PT_PT,
      CATEGORY_INPUT_INFO_0_PT_PT
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_PT_PT,
      CATEGORY_VMU_INFO_0_PT_PT
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_pt_pt[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_PT_PT },
         { "USA",     OPTION_VAL_USA_PT_PT },
         { "Europe",  OPTION_VAL_EUROPE_PT_PT },
         { "Default", OPTION_VAL_DEFAULT_PT_PT },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_PT_PT,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_PT_PT },
         { "English",  OPTION_VAL_ENGLISH_PT_PT },
         { "German",   OPTION_VAL_GERMAN_PT_PT },
         { "French",   OPTION_VAL_FRENCH_PT_PT },
         { "Spanish",  OPTION_VAL_SPANISH_PT_PT },
         { "Italian",  OPTION_VAL_ITALIAN_PT_PT },
         { "Default",  OPTION_VAL_DEFAULT_PT_PT },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_PT_PT,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_boot_to_bios",
      CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_PT_PT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_PT_PT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_force_wince",
      CORE_OPTION_NAME_FORCE_WINCE_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_FORCE_WINCE_INFO_0_PT_PT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_PT_PT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_PT_PT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_PT_PT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_PT_PT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_PT_PT },
         { "640x480",    OPTION_VAL_640X480_PT_PT },
         { "800x600",    OPTION_VAL_800X600_PT_PT },
         { "960x720",    OPTION_VAL_960X720_PT_PT },
         { "1024x768",   OPTION_VAL_1024X768_PT_PT },
         { "1280x960",   OPTION_VAL_1280X960_PT_PT },
         { "1440x1080",  OPTION_VAL_1440X1080_PT_PT },
         { "1600x1200",  OPTION_VAL_1600X1200_PT_PT },
         { "1920x1440",  OPTION_VAL_1920X1440_PT_PT },
         { "2560x1920",  OPTION_VAL_2560X1920_PT_PT },
         { "2880x2160",  OPTION_VAL_2880X2160_PT_PT },
         { "3200x2400",  OPTION_VAL_3200X2400_PT_PT },
         { "3840x2880",  OPTION_VAL_3840X2880_PT_PT },
         { "4480x3360",  OPTION_VAL_4480X3360_PT_PT },
         { "5120x3840",  OPTION_VAL_5120X3840_PT_PT },
         { "5760x4320",  OPTION_VAL_5760X4320_PT_PT },
         { "6400x4800",  OPTION_VAL_6400X4800_PT_PT },
         { "7040x5280",  OPTION_VAL_7040X5280_PT_PT },
         { "7680x5760",  OPTION_VAL_7680X5760_PT_PT },
         { "8320x6240",  OPTION_VAL_8320X6240_PT_PT },
         { "8960x6720",  OPTION_VAL_8960X6720_PT_PT },
         { "9600x7200",  OPTION_VAL_9600X7200_PT_PT },
         { "10240x7680", OPTION_VAL_10240X7680_PT_PT },
         { "10880x8160", OPTION_VAL_10880X8160_PT_PT },
         { "11520x8640", OPTION_VAL_11520X8640_PT_PT },
         { "12160x9120", OPTION_VAL_12160X9120_PT_PT },
         { "12800x9600", OPTION_VAL_12800X9600_PT_PT },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_PT_PT },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_PT_PT },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_PT_PT },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_PT_PT },
         { "PAL",     OPTION_VAL_PAL_PT_PT },
         { "PAL_N",   OPTION_VAL_PAL_N_PT_PT },
         { "PAL_M",   OPTION_VAL_PAL_M_PT_PT },
         { "Default", OPTION_VAL_DEFAULT_PT_PT },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_PT_PT },
         { "vertical",   OPTION_VAL_VERTICAL_PT_PT },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_PT_PT },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_PT_PT },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_PT_PT },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_PT_PT },
         { "1GB",   OPTION_VAL_1GB_PT_PT },
         { "2GB",   OPTION_VAL_2GB_PT_PT },
         { "4GB",   OPTION_VAL_4GB_PT_PT },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_PT_PT },
         { "1",  OPTION_VAL_1_PT_PT },
         { "2",  OPTION_VAL_2_PT_PT },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O27_PT_PT },
         { "4",  OPTION_VAL_4_PT_PT },
         { "6",  OPTION_VAL_6_PT_PT },
         { NULL, NULL },
      },
      "1",
   },
   {/* TODO: needs clarification */
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_PT_PT,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_PT_PT,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_PT_PT,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_PT_PT,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_PT_PT },
         { "more", OPTION_VAL_MORE_PT_PT },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_PT_PT,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_PT_PT,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_PT_PT,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_PT_PT,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: probably needs explanation */
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_PT_PT },
         { "5%",  OPTION_VAL_5_PT_PT },
         { "10%", OPTION_VAL_10_PT_PT },
         { "15%", OPTION_VAL_15_PT_PT },
         { "20%", OPTION_VAL_20_PT_PT },
         { "25%", OPTION_VAL_25_PT_PT },
         { "30%", OPTION_VAL_30_PT_PT },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_PT_PT },
         { "5%",  OPTION_VAL_5_PT_PT },
         { "10%", OPTION_VAL_10_PT_PT },
         { "15%", OPTION_VAL_15_PT_PT },
         { "20%", OPTION_VAL_20_PT_PT },
         { "25%", OPTION_VAL_25_PT_PT },
         { "30%", OPTION_VAL_30_PT_PT },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_purupuru",
      CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_PT_PT,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_PT_PT,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_PT_PT,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_PT_PT },
         { "Red",      OPTION_VAL_RED_PT_PT },
         { "Green",    OPTION_VAL_GREEN_PT_PT },
         { "Blue",     OPTION_VAL_BLUE_PT_PT },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_PT_PT },
         { "Red",      OPTION_VAL_RED_PT_PT },
         { "Green",    OPTION_VAL_GREEN_PT_PT },
         { "Blue",     OPTION_VAL_BLUE_PT_PT },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_PT_PT },
         { "Red",      OPTION_VAL_RED_PT_PT },
         { "Green",    OPTION_VAL_GREEN_PT_PT },
         { "Blue",     OPTION_VAL_BLUE_PT_PT },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_PT_PT },
         { "Red",      OPTION_VAL_RED_PT_PT },
         { "Green",    OPTION_VAL_GREEN_PT_PT },
         { "Blue",     OPTION_VAL_BLUE_PT_PT },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_PT_PT,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_PT_PT,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_PT_PT,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_PT_PT },
         { "All VMUs", OPTION_VAL_ALL_VMUS_PT_PT },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_PT_PT,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_PT_PT,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_PT_PT,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_PT_PT },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_PT_PT },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_PT_PT },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_PT_PT },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_PT_PT },
         { "2x", OPTION_VAL_2_O27_PT_PT },
         { "3x", OPTION_VAL_3X_PT_PT },
         { "4x", OPTION_VAL_4_PT_PT },
         { "5x", OPTION_VAL_5X_PT_PT },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PT_PT },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PT_PT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PT_PT },
         { "BLUE 03",           OPTION_VAL_BLUE_PT_PT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PT_PT },
         { "GREEN 05",          OPTION_VAL_GREEN_PT_PT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PT_PT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PT_PT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PT_PT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PT_PT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PT_PT },
         { "RED 11",            OPTION_VAL_RED_PT_PT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PT_PT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PT_PT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PT_PT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PT_PT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PT_PT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PT_PT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PT_PT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PT_PT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PT_PT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PT_PT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PT_PT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PT_PT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PT_PT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PT_PT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PT_PT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PT_PT },
         { "WHITE 28",          OPTION_VAL_WHITE_PT_PT },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PT_PT },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PT_PT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PT_PT },
         { "BLUE 03",           OPTION_VAL_BLUE_PT_PT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PT_PT },
         { "GREEN 05",          OPTION_VAL_GREEN_PT_PT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PT_PT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PT_PT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PT_PT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PT_PT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PT_PT },
         { "RED 11",            OPTION_VAL_RED_PT_PT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PT_PT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PT_PT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PT_PT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PT_PT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PT_PT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PT_PT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PT_PT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PT_PT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PT_PT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PT_PT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PT_PT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PT_PT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PT_PT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PT_PT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PT_PT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PT_PT },
         { "WHITE 28",          OPTION_VAL_WHITE_PT_PT },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_PT_PT },
         { "20%",  OPTION_VAL_20_PT_PT },
         { "30%",  OPTION_VAL_30_PT_PT },
         { "40%",  OPTION_VAL_40_PT_PT },
         { "50%",  OPTION_VAL_50_PT_PT },
         { "60%",  OPTION_VAL_60_PT_PT },
         { "70%",  OPTION_VAL_70_PT_PT },
         { "80%",  OPTION_VAL_80_PT_PT },
         { "90%",  OPTION_VAL_90_PT_PT },
         { "100%", OPTION_VAL_100_PT_PT },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_PT_PT },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_PT_PT },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_PT_PT },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_PT_PT },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_PT_PT },
         { "2x", OPTION_VAL_2_O27_PT_PT },
         { "3x", OPTION_VAL_3X_PT_PT },
         { "4x", OPTION_VAL_4_PT_PT },
         { "5x", OPTION_VAL_5X_PT_PT },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PT_PT },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PT_PT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PT_PT },
         { "BLUE 03",           OPTION_VAL_BLUE_PT_PT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PT_PT },
         { "GREEN 05",          OPTION_VAL_GREEN_PT_PT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PT_PT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PT_PT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PT_PT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PT_PT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PT_PT },
         { "RED 11",            OPTION_VAL_RED_PT_PT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PT_PT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PT_PT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PT_PT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PT_PT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PT_PT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PT_PT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PT_PT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PT_PT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PT_PT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PT_PT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PT_PT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PT_PT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PT_PT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PT_PT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PT_PT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PT_PT },
         { "WHITE 28",          OPTION_VAL_WHITE_PT_PT },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PT_PT },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PT_PT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PT_PT },
         { "BLUE 03",           OPTION_VAL_BLUE_PT_PT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PT_PT },
         { "GREEN 05",          OPTION_VAL_GREEN_PT_PT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PT_PT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PT_PT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PT_PT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PT_PT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PT_PT },
         { "RED 11",            OPTION_VAL_RED_PT_PT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PT_PT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PT_PT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PT_PT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PT_PT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PT_PT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PT_PT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PT_PT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PT_PT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PT_PT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PT_PT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PT_PT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PT_PT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PT_PT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PT_PT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PT_PT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PT_PT },
         { "WHITE 28",          OPTION_VAL_WHITE_PT_PT },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_PT_PT },
         { "20%",  OPTION_VAL_20_PT_PT },
         { "30%",  OPTION_VAL_30_PT_PT },
         { "40%",  OPTION_VAL_40_PT_PT },
         { "50%",  OPTION_VAL_50_PT_PT },
         { "60%",  OPTION_VAL_60_PT_PT },
         { "70%",  OPTION_VAL_70_PT_PT },
         { "80%",  OPTION_VAL_80_PT_PT },
         { "90%",  OPTION_VAL_90_PT_PT },
         { "100%", OPTION_VAL_100_PT_PT },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_PT_PT },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_PT_PT },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_PT_PT },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_PT_PT },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_PT_PT },
         { "2x", OPTION_VAL_2_O27_PT_PT },
         { "3x", OPTION_VAL_3X_PT_PT },
         { "4x", OPTION_VAL_4_PT_PT },
         { "5x", OPTION_VAL_5X_PT_PT },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PT_PT },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PT_PT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PT_PT },
         { "BLUE 03",           OPTION_VAL_BLUE_PT_PT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PT_PT },
         { "GREEN 05",          OPTION_VAL_GREEN_PT_PT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PT_PT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PT_PT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PT_PT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PT_PT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PT_PT },
         { "RED 11",            OPTION_VAL_RED_PT_PT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PT_PT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PT_PT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PT_PT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PT_PT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PT_PT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PT_PT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PT_PT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PT_PT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PT_PT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PT_PT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PT_PT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PT_PT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PT_PT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PT_PT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PT_PT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PT_PT },
         { "WHITE 28",          OPTION_VAL_WHITE_PT_PT },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PT_PT },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PT_PT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PT_PT },
         { "BLUE 03",           OPTION_VAL_BLUE_PT_PT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PT_PT },
         { "GREEN 05",          OPTION_VAL_GREEN_PT_PT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PT_PT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PT_PT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PT_PT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PT_PT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PT_PT },
         { "RED 11",            OPTION_VAL_RED_PT_PT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PT_PT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PT_PT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PT_PT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PT_PT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PT_PT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PT_PT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PT_PT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PT_PT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PT_PT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PT_PT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PT_PT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PT_PT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PT_PT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PT_PT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PT_PT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PT_PT },
         { "WHITE 28",          OPTION_VAL_WHITE_PT_PT },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_PT_PT },
         { "20%",  OPTION_VAL_20_PT_PT },
         { "30%",  OPTION_VAL_30_PT_PT },
         { "40%",  OPTION_VAL_40_PT_PT },
         { "50%",  OPTION_VAL_50_PT_PT },
         { "60%",  OPTION_VAL_60_PT_PT },
         { "70%",  OPTION_VAL_70_PT_PT },
         { "80%",  OPTION_VAL_80_PT_PT },
         { "90%",  OPTION_VAL_90_PT_PT },
         { "100%", OPTION_VAL_100_PT_PT },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_PT_PT },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_PT_PT },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_PT_PT },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_PT_PT },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_PT_PT },
         { "2x", OPTION_VAL_2_O27_PT_PT },
         { "3x", OPTION_VAL_3X_PT_PT },
         { "4x", OPTION_VAL_4_PT_PT },
         { "5x", OPTION_VAL_5X_PT_PT },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PT_PT },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PT_PT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PT_PT },
         { "BLUE 03",           OPTION_VAL_BLUE_PT_PT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PT_PT },
         { "GREEN 05",          OPTION_VAL_GREEN_PT_PT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PT_PT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PT_PT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PT_PT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PT_PT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PT_PT },
         { "RED 11",            OPTION_VAL_RED_PT_PT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PT_PT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PT_PT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PT_PT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PT_PT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PT_PT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PT_PT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PT_PT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PT_PT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PT_PT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PT_PT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PT_PT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PT_PT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PT_PT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PT_PT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PT_PT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PT_PT },
         { "WHITE 28",          OPTION_VAL_WHITE_PT_PT },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PT_PT },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PT_PT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PT_PT },
         { "BLUE 03",           OPTION_VAL_BLUE_PT_PT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PT_PT },
         { "GREEN 05",          OPTION_VAL_GREEN_PT_PT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PT_PT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PT_PT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PT_PT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PT_PT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PT_PT },
         { "RED 11",            OPTION_VAL_RED_PT_PT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PT_PT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PT_PT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PT_PT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PT_PT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PT_PT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PT_PT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PT_PT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PT_PT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PT_PT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PT_PT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PT_PT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PT_PT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PT_PT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PT_PT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PT_PT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PT_PT },
         { "WHITE 28",          OPTION_VAL_WHITE_PT_PT },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_PT_PT },
         { "20%",  OPTION_VAL_20_PT_PT },
         { "30%",  OPTION_VAL_30_PT_PT },
         { "40%",  OPTION_VAL_40_PT_PT },
         { "50%",  OPTION_VAL_50_PT_PT },
         { "60%",  OPTION_VAL_60_PT_PT },
         { "70%",  OPTION_VAL_70_PT_PT },
         { "80%",  OPTION_VAL_80_PT_PT },
         { "90%",  OPTION_VAL_90_PT_PT },
         { "100%", OPTION_VAL_100_PT_PT },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_pt_pt = {
   option_cats_pt_pt,
   option_defs_pt_pt
};

/* RETRO_LANGUAGE_RU */

#define CATEGORY_SYSTEM_LABEL_RU "Системные"
#define CATEGORY_SYSTEM_INFO_0_RU "Настройка региона, языка, BIOS и основных параметров оборудования."
#define CATEGORY_VIDEO_LABEL_RU "Видео"
#define CATEGORY_VIDEO_INFO_0_RU NULL
#define CATEGORY_PERFORMANCE_LABEL_RU "Производительность"
#define CATEGORY_PERFORMANCE_INFO_0_RU NULL
#define CATEGORY_HACKS_LABEL_RU "Хаки эмуляции"
#define CATEGORY_HACKS_INFO_0_RU NULL
#define CATEGORY_INPUT_LABEL_RU "Ввод"
#define CATEGORY_INPUT_INFO_0_RU NULL
#define CATEGORY_VMU_LABEL_RU NULL
#define CATEGORY_VMU_INFO_0_RU NULL
#define CORE_OPTION_NAME_REGION_LABEL_RU "Регион"
#define OPTION_VAL_JAPAN_RU "Япония"
#define OPTION_VAL_USA_RU "США"
#define OPTION_VAL_EUROPE_RU "Европа"
#define OPTION_VAL_DEFAULT_RU "Стандартный"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_RU "Язык"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_RU "Меняет язык, используемый BIOS и любыми играми, содержащими несколько языков."
#define OPTION_VAL_JAPANESE_RU "Японский"
#define OPTION_VAL_ENGLISH_RU "Английский"
#define OPTION_VAL_GERMAN_RU "Немецкий"
#define OPTION_VAL_FRENCH_RU "Французский"
#define OPTION_VAL_SPANISH_RU "Испанский"
#define OPTION_VAL_ITALIAN_RU "Итальянский"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_RU "HLE BIOS (требуется перезапуск)"
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_RU "Принудительно использовать высокоуровневую эмуляцию BIOS."
#define CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_RU "Загрузка в BIOS (требуется перезапуск)"
#define CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_RU "Загружаться непосредственно в меню Dreamcast BIOS."
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_RU "Включить DSP"
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_RU "Включить эмуляцию звукового DSP (цифрового сигнального процессора) Dreamcast. Улучшает точность генерируемого звука, но повышает системные требования."
#define CORE_OPTION_NAME_FORCE_WINCE_LABEL_RU "Принудительный режим Windows CE"
#define CORE_OPTION_NAME_FORCE_WINCE_INFO_0_RU "Включает полную эмуляцию MMU (блока управления памятью) и другие настройки для игр Windows CE."
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_RU NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_RU NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_RU NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_RU NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_RU "Эмуляция Broadband Adapter"
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_RU "Эмулировать вместо модема широкополосный Ethernet-адаптер (требуется перезапуск)."
#define CORE_OPTION_NAME_UPNP_LABEL_RU "Включить UPnP"
#define CORE_OPTION_NAME_UPNP_INFO_0_RU "Использовать UPnP для автоматической настройки вашего Интернет-роутера для онлайн-игр."
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_RU "Внутреннее разрешение"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_RU "Настройка разрешения рендеринга."
#define OPTION_VAL_320X240_RU NULL
#define OPTION_VAL_640X480_RU NULL
#define OPTION_VAL_800X600_RU NULL
#define OPTION_VAL_960X720_RU NULL
#define OPTION_VAL_1024X768_RU NULL
#define OPTION_VAL_1280X960_RU NULL
#define OPTION_VAL_1440X1080_RU NULL
#define OPTION_VAL_1600X1200_RU NULL
#define OPTION_VAL_1920X1440_RU NULL
#define OPTION_VAL_2560X1920_RU NULL
#define OPTION_VAL_2880X2160_RU NULL
#define OPTION_VAL_3200X2400_RU NULL
#define OPTION_VAL_3840X2880_RU NULL
#define OPTION_VAL_4480X3360_RU NULL
#define OPTION_VAL_5120X3840_RU NULL
#define OPTION_VAL_5760X4320_RU NULL
#define OPTION_VAL_6400X4800_RU NULL
#define OPTION_VAL_7040X5280_RU NULL
#define OPTION_VAL_7680X5760_RU NULL
#define OPTION_VAL_8320X6240_RU NULL
#define OPTION_VAL_8960X6720_RU NULL
#define OPTION_VAL_9600X7200_RU NULL
#define OPTION_VAL_10240X7680_RU NULL
#define OPTION_VAL_10880X8160_RU NULL
#define OPTION_VAL_11520X8640_RU NULL
#define OPTION_VAL_12160X9120_RU NULL
#define OPTION_VAL_12800X9600_RU NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_RU "Тип кабеля"
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_RU "Тип выводимого сигнала. ТВ (Композитный) наиболее поддерживаемый."
#define OPTION_VAL_VGA_RU NULL
#define OPTION_VAL_TV_RGB_RU "ТВ (RGB)"
#define OPTION_VAL_TV_COMPOSITE_RU "ТВ (композитный)"
#define CORE_OPTION_NAME_BROADCAST_LABEL_RU "Стандарт вещания"
#define OPTION_VAL_NTSC_RU NULL
#define OPTION_VAL_PAL_RU "PAL (Мир)"
#define OPTION_VAL_PAL_N_RU "PAL-N (Аргентина, Парагвай, Уругвай)"
#define OPTION_VAL_PAL_M_RU "PAL-M (Бразилия)"
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_RU "Положение экрана"
#define OPTION_VAL_HORIZONTAL_RU "Горизонтальная"
#define OPTION_VAL_VERTICAL_RU "Вертикальная"
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_RU "Альфа-сортировка"
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_RU "По слоям (самый быстрый, но наименее точный)"
#define OPTION_VAL_PER_TRIANGLE_NORMAL_RU "По треугольникам (обычный)"
#define OPTION_VAL_PER_PIXEL_ACCURATE_RU "По пикселям (самый точный, но наиболее медленный)"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_RU NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_RU "Большие значения могут требоваться для правильного отображения высоких разрешений."
#define OPTION_VAL_512MB_RU "512 МБ"
#define OPTION_VAL_1GB_RU "1 ГБ"
#define OPTION_VAL_2GB_RU "2 ГБ"
#define OPTION_VAL_4GB_RU "4 ГБ"
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_RU NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_RU NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_RU NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_RU NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_RU NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_RU NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_RU "MIP-текстурирование"
#define CORE_OPTION_NAME_FOG_LABEL_RU "Эффекты тумана"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_RU "Модификатор объема"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_RU "Функция графического процессора Dreamcast, которая обычно используется играми для отрисовки теней объектов. По умолчанию она должна быть включена - влияние на производительность обычно минимально или незначительно."
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_RU "Анизотропная фильтрация"
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_RU NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_RU "Сглаживание текстур"
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_RU NULL
#define OPTION_VAL_1_RU NULL
#define OPTION_VAL_2_RU NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_RU "Задержка замены кадров"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_RU "Помогает избегать мерцания экрана или артефактов видео. Не рекомендуется на медленных платформах."
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_RU NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_RU NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_RU "Фильтр постобработки PowerVR2"
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_RU "Делает постобработку отображаемого изображения для имитации эффектов, специфичных для графического процессора PowerVR2 и аналоговых видеосигналов."
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_RU "Масштабирование текстур (xBRZ)"
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_RU NULL
#define OPTION_VAL_2_O27_RU NULL
#define OPTION_VAL_4_RU NULL
#define OPTION_VAL_6_RU NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_RU NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_RU NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_RU NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_RU "Рендеринг в отдельном потоке"
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_RU "Запускает графический и центральный процессоры в разных потоках. Настоятельно рекомендуется."
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_RU "Автоматический пропуск кадров"
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_RU NULL
#define OPTION_VAL_SOME_RU "Нормальный"
#define OPTION_VAL_MORE_RU NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_RU "Пропуск кадров"
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_RU NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_RU "Широкоэкранные чит-коды (требуется перезапуск)"
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_RU "Активирует чит-коды, которые позволяют отображать некоторые игры в широкоэкранном формате."
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_RU "Широкоэкранный хак"
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_RU "Отрисовывать геометрию вне нормального соотношения 4:3. Может вызвать графические артефакты в показываемых областях."
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_RU "Быстрая загрузка GD-ROM (неточно)"
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_RU "Ускоряет загрузку GD-ROM."
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_RU "Загружать пользовательские текстуры"
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_RU "Дамп текстур"
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_RU "Мёртвая зона аналогового стика"
#define OPTION_VAL_0_RU NULL
#define OPTION_VAL_5_RU NULL
#define OPTION_VAL_10_RU NULL
#define OPTION_VAL_15_RU NULL
#define OPTION_VAL_20_RU NULL
#define OPTION_VAL_25_RU NULL
#define OPTION_VAL_30_RU NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_RU "Мёртвая зона триггеров"
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_RU "Цифровые триггеры"
#define CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_RU NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_RU "Включает обратную связь контроллера."
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_RU NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_RU NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_RU "Показать настройки светового пистолета"
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_RU NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_RU NULL
#define OPTION_VAL_WHITE_RU "Белый"
#define OPTION_VAL_RED_RU "Красный"
#define OPTION_VAL_GREEN_RU "Зелёный"
#define OPTION_VAL_BLUE_RU "Голубой"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_RU NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_RU NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_RU NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_RU "Отдельные блоки визуальной памяти (VMU) для игр"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_RU "VMU для каждой игры"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_RU NULL
#define OPTION_VAL_VMU_A1_RU NULL
#define OPTION_VAL_ALL_VMUS_RU NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_RU "Показывать настройки отображения блока визуальной памяти (VMU)"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_RU "Показывать настройки отображения VMU"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_RU NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_RU "Показывать экран VMU 1"
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_RU "Положение экрана VMU 1"
#define OPTION_VAL_UPPER_LEFT_RU "Вверху слева"
#define OPTION_VAL_UPPER_RIGHT_RU "Вверху справа"
#define OPTION_VAL_LOWER_LEFT_RU "Внизу слева"
#define OPTION_VAL_LOWER_RIGHT_RU "Внизу справа"
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_RU "Размер экрана VMU 1"
#define OPTION_VAL_1X_RU NULL
#define OPTION_VAL_3X_RU NULL
#define OPTION_VAL_5X_RU NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_RU "Цвет вкл. пикселей VMU 1"
#define OPTION_VAL_DEFAULT_ON_00_RU "По умолчанию ВКЛ"
#define OPTION_VAL_DEFAULT_OFF_01_RU "По умолчанию ВЫКЛ"
#define OPTION_VAL_BLACK_02_RU "Чёрный"
#define OPTION_VAL_LIGHT_BLUE_04_RU "Светло-синий"
#define OPTION_VAL_CYAN_06_RU "Сине-зелёный"
#define OPTION_VAL_CYAN_BLUE_07_RU "Голубовато-синий"
#define OPTION_VAL_LIGHT_GREEN_08_RU "Светло-зелёный"
#define OPTION_VAL_CYAN_GREEN_09_RU "Голубовато-зелёный"
#define OPTION_VAL_LIGHT_CYAN_10_RU "Светло-голубой"
#define OPTION_VAL_PURPLE_12_RU "Фиолетовый"
#define OPTION_VAL_LIGHT_PURPLE_13_RU NULL
#define OPTION_VAL_YELLOW_14_RU "Жёлтая"
#define OPTION_VAL_GRAY_15_RU "Серый"
#define OPTION_VAL_LIGHT_PURPLE_2_16_RU NULL
#define OPTION_VAL_LIGHT_GREEN_2_17_RU NULL
#define OPTION_VAL_LIGHT_GREEN_3_18_RU NULL
#define OPTION_VAL_LIGHT_CYAN_2_19_RU NULL
#define OPTION_VAL_LIGHT_RED_2_20_RU NULL
#define OPTION_VAL_MAGENTA_21_RU NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_RU NULL
#define OPTION_VAL_LIGHT_ORANGE_23_RU NULL
#define OPTION_VAL_ORANGE_24_RU "Оранжевый"
#define OPTION_VAL_LIGHT_PURPLE_4_25_RU NULL
#define OPTION_VAL_LIGHT_YELLOW_26_RU NULL
#define OPTION_VAL_LIGHT_YELLOW_2_27_RU NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_RU "Цвет выкл. пикселей VMU 1"
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_RU "Видимость экрана VMU 1"
#define OPTION_VAL_40_RU NULL
#define OPTION_VAL_50_RU NULL
#define OPTION_VAL_60_RU NULL
#define OPTION_VAL_70_RU NULL
#define OPTION_VAL_80_RU NULL
#define OPTION_VAL_90_RU NULL
#define OPTION_VAL_100_RU NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_RU "Показывать экран VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_RU "Положение экрана VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_RU "Размер экрана VMU 2"
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_RU "Цвет вкл. пикселей VMU 2"
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_RU "Цвет выкл. пикселей VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_RU "Видимость экрана VMU 2"
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_RU "Показывать экран VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_RU "Положение экрана VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_RU "Размер экрана VMU 3"
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_RU "Цвет вкл. пикселей VMU 3"
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_RU "Цвет выкл. пикселей VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_RU "Видимость экрана VMU 3"
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_RU "Показывать экран VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_RU "Положение экрана VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_RU "Размер экрана VMU 4"
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_RU "Цвет вкл. пикселей VMU 4"
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_RU "Цвет выкл. пикселей VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_RU "Видимость экрана VMU 4"

struct retro_core_option_v2_category option_cats_ru[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_RU,
      CATEGORY_SYSTEM_INFO_0_RU
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_RU,
      CATEGORY_VIDEO_INFO_0_RU
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_RU,
      CATEGORY_PERFORMANCE_INFO_0_RU
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_RU,
      CATEGORY_HACKS_INFO_0_RU
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_RU,
      CATEGORY_INPUT_INFO_0_RU
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_RU,
      CATEGORY_VMU_INFO_0_RU
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_ru[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_RU,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_RU },
         { "USA",     OPTION_VAL_USA_RU },
         { "Europe",  OPTION_VAL_EUROPE_RU },
         { "Default", OPTION_VAL_DEFAULT_RU },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_RU,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_RU },
         { "English",  OPTION_VAL_ENGLISH_RU },
         { "German",   OPTION_VAL_GERMAN_RU },
         { "French",   OPTION_VAL_FRENCH_RU },
         { "Spanish",  OPTION_VAL_SPANISH_RU },
         { "Italian",  OPTION_VAL_ITALIAN_RU },
         { "Default",  OPTION_VAL_DEFAULT_RU },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_RU,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_boot_to_bios",
      CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_RU,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_RU,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_force_wince",
      CORE_OPTION_NAME_FORCE_WINCE_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_FORCE_WINCE_INFO_0_RU,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_RU,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_RU,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_RU,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_RU,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_RU,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_RU },
         { "640x480",    OPTION_VAL_640X480_RU },
         { "800x600",    OPTION_VAL_800X600_RU },
         { "960x720",    OPTION_VAL_960X720_RU },
         { "1024x768",   OPTION_VAL_1024X768_RU },
         { "1280x960",   OPTION_VAL_1280X960_RU },
         { "1440x1080",  OPTION_VAL_1440X1080_RU },
         { "1600x1200",  OPTION_VAL_1600X1200_RU },
         { "1920x1440",  OPTION_VAL_1920X1440_RU },
         { "2560x1920",  OPTION_VAL_2560X1920_RU },
         { "2880x2160",  OPTION_VAL_2880X2160_RU },
         { "3200x2400",  OPTION_VAL_3200X2400_RU },
         { "3840x2880",  OPTION_VAL_3840X2880_RU },
         { "4480x3360",  OPTION_VAL_4480X3360_RU },
         { "5120x3840",  OPTION_VAL_5120X3840_RU },
         { "5760x4320",  OPTION_VAL_5760X4320_RU },
         { "6400x4800",  OPTION_VAL_6400X4800_RU },
         { "7040x5280",  OPTION_VAL_7040X5280_RU },
         { "7680x5760",  OPTION_VAL_7680X5760_RU },
         { "8320x6240",  OPTION_VAL_8320X6240_RU },
         { "8960x6720",  OPTION_VAL_8960X6720_RU },
         { "9600x7200",  OPTION_VAL_9600X7200_RU },
         { "10240x7680", OPTION_VAL_10240X7680_RU },
         { "10880x8160", OPTION_VAL_10880X8160_RU },
         { "11520x8640", OPTION_VAL_11520X8640_RU },
         { "12160x9120", OPTION_VAL_12160X9120_RU },
         { "12800x9600", OPTION_VAL_12800X9600_RU },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_RU,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_RU },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_RU },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_RU },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_RU,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_RU },
         { "PAL",     OPTION_VAL_PAL_RU },
         { "PAL_N",   OPTION_VAL_PAL_N_RU },
         { "PAL_M",   OPTION_VAL_PAL_M_RU },
         { "Default", OPTION_VAL_DEFAULT_RU },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_RU,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_RU },
         { "vertical",   OPTION_VAL_VERTICAL_RU },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_RU,
      NULL,
      "",
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_RU },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_RU },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_RU },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_RU,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_RU },
         { "1GB",   OPTION_VAL_1GB_RU },
         { "2GB",   OPTION_VAL_2GB_RU },
         { "4GB",   OPTION_VAL_4GB_RU },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_RU,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_RU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_RU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_RU,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_RU,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_RU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_RU,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_RU,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_RU },
         { "1",  OPTION_VAL_1_RU },
         { "2",  OPTION_VAL_2_RU },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_RU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_RU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_RU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_RU,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O27_RU },
         { "4",  OPTION_VAL_4_RU },
         { "6",  OPTION_VAL_6_RU },
         { NULL, NULL },
      },
      "1",
   },
   {/* TODO: needs clarification */
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_RU,
      NULL,
      "",
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_RU,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_RU,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_RU,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_RU,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_RU },
         { "more", OPTION_VAL_MORE_RU },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_RU,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_RU,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_RU,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_RU,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_RU,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: probably needs explanation */
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_RU,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_RU,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_RU },
         { "5%",  OPTION_VAL_5_RU },
         { "10%", OPTION_VAL_10_RU },
         { "15%", OPTION_VAL_15_RU },
         { "20%", OPTION_VAL_20_RU },
         { "25%", OPTION_VAL_25_RU },
         { "30%", OPTION_VAL_30_RU },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_RU,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_RU },
         { "5%",  OPTION_VAL_5_RU },
         { "10%", OPTION_VAL_10_RU },
         { "15%", OPTION_VAL_15_RU },
         { "20%", OPTION_VAL_20_RU },
         { "25%", OPTION_VAL_25_RU },
         { "30%", OPTION_VAL_30_RU },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_RU,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_purupuru",
      CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_RU,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_RU,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_RU,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_RU,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_RU },
         { "Red",      OPTION_VAL_RED_RU },
         { "Green",    OPTION_VAL_GREEN_RU },
         { "Blue",     OPTION_VAL_BLUE_RU },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_RU,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_RU },
         { "Red",      OPTION_VAL_RED_RU },
         { "Green",    OPTION_VAL_GREEN_RU },
         { "Blue",     OPTION_VAL_BLUE_RU },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_RU,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_RU },
         { "Red",      OPTION_VAL_RED_RU },
         { "Green",    OPTION_VAL_GREEN_RU },
         { "Blue",     OPTION_VAL_BLUE_RU },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_RU,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_RU },
         { "Red",      OPTION_VAL_RED_RU },
         { "Green",    OPTION_VAL_GREEN_RU },
         { "Blue",     OPTION_VAL_BLUE_RU },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_RU,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_RU,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_RU,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_RU },
         { "All VMUs", OPTION_VAL_ALL_VMUS_RU },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_RU,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_RU,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_RU,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_RU },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_RU },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_RU },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_RU },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_RU },
         { "2x", OPTION_VAL_2_O27_RU },
         { "3x", OPTION_VAL_3X_RU },
         { "4x", OPTION_VAL_4_RU },
         { "5x", OPTION_VAL_5X_RU },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_RU },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_RU },
         { "BLACK 02",          OPTION_VAL_BLACK_02_RU },
         { "BLUE 03",           OPTION_VAL_BLUE_RU },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_RU },
         { "GREEN 05",          OPTION_VAL_GREEN_RU },
         { "CYAN 06",           OPTION_VAL_CYAN_06_RU },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_RU },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_RU },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_RU },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_RU },
         { "RED 11",            OPTION_VAL_RED_RU },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_RU },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_RU },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_RU },
         { "GRAY 15",           OPTION_VAL_GRAY_15_RU },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_RU },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_RU },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_RU },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_RU },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_RU },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_RU },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_RU },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_RU },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_RU },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_RU },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_RU },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_RU },
         { "WHITE 28",          OPTION_VAL_WHITE_RU },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_RU },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_RU },
         { "BLACK 02",          OPTION_VAL_BLACK_02_RU },
         { "BLUE 03",           OPTION_VAL_BLUE_RU },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_RU },
         { "GREEN 05",          OPTION_VAL_GREEN_RU },
         { "CYAN 06",           OPTION_VAL_CYAN_06_RU },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_RU },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_RU },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_RU },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_RU },
         { "RED 11",            OPTION_VAL_RED_RU },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_RU },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_RU },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_RU },
         { "GRAY 15",           OPTION_VAL_GRAY_15_RU },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_RU },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_RU },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_RU },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_RU },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_RU },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_RU },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_RU },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_RU },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_RU },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_RU },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_RU },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_RU },
         { "WHITE 28",          OPTION_VAL_WHITE_RU },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_RU },
         { "20%",  OPTION_VAL_20_RU },
         { "30%",  OPTION_VAL_30_RU },
         { "40%",  OPTION_VAL_40_RU },
         { "50%",  OPTION_VAL_50_RU },
         { "60%",  OPTION_VAL_60_RU },
         { "70%",  OPTION_VAL_70_RU },
         { "80%",  OPTION_VAL_80_RU },
         { "90%",  OPTION_VAL_90_RU },
         { "100%", OPTION_VAL_100_RU },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_RU },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_RU },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_RU },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_RU },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_RU },
         { "2x", OPTION_VAL_2_O27_RU },
         { "3x", OPTION_VAL_3X_RU },
         { "4x", OPTION_VAL_4_RU },
         { "5x", OPTION_VAL_5X_RU },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_RU },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_RU },
         { "BLACK 02",          OPTION_VAL_BLACK_02_RU },
         { "BLUE 03",           OPTION_VAL_BLUE_RU },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_RU },
         { "GREEN 05",          OPTION_VAL_GREEN_RU },
         { "CYAN 06",           OPTION_VAL_CYAN_06_RU },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_RU },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_RU },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_RU },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_RU },
         { "RED 11",            OPTION_VAL_RED_RU },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_RU },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_RU },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_RU },
         { "GRAY 15",           OPTION_VAL_GRAY_15_RU },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_RU },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_RU },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_RU },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_RU },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_RU },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_RU },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_RU },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_RU },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_RU },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_RU },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_RU },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_RU },
         { "WHITE 28",          OPTION_VAL_WHITE_RU },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_RU },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_RU },
         { "BLACK 02",          OPTION_VAL_BLACK_02_RU },
         { "BLUE 03",           OPTION_VAL_BLUE_RU },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_RU },
         { "GREEN 05",          OPTION_VAL_GREEN_RU },
         { "CYAN 06",           OPTION_VAL_CYAN_06_RU },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_RU },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_RU },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_RU },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_RU },
         { "RED 11",            OPTION_VAL_RED_RU },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_RU },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_RU },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_RU },
         { "GRAY 15",           OPTION_VAL_GRAY_15_RU },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_RU },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_RU },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_RU },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_RU },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_RU },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_RU },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_RU },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_RU },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_RU },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_RU },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_RU },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_RU },
         { "WHITE 28",          OPTION_VAL_WHITE_RU },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_RU },
         { "20%",  OPTION_VAL_20_RU },
         { "30%",  OPTION_VAL_30_RU },
         { "40%",  OPTION_VAL_40_RU },
         { "50%",  OPTION_VAL_50_RU },
         { "60%",  OPTION_VAL_60_RU },
         { "70%",  OPTION_VAL_70_RU },
         { "80%",  OPTION_VAL_80_RU },
         { "90%",  OPTION_VAL_90_RU },
         { "100%", OPTION_VAL_100_RU },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_RU },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_RU },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_RU },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_RU },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_RU },
         { "2x", OPTION_VAL_2_O27_RU },
         { "3x", OPTION_VAL_3X_RU },
         { "4x", OPTION_VAL_4_RU },
         { "5x", OPTION_VAL_5X_RU },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_RU },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_RU },
         { "BLACK 02",          OPTION_VAL_BLACK_02_RU },
         { "BLUE 03",           OPTION_VAL_BLUE_RU },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_RU },
         { "GREEN 05",          OPTION_VAL_GREEN_RU },
         { "CYAN 06",           OPTION_VAL_CYAN_06_RU },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_RU },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_RU },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_RU },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_RU },
         { "RED 11",            OPTION_VAL_RED_RU },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_RU },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_RU },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_RU },
         { "GRAY 15",           OPTION_VAL_GRAY_15_RU },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_RU },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_RU },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_RU },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_RU },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_RU },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_RU },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_RU },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_RU },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_RU },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_RU },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_RU },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_RU },
         { "WHITE 28",          OPTION_VAL_WHITE_RU },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_RU },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_RU },
         { "BLACK 02",          OPTION_VAL_BLACK_02_RU },
         { "BLUE 03",           OPTION_VAL_BLUE_RU },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_RU },
         { "GREEN 05",          OPTION_VAL_GREEN_RU },
         { "CYAN 06",           OPTION_VAL_CYAN_06_RU },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_RU },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_RU },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_RU },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_RU },
         { "RED 11",            OPTION_VAL_RED_RU },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_RU },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_RU },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_RU },
         { "GRAY 15",           OPTION_VAL_GRAY_15_RU },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_RU },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_RU },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_RU },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_RU },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_RU },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_RU },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_RU },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_RU },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_RU },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_RU },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_RU },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_RU },
         { "WHITE 28",          OPTION_VAL_WHITE_RU },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_RU },
         { "20%",  OPTION_VAL_20_RU },
         { "30%",  OPTION_VAL_30_RU },
         { "40%",  OPTION_VAL_40_RU },
         { "50%",  OPTION_VAL_50_RU },
         { "60%",  OPTION_VAL_60_RU },
         { "70%",  OPTION_VAL_70_RU },
         { "80%",  OPTION_VAL_80_RU },
         { "90%",  OPTION_VAL_90_RU },
         { "100%", OPTION_VAL_100_RU },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_RU },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_RU },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_RU },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_RU },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_RU },
         { "2x", OPTION_VAL_2_O27_RU },
         { "3x", OPTION_VAL_3X_RU },
         { "4x", OPTION_VAL_4_RU },
         { "5x", OPTION_VAL_5X_RU },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_RU },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_RU },
         { "BLACK 02",          OPTION_VAL_BLACK_02_RU },
         { "BLUE 03",           OPTION_VAL_BLUE_RU },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_RU },
         { "GREEN 05",          OPTION_VAL_GREEN_RU },
         { "CYAN 06",           OPTION_VAL_CYAN_06_RU },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_RU },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_RU },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_RU },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_RU },
         { "RED 11",            OPTION_VAL_RED_RU },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_RU },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_RU },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_RU },
         { "GRAY 15",           OPTION_VAL_GRAY_15_RU },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_RU },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_RU },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_RU },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_RU },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_RU },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_RU },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_RU },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_RU },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_RU },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_RU },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_RU },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_RU },
         { "WHITE 28",          OPTION_VAL_WHITE_RU },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_RU },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_RU },
         { "BLACK 02",          OPTION_VAL_BLACK_02_RU },
         { "BLUE 03",           OPTION_VAL_BLUE_RU },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_RU },
         { "GREEN 05",          OPTION_VAL_GREEN_RU },
         { "CYAN 06",           OPTION_VAL_CYAN_06_RU },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_RU },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_RU },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_RU },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_RU },
         { "RED 11",            OPTION_VAL_RED_RU },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_RU },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_RU },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_RU },
         { "GRAY 15",           OPTION_VAL_GRAY_15_RU },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_RU },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_RU },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_RU },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_RU },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_RU },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_RU },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_RU },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_RU },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_RU },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_RU },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_RU },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_RU },
         { "WHITE 28",          OPTION_VAL_WHITE_RU },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_RU },
         { "20%",  OPTION_VAL_20_RU },
         { "30%",  OPTION_VAL_30_RU },
         { "40%",  OPTION_VAL_40_RU },
         { "50%",  OPTION_VAL_50_RU },
         { "60%",  OPTION_VAL_60_RU },
         { "70%",  OPTION_VAL_70_RU },
         { "80%",  OPTION_VAL_80_RU },
         { "90%",  OPTION_VAL_90_RU },
         { "100%", OPTION_VAL_100_RU },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_ru = {
   option_cats_ru,
   option_defs_ru
};

/* RETRO_LANGUAGE_SK */

#define CATEGORY_SYSTEM_LABEL_SK "Systém"
#define CATEGORY_SYSTEM_INFO_0_SK NULL
#define CATEGORY_VIDEO_LABEL_SK NULL
#define CATEGORY_VIDEO_INFO_0_SK NULL
#define CATEGORY_PERFORMANCE_LABEL_SK "Výkon"
#define CATEGORY_PERFORMANCE_INFO_0_SK NULL
#define CATEGORY_HACKS_LABEL_SK "Emulačné hacky"
#define CATEGORY_HACKS_INFO_0_SK NULL
#define CATEGORY_INPUT_LABEL_SK "Vstup"
#define CATEGORY_INPUT_INFO_0_SK NULL
#define CATEGORY_VMU_LABEL_SK NULL
#define CATEGORY_VMU_INFO_0_SK NULL
#define CORE_OPTION_NAME_REGION_LABEL_SK "Región"
#define OPTION_VAL_JAPAN_SK "Japonsko"
#define OPTION_VAL_USA_SK NULL
#define OPTION_VAL_EUROPE_SK "Európa"
#define OPTION_VAL_DEFAULT_SK "Predvolené"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_SK "Jazyk"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_SK NULL
#define OPTION_VAL_JAPANESE_SK "Japončina"
#define OPTION_VAL_ENGLISH_SK "Angličtina"
#define OPTION_VAL_GERMAN_SK "Nemčina"
#define OPTION_VAL_FRENCH_SK "Francúzština"
#define OPTION_VAL_SPANISH_SK "Španielčina"
#define OPTION_VAL_ITALIAN_SK "Taliančina"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_SK NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_SK NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_SK NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_SK NULL
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_SK "Povoliť DSP"
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_SK NULL
#define CORE_OPTION_NAME_FORCE_WINCE_LABEL_SK NULL
#define CORE_OPTION_NAME_FORCE_WINCE_INFO_0_SK NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_SK NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_SK NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_SK NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_SK NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_SK NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_SK NULL
#define CORE_OPTION_NAME_UPNP_LABEL_SK NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_SK NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_SK "Vnútorné rozlíšenie"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_SK NULL
#define OPTION_VAL_320X240_SK NULL
#define OPTION_VAL_640X480_SK NULL
#define OPTION_VAL_800X600_SK NULL
#define OPTION_VAL_960X720_SK NULL
#define OPTION_VAL_1024X768_SK NULL
#define OPTION_VAL_1280X960_SK NULL
#define OPTION_VAL_1440X1080_SK NULL
#define OPTION_VAL_1600X1200_SK NULL
#define OPTION_VAL_1920X1440_SK NULL
#define OPTION_VAL_2560X1920_SK NULL
#define OPTION_VAL_2880X2160_SK NULL
#define OPTION_VAL_3200X2400_SK NULL
#define OPTION_VAL_3840X2880_SK NULL
#define OPTION_VAL_4480X3360_SK NULL
#define OPTION_VAL_5120X3840_SK NULL
#define OPTION_VAL_5760X4320_SK NULL
#define OPTION_VAL_6400X4800_SK NULL
#define OPTION_VAL_7040X5280_SK NULL
#define OPTION_VAL_7680X5760_SK NULL
#define OPTION_VAL_8320X6240_SK NULL
#define OPTION_VAL_8960X6720_SK NULL
#define OPTION_VAL_9600X7200_SK NULL
#define OPTION_VAL_10240X7680_SK NULL
#define OPTION_VAL_10880X8160_SK NULL
#define OPTION_VAL_11520X8640_SK NULL
#define OPTION_VAL_12160X9120_SK NULL
#define OPTION_VAL_12800X9600_SK NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_SK NULL
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_SK NULL
#define OPTION_VAL_VGA_SK NULL
#define OPTION_VAL_TV_RGB_SK NULL
#define OPTION_VAL_TV_COMPOSITE_SK "TV (composite)"
#define CORE_OPTION_NAME_BROADCAST_LABEL_SK NULL
#define OPTION_VAL_NTSC_SK NULL
#define OPTION_VAL_PAL_SK "PAL (svet)"
#define OPTION_VAL_PAL_N_SK "PAL-N (Argentína, Paraguaj, Uruguaj)"
#define OPTION_VAL_PAL_M_SK "PAL-M (Brazília)"
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_SK "Orientácia Obrazovky"
#define OPTION_VAL_HORIZONTAL_SK "Vodorovne"
#define OPTION_VAL_VERTICAL_SK "Zvislo"
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_SK NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_SK NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_SK NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_SK NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_SK NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_SK NULL
#define OPTION_VAL_512MB_SK "512 MB"
#define OPTION_VAL_1GB_SK NULL
#define OPTION_VAL_2GB_SK NULL
#define OPTION_VAL_4GB_SK NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_SK NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_SK NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_SK NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_SK NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_SK NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_SK NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_SK NULL
#define CORE_OPTION_NAME_FOG_LABEL_SK NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_SK NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_SK NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_SK NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_SK NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_SK "Filtrovanie textúr"
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_SK NULL
#define OPTION_VAL_1_SK NULL
#define OPTION_VAL_2_SK NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_SK NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_SK NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_SK NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_SK NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_SK NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_SK NULL
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_SK NULL
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_SK NULL
#define OPTION_VAL_2_O27_SK NULL
#define OPTION_VAL_4_SK NULL
#define OPTION_VAL_6_SK NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_SK NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_SK NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_SK NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_SK NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_SK NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_SK NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_SK NULL
#define OPTION_VAL_SOME_SK "Normálne"
#define OPTION_VAL_MORE_SK "Maximálne"
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_SK "Preskakovanie snímkov"
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_SK NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_SK NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_SK NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_SK NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_SK NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_SK NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_SK NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_SK NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_SK NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_SK NULL
#define OPTION_VAL_0_SK NULL
#define OPTION_VAL_5_SK NULL
#define OPTION_VAL_10_SK NULL
#define OPTION_VAL_15_SK NULL
#define OPTION_VAL_20_SK NULL
#define OPTION_VAL_25_SK NULL
#define OPTION_VAL_30_SK NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_SK NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_SK NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_SK NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_SK NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_SK NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_SK NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_SK NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_SK NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_SK NULL
#define OPTION_VAL_WHITE_SK "Biela"
#define OPTION_VAL_RED_SK "Červená"
#define OPTION_VAL_GREEN_SK "Zelená"
#define OPTION_VAL_BLUE_SK "Modrá"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_SK NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_SK NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_SK NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_SK NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_SK NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_SK NULL
#define OPTION_VAL_VMU_A1_SK NULL
#define OPTION_VAL_ALL_VMUS_SK "Všetky VMU"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_SK NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_SK NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_SK NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_SK NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_SK NULL
#define OPTION_VAL_UPPER_LEFT_SK "Hore vľavo"
#define OPTION_VAL_UPPER_RIGHT_SK "Hore vpravo"
#define OPTION_VAL_LOWER_LEFT_SK NULL
#define OPTION_VAL_LOWER_RIGHT_SK NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_SK NULL
#define OPTION_VAL_1X_SK NULL
#define OPTION_VAL_3X_SK NULL
#define OPTION_VAL_5X_SK NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_SK NULL
#define OPTION_VAL_DEFAULT_ON_00_SK "Predvolene zapnuté"
#define OPTION_VAL_DEFAULT_OFF_01_SK "Predvolene vypnuté"
#define OPTION_VAL_BLACK_02_SK "Čierna"
#define OPTION_VAL_LIGHT_BLUE_04_SK "Svetlomodrá"
#define OPTION_VAL_CYAN_06_SK "Azúrová"
#define OPTION_VAL_CYAN_BLUE_07_SK "Tyrkysová modrá"
#define OPTION_VAL_LIGHT_GREEN_08_SK "Svetlozelená"
#define OPTION_VAL_CYAN_GREEN_09_SK "Tyrkysová zelená"
#define OPTION_VAL_LIGHT_CYAN_10_SK "Svetloazúrová"
#define OPTION_VAL_PURPLE_12_SK "Fialová"
#define OPTION_VAL_LIGHT_PURPLE_13_SK "Svetlofialová"
#define OPTION_VAL_YELLOW_14_SK "Žltá"
#define OPTION_VAL_GRAY_15_SK "Šedá"
#define OPTION_VAL_LIGHT_PURPLE_2_16_SK "Svetlofialová (2)"
#define OPTION_VAL_LIGHT_GREEN_2_17_SK "Svetlozelená (2)"
#define OPTION_VAL_LIGHT_GREEN_3_18_SK "Svetlozelená (3)"
#define OPTION_VAL_LIGHT_CYAN_2_19_SK NULL
#define OPTION_VAL_LIGHT_RED_2_20_SK "Svetločervená (2)"
#define OPTION_VAL_MAGENTA_21_SK "Purpurová"
#define OPTION_VAL_LIGHT_PURPLE_3_22_SK "Svetlofialová (3)"
#define OPTION_VAL_LIGHT_ORANGE_23_SK "Svetlooranžová"
#define OPTION_VAL_ORANGE_24_SK "Oranžová"
#define OPTION_VAL_LIGHT_PURPLE_4_25_SK "Svetlofialová (4)"
#define OPTION_VAL_LIGHT_YELLOW_26_SK "Svetložltá"
#define OPTION_VAL_LIGHT_YELLOW_2_27_SK "Svetložltá (2)"
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_SK NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_SK NULL
#define OPTION_VAL_40_SK NULL
#define OPTION_VAL_50_SK NULL
#define OPTION_VAL_60_SK NULL
#define OPTION_VAL_70_SK NULL
#define OPTION_VAL_80_SK NULL
#define OPTION_VAL_90_SK NULL
#define OPTION_VAL_100_SK NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_SK NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_SK NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_SK NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_SK NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_SK NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_SK NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_SK NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_SK NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_SK NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_SK NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_SK NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_SK NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_SK NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_SK NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_SK NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_SK NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_SK NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_SK NULL

struct retro_core_option_v2_category option_cats_sk[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_SK,
      CATEGORY_SYSTEM_INFO_0_SK
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_SK,
      CATEGORY_VIDEO_INFO_0_SK
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_SK,
      CATEGORY_PERFORMANCE_INFO_0_SK
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_SK,
      CATEGORY_HACKS_INFO_0_SK
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_SK,
      CATEGORY_INPUT_INFO_0_SK
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_SK,
      CATEGORY_VMU_INFO_0_SK
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_sk[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_SK,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_SK },
         { "USA",     OPTION_VAL_USA_SK },
         { "Europe",  OPTION_VAL_EUROPE_SK },
         { "Default", OPTION_VAL_DEFAULT_SK },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_SK,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_SK },
         { "English",  OPTION_VAL_ENGLISH_SK },
         { "German",   OPTION_VAL_GERMAN_SK },
         { "French",   OPTION_VAL_FRENCH_SK },
         { "Spanish",  OPTION_VAL_SPANISH_SK },
         { "Italian",  OPTION_VAL_ITALIAN_SK },
         { "Default",  OPTION_VAL_DEFAULT_SK },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_SK,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_boot_to_bios",
      CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_SK,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_SK,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_force_wince",
      CORE_OPTION_NAME_FORCE_WINCE_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_FORCE_WINCE_INFO_0_SK,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_SK,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_SK,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_SK,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_SK,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_SK,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_SK },
         { "640x480",    OPTION_VAL_640X480_SK },
         { "800x600",    OPTION_VAL_800X600_SK },
         { "960x720",    OPTION_VAL_960X720_SK },
         { "1024x768",   OPTION_VAL_1024X768_SK },
         { "1280x960",   OPTION_VAL_1280X960_SK },
         { "1440x1080",  OPTION_VAL_1440X1080_SK },
         { "1600x1200",  OPTION_VAL_1600X1200_SK },
         { "1920x1440",  OPTION_VAL_1920X1440_SK },
         { "2560x1920",  OPTION_VAL_2560X1920_SK },
         { "2880x2160",  OPTION_VAL_2880X2160_SK },
         { "3200x2400",  OPTION_VAL_3200X2400_SK },
         { "3840x2880",  OPTION_VAL_3840X2880_SK },
         { "4480x3360",  OPTION_VAL_4480X3360_SK },
         { "5120x3840",  OPTION_VAL_5120X3840_SK },
         { "5760x4320",  OPTION_VAL_5760X4320_SK },
         { "6400x4800",  OPTION_VAL_6400X4800_SK },
         { "7040x5280",  OPTION_VAL_7040X5280_SK },
         { "7680x5760",  OPTION_VAL_7680X5760_SK },
         { "8320x6240",  OPTION_VAL_8320X6240_SK },
         { "8960x6720",  OPTION_VAL_8960X6720_SK },
         { "9600x7200",  OPTION_VAL_9600X7200_SK },
         { "10240x7680", OPTION_VAL_10240X7680_SK },
         { "10880x8160", OPTION_VAL_10880X8160_SK },
         { "11520x8640", OPTION_VAL_11520X8640_SK },
         { "12160x9120", OPTION_VAL_12160X9120_SK },
         { "12800x9600", OPTION_VAL_12800X9600_SK },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_SK,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_SK },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_SK },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_SK },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_SK,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_SK },
         { "PAL",     OPTION_VAL_PAL_SK },
         { "PAL_N",   OPTION_VAL_PAL_N_SK },
         { "PAL_M",   OPTION_VAL_PAL_M_SK },
         { "Default", OPTION_VAL_DEFAULT_SK },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_SK,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_SK },
         { "vertical",   OPTION_VAL_VERTICAL_SK },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_SK,
      NULL,
      "",
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_SK },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_SK },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_SK },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_SK,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_SK },
         { "1GB",   OPTION_VAL_1GB_SK },
         { "2GB",   OPTION_VAL_2GB_SK },
         { "4GB",   OPTION_VAL_4GB_SK },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_SK,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_SK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_SK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_SK,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_SK,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_SK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_SK,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_SK,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_SK },
         { "1",  OPTION_VAL_1_SK },
         { "2",  OPTION_VAL_2_SK },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_SK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_SK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_SK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_SK,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O27_SK },
         { "4",  OPTION_VAL_4_SK },
         { "6",  OPTION_VAL_6_SK },
         { NULL, NULL },
      },
      "1",
   },
   {/* TODO: needs clarification */
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_SK,
      NULL,
      "",
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_SK,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_SK,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_SK,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_SK,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_SK },
         { "more", OPTION_VAL_MORE_SK },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_SK,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_SK,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_SK,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_SK,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_SK,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: probably needs explanation */
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_SK,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_SK,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_SK },
         { "5%",  OPTION_VAL_5_SK },
         { "10%", OPTION_VAL_10_SK },
         { "15%", OPTION_VAL_15_SK },
         { "20%", OPTION_VAL_20_SK },
         { "25%", OPTION_VAL_25_SK },
         { "30%", OPTION_VAL_30_SK },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_SK,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_SK },
         { "5%",  OPTION_VAL_5_SK },
         { "10%", OPTION_VAL_10_SK },
         { "15%", OPTION_VAL_15_SK },
         { "20%", OPTION_VAL_20_SK },
         { "25%", OPTION_VAL_25_SK },
         { "30%", OPTION_VAL_30_SK },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_SK,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_purupuru",
      CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_SK,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_SK,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_SK,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_SK,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_SK },
         { "Red",      OPTION_VAL_RED_SK },
         { "Green",    OPTION_VAL_GREEN_SK },
         { "Blue",     OPTION_VAL_BLUE_SK },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_SK,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_SK },
         { "Red",      OPTION_VAL_RED_SK },
         { "Green",    OPTION_VAL_GREEN_SK },
         { "Blue",     OPTION_VAL_BLUE_SK },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_SK,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_SK },
         { "Red",      OPTION_VAL_RED_SK },
         { "Green",    OPTION_VAL_GREEN_SK },
         { "Blue",     OPTION_VAL_BLUE_SK },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_SK,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_SK },
         { "Red",      OPTION_VAL_RED_SK },
         { "Green",    OPTION_VAL_GREEN_SK },
         { "Blue",     OPTION_VAL_BLUE_SK },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_SK,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_SK,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_SK,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_SK },
         { "All VMUs", OPTION_VAL_ALL_VMUS_SK },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_SK,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_SK,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_SK,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_SK },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_SK },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_SK },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_SK },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_SK },
         { "2x", OPTION_VAL_2_O27_SK },
         { "3x", OPTION_VAL_3X_SK },
         { "4x", OPTION_VAL_4_SK },
         { "5x", OPTION_VAL_5X_SK },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SK },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SK },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SK },
         { "BLUE 03",           OPTION_VAL_BLUE_SK },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SK },
         { "GREEN 05",          OPTION_VAL_GREEN_SK },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SK },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SK },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SK },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SK },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SK },
         { "RED 11",            OPTION_VAL_RED_SK },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SK },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SK },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SK },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SK },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SK },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SK },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SK },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SK },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SK },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SK },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SK },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SK },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SK },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SK },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SK },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SK },
         { "WHITE 28",          OPTION_VAL_WHITE_SK },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SK },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SK },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SK },
         { "BLUE 03",           OPTION_VAL_BLUE_SK },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SK },
         { "GREEN 05",          OPTION_VAL_GREEN_SK },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SK },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SK },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SK },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SK },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SK },
         { "RED 11",            OPTION_VAL_RED_SK },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SK },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SK },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SK },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SK },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SK },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SK },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SK },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SK },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SK },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SK },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SK },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SK },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SK },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SK },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SK },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SK },
         { "WHITE 28",          OPTION_VAL_WHITE_SK },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_SK },
         { "20%",  OPTION_VAL_20_SK },
         { "30%",  OPTION_VAL_30_SK },
         { "40%",  OPTION_VAL_40_SK },
         { "50%",  OPTION_VAL_50_SK },
         { "60%",  OPTION_VAL_60_SK },
         { "70%",  OPTION_VAL_70_SK },
         { "80%",  OPTION_VAL_80_SK },
         { "90%",  OPTION_VAL_90_SK },
         { "100%", OPTION_VAL_100_SK },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_SK },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_SK },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_SK },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_SK },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_SK },
         { "2x", OPTION_VAL_2_O27_SK },
         { "3x", OPTION_VAL_3X_SK },
         { "4x", OPTION_VAL_4_SK },
         { "5x", OPTION_VAL_5X_SK },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SK },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SK },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SK },
         { "BLUE 03",           OPTION_VAL_BLUE_SK },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SK },
         { "GREEN 05",          OPTION_VAL_GREEN_SK },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SK },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SK },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SK },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SK },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SK },
         { "RED 11",            OPTION_VAL_RED_SK },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SK },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SK },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SK },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SK },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SK },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SK },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SK },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SK },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SK },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SK },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SK },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SK },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SK },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SK },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SK },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SK },
         { "WHITE 28",          OPTION_VAL_WHITE_SK },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SK },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SK },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SK },
         { "BLUE 03",           OPTION_VAL_BLUE_SK },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SK },
         { "GREEN 05",          OPTION_VAL_GREEN_SK },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SK },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SK },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SK },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SK },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SK },
         { "RED 11",            OPTION_VAL_RED_SK },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SK },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SK },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SK },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SK },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SK },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SK },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SK },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SK },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SK },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SK },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SK },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SK },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SK },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SK },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SK },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SK },
         { "WHITE 28",          OPTION_VAL_WHITE_SK },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_SK },
         { "20%",  OPTION_VAL_20_SK },
         { "30%",  OPTION_VAL_30_SK },
         { "40%",  OPTION_VAL_40_SK },
         { "50%",  OPTION_VAL_50_SK },
         { "60%",  OPTION_VAL_60_SK },
         { "70%",  OPTION_VAL_70_SK },
         { "80%",  OPTION_VAL_80_SK },
         { "90%",  OPTION_VAL_90_SK },
         { "100%", OPTION_VAL_100_SK },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_SK },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_SK },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_SK },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_SK },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_SK },
         { "2x", OPTION_VAL_2_O27_SK },
         { "3x", OPTION_VAL_3X_SK },
         { "4x", OPTION_VAL_4_SK },
         { "5x", OPTION_VAL_5X_SK },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SK },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SK },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SK },
         { "BLUE 03",           OPTION_VAL_BLUE_SK },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SK },
         { "GREEN 05",          OPTION_VAL_GREEN_SK },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SK },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SK },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SK },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SK },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SK },
         { "RED 11",            OPTION_VAL_RED_SK },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SK },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SK },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SK },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SK },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SK },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SK },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SK },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SK },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SK },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SK },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SK },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SK },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SK },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SK },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SK },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SK },
         { "WHITE 28",          OPTION_VAL_WHITE_SK },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SK },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SK },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SK },
         { "BLUE 03",           OPTION_VAL_BLUE_SK },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SK },
         { "GREEN 05",          OPTION_VAL_GREEN_SK },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SK },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SK },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SK },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SK },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SK },
         { "RED 11",            OPTION_VAL_RED_SK },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SK },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SK },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SK },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SK },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SK },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SK },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SK },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SK },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SK },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SK },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SK },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SK },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SK },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SK },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SK },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SK },
         { "WHITE 28",          OPTION_VAL_WHITE_SK },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_SK },
         { "20%",  OPTION_VAL_20_SK },
         { "30%",  OPTION_VAL_30_SK },
         { "40%",  OPTION_VAL_40_SK },
         { "50%",  OPTION_VAL_50_SK },
         { "60%",  OPTION_VAL_60_SK },
         { "70%",  OPTION_VAL_70_SK },
         { "80%",  OPTION_VAL_80_SK },
         { "90%",  OPTION_VAL_90_SK },
         { "100%", OPTION_VAL_100_SK },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_SK },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_SK },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_SK },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_SK },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_SK },
         { "2x", OPTION_VAL_2_O27_SK },
         { "3x", OPTION_VAL_3X_SK },
         { "4x", OPTION_VAL_4_SK },
         { "5x", OPTION_VAL_5X_SK },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SK },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SK },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SK },
         { "BLUE 03",           OPTION_VAL_BLUE_SK },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SK },
         { "GREEN 05",          OPTION_VAL_GREEN_SK },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SK },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SK },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SK },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SK },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SK },
         { "RED 11",            OPTION_VAL_RED_SK },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SK },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SK },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SK },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SK },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SK },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SK },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SK },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SK },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SK },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SK },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SK },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SK },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SK },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SK },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SK },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SK },
         { "WHITE 28",          OPTION_VAL_WHITE_SK },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SK },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SK },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SK },
         { "BLUE 03",           OPTION_VAL_BLUE_SK },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SK },
         { "GREEN 05",          OPTION_VAL_GREEN_SK },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SK },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SK },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SK },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SK },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SK },
         { "RED 11",            OPTION_VAL_RED_SK },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SK },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SK },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SK },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SK },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SK },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SK },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SK },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SK },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SK },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SK },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SK },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SK },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SK },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SK },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SK },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SK },
         { "WHITE 28",          OPTION_VAL_WHITE_SK },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_SK },
         { "20%",  OPTION_VAL_20_SK },
         { "30%",  OPTION_VAL_30_SK },
         { "40%",  OPTION_VAL_40_SK },
         { "50%",  OPTION_VAL_50_SK },
         { "60%",  OPTION_VAL_60_SK },
         { "70%",  OPTION_VAL_70_SK },
         { "80%",  OPTION_VAL_80_SK },
         { "90%",  OPTION_VAL_90_SK },
         { "100%", OPTION_VAL_100_SK },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_sk = {
   option_cats_sk,
   option_defs_sk
};

/* RETRO_LANGUAGE_SR */

#define CATEGORY_SYSTEM_LABEL_SR NULL
#define CATEGORY_SYSTEM_INFO_0_SR NULL
#define CATEGORY_VIDEO_LABEL_SR NULL
#define CATEGORY_VIDEO_INFO_0_SR NULL
#define CATEGORY_PERFORMANCE_LABEL_SR NULL
#define CATEGORY_PERFORMANCE_INFO_0_SR NULL
#define CATEGORY_HACKS_LABEL_SR NULL
#define CATEGORY_HACKS_INFO_0_SR NULL
#define CATEGORY_INPUT_LABEL_SR "Ulaz"
#define CATEGORY_INPUT_INFO_0_SR NULL
#define CATEGORY_VMU_LABEL_SR NULL
#define CATEGORY_VMU_INFO_0_SR NULL
#define CORE_OPTION_NAME_REGION_LABEL_SR NULL
#define OPTION_VAL_JAPAN_SR NULL
#define OPTION_VAL_USA_SR NULL
#define OPTION_VAL_EUROPE_SR NULL
#define OPTION_VAL_DEFAULT_SR NULL
#define CORE_OPTION_NAME_LANGUAGE_LABEL_SR NULL
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_SR NULL
#define OPTION_VAL_JAPANESE_SR NULL
#define OPTION_VAL_ENGLISH_SR NULL
#define OPTION_VAL_GERMAN_SR NULL
#define OPTION_VAL_FRENCH_SR NULL
#define OPTION_VAL_SPANISH_SR NULL
#define OPTION_VAL_ITALIAN_SR NULL
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_SR NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_SR NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_SR NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_SR NULL
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_SR NULL
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_SR NULL
#define CORE_OPTION_NAME_FORCE_WINCE_LABEL_SR NULL
#define CORE_OPTION_NAME_FORCE_WINCE_INFO_0_SR NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_SR NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_SR NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_SR NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_SR NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_SR NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_SR NULL
#define CORE_OPTION_NAME_UPNP_LABEL_SR NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_SR NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_SR NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_SR NULL
#define OPTION_VAL_320X240_SR NULL
#define OPTION_VAL_640X480_SR NULL
#define OPTION_VAL_800X600_SR NULL
#define OPTION_VAL_960X720_SR NULL
#define OPTION_VAL_1024X768_SR NULL
#define OPTION_VAL_1280X960_SR NULL
#define OPTION_VAL_1440X1080_SR NULL
#define OPTION_VAL_1600X1200_SR NULL
#define OPTION_VAL_1920X1440_SR NULL
#define OPTION_VAL_2560X1920_SR NULL
#define OPTION_VAL_2880X2160_SR NULL
#define OPTION_VAL_3200X2400_SR NULL
#define OPTION_VAL_3840X2880_SR NULL
#define OPTION_VAL_4480X3360_SR NULL
#define OPTION_VAL_5120X3840_SR NULL
#define OPTION_VAL_5760X4320_SR NULL
#define OPTION_VAL_6400X4800_SR NULL
#define OPTION_VAL_7040X5280_SR NULL
#define OPTION_VAL_7680X5760_SR NULL
#define OPTION_VAL_8320X6240_SR NULL
#define OPTION_VAL_8960X6720_SR NULL
#define OPTION_VAL_9600X7200_SR NULL
#define OPTION_VAL_10240X7680_SR NULL
#define OPTION_VAL_10880X8160_SR NULL
#define OPTION_VAL_11520X8640_SR NULL
#define OPTION_VAL_12160X9120_SR NULL
#define OPTION_VAL_12800X9600_SR NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_SR NULL
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_SR NULL
#define OPTION_VAL_VGA_SR NULL
#define OPTION_VAL_TV_RGB_SR NULL
#define OPTION_VAL_TV_COMPOSITE_SR NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_SR NULL
#define OPTION_VAL_NTSC_SR NULL
#define OPTION_VAL_PAL_SR NULL
#define OPTION_VAL_PAL_N_SR NULL
#define OPTION_VAL_PAL_M_SR NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_SR "Orijentacija ekrana"
#define OPTION_VAL_HORIZONTAL_SR NULL
#define OPTION_VAL_VERTICAL_SR NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_SR NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_SR NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_SR NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_SR NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_SR NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_SR NULL
#define OPTION_VAL_512MB_SR NULL
#define OPTION_VAL_1GB_SR NULL
#define OPTION_VAL_2GB_SR NULL
#define OPTION_VAL_4GB_SR NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_SR NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_SR NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_SR NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_SR NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_SR NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_SR NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_SR NULL
#define CORE_OPTION_NAME_FOG_LABEL_SR NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_SR NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_SR NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_SR NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_SR NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_SR NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_SR NULL
#define OPTION_VAL_1_SR NULL
#define OPTION_VAL_2_SR NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_SR NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_SR NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_SR NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_SR NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_SR NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_SR NULL
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_SR NULL
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_SR NULL
#define OPTION_VAL_2_O27_SR NULL
#define OPTION_VAL_4_SR NULL
#define OPTION_VAL_6_SR NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_SR NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_SR NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_SR NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_SR NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_SR NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_SR NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_SR NULL
#define OPTION_VAL_SOME_SR NULL
#define OPTION_VAL_MORE_SR NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_SR NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_SR NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_SR NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_SR NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_SR NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_SR NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_SR NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_SR NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_SR NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_SR NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_SR NULL
#define OPTION_VAL_0_SR NULL
#define OPTION_VAL_5_SR NULL
#define OPTION_VAL_10_SR NULL
#define OPTION_VAL_15_SR NULL
#define OPTION_VAL_20_SR NULL
#define OPTION_VAL_25_SR NULL
#define OPTION_VAL_30_SR NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_SR NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_SR NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_SR NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_SR NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_SR NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_SR NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_SR NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_SR NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_SR NULL
#define OPTION_VAL_WHITE_SR NULL
#define OPTION_VAL_RED_SR NULL
#define OPTION_VAL_GREEN_SR NULL
#define OPTION_VAL_BLUE_SR NULL
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_SR NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_SR NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_SR NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_SR NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_SR NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_SR NULL
#define OPTION_VAL_VMU_A1_SR NULL
#define OPTION_VAL_ALL_VMUS_SR NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_SR NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_SR NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_SR NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_SR NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_SR NULL
#define OPTION_VAL_UPPER_LEFT_SR NULL
#define OPTION_VAL_UPPER_RIGHT_SR NULL
#define OPTION_VAL_LOWER_LEFT_SR NULL
#define OPTION_VAL_LOWER_RIGHT_SR NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_SR NULL
#define OPTION_VAL_1X_SR NULL
#define OPTION_VAL_3X_SR NULL
#define OPTION_VAL_5X_SR NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_SR NULL
#define OPTION_VAL_DEFAULT_ON_00_SR NULL
#define OPTION_VAL_DEFAULT_OFF_01_SR NULL
#define OPTION_VAL_BLACK_02_SR NULL
#define OPTION_VAL_LIGHT_BLUE_04_SR NULL
#define OPTION_VAL_CYAN_06_SR NULL
#define OPTION_VAL_CYAN_BLUE_07_SR NULL
#define OPTION_VAL_LIGHT_GREEN_08_SR NULL
#define OPTION_VAL_CYAN_GREEN_09_SR NULL
#define OPTION_VAL_LIGHT_CYAN_10_SR NULL
#define OPTION_VAL_PURPLE_12_SR NULL
#define OPTION_VAL_LIGHT_PURPLE_13_SR NULL
#define OPTION_VAL_YELLOW_14_SR NULL
#define OPTION_VAL_GRAY_15_SR NULL
#define OPTION_VAL_LIGHT_PURPLE_2_16_SR NULL
#define OPTION_VAL_LIGHT_GREEN_2_17_SR NULL
#define OPTION_VAL_LIGHT_GREEN_3_18_SR NULL
#define OPTION_VAL_LIGHT_CYAN_2_19_SR NULL
#define OPTION_VAL_LIGHT_RED_2_20_SR NULL
#define OPTION_VAL_MAGENTA_21_SR NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_SR NULL
#define OPTION_VAL_LIGHT_ORANGE_23_SR NULL
#define OPTION_VAL_ORANGE_24_SR NULL
#define OPTION_VAL_LIGHT_PURPLE_4_25_SR NULL
#define OPTION_VAL_LIGHT_YELLOW_26_SR NULL
#define OPTION_VAL_LIGHT_YELLOW_2_27_SR NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_SR NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_SR NULL
#define OPTION_VAL_40_SR NULL
#define OPTION_VAL_50_SR NULL
#define OPTION_VAL_60_SR NULL
#define OPTION_VAL_70_SR NULL
#define OPTION_VAL_80_SR NULL
#define OPTION_VAL_90_SR NULL
#define OPTION_VAL_100_SR NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_SR NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_SR NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_SR NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_SR NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_SR NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_SR NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_SR NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_SR NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_SR NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_SR NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_SR NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_SR NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_SR NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_SR NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_SR NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_SR NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_SR NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_SR NULL

struct retro_core_option_v2_category option_cats_sr[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_SR,
      CATEGORY_SYSTEM_INFO_0_SR
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_SR,
      CATEGORY_VIDEO_INFO_0_SR
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_SR,
      CATEGORY_PERFORMANCE_INFO_0_SR
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_SR,
      CATEGORY_HACKS_INFO_0_SR
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_SR,
      CATEGORY_INPUT_INFO_0_SR
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_SR,
      CATEGORY_VMU_INFO_0_SR
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_sr[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_SR,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_SR },
         { "USA",     OPTION_VAL_USA_SR },
         { "Europe",  OPTION_VAL_EUROPE_SR },
         { "Default", OPTION_VAL_DEFAULT_SR },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_SR,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_SR },
         { "English",  OPTION_VAL_ENGLISH_SR },
         { "German",   OPTION_VAL_GERMAN_SR },
         { "French",   OPTION_VAL_FRENCH_SR },
         { "Spanish",  OPTION_VAL_SPANISH_SR },
         { "Italian",  OPTION_VAL_ITALIAN_SR },
         { "Default",  OPTION_VAL_DEFAULT_SR },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_SR,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_boot_to_bios",
      CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_SR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_SR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_force_wince",
      CORE_OPTION_NAME_FORCE_WINCE_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_FORCE_WINCE_INFO_0_SR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_SR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_SR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_SR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_SR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_SR,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_SR },
         { "640x480",    OPTION_VAL_640X480_SR },
         { "800x600",    OPTION_VAL_800X600_SR },
         { "960x720",    OPTION_VAL_960X720_SR },
         { "1024x768",   OPTION_VAL_1024X768_SR },
         { "1280x960",   OPTION_VAL_1280X960_SR },
         { "1440x1080",  OPTION_VAL_1440X1080_SR },
         { "1600x1200",  OPTION_VAL_1600X1200_SR },
         { "1920x1440",  OPTION_VAL_1920X1440_SR },
         { "2560x1920",  OPTION_VAL_2560X1920_SR },
         { "2880x2160",  OPTION_VAL_2880X2160_SR },
         { "3200x2400",  OPTION_VAL_3200X2400_SR },
         { "3840x2880",  OPTION_VAL_3840X2880_SR },
         { "4480x3360",  OPTION_VAL_4480X3360_SR },
         { "5120x3840",  OPTION_VAL_5120X3840_SR },
         { "5760x4320",  OPTION_VAL_5760X4320_SR },
         { "6400x4800",  OPTION_VAL_6400X4800_SR },
         { "7040x5280",  OPTION_VAL_7040X5280_SR },
         { "7680x5760",  OPTION_VAL_7680X5760_SR },
         { "8320x6240",  OPTION_VAL_8320X6240_SR },
         { "8960x6720",  OPTION_VAL_8960X6720_SR },
         { "9600x7200",  OPTION_VAL_9600X7200_SR },
         { "10240x7680", OPTION_VAL_10240X7680_SR },
         { "10880x8160", OPTION_VAL_10880X8160_SR },
         { "11520x8640", OPTION_VAL_11520X8640_SR },
         { "12160x9120", OPTION_VAL_12160X9120_SR },
         { "12800x9600", OPTION_VAL_12800X9600_SR },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_SR,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_SR },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_SR },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_SR },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_SR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_SR },
         { "PAL",     OPTION_VAL_PAL_SR },
         { "PAL_N",   OPTION_VAL_PAL_N_SR },
         { "PAL_M",   OPTION_VAL_PAL_M_SR },
         { "Default", OPTION_VAL_DEFAULT_SR },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_SR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_SR },
         { "vertical",   OPTION_VAL_VERTICAL_SR },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_SR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_SR },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_SR },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_SR },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_SR,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_SR },
         { "1GB",   OPTION_VAL_1GB_SR },
         { "2GB",   OPTION_VAL_2GB_SR },
         { "4GB",   OPTION_VAL_4GB_SR },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_SR,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_SR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_SR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_SR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_SR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_SR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_SR,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_SR,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_SR },
         { "1",  OPTION_VAL_1_SR },
         { "2",  OPTION_VAL_2_SR },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_SR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_SR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_SR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_SR,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O27_SR },
         { "4",  OPTION_VAL_4_SR },
         { "6",  OPTION_VAL_6_SR },
         { NULL, NULL },
      },
      "1",
   },
   {/* TODO: needs clarification */
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_SR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_SR,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_SR,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_SR,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_SR,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_SR },
         { "more", OPTION_VAL_MORE_SR },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_SR,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_SR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_SR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_SR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_SR,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: probably needs explanation */
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_SR,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_SR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_SR },
         { "5%",  OPTION_VAL_5_SR },
         { "10%", OPTION_VAL_10_SR },
         { "15%", OPTION_VAL_15_SR },
         { "20%", OPTION_VAL_20_SR },
         { "25%", OPTION_VAL_25_SR },
         { "30%", OPTION_VAL_30_SR },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_SR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_SR },
         { "5%",  OPTION_VAL_5_SR },
         { "10%", OPTION_VAL_10_SR },
         { "15%", OPTION_VAL_15_SR },
         { "20%", OPTION_VAL_20_SR },
         { "25%", OPTION_VAL_25_SR },
         { "30%", OPTION_VAL_30_SR },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_SR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_purupuru",
      CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_SR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_SR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_SR,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_SR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_SR },
         { "Red",      OPTION_VAL_RED_SR },
         { "Green",    OPTION_VAL_GREEN_SR },
         { "Blue",     OPTION_VAL_BLUE_SR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_SR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_SR },
         { "Red",      OPTION_VAL_RED_SR },
         { "Green",    OPTION_VAL_GREEN_SR },
         { "Blue",     OPTION_VAL_BLUE_SR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_SR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_SR },
         { "Red",      OPTION_VAL_RED_SR },
         { "Green",    OPTION_VAL_GREEN_SR },
         { "Blue",     OPTION_VAL_BLUE_SR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_SR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_SR },
         { "Red",      OPTION_VAL_RED_SR },
         { "Green",    OPTION_VAL_GREEN_SR },
         { "Blue",     OPTION_VAL_BLUE_SR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_SR,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_SR,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_SR,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_SR },
         { "All VMUs", OPTION_VAL_ALL_VMUS_SR },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_SR,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_SR,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_SR,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_SR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_SR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_SR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_SR },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_SR },
         { "2x", OPTION_VAL_2_O27_SR },
         { "3x", OPTION_VAL_3X_SR },
         { "4x", OPTION_VAL_4_SR },
         { "5x", OPTION_VAL_5X_SR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SR },
         { "BLUE 03",           OPTION_VAL_BLUE_SR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SR },
         { "GREEN 05",          OPTION_VAL_GREEN_SR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SR },
         { "RED 11",            OPTION_VAL_RED_SR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SR },
         { "WHITE 28",          OPTION_VAL_WHITE_SR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SR },
         { "BLUE 03",           OPTION_VAL_BLUE_SR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SR },
         { "GREEN 05",          OPTION_VAL_GREEN_SR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SR },
         { "RED 11",            OPTION_VAL_RED_SR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SR },
         { "WHITE 28",          OPTION_VAL_WHITE_SR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_SR },
         { "20%",  OPTION_VAL_20_SR },
         { "30%",  OPTION_VAL_30_SR },
         { "40%",  OPTION_VAL_40_SR },
         { "50%",  OPTION_VAL_50_SR },
         { "60%",  OPTION_VAL_60_SR },
         { "70%",  OPTION_VAL_70_SR },
         { "80%",  OPTION_VAL_80_SR },
         { "90%",  OPTION_VAL_90_SR },
         { "100%", OPTION_VAL_100_SR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_SR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_SR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_SR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_SR },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_SR },
         { "2x", OPTION_VAL_2_O27_SR },
         { "3x", OPTION_VAL_3X_SR },
         { "4x", OPTION_VAL_4_SR },
         { "5x", OPTION_VAL_5X_SR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SR },
         { "BLUE 03",           OPTION_VAL_BLUE_SR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SR },
         { "GREEN 05",          OPTION_VAL_GREEN_SR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SR },
         { "RED 11",            OPTION_VAL_RED_SR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SR },
         { "WHITE 28",          OPTION_VAL_WHITE_SR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SR },
         { "BLUE 03",           OPTION_VAL_BLUE_SR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SR },
         { "GREEN 05",          OPTION_VAL_GREEN_SR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SR },
         { "RED 11",            OPTION_VAL_RED_SR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SR },
         { "WHITE 28",          OPTION_VAL_WHITE_SR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_SR },
         { "20%",  OPTION_VAL_20_SR },
         { "30%",  OPTION_VAL_30_SR },
         { "40%",  OPTION_VAL_40_SR },
         { "50%",  OPTION_VAL_50_SR },
         { "60%",  OPTION_VAL_60_SR },
         { "70%",  OPTION_VAL_70_SR },
         { "80%",  OPTION_VAL_80_SR },
         { "90%",  OPTION_VAL_90_SR },
         { "100%", OPTION_VAL_100_SR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_SR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_SR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_SR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_SR },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_SR },
         { "2x", OPTION_VAL_2_O27_SR },
         { "3x", OPTION_VAL_3X_SR },
         { "4x", OPTION_VAL_4_SR },
         { "5x", OPTION_VAL_5X_SR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SR },
         { "BLUE 03",           OPTION_VAL_BLUE_SR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SR },
         { "GREEN 05",          OPTION_VAL_GREEN_SR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SR },
         { "RED 11",            OPTION_VAL_RED_SR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SR },
         { "WHITE 28",          OPTION_VAL_WHITE_SR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SR },
         { "BLUE 03",           OPTION_VAL_BLUE_SR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SR },
         { "GREEN 05",          OPTION_VAL_GREEN_SR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SR },
         { "RED 11",            OPTION_VAL_RED_SR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SR },
         { "WHITE 28",          OPTION_VAL_WHITE_SR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_SR },
         { "20%",  OPTION_VAL_20_SR },
         { "30%",  OPTION_VAL_30_SR },
         { "40%",  OPTION_VAL_40_SR },
         { "50%",  OPTION_VAL_50_SR },
         { "60%",  OPTION_VAL_60_SR },
         { "70%",  OPTION_VAL_70_SR },
         { "80%",  OPTION_VAL_80_SR },
         { "90%",  OPTION_VAL_90_SR },
         { "100%", OPTION_VAL_100_SR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_SR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_SR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_SR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_SR },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_SR },
         { "2x", OPTION_VAL_2_O27_SR },
         { "3x", OPTION_VAL_3X_SR },
         { "4x", OPTION_VAL_4_SR },
         { "5x", OPTION_VAL_5X_SR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SR },
         { "BLUE 03",           OPTION_VAL_BLUE_SR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SR },
         { "GREEN 05",          OPTION_VAL_GREEN_SR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SR },
         { "RED 11",            OPTION_VAL_RED_SR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SR },
         { "WHITE 28",          OPTION_VAL_WHITE_SR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SR },
         { "BLUE 03",           OPTION_VAL_BLUE_SR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SR },
         { "GREEN 05",          OPTION_VAL_GREEN_SR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SR },
         { "RED 11",            OPTION_VAL_RED_SR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SR },
         { "WHITE 28",          OPTION_VAL_WHITE_SR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_SR },
         { "20%",  OPTION_VAL_20_SR },
         { "30%",  OPTION_VAL_30_SR },
         { "40%",  OPTION_VAL_40_SR },
         { "50%",  OPTION_VAL_50_SR },
         { "60%",  OPTION_VAL_60_SR },
         { "70%",  OPTION_VAL_70_SR },
         { "80%",  OPTION_VAL_80_SR },
         { "90%",  OPTION_VAL_90_SR },
         { "100%", OPTION_VAL_100_SR },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_sr = {
   option_cats_sr,
   option_defs_sr
};

/* RETRO_LANGUAGE_SV */

#define CATEGORY_SYSTEM_LABEL_SV NULL
#define CATEGORY_SYSTEM_INFO_0_SV NULL
#define CATEGORY_VIDEO_LABEL_SV "Bild"
#define CATEGORY_VIDEO_INFO_0_SV NULL
#define CATEGORY_PERFORMANCE_LABEL_SV "Prestanda"
#define CATEGORY_PERFORMANCE_INFO_0_SV NULL
#define CATEGORY_HACKS_LABEL_SV NULL
#define CATEGORY_HACKS_INFO_0_SV NULL
#define CATEGORY_INPUT_LABEL_SV "Inmatning"
#define CATEGORY_INPUT_INFO_0_SV NULL
#define CATEGORY_VMU_LABEL_SV NULL
#define CATEGORY_VMU_INFO_0_SV NULL
#define CORE_OPTION_NAME_REGION_LABEL_SV NULL
#define OPTION_VAL_JAPAN_SV NULL
#define OPTION_VAL_USA_SV NULL
#define OPTION_VAL_EUROPE_SV "Europa"
#define OPTION_VAL_DEFAULT_SV NULL
#define CORE_OPTION_NAME_LANGUAGE_LABEL_SV "Språk"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_SV NULL
#define OPTION_VAL_JAPANESE_SV "Japanska"
#define OPTION_VAL_ENGLISH_SV "Engelska"
#define OPTION_VAL_GERMAN_SV "Tyska"
#define OPTION_VAL_FRENCH_SV "Franska"
#define OPTION_VAL_SPANISH_SV "Spanska"
#define OPTION_VAL_ITALIAN_SV "Italienska"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_SV NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_SV NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_SV NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_SV NULL
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_SV "Aktivera DSP"
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_SV NULL
#define CORE_OPTION_NAME_FORCE_WINCE_LABEL_SV NULL
#define CORE_OPTION_NAME_FORCE_WINCE_INFO_0_SV NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_SV NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_SV NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_SV NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_SV NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_SV NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_SV NULL
#define CORE_OPTION_NAME_UPNP_LABEL_SV NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_SV NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_SV "Intern upplösning"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_SV NULL
#define OPTION_VAL_320X240_SV NULL
#define OPTION_VAL_640X480_SV NULL
#define OPTION_VAL_800X600_SV NULL
#define OPTION_VAL_960X720_SV NULL
#define OPTION_VAL_1024X768_SV NULL
#define OPTION_VAL_1280X960_SV NULL
#define OPTION_VAL_1440X1080_SV NULL
#define OPTION_VAL_1600X1200_SV NULL
#define OPTION_VAL_1920X1440_SV NULL
#define OPTION_VAL_2560X1920_SV NULL
#define OPTION_VAL_2880X2160_SV NULL
#define OPTION_VAL_3200X2400_SV NULL
#define OPTION_VAL_3840X2880_SV NULL
#define OPTION_VAL_4480X3360_SV NULL
#define OPTION_VAL_5120X3840_SV NULL
#define OPTION_VAL_5760X4320_SV NULL
#define OPTION_VAL_6400X4800_SV NULL
#define OPTION_VAL_7040X5280_SV NULL
#define OPTION_VAL_7680X5760_SV NULL
#define OPTION_VAL_8320X6240_SV NULL
#define OPTION_VAL_8960X6720_SV NULL
#define OPTION_VAL_9600X7200_SV NULL
#define OPTION_VAL_10240X7680_SV NULL
#define OPTION_VAL_10880X8160_SV NULL
#define OPTION_VAL_11520X8640_SV NULL
#define OPTION_VAL_12160X9120_SV NULL
#define OPTION_VAL_12800X9600_SV NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_SV "Kabeltyp"
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_SV NULL
#define OPTION_VAL_VGA_SV NULL
#define OPTION_VAL_TV_RGB_SV NULL
#define OPTION_VAL_TV_COMPOSITE_SV NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_SV NULL
#define OPTION_VAL_NTSC_SV NULL
#define OPTION_VAL_PAL_SV NULL
#define OPTION_VAL_PAL_N_SV NULL
#define OPTION_VAL_PAL_M_SV NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_SV "Skärmorientering"
#define OPTION_VAL_HORIZONTAL_SV "Horisontell"
#define OPTION_VAL_VERTICAL_SV NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_SV NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_SV NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_SV NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_SV NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_SV NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_SV NULL
#define OPTION_VAL_512MB_SV NULL
#define OPTION_VAL_1GB_SV NULL
#define OPTION_VAL_2GB_SV NULL
#define OPTION_VAL_4GB_SV NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_SV NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_SV NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_SV NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_SV NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_SV NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_SV NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_SV NULL
#define CORE_OPTION_NAME_FOG_LABEL_SV "Dimeffekter"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_SV NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_SV NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_SV NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_SV NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_SV NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_SV NULL
#define OPTION_VAL_1_SV NULL
#define OPTION_VAL_2_SV NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_SV NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_SV NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_SV NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_SV NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_SV NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_SV NULL
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_SV NULL
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_SV NULL
#define OPTION_VAL_2_O27_SV NULL
#define OPTION_VAL_4_SV NULL
#define OPTION_VAL_6_SV NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_SV NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_SV NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_SV NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_SV NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_SV NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_SV NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_SV NULL
#define OPTION_VAL_SOME_SV NULL
#define OPTION_VAL_MORE_SV NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_SV NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_SV NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_SV NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_SV NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_SV NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_SV NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_SV NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_SV NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_SV NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_SV "Dumpa texturer"
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_SV NULL
#define OPTION_VAL_0_SV "0 %"
#define OPTION_VAL_5_SV "5 %"
#define OPTION_VAL_10_SV "10 %"
#define OPTION_VAL_15_SV "15 %"
#define OPTION_VAL_20_SV "20 %"
#define OPTION_VAL_25_SV "25 %"
#define OPTION_VAL_30_SV "30 %"
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_SV NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_SV NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_SV NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_SV NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_SV NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_SV NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_SV NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_SV NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_SV NULL
#define OPTION_VAL_WHITE_SV "Vit"
#define OPTION_VAL_RED_SV "Röd"
#define OPTION_VAL_GREEN_SV "Grön"
#define OPTION_VAL_BLUE_SV "Blå"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_SV NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_SV NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_SV NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_SV NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_SV NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_SV NULL
#define OPTION_VAL_VMU_A1_SV NULL
#define OPTION_VAL_ALL_VMUS_SV "Alla VMU:er"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_SV NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_SV NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_SV NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_SV NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_SV NULL
#define OPTION_VAL_UPPER_LEFT_SV NULL
#define OPTION_VAL_UPPER_RIGHT_SV NULL
#define OPTION_VAL_LOWER_LEFT_SV NULL
#define OPTION_VAL_LOWER_RIGHT_SV NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_SV NULL
#define OPTION_VAL_1X_SV NULL
#define OPTION_VAL_3X_SV NULL
#define OPTION_VAL_5X_SV NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_SV NULL
#define OPTION_VAL_DEFAULT_ON_00_SV NULL
#define OPTION_VAL_DEFAULT_OFF_01_SV NULL
#define OPTION_VAL_BLACK_02_SV "Svart"
#define OPTION_VAL_LIGHT_BLUE_04_SV "Ljusblå"
#define OPTION_VAL_CYAN_06_SV NULL
#define OPTION_VAL_CYAN_BLUE_07_SV NULL
#define OPTION_VAL_LIGHT_GREEN_08_SV "Ljusgrön"
#define OPTION_VAL_CYAN_GREEN_09_SV NULL
#define OPTION_VAL_LIGHT_CYAN_10_SV NULL
#define OPTION_VAL_PURPLE_12_SV "Lila"
#define OPTION_VAL_LIGHT_PURPLE_13_SV "Ljuslila"
#define OPTION_VAL_YELLOW_14_SV "Gul"
#define OPTION_VAL_GRAY_15_SV "Grå"
#define OPTION_VAL_LIGHT_PURPLE_2_16_SV "Ljuslila (2)"
#define OPTION_VAL_LIGHT_GREEN_2_17_SV "Ljusgrön (2)"
#define OPTION_VAL_LIGHT_GREEN_3_18_SV "Ljusgrön (3)"
#define OPTION_VAL_LIGHT_CYAN_2_19_SV NULL
#define OPTION_VAL_LIGHT_RED_2_20_SV NULL
#define OPTION_VAL_MAGENTA_21_SV NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_SV "Ljuslila (3)"
#define OPTION_VAL_LIGHT_ORANGE_23_SV "Ljusorange"
#define OPTION_VAL_ORANGE_24_SV NULL
#define OPTION_VAL_LIGHT_PURPLE_4_25_SV "Ljuslila (4)"
#define OPTION_VAL_LIGHT_YELLOW_26_SV "Ljusgul"
#define OPTION_VAL_LIGHT_YELLOW_2_27_SV NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_SV NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_SV NULL
#define OPTION_VAL_40_SV "40 %"
#define OPTION_VAL_50_SV "50 %"
#define OPTION_VAL_60_SV "60 %"
#define OPTION_VAL_70_SV "70 %"
#define OPTION_VAL_80_SV "80 %"
#define OPTION_VAL_90_SV "90 %"
#define OPTION_VAL_100_SV "100 %"
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_SV NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_SV NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_SV NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_SV NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_SV NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_SV NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_SV NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_SV NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_SV NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_SV NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_SV NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_SV NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_SV NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_SV NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_SV NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_SV NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_SV NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_SV NULL

struct retro_core_option_v2_category option_cats_sv[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_SV,
      CATEGORY_SYSTEM_INFO_0_SV
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_SV,
      CATEGORY_VIDEO_INFO_0_SV
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_SV,
      CATEGORY_PERFORMANCE_INFO_0_SV
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_SV,
      CATEGORY_HACKS_INFO_0_SV
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_SV,
      CATEGORY_INPUT_INFO_0_SV
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_SV,
      CATEGORY_VMU_INFO_0_SV
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_sv[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_SV,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_SV },
         { "USA",     OPTION_VAL_USA_SV },
         { "Europe",  OPTION_VAL_EUROPE_SV },
         { "Default", OPTION_VAL_DEFAULT_SV },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_SV,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_SV },
         { "English",  OPTION_VAL_ENGLISH_SV },
         { "German",   OPTION_VAL_GERMAN_SV },
         { "French",   OPTION_VAL_FRENCH_SV },
         { "Spanish",  OPTION_VAL_SPANISH_SV },
         { "Italian",  OPTION_VAL_ITALIAN_SV },
         { "Default",  OPTION_VAL_DEFAULT_SV },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_SV,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_boot_to_bios",
      CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_SV,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_SV,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_force_wince",
      CORE_OPTION_NAME_FORCE_WINCE_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_FORCE_WINCE_INFO_0_SV,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_SV,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_SV,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_SV,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_SV,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_SV,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_SV },
         { "640x480",    OPTION_VAL_640X480_SV },
         { "800x600",    OPTION_VAL_800X600_SV },
         { "960x720",    OPTION_VAL_960X720_SV },
         { "1024x768",   OPTION_VAL_1024X768_SV },
         { "1280x960",   OPTION_VAL_1280X960_SV },
         { "1440x1080",  OPTION_VAL_1440X1080_SV },
         { "1600x1200",  OPTION_VAL_1600X1200_SV },
         { "1920x1440",  OPTION_VAL_1920X1440_SV },
         { "2560x1920",  OPTION_VAL_2560X1920_SV },
         { "2880x2160",  OPTION_VAL_2880X2160_SV },
         { "3200x2400",  OPTION_VAL_3200X2400_SV },
         { "3840x2880",  OPTION_VAL_3840X2880_SV },
         { "4480x3360",  OPTION_VAL_4480X3360_SV },
         { "5120x3840",  OPTION_VAL_5120X3840_SV },
         { "5760x4320",  OPTION_VAL_5760X4320_SV },
         { "6400x4800",  OPTION_VAL_6400X4800_SV },
         { "7040x5280",  OPTION_VAL_7040X5280_SV },
         { "7680x5760",  OPTION_VAL_7680X5760_SV },
         { "8320x6240",  OPTION_VAL_8320X6240_SV },
         { "8960x6720",  OPTION_VAL_8960X6720_SV },
         { "9600x7200",  OPTION_VAL_9600X7200_SV },
         { "10240x7680", OPTION_VAL_10240X7680_SV },
         { "10880x8160", OPTION_VAL_10880X8160_SV },
         { "11520x8640", OPTION_VAL_11520X8640_SV },
         { "12160x9120", OPTION_VAL_12160X9120_SV },
         { "12800x9600", OPTION_VAL_12800X9600_SV },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_SV,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_SV },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_SV },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_SV },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_SV,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_SV },
         { "PAL",     OPTION_VAL_PAL_SV },
         { "PAL_N",   OPTION_VAL_PAL_N_SV },
         { "PAL_M",   OPTION_VAL_PAL_M_SV },
         { "Default", OPTION_VAL_DEFAULT_SV },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_SV,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_SV },
         { "vertical",   OPTION_VAL_VERTICAL_SV },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_SV,
      NULL,
      "",
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_SV },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_SV },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_SV },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_SV,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_SV },
         { "1GB",   OPTION_VAL_1GB_SV },
         { "2GB",   OPTION_VAL_2GB_SV },
         { "4GB",   OPTION_VAL_4GB_SV },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_SV,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_SV,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_SV,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_SV,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_SV,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_SV,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_SV,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_SV,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_SV },
         { "1",  OPTION_VAL_1_SV },
         { "2",  OPTION_VAL_2_SV },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_SV,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_SV,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_SV,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_SV,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O27_SV },
         { "4",  OPTION_VAL_4_SV },
         { "6",  OPTION_VAL_6_SV },
         { NULL, NULL },
      },
      "1",
   },
   {/* TODO: needs clarification */
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_SV,
      NULL,
      "",
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_SV,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_SV,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_SV,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_SV,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_SV },
         { "more", OPTION_VAL_MORE_SV },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_SV,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_SV,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_SV,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_SV,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_SV,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: probably needs explanation */
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_SV,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_SV,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_SV },
         { "5%",  OPTION_VAL_5_SV },
         { "10%", OPTION_VAL_10_SV },
         { "15%", OPTION_VAL_15_SV },
         { "20%", OPTION_VAL_20_SV },
         { "25%", OPTION_VAL_25_SV },
         { "30%", OPTION_VAL_30_SV },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_SV,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_SV },
         { "5%",  OPTION_VAL_5_SV },
         { "10%", OPTION_VAL_10_SV },
         { "15%", OPTION_VAL_15_SV },
         { "20%", OPTION_VAL_20_SV },
         { "25%", OPTION_VAL_25_SV },
         { "30%", OPTION_VAL_30_SV },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_SV,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_purupuru",
      CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_SV,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_SV,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_SV,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_SV,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_SV },
         { "Red",      OPTION_VAL_RED_SV },
         { "Green",    OPTION_VAL_GREEN_SV },
         { "Blue",     OPTION_VAL_BLUE_SV },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_SV,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_SV },
         { "Red",      OPTION_VAL_RED_SV },
         { "Green",    OPTION_VAL_GREEN_SV },
         { "Blue",     OPTION_VAL_BLUE_SV },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_SV,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_SV },
         { "Red",      OPTION_VAL_RED_SV },
         { "Green",    OPTION_VAL_GREEN_SV },
         { "Blue",     OPTION_VAL_BLUE_SV },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_SV,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_SV },
         { "Red",      OPTION_VAL_RED_SV },
         { "Green",    OPTION_VAL_GREEN_SV },
         { "Blue",     OPTION_VAL_BLUE_SV },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_SV,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_SV,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_SV,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_SV },
         { "All VMUs", OPTION_VAL_ALL_VMUS_SV },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_SV,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_SV,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_SV,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_SV },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_SV },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_SV },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_SV },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_SV },
         { "2x", OPTION_VAL_2_O27_SV },
         { "3x", OPTION_VAL_3X_SV },
         { "4x", OPTION_VAL_4_SV },
         { "5x", OPTION_VAL_5X_SV },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SV },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SV },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SV },
         { "BLUE 03",           OPTION_VAL_BLUE_SV },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SV },
         { "GREEN 05",          OPTION_VAL_GREEN_SV },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SV },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SV },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SV },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SV },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SV },
         { "RED 11",            OPTION_VAL_RED_SV },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SV },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SV },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SV },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SV },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SV },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SV },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SV },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SV },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SV },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SV },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SV },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SV },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SV },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SV },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SV },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SV },
         { "WHITE 28",          OPTION_VAL_WHITE_SV },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SV },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SV },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SV },
         { "BLUE 03",           OPTION_VAL_BLUE_SV },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SV },
         { "GREEN 05",          OPTION_VAL_GREEN_SV },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SV },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SV },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SV },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SV },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SV },
         { "RED 11",            OPTION_VAL_RED_SV },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SV },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SV },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SV },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SV },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SV },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SV },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SV },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SV },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SV },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SV },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SV },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SV },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SV },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SV },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SV },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SV },
         { "WHITE 28",          OPTION_VAL_WHITE_SV },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_SV },
         { "20%",  OPTION_VAL_20_SV },
         { "30%",  OPTION_VAL_30_SV },
         { "40%",  OPTION_VAL_40_SV },
         { "50%",  OPTION_VAL_50_SV },
         { "60%",  OPTION_VAL_60_SV },
         { "70%",  OPTION_VAL_70_SV },
         { "80%",  OPTION_VAL_80_SV },
         { "90%",  OPTION_VAL_90_SV },
         { "100%", OPTION_VAL_100_SV },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_SV },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_SV },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_SV },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_SV },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_SV },
         { "2x", OPTION_VAL_2_O27_SV },
         { "3x", OPTION_VAL_3X_SV },
         { "4x", OPTION_VAL_4_SV },
         { "5x", OPTION_VAL_5X_SV },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SV },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SV },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SV },
         { "BLUE 03",           OPTION_VAL_BLUE_SV },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SV },
         { "GREEN 05",          OPTION_VAL_GREEN_SV },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SV },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SV },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SV },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SV },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SV },
         { "RED 11",            OPTION_VAL_RED_SV },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SV },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SV },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SV },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SV },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SV },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SV },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SV },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SV },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SV },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SV },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SV },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SV },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SV },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SV },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SV },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SV },
         { "WHITE 28",          OPTION_VAL_WHITE_SV },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SV },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SV },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SV },
         { "BLUE 03",           OPTION_VAL_BLUE_SV },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SV },
         { "GREEN 05",          OPTION_VAL_GREEN_SV },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SV },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SV },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SV },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SV },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SV },
         { "RED 11",            OPTION_VAL_RED_SV },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SV },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SV },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SV },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SV },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SV },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SV },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SV },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SV },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SV },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SV },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SV },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SV },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SV },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SV },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SV },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SV },
         { "WHITE 28",          OPTION_VAL_WHITE_SV },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_SV },
         { "20%",  OPTION_VAL_20_SV },
         { "30%",  OPTION_VAL_30_SV },
         { "40%",  OPTION_VAL_40_SV },
         { "50%",  OPTION_VAL_50_SV },
         { "60%",  OPTION_VAL_60_SV },
         { "70%",  OPTION_VAL_70_SV },
         { "80%",  OPTION_VAL_80_SV },
         { "90%",  OPTION_VAL_90_SV },
         { "100%", OPTION_VAL_100_SV },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_SV },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_SV },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_SV },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_SV },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_SV },
         { "2x", OPTION_VAL_2_O27_SV },
         { "3x", OPTION_VAL_3X_SV },
         { "4x", OPTION_VAL_4_SV },
         { "5x", OPTION_VAL_5X_SV },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SV },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SV },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SV },
         { "BLUE 03",           OPTION_VAL_BLUE_SV },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SV },
         { "GREEN 05",          OPTION_VAL_GREEN_SV },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SV },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SV },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SV },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SV },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SV },
         { "RED 11",            OPTION_VAL_RED_SV },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SV },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SV },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SV },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SV },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SV },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SV },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SV },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SV },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SV },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SV },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SV },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SV },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SV },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SV },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SV },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SV },
         { "WHITE 28",          OPTION_VAL_WHITE_SV },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SV },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SV },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SV },
         { "BLUE 03",           OPTION_VAL_BLUE_SV },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SV },
         { "GREEN 05",          OPTION_VAL_GREEN_SV },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SV },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SV },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SV },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SV },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SV },
         { "RED 11",            OPTION_VAL_RED_SV },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SV },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SV },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SV },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SV },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SV },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SV },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SV },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SV },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SV },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SV },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SV },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SV },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SV },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SV },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SV },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SV },
         { "WHITE 28",          OPTION_VAL_WHITE_SV },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_SV },
         { "20%",  OPTION_VAL_20_SV },
         { "30%",  OPTION_VAL_30_SV },
         { "40%",  OPTION_VAL_40_SV },
         { "50%",  OPTION_VAL_50_SV },
         { "60%",  OPTION_VAL_60_SV },
         { "70%",  OPTION_VAL_70_SV },
         { "80%",  OPTION_VAL_80_SV },
         { "90%",  OPTION_VAL_90_SV },
         { "100%", OPTION_VAL_100_SV },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_SV },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_SV },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_SV },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_SV },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_SV },
         { "2x", OPTION_VAL_2_O27_SV },
         { "3x", OPTION_VAL_3X_SV },
         { "4x", OPTION_VAL_4_SV },
         { "5x", OPTION_VAL_5X_SV },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SV },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SV },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SV },
         { "BLUE 03",           OPTION_VAL_BLUE_SV },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SV },
         { "GREEN 05",          OPTION_VAL_GREEN_SV },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SV },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SV },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SV },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SV },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SV },
         { "RED 11",            OPTION_VAL_RED_SV },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SV },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SV },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SV },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SV },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SV },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SV },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SV },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SV },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SV },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SV },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SV },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SV },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SV },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SV },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SV },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SV },
         { "WHITE 28",          OPTION_VAL_WHITE_SV },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SV },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SV },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SV },
         { "BLUE 03",           OPTION_VAL_BLUE_SV },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SV },
         { "GREEN 05",          OPTION_VAL_GREEN_SV },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SV },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SV },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SV },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SV },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SV },
         { "RED 11",            OPTION_VAL_RED_SV },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SV },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SV },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SV },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SV },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SV },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SV },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SV },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SV },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SV },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SV },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SV },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SV },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SV },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SV },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SV },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SV },
         { "WHITE 28",          OPTION_VAL_WHITE_SV },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_SV },
         { "20%",  OPTION_VAL_20_SV },
         { "30%",  OPTION_VAL_30_SV },
         { "40%",  OPTION_VAL_40_SV },
         { "50%",  OPTION_VAL_50_SV },
         { "60%",  OPTION_VAL_60_SV },
         { "70%",  OPTION_VAL_70_SV },
         { "80%",  OPTION_VAL_80_SV },
         { "90%",  OPTION_VAL_90_SV },
         { "100%", OPTION_VAL_100_SV },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_sv = {
   option_cats_sv,
   option_defs_sv
};

/* RETRO_LANGUAGE_TR */

#define CATEGORY_SYSTEM_LABEL_TR "Sistem"
#define CATEGORY_SYSTEM_INFO_0_TR "Bölge, dil, BIOS ve temel donanım ayarlarını yapılandırın."
#define CATEGORY_VIDEO_LABEL_TR NULL
#define CATEGORY_VIDEO_INFO_0_TR "Çözünürlüğü, istenenden bağımsız şeffaflığı ve görsel efekt ayarlarını yapılandırın."
#define CATEGORY_PERFORMANCE_LABEL_TR "Performans"
#define CATEGORY_PERFORMANCE_INFO_0_TR "Zorlu işleme ve kare atlama ayarlarını yapılandırın."
#define CATEGORY_HACKS_LABEL_TR "Emülatör Geliştirmeleri"
#define CATEGORY_HACKS_INFO_0_TR "Geniş ekran geçersiz kılmaları, GD-ROM yükleme hızı ve doku değiştirme ayarlarını yapılandırın."
#define CATEGORY_INPUT_LABEL_TR "Girdi"
#define CATEGORY_INPUT_INFO_0_TR "Oyun kolu ve light gun ayarlarını yapılandırın."
#define CATEGORY_VMU_LABEL_TR "Görsel Bellek Birimi"
#define CATEGORY_VMU_INFO_0_TR "Oyun başına VMU kayıt dosyaları ve ekrandaki VMU görünürlük ayarlarını yapılandırın."
#define CORE_OPTION_NAME_REGION_LABEL_TR "Bölge"
#define OPTION_VAL_JAPAN_TR "Japonya"
#define OPTION_VAL_USA_TR "ABD"
#define OPTION_VAL_EUROPE_TR "Avrupa"
#define OPTION_VAL_DEFAULT_TR "Varsayılan"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_TR "Dil"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_TR "BIOS ve birden çok dil içeren oyunlar tarafından kullanılan dili değiştirir."
#define OPTION_VAL_JAPANESE_TR "Japonca"
#define OPTION_VAL_ENGLISH_TR "İngilizce"
#define OPTION_VAL_GERMAN_TR "Almanca"
#define OPTION_VAL_FRENCH_TR "Fransızca"
#define OPTION_VAL_SPANISH_TR "İspanyolca"
#define OPTION_VAL_ITALIAN_TR "İtalyanca"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_TR "HLE BIOS (Yeniden Başlatılmalı)"
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_TR "Yüksek seviyeli taklit BIOS kullanımını zorunlu kılın."
#define CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_TR "BIOS Önyükleme (Yeniden Başlatılmalı)"
#define CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_TR "Doğrudan Dreamcast BIOS menüsüne önyükleme yapın."
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_TR "DSP'yi Etkinleştir"
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_TR "Dreamcast ses DSP (dijital sinyal işlemcisi) taklidini etkinleştirin. Üretilen sesin doğruluğunu artırır, ancak performans gereksinimlerini artırır."
#define CORE_OPTION_NAME_FORCE_WINCE_LABEL_TR "Windows CE Kipini Zorla"
#define CORE_OPTION_NAME_FORCE_WINCE_INFO_0_TR "Windows CE oyunları için tam MMU (Bellek Yönetim Birimi) öykünmesini ve diğer ayarları etkinleştirin."
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_TR "NAOMI Hizmet Düğmelerine İzin Ver"
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_TR "Kabin ayarlarına girmek için NAOMI için SERVICE düğmesini etkinleştirir."
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_TR "NAOMI Oyunlarını Serbest Oynamaya Ayarlayın"
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_TR "Serbest oynamak için oyunun jeton ayarlarını değiştirin."
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_TR "Geniş Bant Adaptör Taklidi"
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_TR "Modem yerine ethernet geniş bant adaptörünü taklit edin. (Yeniden Başlatılmalı)"
#define CORE_OPTION_NAME_UPNP_LABEL_TR "UPnP Etkinleştir"
#define CORE_OPTION_NAME_UPNP_INFO_0_TR "İnternet yönlendiricinizi çevrimiçi oyunlar için otomatik olarak yapılandırmak üzere UPnP kullanın."
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_TR "Dahili Çözünürlük"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_TR "İşleme çözünürlüğünü değiştirin."
#define OPTION_VAL_320X240_TR NULL
#define OPTION_VAL_640X480_TR NULL
#define OPTION_VAL_800X600_TR NULL
#define OPTION_VAL_960X720_TR NULL
#define OPTION_VAL_1024X768_TR NULL
#define OPTION_VAL_1280X960_TR NULL
#define OPTION_VAL_1440X1080_TR NULL
#define OPTION_VAL_1600X1200_TR NULL
#define OPTION_VAL_1920X1440_TR NULL
#define OPTION_VAL_2560X1920_TR NULL
#define OPTION_VAL_2880X2160_TR NULL
#define OPTION_VAL_3200X2400_TR NULL
#define OPTION_VAL_3840X2880_TR NULL
#define OPTION_VAL_4480X3360_TR NULL
#define OPTION_VAL_5120X3840_TR NULL
#define OPTION_VAL_5760X4320_TR NULL
#define OPTION_VAL_6400X4800_TR NULL
#define OPTION_VAL_7040X5280_TR NULL
#define OPTION_VAL_7680X5760_TR NULL
#define OPTION_VAL_8320X6240_TR NULL
#define OPTION_VAL_8960X6720_TR NULL
#define OPTION_VAL_9600X7200_TR NULL
#define OPTION_VAL_10240X7680_TR NULL
#define OPTION_VAL_10880X8160_TR NULL
#define OPTION_VAL_11520X8640_TR NULL
#define OPTION_VAL_12160X9120_TR NULL
#define OPTION_VAL_12800X9600_TR NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_TR "Kablo Tipi"
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_TR "Çıkış sinyali türü. 'TV (Kompozit)' en yaygın olarak desteklenendir."
#define OPTION_VAL_VGA_TR NULL
#define OPTION_VAL_TV_RGB_TR NULL
#define OPTION_VAL_TV_COMPOSITE_TR NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_TR "Standart Yayın"
#define OPTION_VAL_NTSC_TR NULL
#define OPTION_VAL_PAL_TR "PAL (Dünya)"
#define OPTION_VAL_PAL_N_TR "PAL-N (Arjantin, Paraguay, Uruguay)"
#define OPTION_VAL_PAL_M_TR "PAL-M (Brezilya)"
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_TR "Ekran Yönü"
#define OPTION_VAL_HORIZONTAL_TR "Yatay"
#define OPTION_VAL_VERTICAL_TR "Dikey"
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_TR "Alfabetik Sıralama"
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_TR "Şerit Başına (hızlı, en az doğru)"
#define OPTION_VAL_PER_TRIANGLE_NORMAL_TR "Üçgen Başına (normal)"
#define OPTION_VAL_PER_PIXEL_ACCURATE_TR "Piksel Başına (doğru, ancak en yavaş)"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_TR "Toplama Piksel Tampon Boyutu"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_TR "Daha yüksek çözünürlüklerin doğru çıktı alması için daha yüksek değerler gerekebilir."
#define OPTION_VAL_512MB_TR NULL
#define OPTION_VAL_1GB_TR NULL
#define OPTION_VAL_2GB_TR NULL
#define OPTION_VAL_4GB_TR NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_TR "En Fazla Şeffaf Katman"
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_TR "Karmaşık sahneler için daha yüksek değerler gerekebilir."
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_TR "Tam kare önbellek taklidi"
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_TR "VRAM üstünde tam kare önbellek taklidini etkinleştirin. Kare önbelleğini doğrudan VRAM üstünden okuyan veya yazan oyunlar için kullanışlıdır. Etkinleştirildiğinde, Dahili Çözünürlük 640x480 olarak zorlanır ve performans ciddi şekilde etkilenebilir."
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_TR "RTT (Dokuya Dönüştür) Önbelleğini Etkinleştir"
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_TR "İşlenen dokuları GPU'dan VRAM'e geri kopyalayın. Bu seçenek normalde bunu gerektiren oyunlar için etkinleştirilir. Etkinleştirildiğinde, doku işleme yükseltme devre dışı bırakılır ve performans etkilenebilir."
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_TR "Görüş Açısı"
#define CORE_OPTION_NAME_FOG_LABEL_TR "Sis Efektleri"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_TR "Hacim Değiştirici"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_TR "Genellikle oyunlar tarafından nesne gölgeleri çizmek için kullanılan bir Dreamcast GPU özelliği. Bu normalde etkinleştirilmelidir - performansa olumsuz etkisi yok denilebilecek düzeydedir."
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_TR "Eşyönsüz Filtreleme"
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_TR "Kameraya göre eğik görüş açılarında olan yüzeylerdeki dokuların kalitesini iyileştirin. Daha yüksek değerler GPU'da daha fazla talepkardır. Bu ayarda yapılan değişiklikler yalnızca yeniden başlatmanın ardından geçerlidir."
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_TR "Doku Filtreleme"
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_TR "Kullanılacak doku filtreleme kipi. Bu, Varsayılan'dan daha net (veya pürüzsüz) bir görünüm elde etmek için tüm dokularda belirli bir doku filtreleme kipini zorlamak için kullanılabilir. Varsayılan dışındaki değerler, çeşitli işleme sorunlarına neden olabilir. Bu ayarda yapılan değişiklikler yalnızca yeniden başlatmanın ardından geçerlidir."
#define OPTION_VAL_1_TR "En Yakına Zorla"
#define OPTION_VAL_2_TR "Doğrusala Zorla"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_TR "Gecikmeli Kare Değişimi"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_TR "Yanıp sönen ekranlardan veya sorunlu videolardan kaçınmak için kullanışlıdır. Yavaş platformlarda önerilmez."
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_TR "Kare Hızı Değişikliklerini Algıla"
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_TR "Dahili kare hızı değiştiğinde (örneğin, 60 fps'den 30 fps'ye) ön ucu bilgilendirin. Kilitli 30 fps veya 20 fps'de çalışan oyunlarda kare hızını iyileştirir, ancak kilitlenmemiş (dengesiz) kare hızlarına sahip oyunlar için devre dışı bırakılması gerekir (örn. Ecco the Dolphin, Unreal Tournament). Not: 'Otomatik Kare Atla' etkinleştirildiğinde kullanılamaz."
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_TR "PowerVR2 Rötuş Filtresi"
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_TR "PowerVR2 GPU'ya ve analog video sinyallerine özgü efektleri simüle etmek için işlenen görüntüyü son işlemden geçirin."
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_TR "Doku Yükseltme (xBRZ)"
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_TR "Elle çizilmiş 2B piksel sanatsal grafiklerini geliştirin. Yalnızca 2D piksel oyunlarla kullanılmalıdır."
#define OPTION_VAL_2_O27_TR NULL
#define OPTION_VAL_4_TR NULL
#define OPTION_VAL_6_TR NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_TR "Doku Yükseltme Azami Filtre Boyutu"
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_TR "Yerel Derinlik Enterpolasyonu"
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_TR "AMD GPU'larda doku bozulması ve derinlik sorunlarına yardımcı olur. Bazı durumlarda Intel GPU'lara da yardımcı olabilir."
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_TR "Zorlu İşleme"
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_TR "GPU ve CPU'yu farklı iş parçacıklarında çalıştırır. Şiddetle tavsiye edilir."
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_TR "Kareyi Otomatik Atla"
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_TR "Emulatör yavaş çalışırken kareleri otomatik olarak atlayın. Not: Bu ayar yalnızca 'Zorlu İşleme' etkinleştirildiğinde geçerlidir."
#define OPTION_VAL_SOME_TR NULL
#define OPTION_VAL_MORE_TR "Azami"
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_TR "Kare Atlama"
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_TR "Görüntülenen her kare arasında atlanacak kare sayısını ayarlar."
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_TR "Geniş Ekran Hileleri (Yeniden Başlatılmalı)"
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_TR "Belirli oyunların geniş ekran biçiminde görüntülenmesine izin veren hileleri etkinleştirir."
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_TR "Geniş Ekran Geliştirmesi"
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_TR "Normal 4:3 en boy oranının dışına geometri çizin. Ortaya çıkan alanlarda grafiksel hatalara neden olabilir."
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_TR "GDROM Hızlı Yükleme (kusurlu)"
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_TR "GD-ROM yüklemesini hızlandırır."
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_TR "Özel Dokular Yükle"
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_TR "Dokuları Sakla"
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_TR "Analog Çubuğu Ölü Bölge"
#define OPTION_VAL_0_TR "%0"
#define OPTION_VAL_5_TR "%5"
#define OPTION_VAL_10_TR "%10"
#define OPTION_VAL_15_TR "%15"
#define OPTION_VAL_20_TR "%20"
#define OPTION_VAL_25_TR "%25"
#define OPTION_VAL_30_TR "%30"
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_TR "Tetik Ölü Bölge"
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_TR "Dijital Tetikleyiciler"
#define CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_TR "Purupuru Paketi/Titreşim Paketi"
#define CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_TR "Kontrolcü için zorla titreşimi etkinleştirir."
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_TR "Yayın Dijital Çıkışları"
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_TR "TCP port 8000'de dijital çıkışları ve zorla geri bildirim durumunu yayınlayın. \"-output network\" MAME seçeneğiyle uyumludur."
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_TR "Light Gun Ayarlarını Göster"
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_TR "Light gun artı gösterge seçeneklerinin yapılandırılmasını etkinleştirin. NOT: Bu ayarın etkili olması için Hızlı Menü'nün değiştirilmesi gerekebilir."
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_TR "Silah Artı Göstergesi 1 Görüntüle"
#define OPTION_VAL_WHITE_TR "Beyaz"
#define OPTION_VAL_RED_TR "Kırmızı"
#define OPTION_VAL_GREEN_TR "Yeşil"
#define OPTION_VAL_BLUE_TR "Mavi"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_TR "Silah Artı Göstergesi 2 Görüntüle"
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_TR "Silah Artı Göstergesi 3 Görüntüle"
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_TR "Silah Artı Göstergesi 4 Görüntüle"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_TR "Oyun Başına Görsel Bellek Birimleri/Sistemleri (VMU)"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_TR "Oyun Başına VMU"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_TR "Devre dışı bırakıldığında, tüm oyunlar RetroArch sistem dizininde bulunan 4 VMU kayıt dosyasını (A1, B1, C1, D1) paylaşır. 'VMU A1' ayarı, başlatılan her oyun için RetroArch kayıt dizininde benzersiz bir VMU 'A1' dosyası oluşturur. 'Tüm VMU'' ayarı, başlatılan her oyun için 4 benzersiz VMU dosyası (A1, B1, C1, D1) oluşturur."
#define OPTION_VAL_VMU_A1_TR NULL
#define OPTION_VAL_ALL_VMUS_TR "Tüm VMU"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_TR "Görsel Bellek Birimi/Sistem (VMU) Görüntü Ayarlarını Göster"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_TR "VMU Görüntü Ayarlarını Göster"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_TR "Taklit edilmiş VMU LCD ekran görünürlüğünün, boyutunun, konumunun ve renginin yapılandırılmasını etkinleştirin. NOT: Bu ayarın etkili olması için Hızlı Menü'nün değiştirilmesi gerekebilir."
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_TR "VMU Ekranı 1 Görüntüle"
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_TR "VMU Ekranı 1 Pozisyon"
#define OPTION_VAL_UPPER_LEFT_TR "Sol Üst"
#define OPTION_VAL_UPPER_RIGHT_TR "Sağ Üst"
#define OPTION_VAL_LOWER_LEFT_TR "Sol Alt"
#define OPTION_VAL_LOWER_RIGHT_TR "Sağ Alt"
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_TR "VMU Ekranı 1 Boyutu"
#define OPTION_VAL_1X_TR NULL
#define OPTION_VAL_3X_TR NULL
#define OPTION_VAL_5X_TR NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_TR "VMU Ekranı 1 Piksel Açık Renkli"
#define OPTION_VAL_DEFAULT_ON_00_TR "Varsayılan AÇIK"
#define OPTION_VAL_DEFAULT_OFF_01_TR "Varsayılan KAPALI"
#define OPTION_VAL_BLACK_02_TR "Siyah"
#define OPTION_VAL_LIGHT_BLUE_04_TR "Açık Mavi"
#define OPTION_VAL_CYAN_06_TR "Camgöbeği"
#define OPTION_VAL_CYAN_BLUE_07_TR "Camgöbeği Mavi"
#define OPTION_VAL_LIGHT_GREEN_08_TR "Açık Yeşil"
#define OPTION_VAL_CYAN_GREEN_09_TR "Camgöbeği Yeşil"
#define OPTION_VAL_LIGHT_CYAN_10_TR "Açık Camgöbeği"
#define OPTION_VAL_PURPLE_12_TR "Mor"
#define OPTION_VAL_LIGHT_PURPLE_13_TR "Açık Mor"
#define OPTION_VAL_YELLOW_14_TR "Sarı"
#define OPTION_VAL_GRAY_15_TR "Gri"
#define OPTION_VAL_LIGHT_PURPLE_2_16_TR "Açık Mor (2)"
#define OPTION_VAL_LIGHT_GREEN_2_17_TR "Açık Yeşil (2)"
#define OPTION_VAL_LIGHT_GREEN_3_18_TR "Açık Yeşil (3)"
#define OPTION_VAL_LIGHT_CYAN_2_19_TR "Açık Camgöbeği (2)"
#define OPTION_VAL_LIGHT_RED_2_20_TR "Açık Kırmızı (2)"
#define OPTION_VAL_MAGENTA_21_TR "Eflatun"
#define OPTION_VAL_LIGHT_PURPLE_3_22_TR "Açık Mor (3)"
#define OPTION_VAL_LIGHT_ORANGE_23_TR "Açık Turuncu"
#define OPTION_VAL_ORANGE_24_TR "Turuncu"
#define OPTION_VAL_LIGHT_PURPLE_4_25_TR "Açık Mor (4)"
#define OPTION_VAL_LIGHT_YELLOW_26_TR "Açık Sarı"
#define OPTION_VAL_LIGHT_YELLOW_2_27_TR "Açık Sarı (2)"
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_TR "VMU Ekranı 1 Piksel Kapalı Renk"
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_TR "VMU Ekranı 1 Şeffaflığı"
#define OPTION_VAL_40_TR "%40"
#define OPTION_VAL_50_TR "%50"
#define OPTION_VAL_60_TR "%60"
#define OPTION_VAL_70_TR "%70"
#define OPTION_VAL_80_TR "%80"
#define OPTION_VAL_90_TR "%90"
#define OPTION_VAL_100_TR "%100"
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_TR "VMU Ekranı 2 Görüntüle"
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_TR "VMU Ekranı 2 Pozisyon"
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_TR "VMU Ekranı 2 Boyut"
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_TR "VMU Ekranı 2 Piksel Açık Renkli"
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_TR "VMU Ekranı 2 Piksel Kapalı Renk"
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_TR "VMU Ekranı 2 Şeffaflığı"
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_TR "VMU Ekranı 3 Görüntüle"
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_TR "VMU Ekranı 3 Pozisyon"
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_TR "VMU Ekranı 3 Boyut"
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_TR "VMU Ekranı 3 Piksel Açık Renkli"
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_TR "VMU Ekranı 3 Piksel Kapalı Renk"
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_TR "VMU Ekranı 3 Şeffaflığı"
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_TR "VMU Ekranı 4 Görüntüle"
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_TR "VMU Ekranı 4 Pozisyon"
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_TR "VMU Ekranı 4 Boyut"
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_TR "VMU Ekranı 4 Piksel Açık Renkli"
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_TR "VMU Ekranı 4 Piksel Kapalı Renk"
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_TR "VMU Ekranı 4 Şeffaflığı"

struct retro_core_option_v2_category option_cats_tr[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_TR,
      CATEGORY_SYSTEM_INFO_0_TR
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_TR,
      CATEGORY_VIDEO_INFO_0_TR
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_TR,
      CATEGORY_PERFORMANCE_INFO_0_TR
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_TR,
      CATEGORY_HACKS_INFO_0_TR
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_TR,
      CATEGORY_INPUT_INFO_0_TR
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_TR,
      CATEGORY_VMU_INFO_0_TR
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_tr[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_TR,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_TR },
         { "USA",     OPTION_VAL_USA_TR },
         { "Europe",  OPTION_VAL_EUROPE_TR },
         { "Default", OPTION_VAL_DEFAULT_TR },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_TR,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_TR },
         { "English",  OPTION_VAL_ENGLISH_TR },
         { "German",   OPTION_VAL_GERMAN_TR },
         { "French",   OPTION_VAL_FRENCH_TR },
         { "Spanish",  OPTION_VAL_SPANISH_TR },
         { "Italian",  OPTION_VAL_ITALIAN_TR },
         { "Default",  OPTION_VAL_DEFAULT_TR },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_TR,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_boot_to_bios",
      CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_TR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_TR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_force_wince",
      CORE_OPTION_NAME_FORCE_WINCE_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_FORCE_WINCE_INFO_0_TR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_TR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_TR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_TR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_TR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_TR,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_TR },
         { "640x480",    OPTION_VAL_640X480_TR },
         { "800x600",    OPTION_VAL_800X600_TR },
         { "960x720",    OPTION_VAL_960X720_TR },
         { "1024x768",   OPTION_VAL_1024X768_TR },
         { "1280x960",   OPTION_VAL_1280X960_TR },
         { "1440x1080",  OPTION_VAL_1440X1080_TR },
         { "1600x1200",  OPTION_VAL_1600X1200_TR },
         { "1920x1440",  OPTION_VAL_1920X1440_TR },
         { "2560x1920",  OPTION_VAL_2560X1920_TR },
         { "2880x2160",  OPTION_VAL_2880X2160_TR },
         { "3200x2400",  OPTION_VAL_3200X2400_TR },
         { "3840x2880",  OPTION_VAL_3840X2880_TR },
         { "4480x3360",  OPTION_VAL_4480X3360_TR },
         { "5120x3840",  OPTION_VAL_5120X3840_TR },
         { "5760x4320",  OPTION_VAL_5760X4320_TR },
         { "6400x4800",  OPTION_VAL_6400X4800_TR },
         { "7040x5280",  OPTION_VAL_7040X5280_TR },
         { "7680x5760",  OPTION_VAL_7680X5760_TR },
         { "8320x6240",  OPTION_VAL_8320X6240_TR },
         { "8960x6720",  OPTION_VAL_8960X6720_TR },
         { "9600x7200",  OPTION_VAL_9600X7200_TR },
         { "10240x7680", OPTION_VAL_10240X7680_TR },
         { "10880x8160", OPTION_VAL_10880X8160_TR },
         { "11520x8640", OPTION_VAL_11520X8640_TR },
         { "12160x9120", OPTION_VAL_12160X9120_TR },
         { "12800x9600", OPTION_VAL_12800X9600_TR },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_TR,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_TR },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_TR },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_TR },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_TR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_TR },
         { "PAL",     OPTION_VAL_PAL_TR },
         { "PAL_N",   OPTION_VAL_PAL_N_TR },
         { "PAL_M",   OPTION_VAL_PAL_M_TR },
         { "Default", OPTION_VAL_DEFAULT_TR },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_TR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_TR },
         { "vertical",   OPTION_VAL_VERTICAL_TR },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_TR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_TR },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_TR },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_TR },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_TR,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_TR },
         { "1GB",   OPTION_VAL_1GB_TR },
         { "2GB",   OPTION_VAL_2GB_TR },
         { "4GB",   OPTION_VAL_4GB_TR },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_TR,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_TR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_TR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_TR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_TR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_TR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_TR,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_TR,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_TR },
         { "1",  OPTION_VAL_1_TR },
         { "2",  OPTION_VAL_2_TR },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_TR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_TR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_TR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_TR,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O27_TR },
         { "4",  OPTION_VAL_4_TR },
         { "6",  OPTION_VAL_6_TR },
         { NULL, NULL },
      },
      "1",
   },
   {/* TODO: needs clarification */
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_TR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_TR,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_TR,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_TR,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_TR,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_TR },
         { "more", OPTION_VAL_MORE_TR },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_TR,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_TR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_TR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_TR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_TR,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: probably needs explanation */
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_TR,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_TR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_TR },
         { "5%",  OPTION_VAL_5_TR },
         { "10%", OPTION_VAL_10_TR },
         { "15%", OPTION_VAL_15_TR },
         { "20%", OPTION_VAL_20_TR },
         { "25%", OPTION_VAL_25_TR },
         { "30%", OPTION_VAL_30_TR },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_TR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_TR },
         { "5%",  OPTION_VAL_5_TR },
         { "10%", OPTION_VAL_10_TR },
         { "15%", OPTION_VAL_15_TR },
         { "20%", OPTION_VAL_20_TR },
         { "25%", OPTION_VAL_25_TR },
         { "30%", OPTION_VAL_30_TR },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_TR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_purupuru",
      CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_TR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_TR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_TR,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_TR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_TR },
         { "Red",      OPTION_VAL_RED_TR },
         { "Green",    OPTION_VAL_GREEN_TR },
         { "Blue",     OPTION_VAL_BLUE_TR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_TR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_TR },
         { "Red",      OPTION_VAL_RED_TR },
         { "Green",    OPTION_VAL_GREEN_TR },
         { "Blue",     OPTION_VAL_BLUE_TR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_TR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_TR },
         { "Red",      OPTION_VAL_RED_TR },
         { "Green",    OPTION_VAL_GREEN_TR },
         { "Blue",     OPTION_VAL_BLUE_TR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_TR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_TR },
         { "Red",      OPTION_VAL_RED_TR },
         { "Green",    OPTION_VAL_GREEN_TR },
         { "Blue",     OPTION_VAL_BLUE_TR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_TR,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_TR,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_TR,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_TR },
         { "All VMUs", OPTION_VAL_ALL_VMUS_TR },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_TR,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_TR,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_TR,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_TR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_TR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_TR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_TR },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_TR },
         { "2x", OPTION_VAL_2_O27_TR },
         { "3x", OPTION_VAL_3X_TR },
         { "4x", OPTION_VAL_4_TR },
         { "5x", OPTION_VAL_5X_TR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_TR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_TR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_TR },
         { "BLUE 03",           OPTION_VAL_BLUE_TR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_TR },
         { "GREEN 05",          OPTION_VAL_GREEN_TR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_TR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_TR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_TR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_TR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_TR },
         { "RED 11",            OPTION_VAL_RED_TR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_TR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_TR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_TR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_TR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_TR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_TR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_TR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_TR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_TR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_TR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_TR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_TR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_TR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_TR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_TR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_TR },
         { "WHITE 28",          OPTION_VAL_WHITE_TR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_TR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_TR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_TR },
         { "BLUE 03",           OPTION_VAL_BLUE_TR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_TR },
         { "GREEN 05",          OPTION_VAL_GREEN_TR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_TR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_TR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_TR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_TR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_TR },
         { "RED 11",            OPTION_VAL_RED_TR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_TR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_TR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_TR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_TR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_TR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_TR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_TR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_TR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_TR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_TR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_TR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_TR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_TR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_TR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_TR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_TR },
         { "WHITE 28",          OPTION_VAL_WHITE_TR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_TR },
         { "20%",  OPTION_VAL_20_TR },
         { "30%",  OPTION_VAL_30_TR },
         { "40%",  OPTION_VAL_40_TR },
         { "50%",  OPTION_VAL_50_TR },
         { "60%",  OPTION_VAL_60_TR },
         { "70%",  OPTION_VAL_70_TR },
         { "80%",  OPTION_VAL_80_TR },
         { "90%",  OPTION_VAL_90_TR },
         { "100%", OPTION_VAL_100_TR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_TR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_TR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_TR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_TR },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_TR },
         { "2x", OPTION_VAL_2_O27_TR },
         { "3x", OPTION_VAL_3X_TR },
         { "4x", OPTION_VAL_4_TR },
         { "5x", OPTION_VAL_5X_TR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_TR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_TR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_TR },
         { "BLUE 03",           OPTION_VAL_BLUE_TR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_TR },
         { "GREEN 05",          OPTION_VAL_GREEN_TR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_TR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_TR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_TR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_TR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_TR },
         { "RED 11",            OPTION_VAL_RED_TR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_TR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_TR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_TR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_TR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_TR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_TR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_TR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_TR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_TR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_TR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_TR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_TR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_TR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_TR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_TR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_TR },
         { "WHITE 28",          OPTION_VAL_WHITE_TR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_TR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_TR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_TR },
         { "BLUE 03",           OPTION_VAL_BLUE_TR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_TR },
         { "GREEN 05",          OPTION_VAL_GREEN_TR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_TR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_TR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_TR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_TR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_TR },
         { "RED 11",            OPTION_VAL_RED_TR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_TR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_TR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_TR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_TR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_TR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_TR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_TR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_TR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_TR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_TR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_TR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_TR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_TR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_TR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_TR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_TR },
         { "WHITE 28",          OPTION_VAL_WHITE_TR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_TR },
         { "20%",  OPTION_VAL_20_TR },
         { "30%",  OPTION_VAL_30_TR },
         { "40%",  OPTION_VAL_40_TR },
         { "50%",  OPTION_VAL_50_TR },
         { "60%",  OPTION_VAL_60_TR },
         { "70%",  OPTION_VAL_70_TR },
         { "80%",  OPTION_VAL_80_TR },
         { "90%",  OPTION_VAL_90_TR },
         { "100%", OPTION_VAL_100_TR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_TR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_TR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_TR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_TR },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_TR },
         { "2x", OPTION_VAL_2_O27_TR },
         { "3x", OPTION_VAL_3X_TR },
         { "4x", OPTION_VAL_4_TR },
         { "5x", OPTION_VAL_5X_TR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_TR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_TR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_TR },
         { "BLUE 03",           OPTION_VAL_BLUE_TR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_TR },
         { "GREEN 05",          OPTION_VAL_GREEN_TR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_TR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_TR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_TR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_TR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_TR },
         { "RED 11",            OPTION_VAL_RED_TR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_TR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_TR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_TR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_TR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_TR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_TR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_TR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_TR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_TR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_TR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_TR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_TR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_TR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_TR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_TR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_TR },
         { "WHITE 28",          OPTION_VAL_WHITE_TR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_TR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_TR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_TR },
         { "BLUE 03",           OPTION_VAL_BLUE_TR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_TR },
         { "GREEN 05",          OPTION_VAL_GREEN_TR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_TR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_TR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_TR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_TR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_TR },
         { "RED 11",            OPTION_VAL_RED_TR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_TR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_TR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_TR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_TR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_TR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_TR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_TR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_TR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_TR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_TR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_TR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_TR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_TR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_TR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_TR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_TR },
         { "WHITE 28",          OPTION_VAL_WHITE_TR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_TR },
         { "20%",  OPTION_VAL_20_TR },
         { "30%",  OPTION_VAL_30_TR },
         { "40%",  OPTION_VAL_40_TR },
         { "50%",  OPTION_VAL_50_TR },
         { "60%",  OPTION_VAL_60_TR },
         { "70%",  OPTION_VAL_70_TR },
         { "80%",  OPTION_VAL_80_TR },
         { "90%",  OPTION_VAL_90_TR },
         { "100%", OPTION_VAL_100_TR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_TR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_TR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_TR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_TR },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_TR },
         { "2x", OPTION_VAL_2_O27_TR },
         { "3x", OPTION_VAL_3X_TR },
         { "4x", OPTION_VAL_4_TR },
         { "5x", OPTION_VAL_5X_TR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_TR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_TR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_TR },
         { "BLUE 03",           OPTION_VAL_BLUE_TR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_TR },
         { "GREEN 05",          OPTION_VAL_GREEN_TR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_TR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_TR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_TR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_TR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_TR },
         { "RED 11",            OPTION_VAL_RED_TR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_TR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_TR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_TR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_TR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_TR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_TR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_TR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_TR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_TR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_TR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_TR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_TR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_TR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_TR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_TR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_TR },
         { "WHITE 28",          OPTION_VAL_WHITE_TR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_TR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_TR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_TR },
         { "BLUE 03",           OPTION_VAL_BLUE_TR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_TR },
         { "GREEN 05",          OPTION_VAL_GREEN_TR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_TR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_TR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_TR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_TR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_TR },
         { "RED 11",            OPTION_VAL_RED_TR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_TR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_TR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_TR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_TR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_TR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_TR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_TR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_TR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_TR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_TR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_TR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_TR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_TR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_TR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_TR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_TR },
         { "WHITE 28",          OPTION_VAL_WHITE_TR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_TR },
         { "20%",  OPTION_VAL_20_TR },
         { "30%",  OPTION_VAL_30_TR },
         { "40%",  OPTION_VAL_40_TR },
         { "50%",  OPTION_VAL_50_TR },
         { "60%",  OPTION_VAL_60_TR },
         { "70%",  OPTION_VAL_70_TR },
         { "80%",  OPTION_VAL_80_TR },
         { "90%",  OPTION_VAL_90_TR },
         { "100%", OPTION_VAL_100_TR },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_tr = {
   option_cats_tr,
   option_defs_tr
};

/* RETRO_LANGUAGE_UK */

#define CATEGORY_SYSTEM_LABEL_UK "Система"
#define CATEGORY_SYSTEM_INFO_0_UK NULL
#define CATEGORY_VIDEO_LABEL_UK "Відео"
#define CATEGORY_VIDEO_INFO_0_UK NULL
#define CATEGORY_PERFORMANCE_LABEL_UK "Продуктивність"
#define CATEGORY_PERFORMANCE_INFO_0_UK "Налаштування потокового рендерингу та пропуску кадрів."
#define CATEGORY_HACKS_LABEL_UK "Гаки емуляції"
#define CATEGORY_HACKS_INFO_0_UK NULL
#define CATEGORY_INPUT_LABEL_UK "Ввід"
#define CATEGORY_INPUT_INFO_0_UK NULL
#define CATEGORY_VMU_LABEL_UK NULL
#define CATEGORY_VMU_INFO_0_UK NULL
#define CORE_OPTION_NAME_REGION_LABEL_UK "Регіон"
#define OPTION_VAL_JAPAN_UK NULL
#define OPTION_VAL_USA_UK NULL
#define OPTION_VAL_EUROPE_UK NULL
#define OPTION_VAL_DEFAULT_UK "За замовчуванням"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_UK "Мова"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_UK NULL
#define OPTION_VAL_JAPANESE_UK NULL
#define OPTION_VAL_ENGLISH_UK NULL
#define OPTION_VAL_GERMAN_UK NULL
#define OPTION_VAL_FRENCH_UK NULL
#define OPTION_VAL_SPANISH_UK NULL
#define OPTION_VAL_ITALIAN_UK NULL
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_UK NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_UK "Примусово використовувати BIOS високого рівня."
#define CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_UK "Завантаження в BIOS (потрібний перезапуск)"
#define CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_UK "Завантаження безпосередньо в меню Dreamcast BIOS."
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_UK "Увімкнути DSP"
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_UK "Увімкнути емуляцію аудіо DSP(цифровий процесор сигналів). Покращує точність згенерованого звуку, але збільшує вимоги до продуктивності."
#define CORE_OPTION_NAME_FORCE_WINCE_LABEL_UK NULL
#define CORE_OPTION_NAME_FORCE_WINCE_INFO_0_UK NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_UK NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_UK NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_UK "Встановити NAOMI ігри для безкоштовної гри"
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_UK NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_UK NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_UK NULL
#define CORE_OPTION_NAME_UPNP_LABEL_UK NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_UK NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_UK "Роздільна здатність"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_UK "Змінити роздільну здатність візуалізації."
#define OPTION_VAL_320X240_UK NULL
#define OPTION_VAL_640X480_UK "640х480"
#define OPTION_VAL_800X600_UK NULL
#define OPTION_VAL_960X720_UK "960х720"
#define OPTION_VAL_1024X768_UK NULL
#define OPTION_VAL_1280X960_UK NULL
#define OPTION_VAL_1440X1080_UK "1440х1080"
#define OPTION_VAL_1600X1200_UK NULL
#define OPTION_VAL_1920X1440_UK "1920х1440"
#define OPTION_VAL_2560X1920_UK "2560х1920"
#define OPTION_VAL_2880X2160_UK "2880х2160"
#define OPTION_VAL_3200X2400_UK "3200х2400"
#define OPTION_VAL_3840X2880_UK "3840х2880"
#define OPTION_VAL_4480X3360_UK "4480х3360"
#define OPTION_VAL_5120X3840_UK "5120х3840"
#define OPTION_VAL_5760X4320_UK "5760х4320"
#define OPTION_VAL_6400X4800_UK "6400х4800"
#define OPTION_VAL_7040X5280_UK "7040х5280"
#define OPTION_VAL_7680X5760_UK "7680х5760"
#define OPTION_VAL_8320X6240_UK "8320х6240"
#define OPTION_VAL_8960X6720_UK "8960х6720"
#define OPTION_VAL_9600X7200_UK "9600х7200"
#define OPTION_VAL_10240X7680_UK "10240х7680"
#define OPTION_VAL_10880X8160_UK "10880х8160"
#define OPTION_VAL_11520X8640_UK "11520х8640"
#define OPTION_VAL_12160X9120_UK "12160х9120"
#define OPTION_VAL_12800X9600_UK "12800х9600"
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_UK NULL
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_UK NULL
#define OPTION_VAL_VGA_UK NULL
#define OPTION_VAL_TV_RGB_UK NULL
#define OPTION_VAL_TV_COMPOSITE_UK NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_UK NULL
#define OPTION_VAL_NTSC_UK NULL
#define OPTION_VAL_PAL_UK NULL
#define OPTION_VAL_PAL_N_UK NULL
#define OPTION_VAL_PAL_M_UK NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_UK "Орієнтація екрану"
#define OPTION_VAL_HORIZONTAL_UK "Горизонтально"
#define OPTION_VAL_VERTICAL_UK "Вертикально"
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_UK "Альфа-Сортування"
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_UK NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_UK NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_UK NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_UK NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_UK "Для коректного виводу можуть бути потрібні більші значення."
#define OPTION_VAL_512MB_UK "512 МБ"
#define OPTION_VAL_1GB_UK "1 ГБ"
#define OPTION_VAL_2GB_UK "2 ГБ"
#define OPTION_VAL_4GB_UK "4 ГБ"
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_UK NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_UK NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_UK NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_UK NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_UK NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_UK NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_UK NULL
#define CORE_OPTION_NAME_FOG_LABEL_UK NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_UK NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_UK NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_UK "Анізотропна фільтрація"
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_UK NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_UK "Фільтрація текстур"
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_UK NULL
#define OPTION_VAL_1_UK NULL
#define OPTION_VAL_2_UK NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_UK NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_UK NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_UK NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_UK NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_UK NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_UK "Постобробка зображення для імітації ефектів відеоданих на PowerVR2 та аналогових відео сигналів."
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_UK "Вдосконалення текстур (xBRZ)"
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_UK "Покращення графіки 2D пікселів. Має використовуватися лише з 2D піксельними іграми."
#define OPTION_VAL_2_O27_UK NULL
#define OPTION_VAL_4_UK NULL
#define OPTION_VAL_6_UK NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_UK "Збільшення масштабу текстур. Відфільтрований розмір"
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_UK NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_UK NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_UK NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_UK NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_UK NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_UK NULL
#define OPTION_VAL_SOME_UK "Нормально"
#define OPTION_VAL_MORE_UK NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_UK "Пропуск кадрів"
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_UK NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_UK NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_UK NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_UK NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_UK NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_UK NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_UK NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_UK NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_UK NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_UK NULL
#define OPTION_VAL_0_UK NULL
#define OPTION_VAL_5_UK NULL
#define OPTION_VAL_10_UK NULL
#define OPTION_VAL_15_UK NULL
#define OPTION_VAL_20_UK NULL
#define OPTION_VAL_25_UK NULL
#define OPTION_VAL_30_UK NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_UK NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_UK NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_UK NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_UK NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_UK NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_UK NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_UK NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_UK NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_UK NULL
#define OPTION_VAL_WHITE_UK NULL
#define OPTION_VAL_RED_UK "Червоний"
#define OPTION_VAL_GREEN_UK "Зелений"
#define OPTION_VAL_BLUE_UK "Синій"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_UK NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_UK NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_UK NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_UK NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_UK NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_UK NULL
#define OPTION_VAL_VMU_A1_UK NULL
#define OPTION_VAL_ALL_VMUS_UK NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_UK NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_UK NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_UK NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_UK NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_UK NULL
#define OPTION_VAL_UPPER_LEFT_UK NULL
#define OPTION_VAL_UPPER_RIGHT_UK NULL
#define OPTION_VAL_LOWER_LEFT_UK NULL
#define OPTION_VAL_LOWER_RIGHT_UK NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_UK NULL
#define OPTION_VAL_1X_UK NULL
#define OPTION_VAL_3X_UK NULL
#define OPTION_VAL_5X_UK NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_UK NULL
#define OPTION_VAL_DEFAULT_ON_00_UK NULL
#define OPTION_VAL_DEFAULT_OFF_01_UK NULL
#define OPTION_VAL_BLACK_02_UK NULL
#define OPTION_VAL_LIGHT_BLUE_04_UK NULL
#define OPTION_VAL_CYAN_06_UK NULL
#define OPTION_VAL_CYAN_BLUE_07_UK NULL
#define OPTION_VAL_LIGHT_GREEN_08_UK NULL
#define OPTION_VAL_CYAN_GREEN_09_UK NULL
#define OPTION_VAL_LIGHT_CYAN_10_UK NULL
#define OPTION_VAL_PURPLE_12_UK NULL
#define OPTION_VAL_LIGHT_PURPLE_13_UK NULL
#define OPTION_VAL_YELLOW_14_UK "Жовтий"
#define OPTION_VAL_GRAY_15_UK NULL
#define OPTION_VAL_LIGHT_PURPLE_2_16_UK NULL
#define OPTION_VAL_LIGHT_GREEN_2_17_UK NULL
#define OPTION_VAL_LIGHT_GREEN_3_18_UK NULL
#define OPTION_VAL_LIGHT_CYAN_2_19_UK NULL
#define OPTION_VAL_LIGHT_RED_2_20_UK NULL
#define OPTION_VAL_MAGENTA_21_UK NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_UK NULL
#define OPTION_VAL_LIGHT_ORANGE_23_UK NULL
#define OPTION_VAL_ORANGE_24_UK NULL
#define OPTION_VAL_LIGHT_PURPLE_4_25_UK NULL
#define OPTION_VAL_LIGHT_YELLOW_26_UK NULL
#define OPTION_VAL_LIGHT_YELLOW_2_27_UK NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_UK NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_UK NULL
#define OPTION_VAL_40_UK NULL
#define OPTION_VAL_50_UK NULL
#define OPTION_VAL_60_UK NULL
#define OPTION_VAL_70_UK NULL
#define OPTION_VAL_80_UK NULL
#define OPTION_VAL_90_UK NULL
#define OPTION_VAL_100_UK NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_UK NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_UK NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_UK NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_UK NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_UK NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_UK NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_UK NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_UK NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_UK NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_UK NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_UK NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_UK NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_UK NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_UK NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_UK NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_UK NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_UK NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_UK NULL

struct retro_core_option_v2_category option_cats_uk[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_UK,
      CATEGORY_SYSTEM_INFO_0_UK
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_UK,
      CATEGORY_VIDEO_INFO_0_UK
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_UK,
      CATEGORY_PERFORMANCE_INFO_0_UK
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_UK,
      CATEGORY_HACKS_INFO_0_UK
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_UK,
      CATEGORY_INPUT_INFO_0_UK
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_UK,
      CATEGORY_VMU_INFO_0_UK
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_uk[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_UK,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_UK },
         { "USA",     OPTION_VAL_USA_UK },
         { "Europe",  OPTION_VAL_EUROPE_UK },
         { "Default", OPTION_VAL_DEFAULT_UK },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_UK,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_UK },
         { "English",  OPTION_VAL_ENGLISH_UK },
         { "German",   OPTION_VAL_GERMAN_UK },
         { "French",   OPTION_VAL_FRENCH_UK },
         { "Spanish",  OPTION_VAL_SPANISH_UK },
         { "Italian",  OPTION_VAL_ITALIAN_UK },
         { "Default",  OPTION_VAL_DEFAULT_UK },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_UK,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_boot_to_bios",
      CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_UK,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_UK,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_force_wince",
      CORE_OPTION_NAME_FORCE_WINCE_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_FORCE_WINCE_INFO_0_UK,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_UK,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_UK,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_UK,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_UK,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_UK,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_UK },
         { "640x480",    OPTION_VAL_640X480_UK },
         { "800x600",    OPTION_VAL_800X600_UK },
         { "960x720",    OPTION_VAL_960X720_UK },
         { "1024x768",   OPTION_VAL_1024X768_UK },
         { "1280x960",   OPTION_VAL_1280X960_UK },
         { "1440x1080",  OPTION_VAL_1440X1080_UK },
         { "1600x1200",  OPTION_VAL_1600X1200_UK },
         { "1920x1440",  OPTION_VAL_1920X1440_UK },
         { "2560x1920",  OPTION_VAL_2560X1920_UK },
         { "2880x2160",  OPTION_VAL_2880X2160_UK },
         { "3200x2400",  OPTION_VAL_3200X2400_UK },
         { "3840x2880",  OPTION_VAL_3840X2880_UK },
         { "4480x3360",  OPTION_VAL_4480X3360_UK },
         { "5120x3840",  OPTION_VAL_5120X3840_UK },
         { "5760x4320",  OPTION_VAL_5760X4320_UK },
         { "6400x4800",  OPTION_VAL_6400X4800_UK },
         { "7040x5280",  OPTION_VAL_7040X5280_UK },
         { "7680x5760",  OPTION_VAL_7680X5760_UK },
         { "8320x6240",  OPTION_VAL_8320X6240_UK },
         { "8960x6720",  OPTION_VAL_8960X6720_UK },
         { "9600x7200",  OPTION_VAL_9600X7200_UK },
         { "10240x7680", OPTION_VAL_10240X7680_UK },
         { "10880x8160", OPTION_VAL_10880X8160_UK },
         { "11520x8640", OPTION_VAL_11520X8640_UK },
         { "12160x9120", OPTION_VAL_12160X9120_UK },
         { "12800x9600", OPTION_VAL_12800X9600_UK },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_UK,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_UK },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_UK },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_UK },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_UK,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_UK },
         { "PAL",     OPTION_VAL_PAL_UK },
         { "PAL_N",   OPTION_VAL_PAL_N_UK },
         { "PAL_M",   OPTION_VAL_PAL_M_UK },
         { "Default", OPTION_VAL_DEFAULT_UK },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_UK,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_UK },
         { "vertical",   OPTION_VAL_VERTICAL_UK },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_UK,
      NULL,
      "",
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_UK },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_UK },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_UK },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_UK,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_UK },
         { "1GB",   OPTION_VAL_1GB_UK },
         { "2GB",   OPTION_VAL_2GB_UK },
         { "4GB",   OPTION_VAL_4GB_UK },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_UK,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_UK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_UK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_UK,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_UK,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_UK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_UK,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_UK,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_UK },
         { "1",  OPTION_VAL_1_UK },
         { "2",  OPTION_VAL_2_UK },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_UK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_UK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_UK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_UK,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O27_UK },
         { "4",  OPTION_VAL_4_UK },
         { "6",  OPTION_VAL_6_UK },
         { NULL, NULL },
      },
      "1",
   },
   {/* TODO: needs clarification */
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_UK,
      NULL,
      "",
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_UK,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_UK,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_UK,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_UK,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_UK },
         { "more", OPTION_VAL_MORE_UK },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_UK,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_UK,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_UK,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_UK,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_UK,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: probably needs explanation */
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_UK,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_UK,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_UK },
         { "5%",  OPTION_VAL_5_UK },
         { "10%", OPTION_VAL_10_UK },
         { "15%", OPTION_VAL_15_UK },
         { "20%", OPTION_VAL_20_UK },
         { "25%", OPTION_VAL_25_UK },
         { "30%", OPTION_VAL_30_UK },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_UK,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_UK },
         { "5%",  OPTION_VAL_5_UK },
         { "10%", OPTION_VAL_10_UK },
         { "15%", OPTION_VAL_15_UK },
         { "20%", OPTION_VAL_20_UK },
         { "25%", OPTION_VAL_25_UK },
         { "30%", OPTION_VAL_30_UK },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_UK,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_purupuru",
      CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_UK,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_UK,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_UK,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_UK,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_UK },
         { "Red",      OPTION_VAL_RED_UK },
         { "Green",    OPTION_VAL_GREEN_UK },
         { "Blue",     OPTION_VAL_BLUE_UK },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_UK,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_UK },
         { "Red",      OPTION_VAL_RED_UK },
         { "Green",    OPTION_VAL_GREEN_UK },
         { "Blue",     OPTION_VAL_BLUE_UK },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_UK,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_UK },
         { "Red",      OPTION_VAL_RED_UK },
         { "Green",    OPTION_VAL_GREEN_UK },
         { "Blue",     OPTION_VAL_BLUE_UK },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_UK,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_UK },
         { "Red",      OPTION_VAL_RED_UK },
         { "Green",    OPTION_VAL_GREEN_UK },
         { "Blue",     OPTION_VAL_BLUE_UK },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_UK,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_UK,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_UK,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_UK },
         { "All VMUs", OPTION_VAL_ALL_VMUS_UK },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_UK,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_UK,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_UK,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_UK },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_UK },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_UK },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_UK },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_UK },
         { "2x", OPTION_VAL_2_O27_UK },
         { "3x", OPTION_VAL_3X_UK },
         { "4x", OPTION_VAL_4_UK },
         { "5x", OPTION_VAL_5X_UK },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_UK },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_UK },
         { "BLACK 02",          OPTION_VAL_BLACK_02_UK },
         { "BLUE 03",           OPTION_VAL_BLUE_UK },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_UK },
         { "GREEN 05",          OPTION_VAL_GREEN_UK },
         { "CYAN 06",           OPTION_VAL_CYAN_06_UK },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_UK },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_UK },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_UK },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_UK },
         { "RED 11",            OPTION_VAL_RED_UK },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_UK },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_UK },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_UK },
         { "GRAY 15",           OPTION_VAL_GRAY_15_UK },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_UK },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_UK },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_UK },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_UK },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_UK },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_UK },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_UK },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_UK },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_UK },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_UK },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_UK },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_UK },
         { "WHITE 28",          OPTION_VAL_WHITE_UK },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_UK },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_UK },
         { "BLACK 02",          OPTION_VAL_BLACK_02_UK },
         { "BLUE 03",           OPTION_VAL_BLUE_UK },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_UK },
         { "GREEN 05",          OPTION_VAL_GREEN_UK },
         { "CYAN 06",           OPTION_VAL_CYAN_06_UK },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_UK },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_UK },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_UK },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_UK },
         { "RED 11",            OPTION_VAL_RED_UK },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_UK },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_UK },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_UK },
         { "GRAY 15",           OPTION_VAL_GRAY_15_UK },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_UK },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_UK },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_UK },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_UK },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_UK },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_UK },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_UK },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_UK },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_UK },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_UK },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_UK },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_UK },
         { "WHITE 28",          OPTION_VAL_WHITE_UK },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_UK },
         { "20%",  OPTION_VAL_20_UK },
         { "30%",  OPTION_VAL_30_UK },
         { "40%",  OPTION_VAL_40_UK },
         { "50%",  OPTION_VAL_50_UK },
         { "60%",  OPTION_VAL_60_UK },
         { "70%",  OPTION_VAL_70_UK },
         { "80%",  OPTION_VAL_80_UK },
         { "90%",  OPTION_VAL_90_UK },
         { "100%", OPTION_VAL_100_UK },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_UK },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_UK },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_UK },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_UK },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_UK },
         { "2x", OPTION_VAL_2_O27_UK },
         { "3x", OPTION_VAL_3X_UK },
         { "4x", OPTION_VAL_4_UK },
         { "5x", OPTION_VAL_5X_UK },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_UK },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_UK },
         { "BLACK 02",          OPTION_VAL_BLACK_02_UK },
         { "BLUE 03",           OPTION_VAL_BLUE_UK },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_UK },
         { "GREEN 05",          OPTION_VAL_GREEN_UK },
         { "CYAN 06",           OPTION_VAL_CYAN_06_UK },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_UK },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_UK },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_UK },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_UK },
         { "RED 11",            OPTION_VAL_RED_UK },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_UK },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_UK },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_UK },
         { "GRAY 15",           OPTION_VAL_GRAY_15_UK },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_UK },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_UK },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_UK },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_UK },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_UK },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_UK },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_UK },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_UK },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_UK },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_UK },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_UK },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_UK },
         { "WHITE 28",          OPTION_VAL_WHITE_UK },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_UK },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_UK },
         { "BLACK 02",          OPTION_VAL_BLACK_02_UK },
         { "BLUE 03",           OPTION_VAL_BLUE_UK },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_UK },
         { "GREEN 05",          OPTION_VAL_GREEN_UK },
         { "CYAN 06",           OPTION_VAL_CYAN_06_UK },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_UK },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_UK },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_UK },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_UK },
         { "RED 11",            OPTION_VAL_RED_UK },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_UK },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_UK },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_UK },
         { "GRAY 15",           OPTION_VAL_GRAY_15_UK },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_UK },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_UK },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_UK },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_UK },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_UK },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_UK },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_UK },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_UK },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_UK },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_UK },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_UK },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_UK },
         { "WHITE 28",          OPTION_VAL_WHITE_UK },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_UK },
         { "20%",  OPTION_VAL_20_UK },
         { "30%",  OPTION_VAL_30_UK },
         { "40%",  OPTION_VAL_40_UK },
         { "50%",  OPTION_VAL_50_UK },
         { "60%",  OPTION_VAL_60_UK },
         { "70%",  OPTION_VAL_70_UK },
         { "80%",  OPTION_VAL_80_UK },
         { "90%",  OPTION_VAL_90_UK },
         { "100%", OPTION_VAL_100_UK },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_UK },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_UK },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_UK },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_UK },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_UK },
         { "2x", OPTION_VAL_2_O27_UK },
         { "3x", OPTION_VAL_3X_UK },
         { "4x", OPTION_VAL_4_UK },
         { "5x", OPTION_VAL_5X_UK },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_UK },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_UK },
         { "BLACK 02",          OPTION_VAL_BLACK_02_UK },
         { "BLUE 03",           OPTION_VAL_BLUE_UK },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_UK },
         { "GREEN 05",          OPTION_VAL_GREEN_UK },
         { "CYAN 06",           OPTION_VAL_CYAN_06_UK },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_UK },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_UK },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_UK },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_UK },
         { "RED 11",            OPTION_VAL_RED_UK },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_UK },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_UK },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_UK },
         { "GRAY 15",           OPTION_VAL_GRAY_15_UK },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_UK },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_UK },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_UK },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_UK },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_UK },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_UK },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_UK },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_UK },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_UK },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_UK },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_UK },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_UK },
         { "WHITE 28",          OPTION_VAL_WHITE_UK },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_UK },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_UK },
         { "BLACK 02",          OPTION_VAL_BLACK_02_UK },
         { "BLUE 03",           OPTION_VAL_BLUE_UK },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_UK },
         { "GREEN 05",          OPTION_VAL_GREEN_UK },
         { "CYAN 06",           OPTION_VAL_CYAN_06_UK },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_UK },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_UK },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_UK },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_UK },
         { "RED 11",            OPTION_VAL_RED_UK },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_UK },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_UK },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_UK },
         { "GRAY 15",           OPTION_VAL_GRAY_15_UK },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_UK },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_UK },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_UK },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_UK },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_UK },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_UK },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_UK },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_UK },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_UK },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_UK },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_UK },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_UK },
         { "WHITE 28",          OPTION_VAL_WHITE_UK },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_UK },
         { "20%",  OPTION_VAL_20_UK },
         { "30%",  OPTION_VAL_30_UK },
         { "40%",  OPTION_VAL_40_UK },
         { "50%",  OPTION_VAL_50_UK },
         { "60%",  OPTION_VAL_60_UK },
         { "70%",  OPTION_VAL_70_UK },
         { "80%",  OPTION_VAL_80_UK },
         { "90%",  OPTION_VAL_90_UK },
         { "100%", OPTION_VAL_100_UK },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_UK },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_UK },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_UK },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_UK },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_UK },
         { "2x", OPTION_VAL_2_O27_UK },
         { "3x", OPTION_VAL_3X_UK },
         { "4x", OPTION_VAL_4_UK },
         { "5x", OPTION_VAL_5X_UK },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_UK },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_UK },
         { "BLACK 02",          OPTION_VAL_BLACK_02_UK },
         { "BLUE 03",           OPTION_VAL_BLUE_UK },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_UK },
         { "GREEN 05",          OPTION_VAL_GREEN_UK },
         { "CYAN 06",           OPTION_VAL_CYAN_06_UK },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_UK },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_UK },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_UK },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_UK },
         { "RED 11",            OPTION_VAL_RED_UK },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_UK },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_UK },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_UK },
         { "GRAY 15",           OPTION_VAL_GRAY_15_UK },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_UK },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_UK },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_UK },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_UK },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_UK },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_UK },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_UK },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_UK },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_UK },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_UK },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_UK },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_UK },
         { "WHITE 28",          OPTION_VAL_WHITE_UK },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_UK },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_UK },
         { "BLACK 02",          OPTION_VAL_BLACK_02_UK },
         { "BLUE 03",           OPTION_VAL_BLUE_UK },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_UK },
         { "GREEN 05",          OPTION_VAL_GREEN_UK },
         { "CYAN 06",           OPTION_VAL_CYAN_06_UK },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_UK },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_UK },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_UK },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_UK },
         { "RED 11",            OPTION_VAL_RED_UK },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_UK },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_UK },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_UK },
         { "GRAY 15",           OPTION_VAL_GRAY_15_UK },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_UK },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_UK },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_UK },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_UK },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_UK },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_UK },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_UK },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_UK },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_UK },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_UK },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_UK },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_UK },
         { "WHITE 28",          OPTION_VAL_WHITE_UK },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_UK },
         { "20%",  OPTION_VAL_20_UK },
         { "30%",  OPTION_VAL_30_UK },
         { "40%",  OPTION_VAL_40_UK },
         { "50%",  OPTION_VAL_50_UK },
         { "60%",  OPTION_VAL_60_UK },
         { "70%",  OPTION_VAL_70_UK },
         { "80%",  OPTION_VAL_80_UK },
         { "90%",  OPTION_VAL_90_UK },
         { "100%", OPTION_VAL_100_UK },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_uk = {
   option_cats_uk,
   option_defs_uk
};

/* RETRO_LANGUAGE_VAL */

#define CATEGORY_SYSTEM_LABEL_VAL "Sistema"
#define CATEGORY_SYSTEM_INFO_0_VAL "Configura regió, llengua, BIOS i ajustos de maquinari."
#define CATEGORY_VIDEO_LABEL_VAL "Vídeo"
#define CATEGORY_VIDEO_INFO_0_VAL NULL
#define CATEGORY_PERFORMANCE_LABEL_VAL NULL
#define CATEGORY_PERFORMANCE_INFO_0_VAL NULL
#define CATEGORY_HACKS_LABEL_VAL NULL
#define CATEGORY_HACKS_INFO_0_VAL NULL
#define CATEGORY_INPUT_LABEL_VAL "Entrada"
#define CATEGORY_INPUT_INFO_0_VAL NULL
#define CATEGORY_VMU_LABEL_VAL NULL
#define CATEGORY_VMU_INFO_0_VAL NULL
#define CORE_OPTION_NAME_REGION_LABEL_VAL "Regió"
#define OPTION_VAL_JAPAN_VAL NULL
#define OPTION_VAL_USA_VAL NULL
#define OPTION_VAL_EUROPE_VAL NULL
#define OPTION_VAL_DEFAULT_VAL NULL
#define CORE_OPTION_NAME_LANGUAGE_LABEL_VAL "Llengua"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_VAL NULL
#define OPTION_VAL_JAPANESE_VAL NULL
#define OPTION_VAL_ENGLISH_VAL NULL
#define OPTION_VAL_GERMAN_VAL NULL
#define OPTION_VAL_FRENCH_VAL NULL
#define OPTION_VAL_SPANISH_VAL NULL
#define OPTION_VAL_ITALIAN_VAL NULL
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_VAL NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_VAL NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_VAL NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_VAL NULL
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_VAL NULL
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_VAL NULL
#define CORE_OPTION_NAME_FORCE_WINCE_LABEL_VAL NULL
#define CORE_OPTION_NAME_FORCE_WINCE_INFO_0_VAL NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_VAL NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_VAL NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_VAL NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_VAL NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_VAL NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_VAL NULL
#define CORE_OPTION_NAME_UPNP_LABEL_VAL NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_VAL NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_VAL NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_VAL NULL
#define OPTION_VAL_320X240_VAL "320×240"
#define OPTION_VAL_640X480_VAL "640×480"
#define OPTION_VAL_800X600_VAL "800×600"
#define OPTION_VAL_960X720_VAL "960×720"
#define OPTION_VAL_1024X768_VAL "1024×768"
#define OPTION_VAL_1280X960_VAL "1280×960"
#define OPTION_VAL_1440X1080_VAL "1440×1080"
#define OPTION_VAL_1600X1200_VAL "1600×1200"
#define OPTION_VAL_1920X1440_VAL "1920×1440"
#define OPTION_VAL_2560X1920_VAL "2560×1920"
#define OPTION_VAL_2880X2160_VAL "2880×2160"
#define OPTION_VAL_3200X2400_VAL "3200×2400"
#define OPTION_VAL_3840X2880_VAL "3840×2880"
#define OPTION_VAL_4480X3360_VAL "4480×3360"
#define OPTION_VAL_5120X3840_VAL "5120×3840"
#define OPTION_VAL_5760X4320_VAL "5760×4320"
#define OPTION_VAL_6400X4800_VAL "6400×4800"
#define OPTION_VAL_7040X5280_VAL "7040×5280"
#define OPTION_VAL_7680X5760_VAL "7680×5760"
#define OPTION_VAL_8320X6240_VAL "8320×6240"
#define OPTION_VAL_8960X6720_VAL "8960×6720"
#define OPTION_VAL_9600X7200_VAL "9600×7200"
#define OPTION_VAL_10240X7680_VAL "10240×7680"
#define OPTION_VAL_10880X8160_VAL "10880×8160"
#define OPTION_VAL_11520X8640_VAL "11520×8640"
#define OPTION_VAL_12160X9120_VAL "12160×9120"
#define OPTION_VAL_12800X9600_VAL "12800×9600"
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_VAL NULL
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_VAL NULL
#define OPTION_VAL_VGA_VAL NULL
#define OPTION_VAL_TV_RGB_VAL NULL
#define OPTION_VAL_TV_COMPOSITE_VAL NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_VAL NULL
#define OPTION_VAL_NTSC_VAL NULL
#define OPTION_VAL_PAL_VAL NULL
#define OPTION_VAL_PAL_N_VAL NULL
#define OPTION_VAL_PAL_M_VAL NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_VAL "Orientació de pantalla"
#define OPTION_VAL_HORIZONTAL_VAL NULL
#define OPTION_VAL_VERTICAL_VAL NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_VAL NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_VAL NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_VAL NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_VAL NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_VAL NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_VAL NULL
#define OPTION_VAL_512MB_VAL NULL
#define OPTION_VAL_1GB_VAL NULL
#define OPTION_VAL_2GB_VAL NULL
#define OPTION_VAL_4GB_VAL NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_VAL NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_VAL NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_VAL NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_VAL NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_VAL NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_VAL NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_VAL NULL
#define CORE_OPTION_NAME_FOG_LABEL_VAL NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_VAL NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_VAL NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_VAL NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_VAL NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_VAL NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_VAL NULL
#define OPTION_VAL_1_VAL NULL
#define OPTION_VAL_2_VAL NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_VAL NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_VAL NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_VAL NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_VAL NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_VAL NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_VAL NULL
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_VAL NULL
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_VAL NULL
#define OPTION_VAL_2_O27_VAL NULL
#define OPTION_VAL_4_VAL NULL
#define OPTION_VAL_6_VAL NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_VAL NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_VAL NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_VAL NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_VAL NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_VAL NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_VAL NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_VAL NULL
#define OPTION_VAL_SOME_VAL NULL
#define OPTION_VAL_MORE_VAL NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_VAL NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_VAL NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_VAL NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_VAL NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_VAL NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_VAL NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_VAL NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_VAL NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_VAL NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_VAL NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_VAL NULL
#define OPTION_VAL_0_VAL NULL
#define OPTION_VAL_5_VAL NULL
#define OPTION_VAL_10_VAL NULL
#define OPTION_VAL_15_VAL NULL
#define OPTION_VAL_20_VAL NULL
#define OPTION_VAL_25_VAL NULL
#define OPTION_VAL_30_VAL NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_VAL NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_VAL NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_VAL NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_VAL NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_VAL NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_VAL NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_VAL NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_VAL NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_VAL NULL
#define OPTION_VAL_WHITE_VAL NULL
#define OPTION_VAL_RED_VAL NULL
#define OPTION_VAL_GREEN_VAL NULL
#define OPTION_VAL_BLUE_VAL NULL
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_VAL NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_VAL NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_VAL NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_VAL NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_VAL NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_VAL NULL
#define OPTION_VAL_VMU_A1_VAL NULL
#define OPTION_VAL_ALL_VMUS_VAL NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_VAL NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_VAL NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_VAL NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_VAL NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_VAL NULL
#define OPTION_VAL_UPPER_LEFT_VAL NULL
#define OPTION_VAL_UPPER_RIGHT_VAL NULL
#define OPTION_VAL_LOWER_LEFT_VAL NULL
#define OPTION_VAL_LOWER_RIGHT_VAL NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_VAL NULL
#define OPTION_VAL_1X_VAL NULL
#define OPTION_VAL_3X_VAL NULL
#define OPTION_VAL_5X_VAL NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_VAL NULL
#define OPTION_VAL_DEFAULT_ON_00_VAL NULL
#define OPTION_VAL_DEFAULT_OFF_01_VAL NULL
#define OPTION_VAL_BLACK_02_VAL NULL
#define OPTION_VAL_LIGHT_BLUE_04_VAL NULL
#define OPTION_VAL_CYAN_06_VAL NULL
#define OPTION_VAL_CYAN_BLUE_07_VAL NULL
#define OPTION_VAL_LIGHT_GREEN_08_VAL NULL
#define OPTION_VAL_CYAN_GREEN_09_VAL NULL
#define OPTION_VAL_LIGHT_CYAN_10_VAL NULL
#define OPTION_VAL_PURPLE_12_VAL NULL
#define OPTION_VAL_LIGHT_PURPLE_13_VAL NULL
#define OPTION_VAL_YELLOW_14_VAL NULL
#define OPTION_VAL_GRAY_15_VAL NULL
#define OPTION_VAL_LIGHT_PURPLE_2_16_VAL NULL
#define OPTION_VAL_LIGHT_GREEN_2_17_VAL NULL
#define OPTION_VAL_LIGHT_GREEN_3_18_VAL NULL
#define OPTION_VAL_LIGHT_CYAN_2_19_VAL NULL
#define OPTION_VAL_LIGHT_RED_2_20_VAL NULL
#define OPTION_VAL_MAGENTA_21_VAL NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_VAL NULL
#define OPTION_VAL_LIGHT_ORANGE_23_VAL NULL
#define OPTION_VAL_ORANGE_24_VAL NULL
#define OPTION_VAL_LIGHT_PURPLE_4_25_VAL NULL
#define OPTION_VAL_LIGHT_YELLOW_26_VAL NULL
#define OPTION_VAL_LIGHT_YELLOW_2_27_VAL NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_VAL NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_VAL NULL
#define OPTION_VAL_40_VAL NULL
#define OPTION_VAL_50_VAL NULL
#define OPTION_VAL_60_VAL NULL
#define OPTION_VAL_70_VAL NULL
#define OPTION_VAL_80_VAL NULL
#define OPTION_VAL_90_VAL NULL
#define OPTION_VAL_100_VAL NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_VAL NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_VAL NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_VAL NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_VAL NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_VAL NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_VAL NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_VAL NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_VAL NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_VAL NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_VAL NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_VAL NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_VAL NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_VAL NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_VAL NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_VAL NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_VAL NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_VAL NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_VAL NULL

struct retro_core_option_v2_category option_cats_val[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_VAL,
      CATEGORY_SYSTEM_INFO_0_VAL
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_VAL,
      CATEGORY_VIDEO_INFO_0_VAL
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_VAL,
      CATEGORY_PERFORMANCE_INFO_0_VAL
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_VAL,
      CATEGORY_HACKS_INFO_0_VAL
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_VAL,
      CATEGORY_INPUT_INFO_0_VAL
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_VAL,
      CATEGORY_VMU_INFO_0_VAL
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_val[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_VAL,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_VAL },
         { "USA",     OPTION_VAL_USA_VAL },
         { "Europe",  OPTION_VAL_EUROPE_VAL },
         { "Default", OPTION_VAL_DEFAULT_VAL },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_VAL,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_VAL },
         { "English",  OPTION_VAL_ENGLISH_VAL },
         { "German",   OPTION_VAL_GERMAN_VAL },
         { "French",   OPTION_VAL_FRENCH_VAL },
         { "Spanish",  OPTION_VAL_SPANISH_VAL },
         { "Italian",  OPTION_VAL_ITALIAN_VAL },
         { "Default",  OPTION_VAL_DEFAULT_VAL },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_VAL,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_boot_to_bios",
      CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_VAL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_VAL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_force_wince",
      CORE_OPTION_NAME_FORCE_WINCE_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_FORCE_WINCE_INFO_0_VAL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_VAL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_VAL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_VAL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_VAL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_VAL,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_VAL },
         { "640x480",    OPTION_VAL_640X480_VAL },
         { "800x600",    OPTION_VAL_800X600_VAL },
         { "960x720",    OPTION_VAL_960X720_VAL },
         { "1024x768",   OPTION_VAL_1024X768_VAL },
         { "1280x960",   OPTION_VAL_1280X960_VAL },
         { "1440x1080",  OPTION_VAL_1440X1080_VAL },
         { "1600x1200",  OPTION_VAL_1600X1200_VAL },
         { "1920x1440",  OPTION_VAL_1920X1440_VAL },
         { "2560x1920",  OPTION_VAL_2560X1920_VAL },
         { "2880x2160",  OPTION_VAL_2880X2160_VAL },
         { "3200x2400",  OPTION_VAL_3200X2400_VAL },
         { "3840x2880",  OPTION_VAL_3840X2880_VAL },
         { "4480x3360",  OPTION_VAL_4480X3360_VAL },
         { "5120x3840",  OPTION_VAL_5120X3840_VAL },
         { "5760x4320",  OPTION_VAL_5760X4320_VAL },
         { "6400x4800",  OPTION_VAL_6400X4800_VAL },
         { "7040x5280",  OPTION_VAL_7040X5280_VAL },
         { "7680x5760",  OPTION_VAL_7680X5760_VAL },
         { "8320x6240",  OPTION_VAL_8320X6240_VAL },
         { "8960x6720",  OPTION_VAL_8960X6720_VAL },
         { "9600x7200",  OPTION_VAL_9600X7200_VAL },
         { "10240x7680", OPTION_VAL_10240X7680_VAL },
         { "10880x8160", OPTION_VAL_10880X8160_VAL },
         { "11520x8640", OPTION_VAL_11520X8640_VAL },
         { "12160x9120", OPTION_VAL_12160X9120_VAL },
         { "12800x9600", OPTION_VAL_12800X9600_VAL },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_VAL,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_VAL },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_VAL },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_VAL },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_VAL,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_VAL },
         { "PAL",     OPTION_VAL_PAL_VAL },
         { "PAL_N",   OPTION_VAL_PAL_N_VAL },
         { "PAL_M",   OPTION_VAL_PAL_M_VAL },
         { "Default", OPTION_VAL_DEFAULT_VAL },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_VAL,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_VAL },
         { "vertical",   OPTION_VAL_VERTICAL_VAL },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_VAL,
      NULL,
      "",
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_VAL },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_VAL },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_VAL },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_VAL,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_VAL },
         { "1GB",   OPTION_VAL_1GB_VAL },
         { "2GB",   OPTION_VAL_2GB_VAL },
         { "4GB",   OPTION_VAL_4GB_VAL },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_VAL,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_VAL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_VAL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_VAL,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_VAL,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_VAL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_VAL,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_VAL,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_VAL },
         { "1",  OPTION_VAL_1_VAL },
         { "2",  OPTION_VAL_2_VAL },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_VAL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_VAL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_VAL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_VAL,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O27_VAL },
         { "4",  OPTION_VAL_4_VAL },
         { "6",  OPTION_VAL_6_VAL },
         { NULL, NULL },
      },
      "1",
   },
   {/* TODO: needs clarification */
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_VAL,
      NULL,
      "",
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_VAL,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_VAL,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_VAL,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_VAL,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_VAL },
         { "more", OPTION_VAL_MORE_VAL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_VAL,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_VAL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_VAL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_VAL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_VAL,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: probably needs explanation */
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_VAL,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_VAL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_VAL },
         { "5%",  OPTION_VAL_5_VAL },
         { "10%", OPTION_VAL_10_VAL },
         { "15%", OPTION_VAL_15_VAL },
         { "20%", OPTION_VAL_20_VAL },
         { "25%", OPTION_VAL_25_VAL },
         { "30%", OPTION_VAL_30_VAL },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_VAL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_VAL },
         { "5%",  OPTION_VAL_5_VAL },
         { "10%", OPTION_VAL_10_VAL },
         { "15%", OPTION_VAL_15_VAL },
         { "20%", OPTION_VAL_20_VAL },
         { "25%", OPTION_VAL_25_VAL },
         { "30%", OPTION_VAL_30_VAL },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_VAL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_purupuru",
      CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_VAL,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_VAL,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_VAL,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_VAL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_VAL },
         { "Red",      OPTION_VAL_RED_VAL },
         { "Green",    OPTION_VAL_GREEN_VAL },
         { "Blue",     OPTION_VAL_BLUE_VAL },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_VAL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_VAL },
         { "Red",      OPTION_VAL_RED_VAL },
         { "Green",    OPTION_VAL_GREEN_VAL },
         { "Blue",     OPTION_VAL_BLUE_VAL },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_VAL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_VAL },
         { "Red",      OPTION_VAL_RED_VAL },
         { "Green",    OPTION_VAL_GREEN_VAL },
         { "Blue",     OPTION_VAL_BLUE_VAL },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_VAL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_VAL },
         { "Red",      OPTION_VAL_RED_VAL },
         { "Green",    OPTION_VAL_GREEN_VAL },
         { "Blue",     OPTION_VAL_BLUE_VAL },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_VAL,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_VAL,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_VAL,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_VAL },
         { "All VMUs", OPTION_VAL_ALL_VMUS_VAL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_VAL,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_VAL,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_VAL,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_VAL },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_VAL },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_VAL },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_VAL },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_VAL },
         { "2x", OPTION_VAL_2_O27_VAL },
         { "3x", OPTION_VAL_3X_VAL },
         { "4x", OPTION_VAL_4_VAL },
         { "5x", OPTION_VAL_5X_VAL },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_VAL },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_VAL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_VAL },
         { "BLUE 03",           OPTION_VAL_BLUE_VAL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_VAL },
         { "GREEN 05",          OPTION_VAL_GREEN_VAL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_VAL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_VAL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_VAL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_VAL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_VAL },
         { "RED 11",            OPTION_VAL_RED_VAL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_VAL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_VAL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_VAL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_VAL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_VAL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_VAL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_VAL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_VAL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_VAL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_VAL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_VAL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_VAL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_VAL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_VAL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_VAL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_VAL },
         { "WHITE 28",          OPTION_VAL_WHITE_VAL },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_VAL },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_VAL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_VAL },
         { "BLUE 03",           OPTION_VAL_BLUE_VAL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_VAL },
         { "GREEN 05",          OPTION_VAL_GREEN_VAL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_VAL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_VAL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_VAL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_VAL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_VAL },
         { "RED 11",            OPTION_VAL_RED_VAL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_VAL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_VAL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_VAL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_VAL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_VAL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_VAL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_VAL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_VAL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_VAL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_VAL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_VAL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_VAL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_VAL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_VAL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_VAL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_VAL },
         { "WHITE 28",          OPTION_VAL_WHITE_VAL },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_VAL },
         { "20%",  OPTION_VAL_20_VAL },
         { "30%",  OPTION_VAL_30_VAL },
         { "40%",  OPTION_VAL_40_VAL },
         { "50%",  OPTION_VAL_50_VAL },
         { "60%",  OPTION_VAL_60_VAL },
         { "70%",  OPTION_VAL_70_VAL },
         { "80%",  OPTION_VAL_80_VAL },
         { "90%",  OPTION_VAL_90_VAL },
         { "100%", OPTION_VAL_100_VAL },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_VAL },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_VAL },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_VAL },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_VAL },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_VAL },
         { "2x", OPTION_VAL_2_O27_VAL },
         { "3x", OPTION_VAL_3X_VAL },
         { "4x", OPTION_VAL_4_VAL },
         { "5x", OPTION_VAL_5X_VAL },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_VAL },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_VAL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_VAL },
         { "BLUE 03",           OPTION_VAL_BLUE_VAL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_VAL },
         { "GREEN 05",          OPTION_VAL_GREEN_VAL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_VAL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_VAL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_VAL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_VAL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_VAL },
         { "RED 11",            OPTION_VAL_RED_VAL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_VAL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_VAL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_VAL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_VAL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_VAL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_VAL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_VAL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_VAL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_VAL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_VAL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_VAL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_VAL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_VAL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_VAL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_VAL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_VAL },
         { "WHITE 28",          OPTION_VAL_WHITE_VAL },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_VAL },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_VAL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_VAL },
         { "BLUE 03",           OPTION_VAL_BLUE_VAL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_VAL },
         { "GREEN 05",          OPTION_VAL_GREEN_VAL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_VAL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_VAL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_VAL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_VAL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_VAL },
         { "RED 11",            OPTION_VAL_RED_VAL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_VAL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_VAL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_VAL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_VAL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_VAL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_VAL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_VAL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_VAL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_VAL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_VAL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_VAL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_VAL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_VAL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_VAL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_VAL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_VAL },
         { "WHITE 28",          OPTION_VAL_WHITE_VAL },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_VAL },
         { "20%",  OPTION_VAL_20_VAL },
         { "30%",  OPTION_VAL_30_VAL },
         { "40%",  OPTION_VAL_40_VAL },
         { "50%",  OPTION_VAL_50_VAL },
         { "60%",  OPTION_VAL_60_VAL },
         { "70%",  OPTION_VAL_70_VAL },
         { "80%",  OPTION_VAL_80_VAL },
         { "90%",  OPTION_VAL_90_VAL },
         { "100%", OPTION_VAL_100_VAL },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_VAL },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_VAL },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_VAL },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_VAL },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_VAL },
         { "2x", OPTION_VAL_2_O27_VAL },
         { "3x", OPTION_VAL_3X_VAL },
         { "4x", OPTION_VAL_4_VAL },
         { "5x", OPTION_VAL_5X_VAL },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_VAL },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_VAL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_VAL },
         { "BLUE 03",           OPTION_VAL_BLUE_VAL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_VAL },
         { "GREEN 05",          OPTION_VAL_GREEN_VAL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_VAL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_VAL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_VAL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_VAL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_VAL },
         { "RED 11",            OPTION_VAL_RED_VAL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_VAL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_VAL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_VAL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_VAL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_VAL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_VAL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_VAL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_VAL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_VAL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_VAL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_VAL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_VAL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_VAL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_VAL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_VAL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_VAL },
         { "WHITE 28",          OPTION_VAL_WHITE_VAL },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_VAL },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_VAL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_VAL },
         { "BLUE 03",           OPTION_VAL_BLUE_VAL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_VAL },
         { "GREEN 05",          OPTION_VAL_GREEN_VAL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_VAL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_VAL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_VAL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_VAL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_VAL },
         { "RED 11",            OPTION_VAL_RED_VAL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_VAL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_VAL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_VAL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_VAL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_VAL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_VAL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_VAL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_VAL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_VAL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_VAL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_VAL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_VAL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_VAL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_VAL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_VAL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_VAL },
         { "WHITE 28",          OPTION_VAL_WHITE_VAL },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_VAL },
         { "20%",  OPTION_VAL_20_VAL },
         { "30%",  OPTION_VAL_30_VAL },
         { "40%",  OPTION_VAL_40_VAL },
         { "50%",  OPTION_VAL_50_VAL },
         { "60%",  OPTION_VAL_60_VAL },
         { "70%",  OPTION_VAL_70_VAL },
         { "80%",  OPTION_VAL_80_VAL },
         { "90%",  OPTION_VAL_90_VAL },
         { "100%", OPTION_VAL_100_VAL },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_VAL },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_VAL },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_VAL },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_VAL },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_VAL },
         { "2x", OPTION_VAL_2_O27_VAL },
         { "3x", OPTION_VAL_3X_VAL },
         { "4x", OPTION_VAL_4_VAL },
         { "5x", OPTION_VAL_5X_VAL },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_VAL },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_VAL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_VAL },
         { "BLUE 03",           OPTION_VAL_BLUE_VAL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_VAL },
         { "GREEN 05",          OPTION_VAL_GREEN_VAL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_VAL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_VAL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_VAL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_VAL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_VAL },
         { "RED 11",            OPTION_VAL_RED_VAL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_VAL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_VAL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_VAL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_VAL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_VAL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_VAL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_VAL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_VAL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_VAL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_VAL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_VAL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_VAL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_VAL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_VAL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_VAL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_VAL },
         { "WHITE 28",          OPTION_VAL_WHITE_VAL },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_VAL },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_VAL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_VAL },
         { "BLUE 03",           OPTION_VAL_BLUE_VAL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_VAL },
         { "GREEN 05",          OPTION_VAL_GREEN_VAL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_VAL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_VAL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_VAL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_VAL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_VAL },
         { "RED 11",            OPTION_VAL_RED_VAL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_VAL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_VAL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_VAL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_VAL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_VAL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_VAL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_VAL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_VAL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_VAL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_VAL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_VAL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_VAL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_VAL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_VAL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_VAL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_VAL },
         { "WHITE 28",          OPTION_VAL_WHITE_VAL },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_VAL },
         { "20%",  OPTION_VAL_20_VAL },
         { "30%",  OPTION_VAL_30_VAL },
         { "40%",  OPTION_VAL_40_VAL },
         { "50%",  OPTION_VAL_50_VAL },
         { "60%",  OPTION_VAL_60_VAL },
         { "70%",  OPTION_VAL_70_VAL },
         { "80%",  OPTION_VAL_80_VAL },
         { "90%",  OPTION_VAL_90_VAL },
         { "100%", OPTION_VAL_100_VAL },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_val = {
   option_cats_val,
   option_defs_val
};

/* RETRO_LANGUAGE_VN */

#define CATEGORY_SYSTEM_LABEL_VN "Hệ thống"
#define CATEGORY_SYSTEM_INFO_0_VN NULL
#define CATEGORY_VIDEO_LABEL_VN "Hình ảnh"
#define CATEGORY_VIDEO_INFO_0_VN NULL
#define CATEGORY_PERFORMANCE_LABEL_VN NULL
#define CATEGORY_PERFORMANCE_INFO_0_VN NULL
#define CATEGORY_HACKS_LABEL_VN "Hacks giả lập"
#define CATEGORY_HACKS_INFO_0_VN NULL
#define CATEGORY_INPUT_LABEL_VN "Input Driver"
#define CATEGORY_INPUT_INFO_0_VN NULL
#define CATEGORY_VMU_LABEL_VN NULL
#define CATEGORY_VMU_INFO_0_VN NULL
#define CORE_OPTION_NAME_REGION_LABEL_VN "Khu vực"
#define OPTION_VAL_JAPAN_VN NULL
#define OPTION_VAL_USA_VN NULL
#define OPTION_VAL_EUROPE_VN NULL
#define OPTION_VAL_DEFAULT_VN NULL
#define CORE_OPTION_NAME_LANGUAGE_LABEL_VN "Ngôn ngữ"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_VN NULL
#define OPTION_VAL_JAPANESE_VN "Tiếng Nhật"
#define OPTION_VAL_ENGLISH_VN "Tiếng Anh"
#define OPTION_VAL_GERMAN_VN "Tiếng Đức"
#define OPTION_VAL_FRENCH_VN "Tiếng Pháp"
#define OPTION_VAL_SPANISH_VN "Tiếng Tây Ban Nha"
#define OPTION_VAL_ITALIAN_VN "Tiếng Ý"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_VN NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_VN NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_VN NULL
#define CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_VN NULL
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_VN NULL
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_VN NULL
#define CORE_OPTION_NAME_FORCE_WINCE_LABEL_VN NULL
#define CORE_OPTION_NAME_FORCE_WINCE_INFO_0_VN NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_VN NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_VN NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_VN NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_VN NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_VN NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_VN NULL
#define CORE_OPTION_NAME_UPNP_LABEL_VN NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_VN NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_VN NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_VN NULL
#define OPTION_VAL_320X240_VN NULL
#define OPTION_VAL_640X480_VN NULL
#define OPTION_VAL_800X600_VN NULL
#define OPTION_VAL_960X720_VN NULL
#define OPTION_VAL_1024X768_VN NULL
#define OPTION_VAL_1280X960_VN NULL
#define OPTION_VAL_1440X1080_VN NULL
#define OPTION_VAL_1600X1200_VN NULL
#define OPTION_VAL_1920X1440_VN NULL
#define OPTION_VAL_2560X1920_VN NULL
#define OPTION_VAL_2880X2160_VN NULL
#define OPTION_VAL_3200X2400_VN NULL
#define OPTION_VAL_3840X2880_VN NULL
#define OPTION_VAL_4480X3360_VN NULL
#define OPTION_VAL_5120X3840_VN NULL
#define OPTION_VAL_5760X4320_VN NULL
#define OPTION_VAL_6400X4800_VN NULL
#define OPTION_VAL_7040X5280_VN NULL
#define OPTION_VAL_7680X5760_VN NULL
#define OPTION_VAL_8320X6240_VN NULL
#define OPTION_VAL_8960X6720_VN NULL
#define OPTION_VAL_9600X7200_VN NULL
#define OPTION_VAL_10240X7680_VN NULL
#define OPTION_VAL_10880X8160_VN NULL
#define OPTION_VAL_11520X8640_VN NULL
#define OPTION_VAL_12160X9120_VN NULL
#define OPTION_VAL_12800X9600_VN NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_VN NULL
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_VN NULL
#define OPTION_VAL_VGA_VN NULL
#define OPTION_VAL_TV_RGB_VN NULL
#define OPTION_VAL_TV_COMPOSITE_VN NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_VN NULL
#define OPTION_VAL_NTSC_VN NULL
#define OPTION_VAL_PAL_VN NULL
#define OPTION_VAL_PAL_N_VN NULL
#define OPTION_VAL_PAL_M_VN NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_VN NULL
#define OPTION_VAL_HORIZONTAL_VN NULL
#define OPTION_VAL_VERTICAL_VN NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_VN NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_VN NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_VN NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_VN NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_VN NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_VN NULL
#define OPTION_VAL_512MB_VN NULL
#define OPTION_VAL_1GB_VN NULL
#define OPTION_VAL_2GB_VN NULL
#define OPTION_VAL_4GB_VN NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_VN NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_VN NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_VN NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_VN NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_VN NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_VN NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_VN NULL
#define CORE_OPTION_NAME_FOG_LABEL_VN NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_VN NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_VN NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_VN NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_VN NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_VN NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_VN NULL
#define OPTION_VAL_1_VN NULL
#define OPTION_VAL_2_VN NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_VN NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_VN NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_VN NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_VN NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_VN NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_VN NULL
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_VN NULL
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_VN NULL
#define OPTION_VAL_2_O27_VN NULL
#define OPTION_VAL_4_VN NULL
#define OPTION_VAL_6_VN NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_VN NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_VN NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_VN NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_VN NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_VN NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_VN NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_VN NULL
#define OPTION_VAL_SOME_VN NULL
#define OPTION_VAL_MORE_VN NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_VN NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_VN NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_VN NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_VN NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_VN NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_VN NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_VN NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_VN NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_VN NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_VN NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_VN NULL
#define OPTION_VAL_0_VN NULL
#define OPTION_VAL_5_VN NULL
#define OPTION_VAL_10_VN NULL
#define OPTION_VAL_15_VN NULL
#define OPTION_VAL_20_VN NULL
#define OPTION_VAL_25_VN NULL
#define OPTION_VAL_30_VN NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_VN NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_VN NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_VN NULL
#define CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_VN NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_VN NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_VN NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_VN NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_VN NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_VN NULL
#define OPTION_VAL_WHITE_VN NULL
#define OPTION_VAL_RED_VN NULL
#define OPTION_VAL_GREEN_VN NULL
#define OPTION_VAL_BLUE_VN NULL
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_VN NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_VN NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_VN NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_VN NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_VN NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_VN NULL
#define OPTION_VAL_VMU_A1_VN NULL
#define OPTION_VAL_ALL_VMUS_VN NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_VN NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_VN NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_VN NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_VN NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_VN NULL
#define OPTION_VAL_UPPER_LEFT_VN NULL
#define OPTION_VAL_UPPER_RIGHT_VN NULL
#define OPTION_VAL_LOWER_LEFT_VN NULL
#define OPTION_VAL_LOWER_RIGHT_VN NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_VN NULL
#define OPTION_VAL_1X_VN NULL
#define OPTION_VAL_3X_VN NULL
#define OPTION_VAL_5X_VN NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_VN NULL
#define OPTION_VAL_DEFAULT_ON_00_VN NULL
#define OPTION_VAL_DEFAULT_OFF_01_VN NULL
#define OPTION_VAL_BLACK_02_VN NULL
#define OPTION_VAL_LIGHT_BLUE_04_VN NULL
#define OPTION_VAL_CYAN_06_VN NULL
#define OPTION_VAL_CYAN_BLUE_07_VN NULL
#define OPTION_VAL_LIGHT_GREEN_08_VN NULL
#define OPTION_VAL_CYAN_GREEN_09_VN NULL
#define OPTION_VAL_LIGHT_CYAN_10_VN NULL
#define OPTION_VAL_PURPLE_12_VN NULL
#define OPTION_VAL_LIGHT_PURPLE_13_VN NULL
#define OPTION_VAL_YELLOW_14_VN NULL
#define OPTION_VAL_GRAY_15_VN NULL
#define OPTION_VAL_LIGHT_PURPLE_2_16_VN NULL
#define OPTION_VAL_LIGHT_GREEN_2_17_VN NULL
#define OPTION_VAL_LIGHT_GREEN_3_18_VN NULL
#define OPTION_VAL_LIGHT_CYAN_2_19_VN NULL
#define OPTION_VAL_LIGHT_RED_2_20_VN NULL
#define OPTION_VAL_MAGENTA_21_VN NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_VN NULL
#define OPTION_VAL_LIGHT_ORANGE_23_VN NULL
#define OPTION_VAL_ORANGE_24_VN NULL
#define OPTION_VAL_LIGHT_PURPLE_4_25_VN NULL
#define OPTION_VAL_LIGHT_YELLOW_26_VN NULL
#define OPTION_VAL_LIGHT_YELLOW_2_27_VN NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_VN NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_VN NULL
#define OPTION_VAL_40_VN NULL
#define OPTION_VAL_50_VN NULL
#define OPTION_VAL_60_VN NULL
#define OPTION_VAL_70_VN NULL
#define OPTION_VAL_80_VN NULL
#define OPTION_VAL_90_VN NULL
#define OPTION_VAL_100_VN NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_VN NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_VN NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_VN NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_VN NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_VN NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_VN NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_VN NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_VN NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_VN NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_VN NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_VN NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_VN NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_VN NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_VN NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_VN NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_VN NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_VN NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_VN NULL

struct retro_core_option_v2_category option_cats_vn[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_VN,
      CATEGORY_SYSTEM_INFO_0_VN
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_VN,
      CATEGORY_VIDEO_INFO_0_VN
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_VN,
      CATEGORY_PERFORMANCE_INFO_0_VN
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_VN,
      CATEGORY_HACKS_INFO_0_VN
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_VN,
      CATEGORY_INPUT_INFO_0_VN
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_VN,
      CATEGORY_VMU_INFO_0_VN
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_vn[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_VN,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_VN },
         { "USA",     OPTION_VAL_USA_VN },
         { "Europe",  OPTION_VAL_EUROPE_VN },
         { "Default", OPTION_VAL_DEFAULT_VN },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_VN,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_VN },
         { "English",  OPTION_VAL_ENGLISH_VN },
         { "German",   OPTION_VAL_GERMAN_VN },
         { "French",   OPTION_VAL_FRENCH_VN },
         { "Spanish",  OPTION_VAL_SPANISH_VN },
         { "Italian",  OPTION_VAL_ITALIAN_VN },
         { "Default",  OPTION_VAL_DEFAULT_VN },
         { NULL, NULL },
      },
      "Default",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_VN,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_boot_to_bios",
      CORE_OPTION_NAME_BOOT_TO_BIOS_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_BOOT_TO_BIOS_INFO_0_VN,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_VN,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_force_wince",
      CORE_OPTION_NAME_FORCE_WINCE_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_FORCE_WINCE_INFO_0_VN,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_VN,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_VN,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_VN,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_VN,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_VN,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_VN },
         { "640x480",    OPTION_VAL_640X480_VN },
         { "800x600",    OPTION_VAL_800X600_VN },
         { "960x720",    OPTION_VAL_960X720_VN },
         { "1024x768",   OPTION_VAL_1024X768_VN },
         { "1280x960",   OPTION_VAL_1280X960_VN },
         { "1440x1080",  OPTION_VAL_1440X1080_VN },
         { "1600x1200",  OPTION_VAL_1600X1200_VN },
         { "1920x1440",  OPTION_VAL_1920X1440_VN },
         { "2560x1920",  OPTION_VAL_2560X1920_VN },
         { "2880x2160",  OPTION_VAL_2880X2160_VN },
         { "3200x2400",  OPTION_VAL_3200X2400_VN },
         { "3840x2880",  OPTION_VAL_3840X2880_VN },
         { "4480x3360",  OPTION_VAL_4480X3360_VN },
         { "5120x3840",  OPTION_VAL_5120X3840_VN },
         { "5760x4320",  OPTION_VAL_5760X4320_VN },
         { "6400x4800",  OPTION_VAL_6400X4800_VN },
         { "7040x5280",  OPTION_VAL_7040X5280_VN },
         { "7680x5760",  OPTION_VAL_7680X5760_VN },
         { "8320x6240",  OPTION_VAL_8320X6240_VN },
         { "8960x6720",  OPTION_VAL_8960X6720_VN },
         { "9600x7200",  OPTION_VAL_9600X7200_VN },
         { "10240x7680", OPTION_VAL_10240X7680_VN },
         { "10880x8160", OPTION_VAL_10880X8160_VN },
         { "11520x8640", OPTION_VAL_11520X8640_VN },
         { "12160x9120", OPTION_VAL_12160X9120_VN },
         { "12800x9600", OPTION_VAL_12800X9600_VN },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_VN,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_VN },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_VN },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_VN },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_VN,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_VN },
         { "PAL",     OPTION_VAL_PAL_VN },
         { "PAL_N",   OPTION_VAL_PAL_N_VN },
         { "PAL_M",   OPTION_VAL_PAL_M_VN },
         { "Default", OPTION_VAL_DEFAULT_VN },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_VN,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_VN },
         { "vertical",   OPTION_VAL_VERTICAL_VN },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_VN,
      NULL,
      "",
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_VN },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_VN },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_VN },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_VN,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_VN },
         { "1GB",   OPTION_VAL_1GB_VN },
         { "2GB",   OPTION_VAL_2GB_VN },
         { "4GB",   OPTION_VAL_4GB_VN },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_VN,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_VN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_VN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_VN,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_VN,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_VN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_VN,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_VN,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_VN },
         { "1",  OPTION_VAL_1_VN },
         { "2",  OPTION_VAL_2_VN },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_VN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_VN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_VN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_VN,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O27_VN },
         { "4",  OPTION_VAL_4_VN },
         { "6",  OPTION_VAL_6_VN },
         { NULL, NULL },
      },
      "1",
   },
   {/* TODO: needs clarification */
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_VN,
      NULL,
      "",
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_VN,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_VN,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_VN,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_VN,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_VN },
         { "more", OPTION_VAL_MORE_VN },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_VN,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_VN,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_VN,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_VN,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_VN,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: probably needs explanation */
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_VN,
      NULL,
      "",
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_VN,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_VN },
         { "5%",  OPTION_VAL_5_VN },
         { "10%", OPTION_VAL_10_VN },
         { "15%", OPTION_VAL_15_VN },
         { "20%", OPTION_VAL_20_VN },
         { "25%", OPTION_VAL_25_VN },
         { "30%", OPTION_VAL_30_VN },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_VN,
      NULL,
      "",
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_VN },
         { "5%",  OPTION_VAL_5_VN },
         { "10%", OPTION_VAL_10_VN },
         { "15%", OPTION_VAL_15_VN },
         { "20%", OPTION_VAL_20_VN },
         { "25%", OPTION_VAL_25_VN },
         { "30%", OPTION_VAL_30_VN },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_VN,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_purupuru",
      CORE_OPTION_NAME_ENABLE_PURUPURU_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_ENABLE_PURUPURU_INFO_0_VN,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_VN,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_VN,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_VN,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_VN },
         { "Red",      OPTION_VAL_RED_VN },
         { "Green",    OPTION_VAL_GREEN_VN },
         { "Blue",     OPTION_VAL_BLUE_VN },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_VN,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_VN },
         { "Red",      OPTION_VAL_RED_VN },
         { "Green",    OPTION_VAL_GREEN_VN },
         { "Blue",     OPTION_VAL_BLUE_VN },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_VN,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_VN },
         { "Red",      OPTION_VAL_RED_VN },
         { "Green",    OPTION_VAL_GREEN_VN },
         { "Blue",     OPTION_VAL_BLUE_VN },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_VN,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_VN },
         { "Red",      OPTION_VAL_RED_VN },
         { "Green",    OPTION_VAL_GREEN_VN },
         { "Blue",     OPTION_VAL_BLUE_VN },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_VN,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_VN,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_VN,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_VN },
         { "All VMUs", OPTION_VAL_ALL_VMUS_VN },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_VN,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_VN,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_VN,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_VN },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_VN },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_VN },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_VN },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_VN },
         { "2x", OPTION_VAL_2_O27_VN },
         { "3x", OPTION_VAL_3X_VN },
         { "4x", OPTION_VAL_4_VN },
         { "5x", OPTION_VAL_5X_VN },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_VN },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_VN },
         { "BLACK 02",          OPTION_VAL_BLACK_02_VN },
         { "BLUE 03",           OPTION_VAL_BLUE_VN },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_VN },
         { "GREEN 05",          OPTION_VAL_GREEN_VN },
         { "CYAN 06",           OPTION_VAL_CYAN_06_VN },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_VN },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_VN },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_VN },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_VN },
         { "RED 11",            OPTION_VAL_RED_VN },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_VN },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_VN },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_VN },
         { "GRAY 15",           OPTION_VAL_GRAY_15_VN },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_VN },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_VN },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_VN },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_VN },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_VN },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_VN },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_VN },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_VN },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_VN },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_VN },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_VN },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_VN },
         { "WHITE 28",          OPTION_VAL_WHITE_VN },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_VN },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_VN },
         { "BLACK 02",          OPTION_VAL_BLACK_02_VN },
         { "BLUE 03",           OPTION_VAL_BLUE_VN },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_VN },
         { "GREEN 05",          OPTION_VAL_GREEN_VN },
         { "CYAN 06",           OPTION_VAL_CYAN_06_VN },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_VN },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_VN },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_VN },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_VN },
         { "RED 11",            OPTION_VAL_RED_VN },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_VN },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_VN },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_VN },
         { "GRAY 15",           OPTION_VAL_GRAY_15_VN },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_VN },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_VN },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_VN },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_VN },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_VN },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_VN },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_VN },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_VN },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_VN },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_VN },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_VN },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_VN },
         { "WHITE 28",          OPTION_VAL_WHITE_VN },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_VN },
         { "20%",  OPTION_VAL_20_VN },
         { "30%",  OPTION_VAL_30_VN },
         { "40%",  OPTION_VAL_40_VN },
         { "50%",  OPTION_VAL_50_VN },
         { "60%",  OPTION_VAL_60_VN },
         { "70%",  OPTION_VAL_70_VN },
         { "80%",  OPTION_VAL_80_VN },
         { "90%",  OPTION_VAL_90_VN },
         { "100%", OPTION_VAL_100_VN },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_VN },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_VN },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_VN },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_VN },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_VN },
         { "2x", OPTION_VAL_2_O27_VN },
         { "3x", OPTION_VAL_3X_VN },
         { "4x", OPTION_VAL_4_VN },
         { "5x", OPTION_VAL_5X_VN },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_VN },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_VN },
         { "BLACK 02",          OPTION_VAL_BLACK_02_VN },
         { "BLUE 03",           OPTION_VAL_BLUE_VN },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_VN },
         { "GREEN 05",          OPTION_VAL_GREEN_VN },
         { "CYAN 06",           OPTION_VAL_CYAN_06_VN },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_VN },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_VN },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_VN },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_VN },
         { "RED 11",            OPTION_VAL_RED_VN },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_VN },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_VN },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_VN },
         { "GRAY 15",           OPTION_VAL_GRAY_15_VN },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_VN },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_VN },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_VN },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_VN },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_VN },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_VN },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_VN },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_VN },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_VN },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_VN },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_VN },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_VN },
         { "WHITE 28",          OPTION_VAL_WHITE_VN },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_VN },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_VN },
         { "BLACK 02",          OPTION_VAL_BLACK_02_VN },
         { "BLUE 03",           OPTION_VAL_BLUE_VN },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_VN },
         { "GREEN 05",          OPTION_VAL_GREEN_VN },
         { "CYAN 06",           OPTION_VAL_CYAN_06_VN },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_VN },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_VN },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_VN },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_VN },
         { "RED 11",            OPTION_VAL_RED_VN },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_VN },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_VN },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_VN },
         { "GRAY 15",           OPTION_VAL_GRAY_15_VN },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_VN },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_VN },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_VN },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_VN },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_VN },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_VN },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_VN },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_VN },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_VN },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_VN },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_VN },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_VN },
         { "WHITE 28",          OPTION_VAL_WHITE_VN },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_VN },
         { "20%",  OPTION_VAL_20_VN },
         { "30%",  OPTION_VAL_30_VN },
         { "40%",  OPTION_VAL_40_VN },
         { "50%",  OPTION_VAL_50_VN },
         { "60%",  OPTION_VAL_60_VN },
         { "70%",  OPTION_VAL_70_VN },
         { "80%",  OPTION_VAL_80_VN },
         { "90%",  OPTION_VAL_90_VN },
         { "100%", OPTION_VAL_100_VN },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_VN },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_VN },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_VN },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_VN },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_VN },
         { "2x", OPTION_VAL_2_O27_VN },
         { "3x", OPTION_VAL_3X_VN },
         { "4x", OPTION_VAL_4_VN },
         { "5x", OPTION_VAL_5X_VN },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_VN },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_VN },
         { "BLACK 02",          OPTION_VAL_BLACK_02_VN },
         { "BLUE 03",           OPTION_VAL_BLUE_VN },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_VN },
         { "GREEN 05",          OPTION_VAL_GREEN_VN },
         { "CYAN 06",           OPTION_VAL_CYAN_06_VN },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_VN },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_VN },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_VN },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_VN },
         { "RED 11",            OPTION_VAL_RED_VN },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_VN },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_VN },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_VN },
         { "GRAY 15",           OPTION_VAL_GRAY_15_VN },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_VN },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_VN },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_VN },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_VN },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_VN },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_VN },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_VN },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_VN },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_VN },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_VN },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_VN },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_VN },
         { "WHITE 28",          OPTION_VAL_WHITE_VN },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_VN },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_VN },
         { "BLACK 02",          OPTION_VAL_BLACK_02_VN },
         { "BLUE 03",           OPTION_VAL_BLUE_VN },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_VN },
         { "GREEN 05",          OPTION_VAL_GREEN_VN },
         { "CYAN 06",           OPTION_VAL_CYAN_06_VN },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_VN },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_VN },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_VN },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_VN },
         { "RED 11",            OPTION_VAL_RED_VN },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_VN },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_VN },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_VN },
         { "GRAY 15",           OPTION_VAL_GRAY_15_VN },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_VN },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_VN },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_VN },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_VN },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_VN },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_VN },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_VN },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_VN },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_VN },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_VN },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_VN },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_VN },
         { "WHITE 28",          OPTION_VAL_WHITE_VN },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_VN },
         { "20%",  OPTION_VAL_20_VN },
         { "30%",  OPTION_VAL_30_VN },
         { "40%",  OPTION_VAL_40_VN },
         { "50%",  OPTION_VAL_50_VN },
         { "60%",  OPTION_VAL_60_VN },
         { "70%",  OPTION_VAL_70_VN },
         { "80%",  OPTION_VAL_80_VN },
         { "90%",  OPTION_VAL_90_VN },
         { "100%", OPTION_VAL_100_VN },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_VN },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_VN },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_VN },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_VN },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_VN },
         { "2x", OPTION_VAL_2_O27_VN },
         { "3x", OPTION_VAL_3X_VN },
         { "4x", OPTION_VAL_4_VN },
         { "5x", OPTION_VAL_5X_VN },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_VN },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_VN },
         { "BLACK 02",          OPTION_VAL_BLACK_02_VN },
         { "BLUE 03",           OPTION_VAL_BLUE_VN },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_VN },
         { "GREEN 05",          OPTION_VAL_GREEN_VN },
         { "CYAN 06",           OPTION_VAL_CYAN_06_VN },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_VN },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_VN },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_VN },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_VN },
         { "RED 11",            OPTION_VAL_RED_VN },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_VN },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_VN },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_VN },
         { "GRAY 15",           OPTION_VAL_GRAY_15_VN },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_VN },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_VN },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_VN },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_VN },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_VN },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_VN },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_VN },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_VN },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_VN },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_VN },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_VN },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_VN },
         { "WHITE 28",          OPTION_VAL_WHITE_VN },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_VN },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_VN },
         { "BLACK 02",          OPTION_VAL_BLACK_02_VN },
         { "BLUE 03",           OPTION_VAL_BLUE_VN },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_VN },
         { "GREEN 05",          OPTION_VAL_GREEN_VN },
         { "CYAN 06",           OPTION_VAL_CYAN_06_VN },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_VN },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_VN },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_VN },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_VN },
         { "RED 11",            OPTION_VAL_RED_VN },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_VN },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_VN },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_VN },
         { "GRAY 15",           OPTION_VAL_GRAY_15_VN },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_VN },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_VN },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_VN },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_VN },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_VN },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_VN },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_VN },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_VN },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_VN },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_VN },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_VN },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_VN },
         { "WHITE 28",          OPTION_VAL_WHITE_VN },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_VN },
         { "20%",  OPTION_VAL_20_VN },
         { "30%",  OPTION_VAL_30_VN },
         { "40%",  OPTION_VAL_40_VN },
         { "50%",  OPTION_VAL_50_VN },
         { "60%",  OPTION_VAL_60_VN },
         { "70%",  OPTION_VAL_70_VN },
         { "80%",  OPTION_VAL_80_VN },
         { "90%",  OPTION_VAL_90_VN },
         { "100%", OPTION_VAL_100_VN },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_vn = {
   option_cats_vn,
   option_defs_vn
};


#ifdef __cplusplus
}
#endif

#endif